/*
 *	FileUtils.c
 *
 *	General file utilities.
 *	Include routines to open a file and read it.
 *
 *	-TA 6/27/90
 */

#include "FileUtils.h"

		/* for file reading routines */
	static	long		EndIndex, StartIndex;
	static	int			OpenRefNum, SaveRefNum;
	static	char		*OpenFName, *SaveFName;
	static	int			SaveVRef;
	static	FileBuffer	Buffer;
	static	OSErr		FileError;

	static	int			DefaultVol;
	static	int			RF_OpenRefNum;
	static	char		*RF_Name;


int	GetOpenRefNum()
{
	return(OpenRefNum);
}

int	GetSaveRefNum()
{
	return(SaveRefNum);
}

	/*	Volume Routines	*/

Boolean	GoodVolumeOp(int myErrNo, char *myActionString)
/*	Requires:
 *		1)	Generic volume alert, ID 301 (VolumeAlertID in FileUtils.h)
 *		2)	Series of FERR resources, strings corresponding to
 *			file error codes
 */
{
	Handle	myAlertHandle, myFErrHandle;
	char	myNumString[32];

	if (myErrNo == noErr) return(TRUE);

	myAlertHandle = GetResource(AlertOSType, VolumeAlertID);
	if (!myAlertHandle) return(FALSE);
	myFErrHandle = GetResource(FErrOSType, myErrNo);
	if (!myFErrHandle)
		myFErrHandle = (Handle) NewString("\PUnknown Error");
	NumToString((long) myErrNo, &myNumString);
	HLock(myFErrHandle);
	ParamText(myActionString,"\P",(char *) myNumString,*myFErrHandle);
	if ( StopAlert(VolumeAlertID, 0L) ) ;
	HUnlock(myFErrHandle);
	HPurge(myFErrHandle);
	return(FALSE);
}

Boolean StoreDefaultVolume()
/*	Should be matched by call to RestoreDefaultVolume	*/
{
	return( GoodVolumeOp(GetVol(NIL, &DefaultVol),
			"\Pgetting the default volume") );
}

Boolean RestoreDefaultVolume()
{
	return( GoodVolumeOp(SetVol(NIL, DefaultVol),
			"\Psetting the default volume") );
}

Boolean SetDefaultVolume(int vRef)
{
	return( GoodVolumeOp(SetVol(NIL, vRef), "\Psetting the default volume") );
}

Boolean	GoodFileOp(int myErrNo, char *myActionString, char *myFName)
/*	Requires:
 *		1)	Generic file alert, ID 300 (FileAlertID in FileUtils.h)
 *		2)	Series of FERR resources, strings corresponding to
 *			file error codes
 */
{
	Handle	myAlertHandle, myFErrHandle;
	char	myNumString[32];

	if (myErrNo == noErr) return(TRUE);

	myAlertHandle = GetResource(AlertOSType, FileAlertID);
	if (!myAlertHandle) return(FALSE);
	myFErrHandle = GetResource(FErrOSType, myErrNo);
	if (!myFErrHandle)
		myFErrHandle = (Handle) NewString("\PUnknown Error");
	NumToString((long) myErrNo, &myNumString);
	HLock(myFErrHandle);
	ParamText(myActionString,myFName,(char *) myNumString,*myFErrHandle);
	if ( StopAlert(FileAlertID, 0L) ) ;
	HUnlock(myFErrHandle);
	HPurge(myFErrHandle);
	return(FALSE);
}

int	SimpleGetParamFile(OSType GoodFType, int *pVRef, char *fname)
{
	int		appMessage, appCount;
	AppFile	theAppFile;

	CountAppFiles(&appMessage, &appCount);
	if (appCount == 0) return(NoAppFile);
	GetAppFiles(1, &theAppFile);
	if (theAppFile.fType == GoodFType) {
		sEquate(fname, (char *) &theAppFile.fName);
		*pVRef = theAppFile.vRefNum;
		ClrAppFiles(1);	/*	only clear if handled successfully	*/
		return(appMessage);
	}
	return(NoAppFile);
}

Boolean SetFileCreatorAndType(char *fname, int theVRef,
		OSType Creator, OSType Type)
{
	FInfo	fndrInfo;

	if ( !GoodFileOp(GetFInfo(fname, theVRef, &fndrInfo),
			"\Pgetting information about", fname) )
		return(FALSE);

	if ( (fndrInfo.fdCreator == Creator) && (fndrInfo.fdType == Type) )
		return(TRUE);

	fndrInfo.fdCreator = Creator;
	fndrInfo.fdType = Type;
	return( GoodFileOp(SetFInfo(fname, theVRef, &fndrInfo),
			"\Psetting type and creator of", fname) );
}

Boolean StdCreateFile(char *fname, int vref, OSType Creator, OSType Type)
{
	OSErr	myErr;

	myErr = Create(fname, vref, Creator, Type);
	if ( (myErr != noErr) && (myErr != dupFNErr) ) {
		if ( !GoodFileOp(myErr, "\Pcreating", fname) )
			return(FALSE);
	}
	if (myErr == dupFNErr)
		SetFileCreatorAndType(fname, vref, Creator, Type);
	return(TRUE);
}

Boolean	StartFileRead(int theVRef, char *fname, Boolean OKToFail)
{
	long	ReadAmt;
	OSErr	myErr;

	myErr = FSOpen(fname,theVRef,&OpenRefNum);
	if (myErr != noErr) {
		if (!OKToFail)
			if (!GoodFileOp(myErr, "\Popening",fname)) ;
		return(FALSE);
	}

	ReadAmt = BufSize;
	FileError = FSRead(OpenRefNum, &ReadAmt, &Buffer.whole);
	if ((FileError!=noErr) && (FileError!=eofErr)) {
		if ( !GoodFileOp(FileError, "\Preading", fname) ) ;
		if ( !GoodFileOp(FSClose(OpenRefNum), "\Pclosing", fname) ) ;
		return(FALSE);
	}
	EndIndex = ReadAmt - 1L;
	StartIndex = 0L;

	OpenFName = fname;

	return(TRUE);
}

ResetFileRead()
{
	StartIndex = 0L;
}

StopFileRead()
{
	if (FSClose(OpenRefNum) == noErr)
		;
}

Boolean CheckNextRead()
{
	long	ReadAmt;

	if (StartIndex >= EndIndex) {
		if (!GoodFileOp(FSClose(OpenRefNum), "\Pclosing", OpenFName)) ;
		return(FALSE);
	}

	if (StartIndex >= HalfBufSize) {
		BlockEquate((char *) Buffer.halves.half1, (char *) Buffer.halves.half2, HalfBufSize);
			/*	Buffer.halves.half1 = Buffer.halves.half2;	*/
		StartIndex -= HalfBufSize;
		EndIndex -= HalfBufSize;
		if (FileError != eofErr) {
			ReadAmt = HalfBufSize;
			FileError = FSRead(OpenRefNum, &ReadAmt, &Buffer.halves.half2);
			if ((FileError!=noErr) && (FileError!=eofErr)) {
				if ( !GoodFileOp(FileError, "\Preading", OpenFName) ) ;
				if ( !GoodFileOp(FSClose(OpenRefNum), "\Pclosing", OpenFName) ) ;
				return(FALSE);
			}
			EndIndex += ReadAmt;
/*	was
			EndIndex = HalfBufSize + ReadAmt - 1L;
*/
		}
	}
	return(TRUE);
}

Boolean	ReadTextFileLine(char *theLine)
{
	if (!CheckNextRead())
		return(FALSE);

	if (!GetLine(theLine,(char *) Buffer.whole,&StartIndex,EndIndex))
		;	/*	return(FALSE);	*/
	return(TRUE);
}

Boolean ReadPackedRow(char **pdstPtr, int rowBytes)
{
	char	*srcPtr, *oSrcPtr;

	if (!CheckNextRead())
		return(FALSE);

	oSrcPtr = srcPtr = (char *) (&Buffer.whole) + StartIndex;
	UnpackBits(&srcPtr, pdstPtr, rowBytes);
	StartIndex += (srcPtr - oSrcPtr);
}

Boolean ReadPtrFromFile(Ptr thePtr, long *pReadAmt)
{
	long	AmtRead, AmtToRead;
	Boolean	GoodOp = TRUE;

	AmtRead = 0L;
	while (AmtRead < *pReadAmt) {
		if (!CheckNextRead()) {
			*pReadAmt = AmtRead;
			return(FALSE);
		}
		AmtToRead = MIN(*pReadAmt - AmtRead, HalfBufSize - StartIndex);
		if (AmtToRead > EndIndex - StartIndex + 1) {
			AmtToRead = EndIndex - StartIndex + 1;
			GoodOp = FALSE;
		}
		BlockEquate(thePtr + AmtRead,
				(char *) &Buffer.whole[StartIndex], AmtToRead);
		StartIndex += AmtToRead;
		AmtRead += AmtToRead;
	}
	*pReadAmt = AmtRead;

	return(GoodOp);
}

Boolean ReadHandleFromFile(Handle theHandle, long *pReadAmt)
{
	Boolean	GoodOp;

	HLock(theHandle);
	GoodOp = ReadPtrFromFile(*theHandle, pReadAmt);
	HUnlock(theHandle);
	return(GoodOp);
}

Boolean ReadPICT(PicHandle *ppic)
{
	long	PicSize;

	PicSize = *( (int *) (&Buffer.whole[StartIndex]) );
	*ppic = (PicHandle) NewHandle( PicSize );
	if (*ppic == NIL)
		return(FALSE);

	HLock( (Handle) *ppic );
	if (!ReadPtrFromFile((char *) **ppic, &PicSize)) {
		HUnlock((Handle) *ppic);
		DisposHandle((Handle) *ppic);
		return(FALSE);
	}
	HUnlock((Handle) *ppic);

	return(TRUE);
}

Boolean StartFileSave(char *fname, int theVRef)
{
	if ( !GoodFileOp(FSOpen(fname, theVRef, &SaveRefNum),"\Popening",fname) )
		return(FALSE);

	if ( !GoodFileOp(SetFPos(SaveRefNum,fsFromStart, 0L),"\Pseeking in",fname) )
	{
		if ( !GoodFileOp(FSClose(SaveRefNum), "\Pclosing", fname) ) ;
		return(FALSE);
	}

	SaveFName = fname;
	SaveVRef = theVRef;

	return(TRUE);
}

Boolean StopFileSave()
{
	long	logEof;

	if ( !GoodFileOp(GetFPos(SaveRefNum,&logEof),"\Pfinding position in",SaveFName) )
	{
		if ( !GoodFileOp(FSClose(SaveRefNum), "\Pclosing", SaveFName) ) ;
		return(FALSE);
	}
	if ( !GoodFileOp(SetEOF(SaveRefNum,logEof),"\Psetting EOF in",SaveFName) )
	{
		if ( !GoodFileOp(FSClose(SaveRefNum), "\Pclosing", SaveFName) ) ;
		return(FALSE);
	}

	if (!GoodFileOp(FSClose(SaveRefNum), "\Pclosing", SaveFName))
		return(FALSE);

	return(TRUE);
}

Boolean DeleteSaveFile()
{
	return(GoodFileOp(FSDelete(SaveFName, SaveVRef), "\Pdeleting", SaveFName));
}

Boolean WritePtrToFile(Ptr thePtr, long WriteAmt)
/*	Writes TotalAmt bytes pointed to by thePtr, to SaveRefNum (fname)	*/
{
	long	amt;
	char	*SrcPtr;

	SrcPtr = thePtr;
	while(WriteAmt>0) {
		amt = (WriteAmt>WriteBufSize) ? WriteBufSize : WriteAmt;
		WriteAmt -= amt;
		if (!GoodFileOp(FSWrite(SaveRefNum,&amt,SrcPtr),"\Pwriting",SaveFName)) {
			if ( !GoodFileOp(FSClose(SaveRefNum), "\Pclosing", SaveFName) ) ;
			return(FALSE);
		}
		SrcPtr += amt;
	}

	return(TRUE);
}

Boolean WriteHandleToFile(Handle theHandle, long WriteAmt)
/*	Writes TotalAmt bytes pointed to by *theHandle, to SaveRefNum (fname)	*/
{
	HLock(theHandle);
	if ( !WritePtrToFile(*theHandle, WriteAmt) )
		return(FALSE);
	HUnlock(theHandle);

	return(TRUE);
}

Boolean WriteTEtoFile(TEHandle WriteTE, long WriteStart)
/*	Writes chars from text edit record **WriteTE, starting with position
 *	WriteStart, to SaveRefNum (fname)
 */
{
	Handle	theHandle = (**WriteTE).hText;

	HLock(theHandle);
	if ( !WritePtrToFile((*theHandle)+WriteStart,
			(long) ((**WriteTE).teLength-WriteStart)) )
		return(FALSE);
	HUnlock(theHandle);

	return(TRUE);
}

Boolean WritePICTToFile(PicHandle pic)
{
	HLock(pic);
	if ( !WritePtrToFile((Ptr) *pic, (long) (**pic).picSize) )
		return(FALSE);
	HUnlock(pic);

	return(TRUE);
}

Boolean Delim_WritePtrToFile(Ptr thePtr, long WriteAmt,
		char *DelimString, int MinSpaces, int *NumSpaces,
		Boolean SubstOptSpace, Boolean UseForcedSpace)
/*		Writes to file as WritePtrToFile, substituting DelimString wherever
 *	it finds MinSpaces or more spaces.
 *		Always deletes leading spaces on a line.
 *		*NumSpaces represents (for situations in which there are consecutive
 *	calls to this routine, all meant to be part of the same text) the 
 *	number of spaces having been reached before and at the end of the
 *	curent block of text.
 *		If SubstOptSpace is TRUE, will substitute regular spaces for
 *	option spaces; these spaces are treated as non-spaces for
 *	delimiter purposes.
 *		If UseForcedSpace is TRUE, will substitute a regular space for
 *	the character FORCED_SPACE (at first 0x01); these spaces are treated
 *	as non-spaces for delimiter purposes.
 */
{
	Ptr		NewText;
	long	l=0L, newl = 0L;
	int		i;
	Boolean	GoodWrite;

	if (DelimString[0] == 0)
		return( WritePtrToFile(thePtr, WriteAmt) );

	if ( (NewText = NewPtr(MAX_POINTER_SIZE)) == 0L )
		return(FALSE);

	for( ; l<WriteAmt; ++l) {
		if (thePtr[l] != SPACE) {
			if (*NumSpaces < MinSpaces)
				for (i=1; i<=*NumSpaces; ++i)
					NewText[newl++] = SPACE;
			*NumSpaces = 0;
			if ( SubstOptSpace && (thePtr[l] == OPTION_SPACE) )
				NewText[newl++] = SPACE;
			else if ( UseForcedSpace && (thePtr[l] == FORCED_SPACE) )
				NewText[newl++] = SPACE;
			else
				NewText[newl++] = thePtr[l];
		}
		switch (thePtr[l]) {
			case SPACE:
				if (++(*NumSpaces) == MinSpaces)
					for(i=1;i<=DelimString[0];++i)
						NewText[newl++] = DelimString[i];
				break;
			case RETURN_CHAR:
				do {
					if (++l >= WriteAmt) {
						*NumSpaces = MinSpaces+1;
						break;
					}
				} while (thePtr[l] == SPACE);
				--l;
		}
	}

	GoodWrite = WritePtrToFile(NewText, newl);
	DisposPtr(NewText);
	return(GoodWrite);
}

Boolean Delim_WriteHandleToFile(Handle theHandle, long WriteAmt,
		char *DelimString, int MinSpaces, int *NumSpaces,
		Boolean SubstOptSpace, Boolean UseForcedSpace)
/*	See WriteHandleToFile and Delim_WritePtrToFile	*/
{
	HLock(theHandle);
	if ( !Delim_WritePtrToFile(*theHandle, WriteAmt,
			DelimString, MinSpaces, NumSpaces, SubstOptSpace, UseForcedSpace) )
		return(FALSE);
	HUnlock(theHandle);

	return(TRUE);
}

Boolean Delim_WriteTEtoFile(TEHandle WriteTE, long WriteStart,
		char *DelimString, int MinSpaces, int *NumSpaces,
		Boolean SubstOptSpace, Boolean UseForcedSpace)
/*	See WriteTEToFile and Delim_WritePtrToFile	*/
{
	Handle	theHandle = (**WriteTE).hText;

	HLock(theHandle);
	if ( !Delim_WritePtrToFile((*theHandle)+WriteStart, (long)
			((**WriteTE).teLength-WriteStart),DelimString,MinSpaces,
			NumSpaces, SubstOptSpace, UseForcedSpace) )
		return(FALSE);
	HUnlock(theHandle);

	return(TRUE);
}

	/*	For handling bitmaps...	*/

Boolean OpenBitMap(BitMap *pbm)
{
	int		i;
	char	*dstPtr = (char *) pbm->baseAddr;

	for(i=0;i<pbm->bounds.bottom - pbm->bounds.top; ++i)
		if (!ReadPackedRow(&dstPtr, pbm->rowBytes))
			return(FALSE);

	return(TRUE);
}

Boolean SaveBitMap(BitMap *pbm)
{
	int		i;
	char	*srcPtr = (char *) pbm->baseAddr;
	char	dest[1024];
	char	*dstPtr;

	for(i=0;i<pbm->bounds.bottom - pbm->bounds.top; ++i) {
		dstPtr = (char *) &dest;
		PackBits(&srcPtr, &dstPtr, pbm->rowBytes);
		if (!WritePtrToFile((char *) &dest,
				(long) (dstPtr - (char *) &dest)))
			return(FALSE);
	}
	return(TRUE);
}


	/*	Resource Fork Routines	*/

Boolean OpenResFork(char *fname, int vRef, Boolean ForceResFork,
		Boolean *Exists)
{
	OSErr	theResError;

	if (!StoreDefaultVolume())
		return(FALSE);
	if (!SetDefaultVolume(vRef)) {
		if (!RestoreDefaultVolume())	;
		return(FALSE);
	}
	if (ForceResFork)
		CreateResFile(fname);
	RF_OpenRefNum = OpenResFile(fname);
	theResError = ResError();
	if ( (theResError == resFNotFound) || (theResError == fnfErr) ||
			(theResError == eofErr) )
		if (!ForceResFork) {
			*Exists = FALSE;
			return(RestoreDefaultVolume());
		}

	if ( !GoodFileOp(theResError, "\Popening resource fork of", fname) ) {
		if (!RestoreDefaultVolume())	;
		return(FALSE);
	}
	RF_Name = fname;
	*Exists = TRUE;

	return(RestoreDefaultVolume());
}

Boolean StopResFile()
{
	CloseResFile(RF_OpenRefNum);
	return( GoodFileOp(ResError(), "\Pclosing resource fork of", RF_Name) );
}

Boolean Delete1Resource(ResType theType, int theID, Boolean *pWasThere)
{
	Handle tmpH;

	if ( (tmpH = Get1Resource(theType, theID)) != NIL ) {
		*pWasThere = TRUE;
		RmveResource(tmpH);
		if ( !GoodFileOp(ResError(), "\Premoving resource from", RF_Name) )
			return(FALSE);
		DisposHandle(tmpH);
		UpdateResFile(RF_OpenRefNum);
		if ( !GoodFileOp(ResError(), "\Pupdating resource fork of", RF_Name) )
			return(FALSE);
	}
	else
		*pWasThere = FALSE;

	return(TRUE);
}

Boolean AddNewResource(Handle h, ResType theType, int theID)
{
/*
	Handle tmpH;

	if ( (tmpH = Get1Resource(theType, theID)) != NIL ) {
		RmveResource(tmpH);
		if ( !GoodFileOp(ResError(), "\Premoving resource from", RF_Name) )
			return(FALSE);
		DisposHandle(tmpH);
		UpdateResFile(RF_OpenRefNum);
		if ( !GoodFileOp(ResError(), "\Pupdating resource fork of", RF_Name) )
			return(FALSE);
	}
*/
/*	was out of if statement . . . not sure why	-TA 6/4/91
	UpdateResFile(RF_OpenRefNum);
	if ( !GoodFileOp(ResError(), "\Pupdating resource fork of", RF_Name) )
		return(FALSE);
*/
	Boolean	WasThere;

	if (!Delete1Resource(theType, theID, &WasThere))
		return(FALSE);
	AddResource(h, theType, theID, NIL);
	return(GoodFileOp(ResError(), "\Padding resource to", RF_Name));
}

Boolean UpdateResFork()
{
	UpdateResFile(RF_OpenRefNum);
	return(GoodFileOp(ResError(), "\Pupdating resource fork of", RF_Name));
}

Boolean StdPutPicture(PicHandle pic, int theID)
{
	return(AddNewResource((Handle) pic, PICT_OSType, theID));
}
