/*	GeneralUtils.c: General Utilities
 *	-TA 6/26/90
 */

#include "GeneralUtils.h"

long	DetachedTickDelayStart;

BlockEquate(char *b1, char *b2, long num)
{
	long l;

	for(l=0L;l<num;++l)
		b1[l] = b2[l];
}

long KeyCodeToOffset(int KeyCode)
/*	Given KeyCode, per Inside Mac, Toolbox Event Manager,
 *	converts to offset in KeyMap which can be used with BitTst
 *	to test if a given key is down.
 *	Does not work automatically presumably due to differences in
 *	high-order bit interpretation.
 *	Basic conversion is this:
 *		 0 ->  7
 *		 1 ->  6
 *		 2 ->  5
 *		  ....
 *		 7 ->  0
 *		 8 -> 15
 *		 9 -> 14
 *		  ....
 *		15 ->  8
 *		16 -> 23
 *		  ....
 */
{
	return( (long) (8*(KeyCode/8) + 7 - (KeyCode%8)) );
}

Boolean ModifierKeyDown(int ModifierMask)
{
	KeyMap	theKeys;
	int		KeyCode;

	switch (ModifierMask) {
		case cmdKey:		KeyCode = CmdKeyCode;		break;
		case shiftKey:		KeyCode = ShiftKeyCode;		break;
		case alphaLock:		KeyCode = CapsLockCode;		break;
		case optionKey:		KeyCode = OptionKeyCode;	break;
		case controlKey:	KeyCode = ControlKeyCode;	break;
		default:			KeyCode = -1L;				break;
	}

	GetKeys(&theKeys);
	return(BitTst(&theKeys, KeyCodeToOffset(KeyCode)));
}

int GetCurrentModifiers()
{
	int	CurMods = 0;

	if (!Button())						CurMods |= btnState;
	if (ModifierKeyDown(cmdKey))		CurMods |= cmdKey;
	if (ModifierKeyDown(shiftKey))		CurMods |= shiftKey;
	if (ModifierKeyDown(alphaLock))		CurMods |= alphaLock;
	if (ModifierKeyDown(optionKey))		CurMods |= optionKey;
	if (ModifierKeyDown(controlKey))	CurMods |= controlKey;
}

WaitTicks(long DelayTicks)
{
	long	TC = TickCount();
	while (TC+DelayTicks > TickCount()) ;
}

StartDetachedTickDelay()
{
	DetachedTickDelayStart = TickCount();
}

StopDetachedTickDelay(long DelayTicks)
{
	while ( TickCount() < (DetachedTickDelayStart+DelayTicks) )
		;
}

/****
 * InitMacintosh()
 *
 * Initialize all the managers & memory
 *
 ****/

InitMacintosh()

{
	MaxApplZone();
	
	InitGraf(&thePort);
	InitFonts();
	FlushEvents(everyEvent, 0);
	InitWindows();
	InitMenus();
	TEInit();
	InitDialogs(0L);
	InitCursor();

}

int StdHandleDiskInsertion(long evtMessage)
{
	int		MountResult = HiWord(evtMessage);
	int		BadMountResult;
	Point	DBoxPoint = {100, 130};

	if (MountResult != noErr)
		BadMountResult = DIBadMount(DBoxPoint, evtMessage);
	else
		BadMountResult = noErr;
	return(BadMountResult);
}

Handle GetFreshResource(OSType theResType, int theResID)
{
	Handle h;

	SetResLoad(FALSE);
	if ((h = GetResource(theResType, theResID)) != 0L)
		ReleaseResource(h);
	SetResLoad(TRUE);
	if ((h = GetResource(theResType, theResID)) == 0L) return(h);
	HNoPurge(h);
	return(h);
}

int MacintoshType()
{
	SysEnvRec	theWorld;

	SysEnvirons(1, &theWorld);
	return(theWorld.machineType);
}

int SystemVersion()
{
	SysEnvRec	theWorld;

	SysEnvirons(1, &theWorld);
	return(theWorld.systemVersion);
}
	
Boolean OldTrapAvailable(int tNumber)
{
	return(GetTrapAddress(tNumber) != GetTrapAddress(_Unimplemented));
}

Boolean NewTrapAvailable(int tNumber, int tType)
{
	return(NGetTrapAddress(tNumber, tType) != GetTrapAddress(_Unimplemented));
}

Boolean TrapAvailable(int tNumber, int tType, Boolean NeedSeparateTrapTables)
/*	Checks if trap available; if it is a trap which requires separate
 *	trap tables (on 128K ROM and higher; not on 64K ROM, 128K and 512K macs),
 *	it first checks to see if there are separate trap tables.
 *	Borrowed from Programmers Guide to MultiFinder.
 *		-TA 10/30/90
 */
{
	if (MacintoshType() < envMachUnknown) {
		if (NeedSeparateTrapTables)
			return(FALSE);
		else
			return( OldTrapAvailable(tNumber) );
	}
	else
		return( NewTrapAvailable(tNumber, tType) );
}
