/*
 *	MathUtils.c
 *	Includes math utilities I often use.
 */

#include <math.h>
#include "MathUtils.h"

int Sign(double d)
{
	return( (d==0.0) ? 0 : ( (int) (d/fabs(d)) ) );
}

int RoundToInt(double d)
{
	return( (int) ((fabs(d)+0.5)*Sign(d)) );
}

double Round(double d, int i)
{
	double	tmpD, tmpMod;

	tmpD = fabs(d) * exp(log(10) * i) - 0.5;
	tmpMod = fmod(tmpD, 1.0);
	if ( (tmpMod == 0.0) || (1.0-tmpMod < EffectiveZero) )
		tmpD += 0.1; /* on .5 border */
	return( Sign(d) * ceil(tmpD) * exp(log(10) * (-i)) );
}

Round2Doubles(double d1, double d2, int Places, double *f1, double *f2)
/*	Rounds d1 and d2 to f1 and f2 such that f1 and f2 differ by Places
 *	decimal places.
 *		-TA 3/21/91
 */
{
	double	diff;
	int		i;

	diff = fabs(d1-d2);
	if (diff == 0.0) {
		*f1 = d1;
		*f2 = d2;
		return;
	}

	i = 0;
	while(diff<1) {
		++i;
		diff *= 10;
	}
	while(diff>10) {
		--i;
		diff /= 10;
	}
	i += (Places - 1);
	*f1 = Round(d1, i);
	*f2 = Round(d2, i);

}
