/*
 *	SoundUtils.c
 *
 *	Sound library based on calls I wrote for Rascal in 1985.
 *		-Tyler Abbott, 12/89
 */

#include <math.h>
#include <SoundDvr.h>
#include "SoundUtils.h"

#define	PI	3.14159265358979323846

Note(frequency, amplitude, ticks, compRtn)
	int		frequency, amplitude, ticks;
	char	*compRtn;
	/*
		Generates square-wave tone of given frequency (in hertz),
		amplitude (0-255), and ticks (1/60ths of a second),
		with compRtn set to 0L for asynchronous (detached), -1L for synchronous
		generation.
	*/
{
	SWSynthRec	mySynth;

    mySynth.mode = swMode;
    mySynth.triplets[0].count		= 783360 / frequency;	/* See "Inside Mac" */
    mySynth.triplets[0].amplitude	= amplitude;
    mySynth.triplets[0].duration	= ticks;
    mySynth.triplets[1].count		= 0;
    mySynth.triplets[1].amplitude	= 0;
    mySynth.triplets[1].duration	= 0;
    StartSound(&mySynth, 14L, compRtn);
};

FourToneFixed(duration, freq1, freq2, freq3, freq4, waveptr, compRtn)
	int		duration;
	Fixed	freq1, freq2, freq3, freq4;
	char	*waveptr;
	char	*compRtn;
{
    FTSoundRec		mySound;
    FTSynthRec		mySynth;
    unsigned char	wave[256];
    				/* Wave s/b defined *unsigned* char in SoundDvr.h (?) */
    int				i;
    char			*realwaveptr;

	if (waveptr == 0L) {
		for (i=-1; ++i<256; )
			wave[i] = 255 *(i/128); /* Square Wave */
		realwaveptr = (char *) &wave;
	}
	else
		realwaveptr = waveptr;

	mySound.duration = duration;
	mySound.sound1Rate  = freq1;
	mySound.sound1Phase = 0L;
	mySound.sound2Rate  = freq2;
	mySound.sound2Phase = 0L;
	mySound.sound3Rate  = freq3;
	mySound.sound3Phase = 0L;
	mySound.sound4Rate  = freq4;
	mySound.sound4Phase = 0L;
	mySound.sound1Wave = (WavePtr) realwaveptr;
	mySound.sound2Wave = (WavePtr) realwaveptr;
	mySound.sound3Wave = (WavePtr) realwaveptr;
	mySound.sound4Wave = (WavePtr) realwaveptr;

	mySynth.mode = ftMode;
	mySynth.sndRec = &mySound;

	StartSound(&mySynth, 6L, compRtn);
}

FourTone(duration, frequency1, frequency2, frequency3, frequency4,
													waveptr, compRtn)
	int		duration;
	int		frequency1, frequency2, frequency3, frequency4;
	char	*waveptr;
	char	*compRtn;
{
	Fixed	freq1, freq2, freq3, freq4;

		/* In sequence following, 87 represents an integer
		   approximation of 10^6/(44.93*256) (see Inside Mac)
		   Shifting is done here to heighten accuracy, with the
		   consideration that the / result is word-sized.
		 */
	freq1 = frequency1;	freq1 = (freq1<<16) / 87;
	freq2 = frequency2;	freq2 = (freq2<<16) / 87;
	freq3 = frequency3;	freq3 = (freq3<<16) / 87;
	freq4 = frequency4;	freq4 = (freq4<<16) / 87;

	FourToneFixed(duration, freq1, freq2, freq3, freq4, waveptr, compRtn);
}

MakeSoundSine(unsigned char *SineWave)
{
	int 	i;

	SineWave[0] = 127; SineWave[64] = 254;
	SineWave[128] = 127; SineWave[192] = 0;
	for(i=1;++i<64;) {
		SineWave[i] =  (int) (127.0 * sin(2*PI*i/256.0) + 0.5) + 127;
			/* RoundToInt(127.0 * sin(2*PI*i/256.0)) + 127; */
		SineWave[128-i] = SineWave[i];
		SineWave[128+i] = 254-SineWave[i];
		SineWave[256-i] = 254-SineWave[i];
	}
}