#include "StringUtils.h"

sEquate(char *s, char *t)	/* sets s = t */
{
	int i;
	for(i=0;i<=(unsigned char) (t[0]) && i<=255;++i)
		s[i] = t[i];
}

sAppend(char *s, char *t)	/* appends t to the end of s*/
{
	int i;
	for(i=1; i<=(unsigned char) (t[0]) && i<=(255-(unsigned char) (s[0])); ++i)
		s[i + (unsigned char) (s[0])] = t[i];
	s[0] += (i-1);
}

sInsertC(char *s, int ch, int pos)
	/* Insert character ch into Pascal string s at position pos */
{
	int	i;

	if (pos <=0) return;
	if (pos > (unsigned char) (++s[0])) pos = s[0];
	if (pos < (unsigned char) (s[0])) 
		for (i=(unsigned char) (s[0]);i>pos;--i) s[i] = s[i-1];
	s[pos] = ch;
}

int cComp(char c, char d)
	/*	Returns 1 if c>d, 0 if c==d, -1 if c<d	*/
{
	return( (c>d) ? 1 : ( (c<d) ? -1 : 0 ) );
}

int	sComp(char *s, char *t)
	/*	Returns 1 if s>t, 0 if s==t, -1 if s<t	*/
{
	int	i;

	if (s[0] == 0)			return( (t[0]==0) ? 0 : -1 );
	if (t[0] == 0)			return( 1 );
	for (i=1;;++i) {
		if (s[i] != t[i])	return( cComp(s[i], t[i]) );
		if (i>=s[0])		return( (i>=t[0]) ? 0 : -1 );
		if (i>=t[0])		return( 1 );
	}
}

int	sUpperComp(char *s, char *t)
	/*	Returns 1 if s>t, 0 if s==t, -1 if s<t	*/
{
	int	i;

	if (s[0] == 0)			return( (t[0]==0) ? 0 : -1 );
	if (t[0] == 0)			return( 1 );
	for (i=1;;++i) {
		if (toupper(s[i]) != toupper(t[i]))
			return( cComp(toupper(s[i]), toupper(t[i])) );
		if (i>=s[0])		return( (i>=t[0]) ? 0 : -1 );
		if (i>=t[0])		return( 1 );
	}
}

sSubstC(char *s, int findch, int subch)
	/*	Substitutes subch for findch in Pascal string s	*/
{
	int	i;

	for(i=0; i<=(unsigned char) (s[0]); ++i)
		if (s[i] == findch)
			s[i] = subch;
}

int sFindC(char *s, char c)
/*	Finds the first occurence of c in s; returns index in s (-1 if not found)	*/
{
	int	i;

	for (i=1; i<=(unsigned char) (s[0]); ++i)
		if (c == s[i])
			return(i);
	return(-1);
}

Boolean	StrFind(char *Sub, char *s)
/*	Returns "string Sub is in string s" */
{
	char	*t;

	PtoCstr(Sub);	PtoCstr(s);
	t = strstr(s, Sub);
	CtoPstr(Sub);	CtoPstr(s);

	return(t != NULL);
}

Boolean	strToF(char *s, double *pf)
	/* Converts string to double */
	/* Was straight from "fgetf", Rascal __SaneIO library */
{
	/*		*** Rascal Version ***
		char	decimal[26];
		Boolean bad;
	
		if (s[0]==0) { *pf=0; return(TRUE); } (* Empty strings return 0.0 *)
		strtodec(s,&decimal,&bad);
		if (bad) return(FALSE);
		DecToF(&decimal, pf);
	*/
	double	tmp;

	if (s[0]==0) { *pf=0; return(TRUE); } /* Empty strings return 0.0 */
	if ( (s[0] == 1) && ( (s[1]=='+') || (s[1]=='-') ) )
		return(FALSE);
	PtoCstr(s);
	if (sscanf(s, "%lf", &tmp) == 0) {
		CtoPstr(s);
		return(FALSE);
	}
	CtoPstr(s);
	*pf = tmp;
	return(TRUE);
}

	/*	The following three float to string routines do as follows:
	 *		myFToStr: string will be no longer than fw chars, unless fw
	 *			too small even with prec 0, unless fw=0, in which case
	 *			fw is ignored, and it will act like myFToStrNoPad.
	 *		myFToStrNoPad: like myFToStr, but no padding ever. Pass 0 for
	 *			fw to get string of minimum length to support prec;
	 *			otherwise prec can be overridden.
	 *		PadFToStr: same as myFToStr, except if Pad, will pad to fw
	 *			with leading spaces
	 *		PadFToStrDec: Tries to line up decimal places; has additional
	 *			argument, WholePlaces, number of places to line up on
	 *			left side of decimal point. Will pad to fw like PadFToStr.
	 *
	 *	Note that in myFToStrNoPad, prec can override fw; everywhere else,
	 *	fw overrides prec.
	 */

myFToStr(double f, char *ps, int fw, int prec)
	/*  Converts double f to string of maximum length fw.  Uses either
	 *  decimal or scientific notation, whichever gives higher number
	 *  of significant figures.  If same number of sig figs from both,
	 *  uses decimal notation.  Not guaranteed to be completely bug
	 *  free, especially if rounding changes the number of sig figs,
	 *  or if fieldwidth is too small to properly display f in either
	 *  notation.
	 *	(Above Rascal comments)
	 */
	/* NOTE: Does not currently treat fw properly--use small (0) fw. -11/89
	 */
	/*	Call with:
	 *		Str255	ts; (* same as char ts[256]; *)
	 *		char	*t = (char *) &ts;
	 *		myFToStr(637.2,t,0,2);
	 */
{
		/*	Loop necessary because sometimes prec is too big for fw */
	do {
		sprintf(ps,"%*.*g",fw, prec, f);
		CtoPstr(ps);
		if (fw==0)	break;
		--prec;
	} while ( (ps[0] > fw) && (prec>-1) );
	return;
}

myFToStrNoPad(double f, char *ps, int fw, int prec)
/*	See comments under myFToStr above	*/
{
	do {
		sprintf(ps,"%*.*g", 0 /*fw*/, prec, f);
		CtoPstr(ps);
		if (fw==0)	break;
		--prec;
	} while ( (ps[0]>fw) && (prec>-1) );
	return;
}

PadFtoStr(double f, char *ps, int fw, int prec, Boolean Pad)
	/* Is this necessary with myFToStr using sprintf? (it pads itself) */
{
	myFToStr(f, ps, fw, prec);
	if (Pad) while (ps[0]<fw) sInsertC(ps,' ',1);
}

PadFtoStrDec(double f, char *ps, int fw, int prec, int WholePlaces, Boolean Pad, Boolean PadRight)
	/* Is this necessary with myFToStr using sprintf? */
{
	int	DotIndex, RightPad, i;

	myFToStrNoPad(f, ps, fw, prec);
	DotIndex = sFindC(ps, '.');
	if ( (DotIndex >= 1) && (DotIndex <= WholePlaces) ) {
		myFToStrNoPad(f, ps, fw - (WholePlaces+1-DotIndex), prec);
		DotIndex = sFindC(ps, '.');
	}
	if (DotIndex == -1)
		RightPad = fw - MAX(ps[0], WholePlaces);
	else if ( DotIndex <= WholePlaces )
		RightPad = fw - (WholePlaces+1-DotIndex) - ps[0];
	else
		RightPad = fw - ps[0];
	
	if (Pad)
		while (ps[0]+RightPad<fw) sInsertC(ps,' ',1);

	if (PadRight)
		for (i=0; i<RightPad; ++i)
			sInsertC(ps, ' ', ps[0]+1);

/*
	if (Pad) {
		if (PadRight)
			while (ps[0]<fw) sInsertC(ps,' ',1);
		else
			while (ps[0]+RightPad<fw) sInsertC(ps,' ',1);
	}
*/
}

Boolean	IsInteger(char *ps)
{
	int	i;

	if (ps[0] == 0) return(TRUE);
	i = 1;
	if (ps[i] == '-')
		if (ps[0] == 1) return(FALSE); else ++i;
	while(i<=ps[0])
		if ( !isdigit( (int) (ps[i++]) ) ) return(FALSE);
	return(TRUE);
}

Boolean IsFloat(char *ps)
{
	double	d;
	return(strToF(ps, &d));
}


Boolean	myStringToLong(char *ps, long *pl)
{
	long	tmp;

	if (ps[0] == 0) { *pl = 0L; return(TRUE); }
	if ( (ps[0] == 1) && ( (ps[1]=='+') || (ps[1]=='-') ) )
		return(FALSE);
	PtoCstr(ps);
	if (sscanf(ps, "%ld", &tmp) == 0) {
		CtoPstr(ps);
		return(FALSE);
	}
	CtoPstr(ps);
	*pl = tmp;
	return(TRUE);
/*
	if ( IsInteger(ps) ) { StringToNum(ps, pl); return(TRUE); }
	else return(FALSE);
*/
}

LongToString(long l, char *ps)
{
	NumToString(l, ps);
/* Standard C version
	sprintf(ps, "%ld", l);
	CtoPstr(ps);
*/
}

Boolean	GetLine(char *s, char *buf, long *pfc, long max)
/*	Gets line (until return) from buf, starting with position *pfc,
 *	subject to maximum max (returns FALSE if no return character),
 *	and puts result in string s.
 *	Includes return character in s.
 */
{
	long l;

	for(l=1; ; ++l) {
		if (*pfc+l-1L>max) {
			*pfc = max;
			s[0] = --l;
			return(FALSE);
		}
		s[l] = *(buf+*pfc+l-1L);
		if ( (*(buf+*pfc+l-1L) == '\r') || (*(buf+*pfc+l-1L) == '\n') )
			break;
	}
	s[0] = l;
	*pfc += l;
	return(TRUE);
}
