/*	bb.c -- Bouncing Ball
 *
 *	Simulates the motion of a ball bouncing on a sinusoidally-
 *	vibrating table, say, a ball bearing on a speaker.
 *
 *	Tyler Abbott and Nick Tufillaro
 *	1987 through 1990
 */

#include <stdlib.h>
#include <MultiFinder.h>
#include <MFUtils.h>
#include "bbMenus.h"
#include "bbPrototypes.h"
#include "bbDebug.h"
#include "bbGraphics.h"	/*	For GetInitialConditions() only	*/

		/*	External variables from bbIteration.c	*/
/*
extern	int		HitNumber, HitInd1, HitInd2, RealHitNumber;
extern	double	HitTime;
extern	double	tstop;
extern	int		StuckStatus;
*/
			/*	These are used in _MAIN only, in	*/
			/*	calls to GraphFrom and AddSolution	*/
/*
extern	double			ti, tj, vi, xi, phase;
*/

		/*	External variables from bbContGraphics.c	*/
extern	Boolean			RedrawSine;
		/*	External variables from bbSpecial.c	*/
extern	Boolean			*pBDGraphInAllWindows, *pBAGraphInAllWindows;
extern	Boolean			*pBDUseColor, *pBAUseColor;
		/*	External variables from bbFiles.c	*/
extern	int				OpenWindex;
		/*	External variables from bbMenus.c	*/
extern	Boolean			NeedToSave;
		/*	External variables from bbGraphics.c	*/
extern	CursHandle		watch;
extern	Windows			*pw;	/*	For GetInitialConditions() only	*/

			/*	Global variables used throughout program	*/
IterationVars		NormalVars, *pCurrentVars;
hBBPreferences		hBBPrefs;
/*
double	v0, delta;		(* Initial conditions *)
double	A, freq, damp;	(* Physical System Parameters *)
double	Period, Omega;	(* Calculated from freq *)
*/

			/*	Global variable used in bbIteration.c, bbGraphics.c, special */
SolutionsHandles	mySolutions;

			/*	Global variables used in bbIteration.c, bbMenus.c, elsewhere */
int		ProgramStatus;
int		theEventMask, UsualEventMask;
Boolean	PrintingOnly;
int		CurrentModifiers;
Boolean	SingleStepOn;
Boolean	UserWaiting;

double	nV0, nDelta;	/* to be next v0, delta */
int		SpecialAccuracy;

int		LastProgramStatus;
int		IterationStatus;
Boolean	ClearSolutions;
Boolean	RestartIteration;
Boolean	PAUSING, OKToContinue;
Boolean	OKToContinueBD = FALSE, OKToContinueBA = FALSE;
Boolean	MemoryFull;
Boolean	SuppressErrorMessages;

int		SolutionMemSize;	/* # solutions memory allocated for */
int		NumSolutionArrays;	/* # solution arrays memory ""  "" */

Boolean MultiFinderRunning;	/*	for MultiFinder	*/
int		BackgroundSleep;
Boolean	InForeground, HasWaitNextEvent;
		/*	If in background, breaks in the middle of _MAIN;		*/
		/*	NeedToFinishIteration tells if need to finish on resume	*/
Boolean	NeedToFinishIteration;

Boolean	UsingColorMac;
Boolean	System7OrLater;
Boolean	WasADefaultsFile;

Boolean	GettingParamFiles;
Boolean	GotATextFile, GotABDFile, GotABAFile;
int		OpenedParamFiles;

long	ParamTest = 0L;

Boolean	RootFindingError = FALSE;

	/*	--------------------------------------------------	*/
	/*	----------	Memory-Checking Routines	----------	*/
	/*	--------------------------------------------------	*/

Boolean CheckMem(Size ExtraMemory)
/*	Require at least (MinMemory + ExtraMemory) free bytes, "breathing space."
 *			-TA 3/23/90
 *	ExtraMemory 0 except in "Special" functions
 *			-TA 3/5/91
 */
{
	Size	grow, MemAvail;

	MemAvail = MaxMem(&grow);
	return(MemAvail > (MinMemory + ExtraMemory));
}

Boolean StartingMemCheck()
{
	if (HitNumber >= SolutionMemSize-1) {
		if (!AddSolutionArray()) {
			MemoryFull = TRUE;
			return(FALSE);
		}
	}
	if (!CheckBothBodyTESizes()) {
		MemoryFull = TRUE;
		return(FALSE);
	}
	if (!CheckMem(0L)) {
		MemoryFull = TRUE;
		return(FALSE);
	}

	MemoryFull = FALSE;
	return(TRUE);
}

	/*	--------------------------------------------------	*/
	/*	----------	Iteration Control Routines	----------	*/
	/*	--------------------------------------------------	*/

Boolean StartIterating(int New_PS)
{
	Boolean	ContinueOK;

	if (HitNumber >= MaxSolutions) {
		ERROR("\PMaximum number of solutions reached");
		return(FALSE);
	}
	if (!StartingMemCheck())	{
		ERROR("\PSorrynot enough memory to perform requested action");
		return(FALSE);
	}

		/*	If already done StopIterating(), ProgramStatus will be WAITING	*/
	if (ProgramStatus != WAITING) {
		switch (ProgramStatus) {
			case BIFURCATION_DIAGRAM:
			case BASINS_OF_ATTRACTION:	ContinueOK = TRUE;	/* ? */		break;
			default:					ContinueOK = HitNumber >= 0;	break;
		}
		/*	Need variable to show in middle of BifD?	*/
		StopIterating(ContinueOK);
	}
	if ( (New_PS < BIFURCATION_DIAGRAM) &&
			(LastProgramStatus >= BIFURCATION_DIAGRAM) ) {
		if ( ((LastProgramStatus == BIFURCATION_DIAGRAM) && (*pBDGraphInAllWindows)) ||
				((LastProgramStatus == BASINS_OF_ATTRACTION) && (*pBAGraphInAllWindows)) ) {
			RedrawSine = TRUE;
			SetValuesFromTableParams(0.0, 0.0, 0.0);
			SetHeaders(A, freq, damp);
			NewGraphics(TRUE, TRUE);
		}
	}

	LastProgramStatus = ProgramStatus = New_PS;
	IterationStatus = WRITING;
	ControlMenuStart(ProgramStatus);
	PAUSING = FALSE;
	switch (ProgramStatus) {
		case BIFURCATION_DIAGRAM:	OKToContinueBD = FALSE;
									if (*pBDGraphInAllWindows) {
										ActivateTextScrolls(FALSE);
										SetTextTitle(TRUE, BifD);
									}
									SetTolerance(SpecialAccuracy);
									break;
		case BASINS_OF_ATTRACTION:	OKToContinueBA = FALSE;
									if (*pBAGraphInAllWindows) {
										ActivateTextScrolls(FALSE);
										SetTextTitle(TRUE, BAttr);
									}
									SetTolerance(SpecialAccuracy);
									break;
		default:					OKToContinue = FALSE;
									ActivateTextScrolls(FALSE);
									SetTolerance(Accuracy);
									break;
	}
	RealHitNumber = -1;
	return(TRUE);
}

StopIterating(Boolean ContinueOK)
{
	Boolean *pOKToContinue;

	switch (ProgramStatus) {
		case BIFURCATION_DIAGRAM:
									pOKToContinue = &OKToContinueBD;
									pCurrentVars = &NormalVars;
									if (*pBDGraphInAllWindows) {
										NewGraphics(TRUE, FALSE);
										SetHeaders(A, freq, damp);
										SetTextTitle(TRUE, TxOut);
											/*	zeroed out in StartIterating()	*/
									}
									break;
		case BASINS_OF_ATTRACTION:
									pOKToContinue = &OKToContinueBA;
									pCurrentVars = &NormalVars;
									if (*pBAGraphInAllWindows) {
										NewGraphics(TRUE, FALSE);
										SetHeaders(A, freq, damp);
										SetTextTitle(TRUE, TxOut);
											/*	zeroed out in StartIterating()	*/
									}
									HiliteBAImpact(-1, -1);
									break;
		default:
									pOKToContinue = &OKToContinue;
									break;
	}
	*pOKToContinue = ContinueOK && (StuckStatus != STUCK_FOREVER);
	ControlMenuStop(ProgramStatus, *pOKToContinue);
	ProgramStatus = WAITING;
	IterationStatus = NOT_ITERATING;
	PAUSING = FALSE;
	ActivateTextScrolls(TRUE);
}

BeginIfAppropriate()
{
	if (RestartIteration)
		StartIterating(NORMAL_ITERATION);
	else if (ProgramStatus == NORMAL_ITERATION)
		StopIterating(FALSE);
}

	/*	--------------------------------------------------	*/
	/*	----------	Solution-Handling Routines	----------	*/
	/*	--------------------------------------------------	*/

double GetSolutionInfo(int HitNum, int WhichDouble)
{
	switch (WhichDouble) {
		case GET_PHASE:
				return((**(mySolutions[SI1(HitNum)]))[SI2(HitNum)].del);
		case GET_VELOCITY:
				return((**(mySolutions[SI1(HitNum)]))[SI2(HitNum)].vel);
		case GET_TIME:
				return((**(mySolutions[SI1(HitNum)]))[SI2(HitNum)].time);
	}
}

ResetSolutions()
{
	int	i=0;

	SetHitNumber(-1);
	RealHitNumber = -1;
	while (++i < MaxSolutionArrays) {
		if (mySolutions[i] != NIL) {
			DisposHandle(mySolutions[i]);
			mySolutions[i] = NIL;
		}
		else break;
	}
	SolutionMemSize = SolutionIncrement;
	NumSolutionArrays = 1;
}

Boolean AddSolutionArray()
{
	mySolutions[NumSolutionArrays] =
			(SolutionsHandle) NewHandle( (sizeof(Solutions)) );
	if ((long) mySolutions[NumSolutionArrays] == NIL) return(FALSE);
	else {
		++NumSolutionArrays;
		SolutionMemSize += SolutionIncrement;
	}

	return(CheckMem(0L));
}

AddSolution(double del, double v, int StickyState)
{
	static	double	OldHitTime;	/*	Used for MakeTone only	*/
			Boolean	WriteOK;

	WriteOK = AddGraphicSolution(HitNumber, del, v, HitTime,
			StickyState, ProgramStatus);

	if (  (!PrintingOnly) &&
			( (InForeground) || (UseSoundInBackground) )  ) {
		if ( HitTime != 0.0 )
			MakeTone(del, v, OldHitTime, HitTime, Period, TRUE);
		else MakeTone(del, v, 0.0, 0.0, 0.0, FALSE);
	}
	OldHitTime = HitTime;

	if (!WriteOK) {
		ERROR("\PNot enough memory to store more solutions");
		MemoryFull = TRUE;
		if (ProgramStatus == OPENING) 
			StopOpeningFile(FALSE);
		StopIterating(HitNumber>=0);
	}

	if (ProgramStatus == PLAYBACK) return;

	if ( (HitNumber >= SolutionMemSize-1) && (!MemoryFull) ) {
		if (!AddSolutionArray()) {
			MemoryFull = TRUE;
			if (!( (ProgramStatus == OPENING) && (FinishedOpening()) )) {
				ERROR("\PNot enough memory to store more solutions");
				if (ProgramStatus == OPENING)
					StopOpeningFile(FALSE);
				StopIterating(HitNumber>=0);
			}
		}
	}

	(**(mySolutions[HitInd1]))[HitInd2].stuck	= StickyState;
	(**(mySolutions[HitInd1]))[HitInd2].vel		= v;
	(**(mySolutions[HitInd1]))[HitInd2].del		= del;
	(**(mySolutions[HitInd1]))[HitInd2].time	= HitTime;

	if (HitNumber >= MaxSolutions) {
		ERROR("\PMaximum number of solutions reached");
		StopIterating(FALSE);
		return;
	}
}

	/*	----------------------------------------------	*/
	/*	----------	Initializing Routines	----------	*/
	/*	----------------------------------------------	*/

SetDefaultPrefs()
{
	int	i;

/*!!*/	DefaultV0 = DEFAULT_V0;
/*!!*/	DefaultDelta = DEFAULT_DELTA;
/*!!*/	DefaultA = DEFAULT_A;
/*!!*/	DefaultFreq = DEFAULT_FREQ;
/*!!*/	DefaultDamp = DEFAULT_DAMP;

/*!!*/	DefaultBDGIAW = FALSE;
/*!!*/	DefaultBAGIAW = FALSE;
/*!!*/	DefaultBDUseColor = UsingColorMac;
/*!!*/	DefaultBAUseColor = UsingColorMac;

/*!!*/	GraphWhileOpening = FALSE;
/*!!*/	DelimiterIndex = SPACE_Delim;

/*!!*/	RunInBackground = FALSE;
/*!!*/	FasterBackground = FALSE;
/*!!*/	UseSoundInBackground = FALSE;
/*!!*/	SuppressSaveWarnings = FALSE;
/*!!*/	AutoCommandKey = FALSE;

/*!!*/	Cycles = DEFAULT_CYCLES;
/*!!*/	Accuracy = DEFAULT_ACCURACY;

/*!!*/	MinPhaseInt = INIT_MinPhaseInt;
/*!!*/	UseAutoVMax = TRUE;
/*!!*/	VMax = 10.0;

/*!!*/	RealPeriodTicks = 0;
/*!!*/	ScreenNumPeriods = DefaultNumPeriods;
/*!!*/	ScreenNumAmplitudes = DefaultNumAmplitudes;
/*!!*/	ScrollTrajectory = FALSE;
/*!!*/	ImpactPhasesOffset = DEFAULT_ImpactPhasesOffset;
/*!!*/	FasterPhaseSpace = FALSE;
/*!!*/	TS_SampleRate = DEFAULT_TS_SAMPLERATE;
/*!!*/	UseTimeSeries = FALSE;

/*!!*/	UseAdvancedSound = FALSE;
/*!!*/	PhaseTonesOn = TRUE;
/*!!*/	VelTonesOn = FALSE;
/*!!*/	NumOctaves = 1;
/*!!*/	CurrentMode = ContinuousMode;

/*!!*/	for(i=0; i<NumSolutionTypes; ++i) {
		psi[i] = csi[i] = i;
	}
}

SetDefaults()
{
	pIterationVars	piv = pCurrentVars;

	pCurrentVars = &NormalVars;

	DefaultV0 = v0;
	DefaultDelta = delta;
	DefaultA = A;
	DefaultFreq = freq;
	DefaultDamp = damp;

	DefaultBDGIAW = *pBDGraphInAllWindows;
	DefaultBAGIAW = *pBAGraphInAllWindows;
	DefaultBDUseColor = *pBDUseColor;
	DefaultBAUseColor = *pBAUseColor;

	pCurrentVars = piv;
}

SaveDefaults()
{
	SetDefaultWindowConfiguration();
	SetDefaults();
	if (!SaveDefaultFile()) {
		ERROR("\PError: Could not save defaults file");
		return;
	}
}

SetDefaultParams()
{
	v0	=	DefaultV0;	delta	=	DefaultDelta;
	A	=	DefaultA;	freq	=	DefaultFreq;	damp = DefaultDamp;
	SetFrequencyParams();
	SpecialAccuracy = MAX(Accuracy, MIN_SPECIAL_ACCURACY);
	tstop = GetStopTime(Cycles);
}

DetermineIfColor()
{
	GDHandle	theGD;
	int			ScreenDepth;

	if ( TrapAvailable(GetMainDeviceTrapID, ToolTrap, TRUE) ) {
		theGD = GetMainDevice();
		ScreenDepth = (**(**theGD).gdPMap).pixelSize;
		UsingColorMac = (ScreenDepth >= 4);
	}
	else
		UsingColorMac = FALSE;
}

InitMultiFinderGlobals()
{
	EventRecord	anEvent;

	InForeground = TRUE;
	MultiFinderRunning = MultiFinderOn();
	HasWaitNextEvent = WNEIsImplemented();
	BackgroundSleep = (FasterBackground) ? DAY_DREAM : CATNAP;

	if (MultiFinderRunning)			/* Brings program to top MF layer */
		if (HasWaitNextEvent) WaitNextEvent(0, &anEvent, 0L, NIL);
		else GetNextEvent(0, &anEvent);
}

_INIT()
{
	int		i;
	Boolean	GotPrefs;

	if (!CheckMem(MinMemToStart-MinMemory)) {
		ERROR("\PSorrynot enough memory to run Bouncing Ball");
		exit(1);
	}

	DILoad();

	InitMultiFinderGlobals();

	UserWaiting = FALSE;
	watch = GetCursor(watchCursor);
	StartWaiting();

	System7OrLater = (SystemVersion() >> 8) >= 7;
	DetermineIfColor();

	InitBBFiles();
	WasADefaultsFile = OpenDefaults(&GotPrefs);
	if (GotPrefs) {
		DefaultBDUseColor = DefaultBDUseColor && UsingColorMac;
		DefaultBAUseColor = DefaultBAUseColor && UsingColorMac;
	}
	else {
		ResrvMem(sizeof(BBPreferences));
		hBBPrefs = (hBBPreferences) NewHandle(sizeof(BBPreferences));
		HLock((Handle) hBBPrefs);	/*	almost always locked	*/
		SetDefaultPrefs();
	}

	CreateWindowRecords();
	pCurrentVars = &NormalVars;
	LastProgramStatus = ProgramStatus = WAITING;
	SuppressErrorMessages = FALSE;
	SetDefaultParams();
	SetTolerance(Accuracy);
/*
	if (!Initialize_Iteration_Globals(FALSE, FALSE)) ;
*/

	InitBBMenus(UsingColorMac);
	InitSounds();
	if (!InitGraphics()) {
		StopWaiting();
		ERROR("\PSorrynot enough memory to run Bouncing Ball");
		exit(1);
	}
/*
	InitBBMenus();
*/
	AddWindowsToMenu();
/*
	InitBBFiles();
*/

	StopIterating(FALSE);		/* Takes care of global vars */

	mySolutions[0] = (SolutionsHandle) NewHandle( (sizeof(Solutions)) );
	for(i=1;i<MaxSolutionArrays;++i)
		mySolutions[i] = NIL;
	ResetSolutions();

	ClearSolutions = FALSE;
	RestartIteration = TRUE;
	SingleStepOn = FALSE;

	InitSpecial();

	if (MemoryFull = !CheckMem(0L)) {
		StopWaiting();
		ERROR("\PSorrynot enough memory to run Bouncing Ball");
		_HALT();
	}

	theEventMask = UsualEventMask = everyEvent;

	PrintingOnly = GettingParamFiles =  FALSE;
	GotATextFile = GotABDFile = GotABAFile = FALSE;
	OpenedParamFiles = 0;
	switch (BBGetParamFiles()) {
		case appPrint:
						PrintingOnly = TRUE;
		case appOpen:
						theEventMask = UsualEventMask = 0;
						GettingParamFiles = TRUE;
						break;
	}

	InitialWindowConfiguration(PrintingOnly);
	StopWaiting();
}

	/*	--------------------------------------------------	*/
	/*	----------	Parameter-Setting Routines	----------	*/
	/*	--------------------------------------------------	*/

GetInitialFromLast()
{
	if (HitNumber >= 0) {
		nV0 = (**(mySolutions[HitInd1]))[HitInd2].vel;
		nDelta = (**(mySolutions[HitInd1]))[HitInd2].del;
	}
}

Boolean	GetInitialConditions(Boolean DefaultSame)
{
	Boolean			GoodValues = TRUE;
	pIterationVars	piv = pCurrentVars;
	Boolean			GoodOp;

	pCurrentVars = &NormalVars;
	if (DefaultSame)	{	nV0 = v0;	nDelta = delta;	}

	RestartIteration = TRUE;
	if ( DefaultSame || !SHIFT_DOWN )
		GoodOp = DoInitCondDialog(&nV0, &nDelta, &ClearSolutions, &RestartIteration);
	else
		GoodOp = TRUE;

	if (GoodOp) {
	 	if (nDelta<0.0)		nDelta += -((int) nDelta)+1;
	 	if (nDelta>=1.0)	nDelta -= (int) nDelta;
		if (nDelta != delta) RedrawSine = TRUE;
		v0 = nV0;	delta = nDelta;
		w[Traj].WindowCleared = FALSE;
		if ( (ProgramStatus < BIFURCATION_DIAGRAM) || (RestartIteration) )
			NewGraphics(ClearSolutions, TRUE);
		if (ClearSolutions)
			ResetSolutions();
		if (RestartIteration) {
			GoodValues = Initialize_Iteration_Globals(RestartIteration, FALSE);
			if (!GoodValues) {
				StopIterating(FALSE);
				return(FALSE);
			}
		}
		else
			pCurrentVars = piv;
		return(TRUE);
/*
		return(RestartIteration);
*/
	}
	else {
		pCurrentVars = piv;
		return(FALSE);
	}
}

SetValuesFromTableParams(double oA, double oFreq, double oDamp)
{
	Boolean	ChangeMade;

	if ( ChangeMade = (freq!=oFreq) )
		SetFrequencyParams();
	if ( ChangeMade = (ChangeMade || (A!=oA)) )
		InitTrajScales();
	if (ChangeMade || (damp!=oDamp)) {
		SetVMax(TRUE);
		InitImpactMapScales();
		SetTrajCoords();
	}
}

Boolean GetTableParameters()
{
	double			oA, oFreq, oDamp;
	Boolean			GoodValues = TRUE, GoodOp;
	pIterationVars	piv = pCurrentVars;

	pCurrentVars = &NormalVars;
	oA = A;	oFreq = freq;	oDamp = damp;
	RestartIteration = TRUE;
	GoodOp = DoPhysSysDialog(&RestartIteration);
	if (GoodOp) {
		if (!CheckNormalSaves(FALSE)) {
			GoodOp = FALSE;
			A = oA; freq = oFreq; damp = oDamp;
		}
	}

	if (GoodOp) {
		pCurrentVars = &NormalVars;
		SetValuesFromTableParams(oA, oFreq, oDamp);
		if ( ((oA==0.0)^(A==0.0)) || ((oFreq==0.0)^(freq==0.0)) )
			RedrawSine = TRUE;
		if (!SpecialGraphing())
			SetHeaders(A, freq, damp);

		if (RestartIteration)
			GoodValues = Initialize_Iteration_Globals(RestartIteration, FALSE);
		if ( (ProgramStatus < BIFURCATION_DIAGRAM) || (RestartIteration) ) {
			NewGraphics(TRUE, TRUE);
			ResetSolutions();
		}
		else
			pCurrentVars = piv;
		if (!GoodValues) {
			StopIterating(FALSE);
			return(FALSE);
		}
		return(TRUE);
	}
	else {
		pCurrentVars = piv;
		return(FALSE);
	}
}

GetTrajScalingParams()
{
	int		oNumPeriods = ScreenNumPeriods;
	int		oNumAmplitudes = ScreenNumAmplitudes;

	if ( DoTrajScalingDialog(&ScreenNumPeriods, &ScreenNumAmplitudes) ) {
		if ( (ScreenNumPeriods != oNumPeriods) ||
					(ScreenNumAmplitudes != oNumAmplitudes) ) {
			RedrawSine = TRUE;
			SetWholeScreenTrajCoords();
			SetTrajCoords();
			SetVMax(TRUE);
			InitImpactMapScales();
			NewGraphics(FALSE, FALSE);
		}
	}
}

GetIterationParams()
{
	int		oAccuracy;

	oAccuracy = Accuracy;
	if ( DoIterationDialog(&Accuracy, &Cycles ) ) {
		tstop = GetStopTime(Cycles);
		SpecialAccuracy = MAX(Accuracy, MIN_SPECIAL_ACCURACY);
		if (ProgramStatus >= BIFURCATION_DIAGRAM)
			SetTolerance(SpecialAccuracy);
		else
			if (Accuracy != oAccuracy)
				SetTolerance(Accuracy);
	}
}

	/*	--------------------------------------------------	*/
	/*	----------	Iteration-Looping Routines	----------	*/
	/*	--------------------------------------------------	*/

AdvanceIterationStatus()
{
	if (IterationStatus > NOT_ITERATING)
		IterationStatus = (IterationStatus + 1) % IterationStatuses;
}

Boolean Handle_ITERATING(int PS)
{
	switch (GetNextSolution(PS)) {
		case BAD_SOLUTION:
							StopIterating((PS==OPENING) && (HitNumber>=0));
							return(FALSE);
		case NO_SOLUTION_YET:
							return(FALSE);
		case FOUND_SOLUTION:
							AdvanceIterationStatus();
							return(TRUE);
		case ROOT_FINDING_ERROR:
							IterationStatus = META_ITERATING;
							RootFindingError = TRUE;
							return(TRUE);
	}
}

Handle_GRAPHING(int PS)
{
	if ( OKtoGraph(PS) ) {
		if (StuckStatus == STUCK)	GraphStuckBall(ti, tj);
		else						GraphFrom(ti, tj, xi, vi);
	}
	AdvanceIterationStatus();
}

Boolean Handle_WRITING(int PS, Boolean *pOKCont)
{
	Boolean	TimeToStop;
static long timer1, timer2;

	*pOKCont = TRUE;
	TimeToStop = !StartNextSolution(PS, pOKCont);
	AddSolution(phase, vi, StuckStatus);
timer2=TickCount()-timer1;
timer1 = TickCount();

	if (PS == NORMAL_ITERATION)
		NeedToSave = TRUE;
	if (PS == BIFURCATION_DIAGRAM)
		GraphBD();
	if (PS == BASINS_OF_ATTRACTION)
		GraphBA();

	AdvanceIterationStatus();
	return(!TimeToStop);
}

Boolean Handle_META_ITERATING(int PS, Boolean *pOKCont)
{
	Boolean	TimeToStop = FALSE, ShouldAdvance = TRUE;

	if (ProgramStatus == BIFURCATION_DIAGRAM) {
		if (RootFindingError) {
			StuckStatus = STUCK;
			CheckBDPeriodicity();
/*
			StuckStatus = NOT_STUCK;
*/
			RootFindingError = FALSE;
		}
		if (!GetNextBD()) {
			*pOKCont = FALSE;
			TimeToStop = TRUE;
		}
	}

	if (ProgramStatus == BASINS_OF_ATTRACTION) {
		if (RootFindingError) {
			StuckStatus = STUCK;
			CheckBAPeriodicity();
/*
			StuckStatus = NOT_STUCK;
*/
			RootFindingError = FALSE;
		}
		if (!GetNextBA(&ShouldAdvance)) {
			*pOKCont = FALSE;
			TimeToStop = TRUE;
		}
	}

	if (SingleStepOn) {
/*
		TimeToStop = TRUE;
*/
		HandleControlMenu(StopContinueItem);
		SingleStepOn = FALSE;
	}

	if (ShouldAdvance)
		AdvanceIterationStatus();
	return(!TimeToStop);
}

HandleParamFiles()
{
	if (ProgramStatus == OPENING)
		return;

	if ((PrintingOnly) && (OpenedParamFiles > 0) ) {
		StdForceUpdate(FrontWindow());
		PrintFrontWindow();
	}
	for(;;) {
		OpenWindex = GetNextParamFile();
		if (OpenWindex == NoWindow) {
			if (PrintingOnly)
				_HALT();
			else {
				GettingParamFiles = FALSE;
				theEventMask = UsualEventMask = everyEvent;
				return;
			}
		}
		switch (OpenWindex) {
			case TxOut:
					if ((GotATextFile) && (!PrintingOnly))	return;
					SelectWindowIndex(OpenWindex);
					if (StartOpeningFile()) {
						GotATextFile = TRUE;
						++OpenedParamFiles;
						IterationStatus = ITERATING;
						return;
					}
					break;
			case BifD:
					if ((GotABDFile) && (!PrintingOnly))	return;
					SelectWindowIndex(OpenWindex);
					if (OpenFile())
						if (OpenSpecialFile(OpenWindex)) {
							GotABDFile = TRUE;
							++OpenedParamFiles;
							HandleAllUpdates();
							return;
						}
					break;
			case BAttr:
					if ((GotABAFile) && (!PrintingOnly))	return;
					SelectWindowIndex(OpenWindex);
					if (OpenFile())
						if (OpenSpecialFile(OpenWindex)) {
							GotABAFile = TRUE;
							++OpenedParamFiles;
							HandleAllUpdates();
							return;
						}
					break;
		}
	}

/*		***	OLD VERSION, pre-Special	***
	if (PrintingOnly) {
		if (ProgramStatus != OPENING) {
			SelectWindowIndex(TextOutput);
			StdForceUpdate(FrontWindow());
			PrintFrontWindow();
			if (GetNextParamFile() != NoWindow) {
				if (StartOpeningFile())
					IterationStatus = ITERATING;
			}
			else
				_HALT();
		}
	}
*/
}

_MAIN()
{
	Boolean	TimeToStop = FALSE, OKCont = FALSE;

	if ( (!InForeground) && (!RunInBackground) ) return;

	if (GettingParamFiles)
		HandleParamFiles();

	if (PAUSING) return;
	if (ProgramStatus == WAITING) return;

	switch (IterationStatus) {
		case ITERATING:
						if (!Handle_ITERATING(ProgramStatus))
							return;
						if (!InForeground) {
							NeedToFinishIteration = TRUE;
							return;
						}
						/*	normally don't returnwhen get solution,
						 *	always do full cycle */
		case GRAPHING:
						Handle_GRAPHING(ProgramStatus);
						if (!InForeground) return;
		case WRITING:
						TimeToStop = !Handle_WRITING(ProgramStatus, &OKCont);
						if (!InForeground) return;
		case META_ITERATING:
						if (!Handle_META_ITERATING(ProgramStatus, &OKCont))
							TimeToStop = TRUE;
					/*		This due to changes in InitNextIteration
						if ((StuckStatus == STUCK_FOREVER) && (ProgramStatus == NORMAL_ITERATION))
							TimeToStop = TRUE;
					*/
						if (TimeToStop)
							{	StopIterating(OKCont);	return;	}
						if (StuckStatus == STUCK)
							_MAIN();
						if (!InForeground)
							NeedToFinishIteration = FALSE;
						break;
	}
}


	/*	--------------------------------------------------	*/
	/*	----------	Event-Handling Routines	----------	*/
	/*	--------------------------------------------------	*/

HandleContent(WindowPtr wp, Point pt, int mods)
{
	int		windex;
	int		WhichHotSpot;

	if ( (windex=WindowIndex(wp)) == NoWindow) return;

	SetPort(wp);		GlobalToLocal(&pt);

	if( HandleControl(windex, pt, &WhichHotSpot) ) {
		if (WhichHotSpot != NO_HotSpot)
			switch (windex) {
				case Trajectory:
								if (WhichHotSpot == 0) {
									if (GetInitialConditions(TRUE))
										BeginIfAppropriate();
								}
								else if (WhichHotSpot == 1)
									GetTrajScalingParams();
								break;
				case TextOutput:
								if (GetTableParameters())
									BeginIfAppropriate();
								break;
				case ImpactMap:
								if (WhichHotSpot == 0)
									GetPhaseRange();
								else if (WhichHotSpot == 1)
									GetNewVMax();
								break;
				case Animation:
								if (WhichHotSpot == 0) {
									if (GetInitialConditions(TRUE))
										BeginIfAppropriate();
								}
								else if (WhichHotSpot == 1)
									if (GetTableParameters())
										BeginIfAppropriate();
								break;
				case ImpactPhases:
								if (WhichHotSpot == 0)
									GetPhaseRange();
								else if (WhichHotSpot == 1)
									GetImpactPhasesOffset();
								break;
				case PhaseSpace:
								if (WhichHotSpot == 0)
									GetNewVMax();
								else if (WhichHotSpot == 1)
									GetTrajScalingParams();
								break;
				case TimeSeries:
								GetTimeSeriesInfo();
								break;
				case BifDiagram:
				case BasinsAttraction:
								HandleSpecialHotSpot(windex, WhichHotSpot);
								break;
			}
		return;
	}

	switch ( FindBBClick(windex, pt) ) {
		case TextOutput:
		case TimeSeries:
						HandleTextSelection(windex, pt, mods);
						break;
		case ImpactMap:
						if ( TrackIMapRect(pt, &nV0, &nDelta) )
							GetIMMagnification(SHIFT_DOWN);
						else
							if (GetInitialConditions(FALSE))
								BeginIfAppropriate();
						break;
		case ImpactPhases:
						if ( TrackIPhsRect(pt) )
							GetIPMagnification(SHIFT_DOWN);
						break;
		case BifDiagram:
						if ( TrackBDRect(pt) )
							GetBifDiagInfo(TRUE);
						else
							GetBifDiagInfo(FALSE);
						break;
		case BasinsAttraction:
						if ( TrackBARect(pt) )
							GetBasAttrInfo(TRUE);
						else
							GetBasAttrInfo(FALSE);
						break;
	}

}

HandleMouseDown(EventRecord *pev)
{
	WindowPtr	WhichWindow;
	int			WindowPart;

	WindowPart = FindWindow(pev->where, &WhichWindow);
	if (WindowIndex(WhichWindow) != NoWindow)
		if (WhichWindow != FrontWindow())
			if (WindowPart != inDrag)
				{	SelectWindow(WhichWindow);	return;	}
	switch (WindowPart) {
		case inSysWindow:
						SystemClick(pev, WhichWindow);
						break;
		case inMenuBar:
						AdjustBBMenus();
						HandleMenu(MenuSelect(pev->where));
						break;
		case inGrow:
						BBGrowWindow(pev->where, WhichWindow);
						break;
		case inDrag:
						BBDragWindow(pev->where, WhichWindow, pev->modifiers);
						break;
		case inGoAway:
						BBGoAwayWindow(pev->where, WhichWindow);
						break;
		case inZoomIn:
		case inZoomOut:
						BBZoomWindow(pev->where, WhichWindow, WindowPart);
						break;
		case inDesk:
						break;
 		case inContent:
 						HandleContent(WhichWindow, pev->where, pev->modifiers);
 						break;
	}
}

CheckVars_HandleUpdate(WindowPtr wp)
/*	Sets pCurrentVars to appropriate value for given window.
 *	By the way, since for BifD and BAttr, all updating is done by
 *	copying offports, the WindowIndex test is not really necessary
 */
{
	pIterationVars	piv = pCurrentVars;

	if ( (ProgramStatus == BIFURCATION_DIAGRAM) && !*pBDGraphInAllWindows )
		if (WindowIndex(wp) != BifDiagram)
			pCurrentVars = &NormalVars;

	if ( (ProgramStatus == BASINS_OF_ATTRACTION) && !*pBAGraphInAllWindows )
		if (WindowIndex(wp) != BasinsAttraction)
			pCurrentVars = &NormalVars;

	HandleUpdate(wp, HitNumber);
	pCurrentVars = piv;
}

HandleApp4Event(EventRecord *pEvent)
{
	WindowPtr	wp;
/*
int	appMessage, appCount,i;
*/

	if (App4Selector(pEvent) == SUSPEND_RESUME_SELECTOR) {
		wp = FrontWindow();
		if (SuspResIsSuspend(pEvent->message)) {
			InForeground = FALSE;
			NeedToFinishIteration = FALSE;
			if (WindowIndex(wp) != NoWindow)
				BBHandleActivate(wp, FALSE);
		}
		else {
/*
CountAppFiles(&appMessage, &appCount);
Error2Longs((long) appMessage, (long) appCount);
Error2Longs((long) pEvent->what, pEvent->message);
Error2Longs(pEvent->when, (long) pEvent->modifiers);
(* NOTE: When the application is resumed by trying to Finder-open a
document, pEvent->message has the "Clipboard Changed" bit, 0x00000002, set.
	-TA 1/3/91
*)
*/
			InForeground = TRUE;
			SetCursor(&arrow);
			if (WindowIndex(wp) != NoWindow)
				BBHandleActivate(wp, TRUE);
			if (NeedToFinishIteration)
				_MAIN();
		}
	}
}

_EVENT(EventRecord *theEvent)
{
	switch (theEvent->what) {
		case mouseDown:
			CurrentModifiers = theEvent->modifiers;
			HandleMouseDown(theEvent);
			break;
		case keyDown:
			CurrentModifiers = theEvent->modifiers;
			HandleKeyDown(theEvent->message, theEvent->modifiers);
			break;
		case updateEvt:
			CheckVars_HandleUpdate((WindowPtr) theEvent->message);
/*	was
			CheckVars_HandleUpdate((WindowPtr) theEvent->message,
					IterationStatus >= ITERATING);
*/
			break;
		case diskEvt:
			if (StdHandleDiskInsertion(theEvent->message) != noErr)
				;
			break;
		case activateEvt:
			if (WindowIndex((WindowPtr) theEvent->message) == NoWindow)
				return;
			else BBHandleActivate((WindowPtr) theEvent->message, theEvent->modifiers & activeFlag);
			break;
		case app4Evt:
			HandleApp4Event(theEvent);
			break;
	}
}

Boolean NextEvent(int eventMask, EventRecord *theEvent)
{
	int	MFSleep;

	if (HasWaitNextEvent) {
		if ( PAUSING || (ProgramStatus == WAITING) )
			MFSleep = ALL_NIGHT;
		else if ( !InForeground )
			MFSleep = (RunInBackground) ? BackgroundSleep : ALL_NIGHT;
		else
			MFSleep = DAY_DREAM;
		return( WaitNextEvent(eventMask, theEvent, MFSleep, NIL) );
	}
	else
		return( GetNextEvent(eventMask, theEvent) );
}

HandleAllActivates()
{
	EventRecord	theEvent;

	for(;;)
		if (GetNextEvent(activMask, &theEvent)) _EVENT(&theEvent);
		else break;
}

HandleAllUpdates()
{
	EventRecord	theEvent;

	for(;;)
		if (GetNextEvent(updateMask, &theEvent)) _EVENT(&theEvent);
		else break;
}

HandleEvent()
{
	Boolean		ok;
	EventRecord	theEvent;

	HiliteMenu(0);
	if (InForeground) {
		if (UserWaiting)	SetCursor(*watch);
		else				SetCursor(&arrow);
	}
	if (!HasWaitNextEvent)
		SystemTask ();		/* Handle desk accessories */
	
	if (NextEvent(theEventMask, &theEvent))
		_EVENT(&theEvent);
	else
		_MAIN();
}

	/*	------------------------------------------	*/
	/*	----------	Main Loop; Quitting	----------	*/
	/*	------------------------------------------	*/

main()
{
	InitMacintosh();
	InitPrinting();
	_INIT();
	
	for (;;)
		HandleEvent();
}

_HALT()
{
	ResetSolutions();
	HaltGraphics();
	HaltPrinting();
	DIUnload();
	exit(1);
}

