/*	bb.h
 *
 *	Bouncing Ball constants, data structures
 */

#include <Color.h>		/*	Currently for bbWindows.c only -TA 10/29/90	*/
#include <GeneralMacros.h>
#include "bbIteration.h"

#define	GetMainDeviceTrapID	0xAA2A	/*	Used in bb.c, bbWindows.c	*/

#define AssumedZero			1e-20
#define	AlmostZero			1e-10

#define	G					981.0	/* Acceleration of Gravity, CGS */

#define	CP(s)				((char *) &(s))
#define	RETURN_IF_FALSE(a)	if (!(a)) return(FALSE)

#define	SHIFT_DOWN			((CurrentModifiers & shiftKey) > 0)
#define	OPTION_DOWN			((CurrentModifiers & optionKey) > 0)

#define	DEFAULT_V0				10.0
#define	DEFAULT_DELTA			 0.0
#define	DEFAULT_A				 0.01
#define	DEFAULT_FREQ			60.0
#define	DEFAULT_DAMP			 0.5
#define	DEFAULT_CYCLES			10000L
#define	DEFAULT_ACCURACY		5
#define	DEFAULT_PERIOD			(1.0/DEFAULT_FREQ)
#define	NON_ZERO_A				( (A==0.0) ? DEFAULT_A : A )
#define	MIN_SPECIAL_ACCURACY	10

			/*	Trajectory window scaling defaults	*/
#define	DefaultNumPeriods		16	/* Number of table cycles seen on the screen */
#define	DefaultNumAmplitudes	20	/* Number of vertical amplitudes for scaling */

#define	DEFAULT_TS_SAMPLERATE	8

	/* MinMemory: 64K + 4K "buffer"--needed, I believe, for TextEdit */
#define	MinMemory			68 * 1024L
#define	MinMemToStart		128 * 1024L		/* at least 128K to start bb */

#define	SolutionIncrement	512
#define	MaxSolutions		32767
#define MaxSolutionArrays	MaxSolutions/SolutionIncrement + 1	/* 64 */
#define	SI1(i)				((i)/SolutionIncrement)
#define	SI2(i)				((i)%SolutionIncrement)

	/* For "Body Overflow," bbText.c (and bbFiles.c) */
#define	BodyOFIncrement		32768L
#define	OVERFLOW_MSG_LINES	4
#define	CharsPerSolution	40L		/* really 38 */
#define	MaxBodyOFs			(MaxSolutions * CharsPerSolution / BodyOFIncrement + 1)

#define	NO_ANSWER			-1

		/*	Range for phase ints; for bbGraphics, bbDialog	*/
#define	PHASE_INT_RANGE			100
#define	INIT_MinPhaseInt		0

		/*	Impact Phases Offset constants; for bbImpactGraphics, bbDialog	*/
#define	MIN_ImpactPhasesOffset		1
#define	MAX_ImpactPhasesOffset		1024
#define	DEFAULT_ImpactPhasesOffset	MIN_ImpactPhasesOffset

		/*	Delimiter Description Length, for bbFiles, bbDialog	*/
#define	DelimDescLength		32
#define	DelimStr(index)		CP(DelimiterStrings[index])

		/*	WaitNextEvent sleep values (in ticks), for MultiFinder */
#define	ALL_NIGHT	300
#define	CATNAP		10
#define	DAY_DREAM	0

#define	NO_HotSpot	-1


enum {
		NoEjection = -2,	STUCK_FOREVER	= NoEjection,
		NormalEjection,		STUCK			= NormalEjection,
		InAreaOfEjection,	NOT_STUCK		= InAreaOfEjection
};

		/*	"Program Statuses"	*/
enum {
		WAITING,
		NORMAL_ITERATION,
		OPENING,
		PLAYBACK,
		BIFURCATION_DIAGRAM,
		BASINS_OF_ATTRACTION
};

		/*	"Iteration Statuses"	*/
enum {
		NOT_ITERATING = -1,
		ITERATING,
		GRAPHING,
		WRITING,
		META_ITERATING,
		IterationStatuses /* sp? */
};

		/*	Windows	*/
enum {
		NoWindow = -1,
		Trajectory,			Traj = Trajectory,
		TextOutput,			TxOut = TextOutput,
		ImpactMap,			IMap = ImpactMap,
		Animation,			Anima = Animation,
		NumDefaultWindows,

		PhaseSpace = NumDefaultWindows,
					 		PSpace = PhaseSpace,
		ImpactPhases,		IPhs = ImpactPhases,
		TimeSeries,			TSer = TimeSeries,
/*	To add back in, also modify STR# resource #128
		PhaseSpace3D,		PS3D = PhaseSpace3D,
*/
		NumNormalWindows,

		BifDiagram = NumNormalWindows,
							BifD = BifDiagram,
		BasinsAttraction,	BAttr = BasinsAttraction,
/*	See PhaseSpace3D note above
		ImpactIteration,	ImpI = ImpactIteration,
*/
		NumWindows
};

		/*	Returned by GetNextSolution() (bbIteration.c)	*/
enum {
		BAD_SOLUTION = -1,
		NO_SOLUTION_YET,
		FOUND_SOLUTION,
		ROOT_FINDING_ERROR
};

		/*	Used in GetBBTEHandle (bbText.c), also called in bbPrint.c	*/
enum {
		HEADER_TEXT,
		BODY_TEXT
};

		/*	Used in bbText.c, also called in bbFiles.c	*/
enum {
		NO_TEXT_INDEX = -1,
		TxOutTextIndex,
		TSerTextIndex,
		NumTexts
};

		/*	Delimiter Indicesused in bbFiles.c, bbDialog.c	*/
enum {
		FIRST_Delim,
		SPACE_Delim = FIRST_Delim,
		TAB_Delim,
		COMMA_Delim,
		CR_Delim,
		LAST_Delim,
		NUM_Delims = LAST_Delim
};

enum {
		GET_PHASE,
		GET_VELOCITY,
		GET_TIME
};

enum {						/*	Special "solution types"	*/
	NoSolutionType = -1,
	StuckSolution,
	Period1Solution,
	Period2Solution,
	Period3Solution,
	Period4Solution,
	Period8Solution,
	OtherSolution,
	UnphysicalSolution,
	NumSolutionTypes
};

typedef	char	myFileNameType[64];

typedef struct {
	double	vel, del, time;
	int		stuck;
}		Solution;

	/*	Use an array of handles because I can't seem to access more than
	 *	32767 bytes in an array of structures pointed to by one handle.
	 *	32767 bytes corresponds to about 1260 solutionsI was getting
	 *	a crash every time I went over 1024 solutions (the memory was
	 *	added in increments of 512 solutions).  Perhaps Think C's
	 *	maximum array size, static or dynamic, is 32767 bytes.
	 *	I could, however, access more than 32767 characters pointed to
	 *	by a handlerefer to BodyOverflow use in CheckBodyTESize() in
	 *	bbGraphics.c.  For this, I think the limit is 8 MB (per Inside Mac);
	 *	that's the maximum amount the NewHandle() command can allocate.
	 *		-TA 3/23/90
	 */

typedef	Solution			Solutions[SolutionIncrement];
typedef	Solutions			**SolutionsHandle;
typedef	SolutionsHandle		SolutionsHandles[MaxSolutionArrays];

typedef	Handle				BodyOFHandles[MaxBodyOFs];
typedef	struct {
			BodyOFHandles	hands;
			long			index, size;
			int				num;
			int				ind1, ind2;
		}	BodyOF;
typedef	struct {
			TEHandle	header;
			TEHandle	body;
			BodyOF		OF;
			int			NumLines, nthLine;
			int			OldNumLines;
			int			MaxLinesOnScreen;	/* Max. number of text lines on screen */
			char		*HeaderLine;
			GrafPtr		port;
		}	BBText, *BBTextPtr;

typedef	struct {				/*	Used in bbWindows, bbFiles	*/
			int		windex;
			Rect	wr;
		}	WindowLocation;

typedef	struct {
			int				num;
			WindowLocation	ws[];
		}	WindowLocs, **hWindowLocs;

	/*	For magnification	*/
enum {
	ymin_i,
	xmin_i,
	ymax_i,
	xmax_i,
	ItemsInRect
};
enum {
	top, left, bottom, right
};
typedef double	DoubleRect[ItemsInRect];	/*	Used in graphics, bbDialog	*/
typedef Boolean	BooleanRect[ItemsInRect];

											/*	Used in special windows	*/
typedef	int				SolutionInfo[NumSolutionTypes];
typedef	SolutionInfo	*pSolutionInfo;
