/*	bbBasAttr.c
 *
 *	Routines for Basins of Attraction
 *
 *		-TA 3/6/91
 */

#include <FileUtils.h>
#include <OffPort.h>
#include "bbPrototypes.h"
#include "bbGraphics.h"
#include "bbSpecial.h"

#define	BAOpenMemErrorString		"\PError: Not enough memory to open this basins of attraction file"
#define	SeriousBAOpenMemErrorString	"\PSerious memory error: cannot open new basins of attraction diagram and cannot restore old one"

				/*	Global external vars from bb.c	*/
extern	IterationVars	NormalVars, *pCurrentVars;
extern	hBBPreferences	hBBPrefs;

extern	int				ProgramStatus, IterationStatus;
extern	int				CurrentModifiers;
extern	Boolean			UsingColorMac;
extern	Boolean			OKToContinueBA;
extern	Boolean			GettingParamFiles;
extern	Boolean			SuppressErrorMessages;

				/*	From bbFiles.c	*/
extern	Boolean			LastParamFile;

				/*	From bbGraphics.c	*/
extern	Windows			*pw;

				/*	From bbWindows.c	*/
extern	int				MaxScreenY, MaxScreenX;
extern	Rect			LWR;

				/*	From bbImpactGraphics.c	*/
extern	double			MinPhase, MaxPhase, VMin;

				/*	From bbContGraphics.c	*/
extern	Boolean			RedrawSine;

				/*	From bbMenus.c	*/
extern	Boolean			NeedToSaveBA;

				/*	From bbSpecial.c	*/
extern	Boolean			NeedToTurnOnGraphing;
extern	Boolean			NeedNextSpecialValue;
extern	Pattern			SolutionPats[NumSolutionTypes];
extern	RGBColor		SolutionColors[NumSolutionTypes];

				/*	From bbText.c	*/
extern	Boolean			NeedToSetTxOutHeader, NeedToSetTSerHeader;

BasAttrRecord	BA_Rec, tmpBA;
pBasAttrRecord	pba;
Rect			BABox, BALatestImpactRect;
int				BAXFieldWidth, BAYFieldWidth;
Boolean			DrawBA;
Point			**BA_Points;
int				CurBASolutionType;
int				GuessBASolutionType, GuessBASolCount;
int				CurrentMaxTimes;
int				CMTimesInc, NumCMTimesIncs;
Str255			BA_Spaces;

						/*	For file operations	*/
		Point		**oBA_Points;
static	Boolean		GraphExists;
static	PicHandle	BackPic, GraphPix[NumSolutionTypes];
static	PicHandle	oGraphPix[NumBAOffPorts];




			/*	------------------------------------------	*/
			/*	----------	Utility Routines	----------	*/
			/*	------------------------------------------	*/

double GetBAOffsetPhase(double del)
{
	del -= BA_Rec.MinPhase;
	if (del>1.0+AlmostZero)		/*	used to be >=	*/
		del -= 1.0;
	if (del<0.0)
		del += 1.0;
	return(del);
}

double GetBAUnOffsetPhase(double del)
{
	if (BA_Rec.MinPhase != 0.0)
		del += BA_Rec.MinPhase;
	return(del);
}

int GetDefaultMaxTimes(double theDamping)
{
	return(RoundToInt(BaseMaxTimes * theDamping + AddedMaxTimes));
}

Boolean CreateBAOffPort(int WhichOP, Rect *pr, int StepSize)
{
	if (!CreateSpecialOffPort(&BA_Rec.Graphs[WhichOP], pr))
		return(FALSE);
	SetPort(BA_Rec.Graphs[WhichOP]);
	PenSize(StepSize, StepSize);
	return(TRUE);
}

Boolean CreateBAOffPorts(Rect *pr, int StepSize)
{
	int	i;

	for(i=0; i<NumBAOffPorts; ++i) {
		if ( (i <= StuckSolution) || (i >= UnphysicalSolution) ) {
			if (!CreateBAOffPort(i, pr, StepSize))
				return(FALSE);
		}
		else {
			if (BA_Rec.Graphs[i] != NOPORT)
				DisposeOffPort(BA_Rec.Graphs[i]);
			BA_Rec.Graphs[i] = NOPORT;
		}
	}
	return(TRUE);
}

ChangeBAPhaseRange(double NewMinPhase)
{
	int		PixelOffset;
	GrafPtr	tmpGP = BA_Rec.Graphs[ScratchBAOP];
	Rect	tr = tmpGP->portRect, tr2, tr3;
	int		i;

/*	Use ScratchBAOP instead
	GrafPtr	tmpGP = NOPORT;
	if (!CreateSpecialOffPort(&tmpGP, &tr)) {
		ERROR("\PNot enough memory to change Basins of Attraction phase range.");
		return;
	}
*/

	if (NewMinPhase > BA_Rec.MinPhase) {
		PixelOffset = xFloatToCoord(BAttr, GetBAOffsetPhase(NewMinPhase));
		NewMinPhase = GetBAUnOffsetPhase(xCoordToFloat(BAttr, PixelOffset));
	}
	else {
		PixelOffset = xFloatToCoord(BAttr, GetBAOffsetPhase(NewMinPhase+1.0));
		NewMinPhase = GetBAUnOffsetPhase(xCoordToFloat(BAttr, PixelOffset))
				- 1.0;
	}
	PixelOffset *= BA_Rec.StepSize;

	tr2 = tr;
	tr2.right = tr2.left + PixelOffset;
	tr3 = tr;
	tr3.left += PixelOffset;
	for (i=0; i < NumBAOffPorts; ++i) {
		if (BA_Rec.Graphs[i] == NOPORT)		continue;
		if (i == ScratchBAOP)				continue;
		myCopyBits(BA_Rec.Graphs[i], tmpGP, &tr2, 0, 0,
				(tr.right - tr.left - PixelOffset), 0, srcCopy, NIL);
		myCopyBits(BA_Rec.Graphs[i], tmpGP, &tr3, 0, 0, -PixelOffset, 0, srcCopy, NIL);
		myCopyBits(tmpGP, BA_Rec.Graphs[i], &tr, 0, 0, 0, 0, srcCopy, NIL);
	}

/*
	DisposeOffPort(tmpGP);
*/

		
	if (BA_Rec.CurrentYInt >= 0) {
		++BA_Rec.CurrentYInt;
		BA_Rec.CurrentXInt = BA_Rec.MaxXInt;
/*
		if (ProgramStatus == BASINS_OF_ATTRACTION)
			IterationStatus = META_ITERATING;
*/
	}

	if (fabs(NewMinPhase) < AlmostZero)
		NewMinPhase = 0.0;
	BA_Rec.MinPhase = NewMinPhase;
/*
	BA_Rec.YAxis[xmin_i] = NewMinPhase;
	BA_Rec.YAxis[xmax_i] = NewMinPhase + 1.0;
*/
	w[BAttr].DrawFromScratch = TRUE;
	StdForceUpdate((WindowPtr) w[BAttr].Peek);
	BAYFieldWidth = StdIMFieldWidth;
		/*	Since y-range is 1.0 (AllPhases); see InitBAScales;
		 *	this is only needed if BA is opened, not created;
		 *	DrawBABackground needs this, but nothing else does,
		 *	so normally it does not matter that BAYFieldWidth is not set.
		 */
}


			/*	------------------------------------------	*/
			/*	----------	Iteration Routines	----------	*/
			/*	------------------------------------------	*/

Boolean InitNextBA()
{
	Boolean	GoodInit;

	ResetSolutions();

	SuppressErrorMessages = TRUE;
	GoodInit = Initialize_Iteration_Globals(TRUE, FALSE);
	SuppressErrorMessages = FALSE;

	DrawCurrentBABackground();
	if (!GoodInit)
		return(FALSE);

	CMTimesInc = CurrentMaxTimes =
			GetSpecialTimes(BA_Rec.MaxTimes, BA_Rec.DefaultMaxTimes);
	NumCMTimesIncs = 1;

	if (NeedToTurnOnGraphing) {
		BA_Rec.GraphInAllWindows = TRUE;
		CheckGraphInAll(BAttr);
		SetTextTitle(TRUE, BAttr);
		ActivateTextScrolls(FALSE);
		NeedToTurnOnGraphing = FALSE;

		RedrawSine = TRUE;
		InitTrajScales();
		SetTrajCoords();
				/*	From bbText.c	*/
		NeedToSetTxOutHeader = NeedToSetTSerHeader = TRUE;
/*
		SetHeaders(A, freq, damp);
*/
	}
	if (BA_Rec.GraphInAllWindows) {
		RedrawSine = TRUE;
/*
		w[BAttr].WindowCleared = FALSE;
*/
		NewGraphics(TRUE, FALSE);
	}

	CurBASolutionType = NoSolutionType;
	GuessBASolutionType = NoSolutionType;
	GuessBASolCount = 0;
	NeedToSaveBA = TRUE;

	SetPort(BA_Rec.Graphs[CurIterBAOP]);
	EraseRect(&w[BAttr].GR);

	return(TRUE);
}

Boolean GetNextBA(Boolean *pShouldAdvance)
{
	*pShouldAdvance = TRUE;
	if (CurBASolutionType == NoSolutionType)
		return(TRUE);

	if (++BA_Rec.CurrentXInt > BA_Rec.MaxXInt) {
		if (--BA_Rec.CurrentYInt < 0)
			return(FALSE);
		else
			BA_Rec.CurrentXInt = 0;
		}

	if (BA_Rec.CurrentYInt == BA_Rec.MaxYInt)
		v0 = w[BAttr].GraphRange[ymax_i];
	else
		v0 = yCoordToFloat(BAttr, BA_Rec.CurrentYInt);
	if ( (BA_Rec.CurrentXInt == BA_Rec.MaxXInt) && (!BA_Rec.AllPhases) )
		delta = GetBAUnOffsetPhase(w[BAttr].GraphRange[xmax_i]);
	else
		delta = GetBAUnOffsetPhase(xCoordToFloat(BAttr, BA_Rec.CurrentXInt));
/*	Deleted because of possible phase range changes;
 *	even if not Faster, may have already found solution.	-TA 6/6/91
	if (BA_Rec.Faster)
*/
		if (FindPrevBASolution(BA_Rec.CurrentXInt,
				BA_Rec.CurrentYInt) != NoSolutionType) {
			DrawCurrentBABackground();
			*pShouldAdvance = FALSE;
			return(TRUE);
		}

	if (!InitNextBA()) {
		HandleAllUpdates();
		PlotInBA(BA_Rec.CurrentXInt, BA_Rec.CurrentYInt);
			/*	Can also "mark" all velocities lower than CurrentYInt's	*/
				/*	Code from PlotInBA()	*/
		SetPort(BA_Rec.Graphs[CurIterBAOP]);
/*
		BackMoveTo(BAttr, xCoord * BA_Rec.StepSize, yCoord * BA_Rec.StepSize);
*/
		Line(0, BA_Rec.CurrentYInt * BA_Rec.StepSize);

		SetBASolutionType(UnphysicalSolution);
		SetPort(BA_Rec.Graphs[CurIterBAOP]);
		EraseRect(&w[BAttr].GR);
		*pShouldAdvance = FALSE;
	}
	else
		IterationStatus = GRAPHING;		/*	will be advanced to WRITING	*/
	return(TRUE);
}

StartBasinsOfAttraction()
{
	Boolean	NeedVMax, OKToIterate;
	int		MaxTimes;

	if (pCurrentVars != &BA_Rec.Vars)
		if (HitNumber > 0) {
			if (!CheckNormalSaves(FALSE))
				return(FALSE);
/*
			if (pCurrentVars == &NormalVars)
				if (!DoSpecialWarningDialog("\Pbasins of attraction"))
					return(FALSE);
*/
			NewGraphics(TRUE, TRUE);
		}

	if (!WindowOpen(BAttr))
		if (DoShowWindowDialog(BAttr))
			BBOpenWindow((WindowPtr) w[BAttr].Peek);

	ResetSolutions();
	StopIterating(ProgramStatus == BIFURCATION_DIAGRAM);
	NeedVMax = damp != BA_Rec.Vars._damp;
/*
	if (!BA_Rec.GraphInAllWindows)	(*	was in "if HitNumber..." above	*)
		HandleAllUpdates();
*/

	pCurrentVars = &BA_Rec.Vars;
	SetFrequencyParams();
	if ( (BA_Rec.GraphInAllWindows) && (NeedVMax) ) {
		SetVMax(TRUE);
		InitImpactMapScales();
	}

	if (BA_Rec.GraphInAllWindows) {
		RedrawSine = TRUE;
		InitTrajScales();
		SetTrajCoords();
		NewGraphics(TRUE, FALSE);
		SetHeaders(A, freq, damp);
	}

	MaxTimes = ( (BA_Rec.MaxTimes != AutoTimes) ?
			BA_Rec.MaxTimes :
			(BA_Rec.DefaultMaxTimes * MaxCMTimesIncs) );
	if (BA_Points != NIL)
		DisposHandle((Handle) BA_Points);
	BA_Points = (Point**) NewHandle( ((long) MaxTimes+1) * sizeof(Point) );

	NewSpecial(BAttr);
	BA_Rec.CurrentXInt = -1;
	BA_Rec.CurrentYInt = BA_Rec.MaxYInt;
	NeedToTurnOnGraphing = FALSE;

	CurBASolutionType = 0;	/*	for GetNextBA()	*/
	if (!GetNextBA(&OKToIterate))
		return(FALSE);
	if (!StartIterating(BASINS_OF_ATTRACTION)) {
		SpecialMenuEnable(BAttr, FALSE, FALSE);
		return(FALSE);
	}
	SpecialMenuEnable(BAttr, TRUE, TRUE);

	if (!OKToIterate)
		IterationStatus = META_ITERATING;
	return(TRUE);
}

ContinueBasinsOfAttraction()
{
	Boolean	NeedVMax, OKToIterate;

	if (ProgramStatus < BIFURCATION_DIAGRAM)
		if (HitNumber > 0) {
/*
			if (!DoSpecialWarningDialog("\Pbasins of attraction"))
*/
			if (!CheckNormalSaves(FALSE))
				return(FALSE);
			NewGraphics(TRUE, TRUE);
		}

	if (!WindowOpen(BAttr))
		if (DoShowWindowDialog(BAttr))
			BBOpenWindow((WindowPtr) w[BAttr].Peek);

	ResetSolutions();
	StopIterating(ProgramStatus == BIFURCATION_DIAGRAM);
	NeedVMax = damp != BA_Rec.Vars._damp;
/*
	if (!BA_Rec.GraphInAllWindows)
		HandleAllUpdates();
*/

	pCurrentVars = &BA_Rec.Vars;
	SetFrequencyParams();

	if ( (BA_Rec.GraphInAllWindows) && (NeedVMax) ) {
		SetVMax(TRUE);
		InitImpactMapScales();
	}
	if (BA_Rec.GraphInAllWindows) {
		RedrawSine = TRUE;
		InitTrajScales();
		SetTrajCoords();
		NewGraphics(TRUE, FALSE);
		SetHeaders(A, freq, damp);
	}

	--BA_Rec.CurrentXInt;	/*	Re-incremented in GetNextGoodBA()	*/
	CurBASolutionType = 0;	/*	for GetNextBA()	*/
	if (!GetNextBA(&OKToIterate))
		return;
	if (!StartIterating(BASINS_OF_ATTRACTION))
		return;
	if (!OKToIterate)
		IterationStatus = META_ITERATING;
}

GetBasAttrInfo(Boolean Magnifying)
{
	Boolean		StartBA;
	DoubleRect	MagRange, NewGraphRange;
	int			MagXStart;
	int			i;
	Boolean		OldGIAW;
	double		min, max;
	Boolean		GoodOp;
	int			tmpCurMods;

	tmpBA = BA_Rec;

	if (Magnifying) {
		NewGraphRange[xmin_i] = xCoordToFloat(BAttr,
				(BABox.left - w[BAttr].GR.left) / BA_Rec.StepSize );
		NewGraphRange[xmax_i] = xCoordToFloat(BAttr,
				(BABox.right - w[BAttr].GR.left) / BA_Rec.StepSize );
		NewGraphRange[ymin_i] = yCoordToFloat(BAttr,
				(w[BAttr].GR.bottom - BABox.bottom) / BA_Rec.StepSize );
		NewGraphRange[ymax_i] = yCoordToFloat(BAttr,
				(w[BAttr].GR.bottom - BABox.top) / BA_Rec.StepSize );
		NewGraphRange[xmin_i] = GetBAUnOffsetPhase(NewGraphRange[xmin_i]);
		NewGraphRange[xmax_i] = GetBAUnOffsetPhase(NewGraphRange[xmax_i]);

		Round2Doubles(NewGraphRange[xmin_i], NewGraphRange[xmax_i], RelPlaces, &min, &max);
		if (NewGraphRange[xmin_i] > BA_Rec.MinPhase)
			NewGraphRange[xmin_i] = min;
		if (NewGraphRange[xmax_i] < BA_Rec.MinPhase + w[BAttr].GraphRange[xmax_i])
			NewGraphRange[xmax_i] = max;
		Round2Doubles(NewGraphRange[ymin_i], NewGraphRange[ymax_i], RelPlaces, &min, &max);
		if (NewGraphRange[ymin_i] > w[BAttr].GraphRange[ymin_i])
			NewGraphRange[ymin_i] = min;
		if (NewGraphRange[ymax_i] < w[BAttr].GraphRange[ymax_i])
			NewGraphRange[ymax_i] = max;
	}
	else if (!DrawBA) {
		tmpBA.Vars = NormalVars;
		NewGraphRange[xmin_i] = GetUnOffsetPhase(w[IMap].GraphRange[xmin_i]);
		NewGraphRange[xmax_i] = GetUnOffsetPhase(w[IMap].GraphRange[xmax_i]);
		NewGraphRange[ymin_i] = w[IMap].GraphRange[ymin_i];
		NewGraphRange[ymax_i] = w[IMap].GraphRange[ymax_i];
		tmpBA.StepSize = DEFAULT_StepSize;
		tmpBA.DefaultMaxTimes = GetDefaultMaxTimes(tmpBA.Vars._damp);
		tmpBA.MaxTimes = AutoTimes;
		tmpBA.Faster = TRUE;
		tmpBA.GraphInAllWindows = BA_Rec.GraphInAllWindows;
		tmpBA.FullWindow = TRUE;
		tmpBA.InColor = BA_Rec.InColor;
	}
	else {
		NewGraphRange[xmin_i] = GetBAUnOffsetPhase(w[BAttr].GraphRange[xmin_i]);
		NewGraphRange[xmax_i] = GetBAUnOffsetPhase(w[BAttr].GraphRange[xmax_i]);
		NewGraphRange[ymin_i] = w[BAttr].GraphRange[ymin_i];
		NewGraphRange[ymax_i] = w[BAttr].GraphRange[ymax_i];
	}

	pba = &tmpBA;
	
	if ( !Magnifying || !SHIFT_DOWN)
		GoodOp = DoBasinsOfAttractionDialog(&StartBA, &NewGraphRange);
	else
		GoodOp = TRUE;
	if (GoodOp) {
		if (Magnifying) {
			tmpCurMods = CurrentModifiers;
			CurrentModifiers = 0;
		}
		if (!CheckOneSave(FALSE, BAttr))
			GoodOp = FALSE;
		else
			NeedToSaveBA = FALSE;
		if (Magnifying)
			CurrentModifiers = tmpCurMods;
	}

	if (GoodOp) {
		DrawBA = TRUE;
		OldGIAW = BA_Rec.GraphInAllWindows;
		BA_Rec = tmpBA;

		for(i=0; i<ItemsInRect; ++i)
			w[BAttr].GraphRange[i] = NewGraphRange[i];
		BA_Rec.MinPhase = NewGraphRange[xmin_i];
		w[BAttr].GraphRange[xmin_i] = 0.0;
		w[BAttr].GraphRange[xmax_i] = GetBAOffsetPhase(NewGraphRange[xmax_i]);
		BA_Rec.AllPhases = (w[BAttr].GraphRange[xmax_i] -
				w[BAttr].GraphRange[xmin_i] >= 1.0-AlmostZero);

		if (BA_Rec.FullWindow)
			BA_Rec.WindowSize = LWR;
		else {
			BA_Rec.WindowSize = w[BAttr].Peek->port.portRect;
			OffsetRect(&BA_Rec.WindowSize, -w[BAttr].Peek->port.portBits.bounds.left,
					-w[BAttr].Peek->port.portBits.bounds.top);
			/*	see GetDefaultWindowConfiguration discussion, bbWindows.c	*/
		}
		SetSpecialScrolling(BAttr);

		w[BAttr].xGraphAroundZero =
				(w[BAttr].GraphRange[xmin_i] * w[BAttr].GraphRange[xmax_i] <= 0);
		w[BAttr].yGraphAroundZero = FALSE;
/*
		w[BAttr].yGraphAroundZero =
				(w[BAttr].GraphRange[ymin_i] * w[BAttr].GraphRange[ymax_i] <= 0);
*/
		w[BAttr].Magnifying = TRUE;
		w[BAttr].DrawFromScratch = TRUE;

		if (ProgramStatus == BASINS_OF_ATTRACTION) {
			BA_Rec.GraphInAllWindows = OldGIAW;
			if (OldGIAW != tmpBA.GraphInAllWindows) {
				if (tmpBA.GraphInAllWindows)
					ProgramStatus = WAITING;
				SwitchGraphInAll(BAttr);
				ProgramStatus = BASINS_OF_ATTRACTION;
			}
		}
		else
			CheckGraphInAll(BAttr);
		CheckSpecialColor(BAttr);

		SetSpecialTitle(FALSE, BAttr);
		BA_Rec.CurrentXInt = BA_Rec.CurrentYInt = -1;
		if (!InitBAScales()) {
			ERROR("\PNot enough memory to run basins of attraction");
			ClearBasinsOfAttraction();
			return;
		}
		SpecialMenuEnable(BAttr, TRUE, FALSE);
		if (StartBA)
			StartBasinsOfAttraction();
		else {
			NewSpecial(BAttr);
			if (ProgramStatus == BASINS_OF_ATTRACTION)
				StopIterating(FALSE);
		}
	}

}

Boolean	TrackBARect(Point pt)
{
	SetPort(w[BAttr].Peek);

	if ( TrackDashedRect(pt.h, pt.v, &BA_Rec.LocalGR, &BABox, TRUE) ) {
		OffsetRect(&BABox,GetCtlValue(w[BAttr].BSB),GetCtlValue(w[BAttr].RSB));

			/*
			 *		The next step is something of a kludge.
			 *	Its necessity derives from the assymetry of
			 *	Mac Rect drawing--because the rect is framed on the
			 *	inside, on the left/top sides, the rect is drawn
			 *	to the right of/below the rect, while on the right/
			 *	bottom sides, the rect is drawn to the left/above.
			 *	Lines or points are drawn to the right/below.
			 *		This problem could perhaps also be solved by
			 *	figuring coordinates ([xy]FloatToCoord()) by truncating
			 *	instead of rounding, but (a) that has problems itself in,
			 *	say, the Bifurcation Diagram, where each x-axis
			 *	pixel corresponds to a single value of the amplitude
			 *	or frequency, instead of a range of values, as it
			 *	would elsewhere, and (b) at this point it would be
			 *	a great deal of work to fix.
			 *		Note that the background frame drawn (DrawBABackground())
			 *	and BA_Rec.LocalGR (SetSpecialControls())
			 *	include corresponding kludges.
			 *		These kludges are not present in similar situations
			 *	of the Impact Map and Impact Phases windows, because
			 *	these windows behave a little differently--in particular,
			 *	the user does not expect a value to be plotted, say,
			 *	from 0.01 to 0.02.
			 *												-TA 3/1/91
			 */
		if (BABox.right > BABox.left)
			if ( !BA_Rec.AllPhases || (BABox.right < w[BAttr].GR.right) )
				--BABox.right;
		if (BABox.top < BABox.bottom)
			++BABox.top;
/*	For BasAttr, lower bottom (velocity) but not right (phase)	*/

		return(TRUE);
	}
	return(FALSE);
}

			/*	------------------------------------------	*/
			/*	----------	Graphics Routines	----------	*/
			/*	------------------------------------------	*/

Boolean InitBAScales()
{
	Rect	tmpRect;
	double	range;
	int		i;
	int		ScaledLength;
	int		vOffset;

	SetPort(w[BAttr].Peek);
	w[BAttr].DrawFromScratch = TRUE;
	StdForceUpdate((GrafPtr) w[BAttr].Peek);

		/*	Calculate how much space needed for axis labels	*/
	if (w[BAttr].Magnifying) {
		range = w[BAttr].GraphRange[ymax_i]-w[BAttr].GraphRange[ymin_i];
		for(i=-1;i<2;++i) {
			if (range > 1.0) break;
			range *= 10;
		}
		BAYFieldWidth = StdIMFieldWidth + MAX(i,0);

		range = w[BAttr].GraphRange[xmax_i]-w[BAttr].GraphRange[xmin_i];
		for(i=-2;i<4;++i) {
			if (range > 1.0) break;
			range *= 10;
		}
		BAXFieldWidth = StdIMFieldWidth + MAX(i,0);
	}
	else
		BAYFieldWidth = BAXFieldWidth = StdIMFieldWidth;

	tmpRect = BA_Rec.WindowSize;
	OffsetRect(&tmpRect, -tmpRect.left, -tmpRect.top);
	tmpRect.left += (StdMargin + (BAYFieldWidth+3)*CharWidth('0'));
	tmpRect.top	+= IM_TopMargin;
	tmpRect.right -= (ScrollBarWidth + StdMargin +
			BAXFieldWidth * CharWidth('0') / 2);
	tmpRect.bottom -= BA_BottomMargin;
	tmpRect.right -= (tmpRect.right - tmpRect.left) % BA_Rec.StepSize;
	vOffset = (tmpRect.bottom - tmpRect.top) % BA_Rec.StepSize;
	tmpRect.top += vOffset/2;
	tmpRect.bottom -= (vOffset/2 + vOffset%2);
	w[BAttr].GR = tmpRect;

	if (!CreateBAOffPorts(&w[BAttr].GR, BA_Rec.StepSize))
		return(FALSE);

	BA_Rec.MaxXInt = (tmpRect.right - tmpRect.left) / BA_Rec.StepSize;
	ScaledLength = BA_Rec.MaxXInt;
	if (!BA_Rec.AllPhases)	/*	Magnifying	*/
		--ScaledLength;
	--BA_Rec.MaxXInt;
	w[BAttr].xScale = ScaledLength /
			(w[BAttr].GraphRange[xmax_i] - w[BAttr].GraphRange[xmin_i]);
				/*	Last part 1.0 if no magnification	*/

	BA_Rec.MaxYInt = (tmpRect.bottom-tmpRect.top) / BA_Rec.StepSize;
	--BA_Rec.MaxYInt;
	w[BAttr].yScale = BA_Rec.MaxYInt /
			(w[BAttr].GraphRange[ymax_i] - w[BAttr].GraphRange[ymin_i]);

	SetRect(&BALatestImpactRect, 0, 0, BA_Rec.StepSize+2, BA_Rec.StepSize+2);

	if (w[BAttr].yGraphAroundZero)
		w[BAttr].Origin.v = w[BAttr].GR.top +
			RoundToInt(w[BAttr].GraphRange[ymax_i]*BA_Rec.StepSize*w[BAttr].yScale);
	else
		w[BAttr].Origin.v = w[BAttr].GR.bottom - BA_Rec.StepSize;
	w[BAttr].Origin.h = w[BAttr].GR.left;
	w[BAttr].BackOrigin = w[BAttr].Origin;
	SetRect(&w[BAttr].PR,
			w[BAttr].GR.left - w[BAttr].Origin.h,
			w[BAttr].Origin.v - w[BAttr].GR.bottom,
			w[BAttr].GR.right - w[BAttr].Origin.h + 1,
			w[BAttr].Origin.v - w[BAttr].GR.top + 1);
			/* +1 to account for pixel hanging below and to left of point */

	tmpRect.top = tmpRect.bottom;
	tmpRect.bottom = BA_Rec.WindowSize.bottom - BA_Rec.WindowSize.top - (ScrollBarWidth-1);
	InsetRect(&tmpRect, -1, 0);
	tmpRect.right -= 1;
	BA_Rec.HS0_Rect = tmpRect;

	tmpRect = w[BAttr].GR;
	tmpRect.right = tmpRect.left-1;
	tmpRect.left = 0;
	InsetRect(&tmpRect, 0, -1);
	tmpRect.bottom -= 1;
	BA_Rec.HS1_Rect = tmpRect;

	SetSpecialControls(BAttr);
	return(TRUE);
}

DrawCurrentBABackground()
{
	Str255	s;
	Rect	tmpRect, theClipRect;
	int		theOffset;

	s[0] = 0;

	SetPort(w[BAttr].Back);
	StringAddVar(CP(s), BA_Rec.Vars._v0, "\PV0: ", TRUE);
	 sAppend(CP(s), CP(BA_Spaces));
	 sAppend(CP(s), CP(BA_Spaces));
	 sAppend(CP(s), CP(BA_Spaces));
	StringAddVar(CP(s), BA_Rec.Vars._delta, "\PDelta: ", TRUE);

	SetRect(&tmpRect, StdMargin,
			w[BAttr].GR.bottom + 5*TextLineHeight + StdMargin,
			w[BAttr].Back->portRect.right, w[BAttr].Back->portRect.bottom);
	theOffset = BA_Rec.WindowSize.right - BA_Rec.WindowSize.left -
			ScrollBarWidth - 2*StdMargin - StringWidth(CP(s));
	if (theOffset > 0)
		OffsetRect(&tmpRect, theOffset/2, 0);

	TextBox(CP(s[1]), (long) (s[0]), &tmpRect, teJustLeft);

	SetPort((WindowPtr) w[BAttr].Peek);
	OffsetRect(&tmpRect, -GetCtlValue(w[BAttr].BSB),
			-GetCtlValue(w[BAttr].RSB));
	GetStdDocWindowClipRect(&theClipRect, (WindowPtr) w[BAttr].Peek);
	if (SectRect(&tmpRect, &theClipRect, &tmpRect))
		InvalRect(&tmpRect);
}

DrawBABackground()
{
	Rect	tmpRect;
	int		i, x, y;
	Str255	s, s1;
	Str255	LocalXMaxString, LocalXMinString;
	Str255	LocalYMaxString, LocalYMinString;
	int		MaxXStrLen, OverflowOnRight;
	int		BottomTextSize, BottomTextOffset;

/*		(Don't make hash marks)
	int		MinXHashLabelCoord, MaxXHashLabelCoord;
	double	ScaledMax, Multiplier, HashMeasurement;
	int		ScaledMaxInt, HashInc, HashIndex;
	int		XWriteCoord, XWriteStart, XWriteStop;
*/


	tmpRect = w[BAttr].GR;
	InsetRect(&tmpRect, -1, -1);
/*		This statement used with Bifurcation Diagram
	tmpRect.right += 1;
*/
/*
	tmpRect.bottom += 1;
*/
	FrameRect(&tmpRect);

	/* Document vertical axis */

			/*	not enough that around zero--around but not 0 endpoint	*/
		if (w[BAttr].GraphRange[ymin_i] * w[BAttr].GraphRange[ymax_i] < 0) {


			y = (w[BAttr].GR.bottom - w[BAttr].GR.top - BA_Rec.StepSize);
			y = (int) (y / (w[BAttr].GraphRange[ymax_i] -
					w[BAttr].GraphRange[ymin_i]) *
					(-w[BAttr].GraphRange[ymin_i]));
/*
			y = yFloatToCoord(BAttr, 0.0) * BA_Rec.StepSize;
*/
/*
			WindowMoveTo(BAttr, 0, 0);
*/
/*
			y += w[BAttr].Origin.v - w[BAttr].GR.bottom;
*/
			WindowMoveTo(BAttr, w[BAttr].GR.left-w[BAttr].Origin.h, y);
			Line(w[BAttr].GR.right - w[BAttr].GR.left, 0);
			if ( (w[BAttr].Origin.v - y > w[BAttr].GR.top + TextLineHeight) &&
					(w[BAttr].Origin.v - y < w[BAttr].GR.bottom - TextLineHeight) ) {
				WindowMoveTo(BAttr, w[BAttr].GR.left-w[BAttr].Origin.h, y);
				Move(-TextMargin - CharWidth('0'), TextHeight/2);
				DrawChar('0');
			}
		}

		sEquate(CP(s), "\PVelocity");
		SetRect(&tmpRect, StdMargin+2,
				(w[BAttr].GR.top+w[BAttr].GR.bottom-((int)s[0])*TextLineHeight)/2,
				StdMargin + 12/*CharWidth('V') + 1*/,
				w[BAttr].GR.bottom + 2*TextLineHeight + StdMargin + 1);
		tmpRect.top = MAX(tmpRect.top, w[BAttr].GR.top);
		TextBox(CP(s[1]), (long) (s[0]), &tmpRect, teJustCenter);

		myFToStrNoPad(w[BAttr].GraphRange[ymin_i], CP(LocalYMinString),
				BAYFieldWidth, BAYFieldWidth-IMPrecDiff);
		myFToStrNoPad(w[BAttr].GraphRange[ymax_i], CP(LocalYMaxString),
				BAYFieldWidth, BAYFieldWidth-IMPrecDiff);
		WindowMoveTo(BAttr, 0, w[BAttr].Origin.v - w[BAttr].GR.top);
			Move(-TextMargin - StringWidth(&LocalYMaxString), TextHeight/2);
			DrawString(&LocalYMaxString);
		WindowMoveTo(BAttr, 0, w[BAttr].Origin.v - w[BAttr].GR.bottom);
			Move(-TextMargin - StringWidth(&LocalYMinString), TextHeight/2-1);
			DrawString(&LocalYMinString);

	/* Document horizontal axis */

		sEquate(CP(s), "\PPhase");
		MoveTo( (w[BAttr].GR.right + w[BAttr].GR.left)/2, w[BAttr].GR.bottom);
		Move(-StringWidth(CP(s))/2, TextMargin+TextLineHeight+TextHeight-1);
		DrawString(CP(s));

		myFToStrNoPad(GetBAUnOffsetPhase(w[BAttr].GraphRange[xmin_i]),
				CP(LocalXMinString),
				BAXFieldWidth, BAXFieldWidth-IMPrecDiff);
		myFToStrNoPad(GetBAUnOffsetPhase(w[BAttr].GraphRange[xmax_i]),
				CP(LocalXMaxString),
				BAXFieldWidth, BAXFieldWidth-IMPrecDiff);
		MoveTo(w[BAttr].GR.left, w[BAttr].GR.bottom);
		Move(-StringWidth(CP(LocalXMinString))/2, TextMargin + TextHeight-1);
		DrawString(CP(LocalXMinString));
		MoveTo(w[BAttr].GR.right, w[BAttr].GR.bottom);
		MaxXStrLen = StringWidth(CP(LocalXMaxString));
		OverflowOnRight = MAX(MaxXStrLen/2 - ImpactRightMargin, 0);
		Move(-MaxXStrLen/2 - OverflowOnRight+2, TextMargin + TextHeight-1);
		DrawString(CP(LocalXMaxString));

		for(i=0;i<4;++i) {
				/*	Will the following become BAxFloatToCoord?	*/
			x = (w[BAttr].GR.right - w[BAttr].GR.left);
			if (!BA_Rec.AllPhases)
				x -= BA_Rec.StepSize;
			x = (int) (x / (w[BAttr].GraphRange[xmax_i] -
					w[BAttr].GraphRange[xmin_i]) * GetBAOffsetPhase(i/4.0));
/*
			x = xFloatToCoord(BAttr, GetBAOffsetPhase(i/4.0)) * BA_Rec.StepSize;
*/
			if ( (x<=0) || (x>=w[BAttr].GR.right-w[BAttr].GR.left-BA_Rec.StepSize) )
				continue;
			WindowMoveTo(BAttr, x, w[BAttr].Origin.v-w[BAttr].GR.bottom+HashSize+1);
			Line(0, HashSize);
			Move(0, -(w[BAttr].GR.bottom - w[BAttr].GR.top - 1));
			Line(0, HashSize);
			if (i==0)
				if ( (x > StringWidth(CP(LocalXMinString))/2+4) &&
						(x < w[BAttr].GR.right - w[BAttr].GR.left -
						StringWidth(CP(LocalXMaxString))/2-4) ) {
					WindowMoveTo(BAttr, x, w[BAttr].Origin.v-w[BAttr].GR.bottom);
					Move(-CharWidth('0')/2+1, TextMargin + TextHeight-1);
					DrawChar('0');
				}
		}


	/*	Document all parameter values	*/

		s[0] = 0;
		StringAddVar(CP(s), BA_Rec.Vars._A,    "\PAmplitude: ", FALSE);
		 sAppend(CP(s), CP(BA_Spaces));
		StringAddVar(CP(s), BA_Rec.Vars._damp, "\PDamping:   ", FALSE);
		 sAppend(CP(s), CP(BA_Spaces));
		sAppend(CP(s), "\PMaximum Iterations: ");
		if (BA_Rec.MaxTimes == AutoTimes)
			sEquate(CP(s1), "\PAuto");
		else
			LongToString((long) BA_Rec.MaxTimes, CP(s1));
		for(i=s1[0]; i<SpecialFieldWidth-2; ++i)
			sInsertC(CP(s), ' ', s[0]+1);
		sAppend(CP(s), CP(s1));
		BottomTextSize = StringWidth(CP(s));
		sInsertC(CP(s), ReturnChar, s[0]+1);

		StringAddVar(CP(s), BA_Rec.Vars._freq, "\PFrequency: ", FALSE);
		 sAppend(CP(s), CP(BA_Spaces));
		sAppend(CP(s), "\P                     ");
		sAppend(CP(s), "\PIncrement Size:    ");
		LongToString((long) BA_Rec.StepSize, CP(s1));
		for(i=s1[0]; i<SpecialFieldWidth-1; ++i)
			sInsertC(CP(s), ' ', s[0]+1);
		sAppend(CP(s), CP(s1));

		SetRect(&tmpRect, StdMargin,
				w[BAttr].GR.bottom + 2*TextLineHeight + 2*StdMargin + 1,
				w[BAttr].Back->portRect.right, w[BAttr].Back->portRect.bottom);
		BottomTextOffset = BA_Rec.WindowSize.right - BA_Rec.WindowSize.left -
				ScrollBarWidth - 2*StdMargin - BottomTextSize;
		if (BottomTextOffset > 0)
			OffsetRect(&tmpRect, BottomTextOffset/2, 0);
		TextBox(CP((s[1])), (long) (s[0]), &tmpRect, teJustLeft);

		MoveTo(0, w[BAttr].GR.bottom + 2*TextLineHeight + StdMargin + 2);
		Line(BA_Rec.WindowSize.right - BA_Rec.WindowSize.left -
				ScrollBarWidth, 0);
		Move(0, 2*TextLineHeight + StdMargin + 2);
		Line(-(BA_Rec.WindowSize.right - BA_Rec.WindowSize.left -
				ScrollBarWidth), 0);
}

RefreshBA()
{
	if (!DrawBA) {
		SetPort(w[BAttr].Peek);
		EraseRect(&w[BAttr].Peek->port.portRect);
		return;
	}

	if (!w[BAttr].DrawFromScratch) {
		ScrollCopyBackToWindow(BAttr);
		return;
	}

	w[BAttr].DrawFromScratch = FALSE;
	SetPort(w[BAttr].Back);
	EraseRect(&w[BAttr].Back->portRect);

	DrawBABackground();
	DrawCurrentBABackground();

	ScrollCopyBackToWindow(BAttr);
}

			/*	----------------------------------------------	*/
			/*	----------	Periodicity Routines	----------	*/
			/*	----------------------------------------------	*/

SetBASolutionType(int SolType)
{
	CurBASolutionType = SolType;
	HiliteBAImpact(-1, -1);
	if (!UpdateBAOffPorts()) {
		ERROR("\PNot enough memory to keep creating basins of attraction diagram.");
		StopIterating(FALSE);
	}
}

CheckBAPeriodicity()
{
	int		i, per, k;
	Point	p0, p1, p2, p3;
	int		NewGuess;
	Boolean	WillBePeriodic = FALSE, ProbPer = FALSE, GotAProbable = FALSE;
	static	int	ProbPeriod, ProbCount;

	if (HitNumber == 0) {
		ProbPeriod = NoSolutionType;
		ProbCount = 0;
	}

	if (CurBASolutionType != NoSolutionType)	return;

	if (StuckStatus != NOT_STUCK)	{
		SetBASolutionType(StuckSolution);
		return;
	}

	if (BA_Rec.Faster) {
		p0 = (*BA_Points)[HitNumber];
		NewGuess = FindPrevBASolution(p0.h, p0.v);
		if (NewGuess == NoSolutionType) {
			GuessBASolutionType = NoSolutionType;
			GuessBASolCount = 0;
		}
		else if (MiddleOfBasin(NewGuess, p0.h, p0.v)) {
			SetBASolutionType(NewGuess);
			return;
		}
		if (NewGuess != GuessBASolutionType) {
			GuessBASolutionType = NewGuess;
			GuessBASolCount = 0;
		}
		else if (++GuessBASolCount >= BA_CorrectGuessSol) {
			SetBASolutionType(GuessBASolutionType);
			return;
		}
	}

	for (i=Period1Solution; i<=Period8Solution; ++i) {
		if (i==5)	per=8;
		else		per=i;
		if (4*per-1 > HitNumber) break;
		for (k=0; k<per; ++k) {
			p0 = (*BA_Points)[HitNumber - 0*per - k];
			p1 = (*BA_Points)[HitNumber - 1*per - k];
			p2 = (*BA_Points)[HitNumber - 2*per - k];
			p3 = (*BA_Points)[HitNumber - 3*per - k];
			if (	(PtSqrDist(p0, p1) > 2L) || (PtSqrDist(p0, p2) > 2L) ||
					(PtSqrDist(p0, p3) > 2L) )
					break;
				
			if (k==per-1) {
				if (!TestPeriodicity(per, &WillBePeriodic, &ProbPer)) {
					if (ProbPer || WillBePeriodic) {
/*
						if (per == ProbPeriod)	++ProbCount;
*/
						if (!GotAProbable) {
							if ( (ProbPeriod != NoSolutionType) &&
									( (ProbPeriod % per == 0) ||
									(per % ProbPeriod == 0) /* ??? */ ) )
										/*	OK to get tricked from 2 to 1	*/
										/*	1 to 2, 2 to 4, etc.			*/
								++ProbCount;
							else
								ProbCount = 1;
							ProbPeriod = per;
						}
						GotAProbable = TRUE;
					}
					if (WillBePeriodic || (ProbCount >= MinProbableCount))
						if (HitNumber >= CurrentMaxTimes)
							if (BA_Rec.MaxTimes == AutoTimes)
								if (++NumCMTimesIncs <= MaxCMTimesIncs)
									CurrentMaxTimes += CMTimesInc;
					break;
				}
				SetBASolutionType(i);
				return;
			}
		}
	}

	if (HitNumber >= CurrentMaxTimes)
		SetBASolutionType(OtherSolution);
}

			/*	------------------------------------------	*/
			/*	----------	Graphing Routines	----------	*/
			/*	------------------------------------------	*/

BACopyBits(int SrcPortIndex, int DstPortIndex, int theMode)
{
	if (BA_Rec.Graphs[SrcPortIndex] == NOPORT)	return;
	if (BA_Rec.Graphs[DstPortIndex] == NOPORT)	return;

	myCopyBits(BA_Rec.Graphs[SrcPortIndex], BA_Rec.Graphs[DstPortIndex],
			&w[BAttr].GR, 0,0,0,0, theMode, (RgnHandle) NIL);
}

CopyBAGraph(Rect *pr, int h, int v)
{
	GrafPtr		gp;
	RgnHandle	oldClip, tmpClipRgn;
	Rect		tmpClipRect;
	int			i;

	GetPort(&gp);
	if (BA_Rec.InColor) {
		for(i=0; i<NumSolutionTypes; ++i) {
			if (BA_Rec.Graphs[i] == NOPORT)			continue;
			SetSpecialColor(csi[i], gp == (GrafPtr) w[BAttr].Peek);

/*	BEGIN changes for System 7, BB 2.01
 *		Change 1 of 2
 */
 
			/* OLD: change adMin to srcOr for Sys 7.
			 *	Can't use arithmetic transfer modes to colorize bit maps.
						myCopyBits(BA_Rec.Graphs[i], gp,
								pr, h, v, 0, 0, adMin, (RgnHandle) NIL);
			 */
			
			SetPort(BA_Rec.Graphs[ScratchBAOP]);
			myCopyBits(BA_Rec.Graphs[i],BA_Rec.Graphs[ScratchBAOP], &BA_Rec.Graphs[i]->portRect,
					0, 0, 0, 0, srcCopy, (RgnHandle) NIL);
			myCopyBits(w[BAttr].Back, BA_Rec.Graphs[ScratchBAOP],
					pr, 0, 0, 0, 0, srcBic, (RgnHandle) NIL);
			SetPort(gp);
			myCopyBits(BA_Rec.Graphs[ScratchBAOP], gp,
					pr, h, v, 0, 0, srcOr, (RgnHandle) NIL);

/*	END changes for System 7, BB 2.01
 *		Change 1 of 2
 */

		}
		SetSpecialColor(BlackColorIndex, gp == (GrafPtr) w[BAttr].Peek);
	}
	else
		myCopyBits(BA_Rec.Graphs[CompBAOP], gp,
				pr, h, v, 0, 0, srcOr, (RgnHandle) NIL);

	if (gp == (GrafPtr) w[BAttr].Peek)
		if (BALatestImpactRect.left != 0) {
			SetPort(w[BAttr].Peek);	/*	redundant	*/
			oldClip = NewRgn();
			tmpClipRgn = NewRgn();
			GetClip(oldClip);
			tmpClipRect = w[BAttr].GR;
			OffsetRect(&tmpClipRect, -h, -v);
			SectRect(&tmpClipRect, pr, &tmpClipRect);
			RectRgn(tmpClipRgn, &tmpClipRect);
			SectRgn(oldClip, tmpClipRgn, tmpClipRgn);
			SetClip(tmpClipRgn);
	
			OffsetRect(&BALatestImpactRect, -h, -v);
			InvertRect(&BALatestImpactRect);
			OffsetRect(&BALatestImpactRect, h, v);
	
			SetClip(oldClip);
			DisposeRgn(tmpClipRgn);
			DisposeRgn(oldClip);
		}
}

Boolean TestBAPoint(int WhichPort, int xCoord, int yCoord)
{
	Point	pt;

	if (BA_Rec.Graphs[WhichPort] == NOPORT)
		return(FALSE);
	if( (xCoord < 0) || (xCoord > BA_Rec.MaxXInt) || (yCoord < 0) ||
			(yCoord > BA_Rec.MaxYInt) )
		return(FALSE);
		/*	See BackMoveTo()	*/

	pt.h = w[BAttr].BackOrigin.h + xCoord * BA_Rec.StepSize - w[BAttr].GR.left;
	pt.v = w[BAttr].BackOrigin.v - yCoord * BA_Rec.StepSize - w[BAttr].GR.top;
	return(BitTst((Ptr) BA_Rec.Graphs[WhichPort]->portBits.baseAddr,
		(long) pt.v * BA_Rec.Graphs[WhichPort]->portBits.rowBytes*8 + pt.h));
}

int FindPrevBASolution(int xCoord, int yCoord)
{
	int	i;

	if (!TestBAPoint(AllBAOP, xCoord, yCoord))
		return(NoSolutionType);

	for(i=0; i<NumSolutionTypes; ++i)
		if (TestBAPoint(i, xCoord, yCoord))
			return(i);

	return(NoSolutionType);	/*	should never get here	*/
}

Boolean MiddleOfBasin(int WhichBasin, int xCoord, int yCoord)
{
	int	i,j, x;

	for(i=xCoord-1; i<=xCoord+1; ++i) {
			/*	First handle wraparound phases	*/
		if ( (i<0) && (BA_Rec.AllPhases) )
			x = BA_Rec.MaxXInt;
		else if ( (i > BA_Rec.MaxXInt) && (BA_Rec.AllPhases) )
			x = 0;
		else
			x = i;
		for(j=yCoord-1; j<=yCoord+1; ++j)
			if (FindPrevBASolution(x, j) != WhichBasin)
				return(FALSE);
	}
	return(TRUE);
}

PlotBAPoint(int WhichPort, int xCoord, int yCoord)
/*	Apparently only called with WhichPort == CurIterBAOP	-TA 6/4/91	*/
{
	SetPort(BA_Rec.Graphs[WhichPort]);
	BackMoveTo(BAttr, xCoord * BA_Rec.StepSize, yCoord * BA_Rec.StepSize);
	Line(0, 0);
}

HiliteBAImpact(int xCoord, int yCoord)
{
	Point		pt;
	int			hOffset = GetCtlValue(w[BAttr].BSB);
	int			vOffset = GetCtlValue(w[BAttr].RSB);
	RgnHandle	oldClip, tmpClipRgn;
	Rect		tmpClipRect;

	SetPort(w[BAttr].Peek);
	oldClip = NewRgn();
	tmpClipRgn = NewRgn();
	GetClip(oldClip);
	tmpClipRect = w[BAttr].GR;
	OffsetRect(&tmpClipRect, -hOffset, -vOffset);
	RectRgn(tmpClipRgn, &tmpClipRect);
	SectRgn(oldClip, tmpClipRgn, tmpClipRgn);
	SetClip(tmpClipRgn);

	if (BALatestImpactRect.left != 0) {
		OffsetRect(&BALatestImpactRect, -hOffset, -vOffset);
		InvertRect(&BALatestImpactRect);
		OffsetRect(&BALatestImpactRect, hOffset, vOffset);
	}

	OffsetRect(&BALatestImpactRect, -BALatestImpactRect.left,
			-BALatestImpactRect.top);
	if ( (xCoord >= 0) && (xCoord <= BA_Rec.MaxXInt) &&
			(yCoord >= 0) && (yCoord <= BA_Rec.MaxYInt) ) {
		pt.h = w[BAttr].Origin.h + xCoord * BA_Rec.StepSize;
		pt.v = w[BAttr].Origin.v - yCoord * BA_Rec.StepSize;
		OffsetRect(&BALatestImpactRect, pt.h-1, pt.v-1);
		OffsetRect(&BALatestImpactRect, -hOffset, -vOffset);
		InvertRect(&BALatestImpactRect);
		OffsetRect(&BALatestImpactRect, hOffset, vOffset);
	}

	SetClip(oldClip);
	DisposeRgn(tmpClipRgn);
	DisposeRgn(oldClip);
}

PlotInBA(int xCoord, int yCoord)
{
	w[BAttr].WindowCleared = FALSE;

	HiliteBAImpact(xCoord, yCoord);

	if ( BA_Rec.Faster || (HitNumber==0) )
		PlotBAPoint(CurIterBAOP, xCoord, yCoord);
}

GetBAPlot(double del, double vel, Point *ppt)
{
	ppt->h = xFloatToCoord(BAttr, GetBAOffsetPhase(del));
	ppt->v = yFloatToCoord(BAttr, vel);
			/*	Next statement: See IMap discussion.	*/
/*
	if (ppt->h >= (* was >= *) w[BAttr].GR.right - w[BAttr].Origin.h)
*/
	if (ppt->h > /* was >= */ BA_Rec.MaxXInt)
		if (BA_Rec.AllPhases)
/*
			ppt->h = w[BAttr].GR.left-w[BAttr].Origin.h;
*/
			ppt->h = 0;
	if (!BA_Rec.AllPhases) {
		if (ppt->h == -MaxScreenX)		ppt->h = -1;
		else if (ppt->h == MaxScreenX)	ppt->h = BA_Rec.MaxXInt + 1;
	}
	if (ppt->v == -MaxScreenY)
		ppt->v = -1;
	else if (ppt->v == MaxScreenY)
		ppt->v = BA_Rec.MaxYInt + 1;
}

ViewBAOffPorts(int Which);
ViewBAOffPorts(int Which)
/*	Diagnostic routine	*/
{
/*
	int		i;
	GrafPtr	gp;

	if (Which == -1) {
		for(i=0; i<NumBAOffPorts; ++i)
			if (BA_Rec.Graphs[i] != NOPORT)
				myCopyBits(BA_Rec.Graphs[i], (GrafPtr) w[BAttr].Peek,
						&w[BAttr].GR, 0, 0, -GetCtlValue(w[BAttr].BSB),
						-GetCtlValue(w[BAttr].RSB), srcCopy, (RgnHandle) NIL);
	}
	else
		if (BA_Rec.Graphs[Which] != NOPORT)
			myCopyBits(BA_Rec.Graphs[Which], (GrafPtr) w[BAttr].Peek,
					&w[BAttr].GR, 0, 0, -GetCtlValue(w[BAttr].BSB),
					-GetCtlValue(w[BAttr].RSB), srcCopy, (RgnHandle) NIL);
	GetPort(&gp);
	StdForceUpdate((WindowPtr) w[BAttr].Peek);
	SetPort(gp);
*/
}

Boolean UpdateBAOffPorts()
{
	int	ToScreenPort;

ViewBAOffPorts(CurIterBAOP);
ViewBAOffPorts(AllBAOP);
	if (BA_Rec.Graphs[CurBASolutionType] == NOPORT)
		if (!CreateBAOffPort(CurBASolutionType, &w[BAttr].GR, BA_Rec.StepSize))
			return(FALSE);

	if (BA_Rec.Faster)
		BACopyBits(AllBAOP, CurIterBAOP, srcBic);
	BACopyBits(CurIterBAOP, AllBAOP, srcOr);
	BACopyBits(CurIterBAOP, CurBASolutionType, srcOr);
	if (BA_Rec.InColor) {
		SetPort(w[BAttr].Peek);
		SetSpecialColor(csi[CurBASolutionType], TRUE);
		ToScreenPort = CurIterBAOP;
	}
	else {
		SetPort(BA_Rec.Graphs[ScratchBAOP]);
		BACopyBits(CurIterBAOP, ScratchBAOP, srcCopy);
ViewBAOffPorts(ScratchBAOP);
		PenMode(notPatBic);
		PenPat(&SolutionPats[psi[CurBASolutionType]]);
		PaintRect(&w[BAttr].GR);
ViewBAOffPorts(ScratchBAOP);
		PenMode(patCopy);
		BACopyBits(ScratchBAOP, CompBAOP, srcOr);
ViewBAOffPorts(CompBAOP);
		ToScreenPort = ScratchBAOP;
	}
	SetPort(w[BAttr].Peek);

/*	BEGIN changes for System 7, BB 2.01
 *		Change 1 of 2
 */
 
  /* OLD: change adMin to srcOr for Sys 7.
   *	Can't use arithmetic transfer modes to colorize bit maps.
	if (BA_Rec.InColor)
		myCopyBits(BA_Rec.Graphs[ToScreenPort], (GrafPtr) w[BAttr].Peek,
				&w[BAttr].GR, 0, 0, -GetCtlValue(w[BAttr].BSB),
				-GetCtlValue(w[BAttr].RSB), adMin, (RgnHandle) NIL);
   */

	if (BA_Rec.InColor) {
		SetPort(BA_Rec.Graphs[ScratchBAOP]);
		myCopyBits(BA_Rec.Graphs[ToScreenPort],BA_Rec.Graphs[ScratchBAOP],
				&BA_Rec.Graphs[ToScreenPort]->portRect,
				0, 0, 0, 0, srcCopy, (RgnHandle) NIL);
		myCopyBits(w[BAttr].Back, BA_Rec.Graphs[ScratchBAOP],
				&w[BAttr].GR, 0, 0, 0, 0, srcBic, (RgnHandle) NIL);
		SetPort(w[BAttr].Peek);
		myCopyBits(BA_Rec.Graphs[ScratchBAOP], (GrafPtr) w[BAttr].Peek,
				&w[BAttr].GR, 0, 0, -GetCtlValue(w[BAttr].BSB),
				-GetCtlValue(w[BAttr].RSB), srcOr, (RgnHandle) NIL);
	}

/*	END changes for System 7, BB 2.01
 *		Change 1 of 2
 */

	else
		myCopyBits(BA_Rec.Graphs[ToScreenPort], (GrafPtr) w[BAttr].Peek,
				&w[BAttr].GR, 0, 0, -GetCtlValue(w[BAttr].BSB),
				-GetCtlValue(w[BAttr].RSB), srcOr, (RgnHandle) NIL);
	if (BA_Rec.InColor) {
		SetPort(w[BAttr].Peek);
		SetSpecialColor(BlackColorIndex, TRUE);
	}

	return(TRUE);
}

DrawBAComposite()
{
	int	i;

	SetPort(BA_Rec.Graphs[CompBAOP]);
	EraseRect(&w[BAttr].GR);

	SetPort(BA_Rec.Graphs[ScratchBAOP]);
	for(i=0; i<NumSolutionTypes; ++i) {
		if (BA_Rec.Graphs[i] == NOPORT)		continue;

		BACopyBits(i, ScratchBAOP, srcCopy);
		PenMode(notPatBic);
		PenPat(&SolutionPats[psi[i]]);
		PaintRect(&w[BAttr].GR);
		PenMode(patCopy);
		BACopyBits(ScratchBAOP, CompBAOP, srcOr);
	}
}

GraphBA()
{
	Point	pt;

	GetBAPlot(phase, vi, &pt);

	(*BA_Points)[HitNumber] = pt;

	PlotInBA(pt.h, pt.v);
	CheckBAPeriodicity();

	if (StuckStatus == STUCK_FOREVER)
		StopStickingMessage();
	if (NeedNextSpecialValue && (CurBASolutionType == NoSolutionType)) {
		NeedNextSpecialValue = FALSE;
		SetBASolutionType(OtherSolution);
	}
}

ClearBasinsOfAttraction()
{
	int	i;

	if (ProgramStatus == BASINS_OF_ATTRACTION)
		StopIterating(FALSE);
	SetPort(w[BAttr].Back);
	EraseRect(&w[BAttr].Back->portRect);
	StdForceUpdate((WindowPtr) w[BAttr].Peek);
	DrawBA = FALSE;
	NeedToSaveBA = FALSE;
	SetSpecialTitle(FALSE, BAttr);
	BA_Rec.GraphInAllWindows = BA_Rec.InColor = FALSE;
	CheckGraphInAll(BAttr);
	CheckSpecialColor(BAttr);
	for(i=0; i<NumBAOffPorts; ++i)
		if (BA_Rec.Graphs[i] != NOPORT) {
			DisposeOffPort(BA_Rec.Graphs[i]);
			BA_Rec.Graphs[i] = NOPORT;
		}

	SpecialMenuEnable(BAttr, FALSE, FALSE);
	SetSpecialScrolling(BAttr);
}


			/*	--------------------------------------	*/
			/*	----------	File Routines	----------	*/
			/*	--------------------------------------	*/


Boolean SaveBAReturn(Boolean SaveOKSoFar)
{
	Boolean	GoodOp;
	int		i;

	GoodOp = CloseSaveFile(!SaveOKSoFar);

/* deleted 11/11/91--causing resource problems?
	if ( ((**BackPic).picSize == GetHandleSize((Handle) BackPic)) &&
			((**BackPic).picSize != 0L) )
		KillPicture(BackPic);

	for(i=0; i<NumSolutionTypes; ++i)
		if ( ((**GraphPix[i]).picSize == GetHandleSize((Handle) GraphPix[i])) &&
				((**GraphPix[i]).picSize != 0L) )
			KillPicture(GraphPix[i]);
*/

	return(SaveOKSoFar && GoodOp);
}

Boolean SaveBAFile()
{
	Boolean			GoodOp;
	int				PicID;
	int				i;
	Ptr				SolInfoPtr;
	ControlHandle	RSch, BSch;
	Boolean			WasThere;

	BackPic = NIL;
	for(i=0; i<NumSolutionTypes; ++i)
		GraphPix[i] = NIL;

	if ( !WritePtrToFile((Ptr) &BA_Rec, sizeof(BasAttrRecord)) )
		return(SaveBAReturn(FALSE));

	RSch = w[BAttr].RSB;
	w[BAttr].RSB = (ControlHandle) GetCtlValue(RSch);
	BSch = w[BAttr].BSB;
	w[BAttr].BSB = (ControlHandle) GetCtlValue(BSch);
	GoodOp = WritePtrToFile((Ptr) &w[BAttr], sizeof(BBWindow));
	w[BAttr].RSB = RSch;	w[BAttr].BSB = BSch;
	if (!GoodOp)
		return(SaveBAReturn(FALSE));

	SolInfoPtr = (BA_Rec.InColor) ? (Ptr) &csi : (Ptr) &psi;
	if ( !WritePtrToFile(SolInfoPtr, sizeof(SolutionInfo)) )
		return(SaveBAReturn(FALSE));

	PicID = FirstBAPicture;
	BackPic = MakeSpecialPicture(w[BAttr].Back);
	if (BackPic == NIL)
		return(SaveBAReturn(FALSE));
	if (!StdPutPicture(BackPic, PicID))
		return(SaveBAReturn(FALSE));

	for(i=0; i<NumSolutionTypes; ++i) {
		++PicID;
		if (!Delete1Resource(PICT_OSType, PicID, &WasThere))
			return(SaveBAReturn(FALSE));
		if (BA_Rec.Graphs[i] == NOPORT)		continue;
		GraphPix[i] = MakeSpecialPicture(BA_Rec.Graphs[i]);
		if (GraphPix[i] == NIL)
			return(SaveBAReturn(FALSE));
		if (!StdPutPicture(GraphPix[i], PicID))
			return(SaveBAReturn(FALSE));
	}

	return(SaveBAReturn(TRUE));
}

Boolean BadOpenBAReturn()
{
	Boolean	SeriousError = FALSE;
	int		i;

/*	Don't delete--from resource, so closing file will do
	if (BackPic != NIL)
		KillPicture(BackPic);
	for(i=0; i<NumSolutionTypes; ++i)
		if (GraphPix[i] != NIL)
			KillPicture(GraphPix[i]);
*/
	if (BA_Points != NIL)
		DisposHandle((Handle) BA_Points);
	BA_Points = oBA_Points;

		/*	OffPort Stuff	*/

	for(i=0; i<NumBAOffPorts; ++i)
		if (tmpBA.Graphs[i] != NOPORT) {
			DisposeOffPort(tmpBA.Graphs[i]);
			tmpBA.Graphs[i] = NOPORT;
		}
	if (GraphExists)
		for(i=0; i<NumBAOffPorts; ++i) {
			if (oGraphPix[i] != NIL)
				if (BA_Rec.Graphs[i] == NOPORT) {
					if (!CreateBAOffPort(i, &w[BAttr].GR, BA_Rec.StepSize))
						SeriousError = TRUE;
					else {
						EraseRect(&BA_Rec.Graphs[i]->portRect);
						DrawPicture(oGraphPix[i], &BA_Rec.Graphs[i]->portRect);
					}
				}
			if (oGraphPix[i] != NIL)
				KillPicture(oGraphPix[i]);
		}
	if (SeriousError) {
		DrawBA = FALSE;
		w[BAttr].DrawFromScratch = TRUE;
		StdForceUpdate((WindowPtr) w[BAttr].Peek);
		SetSpecialScrolling(BAttr);
	}

/*
	CloseOpenFile();
*/

	SetANilFileName(BAttr);

	if (SeriousError)	ERROR(SeriousBAOpenMemErrorString);
	else				ERROR(BAOpenMemErrorString);

	return(FALSE);
}

Boolean OpenBAFile()
{
	long			ReadAmt;
	BBWindow		tmpW;
	Rect			nBackRect, oBackRect, CompositeBackRect, tmpRect;
	Boolean			NeedNewBackPort;
	int				PicID;
	Size			MinExtraMemory;
	Boolean			oGraphInAllWindows;
	SolutionInfo	tsi;
	int				i;
	int				RScrollValue, BScrollValue;
	int				MaxTimes;
	long			NewBackMemAddition;

		/*	Get Main Records: BA_Rec, w[BAttr]	*/

	ReadAmt = sizeof(BasAttrRecord);
	if ( !ReadPtrFromFile((Ptr) &tmpBA, &ReadAmt) ) {
		ERROR("\PBasins of Attraction file not intact. Cannot open.");
		return(FALSE);
	}

	ReadAmt = sizeof(BBWindow);
	if ( !ReadPtrFromFile((Ptr) &tmpW, &ReadAmt) ) {
		ERROR("\PBasins of Attraction file not intact. Cannot open.");
		return(FALSE);
	}

	ReadAmt = sizeof(SolutionInfo);
	if ( !ReadPtrFromFile((Ptr) &tsi, &ReadAmt) ) {
		ERROR("\PBasins of Attraction file not intact. Cannot open.");
		return(FALSE);
	}

		/*	Initialize New Variables	*/

		/*	The following information, if created, needs to be cleared
		 *	if open errorsee BadOpenBAReturn().
		 */
	for(i=0; i<NumBAOffPorts; ++i) {
		tmpBA.Graphs[i] = NOPORT;
		oGraphPix[i] = NIL;
		if (i<NumSolutionTypes)
			GraphPix[i] = NIL;
	}
	BackPic = NIL;
	oBA_Points = BA_Points;
	BA_Points = NIL;

		/*	Test if sufficient memory	*/
	
	oBackRect = w[BAttr].Back->portRect;
	nBackRect = tmpBA.WindowSize;
	OffsetRect(&nBackRect, -nBackRect.left, -nBackRect.top);
		/*	Need to enlarge w[BAttr].Back?	*/
	NeedNewBackPort = (	(nBackRect.right > oBackRect.right) ||
						(nBackRect.bottom > oBackRect.bottom) );
	GraphExists = (BA_Rec.Graphs[0] != NOPORT);

	MinExtraMemory = ((long) (tmpW.GR.right - tmpW.GR.left)) *
			((long) (tmpW.GR.bottom - tmpW.GR.top));
	MinExtraMemory /= 8;		/*	bits in a byte	*/
	MinExtraMemory *= (NumBAOffPorts-4 + NumSolutionTypes-4);	/* offports, graphpix	*/
	MinExtraMemory += 5*1024L;	/*	Approx. size of BackPic	*/
/*		Delete, because oGraphPix creation accompanied by BA_Rec.Graphs disposal
	if (GraphExists)
		MinExtraMemory += ((long) (w[BAttr].GR.right - w[BAttr].GR.left)) *
				((long) (w[BAttr].GR.bottom - w[BAttr].GR.top)) / 8 * 8;	(*	oGraphPix	*)
*/
	if (NeedNewBackPort) {
		SetRect(&CompositeBackRect, 0, 0, MAX(nBackRect.right, oBackRect.right),
				MAX(nBackRect.bottom, oBackRect.bottom));
		NewBackMemAddition = ((long) CompositeBackRect.right) *
				((long) CompositeBackRect.bottom) / 8;		/*	new BackPort	*/
		NewBackMemAddition += NewBackMemAddition * 3 / 4;	/*	Cushion for drawing backgd?--bug 11/12/91	*/
		MinExtraMemory += NewBackMemAddition;
	}
	if (!CheckMem(MinExtraMemory))
		return(BadOpenBAReturn());

		/*	Get Pictures, Back and Graphs[i]	*/

	PicID = FirstBAPicture;
	if ( (BackPic = GetPicture(PicID)) == NIL)
		return(BadOpenBAReturn());
	for(i=0; i<NumSolutionTypes; ++i) {
		++PicID;
		if (Get1Resource(PICT_OSType, PicID) == NIL)		continue;
		if ( (GraphPix[i] = GetPicture(PicID)) == NIL)
			return(BadOpenBAReturn());
	}

		/*	Create BA_Points Array (of graphed points)	*/

	MaxTimes = ( (tmpBA.MaxTimes != AutoTimes) ?
			tmpBA.MaxTimes : (tmpBA.DefaultMaxTimes * MaxCMTimesIncs) );
	BA_Points = (Point**) NewHandle( ((long) MaxTimes+1) * sizeof(Point) );
	if (BA_Points == NIL)
		return(BadOpenBAReturn());

		/*	Handle Offports	*/

	if (GraphExists)
		for(i=0; i<NumBAOffPorts; ++i) {
			if (BA_Rec.Graphs[i] == NOPORT)		continue;
			oGraphPix[i] = MakeSpecialPicture(BA_Rec.Graphs[i]);	/*	Backup	*/
			if (oGraphPix[i] == NIL)
				return(BadOpenBAReturn());
			DisposeOffPort(BA_Rec.Graphs[i]);
			BA_Rec.Graphs[i] = NOPORT;
		}
	for(i=0; i<NumBAOffPorts; ++i) {
		tmpBA.Graphs[i] = NOPORT;
		if ( (i<NumSolutionTypes) && (GraphPix[i] == NIL) )
			continue;

		if (!CreateSpecialOffPort(&tmpBA.Graphs[i], &tmpW.GR))
			return(BadOpenBAReturn());
		SetPort(tmpBA.Graphs[i]);
		PenSize(tmpBA.StepSize, tmpBA.StepSize);
	}

	if (NeedNewBackPort) {
			/*	See comment, etc., in bbBifDiag.c */
		tmpW.Back = NewOffPort(&CompositeBackRect);
		SetPort(tmpW.Back);	SetStdTextAttributes();
		if (tmpW.Back == NIL)
			return(BadOpenBAReturn());
	}
	else
		tmpW.Back = w[BAttr].Back;

		/*	Open is OK!--Clean up fields	*/

	SetOpenWindows();

		/*	Scroll values saved in ControlHandle fields (except early BB)	*/
	RScrollValue = (int) tmpW.RSB;
	if ( (RScrollValue != (long) tmpW.RSB) || (RScrollValue < 0) )
		RScrollValue = 0;
	BScrollValue = (int) tmpW.BSB;
	if ( (BScrollValue != (long) tmpW.BSB) || (BScrollValue < 0) )
		BScrollValue = 0;

	if (tmpBA.InColor)
		tmpBA.InColor = UsingColorMac;
	tmpBA.FullWindow =	(tmpBA.WindowSize.left == LWR.left) &&
						(tmpBA.WindowSize.top == LWR.top) &&
						(tmpBA.WindowSize.right == LWR.right) &&
						(tmpBA.WindowSize.bottom == LWR.bottom);
	tmpW.Peek = w[BAttr].Peek;
	tmpW.RSB = w[BAttr].RSB;
	tmpW.BSB = w[BAttr].BSB;
	tmpW.HS[0] = w[BAttr].HS[0];
	tmpW.HS[1] = w[BAttr].HS[1];

		/*	Dispose of all old BA info	*/

	for(i=0; i<NumBAOffPorts; ++i)
		if (oGraphPix[i] != NIL)
			KillPicture(oGraphPix[i]);
	if (oBA_Points != NIL)
		DisposHandle((Handle) oBA_Points);
	if (NeedNewBackPort) {
		DisposeOffPort(w[BAttr].Back);
		w[BAttr].Back = NOPORT;
	}

		/*	Set Global Variables	*/

	oGraphInAllWindows = BA_Rec.GraphInAllWindows;
	BA_Rec = tmpBA;
	BA_Rec.GraphInAllWindows = oGraphInAllWindows;
	w[BAttr] = tmpW;
	SetRect(&BALatestImpactRect, 0, 0, BA_Rec.StepSize+2, BA_Rec.StepSize+2);
	DrawBA = TRUE;
	w[BAttr].DrawFromScratch = FALSE;
	w[BAttr].WindowCleared = FALSE;

	SetSpecialScrolling(BAttr);
	if (RScrollValue > GetCtlMax(w[BAttr].RSB))	RScrollValue = 0;
	if (BScrollValue > GetCtlMax(w[BAttr].BSB))	BScrollValue = 0;
	SetCtlValue(w[BAttr].RSB, RScrollValue);
	SetCtlValue(w[BAttr].BSB, BScrollValue);

	CheckSpecialColor(BAttr);
	StdForceUpdate((WindowPtr) w[BAttr].Peek);

		/*	Draw into OffPorts	*/

	SetPort(w[BAttr].Back);
	EraseRect(&w[BAttr].Back->portRect);
	DrawPicture(BackPic, &(**BackPic).picFrame);
/*
	KillPicture(BackPic);
*/
	for(i=0; i<NumSolutionTypes; ++i) {
		if (BA_Rec.Graphs[i] == NOPORT)		continue;
		SetPort(BA_Rec.Graphs[i]);
		EraseRect(&BA_Rec.Graphs[i]->portRect);
		DrawPicture(GraphPix[i], &BA_Rec.Graphs[i]->portRect);
/*
		KillPicture(GraphPix[i]);
*/
	}
	for(i=0; i<NumSolutionTypes; ++i) {
		if (BA_Rec.Graphs[i] == NOPORT)		continue;
		BACopyBits(i, AllBAOP, srcOr);
	}

		/*	Handle Iteration, Menus	*/

/*	CloseOpenFile();
*/	/*	Delete before call other resources--unloc. bug	FOUND 3/23/91 -TA */

	SetSpecialTitle(TRUE, BAttr);

	if (tmpBA.InColor) {
		for(i=0; i<NumSolutionTypes; ++i)
			if (csi[i] != tsi[i]) break;
		if (i<NumSolutionTypes) 
			if (ReplaceSolutionInfoDialog("\Pcolors", "\Pbasins of attraction")) {
				for(i=0; i<NumSolutionTypes; ++i)
					csi[i] = tsi[i];
				StdForceUpdate((WindowPtr) w[BifD].Peek);
			}
	}
	else {
		for(i=0; i<NumSolutionTypes; ++i)
			if (psi[i] != tsi[i]) break;
		if (i<NumSolutionTypes) 
			if (ReplaceSolutionInfoDialog("\Ppatterns", "\Pbasins of attraction"))
				for(i=0; i<NumSolutionTypes; ++i)
					psi[i] = tsi[i];
	}
	DrawBAComposite();

	SpecialMenuEnable(BAttr, TRUE, FALSE);
	if (ProgramStatus == BASINS_OF_ATTRACTION)
		StopIterating(FALSE);
			/*	Do next here due to StopIterating(), ProgramStatus	*/
	if (oGraphInAllWindows != tmpBA.GraphInAllWindows)
		SwitchGraphInAll(BAttr);
	NeedToSaveBA = FALSE;
	OKToContinueBA = (BA_Rec.CurrentXInt >= 0) &&
			(BA_Rec.CurrentXInt <= BA_Rec.MaxXInt);
	ControlMenuStop(BASINS_OF_ATTRACTION, OKToContinueBA);
	if ((OKToContinueBA) && (!GettingParamFiles || LastParamFile)) {
		HandleAllActivates();
		HandleAllUpdates();
		if (DoSpecialOpenContinueDialog("\Pbasins of attraction"))
			ContinueBasinsOfAttraction();
	}

	return(TRUE);
}
