/*	bbBifDiag.c
 *
 *	Routines for Bifurcation Diagram
 *
 *		-TA 2/8/91
 */

#include <FileUtils.h>
#include <OffPort.h>
#include "bbPrototypes.h"
#include "bbGraphics.h"
#include "bbSpecial.h"

/*	BEGIN changes for System 7, BB 2.01					Change 4 of 5
 */

/*
pascal OSErr BitMapToRegionGlue(RgnHandle region,const BitMap *bMap); 
pascal OSErr BitMapToRegion(RgnHandle region,const BitMap *bMap)
    = 0xA8D7; 
*/
pascal OSErr BitMapToRegion(RgnHandle region, BitMap *bMap)
    = 0xA8D7; 

/*	END changes for System 7, BB 2.01					Change 4 of 5
 */

#define	BDOpenMemErrorString		"\PError: Not enough memory to open this bifurcation diagram file"
#define	SeriousBDOpenMemErrorString	"\PSerious memory error: cannot open new bifurcation diagram and cannot restore old one"

				/*	Global external vars from bb.c	*/
extern	IterationVars	NormalVars, *pCurrentVars;
extern	hBBPreferences	hBBPrefs;

extern	int				ProgramStatus, IterationStatus;
extern	int				CurrentModifiers;
extern	Boolean			UsingColorMac;

/*	BEGIN changes for System 7, BB 2.01					Change 4 of 5
 */

extern	Boolean			System7OrLater;

/*	END changes for System 7, BB 2.01					Change 4 of 5
 */

extern	Boolean			OKToContinueBD;
extern	Boolean			GettingParamFiles;
extern	Boolean			RootFindingError;

				/*	From bbFiles.c	*/
extern	Boolean			LastParamFile;

				/*	From bbGraphics.c	*/
extern	Windows			*pw;

				/*	From bbWindows.c	*/
extern	int				MaxScreenY;
extern	Rect			LWR;

				/*	From bbImpactGraphics.c	*/
extern	double			MinPhase, MaxPhase, VMin;

				/*	From bbContGraphics.c	*/
extern	Boolean			RedrawSine;

				/*	From bbMenus.c	*/
extern	Boolean			NeedToSaveBD;

				/*	From bbSpecial.c	*/
extern	Boolean			NeedToTurnOnGraphing;
extern	Boolean			NeedNextSpecialValue;
extern	RGBColor		SolutionColors[NumSolutionTypes];

				/*	From bbText.c	*/
extern	Boolean			NeedToSetTxOutHeader, NeedToSetTSerHeader;

BifDiagRecord		BD_Rec, tmpBD;
pBifDiagRecord		pbd;
int					CurrentInitIterTimes, CurrentShowTimes = -1; /* why -1? -TA 3/4/91*/
Rect				BDBox, BDLatestImpactRect;
int					BDXFieldWidth, BDYFieldWidth;
Boolean				DrawBD;
Boolean				NeedOldInitCond;
int					BD_MaxTimes, MaxBDShowTimes;
int					**BD_Points;
int					CurBDSolutionType;
Boolean				NeedToColorAfter;
int					CIITimesInc, NumCIITimesIncs;
Boolean				BDChanging = TRUE;
Str255				BD_Spaces;

						/*	Following for Opening	*/
int					oBD_MaxTimes, **oBD_Points, oMaxBDShowTimes;
PicHandle			oGraphPic;
static	Boolean		GraphExists;
static	PicHandle	BackPic, GraphPic;	/*	for file operations	*/


			/*	------------------------------------------	*/
			/*	----------	Utility Routines	----------	*/
			/*	------------------------------------------	*/

double GetBDOffsetPhase(double del)
{
	del -= BD_Rec.MinPhase;
	if (del>1.0+AlmostZero)		/*	used to be >=	*/
		del -= 1.0;
	if (del<0.0)
		del += 1.0;
	return(del);
}

double GetBDUnOffsetPhase(double del)
{
	if (BD_Rec.MinPhase != 0.0)
		del += BD_Rec.MinPhase;
	return(del);
}

int GetDefaultInitIterTimes(double theDamping)
{
	return(RoundToInt(BaseInitIterTimes * theDamping + AddedInitIterTimes));
}

CopyBDGraph(Rect *pr, int h, int v)
{
	GrafPtr		gp;
	int			i;
	int			LeftEnd, RightEnd;
	Rect		tmpRect;
	int			TheColor;


/*	BEGIN changes for System 7, BB 2.01					Change 4 of 5
 */

	RgnHandle	tmprgn, sys7mask;

/*	END changes for System 7, BB 2.01					Change 4 of 5
 */

	GetPort(&gp);

	if (BD_Rec.InColor) {
		RightEnd = 0;

/*	BEGIN changes for System 7, BB 2.01					Change 4 of 5
 */

		if (System7OrLater) {
			sys7mask = NewRgn();
			tmprgn = NewRgn();
			if (BitMapToRegion(tmprgn, &(w[BifD].Back->portBits)) != noErr) {
				DisposeRgn(sys7mask);
				sys7mask = NIL;
			}
			else {
				RectRgn(sys7mask, &w[BifD].GR);
				DiffRgn(sys7mask, tmprgn, sys7mask);
			}
			DisposeRgn(tmprgn);
		}

/*	END changes for System 7, BB 2.01					Change 4 of 5
 */

		for(i=0; i<=BD_Rec.DiffSolutions+1; ++i) {
			LeftEnd = RightEnd;
			if (i==BD_Rec.DiffSolutions+1) {
				RightEnd = w[BifD].GR.right - w[BifD].GR.left;
				TheColor = BlackColorIndex;
			}
			else {
				if (i==BD_Rec.DiffSolutions) {
/* change 8/27/91, re-beta
					if (NeedToColorAfter)
						RightEnd = BD_Rec.CurrentXInt - 1;
					else
						RightEnd = BD_Rec.CurrentXInt;
*/
					if (NeedToColorAfter)
						RightEnd = BD_Rec.CurrentXInt;
					else
						RightEnd = BD_Rec.CurrentXInt + 1;
				}
				else
					RightEnd = (*BD_Rec.SolutionTypes)[i+1].h  /* - 1 (1.01 change 5/20/92; beta?) */ ;
				TheColor = csi[(*BD_Rec.SolutionTypes)[i].v];
			}
			tmpRect = w[BifD].GR;
			tmpRect.left = w[BifD].GR.left + LeftEnd;
			tmpRect.right = w[BifD].GR.left + RightEnd;

/*	1.01 change 5/20/92. Not necessary, as will treat as null rect anyway.
 */
			if (tmpRect.right < tmpRect.left)
				tmpRect.right = tmpRect.left;
/*	End, 1.01 change
 */

			OffsetRect(&tmpRect, -h, -v);
			if (SectRect(&tmpRect, pr, &tmpRect)) {
				SetSpecialColor(TheColor, gp == (GrafPtr) w[BifD].Peek);

/*	BEGIN changes for System 7, BB 2.01					Change 4 of 5
 */
 			/* OLD: change adMin to srcOr/CopyMask for Sys 7.
			 *	Can't use arithmetic transfer modes to colorize bit maps.
				myCopyBits(BD_Rec.Graph, gp,
						&tmpRect, h, v, 0, 0, adMin, (RgnHandle) NIL);
			 */
				if (System7OrLater)
					myCopyBits(BD_Rec.Graph, gp,
							&tmpRect, h, v, 0, 0, srcOr, sys7mask);
				else
					myCopyBits(BD_Rec.Graph, gp,
							&tmpRect, h, v, 0, 0, adMin, (RgnHandle) NIL);

/*	END changes for System 7, BB 2.01					Change 4 of 5
 */

				SetSpecialColor(BlackColorIndex, gp == (GrafPtr) w[BifD].Peek);
/*
myCopyBits(BD_Rec.Graph, gp,
		&tmpRect, h, v, 0, 0, srcOr, (RgnHandle) NIL);
PenMode(patXor);
FrameRect(&tmpRect);
PenMode(patCopy);
*/
			}
		}

/*	BEGIN changes for System 7, BB 2.01					Change 4 of 5
 */

		if (System7OrLater) {
			if (sys7mask != NIL)
				DisposeRgn(sys7mask);
		}

/*	END changes for System 7, BB 2.01					Change 4 of 5
 */

	}
	else
		myCopyBits(BD_Rec.Graph, gp,
				pr, h, v, 0, 0, srcOr, (RgnHandle) NIL);
}

ChangeBDPhaseRange(double NewMinPhase)
{
	int		PixelOffset;
	GrafPtr	tmpGP = NOPORT;
	Rect	tr = BD_Rec.Graph->portRect, tr2;

	if (!CreateSpecialOffPort(&tmpGP, &tr)) {
		ERROR("\PNot enough memory to change Bifurcation Diagram phase range.");
		return;
	}

	if (NewMinPhase > BD_Rec.MinPhase) {
		PixelOffset = yFloatToCoord(BifD, GetBDOffsetPhase(NewMinPhase));
		NewMinPhase = GetBDUnOffsetPhase(yCoordToFloat(BifD, PixelOffset));
	}
	else {
		PixelOffset = yFloatToCoord(BifD, GetBDOffsetPhase(NewMinPhase+1.0));
		NewMinPhase = GetBDUnOffsetPhase(yCoordToFloat(BifD, PixelOffset))
				- 1.0;
	}

	tr2 = tr;
	tr2.top = tr2.bottom - PixelOffset;
	myCopyBits(BD_Rec.Graph, tmpGP, &tr2, 0, 0, 0,
			-(tr.bottom - tr.top - PixelOffset), srcCopy, NIL);
	tr2 = tr;
	tr2.bottom -= PixelOffset;
	myCopyBits(BD_Rec.Graph, tmpGP, &tr2, 0, 0, 0, PixelOffset, srcCopy, NIL);
	myCopyBits(tmpGP, BD_Rec.Graph, &tr, 0, 0, 0, 0, srcCopy, NIL);

	DisposeOffPort(tmpGP);

	if (fabs(NewMinPhase) < AlmostZero)
		NewMinPhase = 0.0;
	BD_Rec.MinPhase = NewMinPhase;
	BD_Rec.YAxis[xmin_i] = NewMinPhase;
	BD_Rec.YAxis[xmax_i] = NewMinPhase + 1.0;
	if (fabs(BD_Rec.YAxis[xmax_i]) < AlmostZero)
		BD_Rec.YAxis[xmax_i] = 0.0;
	w[BifD].DrawFromScratch = TRUE;
	StdForceUpdate((WindowPtr) w[BifD].Peek);
	BDYFieldWidth = StdIMFieldWidth;
		/*	Since y-range is 1.0 (AllPhases); see InitBDScales;
		 *	this is only needed if BD is opened, not created;
		 *	DrawBDBackground needs this, but nothing else does,
		 *	so normally it does not matter that BDYFieldWidth is not set.
		 */
}


			/*	------------------------------------------	*/
			/*	----------	Iteration Routines	----------	*/
			/*	------------------------------------------	*/

Boolean InitNextBD()
{
	Boolean	GoodInit;
	double	f;

	ResetSolutions();
	if (BD_Rec.CurrentXInt == BD_Rec.MaxXInt)
		f = w[BifD].GraphRange[xmax_i];
	else
		f = xCoordToFloat(BifD, BD_Rec.CurrentXInt);

	if (BD_Rec.XAxisIsAmplitude)
		A = f;
	else {
		freq = f;
		SetFrequencyParams();
	}
	if (NeedOldInitCond) {
		v0 = BD_Rec.v00;
		delta = BD_Rec.delta0;
	}
	else {
		v0 = vj;
		delta = phase;
	}
	GoodInit = Initialize_Iteration_Globals(TRUE, TRUE);

	DrawCurrentBDBackground();
	if (!GoodInit) {
		ERROR("\PCannot continue bifurcation diagram at new x-axis value");
		return(FALSE);
	}

	(*BD_Rec.InitConds)[BD_Rec.CurrentXInt].vel = v0;
	(*BD_Rec.InitConds)[BD_Rec.CurrentXInt].del = delta;

	CIITimesInc = CurrentInitIterTimes =
			GetSpecialTimes(BD_Rec.InitIterTimes, BD_Rec.DefaultIITimes);
	CurrentShowTimes = MaxBDShowTimes;
	NumCIITimesIncs = 1;

	if (NeedToTurnOnGraphing) {
		BD_Rec.GraphInAllWindows = TRUE;
		CheckGraphInAll(BifD);
		SetTextTitle(TRUE, BifD);
		ActivateTextScrolls(FALSE);
		NeedToTurnOnGraphing = FALSE;
	}
	if (BD_Rec.GraphInAllWindows) {
		RedrawSine = TRUE;
		InitTrajScales();
		SetTrajCoords();
/*
		w[BifD].WindowCleared = FALSE;
*/
		NewGraphics(TRUE, FALSE);
		NeedToSetTxOutHeader = NeedToSetTSerHeader = TRUE;
/*
		SetHeaders(A, freq, damp);
*/
	}

	CurBDSolutionType = NoSolutionType;
	NeedOldInitCond = FALSE;
	NeedToSaveBD = TRUE;
	NeedToColorAfter = FALSE;
	return(TRUE);
}

Boolean GetNextBD()
{
	Rect	tr;

	if (HitNumber < CurrentInitIterTimes + CurrentShowTimes)
		return(TRUE);

	if (BD_Rec.InColor && NeedToColorAfter) {
		tr = w[BifD].GR;
		tr.left = w[BifD].GR.left + BD_Rec.CurrentXInt;
		tr.right = tr.left + 1;
		OffsetRect(&tr, -GetCtlValue(w[BifD].BSB), -GetCtlValue(w[BifD].RSB));
/* added 8/27/91, re-beta */
NeedToColorAfter = FALSE;
		ScrollCopyBackRect(BifD, &tr, TRUE);
	}

	if ( ++BD_Rec.CurrentXInt > BD_Rec.MaxXInt )
		return(FALSE);
	else {
		IterationStatus = GRAPHING;	/*	will be advanced	*/
		return(InitNextBD());
	}
}

StartBifurcationDiagram()
{
	Boolean	NeedVMax;

	if (pCurrentVars != &BD_Rec.Vars)
		if (HitNumber > 0) {
			if (!CheckNormalSaves(FALSE))
				return(FALSE);
/*
			if (pCurrentVars == &NormalVars)
				if (!DoSpecialWarningDialog("\Pbifurcation"))
					return(FALSE);
*/
			NewGraphics(TRUE, TRUE);
		}

	if (!WindowOpen(BifD))
		if (DoShowWindowDialog(BifD))
			BBOpenWindow((WindowPtr) w[BifD].Peek);

	ResetSolutions();
	StopIterating(ProgramStatus == BASINS_OF_ATTRACTION);
	NeedVMax = damp != BD_Rec.Vars._damp;
/*
	if (!BD_Rec.GraphInAllWindows)
		HandleAllUpdates();
*/

	pCurrentVars = &BD_Rec.Vars;
	SetFrequencyParams();
	if ( (BD_Rec.GraphInAllWindows) && (NeedVMax) ) {
		SetVMax(TRUE);
		InitImpactMapScales();
	}

	MaxBDShowTimes = ( (BD_Rec.ShowTimes != AutoTimes) ?
			BD_Rec.ShowTimes :
			MAX( MinAutoShow,
			AutoShowMultiplier * (w[BifD].GR.bottom-w[BifD].GR.top+2) ) );
	BD_MaxTimes = ( (BD_Rec.InitIterTimes != AutoTimes) ?
			BD_Rec.InitIterTimes :
			(BD_Rec.DefaultIITimes * MaxCIITimesIncs) ) +
			MaxBDShowTimes;
	if (BD_Points != NIL)
		DisposHandle((Handle) BD_Points);
	BD_Points = (int**)
			NewHandle( ((long) BD_MaxTimes+1) * sizeof(int) );

	if (BD_Rec.InitConds != NIL)
		DisposHandle((Handle) BD_Rec.InitConds);
	BD_Rec.InitConds = (ImpactSet**)
			NewHandle( ((long) (w[BifD].GR.right-w[BifD].GR.left+1)) *
			sizeof(ImpactSet) );

	BD_Rec.MaxDiffSolutions = DiffSolutionInc;
	if (BD_Rec.SolutionTypes != NIL)
		DisposHandle((Handle) BD_Rec.SolutionTypes);
	BD_Rec.SolutionTypes = (Point**)
			NewHandle( ((long) BD_Rec.MaxDiffSolutions) * sizeof(Point) );
	BD_Rec.DiffSolutions = -1;

	NewSpecial(BifD);
	BD_Rec.CurrentXInt = 0;
	NeedToTurnOnGraphing = FALSE;
	NeedOldInitCond = FALSE;

	vj = BD_Rec.v00;					/*	vj, phase for InitNextBD()	*/
	phase = BD_Rec.delta0;
	NeedToColorAfter = FALSE;
	if (!InitNextBD())
		return(FALSE);
	if (!StartIterating(BIFURCATION_DIAGRAM)) {
		SpecialMenuEnable(BifD, FALSE, FALSE);
		return(FALSE);
	}
	SpecialMenuEnable(BifD, TRUE, TRUE);

	return(TRUE);
}

ContinueBifurcationDiagram()
{
	Boolean	NeedVMax;

	if (ProgramStatus < BIFURCATION_DIAGRAM)
		if (HitNumber > 0) {
/*
			if (!DoSpecialWarningDialog("\Pbifurcation"))
*/
			if (!CheckNormalSaves(FALSE))
				return(FALSE);
			NewGraphics(TRUE, TRUE);
		}

	if (!WindowOpen(BifD))
		if (DoShowWindowDialog(BifD))
			BBOpenWindow((WindowPtr) w[BifD].Peek);

	ResetSolutions();
	StopIterating(ProgramStatus == BASINS_OF_ATTRACTION);
	NeedVMax = damp != BD_Rec.Vars._damp;
/*
	if (!BD_Rec.GraphInAllWindows)
		HandleAllUpdates();
*/

	pCurrentVars = &BD_Rec.Vars;
	SetFrequencyParams();

	if ( (BD_Rec.GraphInAllWindows) && (NeedVMax) ) {
		SetVMax(TRUE);
		InitImpactMapScales();
	}

		/*	For InitNextBD()	*/
	vj = v0;
	phase = delta;

	NeedToColorAfter = FALSE;
	if (!InitNextBD())
		return;
	if (!StartIterating(BIFURCATION_DIAGRAM))
		return;
}

GetBifDiagInfo(Boolean Magnifying)
{
	Boolean		StartBD;
	double		tnd;	/*	tmp new delta	*/
	DoubleRect	MagRange;
	int			MagXStart;
	Boolean		OldGIAW;
	double		min, max;
	Boolean		GoodOp;
	int			tmpCurMods;

	tmpBD = BD_Rec;

	if (Magnifying) {
		MagXStart = BDBox.left - w[BifD].Origin.h;
		MagRange[xmin_i] = xCoordToFloat(BifD, MagXStart);
		MagRange[xmax_i] = xCoordToFloat(BifD, BDBox.right-w[BifD].Origin.h);
		MagRange[ymin_i] = yCoordToFloat(BifD, w[BifD].Origin.v-BDBox.bottom);
		MagRange[ymax_i] = yCoordToFloat(BifD, w[BifD].Origin.v-BDBox.top);

		if (tmpBD.XAxisIsAmplitude) {
			Round2Doubles(MagRange[xmin_i], MagRange[xmax_i], RelPlaces, &min, &max);
			MagRange[xmin_i] = (BDBox.left == w[BifD].GR.left) ?
					tmpBD.XAxis[xmin_i] : min;
			MagRange[xmax_i] = (BDBox.right == w[BifD].GR.right-1) ?
					tmpBD.XAxis[xmax_i] : max;
/*
			if (MagRange[xmin_i] != tmpBD.XAxis[xmin_i])
				MagRange[xmin_i] = min;
			if (MagRange[xmax_i] != tmpBD.XAxis[xmax_i])
				MagRange[xmax_i] = max;
*/
			tmpBD.XAxis[xmin_i] = MagRange[xmin_i];
			tmpBD.XAxis[xmax_i] = MagRange[xmax_i];
		}
		else {
			Round2Doubles(MagRange[xmin_i], MagRange[xmax_i], RelPlaces, &min, &max);
			MagRange[xmin_i] = (BDBox.left == w[BifD].GR.left) ?
					tmpBD.XAxis[ymin_i] : min;
			MagRange[xmax_i] = (BDBox.right == w[BifD].GR.right-1) ?
					tmpBD.XAxis[ymax_i] : max;
/*
			if (MagRange[xmin_i] != tmpBD.XAxis[ymin_i])
				MagRange[xmin_i] = min;
			if (MagRange[xmax_i] != tmpBD.XAxis[ymax_i])
				MagRange[xmax_i] = max;
*/
			tmpBD.XAxis[ymin_i] = MagRange[xmin_i];
			tmpBD.XAxis[ymax_i] = MagRange[xmax_i];
		}
		if (tmpBD.YAxisIsPhase) {
			MagRange[ymin_i] = GetBDUnOffsetPhase(MagRange[ymin_i]);
			MagRange[ymax_i] = GetBDUnOffsetPhase(MagRange[ymax_i]);

			Round2Doubles(MagRange[ymin_i], MagRange[ymax_i], RelPlaces, &min, &max);
			MagRange[ymin_i] = (BDBox.bottom == w[BifD].GR.bottom-1) ?
					BD_Rec.MinPhase : min;
			MagRange[ymax_i] = (BDBox.top == w[BifD].GR.top) ?
					BD_Rec.MinPhase + w[BifD].GraphRange[ymax_i] : max;
/*
			if (MagRange[ymin_i] > BD_Rec.MinPhase)
				MagRange[ymin_i] = min;
			if (MagRange[ymax_i] < BD_Rec.MinPhase + w[BifD].GraphRange[ymax_i])
				MagRange[ymax_i] = max;
*/

			tmpBD.YAxis[xmin_i] = MagRange[ymin_i];
			tmpBD.YAxis[xmax_i] = MagRange[ymax_i];
		}
		else {
			Round2Doubles(MagRange[ymin_i], MagRange[ymax_i], RelPlaces, &min, &max);
			MagRange[ymin_i] = (BDBox.bottom == w[BifD].GR.bottom-1) ?
					tmpBD.YAxis[ymin_i] : min;
			MagRange[ymax_i] = (BDBox.top == w[BifD].GR.top) ?
					tmpBD.YAxis[ymax_i] : max;
/*
			if (MagRange[ymin_i] != tmpBD.YAxis[ymin_i])
				MagRange[ymin_i] = min;
			if (MagRange[ymax_i] != tmpBD.YAxis[ymax_i])
				MagRange[ymax_i] = max;
*/
			tmpBD.YAxis[ymin_i] = MagRange[ymin_i];
			tmpBD.YAxis[ymax_i] = MagRange[ymax_i];
		}

		MagXStart = MIN(MagXStart, tmpBD.CurrentXInt);
		if (MagXStart >= 0) {
			tmpBD.v00 = (*tmpBD.InitConds)[MagXStart].vel;
			tmpBD.delta0 = (*tmpBD.InitConds)[MagXStart].del;
		}
	}
	else if (!DrawBD) {
		tmpBD.Vars = NormalVars;
		tmpBD.XAxisIsAmplitude = TRUE;
		tmpBD.YAxisIsPhase = TRUE;

			/*	x, y both apply to x-axis here	*/
		tmpBD.XAxis[xmin_i] = tmpBD.Vars._A;
		tmpBD.XAxis[xmax_i] = 2.0 * tmpBD.XAxis[xmin_i];
		tmpBD.XAxis[ymin_i] = tmpBD.Vars._freq;
		tmpBD.XAxis[ymax_i] = 2.0 * tmpBD.XAxis[ymin_i];

		tmpBD.YAxis[xmin_i] = tmpBD.MinPhase = MinPhase;
		tmpBD.YAxis[xmax_i] = MaxPhase;
		tmpBD.YAxis[ymin_i] = VMin;
		tmpBD.YAxis[ymax_i] = VMax;

		tmpBD.v00 = tmpBD.Vars._v0;
		tmpBD.delta0 = tmpBD.Vars._delta;
		tmpBD.DefaultIITimes = GetDefaultInitIterTimes(tmpBD.Vars._damp);
		tmpBD.InitIterTimes = AutoTimes;
		tmpBD.ShowTimes = AutoTimes;
		tmpBD.GraphInAllWindows = BD_Rec.GraphInAllWindows;
		tmpBD.FullWindow = TRUE;
		tmpBD.InColor = BD_Rec.InColor;
	}

	pbd = &tmpBD;
	
	if ( !Magnifying || !SHIFT_DOWN)
		GoodOp = DoBifurcationDiagramDialog(&StartBD);
	else
		GoodOp = TRUE;
	if (GoodOp) {
		if (Magnifying) {
			tmpCurMods = CurrentModifiers;
			CurrentModifiers = 0;
		}
		if (!CheckOneSave(FALSE, BifD))
			GoodOp = FALSE;
		else
			NeedToSaveBD = FALSE;
		if (Magnifying)
			CurrentModifiers = tmpCurMods;
	}

	if (GoodOp) {
		DrawBD = TRUE;
		OldGIAW = BD_Rec.GraphInAllWindows;
		BD_Rec = tmpBD;
		BD_Rec.DefaultIITimes =
			GetDefaultInitIterTimes(BD_Rec.Vars._damp);
		tnd = BD_Rec.delta0;
	 	if (tnd<0.0)	tnd += -((int) tnd)+1;
	 	if (tnd>=1.0)	tnd -= (int) tnd;
	 	BD_Rec.delta0 = tnd;
	 		/* Set the next two and _A, _freq below for DrawBDBackground()	*/
	 	BD_Rec.Vars._v0 = BD_Rec.v00;
	 	BD_Rec.Vars._delta = BD_Rec.delta0;

	 	if (BD_Rec.XAxisIsAmplitude) {
	 		w[BifD].GraphRange[xmin_i] = BD_Rec.XAxis[xmin_i];
	 		w[BifD].GraphRange[xmax_i] = BD_Rec.XAxis[xmax_i];
	 		BD_Rec.Vars._A = BD_Rec.XAxis[xmin_i];
		}
		else {
	 		w[BifD].GraphRange[xmin_i] = BD_Rec.XAxis[ymin_i];
	 		w[BifD].GraphRange[xmax_i] = BD_Rec.XAxis[ymax_i];
	 		BD_Rec.Vars._freq = BD_Rec.XAxis[ymin_i];
		}
	 	if (BD_Rec.YAxisIsPhase) {
	 			/*	GraphRange is Offset, YAxis is not	*/
			BD_Rec.MinPhase = BD_Rec.YAxis[xmin_i];
	 		w[BifD].GraphRange[ymin_i] = 0.0;
	 				/*	GetBDOffsetPhase(BD_Rec.YAxis[xmin_i]);	*/
	 		w[BifD].GraphRange[ymax_i] = GetBDOffsetPhase(BD_Rec.YAxis[xmax_i]);
		}
		else {
	 		w[BifD].GraphRange[ymin_i] = BD_Rec.YAxis[ymin_i];
	 		w[BifD].GraphRange[ymax_i] = BD_Rec.YAxis[ymax_i];
		}
		BD_Rec.AllPhases = BD_Rec.YAxisIsPhase &&
				(w[BifD].GraphRange[ymax_i] -
				w[BifD].GraphRange[ymin_i] >= 1.0-AlmostZero);

		if (BD_Rec.FullWindow)
			BD_Rec.WindowSize = LWR;
		else {
			BD_Rec.WindowSize = w[BifD].Peek->port.portRect;
			OffsetRect(&BD_Rec.WindowSize, -w[BifD].Peek->port.portBits.bounds.left,
					-w[BifD].Peek->port.portBits.bounds.top);
			/*	see GetDefaultWindowConfiguration discussion, bbWindows.c	*/
		}
		SetSpecialScrolling(BifD);

		w[BifD].xGraphAroundZero = FALSE;
		w[BifD].yGraphAroundZero =
				(w[BifD].GraphRange[ymin_i] * w[BifD].GraphRange[ymax_i] <= 0);
		w[BifD].DrawFromScratch = TRUE;

		if (ProgramStatus == BIFURCATION_DIAGRAM) {
			BD_Rec.GraphInAllWindows = OldGIAW;
			if (OldGIAW != tmpBD.GraphInAllWindows) {
				if (tmpBD.GraphInAllWindows)
					ProgramStatus = WAITING;
				SwitchGraphInAll(BifD);
				ProgramStatus = BIFURCATION_DIAGRAM;
			}
		}
		else
			CheckGraphInAll(BifD);
		CheckSpecialColor(BifD);

		SetSpecialTitle(FALSE, BifD);
		BD_Rec.CurrentXInt = -1;
		BD_Rec.DiffSolutions = -1;
		if (!InitBDScales()) {
			ERROR("\PNot enough memory to run bifurcation diagram");
			ClearBifurcationDiagram();
			return;
		}
		SpecialMenuEnable(BifD, TRUE, FALSE);
		if (StartBD)
			StartBifurcationDiagram();
		else {
			NewSpecial(BifD);
			if (ProgramStatus == BIFURCATION_DIAGRAM)
				StopIterating(FALSE);
		}
	}

}

Boolean	TrackBDRect(Point pt)
{
	SetPort(w[BifD].Peek);

	if ( TrackDashedRect(pt.h, pt.v, &BD_Rec.LocalGR, &BDBox, TRUE) ) {
		OffsetRect(&BDBox,GetCtlValue(w[BifD].BSB),GetCtlValue(w[BifD].RSB));

			/*
			 *		The next step is something of a kludge.
			 *	Its necessity derives from the assymetry of
			 *	Mac Rect drawing--because the rect is framed on the
			 *	inside, on the left/top sides, the rect is drawn
			 *	to the right of/below the rect, while on the right/
			 *	bottom sides, the rect is drawn to the left/above.
			 *	Lines or points are drawn to the right/below.
			 *		This problem could perhaps also be solved by
			 *	figuring coordinates ([xy]FloatToCoord()) by truncating
			 *	instead of rounding, but (a) that has problems itself in,
			 *	say, the Bifurcation Diagram, where each x-axis
			 *	pixel corresponds to a single value of the amplitude
			 *	or frequency, instead of a range of values, as it
			 *	would elsewhere, and (b) at this point it would be
			 *	a great deal of work to fix.
			 *		Note that the background frame drawn (DrawBDBackground())
			 *	and BD_Rec.LocalGR (SetSpecialControls())
			 *	include corresponding kludges.
			 *		These kludges are not present in similar situations
			 *	of the Impact Map and Impact Phases windows, because
			 *	these windows behave a little differently--in particular,
			 *	the user does not expect a value to be plotted, say,
			 *	from 0.01 to 0.02.
			 *												-TA 3/1/91
			 */
		if (BDBox.right > BDBox.left)
			--BDBox.right;
		if (BDBox.bottom > BDBox.top)
			--BDBox.bottom;
/*	For BasAttr, lower bottom (velocity) but not right (phase)	*/

		return(TRUE);
	}
	return(FALSE);
}

			/*	------------------------------------------	*/
			/*	----------	Graphics Routines	----------	*/
			/*	------------------------------------------	*/

Boolean InitBDScales()
{
	Rect	tmpRect;
	double	range;
	int		i;
	int		MaxYInt;

	SetPort(w[BifD].Peek);
	w[BifD].DrawFromScratch = TRUE;
	StdForceUpdate((GrafPtr) w[BifD].Peek);

		/*	Calculate how much space needed for axis labels	*/
	if (w[BifD].Magnifying) {
		range = w[BifD].GraphRange[ymax_i]-w[BifD].GraphRange[ymin_i];
		for(i=-1;i<2;++i) {
			if (range > 1.0) break;
			range *= 10;
		}
		BDYFieldWidth = StdIMFieldWidth + MAX(i,0);

		range = w[BifD].GraphRange[xmax_i]-w[BifD].GraphRange[xmin_i];
		for(i=-2;i<4;++i) {
			if (range > 1.0) break;
			range *= 10;
		}
		BDXFieldWidth = StdIMFieldWidth + MAX(i,0);
	}
	else
		BDYFieldWidth = BDXFieldWidth = StdIMFieldWidth;

	tmpRect = BD_Rec.WindowSize;
	OffsetRect(&tmpRect, -tmpRect.left, -tmpRect.top);
	tmpRect.left += (StdMargin + (BDYFieldWidth+3)*CharWidth('0'));
	tmpRect.top	+= IM_TopMargin;
	tmpRect.right -= (ScrollBarWidth + StdMargin +
			BDXFieldWidth * CharWidth('0') / 2);
	tmpRect.bottom -= BD_BottomMargin;
	w[BifD].GR = tmpRect;

	if (!CreateSpecialOffPort(&BD_Rec.Graph, &w[BifD].GR))
		return(FALSE);

	BD_Rec.MaxXInt = (tmpRect.right - tmpRect.left);
	--BD_Rec.MaxXInt;
	w[BifD].xScale = (double) BD_Rec.MaxXInt /
			(w[BifD].GraphRange[xmax_i] - w[BifD].GraphRange[xmin_i]);

/*
	ScaledLength = tmpRect.bottom - tmpRect.top;
	if (!BD_Rec.AllPhases)
		--ScaledLength;
*/
	MaxYInt = tmpRect.bottom - tmpRect.top;
	if ( (!BD_Rec.YAxisIsPhase) || (!BD_Rec.AllPhases) )
		--MaxYInt;
	w[BifD].yScale = (double) (MaxYInt) /
			(w[BifD].GraphRange[ymax_i] - w[BifD].GraphRange[ymin_i]);
	if ( (!BD_Rec.YAxisIsPhase) && (w[BifD].yGraphAroundZero) )
		w[BifD].Origin.v = w[BifD].GR.top +
				RoundToInt(w[BifD].GraphRange[ymax_i]*w[BifD].yScale);
	else
		w[BifD].Origin.v = w[BifD].GR.bottom - 1;
	w[BifD].Origin.h = w[BifD].GR.left;
	w[BifD].BackOrigin = w[BifD].Origin;
	SetRect(&w[BifD].PR,
			w[BifD].GR.left - w[BifD].Origin.h,
			w[BifD].Origin.v - w[BifD].GR.bottom,
			w[BifD].GR.right - w[BifD].Origin.h,
			w[BifD].Origin.v - w[BifD].GR.top);
			/* +1 to account for pixel hanging below and to left of point */

	tmpRect.top = tmpRect.bottom;
	tmpRect.bottom = BD_Rec.WindowSize.bottom - BD_Rec.WindowSize.top - (ScrollBarWidth-1);
	InsetRect(&tmpRect, -1, 0);
	--tmpRect.right;
	BD_Rec.HS0_Rect = tmpRect;

	tmpRect = w[BifD].GR;
	tmpRect.right = tmpRect.left-1;
	tmpRect.left = 0;
	InsetRect(&tmpRect, 0, -1);
	--tmpRect.bottom;
	BD_Rec.HS1_Rect = tmpRect;

	SetSpecialControls(BifD);
	return(TRUE);
}

DrawCurrentBDBackground()
{
	Str255	s;
	Rect	tmpRect, theClipRect;
	int		theOffset;

	s[0] = 0;

	SetPort(w[BifD].Back);
	StringAddVar(CP(s), BD_Rec.Vars._v0, "\PV0: ", TRUE);
	 sAppend(CP(s), CP(BD_Spaces));
	StringAddVar(CP(s), BD_Rec.Vars._delta, "\PDelta: ", TRUE);
	 sAppend(CP(s), CP(BD_Spaces));
	if (BD_Rec.XAxisIsAmplitude)
		StringAddVar(CP(s), BD_Rec.Vars._A, "\PAmplitude: ", TRUE);
	else
		StringAddVar(CP(s), BD_Rec.Vars._freq, "\PFrequency: ", TRUE);

	SetRect(&tmpRect, StdMargin,
			w[BifD].GR.bottom + 5*TextLineHeight + StdMargin,
			w[BifD].Back->portRect.right, w[BifD].Back->portRect.bottom);
	theOffset = BD_Rec.WindowSize.right - BD_Rec.WindowSize.left -
			ScrollBarWidth - 2*StdMargin - StringWidth(CP(s));
	if (theOffset > 0)
		OffsetRect(&tmpRect, theOffset/2, 0);

	TextBox(CP(s[1]), (long) (s[0]), &tmpRect, teJustLeft);

	SetPort((WindowPtr) w[BifD].Peek);
	OffsetRect(&tmpRect, -GetCtlValue(w[BifD].BSB),
			-GetCtlValue(w[BifD].RSB));
	GetStdDocWindowClipRect(&theClipRect, (WindowPtr) w[BifD].Peek);
	if (SectRect(&tmpRect, &theClipRect, &tmpRect))
		InvalRect(&tmpRect);
}

DrawBDBackground()
{
	Rect		tmpRect;
	int			i, x, y;
	Str255		s, s1, AxisLabel;
	Str255		LocalXMaxString, LocalXMinString;
	Str255		LocalYMaxString, LocalYMinString;
	double		ymin, ymax;
	int			MaxXStrLen, OverflowOnRight;
	int			BottomTextSize, BottomTextOffset;

	int			MinXHashLabelCoord, MaxXHashLabelCoord;
	double		ScaledMax, Multiplier, HashMeasurement;
	int			ScaledMaxInt, HashInc, HashIndex;
	int			XWriteCoord, XWriteStart, XWriteStop;
	int			swidth;


	tmpRect = w[BifD].GR;
	InsetRect(&tmpRect, -1, -1);
/*
	tmpRect.right += 1;
*/
/*	old comment
	tmpRect.bottom += 1;
*/
	FrameRect(&tmpRect);

	/* Document vertical axis */

		ymin = w[BifD].GraphRange[ymin_i];
		ymax = w[BifD].GraphRange[ymax_i];
		if (BD_Rec.YAxisIsPhase) {
			ymin = GetBDUnOffsetPhase(ymin);
			ymax = GetBDUnOffsetPhase(ymax);
			for(i=0;i<4;++i) {
				y = yFloatToCoord(BifD, GetBDOffsetPhase(i/4.0));
				if ( (y<=0) || (y>=w[BifD].GR.bottom-w[BifD].GR.top-1) )
					continue;
				WindowMoveTo(BifD, w[BifD].GR.left-w[BifD].Origin.h+HashSize, y);
				Line(-HashSize, 0);
				Move(w[BifD].GR.right - w[BifD].GR.left-1, 0);
				Line(-HashSize, 0);
			}
			sEquate(CP(AxisLabel), "\PPhase");
			if (w[BifD].yGraphAroundZero) {
				y=yFloatToCoord(BifD, GetBDOffsetPhase(0.0));
				if ( (y > TextHeight) && (y <
						w[BifD].GR.bottom - w[BifD].GR.top - TextHeight+1) ) {
					WindowMoveTo(BifD, w[BifD].GR.left-w[BifD].Origin.h, y);
					Move(-TextMargin - CharWidth('0'), TextHeight/2);
					DrawChar('0');
				}
			}
		}
		else {
			if (w[BifD].yGraphAroundZero) {
				y = yFloatToCoord(BifD, 0.0);
				WindowMoveTo(BifD, w[BifD].GR.left-w[BifD].Origin.h+HashSize, y);
				Line(-HashSize, 0);
				Move(w[BifD].GR.right - w[BifD].GR.left-1, 0);
				Line(-HashSize, 0);
				if ( (w[BifD].Origin.v - w[BifD].GR.top > TextHeight) &&
						(w[BifD].GR.bottom - w[BifD].Origin.v > TextHeight) ) {
					WindowMoveTo(BifD, 0, 0);
						Move(-TextMargin - CharWidth('0'), TextHeight/2);
						DrawChar('0');
				}
			}
			sEquate(CP(AxisLabel), "\PVelocity");
		}
		SetRect(&tmpRect, StdMargin+2,
				(w[BifD].GR.top+w[BifD].GR.bottom-((int)AxisLabel[0])*TextLineHeight)/2,
				StdMargin + 12/*CharWidth('V') + 1*/,
				w[BifD].GR.bottom + 2*TextLineHeight + StdMargin + 1);
		tmpRect.top = MAX(tmpRect.top, w[BifD].GR.top);
		TextBox(CP(AxisLabel[1]), (long) (AxisLabel[0]), &tmpRect, teJustCenter);

		myFToStrNoPad(ymin, CP(LocalYMinString),
				BDYFieldWidth, BDYFieldWidth-IMPrecDiff);
		myFToStrNoPad(ymax, CP(LocalYMaxString),
				BDYFieldWidth, BDYFieldWidth-IMPrecDiff);
		WindowMoveTo(BifD, 0, w[BifD].Origin.v - w[BifD].GR.top);
			Move(-TextMargin - StringWidth(&LocalYMaxString), TextHeight/2);
			DrawString(&LocalYMaxString);
		WindowMoveTo(BifD, 0, w[BifD].Origin.v - w[BifD].GR.bottom);
			Move(-TextMargin - StringWidth(&LocalYMinString), TextHeight/2-1);
			DrawString(&LocalYMinString);

	/* Document horizontal axis */

		if (BD_Rec.XAxisIsAmplitude) {
			sEquate(CP(AxisLabel), "\PAmplitude");
		}
		else {
			sEquate(CP(AxisLabel), "\PFrequency");
		}
		MoveTo( (w[BifD].GR.right + w[BifD].GR.left)/2, w[BifD].GR.bottom);
		Move(-StringWidth(CP(AxisLabel))/2,
				TextMargin + TextLineHeight + TextHeight - 1);
		DrawString(CP(AxisLabel));

		myFToStrNoPad(w[BifD].GraphRange[xmin_i],
				CP(LocalXMinString),
				BDXFieldWidth, BDXFieldWidth - IMPrecDiff);
		myFToStrNoPad(w[BifD].GraphRange[xmax_i],
				CP(LocalXMaxString),
				BDXFieldWidth, BDXFieldWidth - IMPrecDiff);
		MoveTo(w[BifD].GR.left, w[BifD].GR.bottom);
		Move(-StringWidth(CP(LocalXMinString))/2, TextMargin + TextHeight-1);
		DrawString(CP(LocalXMinString));
		MoveTo(w[BifD].GR.right, w[BifD].GR.bottom);
		MaxXStrLen = StringWidth(CP(LocalXMaxString));
		OverflowOnRight = MAX(MaxXStrLen/2 - ImpactRightMargin, 0);
		Move(-MaxXStrLen/2 - OverflowOnRight+2, TextMargin + TextHeight-1);
		DrawString(CP(LocalXMaxString));
		MinXHashLabelCoord = StringWidth(CP(LocalXMinString))/2 +
				MinSpecialLabelSeparation;
		MaxXHashLabelCoord = w[BifD].GR.right - w[BifD].GR.left -
				MaxXStrLen/2 - OverflowOnRight - MinSpecialLabelSeparation;

		ScaledMax = w[BifD].GraphRange[xmax_i] - w[BifD].GraphRange[xmin_i];
		while (fabs(ScaledMax) > 50.0)	ScaledMax /= 10.0;
		while (fabs(ScaledMax) < 5.0)	ScaledMax *= 10.0;
		Multiplier = (w[BifD].GraphRange[xmax_i] - w[BifD].GraphRange[xmin_i]) /
				ScaledMax;
		ScaledMaxInt = RoundToInt(fabs(ScaledMax));
		HashInc = ScaledMaxInt / 10 + 1;
		if (w[BifD].GraphRange[xmin_i] > w[BifD].GraphRange[xmax_i])
			HashInc = -HashInc;
	
						/*	Draw, label hash marks	*/
/*
		for (HashIndex = (int) (w[BifD].GraphRange[xmin_i]/Multiplier);;) {
*/
/* Problem when (basically) ScaledMaxInt % HashInc != 0
		HashIndex = RoundToInt(w[BifD].GraphRange[xmin_i]/Multiplier);
		for (i=1; i<ScaledMaxInt/HashInc; ++i) {
*/
		if (w[BifD].GraphRange[xmin_i] < 0) {	/*	Should *never* be	*/
			if (w[BifD].xScale < 0.0) {	/*	Backwards	*/
				HashIndex = (int)
						ceil(w[BifD].GraphRange[xmin_i]/Multiplier-AlmostZero);
				HashIndex = (HashIndex/HashInc) * HashInc;
			}
			else {
				HashIndex = (int)
						floor(w[BifD].GraphRange[xmin_i]/Multiplier+AlmostZero);
				HashIndex = (HashIndex/HashInc) * HashInc;
				HashIndex -= HashInc;
			}
		}
		else {
			if (w[BifD].xScale < 0.0) {	/*	Backwards	*/
				HashIndex = (int)
						ceil(w[BifD].GraphRange[xmin_i]/Multiplier-AlmostZero);
				HashIndex = ((HashIndex-HashInc-1)/HashInc) * HashInc;
			}
			else {
				HashIndex = (int)
						floor(w[BifD].GraphRange[xmin_i]/Multiplier+AlmostZero);
				HashIndex = (HashIndex/HashInc) * HashInc;
			}
		}
/*
		HashIndex = RoundToInt(w[BifD].GraphRange[xmin_i]/Multiplier);
*/
		for(;;) {
			HashIndex += HashInc;
			HashMeasurement = HashIndex * Multiplier;
			XWriteCoord = xFloatToCoord(BifD, HashMeasurement);
			if (XWriteCoord >= BD_Rec.MaxXInt) break;
	
			myFToStrNoPad(HashMeasurement, CP(s), 0, PS_precision);
			WindowMoveTo(BifD, XWriteCoord,
					w[BifD].Origin.v - w[BifD].GR.top);
			Line(0, HashSize);
			Move(0, w[BifD].GR.bottom - w[BifD].GR.top - 2*HashSize-1/*!*/);
			Line(0, HashSize);
			swidth = StringWidth(CP(s)) - 1;
			XWriteStart = XWriteCoord - swidth/2;
			XWriteStop = XWriteCoord + swidth/2;
			if ( (XWriteStart > MinXHashLabelCoord) &&
					(XWriteStop < MaxXHashLabelCoord) ) {
				Move(-swidth/2, TextMargin + TextHeight /*-1*/);
				DrawString(CP(s));
				MinXHashLabelCoord = XWriteStop + MinSpecialLabelSeparation;
			}
		}

	/*	Document all parameter values	*/

		s[0] = 0;
		StringAddVar(CP(s), BD_Rec.v00, "\PInitial V0:    ", FALSE);
		 sAppend(CP(s), CP(BD_Spaces));
		StringAddVar(CP(s), BD_Rec.Vars._freq, "\PFrequency: ", FALSE);
		 sAppend(CP(s), CP(BD_Spaces));
		sAppend(CP(s), "\PInitial Iterations: ");
		if (BD_Rec.InitIterTimes == AutoTimes)
			sEquate(CP(s1), "\PAuto");
		else
			LongToString((long) BD_Rec.InitIterTimes, CP(s1));
		for(i=s1[0]; i<SpecialFieldWidth-1; ++i)
			sInsertC(CP(s), ' ', s[0]+1);
		sAppend(CP(s), CP(s1));
		sInsertC(CP(s), ReturnChar, s[0]+1);
		BottomTextSize = StringWidth(CP(s));

		StringAddVar(CP(s), BD_Rec.delta0, "\PInitial Delta: ", FALSE);
		 sAppend(CP(s), CP(BD_Spaces));
		StringAddVar(CP(s), BD_Rec.Vars._damp, "\PDamping:   ", FALSE);
		 sAppend(CP(s), CP(BD_Spaces));
		sAppend(CP(s), "\PShow Iterations:    ");
		if (BD_Rec.ShowTimes == AutoTimes)
			sEquate(CP(s1), "\PAuto");
		else
			LongToString((long) BD_Rec.ShowTimes, CP(s1));
		for(i=s1[0]; i<SpecialFieldWidth-1; ++i)
			sInsertC(CP(s), ' ', s[0]+1);
		sAppend(CP(s), CP(s1));

		SetRect(&tmpRect, StdMargin,
				w[BifD].GR.bottom + 2*TextLineHeight + 2*StdMargin + 1,
				w[BifD].Back->portRect.right, w[BifD].Back->portRect.bottom);
		BottomTextOffset = BD_Rec.WindowSize.right - BD_Rec.WindowSize.left -
				ScrollBarWidth - 2*StdMargin - BottomTextSize;
		if (BottomTextOffset > 0)
			OffsetRect(&tmpRect, BottomTextOffset/2, 0);
		TextBox(CP((s[1])), (long) (s[0]), &tmpRect, teJustLeft);

		MoveTo(0, w[BifD].GR.bottom + 2*TextLineHeight + StdMargin + 2);
		Line(BD_Rec.WindowSize.right - BD_Rec.WindowSize.left -
				ScrollBarWidth, 0);
		Move(0, 2*TextLineHeight + StdMargin + 2);
		Line(-(BD_Rec.WindowSize.right - BD_Rec.WindowSize.left -
				ScrollBarWidth), 0);
}

RefreshBD()
{
	if (!DrawBD) {
		SetPort(w[BifD].Peek);
		EraseRect(&w[BifD].Peek->port.portRect);
/*
		ScrollCopyBackToWindow(BifD);
*/
		return;
	}

	if (!w[BifD].DrawFromScratch) {
		ScrollCopyBackToWindow(BifD);
		return;
	}

	w[BifD].DrawFromScratch = FALSE;
	SetPort(w[BifD].Back);
	EraseRect(&w[BifD].Back->portRect);

	DrawBDBackground();
	DrawCurrentBDBackground();

	ScrollCopyBackToWindow(BifD);
}

PlotInBD(int yCoord)
{
	Point		pt;
	int			hOffset = GetCtlValue(w[BifD].BSB);
	int			vOffset = GetCtlValue(w[BifD].RSB);
	PenState	PS;

	w[BifD].WindowCleared = FALSE;

	SetPort(w[BifD].Peek);
	if (BD_Rec.InColor && !NeedToColorAfter) {
		GetPenState(&PS);
		PenMode(adMin);
		SetSpecialColor(csi[CurBDSolutionType], TRUE);
	}

	pt.h = BD_Rec.CurrentXInt;
	pt.v = yCoord;
	
	pt.h -= hOffset;	pt.v += vOffset;
	SetPort(w[BifD].Peek);	 WindowMoveTo(BifD, pt.h, pt.v);
	pt.h += hOffset;	pt.v -= vOffset;
	SetPort(BD_Rec.Graph);	 BackMoveTo(BifD, pt.h, pt.v);

	SetPort(w[BifD].Peek);	Line(0, 0);
	SetPort(BD_Rec.Graph);	Line(0, 0);

	SetPort(w[BifD].Peek);
	if (BD_Rec.InColor && !NeedToColorAfter) {
		GetPenState(&PS);
		PenMode(PS.pnMode);
		SetSpecialColor(BlackColorIndex, TRUE);
	}
}

int GetBDYPlot(double del, double vel)
{
	double	y = (BD_Rec.YAxisIsPhase) ? GetBDOffsetPhase(del) : vel;
	int		yint;

	yint = yFloatToCoord(BifD, y);
			/*	Next statement: See IMap discussion.	*/
	if (BD_Rec.YAxisIsPhase)
		if (yint > /* was >= */ w[BifD].Origin.v - w[BifD].GR.top)
/*
			if (	(!w[BifD].Magnifying) ||
					(w[BifD].GraphRange[ymax_i] -
					w[BifD].GraphRange[ymin_i] > 1.0-AlmostZero)	)
*/
			if (BD_Rec.AllPhases)
				yint = w[BifD].Origin.v - (w[BifD].GR.bottom - 1);
	if (yint == -MaxScreenY)
		yint = w[BifD].Origin.v+1 - w[BifD].GR.bottom - 1;
	else if (yint == MaxScreenY)
		yint = w[BifD].Origin.v - w[BifD].GR.top + 1;
	return(yint);
}

			/*	----------------------------------------------	*/
			/*	----------	Periodicity Routines	----------	*/
			/*	----------------------------------------------	*/

/*
#define MaxHNs	500
int	HNs[MaxHNs];
*/
SetBDSolutionType(int SolType)
{
	CurBDSolutionType = SolType;

	BDChanging = (CurrentInitIterTimes > CIITimesInc);

/*
if (!StopFileSave())
	SolType = SolType;
*/
/*
if (BD_Rec.CurrentXInt < MaxHNs)
	HNs[BD_Rec.CurrentXInt] = HitNumber;
*/
	if (BD_Rec.CurrentXInt != 0)
		if (SolType == (*BD_Rec.SolutionTypes)[BD_Rec.DiffSolutions].v)
			return;

	if (++BD_Rec.DiffSolutions >= BD_Rec.MaxDiffSolutions) {
		BD_Rec.MaxDiffSolutions += DiffSolutionInc;
		SetHandleSize( (Handle) BD_Rec.SolutionTypes,
				((long) BD_Rec.MaxDiffSolutions) * sizeof(Point) );
	}
	(*BD_Rec.SolutionTypes)[BD_Rec.DiffSolutions].h = BD_Rec.CurrentXInt;
	(*BD_Rec.SolutionTypes)[BD_Rec.DiffSolutions].v = SolType;
	BDChanging = TRUE;
}

/*
extern int SysVRef;
*/

CheckBDPeriodicity()
{
	int		i, per, k;
	int		yi0, yi1, yi2, yi3;
	int		ymin, ymax;
	int		MoreTimes;
	Boolean	NeedMore;
	int		FirstShow;
	Boolean	WillBePeriodic = FALSE, ProbPer = FALSE, GotAProbable = FALSE;
	static	int	ProbPeriod, ProbCount;

/*
Str255	s;
char	*pf = "\PP4 Conv. Test";
int		theVRef = SysVRef;
if (HitNumber == 0) {
	if ( !StdCreateFile(pf, theVRef, 'BBAL', 'TEXT') )
		i=i;
	if (!StartFileSave(pf, theVRef))
		i=i;
}
sprintf(CP(s), "%d\t%20.20g\r", HitNumber, phase);
CtoPstr(CP(s));
if (!WritePtrToFile(CP(s), (long) s[0]))
	i=i;
*/

	if (HitNumber == 0) {
		ProbPeriod = NoSolutionType;
		ProbCount = 0;
	}

	if (CurBDSolutionType != NoSolutionType)	return;

	if (StuckStatus != NOT_STUCK)	{
		SetBDSolutionType(StuckSolution);
		if (HitNumber <= CurrentInitIterTimes)
			CurrentInitIterTimes = HitNumber - 1;
		if (RootFindingError)
			CurrentShowTimes = 0;
		else
			CurrentShowTimes = (HitNumber - CurrentInitIterTimes) + 2;
		return;
	}

	for (i=Period1Solution; i<=Period8Solution; ++i) {
		switch (i) {
			case Period8Solution:	per = 8;	break;
			default:				per = i;	break;
		}
		if ( BDChanging && (HitNumber < CIITimesInc-MinProbableCount-2) )
			break;
		if (BD_Rec.DiffSolutions >= 0)
			if (i != (*BD_Rec.SolutionTypes)[BD_Rec.DiffSolutions].v)
				if (HitNumber < CIITimesInc-MinProbableCount-2)  continue;
		if (4*per-1 > HitNumber) break;

		MoreTimes = per;
		NeedMore = FALSE;
		for (k=0; k<per; ++k) {
			yi0 = (*BD_Points)[HitNumber - 0*per - k];
			yi1 = (*BD_Points)[HitNumber - 1*per - k];
			yi2 = (*BD_Points)[HitNumber - 2*per - k];
			yi3 = (*BD_Points)[HitNumber - 3*per - k];
			if (	(yi0 != yi1) || (yi0 != yi2) || (yi0 != yi3)	) {
				if ( (ABS(yi0-yi1)>1) || (ABS(yi0-yi2)>1) || (ABS(yi0-yi3)>1) )
					break;
				else
					if (HitNumber <= CurrentInitIterTimes)
						NeedMore = TRUE;
					else
						MoreTimes = per*3;
			}
				
			if (k==per-1) {
				if (!TestPeriodicity(per, &WillBePeriodic, &ProbPer)) {
					if (ProbPer || WillBePeriodic) {
/*
						if (per == ProbPeriod)	++ProbCount;
*/
						if (!GotAProbable) {
							if ( (ProbPeriod != NoSolutionType) &&
									( (ProbPeriod % per == 0) ||
									(per % ProbPeriod == 0) /* ??? */ ) )
										/*	OK to get tricked from 2 to 1	*/
										/*	1 to 2, 2 to 4, etc.;			*/
										/*	does ProbPeriod matterat all?	*/
								++ProbCount;
							else
								ProbCount = 1;
							ProbPeriod = per;
						}
						GotAProbable = TRUE;
					}
					if (WillBePeriodic || (ProbCount >= MinProbableCount))
						if (HitNumber >= CurrentInitIterTimes)
							if (BD_Rec.InitIterTimes == AutoTimes)
								if (++NumCIITimesIncs <= MaxCIITimesIncs)
									CurrentInitIterTimes += CIITimesInc;
					break;
				}
						/*	periodic, but iterate more	*/
				if (NeedMore)				return;
				SetBDSolutionType(i);
				FirstShow = HitNumber + (MoreTimes-per);
				CurrentInitIterTimes =
						MIN(FirstShow - 1, CurrentInitIterTimes);
				CurrentShowTimes = MIN(CurrentShowTimes,
						(FirstShow - 1 - CurrentInitIterTimes) + MoreTimes);
/*
				if (HitNumber <= CurrentInitIterTimes)
					CurrentInitIterTimes = HitNumber - 1;
				CurrentShowTimes = MIN(CurrentShowTimes,
						(HitNumber - CurrentInitIterTimes) + MoreTimes);
*/
				return;
			}
		}
	}

	if (!GotAProbable) {
		ProbPeriod = NoSolutionType;
		ProbCount = 0;
	}

	if (BD_Rec.InColor)
		if (HitNumber == CurrentInitIterTimes)
			NeedToColorAfter = TRUE;

	if (BD_Rec.ShowTimes == AutoTimes)
		if (HitNumber == CurrentInitIterTimes + MinAutoShow) {
			ymin = ymax = (*BD_Points)[0];
			for(i=1; i<=HitNumber; ++i) {
				if ((*BD_Points)[i] < ymin)	ymin = (*BD_Points)[i];
				if ((*BD_Points)[i] > ymax)	ymax = (*BD_Points)[i];
			}
			CurrentShowTimes = MAX(HitNumber - CurrentInitIterTimes,
					(ymax-ymin)*AutoShowMultiplier);
		}

	if (HitNumber == CurrentInitIterTimes + CurrentShowTimes)
		SetBDSolutionType(OtherSolution);
}

			/*	------------------------------------------	*/
			/*	----------	Graphing Routine	----------	*/
			/*	------------------------------------------	*/

GraphBD()
{
	int		yCoord = GetBDYPlot(phase, vj);

	(*BD_Points)[HitNumber] = yCoord;

	CheckBDPeriodicity();

	if (StuckStatus != NOT_STUCK)	/*	if stuck, don't use latest i.c.	*/
		NeedOldInitCond = TRUE;
	if (StuckStatus == STUCK_FOREVER) {
		StopStickingMessage();
		CurrentInitIterTimes = CurrentShowTimes = -1;	/* forces following	*/
	}
	if (NeedNextSpecialValue) {
		CurrentInitIterTimes = CurrentShowTimes = -1;	/* forces following	*/
		NeedNextSpecialValue = FALSE;
	}

	if (HitNumber > CurrentInitIterTimes)
		PlotInBD(yCoord);
}

ClearBifurcationDiagram()
{
	if (ProgramStatus == BIFURCATION_DIAGRAM)
		StopIterating(FALSE);
	SetPort(w[BifD].Back);
	EraseRect(&w[BifD].Back->portRect);
	StdForceUpdate((WindowPtr) w[BifD].Peek);
	DrawBD = FALSE;
	NeedToSaveBD = FALSE;
	SetSpecialTitle(FALSE, BifD);
	BD_Rec.GraphInAllWindows = BD_Rec.InColor = FALSE;
	CheckGraphInAll(BifD);
	CheckSpecialColor(BifD);
	if (BD_Rec.Graph != NOPORT) {
		DisposeOffPort(BD_Rec.Graph);
		BD_Rec.Graph = NOPORT;
	}

	SpecialMenuEnable(BifD, FALSE, FALSE);
	SetSpecialScrolling(BifD);
}


			/*	--------------------------------------	*/
			/*	----------	File Routines	----------	*/
			/*	--------------------------------------	*/


Boolean SaveBDReturn(Boolean SaveOKSoFar)
{
	Boolean	GoodOp;

	GoodOp = CloseSaveFile(!SaveOKSoFar);

/* deleted 11/11/91--causing resource problems?
	if ( ((**BackPic).picSize == GetHandleSize((Handle) BackPic)) &&
			((**BackPic).picSize != 0L) )
		KillPicture(BackPic);
	if ( ((**GraphPic).picSize == GetHandleSize((Handle) GraphPic)) &&
			((**GraphPic).picSize != 0L) )
		KillPicture(GraphPic);
*/
			/*	was returning valid but odd value for
			 *	GetHandleSize(GraphPic) some times, seemed to be when
			 *	replacing an existing file	*/

	return(SaveOKSoFar && GoodOp);
}

Boolean SaveBDFile()
{
	Boolean			GoodOp;
	int				PicID;
	ControlHandle	RSch, BSch;

	BackPic = GraphPic = NIL;

	if ( !WritePtrToFile((Ptr) &BD_Rec, sizeof(BifDiagRecord)) )
		return(SaveBDReturn(FALSE));

	RSch = w[BifD].RSB;
	w[BifD].RSB = (ControlHandle) GetCtlValue(RSch);
	BSch = w[BifD].BSB;
	w[BifD].BSB = (ControlHandle) GetCtlValue(BSch);
	GoodOp = WritePtrToFile((Ptr) &w[BifD], sizeof(BBWindow));
	w[BifD].RSB = RSch;	w[BifD].BSB = BSch;
	if (!GoodOp)
		return(SaveBDReturn(FALSE));

	if (BD_Rec.CurrentXInt >= 0)
		if ( !WriteHandleToFile((Handle) BD_Rec.InitConds,
				(MIN(BD_Rec.CurrentXInt,BD_Rec.MaxXInt) + 1) *
				sizeof(ImpactSet)) )
			return(SaveBDReturn(FALSE));

	if (BD_Rec.DiffSolutions >= 0)
		if (!WriteHandleToFile((Handle) BD_Rec.SolutionTypes,
				(BD_Rec.DiffSolutions+1) * sizeof(Point)) )
			return(SaveBDReturn(FALSE));

	if (BD_Rec.InColor)
		if ( !WritePtrToFile((Ptr) &csi, sizeof(SolutionInfo)) )
			return(SaveBDReturn(FALSE));

	PicID = FirstBDPicture;
	BackPic = MakeSpecialPicture(w[BifD].Back);
	if (BackPic == NIL)
		return(SaveBDReturn(FALSE));
	if (!StdPutPicture(BackPic, PicID++))
		return(SaveBDReturn(FALSE));

	GraphPic = MakeSpecialPicture(BD_Rec.Graph);
	if (GraphPic == NIL)
		return(SaveBDReturn(FALSE));
	if (!StdPutPicture(GraphPic, PicID++))
		return(SaveBDReturn(FALSE));

	return(SaveBDReturn(TRUE));
}

Boolean BadOpenBDReturn(Boolean IsMemError)
{
	Boolean	SeriousError = FALSE;

	if (tmpBD.InitConds != NIL)
		DisposHandle((Handle) tmpBD.InitConds);
	if (tmpBD.SolutionTypes != NIL)
		DisposHandle((Handle) tmpBD.SolutionTypes);
/*	See BadOpenBAReturn
	if (BackPic != NIL)
		KillPicture(BackPic);
	if (GraphPic != NIL)
		KillPicture(GraphPic);
*/
	if (BD_Points != NIL)
		DisposHandle((Handle) BD_Points);
	BD_Points = oBD_Points;
	BD_MaxTimes = oBD_MaxTimes;
	MaxBDShowTimes = oMaxBDShowTimes;

		/*	OffPort Stuff	*/
	if (tmpBD.Graph != NOPORT)
		DisposeOffPort(tmpBD.Graph);
	if ( GraphExists && (BD_Rec.Graph == NOPORT) ) {
		if (!CreateSpecialOffPort(&BD_Rec.Graph, &w[BifD].GR) ) {
			DrawBD = FALSE;
			w[BifD].DrawFromScratch = TRUE;
			StdForceUpdate((WindowPtr) w[BifD].Peek);
			SetSpecialScrolling(BifD);
			SeriousError = TRUE;
		}
		else {
			SetPort(BD_Rec.Graph);
			EraseRect(&BD_Rec.Graph->portRect);
			DrawPicture(oGraphPic, &BD_Rec.Graph->portRect);
		}
	}
	if (oGraphPic != NIL)
		KillPicture(oGraphPic);

/*
	CloseOpenFile();
*/

	SetANilFileName(BifD);

	if (IsMemError) {
		if (SeriousError)	ERROR(SeriousBDOpenMemErrorString);
		else				ERROR(BDOpenMemErrorString);
	}
	else
		ERROR("\PBifurcation Diagram file not intact. Cannot open.");

	return(FALSE);
}

Boolean OpenBDFile()
{
	long			ReadAmt;
	BBWindow		tmpW;
	Rect			nBackRect, oBackRect, CompositeBackRect;
	Boolean			NeedNewBackPort;
	int				PicID;
	Size			MinExtraMemory;
	Boolean			oGraphInAllWindows;
	SolutionInfo	tsi;
	int				i;
	Boolean			NeedToCheckColors;
	int				RScrollValue, BScrollValue;
	long			NewBackMemAddition;

		/*	Get Main Records: BD_Rec, w[BifD]	*/

	ReadAmt = sizeof(BifDiagRecord);
	if ( !ReadPtrFromFile((Ptr) &tmpBD, &ReadAmt) )
		return(FALSE);

	ReadAmt = sizeof(BBWindow);
	if ( !ReadPtrFromFile((Ptr) &tmpW, &ReadAmt) )
		return(FALSE);

		/*	Initialize New Variables	*/

		/*	The following information, if created, needs to be cleared
		 *	if open errorsee BadOpenBDReturn().
		 */
	tmpBD.InitConds = NIL;
	tmpBD.SolutionTypes = NIL;
	tmpBD.Graph = NOPORT;
	BackPic = GraphPic = NIL;
	oGraphPic = NIL;
	oBD_MaxTimes = BD_MaxTimes;
	oMaxBDShowTimes = MaxBDShowTimes;
	oBD_Points = BD_Points;
	BD_Points = NIL;

		/*	Test if sufficient memory	*/
	
	oBackRect = w[BifD].Back->portRect;
	nBackRect = tmpBD.WindowSize;
	OffsetRect(&nBackRect, -nBackRect.left, -nBackRect.top);
		/*	Need to enlarge w[BifD].Back?	*/
	NeedNewBackPort = (	(nBackRect.right > oBackRect.right) ||
						(nBackRect.bottom > oBackRect.bottom) );
	GraphExists = (BD_Rec.Graph != NOPORT);

	MinExtraMemory = ((long) (tmpW.GR.right - tmpW.GR.left)) *
			((long) (tmpW.GR.bottom - tmpW.GR.top));
	MinExtraMemory /= 8;		/*	bits in a byte	*/
	MinExtraMemory *= 2;		/*	Double for GraphPic	*/
	MinExtraMemory += 5*1024L;	/*	Approx. size of BackPic	*/
	if (GraphExists)
		MinExtraMemory += ((long) (w[BifD].GR.right - w[BifD].GR.left)) *
				((long) (w[BifD].GR.bottom - w[BifD].GR.top)) / 8;	/*	oGraphPic	*/
	if (NeedNewBackPort) {
		SetRect(&CompositeBackRect, 0, 0, MAX(nBackRect.right, oBackRect.right),
				MAX(nBackRect.bottom, oBackRect.bottom));
		NewBackMemAddition = ((long) CompositeBackRect.right) *
				((long) CompositeBackRect.bottom) / 8;		/*	new BackPort	*/
		NewBackMemAddition += NewBackMemAddition * 3 / 4;	/*	Cushion for drawing backgd?--bug 11/12/91	*/
		MinExtraMemory += NewBackMemAddition;
	}
	if (!CheckMem(MinExtraMemory))
		return(BadOpenBDReturn(TRUE));

		/*	Create Initial Condition Array	*/

	if (tmpBD.CurrentXInt >= 0) {
		tmpBD.InitConds = (ImpactSet**)
				NewHandle( ((long) (tmpBD.MaxXInt+1)) *
				sizeof(ImpactSet) );
		if (tmpBD.InitConds == NIL)
			return(BadOpenBDReturn(TRUE));
		ReadAmt = (MIN(tmpBD.CurrentXInt, tmpBD.MaxXInt) + 1) *
				sizeof(ImpactSet);
		if ( !ReadHandleFromFile((Handle) tmpBD.InitConds, &ReadAmt) )
			return(BadOpenBDReturn(FALSE));
	}

		/*	Create Solution Type Array	*/

	if (tmpBD.DiffSolutions >= 0) {
		tmpBD.SolutionTypes = (Point**)
				NewHandle( ((long) tmpBD.MaxDiffSolutions) * sizeof(Point) );
		if (tmpBD.SolutionTypes == NIL)
			return(BadOpenBDReturn(TRUE));
		ReadAmt = (tmpBD.DiffSolutions+1) * sizeof(Point);
		if (!ReadHandleFromFile((Handle) tmpBD.SolutionTypes, &ReadAmt) )
			return(BadOpenBDReturn(FALSE));
	}

	NeedToCheckColors = FALSE;
	ReadAmt = sizeof(SolutionInfo);
	if (tmpBD.InColor)
		if ( ReadPtrFromFile((Ptr) &tsi, &ReadAmt) )
			NeedToCheckColors = TRUE;

		/*	Get Pictures, Back and Graph	*/

	PicID = FirstBDPicture;
	if ( (BackPic = GetPicture(PicID++)) == NIL)
		return(BadOpenBDReturn(FALSE));
	if ( (GraphPic = GetPicture(PicID++)) == NIL)
		return(BadOpenBDReturn(FALSE));

		/*	Create BD_Points Array (of graphed points for current values)	*/

	MaxBDShowTimes = ( (tmpBD.ShowTimes != AutoTimes) ?
			tmpBD.ShowTimes :
			MAX( MinAutoShow,
			AutoShowMultiplier * (tmpW.GR.bottom - tmpW.GR.top + 2) ) );
	BD_MaxTimes = ( (tmpBD.InitIterTimes != AutoTimes) ?
			tmpBD.InitIterTimes :
			(tmpBD.DefaultIITimes * MaxCIITimesIncs) ) +
			MaxBDShowTimes;
	BD_Points = (int**)
			NewHandle( ((long) BD_MaxTimes+1) * sizeof(int) );
	if (BD_Points == NIL)
		return(BadOpenBDReturn(TRUE));

		/*	Handle Offports	*/

	if (GraphExists) {
		oGraphPic = MakeSpecialPicture(BD_Rec.Graph);	/*	Backup	*/
		if (oGraphPic == NIL)
			return(BadOpenBDReturn(TRUE));
		DisposeOffPort(BD_Rec.Graph);
		BD_Rec.Graph = NOPORT;
	}
	if (!CreateSpecialOffPort(&tmpBD.Graph, &tmpW.GR))
		return(BadOpenBDReturn(TRUE));

	if (NeedNewBackPort) {
/*	Don't do DisposeOffPort() now, because of errors it might cause
 *	NOTE: This will cause fragmented memory.
 *	Hopefully, it won't happen too often
 *	(can only happen when one opens a bif. diag. on a Mac with a smaller
 *	screen than the one it was created on).
 *	Alternative is to quit program?, or at least disable bif. diag.?
 *	Also, do not need oBackPic, because old BackPort is not eliminated
 *	until sure open is successful, so will not lose the data oBackPic
 *	is set up to store, in PICT (packed) format.
 *			-TA, 3/4/91
		oBackPic = MakeSpecialPicture(w[BifD].Back);	(*	Backup	*)
		DisposeOffPort(w[BifD].Back);
 */
		tmpW.Back = NewOffPort(&CompositeBackRect);
		SetPort(tmpW.Back);	SetStdTextAttributes();
		if (tmpW.Back == NIL)
			return(BadOpenBDReturn(TRUE));
/*	Would become part of BadOpenBDReturn() if dispose w[BifD].Back
			w[BifD].Back = NewOffPort(&oBackRect);
			SetPort(w[BifD].Back);	SetStdTextAttributes();
			if (w[BifD].Back == NIL) || ( GraphExists &&
					!CreateSpecialOffPort(&BD_Rec.Graph, &w[BifD].GR) ) ) {
*/
	}
	else
		tmpW.Back = w[BifD].Back;

		/*	Open is OK!--Clean up fields	*/

	SetOpenWindows();

		/*	Scroll values saved in ControlHandle fields (except early BB)	*/
	RScrollValue = (int) tmpW.RSB;
	if ( (RScrollValue != (long) tmpW.RSB) || (RScrollValue < 0) )
		RScrollValue = 0;
	BScrollValue = (int) tmpW.BSB;
	if ( (BScrollValue != (long) tmpW.BSB) || (BScrollValue < 0) )
		BScrollValue = 0;

	if (tmpBD.InColor)
		tmpBD.InColor = UsingColorMac;
	tmpBD.FullWindow =	(tmpBD.WindowSize.left == LWR.left) &&
						(tmpBD.WindowSize.top == LWR.top) &&
						(tmpBD.WindowSize.right == LWR.right) &&
						(tmpBD.WindowSize.bottom == LWR.bottom);
	tmpW.Peek = w[BifD].Peek;
	tmpW.RSB = w[BifD].RSB;
	tmpW.BSB = w[BifD].BSB;
	tmpW.HS[0] = w[BifD].HS[0];
	tmpW.HS[1] = w[BifD].HS[1];

		/*	Draw into OffPorts	*/

	/*	Moved here 11/11/91--sometimes may need freed-up memory for drawing
	 */
	if (NeedNewBackPort)
		DisposeOffPort(w[BifD].Back);

	SetPort(tmpW.Back);
	EraseRect(&tmpW.Back->portRect);
	DrawPicture(BackPic, &(**BackPic).picFrame);
/*
	KillPicture(BackPic);
*/
	SetPort(tmpBD.Graph);
	EraseRect(&tmpBD.Graph->portRect);
	DrawPicture(GraphPic, &tmpBD.Graph->portRect);
/*
	KillPicture(GraphPic);
*/
	if (oGraphPic != NIL)
		KillPicture(oGraphPic);

		/*	Dispose of all old BD info	*/

	DisposHandle( (Handle) BD_Rec.InitConds);
	DisposHandle( (Handle) BD_Rec.SolutionTypes);
	if (oBD_Points != NIL)
		DisposHandle((Handle) oBD_Points);
/*
	if (NeedNewBackPort)
		DisposeOffPort(w[BifD].Back);
*/

		/*	Set Global Variables	*/

	oGraphInAllWindows = BD_Rec.GraphInAllWindows;
	BD_Rec = tmpBD;
	BD_Rec.GraphInAllWindows = oGraphInAllWindows;
	w[BifD] = tmpW;
	NeedOldInitCond = FALSE;
	if (tmpBD.DiffSolutions >= 0)
		if ((*tmpBD.SolutionTypes)[tmpBD.DiffSolutions].v == StuckSolution)
			NeedOldInitCond = TRUE;
	DrawBD = TRUE;
	w[BifD].DrawFromScratch = FALSE;
	w[BifD].WindowCleared = FALSE;

	SetSpecialScrolling(BifD);
	if (RScrollValue > GetCtlMax(w[BifD].RSB))	RScrollValue = 0;
	if (BScrollValue > GetCtlMax(w[BifD].BSB))	BScrollValue = 0;
	SetCtlValue(w[BifD].RSB, RScrollValue);
	SetCtlValue(w[BifD].BSB, BScrollValue);

	CheckSpecialColor(BifD);
	StdForceUpdate((WindowPtr) w[BifD].Peek);

		/*	Handle Iteration, Menus	*/

/*
	CloseOpenFile();
*/	/*	Seem to need to close before call other resources--unloc. bug	*/

	SetSpecialTitle(TRUE, BifD);
	if (NeedToCheckColors) {
		for(i=0; i<NumSolutionTypes; ++i)
			if (csi[i] != tsi[i]) break;
		if (i<NumSolutionTypes) 
			if (ReplaceSolutionInfoDialog("\Pcolors", "\Pbifurcation diagram")) {
				for(i=0; i<NumSolutionTypes; ++i)
					csi[i] = tsi[i];
				StdForceUpdate((WindowPtr) w[BAttr].Peek);
			}
	}

	SpecialMenuEnable(BifD, TRUE, FALSE);
	if (ProgramStatus == BIFURCATION_DIAGRAM)
		StopIterating(FALSE);
			/*	Do next here due to StopIterating(), ProgramStatus	*/
	if (oGraphInAllWindows != tmpBD.GraphInAllWindows)
		SwitchGraphInAll(BifD);
	NeedToSaveBD = FALSE;
	OKToContinueBD = (BD_Rec.CurrentXInt >= 0) &&
			(BD_Rec.CurrentXInt <= BD_Rec.MaxXInt);
	ControlMenuStop(BIFURCATION_DIAGRAM, OKToContinueBD);
	if ((OKToContinueBD) && (!GettingParamFiles || LastParamFile)) {
		HandleAllActivates();
		HandleAllUpdates();
		if (DoSpecialOpenContinueDialog("\Pbifurcation"))
			ContinueBifurcationDiagram();
	}

	return(TRUE);
}
