/*
 *	bbContGraphics.c
 *
 *	Graphics routines for Bouncing Balls continuous windows,
 *	those that in theory run continuously, as opposed to the impact
 *	windows, those which just display information about impacts.
 *
 *	Routines for impact windows are stored in bbImpactGraphics.c.
 *
 *	These windows are:
 *			Trajectory
 *			Animation
 *			Phase Space
 *			Time Series	(most routines are in bbText.c)
 *			3-D Phase Space
 *
 *		-TA 11/8/90
 */

#include "bbPrototypes.h"
#include "bbGraphics.h"
#include "bbMenus.h"
/*
#include "bbDebug.h"
*/

#define	StuckAnimationTime		10	/* "ticks" delay during stuck animation */
#define	GetNext_TS_t			(Cur_TS_Period + (Period*Next_TS_Sample)/TS_SampleRate)

typedef	int		(**ScreenArrayHandle)[];

		/*	Global external variables; declared in bb.c */
/*
extern	double				v0, delta, A, freq, damp, Period, Omega;
*/
				/*	Global external vars from bb.c	*/
extern	pIterationVars		pCurrentVars;
extern	hBBPreferences		hBBPrefs;
								/*	From bbGraphics.c	*/
extern	Windows				*pw;
								/*	From bbWindows.c; screen size	*/
extern	int					MaxScreenX;
extern	int					RightWindowLimit, BottomWindowLimit;
								/*	From bbImpactGraphics.c; for phase space	*/
extern	double				VMin;
								/*	From bbMenus.c	*/
extern	MenuHandle			ViewMenuH;
extern	Boolean				NeedToSaveTS;


	int		XStart, XStop;

				/* For Trajectory Window */
	int		WindowNumPeriods, WindowNumAmplitudes;
	int		PeriodWidth, AmplitudeHeight;	/* Screen dims. of one cycle */
	int		Max_t_traj, BackMax_t_traj;
	Boolean	RedrawSine, RestartTrajGraph;
	ScreenArrayHandle	SineWave;
	double	TrajTimeOffset;
	Rect	TrajClipRect;
	int		BackTraj_tOffset, Traj_tOffset;
	int		NumTrajAmplitudeHashes;
	Boolean	TrajectoryResized;

				/* For Phase Space Window */
	double	MaxPSVel, MaxPSX;
	Boolean	RestartPSGraph;

				/* For Animation Window */
	Rect	Ball, Table, Frame, LineUpdate;
	int		AnimaHOffset;
	int		Last_Animate_xint, Last_Animate_yint;
					/*	See bbGraphUtils, KludgeForPatterns	*/
	long	KludgeForAnimationPats[2*NumAnimationPats];
	Pattern	*AnimationPats = (Pattern *) &KludgeForAnimationPats;
	GrafPtr	AnimationParts;

				/* For Time Series Window */
	int		Next_TS_Sample;
	double	Cur_TS_Period, Next_TS_t;
	Boolean	NeedToResetTSVars;


			/*	-------------------------------------------------	*/
			/*	---------- Window-Independent Routines ----------	*/
			/*	----------  (more or lesssine wave)   ----------	*/
			/*	-------------------------------------------------	*/

MakeSineWave()
{
	int		x;

	if (!RedrawSine)	return;

	RedrawSine = FALSE;
	InitFixedSineParams(A, w[Traj].xScale, w[Traj].yScale);
	for (x = 0; x < PeriodWidth; ++x)
		(**SineWave)[x] = xTableFixed(x);

	Last_Animate_yint = (**SineWave)[Last_Animate_xint % PeriodWidth];
}

GetSimulationSpeed()
{
	int	oRealPeriodTicks = RealPeriodTicks;

	if (DoSimulationSpeedDialog(&RealPeriodTicks)) ;
}


		/*	------------------------------------------------	*/
		/*	---------- Trajectory Window Routines ----------	*/
		/*	------------------------------------------------	*/

SwitchScrollTrajectory()
{
	ScrollTrajectory = !ScrollTrajectory;
	CheckItem(ViewMenuH, ScrollTrajectoryItem, ScrollTrajectory);
	if (!ScrollTrajectory)
		BackTraj_tOffset = Traj_tOffset = 0;
	return(ScrollTrajectory);
}

ForceRefreshTraj()
{
	StdForceUpdate((WindowPtr) w[Traj].Peek);
	HandleUpdate((WindowPtr) w[Traj].Peek, 0);
}

SetTrajTimeOffset(double theHitTime)
{
	int	periods = (int) (theHitTime/Period);

	TrajTimeOffset = periods * Period;
			/*	what's left was ( phase - delta + ((phase>=delta) ? 0.0 : 1.0) ) * Period;	*/
			/*	or fmod(theHitTime, Period)	*/
	if (ScrollTrajectory)
		BackTraj_tOffset = Traj_tOffset = periods * PeriodWidth;

	if (theHitTime == 0.0)
		if (RealPeriodTicks == 0)
			StartDetachedTickDelay();
}

InitTrajScales()
{
	w[Traj].xScale = PeriodWidth/Period;
	w[Traj].yScale = AmplitudeHeight/NON_ZERO_A;

	SetFixed_G_Over2(w[Traj].xScale, w[Traj].yScale);
}

SetTrajCoords()
{
	int		WindowMaxXCoord, WindowMaxYCoord;
	Rect	HotSpotRect;

	WindowMaxXCoord = w[Traj].Peek->port.portRect.right -
		Traj_RightMargin - w[Traj].BackLimits.left;
	if (WindowMaxXCoord >= PeriodWidth)
		WindowNumPeriods = MIN(WindowMaxXCoord / PeriodWidth, ScreenNumPeriods);
	else
		WindowNumPeriods = 1;
	WindowMaxXCoord = w[Traj].BackLimits.left + (WindowNumPeriods * PeriodWidth);

	WindowMaxYCoord = w[Traj].Peek->port.portRect.bottom -
		Traj_BottomMargin - w[Traj].BackLimits.top;
	WindowNumAmplitudes = MIN(WindowMaxYCoord / AmplitudeHeight,
		ScreenNumAmplitudes);
/*
	WindowMaxYCoord=(int) (WindowNumAmplitudes*AmplitudeHeight)+
			w[Traj].BackLimits.top;
*/
	WindowMaxYCoord = w[Traj].Peek->port.portRect.bottom - Traj_BottomMargin;

	SetRect(&w[Traj].GR, w[Traj].BackLimits.left, w[Traj].BackLimits.top,
			WindowMaxXCoord, WindowMaxYCoord);
	w[Traj].Origin.h = w[Traj].GR.left;
	w[Traj].Origin.v = w[Traj].GR.bottom;
	SetRect(&w[Traj].PR,
			w[Traj].GR.left - w[Traj].Origin.h,
			w[Traj].Origin.v - w[Traj].GR.bottom,
			w[Traj].GR.right - w[Traj].Origin.h + 1,
			w[Traj].Origin.v - w[Traj].GR.top + 1);
			/* +1 to account for pixel hanging below and to left of point */
	Max_t_traj = w[Traj].GR.right - w[Traj].GR.left;

	HotSpotRect = w[Traj].Peek->port.portRect;
	HotSpotRect.top = WindowMaxYCoord + 1;
	HotSpotRect.right	-= ScrollBarWidth-1;
	HotSpotRect.bottom	-= ScrollBarWidth-1;
	ResizeControl(w[Traj].HS[0], &HotSpotRect);

	HotSpotRect.bottom = HotSpotRect.top;
	HotSpotRect.top = w[Traj].GR.top;
	HotSpotRect.right = w[Traj].Origin.h;
	ResizeControl(w[Traj].HS[1], &HotSpotRect);

	++w[Traj].GR.right;
}

SetWholeScreenTrajCoords()
{
	w[Traj].BackLimits.left = Traj_LeftMargin;
	w[Traj].BackLimits.top = Traj_TopMargin;
	/*	For multiple screens, use MaxCurScreenX, Y, where these are the
	 *	corresponding maxes for the main device?		-TA 10/29/90
	 */
		/*	w[Traj].BackLimits.right (used to be ScreenMaxXCoord)
		 *	will always allow each period the same # of pixels
		 */
	w[Traj].BackLimits.right = RightWindowLimit -
			Traj_RightMargin - w[Traj].BackLimits.left;
	PeriodWidth = w[Traj].BackLimits.right / ScreenNumPeriods;
	w[Traj].BackLimits.right = (int) (ScreenNumPeriods * PeriodWidth) +
			w[Traj].BackLimits.left;

	w[Traj].BackLimits.bottom = BottomWindowLimit-Traj_BottomMargin;
	AmplitudeHeight = (w[Traj].BackLimits.bottom - w[Traj].BackLimits.top)/
			ScreenNumAmplitudes;

	w[Traj].BackOrigin.h = w[Traj].BackLimits.left;
	w[Traj].BackOrigin.v = w[Traj].BackLimits.bottom;
	BackMax_t_traj = w[Traj].BackLimits.right - w[Traj].BackLimits.left;

	InitFixedCoeffs(w[Traj].BackLimits.bottom - w[Traj].BackLimits.top);

	InitTrajScales();
	StdForceUpdate((WindowPtr) w[Traj].Peek);
	StdForceUpdate((WindowPtr) w[Anima].Peek);
	TrajectoryResized = TRUE;
}

DrawTrajBackground()
{
	int	XWriteCoord, YWriteCoord;
	int	i, yhash;

	SetPort(w[Traj].Peek);
		/* Draw Axes */
	WindowMoveTo(Traj,w[Traj].GR.right - w[Traj].GR.left, 0);
/*
	WindowLineTo(Traj,0,0);
	WindowLineTo(Traj,0,w[Traj].GR.bottom - w[Traj].GR.top);
*/
	WindowLineTo(Traj,-1,0);
	WindowLineTo(Traj,-1,w[Traj].GR.bottom - w[Traj].GR.top);

		/*	Label Axes	*/

	WindowMoveTo(Traj,w[Traj].GR.right - w[Traj].Origin.h, 0);
	Move(HorAxisLabelHOffset, TextHeight/2-TextHeight%2);
	DrawChar('t');
	WindowMoveTo(Traj,0,w[Traj].Origin.v - w[Traj].GR.top);
	Move(-CharWidth('x')/2 + 1, -VerAxisLabelVOffset);
	DrawChar('x');

		/*	Draw position axis (vertical) hash marks	*/

	for (i=2; i<=ScreenNumAmplitudes; i+=2) {
		yhash = yFloatToCoord(Traj, i*A);
		if (yhash > w[Traj].Origin.v - w[Traj].GR.top)
			break;
		WindowMoveTo(Traj, 0, yhash);
		Line(HashSize,0);
	}
	NumTrajAmplitudeHashes = i/2 - 1;

		/* Write out initial conditions */
	XWriteCoord = w[Traj].BackLimits.left +
			MAX((w[Traj].GR.right - w[Traj].GR.left)/2 - 106,0);
	YWriteCoord = w[Traj].GR.bottom+(Traj_BottomMargin-ScrollBarWidth)/2 +
						TextHeight/2 + TextHeight%2;
	MoveTo(XWriteCoord, YWriteCoord);

	WriteVar("\PV0: ", v0); DrawString("\P    ");
	WriteVar("\PDelta: ", delta);
}

RefreshTraj()
{
		DrawTrajBackground();
/*
		ClipRect(&w[Traj].Peek->port.portRect);
*/
		MakeSineWave();
		if (RestartTrajGraph) {
			SetPort(w[Traj].Back);
			EraseRect(&w[Traj].Back->portRect);
			RestartTrajGraph = FALSE;
		}
		else
			myCopyBits(w[Traj].Back, (GrafPtr) w[Traj].Peek, &w[Traj].GR,
				Traj_tOffset-BackTraj_tOffset,
				w[Traj].BackLimits.bottom - w[Traj].Origin.v,
				0, 0, srcOr, NIL );

}

InitTrajGraphing(double Xi, double Vi, double Ti, int yint)
{
	int			CurrentPeriod, ShiftPeriods, ScrollAmount;
	RgnHandle	updateRgn;

	InitFixedTraj(Xi, Vi, Ti-TrajTimeOffset, XStart);

		/* if window has been shrunk beyond cursor since last graph... */
	if (XStart-Traj_tOffset > Max_t_traj) {
		if (ScrollTrajectory) {
			ScrollAmount = XStart - Traj_tOffset - Max_t_traj;
			Traj_tOffset += ScrollAmount;

			SetPort(w[Traj].Peek);
			updateRgn = NewRgn();
			ScrollRect(&w[Traj].GR, -ScrollAmount, 0, updateRgn);
			myCopyBits(w[Traj].Back, (GrafPtr) w[Traj].Peek,
					&(**updateRgn).rgnBBox,
					Traj_tOffset-BackTraj_tOffset,
					w[Traj].BackOrigin.v-w[Traj].Origin.v,
					0,0, srcCopy, NIL);
			DisposeRgn(updateRgn);
		}
		else {
			CurrentPeriod = (XStart - 1) / PeriodWidth	+ 1;
			ShiftPeriods = CurrentPeriod - WindowNumPeriods;
	
			SetPort(w[Traj].Back);
			updateRgn = NewRgn();
			ScrollRect(&w[Traj].Back->portRect, -(int) (ShiftPeriods*PeriodWidth),
					0, updateRgn);
			DisposeRgn(updateRgn);
			ForceRefreshTraj();
			/*	Had to put back in because SetContGraphingClipping before,
			 *	was after	*/
		 	SetPort(w[Traj].Peek);
			ClipRect(&TrajClipRect);
	
			XStart -= ShiftPeriods * PeriodWidth;
			XStop -= ShiftPeriods * PeriodWidth;
			TrajTimeOffset += ShiftPeriods * Period;
		}
	}

	SetPort(w[Traj].Peek);
	WindowMoveTo(Traj, XStart-Traj_tOffset, yint);
	SetPort(w[Traj].Back);
	BackMoveTo(Traj, XStart-BackTraj_tOffset, yint);
}

GraphTrajectory(Boolean IsStuck, int *pxint, int yint)
{
	RgnHandle	updateRgn, tmpRgn;
	Rect		tmpRect;
	int			Traj_xint, Back_xint;
	int			i, yhash;
	Point		pt;

			/* if beyond edge of graph */
		/*	BackTraj_tOffset always 0, except temporarily here--delete?	*/
	if ( ScrollTrajectory &&
			(*pxint-BackTraj_tOffset > BackMax_t_traj+Max_t_traj) ) {
		BackTraj_tOffset += ScreenNumPeriods * PeriodWidth;

		TrajTimeOffset += ScreenNumPeriods * Period;
		XStart -= BackTraj_tOffset;
		XStop -= BackTraj_tOffset;
		*pxint -= BackTraj_tOffset;
		Traj_tOffset -= BackTraj_tOffset;

		SetPort(w[Traj].Back);
		updateRgn = NewRgn();
		ScrollRect(&w[Traj].Back->portRect, -BackTraj_tOffset, 0, updateRgn);
		DisposeRgn(updateRgn);
		Move(-BackTraj_tOffset, 0);
		BackTraj_tOffset = 0;
/*
		BackTraj_tOffset = ++BackTraj_tOffset % PeriodWidth;
		if (BackTraj_tOffset == 0) {
			TrajTimeOffset += Period;
			XStart -= PeriodWidth;
			XStop -= PeriodWidth;
			*pxint -= PeriodWidth;
			Traj_tOffset -= PeriodWidth;
		}
		SetPort(w[Traj].Back);
		updateRgn = NewRgn();
		ScrollRect(&w[Traj].BackLimits, -1, 0, updateRgn);
		DisposeRgn(updateRgn);
		Move(-1, 0);
*/
	}
	if (*pxint-Traj_tOffset >= Max_t_traj) {
		if (ScrollTrajectory) {
			++Traj_tOffset;
			SetPort(w[Traj].Peek);
			updateRgn = NewRgn();
			ScrollRect(&w[Traj].GR, -1, 0, updateRgn);
	
			if ((**updateRgn).rgnBBox.left < w[Traj].GR.right-1) {
				tmpRect = w[Traj].GR;
				tmpRect.left = tmpRect.right-1;
				tmpRgn = NewRgn();
				RectRgn(tmpRgn, &tmpRect);
				DiffRgn(updateRgn, tmpRgn, updateRgn);
				DisposeRgn(tmpRgn);
						/* should not be 0,0,0,0	*/
				myCopyBits(w[Traj].Back, (GrafPtr) w[Traj].Peek,
						&(**updateRgn).rgnBBox,
						Traj_tOffset-BackTraj_tOffset,
						w[Traj].BackOrigin.v-w[Traj].Origin.v,
						0,0, srcCopy, NIL);
			}
			DisposeRgn(updateRgn);
			Move(-1, 0);
		}
		else {
			RestartTrajGraph = TRUE;
			ForceRefreshTraj();
		 	SetPort(w[Traj].Peek);
			ClipRect(&TrajClipRect);

			*pxint -= Max_t_traj;
			BothMoveTo(Traj, *pxint-1, yint);
			XStart -= Max_t_traj;
			XStop -= Max_t_traj;
			TrajTimeOffset += (WindowNumPeriods*Period);
		}
	}

	Traj_xint = *pxint - Traj_tOffset;
	Back_xint = *pxint - BackTraj_tOffset;
	if (IsStuck) {
		SetPort(w[Traj].Peek);
		WindowLineTo(Traj, Traj_xint, yint);

		SetPort(w[Traj].Back);
		BackLineTo(Traj, Back_xint, yint);
	}
	else {
		SetPort(w[Traj].Peek);
		WindowLineTo(Traj, Traj_xint, yint);
		if (*pxint == 0) {
			WindowMoveTo(Traj, Traj_xint, (**SineWave)[*pxint % PeriodWidth]);
			Line(0, 0);
		}
		else {
			WindowMoveTo(Traj, Traj_xint-1, (**SineWave)[(*pxint-1) % PeriodWidth]);
			WindowLineTo(Traj, Traj_xint, (**SineWave)[*pxint % PeriodWidth]);
		}
		WindowMoveTo(Traj, Traj_xint, yint);

		SetPort(w[Traj].Back);
		BackLineTo(Traj, Back_xint, yint);
		if (*pxint == 0) {
			BackMoveTo(Traj, Back_xint, (**SineWave)[*pxint % PeriodWidth]);
			Line(0, 0);
		}
		else {
			BackMoveTo(Traj, Back_xint-1, (**SineWave)[(*pxint-1) % PeriodWidth]);
			BackLineTo(Traj, Back_xint, (**SineWave)[*pxint % PeriodWidth]);
		}
		BackMoveTo(Traj, Back_xint, yint);
	}
}


		/*	-------------------------------------------------	*/
		/*	---------- Phase Space Window Routines ----------	*/
		/*	-------------------------------------------------	*/

Boolean ToggleFasterPhaseSpace()
{
	return(FasterPhaseSpace = !FasterPhaseSpace);
}

ClearPhaseSpace()
{
	SetPort(w[PSpace].Back);
	EraseRect(&w[PSpace].Back->portRect);
	StdForceUpdate((WindowPtr) w[PSpace].Peek);
	w[PSpace].DrawFromScratch = TRUE;
}

InitPSScales()
{
	MaxPSVel = MAX(VMax, -VMin);
	MaxPSX = (w[PSpace].BackLimits.bottom - w[PSpace].BackLimits.top) *
			NON_ZERO_A / AmplitudeHeight;
	w[PSpace].xScale = (w[PSpace].BackLimits.right-w[PSpace].BackLimits.left) /
			2 / MaxPSVel;
	w[PSpace].yScale = w[Traj].yScale;

	SetFixed_G_ps(w[PSpace].xScale, w[PSpace].yScale);
}

InitWholeScreenPSCoords()
{
	SetPort(w[PSpace].Peek);
	SetRect(&w[PSpace].BackLimits, PS_LeftMargin, PS_TopMargin,
			w[PSpace].Back->portRect.right - PS_RightMargin,
			w[PSpace].Back->portRect.bottom - PS_BottomMargin);
	if ( (w[PSpace].BackLimits.right-w[PSpace].BackLimits.left) % 2 == 1 )
		--w[PSpace].BackLimits.right;

	w[PSpace].BackOrigin.h = (w[PSpace].BackLimits.right-w[PSpace].BackLimits.left)/2;
	w[PSpace].BackOrigin.v = w[PSpace].BackLimits.bottom;

	InitPSScales();
}

SetPSCoords()
{
	Rect	HotSpotRect;

	SetPort(w[PSpace].Peek);
	SetRect(&w[PSpace].GR, w[PSpace].BackLimits.left, w[PSpace].BackLimits.top,
			w[PSpace].Peek->port.portRect.right - PS_RightMargin,
			w[PSpace].Peek->port.portRect.bottom - PS_BottomMargin);
	if ( (w[PSpace].GR.right-w[PSpace].GR.left) % 2 == 1 )
		--w[PSpace].GR.right;
	w[PSpace].Origin.h = (w[PSpace].GR.right + w[PSpace].GR.left) / 2;
	w[PSpace].Origin.v = w[PSpace].GR.bottom;
	SetRect(&w[PSpace].PR,
			w[PSpace].GR.left - w[PSpace].Origin.h,
			w[PSpace].Origin.v - w[PSpace].GR.bottom,
			w[PSpace].GR.right - w[PSpace].Origin.h + 1,
			w[PSpace].Origin.v - w[PSpace].GR.top + 1);
			/* +1 to account for pixel hanging below and to left of point */

	HotSpotRect = w[PSpace].GR;
	++HotSpotRect.right;
	HotSpotRect.top = HotSpotRect.bottom + 1;
	HotSpotRect.bottom = w[PSpace].Peek->port.portRect.bottom -
			(ScrollBarWidth-1);
	ResizeControl(w[PSpace].HS[0], &HotSpotRect);
}

DrawPSBackground()
{
	int		XWriteCoord, YWriteCoord;
	Str255	s;
	double	ScaledMax, Multiplier, HashMeasurement;
	int		ScaledMaxInt, HashInc, HashIndex;
	int		MaxXStrLen = 0;
	Rect	HotSpotRect;

	SetPort(w[PSpace].Peek);
	StdDocWindowClipRect((WindowPtr) w[PSpace].Peek);

		/*	Draw Axes	*/

	WindowMoveTo(PSpace,w[PSpace].GR.left - w[PSpace].Origin.h, 0);
	WindowLineTo(PSpace,w[PSpace].GR.right - w[PSpace].Origin.h, 0);
	WindowMoveTo(PSpace,0,0);
	WindowLineTo(PSpace,0,w[PSpace].Origin.v - w[PSpace].GR.top);

		/*	Label Axes	*/

	WindowMoveTo(PSpace,w[PSpace].GR.right - w[PSpace].Origin.h, 0);
	Move(HorAxisLabelHOffset, TextHeight/2-TextHeight%2);
	DrawChar('v');
	WindowMoveTo(PSpace,0,w[PSpace].Origin.v - w[PSpace].GR.top);
	Move(-CharWidth('x')/2 + 1, -VerAxisLabelVOffset);
	DrawChar('x');

		/*	Draw, Label Hash Marks	*/

			/*	Horizontal Hash Marks	*/

				/*	How far apart are the horizontal marks?	*/
	ScaledMax = MaxPSVel;
	while (ScaledMax > 50.0)	ScaledMax /= 10.0;
	while (ScaledMax < 5.0)	ScaledMax *= 10.0;
	Multiplier = MaxPSVel / ScaledMax;
	ScaledMaxInt = RoundToInt(ScaledMax);
	HashInc = ScaledMaxInt / 10 + 1;

	WindowMoveTo(PSpace, 0, 0);
	Move(-CharWidth('0')/2+1, TextLineHeight);
	DrawChar('0');

				/*	Draw, label hash marks	*/
	for (HashIndex = 0;;) {
		HashIndex += HashInc;
		HashMeasurement = HashIndex * Multiplier;
		XWriteCoord = xFloatToCoord(PSpace, HashMeasurement);
		if (XWriteCoord > w[PSpace].GR.right-w[PSpace].Origin.h) break;

		myFToStrNoPad(HashMeasurement, CP(s), 0, PS_precision);
		WindowMoveTo(PSpace, XWriteCoord, HashSize);
		Line(0, HashSize);
		Move(-StringWidth(CP(s))/2+1, TextLineHeight);
		DrawString(CP(s));

		myFToStrNoPad(-HashMeasurement, CP(s), 0, PS_precision);
		WindowMoveTo(PSpace, -XWriteCoord, HashSize);
		Line(0, HashSize);
		Move(-StringWidth(CP(s))/2+1, TextLineHeight);
		DrawString(CP(s));
	}

			/*	Vertical Hash Marks	*/

				/*	How far apart are the vertical marks?	*/
	ScaledMax = MaxPSX;
	while (ScaledMax > 50.0)	ScaledMax /= 10.0;
	while (ScaledMax < 5.0)	ScaledMax *= 10.0;
	Multiplier = MaxPSX / ScaledMax;
	ScaledMaxInt = RoundToInt(ScaledMax);
	HashInc = ScaledMaxInt / 10 + 1;

				/*	Draw, label hash marks	*/
	for (HashIndex = 0;;) {
		HashIndex += HashInc;
		HashMeasurement = HashIndex * Multiplier;
		YWriteCoord = yFloatToCoord(PSpace, HashMeasurement);
		if (YWriteCoord > w[PSpace].Origin.v-w[PSpace].GR.top) break;

		myFToStrNoPad(HashMeasurement, CP(s), 0, PS_precision);
		if (StringWidth(CP(s)) > MaxXStrLen)
			MaxXStrLen = StringWidth(CP(s));
		WindowMoveTo(PSpace, HashSize, YWriteCoord);
		Line(-HashSize-HashSize, 0);
		Move(-StringWidth(CP(s))-3, TextHeight/2);
		DrawString(CP(s));
	}
/*
	XWriteCoord = w[PSpace].BackLimits.left +
			MAX((w[PSpace].GR.right - w[PSpace].GR.left)/2 - 106,0);
	YWriteCoord = w[PSpace].GR.bottom+(Traj_BottomMargin-ScrollBarWidth)/2 +
						TextHeight/2 + TextHeight%2;
	MoveTo(XWriteCoord, YWriteCoord);

	WriteVar("\PV0: ", v0); DrawString("\P    ");
	WriteVar("\PDelta: ", delta);
*/
	HotSpotRect = w[PSpace].GR;
	HotSpotRect.top = 0;
	HotSpotRect.left = w[PSpace].Origin.h - HashSize - MaxXStrLen - 3 - 1;
	HotSpotRect.right = w[PSpace].Origin.h + HashSize + 1 + 1;
	ResizeControl(w[PSpace].HS[1], &HotSpotRect);
}

RefreshPS()
{
		DrawPSBackground();
/*
		ClipRect(&w[PSpace].Peek->port.portRect);
*/
		if (RestartPSGraph) {
			SetPort(w[PSpace].Back);
			EraseRect(&w[PSpace].Back->portRect);
			RestartPSGraph = FALSE;
		}
		else
			myCopyBits(w[PSpace].Back, (GrafPtr) w[PSpace].Peek,
				&w[PSpace].GR,
				w[PSpace].BackOrigin.h - w[PSpace].Origin.h,
				w[PSpace].BackOrigin.v - w[PSpace].Origin.v,
				0, 0, srcOr, NIL );

}


		/*	-----------------------------------------------	*/
		/*	---------- Animation Window Routines ----------	*/
		/*	-----------------------------------------------	*/

InitAnimationParts()
{
	Point	pt;

	SetPort(AnimationParts);

	pt.h = Ball.left;	pt.v = Ball.top;
	OffsetRect(&Ball, -pt.h, -pt.v);
	FillOval(&Ball, &AnimationPats[BallPattern]);
/*
	FrameOval(&Ball);
*/
	OffsetRect(&Ball, pt.h, pt.v);

	pt.h = Table.left-PERMA_TABLE_H;	pt.v = Table.top;
	OffsetRect(&Table, -pt.h, -pt.v);
	FillRect(&Table, &AnimationPats[TablePattern]);
/*
	FrameRect(&Table);
*/
	OffsetRect(&Table, pt.h, pt.v);

	pt.h = Frame.left-PERMA_FRAME_H;	pt.v = Frame.top;
	OffsetRect(&Frame, -pt.h, -pt.v);
	FillRect(&Frame, &AnimationPats[FramePattern]);
/*
	FrameRect(&Frame);
*/
	OffsetRect(&Frame, pt.h, pt.v);
}

DrawAnimation(Boolean DrawingWholeThing)
/*	DrawingWholeThing is TRUE if we are drawing all the
 *	objects in the animation window, instead of just the moving
 *	parts. This occurs when copying the PICT (HandlePICTCopy) or
 *	printing; both call MakePicture().
 *	When it's true, it offsets properly, does not set w[Anima].Peek,
 *	and does not erase previous line connecting table and frame.
 */
{
	GrafPtr	CurPort;
	Rect	tmpRect;
	int		hOffset = (DrawingWholeThing) ? AnimaHOffset : 0;

	GetPort(&CurPort);
	myCopyBits(AnimationParts, CurPort, &Ball, -Ball.left,-Ball.top,
			hOffset, 0, srcCopy, NIL);
	OffsetRect(&Ball, hOffset, 0);
	FrameOval(&Ball);
	OffsetRect(&Ball, -hOffset, 0);

	myCopyBits(AnimationParts, CurPort, &Table,
			-Table.left+PERMA_TABLE_H, -Table.top, hOffset, 0, srcCopy, NIL);
	OffsetRect(&Table, hOffset, 0);
	FrameRect(&Table);
	OffsetRect(&Table, -hOffset, 0);

	if (!DrawingWholeThing)
		SetPort(w[Anima].Peek);
	PenSize(2,2);
	MoveTo(Frame.right, (Frame.top+Frame.bottom)/2-1);
	if (!DrawingWholeThing) {
/*
		SetRect(&tmpRect, Frame.right, MIN(Frame.top, Table.top),
				AnimaHOffset, MAX(Frame.bottom, Table.bottom));
		EraseRect(&tmpRect);
*/
		EraseRect(&LineUpdate);
	}
	LineTo(Frame.right + FRAME_DIST-1, (Table.top+Table.bottom)/2-1);
	PenSize(1,1);
/*
	FillOval(&Ball, &AnimationPats[BallPattern]);
	FrameOval(&Ball);
	FillRect(&Table, &AnimationPats[TablePattern]);
	FrameRect(&Table);
*/
}

Animate(int xint, int yint)
{
	Rect	BallUpdate, TableUpdate;
	int		vOffset;

	if (!WindowOpen(Anima)) return;
	Last_Animate_xint = xint;	Last_Animate_yint = yint;
	BallUpdate = Ball; TableUpdate = Table;

	SetPort(w[Anima].Back);

	EraseOval(&Ball);
	EraseRect(&Table);

	vOffset = w[Anima].Peek->port.portRect.bottom	- ScrollBarWidth -
		(Table.bottom - Table.top) - StdMargin;
	OffsetRect(&Ball, 0, -Ball.top + vOffset - yint - (Ball.bottom - Ball.top) );
	yint = (**SineWave)[xint % PeriodWidth];
	OffsetRect(&Table, 0, -Table.top + vOffset - yint);

	UnionRect(&BallUpdate, &Ball, &BallUpdate);
	UnionRect(&TableUpdate, &Table, &TableUpdate);
	UnionRect(&Frame, &TableUpdate, &LineUpdate);
	LineUpdate.left = Frame.right;	LineUpdate.right = AnimaHOffset;
/*
	SetPort(w[Anima].Peek);	EraseRect(&LineUpdate);
	SetPort(w[Anima].Back);
*/

	DrawAnimation(FALSE);

	if (SectRect(&BallUpdate,&(*w[Anima].Peek->port.clipRgn)->rgnBBox,&BallUpdate))
		myCopyBits(w[Anima].Back, (GrafPtr) w[Anima].Peek, &BallUpdate,
			0,0,AnimaHOffset,0,srcCopy,w[Anima].Peek->port.visRgn);
	if (SectRect(&TableUpdate,&(*w[Anima].Peek->port.clipRgn)->rgnBBox,&TableUpdate))
		myCopyBits(w[Anima].Back, (GrafPtr) w[Anima].Peek, &TableUpdate,
			0,0,AnimaHOffset,0,srcCopy,w[Anima].Peek->port.visRgn);
/*
	if (SectRect(&LineUpdate,&(*w[Anima].Peek->port.clipRgn)->rgnBBox,&LineUpdate))
		myCopyBits(w[Anima].Back, (GrafPtr) w[Anima].Peek, &LineUpdate,
			0,0,AnimaHOffset,0,srcCopy,w[Anima].Peek->port.visRgn);
*/
}

DrawAnimationFrame()
{
	GrafPtr	gp;
	Rect	tmpFrame = Frame;

	GetPort(&gp);
	tmpFrame.left = 0;
	myCopyBits(AnimationParts, gp, &tmpFrame,
			-Frame.left+PERMA_FRAME_H+(Frame.left-tmpFrame.left),
			-Frame.top, 0, 0, srcCopy, NIL);
	MoveTo(tmpFrame.left, tmpFrame.top);
	Line((tmpFrame.right - tmpFrame.left - 1), 0);
	Line(0, tmpFrame.bottom - tmpFrame.top - 1);
	Line(-(tmpFrame.right - tmpFrame.left - 1), 0);
/*
	FrameRect(&tmpFrame);
*/
/*
	myCopyBits(AnimationParts, gp, &Frame,
			-Frame.left+PERMA_FRAME_H, -Frame.top, 0, 0, srcCopy, NIL);
	FrameRect(&Frame);
*/
}

RefreshAnimation()
{
	Rect	tmpRect;

	MakeSineWave();
	SetPort(w[Anima].Back);
	EraseRect(&w[Anima].Back->portRect);

	AnimaHOffset = ( w[Anima].Peek->port.portRect.right -
			w[Anima].Peek->port.portRect.left - ScrollBarWidth -
			(Table.right-Table.left) );
	AnimaHOffset = AnimaHOffset/2 + AnimaHOffset%2;

	OffsetRect(&Frame, -Frame.right + AnimaHOffset - FRAME_DIST,
			-Frame.top + w[Anima].Peek->port.portRect.bottom -
			ScrollBarWidth - StdMargin - AmplitudeHeight -
			TABLE_HEIGHT - TABLE_HEIGHT/2);
	SetPort(w[Anima].Peek);
	DrawAnimationFrame();

	tmpRect = w[Anima].Peek->port.portRect;
	tmpRect.left = Frame.right;
	tmpRect.right = AnimaHOffset;
	EraseRect(&tmpRect);

/*
	if ((myHitNumber>=0) || IsIterating)
*/
		Animate(Last_Animate_xint, Last_Animate_yint);
/*
		Animate(0, (**SineWave)[0]);
*/
	ActivateAnimaHS(TRUE, &Ball, &Table);


}


		/*	------------------------------------------	*/
		/*	---------- Time Series Routines ----------	*/
		/*	---------- (see also bbText.c)  ----------	*/
		/*	------------------------------------------	*/

ResetTSVars(double time)
{
	double TS_SampleStatus;

	Cur_TS_Period = ((int) (time/Period)) * Period;
	TS_SampleStatus = (time-Cur_TS_Period)/Period * TS_SampleRate;
	Next_TS_Sample = (int) TS_SampleStatus;
	if (Next_TS_Sample != TS_SampleStatus)
		++Next_TS_Sample;
	if (Next_TS_Sample == TS_SampleRate) {
		Cur_TS_Period += Period;
		Next_TS_Sample = 0;
	}
	Next_TS_t = GetNext_TS_t; 
	NeedToResetTSVars = FALSE;
}

ToggleUseTimeSeries(Boolean CheckWindow)
{
	UseTimeSeries = !UseTimeSeries;
	CheckUseTimeSeries(UseTimeSeries);
	if (UseTimeSeries) {
		if (CheckWindow)
			if (!WindowOpen(TimeSeries))
				if (DoShowWindowDialog(TSer))
					BBOpenWindow((WindowPtr) w[TimeSeries].Peek);
		InitTimeSeries(A, freq, damp, FALSE);
		DeactivateATextScroll(TimeSeries);
		NeedToResetTSVars = TRUE;
	}
	else
		ActivateATextScroll(TimeSeries);
	return;
}

GetTimeSeriesInfo()
{
	int		OldTS_SampleRate = TS_SampleRate;
	Boolean	tmpUTS = UseTimeSeries;

	if (!DoTimeSeriesInfoDialog(&TS_SampleRate, &tmpUTS))	return;

	if (TS_SampleRate != OldTS_SampleRate)
		InitTimeSeries(A, freq, damp, TRUE);
	if (tmpUTS != UseTimeSeries)
		ToggleUseTimeSeries(TRUE);
}

ClearTimeSeries()
{
	InitBodyTE(TimeSeries);
}

SeeIfWantTS(Boolean ToStart)
{
	if (ToStart) {
		if (UseTimeSeries)					return;
		if (!DoStartTimeSeriesDialog())		return;
		ToggleUseTimeSeries(FALSE);
	}
	else {
		if (!UseTimeSeries)					return;
		if (!DoStopTimeSeriesDialog())		return;
		ToggleUseTimeSeries(FALSE);
	}
}

GraphTimeSeries(double time, double Ti, double Xi, double Vi, Boolean IsStuck)
{
	w[TSer].WindowCleared = FALSE;

	while (Next_TS_t <= time) {
		NeedToSaveTS = TRUE;
		if (IsStuck)
			WriteTS(Next_TS_t, vTable(Next_TS_t), xTable(Next_TS_t));
		else
			WriteTS(Next_TS_t, vBall(Next_TS_t-Ti, Vi), xBall(Next_TS_t-Ti, Xi, Vi));
		Next_TS_Sample = (Next_TS_Sample + 1) % TS_SampleRate;
		if (Next_TS_Sample == 0)
			Cur_TS_Period += Period;
		Next_TS_t = GetNext_TS_t;
	}
}


		/*	------------------------------------------------------	*/
		/*	----------      Inter-Window Routines       ----------	*/
		/*	---------- (incl. GraphFrom, Bottom Line) ----------	*/
		/*	------------------------------------------------------	*/

SetContGraphingClipping(Boolean StartingToGraph)
{
	Rect		StdTrajCR, tmpRect;
	RgnHandle	TrajClipRgn, tmpRgn;
	int			i, yhash;

	if (StartingToGraph) {
		if (WindowOpen(Trajectory)) {
			SetPort(w[Traj].Peek);
			GetStdDocWindowClipRect(&StdTrajCR, (WindowPtr) w[Traj].Peek);
			if (SectRect(&w[Traj].GR, &StdTrajCR, &TrajClipRect)) ;
			if (ScrollTrajectory) {
				TrajClipRgn = NewRgn();
				tmpRgn = NewRgn();
				RectRgn(TrajClipRgn, &TrajClipRect);
				for (i=2; i<=NumTrajAmplitudeHashes*2; i+=2) {
					yhash = yFloatToCoord(Traj, i*A);
					SetRect(&tmpRect, 0, 0, HashSize+1, 1);
					OffsetRect(&tmpRect, w[Traj].GR.left,
							w[Traj].Origin.v - yhash);
					RectRgn(tmpRgn, &tmpRect);
					DiffRgn(TrajClipRgn, tmpRgn, TrajClipRgn);
				}
				SetClip(TrajClipRgn);
				DisposeRgn(tmpRgn);
				DisposeRgn(TrajClipRgn);
			}
			else
				ClipRect(&TrajClipRect);
		}
		if (WindowOpen(PhaseSpace)) {
			SetPort(w[PSpace].Peek);
			ClipRect(&w[PSpace].GR);
		}
	}
	else {
		if (WindowOpen(Trajectory))
			StdDocWindowClipRect((GrafPtr) w[Traj].Peek);
		if (WindowOpen(PhaseSpace))
			StdDocWindowClipRect((GrafPtr) w[PSpace].Peek);
	}
}

GraphFrom(Ti, Tj, Xi, Vi)
	double	Ti, Tj, Xi, Vi;
{
	int		xint, yint;
	int		xmod;
	int		ShiftPeriods;

	if (RealPeriodTicks != 0)
		StartDetachedTickDelay();

	if (WindowOpen(Traj))	w[Traj].WindowCleared = FALSE;
	if (WindowOpen(Anima))	w[Anima].WindowCleared = FALSE;
	if (WindowOpen(PSpace))	w[PSpace].WindowCleared = FALSE;

/*
	if ((!WindowOpen(Traj)) && (!WindowOpen(Anima))) return;
*/

	if (TrajectoryResized) {
		ShiftPeriods = (int) ( (Ti-TrajTimeOffset) / Period );
		TrajTimeOffset += ShiftPeriods * Period;
		Traj_tOffset = 0;
		BackTraj_tOffset = 0;
		TrajectoryResized = FALSE;
	}

	XStart = xFloatToCoord(Traj, Ti-TrajTimeOffset);
	XStop = xFloatToCoord(Traj, Tj-TrajTimeOffset);
		/* Get first point */
	yint = (**SineWave)[XStart % PeriodWidth];

	SetContGraphingClipping(TRUE);
	ActivateAnimaHS(FALSE, &Ball, &Table);

	InitTrajGraphing(Xi, Vi, Ti, yint);
	if (WindowOpen(PhaseSpace))
		InitPSGraphing(Xi, Vi, yint, Tj - Ti, &w[PSpace].DrawFromScratch);
	Animate(XStart, yint);
	if (NeedToResetTSVars)
		ResetTSVars(Ti);

	for (xint=XStart+1; xint<=XStop /* -1 */; ++xint) {
		if (xint < XStop)
			yint = xBallFixed(xint-XStart);
		else	/* Last point */
			yint = (**SineWave)[XStop % PeriodWidth];

		GraphTrajectory(FALSE, &xint, yint);
		if (xint < XStop) {
			if (WindowOpen(PhaseSpace))
				GraphPhaseSpace(xint-XStart, yint, FALSE);
			if (UseTimeSeries)
				GraphTimeSeries(xCoordToFloat(Traj, xint) + TrajTimeOffset,
						Ti, Xi, Vi, FALSE);
		}
		Animate(xint, yint);

/*
		xmod = (xint - 1) % PeriodWidth + 1;
*/
		if (RealPeriodTicks != 0)
			StopDetachedTickDelay((long) ((xint-XStart)*RealPeriodTicks/
					PeriodWidth));
/*
			StopDetachedTickDelay((long) ((xint-XStart)*RealPeriodTime*60/
					PeriodWidth));
*/
/*
			StopDetachedTickDelay((long) (RealPeriodTime*60*xmod/PeriodWidth));
		if (xmod == PeriodWidth)
			StartDetachedTickDelay();
*/
	}
		/*	Next two are esp. in case XStart == XStop	*/
	if (WindowOpen(PhaseSpace))
		GraphPhaseSpace(0, 0, TRUE); /* first 2 arguments don't matter */
	if (UseTimeSeries)
		GraphTimeSeries(Tj, Ti, Xi, Vi, FALSE);

	ActivateAnimaHS(TRUE, &Ball, &Table);
	SetContGraphingClipping(FALSE);

	if (RealPeriodTicks != 0)
		StopDetachedTickDelay( (long) ((Tj-Ti)*RealPeriodTicks/Period));
/*
		StopDetachedTickDelay( (long) ((Tj-Ti)*RealPeriodTime*60/Period));
*/
}

GraphStuckBall(double Ti, double Tj)
{
	int		xint, yint;
	int		xmod;

	if (WindowOpen(Traj))	w[Traj].WindowCleared = FALSE;
	if (WindowOpen(Anima))	w[Anima].WindowCleared = FALSE;
	if (WindowOpen(PSpace))	w[PSpace].WindowCleared = FALSE;

/*
	if (!WindowOpen(Animation) &&
		!WindowOpen(PhaseSpace) &&
		!WindowOpen(Trajectory) &&
		!UseTimeSeries) return;
*/

	SetContGraphingClipping(TRUE);
	ActivateAnimaHS(FALSE, &Ball, &Table);

	XStart = xFloatToCoord(Traj, Ti-TrajTimeOffset);
	XStop = xFloatToCoord(Traj, Tj-TrajTimeOffset);

	BothMoveTo(Traj, XStart, (**SineWave)[XStart % PeriodWidth]);
	if (WindowOpen(PhaseSpace))
		InitStuckPhaseSpace(Ti, Tj, Period);
	if (NeedToResetTSVars)
		ResetTSVars(Ti);
		/* Should go through following loop once if XStop = XStart? */
/*
	if (RealPeriodTime == 0.0)
*/
	StartDetachedTickDelay();
	for (xint=XStart+1;xint<=XStop;++xint) {
		yint = (**SineWave)[xint % PeriodWidth];

		GraphTrajectory(TRUE, &xint, yint);
		Animate(xint, yint);
		if (xint < XStop) {
			if (WindowOpen(PhaseSpace))
				GraphStuckPhaseSpace(xCoordToFloat(Traj, xint), yint, A,
						Period, FALSE);
			if (UseTimeSeries)
				GraphTimeSeries(xCoordToFloat(Traj, xint) + TrajTimeOffset,
						Ti, 0.0, 0.0, TRUE);
		}

		xmod = (xint - 1) % PeriodWidth + 1;
		if (RealPeriodTicks == 0)
			StopDetachedTickDelay((xint-XStart)*StuckAnimationTime);
		else {
			StopDetachedTickDelay((long) ((xint-XStart)*RealPeriodTicks/
					PeriodWidth));
/*
			StopDetachedTickDelay((long) (RealPeriodTime*60*xmod/PeriodWidth));
			if (xmod == PeriodWidth)
				StartDetachedTickDelay();
*/
		}
	}
		/*	to be like GraphFrom; could be in loop here;
		 *	would need different first arguments for PhaseSpace	*/
	if (WindowOpen(PhaseSpace))
		GraphStuckPhaseSpace(Tj, yint, A, Period, TRUE);
	if (UseTimeSeries)
		GraphTimeSeries(Tj, Ti, 0.0, 0.0, TRUE);

	ActivateAnimaHS(TRUE, &Ball, &Table);
	SetContGraphingClipping(FALSE);

	if (RealPeriodTicks != 0)
		StopDetachedTickDelay( (long) ((Tj-Ti)*RealPeriodTicks/Period));
}

NewContGraphics()
{
	w[PSpace].DrawFromScratch = TRUE;
	MakeSineWave();

	Last_Animate_xint = 0;	Last_Animate_yint = (**SineWave)[0];
	HandleUpdate((WindowPtr) w[Anima].Peek, 0);
/*
	Animate(0, (**SineWave)[0]);
*/
/*
	ActivateAnimaHS(TRUE, &Ball, &Table);
*/
	if (UseTimeSeries)
		ResetTSVars(0.0);
	if (!w[Traj].WindowCleared)
		NewBBWindow(Trajectory);
}

InitContGraphics()
{
	Handle	PH;

	if ( (PH = GetResource(PatternListResType, AnimationPatsID)) == NIL ) {
		BlockEquate(CP(AnimationPats[BallPattern]), CP(black), sizeof(Pattern));
		BlockEquate(CP(AnimationPats[TablePattern]), CP(gray), sizeof(Pattern));
	}
	else {
		HLock(PH);
		BlockEquate(CP(AnimationPats[0]), *PH+2, NumAnimationPats*sizeof(Pattern));
		HUnlock(PH);
		ReleaseResource(PH);
	}

	RestartTrajGraph = FALSE;
	TrajTimeOffset = 0.0;
	BackTraj_tOffset = Traj_tOffset = 0;
	SetWholeScreenTrajCoords();
	TrajectoryResized = FALSE;

/*
	w[IMap].Peek = NIL;
	w[PSpace].Back = NOPORT;
*/	/*	kludges for SetVMax	*/
	SetVMax(TRUE);				/*	should be under InitImpactGraphics;	*
								 *	here for timing reasons				*/

	SineWave = (ScreenArrayHandle) NewHandle(MaxScreenX * sizeof(int));
	RedrawSine = TRUE;

	SetRect(&Ball, TABLE_WIDTH/2 - BALL_DIAMETER/2, 0,
			TABLE_WIDTH/2 + BALL_DIAMETER/2, BALL_DIAMETER);
	SetRect(&Table, 0, 0, 0+TABLE_WIDTH, TABLE_HEIGHT);
	SetRect(&Frame, 0, 0, MaxScreenX/2, Table.bottom*2);
	Last_Animate_xint = 0;	Last_Animate_yint = (**SineWave)[0];
	InitAnimationParts();
	NeedToResetTSVars = FALSE;
}
