/*	bbControls.c
 *
 *	Handles Bouncing Ball's controls.
 *
 *		-TA, 8/15/90
 */

#include "bbPrototypes.h"
#include "bbGraphics.h"

#define	ACTIVE_CONTROL			0
#define	INACTIVE_CONTROL		254
#define	DEAD_CONTROL			255
#define	SmallTextScroll			(1L<<16)
#define	GraphicsScroll			( (5L<<16) | 50L )
#define	HotSpotPartCode			1
#define	HotSpotVarCode			3
		/* 3-highlight (invert)/track, 2-no highlight/track,
		 * 1-always hit, 0-always hit/return immediately */
#define	ControlHandleParamMask	(0x00FFFFFF)
#define	ControlVarCodeParamMask	(0xFF000000)

#define	GET_VARCODE(ch)			((int)  ( ((long) (**(ch)).contrlDefProc) >> 24 ))


		/*	From bbGraphics.c	*/
extern	Windows	*pw;

		/*	From bbContGraphics.c	*/
extern	int		AnimaHOffset;

	int		CurrentWindex;	/*	Controls belong to window indexed by this */
	Boolean	InHotSpotRect;


SetRScrollBarRect(WindowPtr wp, Rect *pr)
{
	*pr = wp->portRect;
	(*pr).left = ++((*pr).right) - ScrollBarWidth;
	--((*pr).top);
	(*pr).bottom -= ScrollBarWidth-2;
}

SetBScrollBarRect(WindowPtr wp, Rect *pr)
{
	*pr = wp->portRect;
	--((*pr).left);
	(*pr).right -= ScrollBarWidth-2;
	(*pr).top = ++((*pr).bottom) - ScrollBarWidth;
}

ResizeControl(ControlHandle ch, Rect *pr)
{
	(**ch).contrlVis = FALSE;
	MoveControl(ch, (*pr).left, (*pr).top);
	SizeControl(ch, (*pr).right - (*pr).left, (*pr).bottom - (*pr).top);
	(**ch).contrlVis = TRUE;
}

ResetScrollBars(int i /* window index */)
{
	Rect	tr;

	if (w[i].RSB != NIL) {
		NoClippingRect((WindowPtr) w[i].Peek);
		EraseRect(&(**w[i].RSB).contrlRect);
		StdDocWindowClipRect((WindowPtr) w[i].Peek);
		SetRScrollBarRect((WindowPtr) w[i].Peek, &tr);
		ResizeControl(w[i].RSB, &tr);
	}
	if (w[i].BSB != NIL) {
		NoClippingRect((WindowPtr) w[i].Peek);
		EraseRect(&(**w[i].BSB).contrlRect);
		StdDocWindowClipRect((WindowPtr) w[i].Peek);
		SetBScrollBarRect((WindowPtr) w[i].Peek, &tr);
		ResizeControl(w[i].BSB, &tr);
	}
}

CheckHiliteScrollBar(ControlHandle ch, Boolean ControlActive)
{
	if (ch == NIL) return;
	if ( (ControlActive) && ( GetCtlMax(ch) > GetCtlMin(ch) ) )
		HiliteControl(ch, ACTIVE_CONTROL);
	else
		HiliteControl(ch, INACTIVE_CONTROL);
}

BBHiliteControls(int i, Boolean ControlActive)
{
	CheckHiliteScrollBar(w[i].RSB, ControlActive);
	CheckHiliteScrollBar(w[i].BSB, ControlActive);
	DrawControls((WindowPtr) w[i].Peek);
}

HiliteTextScrollBar(int windex, Boolean Activate_It)
{
	CheckHiliteScrollBar(w[windex].RSB, Activate_It);
}

int GetTextScrollMax(int windex)
{
	return(GetCtlMax(w[windex].RSB));
}

SetTextScrollMax(int windex, int NumHiddenLines, Boolean SetValue)
{
	int			ctlmax = MAX(NumHiddenLines,0);
	WindowPtr	wp;

	SetCtlMax(w[windex].RSB, ctlmax);
	if ( (SetValue) || (GetCtlValue(w[windex].RSB) > ctlmax) )
		SetCtlValue(w[windex].RSB, ctlmax);
	if ( (*w[windex].Peek).hilited == 0 ) return;
	HiliteControl(w[windex].RSB,
			(ctlmax>0) ? ACTIVE_CONTROL : INACTIVE_CONTROL); 
	wp = (WindowPtr) w[windex].Peek;
	SetPort((WindowPtr) w[windex].Peek);
	InvalRect(&(**w[windex].RSB).contrlRect);
/*
	NoClippingRect(wp);
	 DrawControls(wp);
	StdDocWindowClipRect(wp);
*/
}

SetTextScrollPage(int windex, int LinesPerPage)
{
	SetCRefCon( w[windex].RSB, SmallTextScroll |
		(long) MAX(0,(LinesPerPage>1) ? LinesPerPage-1 : LinesPerPage) );
}

int GetTextScrollValue(int windex)
{
	return(GetCtlValue(w[windex].RSB));
}

int TextScrollValueFromBottom(int windex)
{
	return(GetCtlMax( w[windex].RSB) - GetCtlValue(w[windex].RSB) );
}

int SetRelScrollVal(ControlHandle ch, int dhv)
/*	Sets scroll value, keeping between min and max; returns lines scrolled */
{
	int	theValue, oldValue = GetCtlValue(ch);

	if ( (oldValue + dhv) < GetCtlMin(ch) )
		theValue = GetCtlMin(ch);
	else if ( (oldValue + dhv) > GetCtlMax(ch) )
		theValue = GetCtlMax(ch);
	else
		theValue = oldValue + dhv;
	SetCtlValue(ch, theValue);
	return(theValue-oldValue);
}

IncTextScrollValue(int windex, int LinesToScroll)
{
	RgnHandle	rgnh = NewRgn();
	Rect		tmpRect = {0,0,0,0};

	if (SetRelScrollVal(w[windex].RSB, LinesToScroll)) {
		SetPort((WindowPtr) w[windex].Peek);
		GetClip(rgnh);
		ClipRect(&tmpRect);
		HideControl(w[windex].RSB);
/*
		InvalRect(&(**w[windex].RSB).contrlRect);
*/
		NoClippingRect((WindowPtr) w[windex].Peek);
		ShowControl(w[windex].RSB);
		SetClip(rgnh);
	}
}

ScrollWindowGraphics(int windex, int HScrollAmount, int VScrollAmount)
{
	RgnHandle	updateRgn = NewRgn();
	Rect		tmpRect;

	GetStdDocWindowClipRect(&tmpRect, (WindowPtr) w[windex].Peek);
	ScrollRect(&tmpRect, HScrollAmount, VScrollAmount, updateRgn);
	ScrollCopyBackRect(windex, &(**updateRgn).rgnBBox, TRUE);
	DisposeRgn(updateRgn);
	SetSpecialControls(windex);
}

ScrollWindow(int windex, int HScrollAmount, int VScrollAmount)
{
	switch (windex) {
		case TextOutput:
		case TimeSeries:
				ScrollWindowText(windex, -VScrollAmount);
				break;
		case BifDiagram:
		case BasinsAttraction:
				ScrollWindowGraphics(windex, -HScrollAmount, -VScrollAmount);
				break;
	}
}

ScrollWindowOneDirection(int windex, ControlHandle ch, int ScrollAmount)
{
	if (ch == w[windex].RSB)
		ScrollWindow(windex, 0, ScrollAmount);
	else if (ch == w[windex].BSB)
		ScrollWindow(windex, ScrollAmount, 0);
}

pascal void MyScrollBarAction(ControlHandle ch, int partCode)
{
	int	ScrollAmount;

	switch (partCode) {
		case inUpButton:	ScrollAmount = -HiWord(GetCRefCon(ch));	break;
		case inDownButton:	ScrollAmount =  HiWord(GetCRefCon(ch));	break;
		case inPageUp:		ScrollAmount = -LoWord(GetCRefCon(ch));	break;
		case inPageDown:	ScrollAmount =  LoWord(GetCRefCon(ch));	break;
		default:			ScrollAmount =  0;						break;
	}
	ScrollWindowOneDirection(CurrentWindex, ch, SetRelScrollVal(ch,ScrollAmount));
}

HandleScrollBar(int windex, ControlHandle ch, int partCode, Point pt)
{
	RgnHandle	rgnh;
	int			oldValue;

	SetPort((WindowPtr) w[windex].Peek);
	rgnh = NewRgn();
	GetClip(rgnh);
	NoClippingRect((WindowPtr) w[windex].Peek);
	if (partCode == inThumb) {
		oldValue = GetCtlValue(ch);
		if (TrackControl(ch, pt, NIL) != 0)
			;
		ScrollWindowOneDirection(windex, ch, GetCtlValue(ch) - oldValue);
	}
	else {
		if (TrackControl(ch, pt, &MyScrollBarAction) != 0)
			;
	}
	SetClip(rgnh);
}

ActivateAnimaHS(Boolean TurnOn, Rect *pBall, Rect *pTable)
{
	Rect	theBall = *pBall, theTable = *pTable;

	OffsetRect(&theBall, AnimaHOffset, 0);
	OffsetRect(&theTable, AnimaHOffset, 0);
	if (TurnOn) {
		ResizeControl(w[Anima].HS[0], &theBall);
		ResizeControl(w[Anima].HS[1], &theTable);
		HiliteControl(w[Anima].HS[0], ACTIVE_CONTROL);
		HiliteControl(w[Anima].HS[1], ACTIVE_CONTROL);
	}
	else {
		HiliteControl(w[Anima].HS[0], INACTIVE_CONTROL);
		HiliteControl(w[Anima].HS[1], INACTIVE_CONTROL);
	}
}

Boolean HandleHotSpot(ControlHandle ch, Point pt)
{
	int		theVarCode = GET_VARCODE(ch);
	Boolean	tmpGoodHotSpotClick;

	InHotSpotRect = TRUE;
	if (theVarCode == 3)
		InvertRect( &((**ch).contrlRect) );
	if (theVarCode == 0)
		tmpGoodHotSpotClick = TRUE;
	else
		tmpGoodHotSpotClick = (TrackControl(ch, pt, (ProcPtr) -1L) != 0) ||
				(theVarCode == 1);
	if (InHotSpotRect)
		if (theVarCode == 3)
			InvertRect( &((**ch).contrlRect) );
	return(tmpGoodHotSpotClick);
}

int GetWhichHotSpot(int windex, ControlHandle ch)
{
	int	i;

	for(i=0;i<MAX_HotSpots;++i)
		if (ch == w[windex].HS[i])
			return(i);
	return(NO_HotSpot);
}

Boolean HandleControl(int i, Point pt, int *pWhichHotSpot)
	/*	Returns TRUE if in a control, FALSE otherwise	*/
{
	int				partCode;
	ControlHandle	ch;

	CurrentWindex = i;
	*pWhichHotSpot = NO_HotSpot;
	if ( (partCode = FindControl(pt, (WindowPtr) w[i].Peek, &ch)) != 0 ) {
		if ( (**ch).contrlHilite == INACTIVE_CONTROL )
			return(TRUE);
		if ( (ch == w[i].RSB) || (ch == w[i].BSB) )
			HandleScrollBar(i, ch, partCode, pt);
		else if (HandleHotSpot(ch, pt))
			*pWhichHotSpot = GetWhichHotSpot(i, ch);
		return(TRUE);
	}
	else
		return(FALSE);
}

pascal long HotSpotControlDef(int varCode,ControlHandle ch,int message,long param)
{
	Point	pt;

	switch (message) {
		case drawCntl:	break;
		case testCntl:	if ( (**ch).contrlHilite == DEAD_CONTROL ) return(0L);
						pt.h = LoWord(param);	pt.v = HiWord(param);
						if (  PtInRect( pt, &((**ch).contrlRect) )   ){
							if( (**ch).contrlHilite == INACTIVE_CONTROL )
								return( (long) (**ch).contrlHilite );
							else
								return((long) HotSpotPartCode);
						}
						else
							return(0L);
		case calcCRgns:	param &= ControlHandleParamMask;
						RectRgn( (RgnHandle) param, &( (**ch).contrlRect ) );
						break;
		case initCntl:	break;
		case dispCntl:	break;
		case posCntl:	break;
		case thumbCntl:	break;
		case dragCntl:	break;
		case autoTrack:	if (varCode != 3) break;
						else {
							GetMouse(&pt);
							param = ((long) pt.h) | ( ((long) pt.v) << 16 );
							if ( (HotSpotControlDef(varCode, ch,testCntl, param)
									!= 0L) ^ (InHotSpotRect) ) {
								InvertRect( &((**ch).contrlRect) );
								InHotSpotRect = !InHotSpotRect;
							}
							break;
						}
	}
	return(0L);
}

CreateHotSpots(int windex, int num)
{
			Rect	boundsRect = {0,0,0,0};
	static	ProcPtr	RoutinePtr;
			int		i;

	RoutinePtr = (ProcPtr) &HotSpotControlDef;

	for(i=0; i<num; ++i) {
		w[windex].HS[i] = NewControl((WindowPtr) w[windex].Peek,
				&boundsRect, "\P", FALSE, 0, 0, 0, pushButProc, NIL );
		(**w[windex].HS[i]).contrlDefProc =
			(Handle) ( (long) &RoutinePtr | (((long) HotSpotVarCode) << 24) );
		(**w[windex].HS[i]).contrlAction = (ProcPtr) -1L;
		(**w[windex].HS[i]).contrlHilite = ACTIVE_CONTROL;
		ShowControl(w[windex].HS[i]);
	}
}

InitBBControls()
{
	int		i;
	Rect	boundsRect = {0,0,0,0};

	for(i=0; i<NumWindows; ++i)
		w[i].RSB = w[i].BSB = w[i].HS[0] = w[i].HS[1] = NIL;

	w[TxOut].RSB = NewControl((WindowPtr) w[TxOut].Peek, &boundsRect, "\P",
		TRUE /* visible */, 0, 0, 0, scrollBarProc, SmallTextScroll );
	(**w[TxOut].RSB).contrlHilite = INACTIVE_CONTROL;

	w[TSer].RSB = NewControl((WindowPtr) w[TSer].Peek, &boundsRect, "\P",
		TRUE /* visible */, 0, 0, 0, scrollBarProc, SmallTextScroll );
	(**w[TSer].RSB).contrlHilite = INACTIVE_CONTROL;

	for(i=BifD; i<=BAttr; ++i) {
		w[i].RSB = NewControl((WindowPtr) w[i].Peek, &boundsRect, "\P",
			TRUE /* visible */, 0, 0, 0, scrollBarProc, GraphicsScroll );
		(**w[i].RSB).contrlHilite = INACTIVE_CONTROL;
		w[i].BSB = NewControl((WindowPtr) w[i].Peek, &boundsRect, "\P",
			TRUE /* visible */, 0, 0, 0, scrollBarProc, GraphicsScroll );
		(**w[i].BSB).contrlHilite = INACTIVE_CONTROL;
	}

	CreateHotSpots(Traj, 2);
	CreateHotSpots(TxOut, 1);
	CreateHotSpots(IMap, 2);
	CreateHotSpots(Anima, 2);
	CreateHotSpots(IPhs, 2);
	CreateHotSpots(PSpace, 2);
	CreateHotSpots(TSer, 1);
	CreateHotSpots(BifD, 2);
	CreateHotSpots(BAttr, 2);

	SetRect(&boundsRect, 0, 0, 0, BodyTETop-TextHeight-1);
	ResizeControl(w[TxOut].HS[0], &boundsRect);
	SetRect(&boundsRect, 0, 0, 0, TS_BodyTETop-TextHeight-1);
	ResizeControl(w[TSer].HS[0], &boundsRect);
}

