/*	bbDialog.c
 *
 *	Routines which handle Bouncing Ball's Dialog Boxes
 *
 *		-TA, 7/87...
 *
 */

#include <ctype.h>
#include <limits.h>
#include <DialogUtils.h>

#include "bbPrototypes.h"

enum {
		GeneralAlertID			= 351,
		BadParamAlertID,
		RToSavedAlertID,
		OutOfRangeAlertID,
		AllBadValsAlertID,
		FixBadValAlertID,
		FixBadIntValAlertID,
		AboutBBDLogID			= 400,
		PrintingDLogID,
		DelimiterDLogID,
		QuitWarningDLogID,
		SaveWarningDLogID,
		BackgroundDLogID,
		InitCondDLogID			= 452,
		PhysSysDLogID,
		TrajScalingDLogID,
		StuckBallDLogID,
		RunInBackgroundDLogID,
		PhaseRangeDLogID,
		SetVMaxDLogID,
		IterationDLogID,
		ImpactPhasesOffsetDLogID,
		MagnifyImpactMapDLogID,
		MagnifyImpactPhasesDLogID,
		oSimulationSpeedDLogID,
		SimulationSpeedDLogID,
		TimeSeriesInfoDLogID	= 471,
		StartTimeSeriesDLogID,
		StopTimeSeriesDLogID,
		ShowWindowDLogID
};

#define	StuckStringsResID		130
#define	ScrollValBarCDEFID		128
#define	PopUpCDEFID				129
#define	DelimPopUpMenuID		128

#define	DLogPrecision			6
#define UserOutlineItemNum		3
#define	FirstEditTextItem		6
#define	ETItem1					FirstEditTextItem
#define	MinStuckMsgTime			120L		/* ticks, = 2 seconds */
#define	DelimItemOffset			4
#define	VMaxItem				5
#define	AutoVMaxItem			6
#define	MAX_RealSimulationTime	100
#define	SimSpeedItem			5
#define	MagIMapFirstET			6
#define	MagIMapNoMag			17
#define	IPO_OffsetItemNum		6
#define	TSI_SampleRateTextItem	6
#define	TSI_UseTSButtonItem		7

				/*	Global external vars from bb.c	*/
extern	pIterationVars		pCurrentVars;
extern	hBBPreferences		hBBPrefs;
extern	int					CurrentModifiers;
extern	Boolean				SuppressErrorMessages;
extern	Boolean				System7OrLater;

	/*	From bbImpactGraphics.c; for SetVMax	*/
extern	double		AutoVMax;

	/*	From bbWindows.c; for DoStopTimeSeriesDialog 	*/
extern	Boolean		ClosingAllWindows;

DialogPtr		StuckBallDialog;
DialogPtr		PhaseRangeDialog;
DialogPtr		PrintingDialog;
DialogPtr		SimSpeedDialog;

double			IC_v0, IC_delta;				/*	Initial Conditions	*/
double			TP_A, TP_freq, TP_damp;			/*	Table Parameters	*/
int				TS_NumPeriods, TS_NumAmplitudes;	/*	Trajectory Scaling	*/
int				It_Accuracy;					/*	Iteration (Params)	*/
long			It_Cycles;
Boolean			VM_UseAuto;						/*	Set Max. Velocity	*/
double			VM_VMax;
Str255			AutoVMaxString;
BooleanRect		MI_UsesFull;					/*	Magnify Impact ...	*/
DoubleRect		MI_Actual, *pMI_Full;
Str255			MI_OffsetString;					/* Phases only	*/
int				IPO_ImpPhsOffset;				/*	Impact Phases Offset	*/
int				TSI_SampleRate;					/*	Time Series Info	*/

long			StickingMessageStartTime;
int				DelimButtonItemNum, DelimTextItemNum;
int				GWO_CheckItemNum, StdFileOutlineItemNum;
Boolean			FirstAccessStdFileDLog;
ControlHandle	SVBch;		/*	for PhaseRangeDialog; ScrollValBar	*/
Str255			GraphRangeStrings[ItemsInRect];


SelectBadField(DialogPtr dp, int index, Boolean *PreviousAllRight)
{
	if (*PreviousAllRight)
		mySelItemText(dp, index);
	*PreviousAllRight = FALSE;
}

	/* ------------------------------------------------------ */
	/* -------- Actual dialog/alert calling routines -------- */
	/* ------------------------------------------------------ */

BadParamAlert(char *s)
{
	ParamText(s,"\P","\P","\P");
	if (!CautionAlert(BadParamAlertID,NIL)) SysBeep(30);
}

AllBadValsAlert(char *s)
{
	ParamText(s,"\P","\P","\P");
	if (!CautionAlert(AllBadValsAlertID,NIL)) SysBeep(30);
}

FixBadValAlert(char *s)
{
	ParamText(s,"\P","\P","\P");
	if (!CautionAlert(FixBadValAlertID,NIL)) SysBeep(30);
}

FixBadIntValAlert(char *s)
{
	ParamText(s,"\P","\P","\P");
	if (!CautionAlert(FixBadIntValAlertID,NIL)) SysBeep(30);
}

ErrorAlert(char *s1, char *s2, char *s3, char *s4)
{
	ParamText(s1,s2,s3,s4);
	if (!StopAlert(GeneralAlertID,NIL))
		SysBeep(30);
}

ERROR(char *s)
{
	if (!SuppressErrorMessages)
		ErrorAlert(s,"\P","\P","\P");
}

DoAboutBBDialog()
{
	DialogPtr	dp;

	dp = GetNewDialog(AboutBBDLogID, NIL, -1L);
	if (HandleDialog(dp, NO_SELECT_ITEM,
			Cancel, UserOutlineItemNum, NIL, NIL, NIL) == OK) ;
	DisposDialog(dp);
}

Boolean CheckInitCondDialog(DialogPtr dp, int itemNo)
{
	Boolean	NumbersOK = TRUE;
	int		i = -1;

	if (itemNo != OK) return(TRUE);

	++i;
	if ( !myGetItemFloat(dp, ETItem1+i+i, &IC_v0) ) {
		FixBadValAlert("\PInitial Velocity");
		SelectBadField(dp, ETItem1+i+i, &NumbersOK);
	}
	++i;
	if ( !myGetItemFloat(dp, ETItem1+i+i, &IC_delta) ) {
		FixBadValAlert("\PInitial Phase");
		SelectBadField(dp, ETItem1+i+i, &NumbersOK);
	}

	return(NumbersOK);
}

Boolean DoInitCondDialog(nV0, nDelta, myClearSolutions, RestartIteration)
	double	*nV0, *nDelta;
	Boolean	*myClearSolutions, *RestartIteration;
{
	DialogPtr	dp;
	int			ClearSolChkBox = 9, RestartChkBox = 10; /* "constants" */
	int			i;

	dp = GetNewDialog(InitCondDLogID, NIL, -1L);

	i = -1;
	++i; mySetItemFloat(dp, ETItem1+i+i,*nV0,DLogPrecision);
	++i; mySetItemFloat(dp, ETItem1+i+i,*nDelta,DLogPrecision);
	mySetItemCtlValue(dp, ClearSolChkBox,(int)*myClearSolutions);
	mySetItemCtlValue(dp, RestartChkBox, (int) *RestartIteration);

	if (HandleDialog(dp, ETItem1, Cancel, UserOutlineItemNum,
			NIL, NIL, &CheckInitCondDialog) == OK) {
		*nV0 = IC_v0;
		*nDelta = IC_delta;

		*myClearSolutions = myGetItemCtlValue(dp,ClearSolChkBox);

		*RestartIteration = myGetItemCtlValue(dp,RestartChkBox);
		DialogReturn(dp, TRUE);
	}
	else DialogReturn(dp, FALSE);
}

Boolean CheckPhysSysDialog(DialogPtr dp, int itemNo)
{
	Boolean	NumbersOK = TRUE;
	int		i = -1;

	if (itemNo != OK) return(TRUE);

	++i;
	if ( !myGetItemFloat(dp, ETItem1+i+i, &TP_A) ) {
		FixBadValAlert("\PTable Amplitude");
		SelectBadField(dp, ETItem1+i+i, &NumbersOK);
	}
	if (TP_A < 0.0) {
		ERROR("\PTable Amplitude must be non-negative.");
		SelectBadField(dp, ETItem1+i+i, &NumbersOK);
	}
	++i;
	if ( !myGetItemFloat(dp, ETItem1+i+i, &TP_freq) ) {
		FixBadValAlert("\PTable Frequency");
		SelectBadField(dp, ETItem1+i+i, &NumbersOK);
	}
	if (TP_freq < 0.0) {
		ERROR("\PTable Frequency must be non-negative.");
		SelectBadField(dp, ETItem1+i+i, &NumbersOK);
	}
	++i;
	if ( !myGetItemFloat(dp, ETItem1+i+i, &TP_damp) ) {
		FixBadValAlert("\PDamping Coefficient");
		SelectBadField(dp, ETItem1+i+i, &NumbersOK);
	}
	return(NumbersOK);
}

Boolean DoPhysSysDialog(Boolean *RestartIteration)
{
	DialogPtr	dp;
	int			RestartChkBox = 11; /* "constant" */
	int			i;

	dp = GetNewDialog(PhysSysDLogID, NIL, -1L);

	i = -1;
	++i; mySetItemFloat(dp, ETItem1+i+i, A, DLogPrecision);
	++i; mySetItemFloat(dp, ETItem1+i+i, freq, DLogPrecision);
	++i; mySetItemFloat(dp, ETItem1+i+i, damp, DLogPrecision);
	mySetItemCtlValue(dp, RestartChkBox, (int) *RestartIteration);

	if (HandleDialog(dp, ETItem1, Cancel, UserOutlineItemNum,
			NIL, NIL, &CheckPhysSysDialog) == OK) {
		A = TP_A;	freq = TP_freq;		damp = TP_damp;
		*RestartIteration = myGetItemCtlValue(dp,RestartChkBox);
		DialogReturn(dp, TRUE);
	}
	else DialogReturn(dp, FALSE);
}

OutOfRangeAlert(char *s, long minL, long maxL)
{
	Str255	s1, s2;

	LongToString(minL, CP(s1));
	LongToString(maxL, CP(s2));
	ParamText(s, s1, s2, "\P");
	if (!CautionAlert(OutOfRangeAlertID,NIL)) SysBeep(30);
}

Boolean myGetLongParamItem(dp, itemNo, pl, s, minL, maxL)
	DialogPtr	dp;
	int			itemNo;
	long		*pl, minL, maxL;
	char		*s;
	/* Returns long from item itemNo, subject to contraints minL & maxL */
{
	double	d = 0.0;

	if (!myGetItemLong(dp, itemNo, pl)) {
		if (myGetItemFloat(dp, itemNo, &d))
			{	FixBadIntValAlert(s); return(FALSE);	}
		else
			{	FixBadValAlert(s); return(FALSE);		}
	}
/*
	if (!myGetItemLong(dp, itemNo, pl)) {
		myGetItemText(dp, itemNo, CP(itemStr));
		if (IsFloat(CP(itemStr)))	FixBadIntValAlert(s);
		else						FixBadValAlert(s);
		return(FALSE);
	}
*/
	if (!myGetItemFloat(dp, itemNo, &d)) ;
	if (fabs(d-*pl) > AlmostZero)
		{	FixBadIntValAlert(s); return(FALSE);	}
	if ( !( (*pl >= minL) && (*pl <= maxL) ) )
		{ OutOfRangeAlert(s, minL, maxL); return(FALSE);}
	return(TRUE);
}

Boolean CheckTrajScalingDialog(DialogPtr dp, int itemNo)
{
	int		i = -1;
	long	tmpL;
	Boolean	NumbersOK = TRUE;

	if (itemNo != OK) return(TRUE);

	++i;
	if ( myGetLongParamItem(dp, ETItem1+i+i, &tmpL,
			"\PNumber of Periods", 1L, 256L) )
		TS_NumPeriods = (int) tmpL;
	else
		SelectBadField(dp, ETItem1+i+i, &NumbersOK);

	++i;
	if ( myGetLongParamItem(dp, ETItem1+i+i, &tmpL,
			"\PNumber of Amplitudes", 1L, 256L) )
		TS_NumAmplitudes = (int) tmpL;
	else
		SelectBadField(dp, ETItem1+i+i, &NumbersOK);

	return(NumbersOK);
}

Boolean DoTrajScalingDialog(int *NumPeriods, int *NumAmplitudes)
{
	DialogPtr	dp;
	int			i;

	dp = GetNewDialog(TrajScalingDLogID, NIL, -1L);

	i = -1;
	++i; mySetItemLong(dp, ETItem1+i+i, (long) *NumPeriods);
	++i; mySetItemLong(dp, ETItem1+i+i, (long) *NumAmplitudes);

	if (HandleDialog(dp, ETItem1, Cancel, UserOutlineItemNum,
			NIL, NIL, &CheckTrajScalingDialog) == OK) {
		*NumPeriods = TS_NumPeriods;
		*NumAmplitudes = TS_NumAmplitudes;
		DialogReturn(dp, TRUE);
	}
	else DialogReturn(dp, FALSE);
}

Boolean CheckIterationDialog(DialogPtr dp, int itemNo)
{
	int		i = -1;
	long	tmpL;
	Boolean	NumbersOK = TRUE;

	if (itemNo != OK) return(TRUE);

	++i;
	if ( myGetLongParamItem(dp, ETItem1+i+i, &tmpL,
			"\PAccuracy", 0L, 14L) )
		It_Accuracy = (int) tmpL;
	else
		SelectBadField(dp, ETItem1+i+i, &NumbersOK);

	++i;
	if ( myGetLongParamItem(dp, ETItem1+i+i, &tmpL,
			"\PNumber of Cycles", 0L, LONG_MAX) )
		It_Cycles = tmpL;
	else
		SelectBadField(dp, ETItem1+i+i, &NumbersOK);

	return(NumbersOK);
}

Boolean DoIterationDialog(int *pAccuracy, long *pCycles)
{
	DialogPtr	dp;
	int			i;

	dp = GetNewDialog(IterationDLogID, NIL, -1L);

	i = -1;
	++i; mySetItemLong(dp, ETItem1+i+i, (long) *pAccuracy);
	++i; mySetItemLong(dp, ETItem1+i+i, *pCycles);

	if (HandleDialog(dp, ETItem1, Cancel, UserOutlineItemNum,
			NIL, NIL, &CheckIterationDialog) == OK) {
		*pAccuracy = It_Accuracy;
		*pCycles = It_Cycles;
		DialogReturn(dp, TRUE);
	}
	else DialogReturn(dp, FALSE);
}

StartStickingMessage(int theStuckCount)
{
	static	int		NumStrings;
	int		TooMuchSticking = 11;
	char	*s;
	Handle	strList;
	int		index;
	Str255	FunnyString;

	if (theStuckCount == 1)	randSeed = TickCount();

	switch (theStuckCount) {
		case STUCK_FOREVER:			s = "\P     Ball Stuck Forever    ";	break;
		case 1:						s = "\P   Ball Sticking to Table  ";	break;
		default:	if (theStuckCount <= TooMuchSticking-1)
									s = "\P  Ball Stuck Twice or More ";
					else if (theStuckCount == TooMuchSticking)
									s = "\P   Aren't you bored yet?   ";
					else {
						if (theStuckCount == TooMuchSticking+1) {
							strList = GetResource(StringListResType,
									StuckStringsResID);
							NumStrings = *(int *) *strList;
						}
						index = ((unsigned) Random()) % NumStrings + 1;
						s = CP(FunnyString);
						GetIndString(s, StuckStringsResID, index);
/*
						switch ( ((unsigned) Random()) % 12 ) {
							case 0:	s = "\P   Aren't you bored yet?   ";	break;
							case 1:	s = "\P    This isn't stopping... ";	break;
							case 2:	s = "\P A moose once bit my sister";	break;
							case 3:	s = "\PThis doesn't get any better";	break;
							case 4:	s = "\P            Ho hum...      ";	break;
							case 5:	s = "\P   Z-Z-Z-Z-Z-Z-Z-Z-Z-Z-Z   ";	break;
							case 6:	s = "\P   Ball Stuck Many Times   ";	break;
							case 7:	s = "\PIt's still getting stuck...";	break;	(*	Rest from Jeremiah	*)
							case 8:	s = "\PWake me up when you're done";	break;
							case 9:	s = "\P Elvis lives. Paul is dead.";	break;
							case 10:s = "\PI bet you love Philip Glass";	break;
							default:s = "\P  Do I hear grass growing? ";	break;
						}
*/
					}
					break;
	}
/*
	if (theStuckCount == 1)	s = "\P Ball Sticking to Table ";
	else					s = "\PBall Stuck Twice or More";
	if (theStuckCount > 10)	s = "\P Aren't you bored yet?  ";
*/
	ParamText(s, "\P", "\P", "\P");
	StuckBallDialog = GetNewDialog(StuckBallDLogID, NIL, -1L);
	ShowWindow(StuckBallDialog);
	DrawDialog(StuckBallDialog);
	StickingMessageStartTime = TickCount();
}

StopStickingMessage()
{
	while ( TickCount() < (StickingMessageStartTime+MinStuckMsgTime) )
		;
	HideWindow(StuckBallDialog);
	DisposDialog(StuckBallDialog);
	StuckBallDialog = NIL;
	return;
}

SetPhaseString(int Phase, char *PhaseString)
/*	Converts Phase/100.0 to string; only works for <=3 digit Phase */
{
	int	i=0;

	if (Phase<0) {
		PhaseString[++i] = '-';
		Phase = -Phase;
	}
	PhaseString[++i] = '0' + Phase/100;
	if (Phase % 100 != 0) {
		PhaseString[++i] = '.';
		PhaseString[++i] = '0' + ( Phase/10 % 10 );
		if (Phase % 10 != 0)
			PhaseString[++i] = '0' + Phase % 10;
	}
	PhaseString[0] = i;
}

SetDialogPhases(ControlHandle ch)
{
	int		MinPhaseItem = 12, MaxPhaseItem = 13;	/* "constants" */
	int		thePhase = GetCtlValue(ch);
	Str255	ps;

	SetPhaseString(thePhase, CP(ps));
	mySetItemText(PhaseRangeDialog, 12, CP(ps));

	SetPhaseString(thePhase+100, CP(ps));
	mySetItemText(PhaseRangeDialog, 13, CP(ps));
}

Boolean PhaseRangeDLogEventFilter(DialogPtr dp, EventRecord *pTheEvent,
		int *pItemHit, Boolean *pHowCKFShouldReturn)
{
	char	theChar;
	long	rc;
	int		ScrollAmount, NewValue;

	switch (pTheEvent->what) {
		case keyDown: 
		case autoKey:
			theChar = BitAnd(pTheEvent->message, charCodeMask);
			rc = GetCRefCon(SVBch);
			switch (theChar) {
				case Left_ArrowKey:		ScrollAmount = -LoWord(rc);	break;
				case Right_ArrowKey:	ScrollAmount =  LoWord(rc);	break;
				case Up_ArrowKey:		ScrollAmount =  HiWord(rc);	break;
				case Down_ArrowKey:		ScrollAmount = -HiWord(rc);	break;
				default:				return(FALSE);
			}
			NewValue = GetCtlValue(SVBch) + ScrollAmount;
			if (NewValue > GetCtlMax(SVBch))	NewValue = GetCtlMax(SVBch);
			if (NewValue < GetCtlMin(SVBch))	NewValue = GetCtlMin(SVBch);
			SetCtlValue(SVBch, NewValue);
			*pHowCKFShouldReturn = TRUE;
			return(TRUE);
	}
	return(FALSE);
}

Boolean PhaseRangeDLogItemFilter(DialogPtr dp, int itemHit)
{
	int		ZeroRange = 8, HalfRange = 9;	/*	"constants"	*/
	int		NewValue;

	if (itemHit == ZeroRange)
		NewValue = GetCtlMax(SVBch);
	else if (itemHit == HalfRange)
		NewValue = -50;
	else
		return(FALSE);
	SetCtlValue(SVBch, NewValue);
	return(FALSE);
}

Boolean DoPhaseRangeDialog(int *pMinPhaseInt, int windex)
{
						/*	next two lines are "constants"	*/
	int				ScrollValBarItem = 5;
	long			PageScrollAmount = 10L, ButtonScrollAmount = 1L;
	int				type;
	Handle			item;
	Rect			box;
	char			*ps;

	PhaseRangeDialog =GetNewDialog(PhaseRangeDLogID, NIL, -1L);

	switch (windex) {
		case BifD:	ps = "\PBifurcation Diagram ";	break;
		case BAttr:	ps = "\PBasins of Attraction ";	break;
		default:	ps = "\P                 ";	break;
	}
	ParamText(ps, "\P", "\P", "\P");

	GetDItem(PhaseRangeDialog, ScrollValBarItem, &type, &item, &box);
	if (type & ctrlItem) {
		DisposeControl((ControlHandle) item);
		SVBch = NewControl(PhaseRangeDialog, &box, "\P", TRUE,
				*pMinPhaseInt, -PHASE_INT_RANGE, 0,
				ScrollValBarCDEFID<<4, (long) &SetDialogPhases);
		SetCRefCon(SVBch, (PageScrollAmount << 16) | (ButtonScrollAmount) );
		SetDItem(PhaseRangeDialog, ScrollValBarItem, ctrlItem+resCtrl, (Handle) SVBch, &box);
	}
	if (HandleDialog(PhaseRangeDialog, NO_SELECT_ITEM,  Cancel,
			UserOutlineItemNum, &PhaseRangeDLogEventFilter,
			&PhaseRangeDLogItemFilter, NIL) == OK) {
		*pMinPhaseInt = GetCtlValue(SVBch);
		DialogReturn(PhaseRangeDialog, TRUE);
	}
	else DialogReturn(PhaseRangeDialog, FALSE);
}

Boolean SetVMaxDLogItemFilter(DialogPtr dp, int itemHit)
{
	Str255	tmpVMaxString;

	switch (itemHit) {
		case VMaxItem:
				if (myGetItemCtlValue(dp, AutoVMaxItem)) {
					myGetItemText(dp, VMaxItem,
							CP(tmpVMaxString));
					if ( sComp(CP(AutoVMaxString),
							CP(tmpVMaxString)) != 0 )
						mySetItemCtlValue(dp, AutoVMaxItem,FALSE);
				}
				break;
		case AutoVMaxItem:
				if (!myGetItemCtlValue(dp, AutoVMaxItem)) {
					mySetItemText(dp, VMaxItem, CP(AutoVMaxString));
					mySelItemText(dp, VMaxItem);
				}
				break;
	}
	return(FALSE);
}

Boolean CheckSetVMaxDialog(DialogPtr dp, int itemNo)
{
	Boolean	GoodValue = TRUE;

	if (itemNo != OK) return(TRUE);

	VM_UseAuto = myGetItemCtlValue(dp,AutoVMaxItem);
	if (!VM_UseAuto) {
		if ( !myGetItemFloat(dp, VMaxItem, &VM_VMax) ) {
			FixBadValAlert("\Pmaximum velocity");
			SelectBadField(dp, VMaxItem, &GoodValue);
		}
		else if (VM_VMax <=0) {
			ERROR("\PMaximum velocity must be greater than 0.");
			SelectBadField(dp, VMaxItem, &GoodValue);
		}
	}
	else
		VM_VMax = AutoVMax;
	return(GoodValue);
}

Boolean DoSetVMaxDialog(double *pVMax, Boolean *pUseAutoVMax)
{
	DialogPtr	dp;

	dp =GetNewDialog(SetVMaxDLogID, NIL, -1L);

	myFToStrNoPad(AutoVMax, CP(AutoVMaxString), 0, DLogPrecision);
	if (*pUseAutoVMax)
		mySetItemText(dp, VMaxItem, CP(AutoVMaxString));
	else
		mySetItemFloat(dp, VMaxItem, *pVMax, DLogPrecision);
	mySetItemCtlValue(dp, AutoVMaxItem, *pUseAutoVMax);

	if (HandleDialog(dp, VMaxItem,  Cancel, UserOutlineItemNum,
			NIL, &SetVMaxDLogItemFilter, &CheckSetVMaxDialog) == OK) {
		*pUseAutoVMax = VM_UseAuto;
		*pVMax = VM_VMax;
		DialogReturn(dp, TRUE);
	}
	else DialogReturn(dp, FALSE);
}

Boolean SimSpeedDLogEventFilter(DialogPtr dp, EventRecord *pTheEvent,
		int *pItemHit, Boolean *pHowCKFShouldReturn)
{
	char	theChar;
	long	rc;
	int		ScrollAmount, NewValue;

	switch (pTheEvent->what) {
		case keyDown: 
		case autoKey:
			theChar = BitAnd(pTheEvent->message, charCodeMask);
			rc = GetCRefCon(SVBch);
			switch (theChar) {
				case Left_ArrowKey:		ScrollAmount = -LoWord(rc);	break;
				case Right_ArrowKey:	ScrollAmount =  LoWord(rc);	break;
				case Up_ArrowKey:		ScrollAmount =  HiWord(rc);	break;
				case Down_ArrowKey:		ScrollAmount = -HiWord(rc);	break;
				default:				return(FALSE);
			}
			NewValue = GetCtlValue(SVBch) + ScrollAmount;
			if (NewValue > GetCtlMax(SVBch))	NewValue = GetCtlMax(SVBch);
			if (NewValue < GetCtlMin(SVBch))	NewValue = GetCtlMin(SVBch);
			SetCtlValue(SVBch, NewValue);
			*pHowCKFShouldReturn = TRUE;
			return(TRUE);
	}
	return(FALSE);
}

Boolean DoSimulationSpeedDialog(long *pRealPerTicks)
{
						/*	next two lines are "constants"	*/
	int				ScrollValBarItem = 5;
	long			PageScrollAmount = 10L, ButtonScrollAmount = 1L;
	int				type;
	Handle			item;
	Rect			box;

	SimSpeedDialog =GetNewDialog(SimulationSpeedDLogID, NIL, -1L);
	GetDItem(SimSpeedDialog, ScrollValBarItem, &type, &item, &box);
	if (type & ctrlItem) {
		DisposeControl((ControlHandle) item);
		SVBch = NewControl(SimSpeedDialog, &box, "\P", TRUE,
				MAX_RealSimulationTime - (int) *pRealPerTicks/6,
				0, MAX_RealSimulationTime,
				ScrollValBarCDEFID<<4, 0L /* (long) &SetDialogPhases */ );
		SetCRefCon(SVBch, (PageScrollAmount << 16) | (ButtonScrollAmount) );
		SetDItem(SimSpeedDialog, ScrollValBarItem, ctrlItem+resCtrl, (Handle) SVBch, &box);
	}
	if (HandleDialog(SimSpeedDialog, NO_SELECT_ITEM,  Cancel,
			UserOutlineItemNum, &SimSpeedDLogEventFilter, NIL, NIL) == OK) {
		*pRealPerTicks = (MAX_RealSimulationTime - GetCtlValue(SVBch)) * 6;
		DialogReturn(SimSpeedDialog, TRUE);
	}
	else DialogReturn(SimSpeedDialog, FALSE);
}

/*
Boolean CheckSimulationSpeedDialog(DialogPtr dp, int itemNo)
{
	Boolean	GoodValue = TRUE;

	if (itemNo != OK) return(TRUE);

	if ( !myGetItemFloat(dp, SimSpeedItem, &SS_RealPerTime) ) {
		FixBadValAlert("\Psimulation speed");
		SelectBadField(dp, SimSpeedItem, &GoodValue);
	}
	else if (SS_RealPerTime < 0.0) {
		ERROR("\PSimulation speed must be greater than or equal to 0.");
		SelectBadField(dp, SimSpeedItem, &GoodValue);
	}
	return(GoodValue);
}

Boolean DoSimulationSpeedDialog(double *pRealPerTime)
{
	DialogPtr	dp;

	dp =GetNewDialog(SimulationSpeedDLogID, NIL, -1L);

	mySetItemFloat(dp, SimSpeedItem, *pRealPerTime, 3);

	if (HandleDialog(dp, SimSpeedItem,  Cancel, UserOutlineItemNum,
			NIL, NIL, &CheckSimulationSpeedDialog) == OK) {
		*pRealPerTime = SS_RealPerTime;
		DialogReturn(dp, TRUE);
	}
	else DialogReturn(dp, FALSE);
}
*/

Boolean MagIMapDLogItemFilter(DialogPtr dp, int itemHit)
{
	Str255	s;
	int		i, index;

	switch (itemHit) {
		case MagIMapFirstET:
		case MagIMapFirstET+3:
		case MagIMapFirstET+6:
		case MagIMapFirstET+9:
				index = (itemHit - MagIMapFirstET) / 3;
				index = (index%2)*2 + (1-index/2);	/*	0->1, 1->3, 2->0, 3->2	*/
				if (myGetItemCtlValue(dp, itemHit+1)) {
					myGetItemText(dp, itemHit, CP(s));
					if ( sComp(CP(GraphRangeStrings[index]),
							CP(s)) != 0 )
						mySetItemCtlValue(dp, itemHit+1, FALSE);
				}
				break;
		case MagIMapFirstET+1:
		case MagIMapFirstET+4:
		case MagIMapFirstET+7:
		case MagIMapFirstET+10:
				index = (itemHit - MagIMapFirstET - 1) / 3;
				index = (index%2)*2 + (1-index/2);	/*	0->1, 1->3, 2->0, 3->2	*/
				if (!myGetItemCtlValue(dp, itemHit)) {
					mySetItemText(dp, itemHit-1,
							CP(GraphRangeStrings[index]));
					mySelItemText(dp, itemHit-1);
				}
				break;
		case MagIMapNoMag:
				for(i=0; i <= 3; ++i) {
					index = (i%2)*2 + (1-i/2);	/*	0->1, 1->3, 2->0, 3->2	*/
					mySetItemText(dp, MagIMapFirstET+3*i,
							CP(GraphRangeStrings[index]));
					mySetItemCtlValue(dp, MagIMapFirstET+3*i+1, TRUE);
				}
				mySelItemText(dp, MagIMapFirstET);
				break;
	}
	return(FALSE);
}

Boolean CheckMagnifyImpactMapDialog(DialogPtr dp, int itemNo)
{
	int		i, index;
	Str255	ErrorStrings[ItemsInRect];
	Boolean	NumbersOK = TRUE;
	double	tmpF;

	if (itemNo != OK) return(TRUE);

	i=0;
	sEquate(CP(ErrorStrings[i++]), "\Pminimum velocity");
	sEquate(CP(ErrorStrings[i++]), "\Pminimum phase");
	sEquate(CP(ErrorStrings[i++]), "\Pmaximum velocity");
	sEquate(CP(ErrorStrings[i++]), "\Pmaximum phase");

	for(i=ymin_i;i<ItemsInRect;++i) {
		index = (1-i%2)*2 + i/2;		/*	0->2, 1->0, 2->3, 3->1	*/
		index = MagIMapFirstET+3*index;
		MI_UsesFull[i] = myGetItemCtlValue(dp, index+1);
		if (!MI_UsesFull[i]) {
			if ( !myGetItemFloat(dp, index, &tmpF) ) {
				FixBadValAlert(CP(ErrorStrings[i]));
				SelectBadField(dp, index, &NumbersOK);
				return(FALSE);
			}
			switch (i) {
				case ymin_i:
					break;
				case xmin_i:
					if ( (tmpF<(*pMI_Full)[xmin_i]) ||
							(tmpF>(*pMI_Full)[xmax_i]) ) {
						ERROR("\PPhase must be between current min and max phase.");
						SelectBadField(dp, index, &NumbersOK);
						return(FALSE);
					}
					break;
				case ymax_i:
					if (tmpF <= MI_Actual[ymin_i]) {
						ERROR("\PMaximum velocity must be greater than minimum velocity.");
						SelectBadField(dp, index, &NumbersOK);
						return(FALSE);
					}
					break;
				case xmax_i:
					if ( (tmpF<(*pMI_Full)[xmin_i]) ||
							(tmpF>(*pMI_Full)[xmax_i]) ) {
						ERROR("\PPhase must be between current min and max phase.");
						SelectBadField(dp, index, &NumbersOK);
						return(FALSE);
					}
					if (tmpF <= MI_Actual[xmin_i]) {
						ERROR("\PMaximum phase must be greater than minimum phase.");
						SelectBadField(dp, index, &NumbersOK);
						return(FALSE);
					}
					if (tmpF-MI_Actual[xmin_i] > 1) {
						ERROR("\PPhase range must be less than or equal to 1.");
						SelectBadField(dp, index, &NumbersOK);
						return(FALSE);
					}
					break;
			}
			MI_Actual[i] = tmpF;
		}
		else {
			MI_Actual[i] = (*pMI_Full)[i];
			if (i == ymax_i)
				if (MI_Actual[ymin_i] >= MI_Actual[i]) {
					ERROR("\PMinimum velocity must be less than maximum velocity.");
					SelectBadField(dp, MagIMapFirstET+3*2, &NumbersOK);
					return(FALSE);
				}
		}
	}
	return(NumbersOK);
}

Boolean DoMagnifyImpactMapDialog(DoubleRect *Actual, DoubleRect *Full,
		BooleanRect *UsesFull)
{
	DialogPtr	dp;
	int			i, index;

	pMI_Full = Full;

	dp =GetNewDialog(MagnifyImpactMapDLogID, NIL, -1L);

	for(i=ymin_i;i<ItemsInRect;++i) {
		MI_Actual[i] = (*Actual)[i];
		index = (1-i%2)*2 + i/2;		/*	0->2, 1->0, 2->3, 3->1	*/
		index = MagIMapFirstET+3*index;
		myFToStrNoPad((*Full)[i], CP(GraphRangeStrings[i]),
				0, DLogPrecision);
		if ((*UsesFull)[i])
			mySetItemText(dp, index, CP(GraphRangeStrings[i]));
		else
			mySetItemFloat(dp, index, MI_Actual[i], DLogPrecision);
		mySetItemCtlValue(dp, index+1, (*UsesFull)[i]);
	}

	if (HandleDialog(dp, MagIMapFirstET,  Cancel, UserOutlineItemNum,
			NIL, &MagIMapDLogItemFilter, &CheckMagnifyImpactMapDialog) == OK) {
		for(i=ymin_i;i<ItemsInRect;++i) {
			(*Actual)[i] = MI_Actual[i];
			(*UsesFull)[i] = MI_UsesFull[i];
		}
		DialogReturn(dp, TRUE);
	}
	else DialogReturn(dp, FALSE);
}

Boolean CheckMagnifyImpactPhasesDialog(DialogPtr dp, int itemNo)
{
	int		i, index;
	Str255	ErrorStrings[ItemsInRect];
	Boolean	NumbersOK = TRUE;
	double	tmpF;

	if (itemNo != OK) return(TRUE);

	i=0;
	sEquate(CP(ErrorStrings[i++]), "\Pminimum phase n+");
	sEquate(CP(ErrorStrings[i++]), "\Pminimum phase n");
	sEquate(CP(ErrorStrings[i++]), "\Pmaximum phase n+");
	sEquate(CP(ErrorStrings[i++]), "\Pmaximum phase n");
	sAppend(CP(ErrorStrings[0]), CP(MI_OffsetString));
	sAppend(CP(ErrorStrings[2]), CP(MI_OffsetString));

	for(i=ymin_i;i<ItemsInRect;++i) {
		index = (1-i%2)*2 + i/2;		/*	0->2, 1->0, 2->3, 3->1	*/
		index = MagIMapFirstET+3*index;
		MI_UsesFull[i] = myGetItemCtlValue(dp, index+1);
		if (!MI_UsesFull[i]) {
			if ( !myGetItemFloat(dp, index, &tmpF) ) {
				FixBadValAlert(CP(ErrorStrings[i]));
				SelectBadField(dp, index, &NumbersOK);
				return(FALSE);
			}
			switch (i) {
				case ymin_i:
					if ( (tmpF<(*pMI_Full)[ymin_i]) ||
							(tmpF>(*pMI_Full)[ymax_i]) ) {
						ERROR("\PPhase must be between current min and max phase.");
						SelectBadField(dp, index, &NumbersOK);
						return(FALSE);
					}
					break;
				case xmin_i:
					if ( (tmpF<(*pMI_Full)[xmin_i]) ||
							(tmpF>(*pMI_Full)[xmax_i]) ) {
						ERROR("\PPhase must be between current min and max phase.");
						SelectBadField(dp, index, &NumbersOK);
						return(FALSE);
					}
					break;
				case ymax_i:
					if ( (tmpF<(*pMI_Full)[ymin_i]) ||
							(tmpF>(*pMI_Full)[ymax_i]) ) {
						ERROR("\PPhase must be between current min and max phase.");
						SelectBadField(dp, index, &NumbersOK);
						return(FALSE);
					}
					if (tmpF <= MI_Actual[ymin_i]) {
						ERROR("\PMaximum phase must be greater than minimum phase.");
						SelectBadField(dp, index, &NumbersOK);
						return(FALSE);
					}
					if (tmpF-MI_Actual[ymin_i] > 1) {
						ERROR("\PPhase range must be less than or equal to 1.");
						SelectBadField(dp, index, &NumbersOK);
						return(FALSE);
					}
					break;
				case xmax_i:
					if ( (tmpF<(*pMI_Full)[xmin_i]) ||
							(tmpF>(*pMI_Full)[xmax_i]) ) {
						ERROR("\PPhase must be between current min and max phase.");
						SelectBadField(dp, index, &NumbersOK);
						return(FALSE);
					}
					if (tmpF <= MI_Actual[xmin_i]) {
						ERROR("\PMaximum phase must be greater than minimum phase.");
						SelectBadField(dp, index, &NumbersOK);
						return(FALSE);
					}
					if (tmpF-MI_Actual[xmin_i] > 1) {
						ERROR("\PPhase range must be less than or equal to 1.");
						SelectBadField(dp, index, &NumbersOK);
						return(FALSE);
					}
					break;
			}
			MI_Actual[i] = tmpF;
		}
		else
			MI_Actual[i] = (*pMI_Full)[i];
	}

	return(NumbersOK);
}

Boolean DoMagnifyImpactPhasesDialog(DoubleRect *Actual, DoubleRect *Full,
		BooleanRect *UsesFull, int IPO)
{
	DialogPtr	dp;
	int			i, index;

	pMI_Full = Full;

	dp =GetNewDialog(MagnifyImpactPhasesDLogID, NIL, -1L);
	LongToString((long) IPO, CP(MI_OffsetString));
	ParamText(CP(MI_OffsetString), "\P", "\P", "\P");

	for(i=ymin_i;i<ItemsInRect;++i) {
		MI_Actual[i] = (*Actual)[i];
		index = (1-i%2)*2 + i/2;		/*	0->2, 1->0, 2->3, 3->1	*/
		index = MagIMapFirstET+3*index;
		myFToStrNoPad((*Full)[i], CP(GraphRangeStrings[i]),
				0, DLogPrecision);
		if ((*UsesFull)[i])
			mySetItemText(dp, index, CP(GraphRangeStrings[i]));
		else
			mySetItemFloat(dp, index, MI_Actual[i], DLogPrecision);
		mySetItemCtlValue(dp, index+1, (*UsesFull)[i]);
	}

	if (HandleDialog(dp, MagIMapFirstET,  Cancel, UserOutlineItemNum,
			NIL, &MagIMapDLogItemFilter, &CheckMagnifyImpactPhasesDialog) == OK) {
		for(i=ymin_i;i<ItemsInRect;++i) {
			(*Actual)[i] = MI_Actual[i];
			(*UsesFull)[i] = MI_UsesFull[i];
		}
		DialogReturn(dp, TRUE);
	}
	else DialogReturn(dp, FALSE);
}

Boolean CheckImpactPhasesOffsetDialog(DialogPtr dp, int itemNo)
{
	long	tmpL;
	Boolean	GoodValue = TRUE;

	if (itemNo != OK) return(TRUE);

	if ( myGetLongParamItem(dp, IPO_OffsetItemNum, &tmpL,
			"\PImpact Phases Offset",
			(long) MIN_ImpactPhasesOffset, (long) MAX_ImpactPhasesOffset) )
		IPO_ImpPhsOffset = (int) tmpL;
	else
		SelectBadField(dp, IPO_OffsetItemNum, &GoodValue);

	return(GoodValue);
}

Boolean DoImpactPhasesOffsetDialog(int *pImpPhsOffset)
{
	DialogPtr	dp;

	dp = GetNewDialog(ImpactPhasesOffsetDLogID, NIL, -1L);

	mySetItemLong(dp, IPO_OffsetItemNum, (long) *pImpPhsOffset);

	if (HandleDialog(dp, IPO_OffsetItemNum, Cancel, UserOutlineItemNum,
			NIL, NIL, &CheckImpactPhasesOffsetDialog) == OK) {
		*pImpPhsOffset = IPO_ImpPhsOffset;
		DialogReturn(dp, TRUE);
	}
	else DialogReturn(dp, FALSE);
}


		/*	------------------------------------------	*/
		/*	----------	Time Series Dialogs	----------	*/
		/*	------------------------------------------	*/

Boolean CheckTimeSeriesInfoDialog(DialogPtr dp, int itemNo)
{
	long	tmpL;
	Boolean	GoodValue = TRUE;

	if (itemNo != OK) return(TRUE);

	if ( myGetLongParamItem(dp, TSI_SampleRateTextItem,
			&tmpL, "\PSample Rate", 1L, 1024L) )
		TSI_SampleRate = (int) tmpL;
	else
		SelectBadField(dp, TSI_SampleRateTextItem, &GoodValue);

	return(GoodValue);
}

Boolean DoTimeSeriesInfoDialog(int *pSampleRate, Boolean *pUseTimeSeries)
{
	DialogPtr	dp;

	dp = GetNewDialog(TimeSeriesInfoDLogID, NIL, -1L);

	mySetItemLong(dp, TSI_SampleRateTextItem, (long) *pSampleRate);
	mySetItemCtlValue(dp, TSI_UseTSButtonItem, (int) (*pUseTimeSeries) );

	if (HandleDialog(dp, TSI_SampleRateTextItem, Cancel, UserOutlineItemNum,
			NIL, NIL, &CheckTimeSeriesInfoDialog) == OK) {
		*pSampleRate = TSI_SampleRate;
		*pUseTimeSeries =
				myGetItemCtlValue(dp, TSI_UseTSButtonItem);
		DialogReturn(dp, TRUE);
	}
	else DialogReturn(dp, FALSE);
}

Boolean DoStartTimeSeriesDialog()
{
	DialogPtr	dp;

	if (SHIFT_DOWN)		return(TRUE);
	if (OPTION_DOWN)	return(FALSE);

	dp = GetNewDialog(StartTimeSeriesDLogID, NIL, -1L);
	if (HandleDialog(dp, NO_EDIT_TEXT,
			Cancel, UserOutlineItemNum, NIL, NIL, NIL) == OK)
		DialogReturn(dp, TRUE)
	else
		DialogReturn(dp, FALSE)
}

Boolean DoStopTimeSeriesDialog()
{
	DialogPtr	dp;

	if (SHIFT_DOWN)		return(TRUE);
	if (!ClosingAllWindows)
		if (OPTION_DOWN)	return(FALSE);

	dp = GetNewDialog(StopTimeSeriesDLogID, NIL, -1L);
	if (HandleDialog(dp, NO_EDIT_TEXT,
			Cancel, UserOutlineItemNum, NIL, NIL, NIL) == OK)
		DialogReturn(dp, TRUE)
	else
		DialogReturn(dp, FALSE)
}

Boolean DoShowWindowDialog(int windex)
{
	DialogPtr	dp;
	char		*p;

	if (SHIFT_DOWN)		return(TRUE);
	if (OPTION_DOWN)	return(FALSE);

	switch (windex) {
		case TSer:	p = "\PTime Series";			break;
		case BifD:	p = "\PBifurcation Diagram";	break;
		case BAttr:	p = "\PBasins of Attraction";	break;
	}
	ParamText(p, "\P", "\P", "\P");

	dp = GetNewDialog(ShowWindowDLogID, NIL, -1L);
	if (HandleDialog(dp, NO_EDIT_TEXT,
			Cancel, UserOutlineItemNum, NIL, NIL, NIL) == OK)
		DialogReturn(dp, TRUE)
	else
		DialogReturn(dp, FALSE)
}


		/*	----------------------------------------------	*/
		/*	----------	System-related dialogs	----------	*/
		/*	----------------------------------------------	*/

Boolean DoRunInBackgroundDLog(long *pBkgdSleep, Boolean *pSuppressBkgdSnd)
{
	DialogPtr	dp;
	int			SuppressBkgdSndBox = 5,
				RIBUserOutlineItemNum = 4,
				RIBLastStopItem = 3,
				RIBSlowButton = 1;	/* "constants" */
	int			theResult;

	if (SHIFT_DOWN)	{
		*pBkgdSleep = CATNAP;
		return(TRUE);
	}

	dp = GetNewDialog(RunInBackgroundDLogID, NIL, -1L);

	mySetItemCtlValue(dp, SuppressBkgdSndBox, *pSuppressBkgdSnd );

	theResult = HandleDialog(dp, NO_SELECT_ITEM,
			RIBLastStopItem, RIBUserOutlineItemNum, NIL, NIL, NIL);
	if (theResult == RIBLastStopItem)
		DialogReturn(dp, FALSE);

	*pBkgdSleep = (theResult == RIBSlowButton) ? CATNAP : DAY_DREAM;
	*pSuppressBkgdSnd =
			myGetItemCtlValue(dp,SuppressBkgdSndBox);
	DialogReturn(dp, TRUE);
}

StartPrintingDialog()
{
	PrintingDialog = GetNewDialog(PrintingDLogID, NIL, -1L);
	ShowWindow(PrintingDialog);
	DrawDialog(PrintingDialog);
}

Boolean HandlePrintingDialog()
{
	EventRecord	theEvent;
	int			itemHit;
	DialogPtr	theDialog;

	if (GetNextEvent((mDownMask | keyDownMask), &theEvent)) {
		if ( (theEvent.modifiers & cmdKey) && 
				((char) (theEvent.message & charCodeMask) == '.') )
			return(FALSE);
		if (IsDialogEvent(&theEvent))
			if (DialogSelect(&theEvent, &theDialog, &itemHit))
				if (theDialog == PrintingDialog)
					if (itemHit == 1)
						return(FALSE);
	}
	return(TRUE);
}

StopPrintingDialog()
{
	HideWindow(PrintingDialog);
	DisposDialog(PrintingDialog);
}

Boolean DoQuitWarningDialog(Boolean *pWantsToSave, int windex)
{
	DialogPtr	dp;
	Str255		TextWindowTitle;
	int			CloseItem;
	char		*ps;

	if (SuppressSaveWarnings) {
		*pWantsToSave = FALSE;
		return(TRUE);
	}

	if (SHIFT_DOWN || OPTION_DOWN) {
		*pWantsToSave = SHIFT_DOWN;
		return(TRUE);
	}

	dp = GetNewDialog(QuitWarningDLogID, NIL, -1L);
	switch (windex) {
		case TimeSeries:
				ps = GetSaveFName(windex);
				if (ps[0] == 0)
					ParamText("\PTime Series", "\P", "\P", "\P");
				else
					ParamText("\PTime Series", "\P ", ps, "\P");
				break;
		case BifDiagram:
				ps = GetSaveFName(windex);
				if (ps[0] == 0)
					ParamText("\PBifurcation Diagram", "\P", "\P", "\P");
				else
					ParamText("\PBifurcation Diagram", "\P ", ps, "\P");
				break;
		case BasinsAttraction:
				ps = GetSaveFName(windex);
				if (ps[0] == 0)
					ParamText("\PBasins of Attraction", "\P", "\P", "\P");
				else
					ParamText("\PBasins of Attraction", "\P ", ps, "\P");
				break;
		default:
				GetTextTitle(CP(TextWindowTitle));
				ParamText("\P", CP(TextWindowTitle), "\P", "\P");
				break;
	}
	CloseItem = HandleDialog(dp, NO_EDIT_TEXT,
			Cancel+1, UserOutlineItemNum+1, NIL, NIL, NIL);
	switch (CloseItem) {
		case OK:
		case OK+1:
			*pWantsToSave = (CloseItem == OK);
			DialogReturn(dp, TRUE)
		case OK+2:
			DialogReturn(dp, FALSE)
	}
}

Boolean DoSaveWarningDialog(Boolean *pWantsToSave, int windex)
{
	DialogPtr	dp;
	char		*p;
	int			CloseItem;

	if (SuppressSaveWarnings) {
		*pWantsToSave = FALSE;
		return(TRUE);
	}

	if (SHIFT_DOWN || OPTION_DOWN) {
		*pWantsToSave = SHIFT_DOWN;
		return(TRUE);
	}

	dp = GetNewDialog(SaveWarningDLogID, NIL, -1L);
	switch (windex) {
		case TimeSeries:		p = "\Ptime series";			break;
		case BifDiagram:		p = "\Pbifurcation diagram";	break;
		case BasinsAttraction:	p = "\Pbasins of attraction";	break;
		default:				p = "\Psimulation";				break;
	}
	ParamText(p, "\P", "\P", "\P");

	CloseItem = HandleDialog(dp, NO_EDIT_TEXT,
			Cancel+1, UserOutlineItemNum+1, NIL, NIL, NIL);
	switch (CloseItem) {
		case OK:
		case OK+1:
			*pWantsToSave = (CloseItem == OK);
			DialogReturn(dp, TRUE)
		case OK+2:
			DialogReturn(dp, FALSE)
	}
}

Boolean DoBackgroundDialog(Boolean *pRunInBackground,
		Boolean *pRunQuickly, Boolean *pUseSoundInBG)
{
	DialogPtr	dp;
	int			FirstCheckBox = 11;	/*	"constant"	*/
	int			i;

	dp = GetNewDialog(BackgroundDLogID, NIL, -1L);
	ParamText("\PBouncing Ball", "\P", "\P", "\P");

	i = FirstCheckBox;
	mySetItemCtlValue(dp, i++, *pRunInBackground);
	mySetItemCtlValue(dp, i++, !*pRunQuickly);
	mySetItemCtlValue(dp, i++, *pRunQuickly);
	mySetItemCtlValue(dp, i++, *pUseSoundInBG);

	if (HandleDialog(dp, NO_EDIT_TEXT,
			Cancel+1, UserOutlineItemNum, NIL, NIL, NIL) == OK) {
		i = FirstCheckBox;
		*pRunInBackground = myGetItemCtlValue(dp, i++);
		i++;
		*pRunQuickly = myGetItemCtlValue(dp, i++);
		*pUseSoundInBG = myGetItemCtlValue(dp, i++);
		DialogReturn(dp, TRUE);
	}
	else
		DialogReturn(dp, FALSE);
}


	/*	----------------------------------------------	*/
	/*	------	Standard File Package Routines,	------	*/
	/*	------	 customized for Bouncing Ball	------	*/
	/*	----------------------------------------------	*/

pascal int MakeStdFileOutline(int item, DialogPtr StdFileDialog)
{
	if (FirstAccessStdFileDLog) {		/*	(item == -1)?	*/
		if (!System7OrLater)
			MakeUserItemOutline(StdFileDialog, StdFileOutlineItemNum);
		FirstAccessStdFileDLog = FALSE;
	}
}

pascal int DelimiterDLog(int itemNo, DialogPtr SaveAsDialog)
{
	int				type;
	Handle			item;
	Rect			box;
	ControlHandle	PUMch;

	if (FirstAccessStdFileDLog) {
		if (!System7OrLater)
			MakeUserItemOutline(SaveAsDialog, StdFileOutlineItemNum);
		FirstAccessStdFileDLog = FALSE;

		GetDItem(SaveAsDialog, DelimButtonItemNum, &type, &item, &box);
		if (type & ctrlItem) {
			DisposeControl((ControlHandle) item);
			PUMch = NewControl(SaveAsDialog, &box, "\PDelimiter:", TRUE,
					0, 0, 0, PopUpCDEFID << 4,
					(((long) DelimPopUpMenuID) << 16) | (DelimiterIndex+1));
			box = (**PUMch).contrlRect;
			SetDItem(SaveAsDialog, DelimButtonItemNum, ctrlItem+resCtrl,
					(Handle) PUMch, &box);
		}
	}
	if (itemNo == DelimButtonItemNum) {
		GetDItem(SaveAsDialog, DelimButtonItemNum, &type, &item, &box);
		DelimiterIndex = GetCtlValue((ControlHandle) item) - 1;
	}
	return(itemNo);
}

/*	From DialogUtils.c	*/
extern	Point	StdFilePoint;

Boolean BBSaveAsDialog(char *prompt, int *vRef, char *fname)
{
	SFReply		mySFReply;
	DialogTHndl	dth;
	GrafPtr		CurPort;
	int			theFont, theSize;
	Rect		DelimButtonRect = {0,0,0,0};
	Rect		OutlineRect = {0,0,0,0};
	int			OldDelimiterIndex = DelimiterIndex;

	if ((dth = (DialogTHndl) GetFreshResource('DLOG', putDlgID)) == 0L) {
		ERROR("\PSystem Error: Unable to create Save As... dialog box");
		return(FALSE);
	}

	if (!System7OrLater)
		if ((StdFileOutlineItemNum = AddDItem(putDlgID, TRUE, userItem,
				&OutlineRect, "\P")) == 0 ) {
			ERROR("\PSystem Error: Unable to create Save As... dialog box");
			return(FALSE);
		}
	
	(**dth).boundsRect.bottom += 25;
	GetPort(&CurPort);
	theFont = CurPort->txFont;	theSize = CurPort->txSize;
	TextFont(systemFont);		TextSize(12);
	OffsetRect(&DelimButtonRect, 14, (**dth).boundsRect.bottom - 25);
	TextFont(theFont);			TextSize(theSize);

	if ((DelimButtonItemNum = AddDItem(putDlgID, FALSE, ctrlItem+radCtrl,
			&DelimButtonRect, "\P")) == 0 ) {
		ERROR("\PSystem Error: Unable to create Save As... dialog box");
		return(FALSE);
	}

	FirstAccessStdFileDLog = TRUE;
	SFPPutFile(StdFilePoint, prompt, fname, &DelimiterDLog, &mySFReply,
			putDlgID, &CommandKeyFilter);
	ReleaseDialogResources(dth);
	if (!mySFReply.good) {
		DelimiterIndex = OldDelimiterIndex;
		return(FALSE);
	}
	*vRef = mySFReply.vRefNum;
	sEquate(fname, CP(mySFReply.fName));
	return(TRUE);
}

Boolean BBSpecialSaveAsDialog(char *prompt, int *vRef, char *fname)
{
	SFReply		mySFReply;
	DialogTHndl	dth;
	Rect		OutlineRect = {0,0,0,0};

	if ((dth = (DialogTHndl) GetFreshResource('DLOG', putDlgID)) == 0L) {
		ERROR("\PSystem Error: Unable to create Save As... dialog box");
		return(FALSE);
	}

	if (!System7OrLater)
		if ((StdFileOutlineItemNum = AddDItem(putDlgID, TRUE, userItem,
				&OutlineRect, "\P")) == 0 ) {
			ERROR("\PSystem Error: Unable to create Save As... dialog box");
			return(FALSE);
		}

	if (!System7OrLater) {
		FirstAccessStdFileDLog = TRUE;
		SFPPutFile(StdFilePoint, prompt, fname, &MakeStdFileOutline, &mySFReply,
				putDlgID, &CommandKeyFilter);
	}
	else
		SFPPutFile(StdFilePoint, prompt, fname, NIL, &mySFReply,
				putDlgID, &CommandKeyFilter);
	ReleaseDialogResources(dth);
	if (!mySFReply.good)
		return(FALSE);
	*vRef = mySFReply.vRefNum;
	sEquate(fname, CP(mySFReply.fName));
	return(TRUE);
}

pascal int GWO_Check(int item, DialogPtr OpenDialog)
{
	if (FirstAccessStdFileDLog) {		/*	(item == -1)?	*/
		mySetItemCtlValue(OpenDialog, GWO_CheckItemNum, GraphWhileOpening);
		if (!System7OrLater)
			MakeUserItemOutline(OpenDialog, StdFileOutlineItemNum);
		FirstAccessStdFileDLog = FALSE;
	}
	if (item == GWO_CheckItemNum) {
		GraphWhileOpening = !GraphWhileOpening;
		mySetItemCtlValue(OpenDialog, item, GraphWhileOpening);
	}
	return(item);
}

Boolean BBOpenDialog(OSType *OpenFileTypes, int NumOSTypes, int *vRef,
		char *fname, OSType *fType)
{
	SFReply		mySFReply;
	SFTypeList	mySFTypeList;
	int			i;
	DialogTHndl	dth;
	GrafPtr		CurPort;
	int			theFont, theSize;
	Rect		OutlineRect  = {0,0,0,0}, GWO_CheckRect = {0,0,0,0};
	Boolean		OldGraphWhileOpening = GraphWhileOpening;
	char		*pGWOString;

	if ((dth = (DialogTHndl) GetFreshResource('DLOG', getDlgID)) == 0L) {
		ERROR("\PSystem Error: Unable to create Open... dialog box");
		return(FALSE);
	}

	if (!System7OrLater)
		if ((StdFileOutlineItemNum = AddDItem(getDlgID, TRUE, userItem,
				&OutlineRect, "\P")) == 0 ) {
			ERROR("\PSystem Error: Unable to create Open... dialog box");
			return(FALSE);
		}

	if (TRUE) {
		pGWOString = "\PGraph while opening (if appropriate)";
		(**dth).boundsRect.bottom += 12;
		GetPort(&CurPort);
		theFont = CurPort->txFont;	theSize = CurPort->txSize;
		TextFont(systemFont);		TextSize(12);
		OffsetRect(&GWO_CheckRect, 11, (**dth).boundsRect.bottom - 18);
		GWO_CheckRect.right += StringWidth(pGWOString) + 20;
		GWO_CheckRect.bottom += 16;
		TextFont(theFont);			TextSize(theSize);
	
		if ((GWO_CheckItemNum = AddDItem(getDlgID, FALSE, ctrlItem+chkCtrl,
				&GWO_CheckRect, pGWOString)) == 0 ) {
			ERROR("\PSystem Error: Unable to create Open... dialog box");
			return(FALSE);
		}
	}

	FirstAccessStdFileDLog = TRUE;
	for(i=0; i<NumOSTypes; ++i)
		mySFTypeList[i] = *OpenFileTypes++;
	SetCheckCharIfNoCommand(FALSE);
	if (TRUE)
		SFPGetFile(StdFilePoint, "\P", 0L, NumOSTypes, &mySFTypeList, &GWO_Check,
				&mySFReply, getDlgID, &CommandKeyFilter);
	else
		SFPGetFile(StdFilePoint, "\P", 0L,1, &mySFTypeList, &MakeStdFileOutline,
				&mySFReply, getDlgID, &CommandKeyFilter);
	SetCheckCharIfNoCommand(TRUE);
	ReleaseDialogResources(dth);
	if (!mySFReply.good) {
		GraphWhileOpening = OldGraphWhileOpening;
		return(FALSE);
	}

	*vRef = mySFReply.vRefNum;
	sEquate(fname, CP(mySFReply.fName));
	*fType = mySFReply.fType;
	return(TRUE);
}

Boolean BBOpenSpecialDialog(OSType *OpenFileTypes, int NumOSTypes,
		int *vRef, char *fname, OSType *fType)
{
	int			i;
	SFReply		mySFReply;
	SFTypeList	mySFTypeList;
	DialogTHndl	dth;
	Rect		OutlineRect  = {0,0,0,0};

	if ((dth = (DialogTHndl) GetFreshResource('DLOG', getDlgID)) == 0L) {
		ERROR("\PSystem Error: Unable to create Open... dialog box");
		return(FALSE);
	}

	if (!System7OrLater)
		if ((StdFileOutlineItemNum = AddDItem(getDlgID, TRUE, userItem,
				&OutlineRect, "\P")) == 0 ) {
			ERROR("\PSystem Error: Unable to create Open... dialog box");
			return(FALSE);
		}

	for(i=0; i<NumOSTypes; ++i)
		mySFTypeList[i] = *OpenFileTypes++;
	SetCheckCharIfNoCommand(FALSE);
	if (!System7OrLater) {
		FirstAccessStdFileDLog = TRUE;
		SFPGetFile(StdFilePoint, "\P", 0L,NumOSTypes, &mySFTypeList, &MakeStdFileOutline,
					&mySFReply, getDlgID, &CommandKeyFilter);
	}
	else
		SFPGetFile(StdFilePoint, "\P", 0L,NumOSTypes, &mySFTypeList, NIL,
					&mySFReply, getDlgID, &CommandKeyFilter);
	SetCheckCharIfNoCommand(TRUE);
	ReleaseDialogResources(dth);
	if (!mySFReply.good)
		return(FALSE);

	*vRef = mySFReply.vRefNum;
	sEquate(fname, CP(mySFReply.fName));
	*fType = mySFReply.fType;
	return(TRUE);
}
