/*	bb segment dealing with files: opening, saving
 */

#include <DialogUtils.h>
#include <FileUtils.h>
#include "bbPrototypes.h"

#define	BBSFType				'BBAL'
#define	TextSFType				'TEXT'
#define	DefaultsSFType			'BBDF'
#define	BifDSFType				'BifD'
#define	BAttrSFType				'BAtt'
#define	BBWindowsResType		'BBWn'
#define	BBPrefsResType			'BBDf'
#define	NumOpenFileTypes		3

enum {	DefaultWindowID = 128, DefaultPrefsID = DefaultWindowID, OpenWindowID };

/*
#define pSaveFName				(CP(SaveFName))
*/
#define	pDefaultsFName			(CP(DefaultsFName))
#define	DefaultDefaultsFName	"\PBouncing Ball Defaults"

#define	DelimiterStringResID	129
#define	TAB_Delimiter			"\P\t"
#define	COMMA_Delimiter			"\P, "
#define	CR_Delimiter			"\P\r"
#define	TimeSeriesString		"\PTime Series"
#define	MIN_SPACES				2
#define	PARAM_LINES				3


	/* From bbText.c */
/* (CurText local var)
extern	BBText			*CurText;
extern	TEHandle		HeaderTE;
extern	TEHandle		BodyTE;
extern	BodyOFHandles	BodyOverflow;
extern	long			BodyOFIndex;
*/

		/*	From bb.c	*/
extern	hBBPreferences		hBBPrefs;
extern	Boolean				PrintingOnly;

		/*	From bbBifDiag.c, bbBasAttr.c	*/
extern	Boolean			DrawBD, DrawBA;

		/*	From bbWindows.c	*/
extern	WindowLocation		wd[NumWindows], CurWD[NumWindows];
extern	int					DefaultWindowNum, CurWindowNum;

int				SaveVRef, oSaveVRef;
int				TSSaveVRef, BDSaveVRef, BASaveVRef;
myFileNameType	oSaveFName;
myFileNameType	NormalSaveFName, TSSaveFName;
myFileNameType	BDSaveFName, BASaveFName;

/*	2.01 addition
 */
int				tmpVRef;
myFileNameType	tmpSaveFName;
/*	end, 2.01 addition
 */

int				SaveWindex;
int				OpenWindex;

char			*pSaveFName;
int				*pVRef;
char			*SaveAsPrompt;
OSType			SaveType;

Boolean			GoodSaveFName, GoodTSSaveFName;
Boolean			GoodBDSaveFName, GoodBASaveFName;
int				Untitleds, TSUntitleds, BDUntitleds, BAUntitleds;

int				NumParamFiles, ParamFileCount;
Boolean			LastParamFile;

int		oDelimiterIndex;
char	DelimiterStrings[NUM_Delims][DelimDescLength];
char	*Delimiters[NUM_Delims];
long	LinesRead;	/*	used with CR-delimiter only	*/

int				SysVRef, DefaultsOpenVRef;
myFileNameType	DefaultsFName;

hWindowLocs		hWLs;
Size			hWLsSize;


SetANilFileName(int theWindex)
{
	switch (theWindex) {
		case TSer:
			TSSaveFName[0] = 0;	break;
		case BifD:
			BDSaveFName[0] = 0;	break;
		case BAttr:
			BASaveFName[0] = 0;	break;
		default:
			NormalSaveFName[0] = 0;
			break;
	}
}

SetNilFileNames()
{
	NormalSaveFName[0] = TSSaveFName[0] = BDSaveFName[0] =
			BASaveFName[0] = oSaveFName[0] = 0;
}

char *GetSaveFName(int windex)
{
	switch(windex) {
		case TSer:	return(CP(TSSaveFName));
		case BifD:	return(CP(BDSaveFName));
		case BAttr:	return(CP(BASaveFName));
		default:	return(CP(NormalSaveFName));
	}
}

SetFileName(int theWindex)
{
	SaveWindex = theWindex;
	switch (theWindex) {
		case TSer:
					pSaveFName = CP(TSSaveFName);
					pVRef = &TSSaveVRef;
					SaveAsPrompt = "\PSave time series as:";
					SaveType = TextSFType;
					break;
		case BifD:
					pSaveFName = CP(BDSaveFName);
					pVRef = &BDSaveVRef;
					SaveAsPrompt = "\PSave bifurcation diag. as:";
					SaveType = BifDSFType;
					break;
		case BAttr:
					pSaveFName = CP(BASaveFName);
					pVRef = &BASaveVRef;
					SaveAsPrompt = "\PSave basins of attr. as:";
					SaveType = BAttrSFType;
					break;
		default:
					pSaveFName = CP(NormalSaveFName);
					pVRef = &SaveVRef;
					SaveAsPrompt = "\PSave impact data as:";
					SaveType = TextSFType;
					break;
	}
}

int DecideWhichSaveWindow()
{
	int	theWindex;

	switch (theWindex = WindowIndex(FrontWindow())) {
		case BifD:	if (!DrawBD)	theWindex = TxOut;	break;
		case BAttr:	if (!DrawBA)	theWindex = TxOut;	break;
		case TSer:										break;
		default:	theWindex = TxOut;					break;
	}
	SetFileName(theWindex);
	return(theWindex);
}

InitBBFiles()
{
	int	i;

	for(i=FIRST_Delim;i<LAST_Delim;++i)
		GetIndString(DelimStr(i),DelimiterStringResID,i+1);
	Delimiters[SPACE_Delim]	= 	"\P";
	Delimiters[TAB_Delim]	= 	TAB_Delimiter;
	Delimiters[COMMA_Delim]	= 	COMMA_Delimiter;
	Delimiters[CR_Delim]	= 	CR_Delimiter;
	/*	SetNilFileNames(); */	/* Done in bbGraphics */
	SetNilFileNames();

	SysVRef = BootDrive;
	DefaultsOpenVRef = SysVRef;
	sEquate(pDefaultsFName, DefaultDefaultsFName);
/*
	DefaultsFileExists = FALSE;
*/
	GoodSaveFName= GoodTSSaveFName= GoodBDSaveFName= GoodBASaveFName= TRUE;
	Untitleds = TSUntitleds = BDUntitleds = BAUntitleds = 0;
}

int	GetNextParamFile()
{
	AppFile	theAppFile;
	int		windex;
	char	*pfname;
	int		*pvref;

	for(;;) {
		if (++ParamFileCount > NumParamFiles) return(NoWindow);
		LastParamFile = (ParamFileCount == NumParamFiles);
		GetAppFiles(ParamFileCount, &theAppFile);
		switch (theAppFile.fType) {
			case BifDSFType:	
							pfname = CP(BDSaveFName);	pvref = &BDSaveVRef;
							windex = BifD;				break;
			case BAttrSFType:	
							pfname = CP(BASaveFName);	pvref = &BASaveVRef;
							windex = BAttr;				break;
			case TextSFType:	
							pfname = CP(NormalSaveFName);	pvref = &SaveVRef;
							windex = TxOut;				break;
			default:
							continue;
		}
		break;
	}

	sEquate(pfname, CP(theAppFile.fName));
	*pvref = theAppFile.vRefNum;
	ClrAppFiles(ParamFileCount);	/*	only clear if handled successfully	*/
	return(windex);
}

int	BBGetParamFiles()
{
	int		appMessage;
	AppFile	theAppFile;
	int		i;
	Boolean	GotADefaultFile;

	CountAppFiles(&appMessage, &NumParamFiles);
	if (NumParamFiles == 0) return(NoAppFile);

	GotADefaultFile = FALSE;
	if (appMessage != appPrint)
		for(i=NumParamFiles; i>=1; --i) {
			GetAppFiles(i, &theAppFile);
			if (theAppFile.fType == DefaultsSFType) {
				if (!GotADefaultFile) {
					sEquate(pDefaultsFName, CP(theAppFile.fName));
					DefaultsOpenVRef = theAppFile.vRefNum;
					GotADefaultFile = TRUE;
				}
				ClrAppFiles(i);
			}
		}

	ParamFileCount = 0;
	return(appMessage);
/*
	*pParamWindex = GetNextParamFile();
	if (*pParamWindex != NoWindow)
		return(appMessage);
	else
		return(NoAppFile);
*/
}

/*
CallHelp()
{
	Help("\PBouncing Ball Help", 0);
}
*/

Boolean	BBGetFileLine(char *theLine)
{
	Boolean GoodLine;
	Str255	NextLine;
	int		i;

	GoodLine = ReadTextFileLine(theLine);
	if (DelimiterIndex == COMMA_Delim)
		sSubstC(theLine, ',', SPACE);
	if ( (DelimiterIndex != CR_Delim) || (!GoodLine) )
		return(GoodLine);

		/*	Only get here if CR_Delim, GoodLine	*/
	++LinesRead;
	if ( (theLine[1] == 'b') || (theLine[1] == 'B') )	/* "Ball stuck" line */
		return(TRUE);
	RETURN_IF_FALSE( ReadTextFileLine(CP(NextLine)) );
	theLine[theLine[0]] = SPACE;
	sAppend(theLine, CP(NextLine));
	if (LinesRead > PARAM_LINES)
		for (i=1; i<=2; ++i) {
			RETURN_IF_FALSE( ReadTextFileLine(CP(NextLine)) );
			theLine[theLine[0]] = SPACE;
			sAppend(theLine, CP(NextLine));
		}
	return(TRUE);
}

Boolean CheckFirstLine(char *theLine)
/*	Returns FALSE if a Time Series file, TRUE (and sets delimiter) otherwise	*/
{
	int	i;

	if ( StrFind("\pmplitude", theLine) ) {
		ResetFileRead();
		return(TRUE);
	}
	if ( StrFind("\peries", theLine) )
		return(FALSE);
	for (i=TAB_Delim; i<=CR_Delim; ++i)
		if ( StrFind(DelimStr(i), theLine) ) {
			DelimiterIndex = i;
			break;
		}
	if (DelimiterIndex == CR_Delim)
		LinesRead = 0;
	return(TRUE);
}

CloseOpenFile()
{
	if ( (OpenWindex == BifD) || (OpenWindex == BAttr) )
		if (!StopResFile())
			;
	StopFileRead();
}

Boolean GetOpenWindows()
{
	int			i;

	hWLs = (hWindowLocs) GetResource(BBWindowsResType, OpenWindowID);
	if (hWLs != NIL) {
		CurWindowNum = (**hWLs).num;
		for(i=0; i<CurWindowNum; ++i)
			CurWD[i] = (**hWLs).ws[i];
	}

	hWLsSize = GetHandleSize((Handle) hWLs);
	if (OpenWindex == TxOut)
		if (!StopResFile())
			;
	return(hWLs != NIL);
}

Boolean	OpenFile()
{
	Boolean	ResForkExists;

	SetFileName(OpenWindex);
	RETURN_IF_FALSE( StartFileRead(*pVRef, pSaveFName, FALSE) );
	if (OpenWindex == TxOut) {
		oDelimiterIndex = DelimiterIndex;
		DelimiterIndex = SPACE_Delim;
	}
	if (!OpenResFork(pSaveFName, *pVRef, OpenWindex != TxOut, &ResForkExists)) {
		StopFileRead();
		return(FALSE);
	}
/*
	RETURN_IF_FALSE( OpenResFork(pSaveFName, *pVRef) );
*/
	if (!PrintingOnly)
		if (ResForkExists)
			if (GetOpenWindows())
				if (OpenWindex == TxOut)
					SetOpenWindows();
	return(TRUE);
}

ResetDelimiter()
{
	DelimiterIndex = oDelimiterIndex;
}

Boolean	GetOpenFileName(int *pwindex)
{
	OSType	theTypes[NumOpenFileTypes], fType;
/*	2.01 deletion
	Str255	s;
	int		vref;
*/

	theTypes[0] = TextSFType;
	theTypes[1] = BifDSFType;
	theTypes[2] = BAttrSFType;
/*	2.01 change
	RETURN_IF_FALSE( BBOpenDialog((OSType *) &theTypes,
			NumOpenFileTypes, &vref, CP(s), &fType) );
*/
	RETURN_IF_FALSE( BBOpenDialog((OSType *) &theTypes,
			NumOpenFileTypes, &tmpVRef, CP(tmpSaveFName), &fType) );
/*	end, 2.01 change
 */

	switch (fType) {
		case TextSFType:
/*	2.01 deletion
						SaveVRef = vref;
						sEquate(CP(NormalSaveFName), CP(s));
*/
						OpenWindex = TxOut;
						break;
		case BifDSFType:
/*	2.01 deletion
						BDSaveVRef = vref;
						sEquate(CP(BDSaveFName), CP(s));
*/
						OpenWindex = BifD;
						break;
		case BAttrSFType:
/*	2.01 deletion
						BASaveVRef = vref;
						sEquate(CP(BASaveFName), CP(s));
*/
						OpenWindex = BAttr;
						break;
	}

	*pwindex = OpenWindex;
	return(TRUE);
}

/* New 2.01 routine
 */

SetOpenFileName(int windex)
{
	switch (windex) {
		case TxOut:
						SaveVRef = tmpVRef;
						sEquate(CP(NormalSaveFName), CP(tmpSaveFName));
						break;
		case BifD:
						BDSaveVRef = tmpVRef;
						sEquate(CP(BDSaveFName), CP(tmpSaveFName));
						break;
		case BAttr:
						BASaveVRef = tmpVRef;
						sEquate(CP(BASaveFName), CP(tmpSaveFName));
						break;
	}
}

/*	End, new 2.01 routine
 */

ResetFileName()
{
	sEquate( pSaveFName, CP(oSaveFName) );
	*pVRef = oSaveVRef;
}

GetTitleString(int theWindex, char *ts)
{
	sEquate(ts, "\pBouncing Ball");
	if (theWindex == TSer) {
		sInsertC(ts, '/', (int) ts[0]+1);
		sAppend(ts, TimeSeriesString);
	}
	if (DelimiterIndex != SPACE_Delim) {
		if (theWindex == TSer)
			sAppend(ts, "\P, ");
		else
			sInsertC(ts, '/', (int) ts[0]+1);
		sAppend(ts, DelimStr(DelimiterIndex));
		sAppend(ts, "\P-delimited");
	}
	sInsertC(ts, RETURN_CHAR, (int) ts[0]+1);
}

Boolean CloseSaveFile(Boolean NeedToDelete)
{
	Boolean	GoodOp = TRUE;

	if (!StopFileSave())
		GoodOp = FALSE;

	if ( (SaveWindex == BifD) || (SaveWindex == BAttr) )
		if (!StopResFile())
			GoodOp = FALSE;
/* deleted 11/11/91--causing resource problems?
	if ( (GetHandleSize((Handle) hWLs) == hWLsSize) &&
			(hWLsSize != 0L) )
		DisposHandle((Handle) hWLs);
*/
	if (NeedToDelete)
		if (DeleteSaveFile()) ;
	return(GoodOp);
}

Boolean SaveCurWindows()
{
	int			i;

	hWLs = (hWindowLocs) NewHandle(WindowLocsSize(CurWindowNum));
	if (hWLs == NIL) return(FALSE);
	(**hWLs).num = CurWindowNum;
	for(i=0; i<CurWindowNum; ++i)
		(**hWLs).ws[i] = CurWD[i];
	RETURN_IF_FALSE( AddNewResource((Handle) hWLs, BBWindowsResType,
			OpenWindowID) );
	hWLsSize = GetHandleSize((Handle) hWLs);

	if (SaveWindex == TxOut)
		if (!StopResFile())
			;
	return(TRUE);
}

Boolean SaveFile()
{
	int			i;
	long		TotalAmt, TEWriteStart;
	int			NumSpaces;
	char		*ds = Delimiters[DelimiterIndex];
	Str255		TitleString;
	BBTextPtr	CurText;
	int			DefaultVol;
	Str255		s;
	Boolean		ResForkExists;
					/*	for time series kludge below	*/
	Handle		tmpHandle;
	long		tmpLen, l;
	Boolean		GoodOp;

	RETURN_IF_FALSE( StartFileSave(pSaveFName, *pVRef) );

	hWLs = NIL;
	if (SaveWindex != TSer) {
		RETURN_IF_FALSE( OpenResFork(pSaveFName, *pVRef, TRUE, &ResForkExists) );
		SetCurWindows(SaveWindex);
		SaveCurWindows();
		if ( (SaveWindex == BifD) || (SaveWindex == BAttr) )
			return(SaveSpecialFile(SaveWindex));
	}

/*
	if (SaveWindex == TxOut) {
		RETURN_IF_FALSE( OpenResFork(pSaveFName, *pVRef) );
		SetCurWindows(SaveWindex);
	}
*/

	CurText = GetTextFromTindex(GetTextIndex(SaveWindex));
	GetTitleString(SaveWindex, CP(TitleString));
	RETURN_IF_FALSE( WritePtrToFile(CP(TitleString) + 1,
			(long) TitleString[0]) );

	NumSpaces = 0;
	if ( (SaveWindex == TimeSeries) && (DelimiterIndex != SPACE_Delim) ) {
			/*	kludge: was putting delimiter between "Sample" and "rate:"
			 *	Another option would be to put two spaces after colons
			 *	in header; this has the advantage that it's compatible
			 *	with current files and looks better.
			 *		-TA 10/23/91
			 */
		tmpLen = (**CurText->header).teLength;
		tmpHandle = NewHandle(tmpLen);
		if (tmpHandle == NIL) return(FALSE);
		for(l=0; l<tmpLen; ++l)
			(*tmpHandle)[l] = (*(**CurText->header).hText)[l];
		if ( (*tmpHandle)[6] == SPACE )		/*	Should always be true	*/
			(*tmpHandle)[6] = FORCED_SPACE;
		GoodOp = Delim_WriteHandleToFile(tmpHandle,
				tmpLen, ds, 1, &NumSpaces, FALSE, TRUE);
		DisposHandle(tmpHandle);
		RETURN_IF_FALSE(GoodOp);
	}
	else
		RETURN_IF_FALSE( Delim_WriteTEtoFile(CurText->header, 0L,
				ds, 1, &NumSpaces, FALSE, FALSE) );

		/* Handle "overflow", for > 32K chars */
	NumSpaces = MIN_SPACES+1;
	if (CurText->OF.index > 0L) {
		for (i=0; i<=CurText->OF.index / BodyOFIncrement; ++i) {
			TotalAmt = (CurText->OF.index >= (i+1)*BodyOFIncrement) ?
					BodyOFIncrement : CurText->OF.index % BodyOFIncrement;
			RETURN_IF_FALSE( Delim_WriteHandleToFile(CurText->OF.hands[i],
					TotalAmt, ds, MIN_SPACES, &NumSpaces, FALSE, FALSE) );
		}
		TEWriteStart = (long) (**CurText->body).lineStarts[OVERFLOW_MSG_LINES];
	}
	else
		TEWriteStart = 0L;

	RETURN_IF_FALSE( Delim_WriteTEtoFile(CurText->body, TEWriteStart,
			ds, MIN_SPACES, &NumSpaces, FALSE, FALSE) );

	RETURN_IF_FALSE( CloseSaveFile(FALSE) );

	return(TRUE);
}

GetUntitledFileName(char *ps)
{
	switch (SaveWindex) {
		case TSer:	sprintf(ps, "Time Series %d", ++TSUntitleds);
					break;
		case BifD:	sprintf(ps, "Bifurcation Diagram %d", ++BDUntitleds);
					break;
		case BAttr:	sprintf(ps, "Basins of Attraction %d", ++BAUntitleds);
					break;
		default:	sprintf(ps, "Untitled %d", ++Untitleds);
					break;
	}
	CtoPstr(ps);
}

SetGoodSaveFName(Boolean NewGoodSaveFName)
{
	switch (SaveWindex) {
		case TSer:	GoodTSSaveFName = NewGoodSaveFName;	break;
		case BifD:	GoodBDSaveFName = NewGoodSaveFName;	break;
		case BAttr:	GoodBASaveFName = NewGoodSaveFName;	break;
		default:	GoodSaveFName = NewGoodSaveFName;	break;
	}
}

Boolean	GetGoodSaveFName()
{
	switch (SaveWindex) {
		case TSer:	return(GoodTSSaveFName);
		case BifD:	return(GoodBDSaveFName);
		case BAttr:	return(GoodBASaveFName);
		default:	return(GoodSaveFName);
	}
}

Boolean	SaveAs()
{
	Str255	s;

	if (pSaveFName[0] == 0)
		GetUntitledFileName(CP(s));
	else
		sEquate(CP(s), pSaveFName);

	sEquate (CP(oSaveFName), pSaveFName);
	oSaveVRef = *pVRef;

	if ( (SaveWindex == BifD) || (SaveWindex == BAttr) ) {
		if ( !BBSpecialSaveAsDialog(SaveAsPrompt, pVRef, CP(s)) ) {
			if (pSaveFName[0] == 0) {
				if (SaveWindex == BifD)
					--BDUntitleds;
				else if (SaveWindex == BAttr)
					--BAUntitleds;
			}
			return(FALSE);
		}
	}
	else if ( !BBSaveAsDialog(SaveAsPrompt, pVRef, CP(s)) ) {
		if (pSaveFName[0] == 0) {
			if (SaveWindex == TSer)
				--TSUntitleds;
			else
				--Untitleds;
		}
		return(FALSE);
	}

	sEquate(pSaveFName, CP(s));

	if ( !StdCreateFile(pSaveFName, *pVRef, BBSFType, SaveType) ) {
		SetGoodSaveFName(FALSE);
		return(FALSE);
	}
	else
		SetGoodSaveFName(TRUE);

	return(TRUE);
}

Boolean	SaveFileExists()
{
	if (GetGoodSaveFName()) {
		RETURN_IF_FALSE( pSaveFName[0] > 0 );
		return( StdCreateFile(pSaveFName, *pVRef, BBSFType, SaveType) );
	}
	else
		return(FALSE);
}


		/*	----------	Routines Relating to defaults file	----------	*/

Boolean CreateDefaultsFile()
{
	OSErr	myErr;
	Boolean	ResForkExists;

	sEquate(pDefaultsFName, DefaultDefaultsFName);
	RETURN_IF_FALSE( StdCreateFile(pDefaultsFName, SysVRef, BBSFType,
			DefaultsSFType));
	return( OpenResFork(pDefaultsFName, SysVRef, TRUE, &ResForkExists) );
}

Boolean OpenDefaultsFile()
{
	Boolean	ResForkExists;

	RETURN_IF_FALSE( OpenResFork(pDefaultsFName, SysVRef, FALSE, &ResForkExists) );
	return(ResForkExists);
}

CloseDefaultsFile()
{
	if (!StopResFile()) ;
}

Boolean OpenDefaults(Boolean *pGotPrefs)
{
	int			i;

	*pGotPrefs = FALSE;
	RETURN_IF_FALSE(OpenDefaultsFile());

	hWLs = (hWindowLocs) GetResource(BBWindowsResType, DefaultWindowID);
	if (hWLs == NIL)
		return(FALSE);
	DefaultWindowNum = (**hWLs).num;
	for(i=0; i<DefaultWindowNum; ++i)
		wd[i] = (**hWLs).ws[i];
	hWLsSize = GetHandleSize((Handle) hWLs);

	ResrvMem(sizeof(BBPreferences));
	hBBPrefs = (hBBPreferences) GetResource(BBPrefsResType, DefaultPrefsID);
	if (hBBPrefs != NIL) {
		*pGotPrefs = TRUE;
		DetachResource(hBBPrefs);
		HLock((Handle) hBBPrefs);	/*	almost always locked	*/
	}

	CloseDefaultsFile();
	return(TRUE);
}

Boolean SaveDefaultFile()
{
	int			i;

	RETURN_IF_FALSE( CreateDefaultsFile() );
	hWLs = (hWindowLocs) NewHandle(WindowLocsSize(DefaultWindowNum));
	(**hWLs).num = DefaultWindowNum;
	for(i=0; i<DefaultWindowNum; ++i)
		(**hWLs).ws[i] = wd[i];
	RETURN_IF_FALSE( AddNewResource((Handle) hWLs, BBWindowsResType,
			DefaultWindowID) );
	hWLsSize = GetHandleSize((Handle) hWLs);

	HUnlock((Handle) hBBPrefs);
	RETURN_IF_FALSE( AddNewResource((Handle) hBBPrefs, BBPrefsResType,
			DefaultPrefsID) );
	RETURN_IF_FALSE(UpdateResFork());
	DetachResource(hBBPrefs);
	HLock((Handle) hBBPrefs);

	CloseDefaultsFile();
/* deleted 11/11/91--causing resource problems?
	if ( (GetHandleSize((Handle) hWLs) == hWLsSize) &&
			(hWLsSize != 0L) )
		DisposHandle((Handle) hWLs);
*/
	return(TRUE);
}


HaltBBFiles()
{
	;
}
