/*	bbGraphUtils.c
 *
 *	Utilities used mainly by bbGraphics, to handle
 *	Bouncing Ball's graphics requirements.
 *
 *	-TA
 */

#include <limits.h>
#include "bbPrototypes.h"
#include "bbGraphics.h"

#define	LabelFieldWidth		6
#define	LabelPrecision		(LabelFieldWidth-2)
#define	DashedLinePatsName	"\PDashed Line"
#define	MAX_INT_LONG		((long) INT_MAX)
#define	SqrtMaxPosInt		181		/* "sqrt" 32767 (really 181.01657...) */

	/*	From bbPhysics.c, elsewhere, used only for sine wave, xTableFixed	*/
/*
extern	double	Omega, delta;
*/
				/*	Global external vars from bb.c	*/
extern	pIterationVars		pCurrentVars;
extern	hBBPreferences		hBBPrefs;

	/*	External variables from bbGraphics.c */
extern	CursHandle	MagnifyingGlass;

int		MaxYRange, MaxTime;
double	traj_xScale, traj_yScale;
Fixed	xif_traj, vif_traj, tofff_traj, Fixed_G_Over2;

Fixed	A_fxd;
double	TwoPi;

double	ps_xScale, ps_yScale;
Fixed	xif_ps, vif_ps, Fixed_G_ps, MaxHalfVSquareDiff_f, dvdtf_ps, dtf_ps;
int		xint_ps, vint_ps, VStop, Last_xint_ps;
double	TStuckStop_ps, viFloat_ps;
Boolean	vif_TooBig;

			/*	To force ThePatterns to line up on even-byte boundary	*/
long	KludgeForPatterns[2*NumDashedLinePatterns];
Pattern	*ThePatterns = (Pattern *) &KludgeForPatterns;

Boolean	UseDashedLine;



WriteVar(char *s, double f)
{
	DrawString(s);
	DrawFloat(f, LabelFieldWidth, LabelPrecision, TRUE);
	DrawString("\P	");
}

Fract AccurateFracSin(double time)
{
	int		sign = 1;

	if (time >= TwoPi)	time = fmod(time, TwoPi);
	if (time > PI) 		{	sign = -1; time  = TwoPi - time;	}
	if (time > PI/2.0)	time = PI - time;
	return( sign*FracSin(X2Fix(&time)) );
}


	/* --------------------------------------------------------- */
	/* ------- Routines Relating to GraphFrom Iteration -------- */
	/* ------- (Fixed version of xBall from bbPhysics)	-------- */
	/* --------------------------------------------------------- */

InitFixedCoeffs(int YRange)
{
	MaxYRange = YRange;
	MaxTime = MAX_INT_LONG / MaxYRange;
}

SetFixed_G_Over2(double xScale, double yScale)
{
	double	tmpF = G * yScale / (2.0*xScale*xScale);
	Fixed_G_Over2 = X2Fix(&tmpF);
	traj_xScale = xScale;	traj_yScale = yScale;
}

InitFixedTraj(double Xi, double Vi, double Ti, int TStart)
{
	double tmpF;

	xif_traj = yFloatToFixed(Traj, Xi);
	tmpF = Vi*traj_yScale/traj_xScale;	vif_traj = X2Fix(&tmpF);

	tmpF = Ti*traj_xScale - TStart;		tofff_traj = X2Fix(&tmpF);
}

int	xBallFixed(int time)
	/* xBall from bbPhysics, to be used with Graphics only */
{
	Fixed	AverageVelocity, tf;

	tf = Long2Fix((long) time) - tofff_traj;
	if (time > MaxTime)
		return( yFloatToCoord( Traj,
				xBall(Fix2X(tf)/traj_xScale,
				Fix2X(xif_traj)/traj_yScale,
				Fix2X(vif_traj)/traj_yScale*traj_xScale) ) );

	AverageVelocity = vif_traj - FixMul(tf, Fixed_G_Over2);
	if (FixRound(AverageVelocity) > MaxYRange) return(MaxYRange);
	return ( FixRound( xif_traj + FixMul(tf, AverageVelocity) ) );
}

		/*	-----------------------------------------------	*/
		/*	---	Routines for fixed calcs for Trajectory	---	*/
		/*	---	   window, table's sine wave graphics 	---	*/
		/*	-----------------------------------------------	*/

InitFixedSineParams(double Ampl, double xScale, double yScale)
{
/*
	tmpF = 2.0*PI*delta;	TwoPiDelta_fxd = X2Fix(&tmpF);
	tmpF = Omega/xScale;	Omega_fxd = X2Fix(&tmpF);
*/
	TwoPi = 2.0*PI;
	A_fxd = yFloatToFixed(Traj, Ampl);
}

int	xTableFixed(int time)
{
	Fract	tmpFract;
	Fixed	tmpFixed;

/*
	tmpFract = FracSin(Omega_fxd*time + TwoPiDelta_fxd);
*/
	tmpFract = AccurateFracSin(Omega*xCoordToFloat(Traj, time) + TwoPi*delta);
	tmpFixed = Frac2Fix(tmpFract) + Long2Fix(1L);
	return(FixRound(FixMul(tmpFixed, A_fxd)));
}

		/*	-----------------------------------------------------------	*/
		/*	---	Routines for fixed calcs for Phase Space graphics	---	*/
		/*	-----------------------------------------------------------	*/

SetFixed_G_ps(double xScale, double yScale)
{
	double	tmpF = G * xScale * xScale / yScale;
	Fixed_G_ps = X2Fix(&tmpF);
	if ( Fixed_G_ps > Long2Fix(1L) )
		MaxHalfVSquareDiff_f = Long2Fix(MAX_INT_LONG);
	else
		MaxHalfVSquareDiff_f = MAX_INT_LONG * Fixed_G_ps;

	ps_xScale = xScale;
	ps_yScale = yScale;
}

int xBallFromV_Fixed(int vint)
{
	Fixed	HalfVSquareDiff;
	double	HVSD_float;
	Boolean	UseFixed = TRUE;
	long	vint_l = ABS(vint);

	if ( (vif_TooBig) || (vint_l > SqrtMaxPosInt) ) {
		HVSD_float = (viFloat_ps*viFloat_ps - vint_l*vint_l)/2.0;
		if (HVSD_float > MAX_INT_LONG)
			UseFixed = FALSE;
		else
			HalfVSquareDiff = X2Fix(&HVSD_float);
	}
	else
		HalfVSquareDiff = ( FixMul(vif_ps,vif_ps) -
				Long2Fix((long) vint*vint) )/2;

	if ( !UseFixed || (HalfVSquareDiff > MaxHalfVSquareDiff_f) )
		return(yFloatToCoord( PSpace,
				xBall((Fix2X(vif_ps)/ps_xScale-xCoordToFloat(PSpace,vint))/G,
				Fix2X(xif_ps)/ps_yScale, Fix2X(vif_ps)/ps_xScale) ) );
	return(FixRound( xif_ps + FixDiv(HalfVSquareDiff, Fixed_G_ps) ));
}

InitPSGraphing(double Xi, double Vi, int yint, double delta_t,
		Boolean *pDrawFromScratch)
{
	double tmpF;
	int		ovint_ps = vint_ps, oxint_ps = xint_ps;

	tmpF = Vi*ps_xScale;	vif_ps = X2Fix(&tmpF);	vint_ps = FixRound(vif_ps);
	viFloat_ps = tmpF;
	vif_TooBig = (tmpF > SqrtMaxPosInt);
	xif_ps = yFloatToFixed(PSpace, Xi);	xint_ps = FixRound(xif_ps);
	VStop = xFloatToCoord(PSpace, vBall(delta_t, Vi));
	Last_xint_ps = yFloatToCoord(PSpace, xBall(delta_t, Xi, Vi));

	dvdtf_ps = xFloatToFixed(PSpace, G*xCoordToFloat(Traj, 1));
	tmpF = xCoordToFloat(Traj, 1)*ps_yScale / ps_xScale;	dtf_ps = X2Fix(&tmpF);

	if (*pDrawFromScratch) {
		*pDrawFromScratch = FALSE;
		BothMoveTo(PSpace, vint_ps, xint_ps);
	}
	else {
		BothMoveTo(PSpace, ovint_ps, oxint_ps);
		BothLineTo(PSpace, vint_ps, xint_ps);
	}
}

GraphPhaseSpace(int delta_t, int xint, Boolean IsLastPoint)
/*	Warning: Does not graph identically to Trajectory window;
 *	Trajectory window takes sine wave values for end points,
 *	while this uses true end point values.
 *		-TA 12/3/90
 */
{
	int		CurVStop;
	Fixed	tf;

	if (IsLastPoint)
		CurVStop = VStop;
	else {
		tf = Long2Fix((long) delta_t) - tofff_traj;
		CurVStop = (int) Fix2Long( vif_ps - FixMul(tf, dvdtf_ps) );
	}
/*
	CurVStop = xFloatToCoord(PSpace, Vi - G * xCoordToFloat(delta_t);
*/

	if (FasterPhaseSpace && !IsLastPoint) {
		vint_ps = FixRound(vif_ps - FixMul(Fixed_G_ps, delta_t*dtf_ps));
		xint_ps = xint;
		BothLineTo(PSpace, vint_ps, xint_ps);
	}
	else for (;;) {
		--vint_ps;
		if (vint_ps <= VStop) {
			++vint_ps;
			break;
		}
		if (vint_ps < CurVStop) {
			++vint_ps;
			break;
		}
		xint_ps = xBallFromV_Fixed(vint_ps);
		BothLineTo(PSpace, vint_ps, xint_ps);
	}
	if (IsLastPoint) {
		vint_ps = VStop;
		xint_ps = Last_xint_ps;
		BothLineTo(PSpace, vint_ps, xint_ps);
	}
}

InitStuckPhaseSpace(double Ti, double Tj, double thePeriod)
{
	xint_ps = yFloatToCoord(PSpace, xTable(Ti));
	vint_ps = xFloatToCoord(PSpace, vTable(Ti));
	BothLineTo(PSpace, vint_ps, xint_ps);
	TStuckStop_ps = fmod(Tj, thePeriod) + thePeriod;
		/* we know Tj phase between 0 and PI/2	*/
}

GraphStuckPhaseSpace(double t_limit, int xint, double Ampl, double thePeriod,
		Boolean LastIteration)
{
	Boolean IncXInt;
	double	x, time, std_t_limit;
	int		oxint_ps;

	if (!WindowOpen(PhaseSpace)) return;

	if (FasterPhaseSpace) {
		xint_ps = xint;
		vint_ps = xFloatToCoord(PSpace, vTable(t_limit));
		BothLineTo(PSpace, vint_ps, xint_ps);
		return;
	}

	std_t_limit = fmod(t_limit, thePeriod);
	if (std_t_limit <= thePeriod/4.0)
		std_t_limit += thePeriod;

	for (;;) {
		oxint_ps = xint_ps;
		IncXInt = TRUE;
		if (vint_ps < 0)
			if (xint_ps > 0) {
				IncXInt = FALSE;
				--xint_ps;
			}
		if (IncXInt)
			++xint_ps;
		if ( (x=yCoordToFloat(PSpace, xint_ps)) > 2.0*Ampl ) {
			xint_ps = oxint_ps;
			break;
		}
		time = tFromXTable(x, IncXInt);
		if (time <= thePeriod/4.0)
			time += thePeriod;
		if ( (time>TStuckStop_ps) || (time>std_t_limit) ) {
			xint_ps = oxint_ps;
			break;
		}
			
		vint_ps = xFloatToCoord(PSpace, vTable(time));
		BothLineTo(PSpace, vint_ps, xint_ps);
	}
	if (LastIteration) {
		vint_ps = xFloatToCoord(PSpace, vTable(TStuckStop_ps));
		xint_ps = yFloatToCoord(PSpace, xTable(TStuckStop_ps));
		BothLineTo(PSpace, vint_ps, xint_ps);
	}
}


	/* ------------------------------------------------------ */
	/* -------    Routines Relating to Tracking in	 -------- */
	/* ------- IMap window, with selection rectangle -------- */
	/* ------------------------------------------------------ */

Boolean LoadDashedPatterns()
{
	Handle	PH;
	int		i;

	if ( (PH = GetNamedResource(PatternListResType, DashedLinePatsName)) == NIL ) {
		for(i = 0; i < NumDashedLinePatterns; ++i)
			BlockEquate(CP(ThePatterns[i]), CP(gray), sizeof(Pattern));
		return(UseDashedLine = FALSE);
	}
	HLock(PH);
	BlockEquate(CP(ThePatterns[0]), *PH+2, NumDashedLinePatterns*sizeof(Pattern));
	HUnlock(PH);
	ReleaseResource(PH);
	return(UseDashedLine = TRUE);
}

Boolean TrackDashedRect(int h, int v, Rect *p_DragRect, Rect *p_Outline,
		Boolean ForMagnification)
{
	long		l, PosLong, DashedLineWaitTicks = 7L;
	Point		oldpt, pt;
	PenState	PS;
	Boolean		MovedSome;
	Rect		tmpDR;

		/*	kludge because PinRect doesn't seem to let point equal
		 *	bottom or right of *p_DragRect
		 *		-TA 1/22/91
		 */
	tmpDR = *p_DragRect;
	++tmpDR.right;
	++tmpDR.bottom;

	GetPenState(&PS);
	PenMode(patXor);
	SetRect(p_Outline, h, v, h, v);
	oldpt.h = -1; oldpt.v = -1;
	WaitTicks(DashedLineWaitTicks);
	MovedSome = FALSE;
	for(l=-1L;StillDown();) {
		++l;
		GetMouse(&pt);
		if ( (pt.h==oldpt.h) && (pt.v==oldpt.v) )
			PenPat(&ThePatterns[l%4+8]);
		else {
			if (l>0) {
				PenPat(&ThePatterns[(l-1)%8]);
				FrameRect(p_Outline);
			}
			PenPat(&ThePatterns[l%8]);
			PosLong = PinRect(&tmpDR, pt);
			pt.h = LoWord(PosLong);	pt.v = HiWord(PosLong);
			if (pt.h>=h)	{	p_Outline->left = h;	p_Outline->right = pt.h;	}
			else			{	p_Outline->left = pt.h;	p_Outline->right = h;		}
			if (pt.v>=v)	{	p_Outline->top = v;		p_Outline->bottom = pt.v;	}
			else			{	p_Outline->top = pt.v;	p_Outline->bottom = v;		}

			if (!MovedSome)
				if ( (pt.h-h)*(pt.h-h) + (pt.v-v)*(pt.v-v) > 2 ) {
					MovedSome = TRUE;
					if (ForMagnification)
						SetCursor(*MagnifyingGlass);
				}
		}
		FrameRect(p_Outline);
		oldpt.h = pt.h; oldpt.v = pt.v;
		WaitTicks(DashedLineWaitTicks);
	}
	if (l>=0) {
		PenPat(&ThePatterns[l%8]);
		FrameRect(p_Outline);
	}
	PenMode(PS.pnMode);
	PenPat(&PS.pnPat);
	if (MovedSome && ForMagnification)
		SetCursor(&arrow);
	if (l>=0)	return(MovedSome);
	else		return(FALSE);
}
