/*
 *	bbGraphics.c
 *	Graphics routines for Bouncing Ball
 */

#include "bbPrototypes.h"
#include "bbGraphics.h"

				/*	Global external vars from bb.c	*/
extern	IterationVars		*pCurrentVars;
extern	hBBPreferences		hBBPrefs;	/*	for MinPhaseInt, InitGraphics only	*/
				/*	From bb.c, for "user waiting" (watch cursor, ticking?)	*/
extern	Boolean	UserWaiting;
extern	Boolean	InForeground;
				/*	From bbContGraphics.c, used for Animation copying	*/
extern	Rect	Ball, Table, Frame;
extern	int		AnimaHOffset;
				/*	From bb.c, for NewGraphics(TRUE, TRUE)	*/
extern	int		ProgramStatus;
				/*	From bbMenus.c, for NewGraphics(TRUE, )	*/
extern	Boolean	NeedToSave, NeedToSaveTS;

	Windows		*pw;

	CursHandle	watch, MagnifyingGlass;

	Boolean		*pTxOutWindowCleared, *pTSerWindowCleared;


StartWaiting()
{
	UserWaiting = TRUE;
	HLock((Handle) watch);
	if (InForeground)
		SetCursor(*watch);
}

StopWaiting()
{
	UserWaiting = FALSE;
	HUnlock((Handle) watch);
	if (InForeground)
		SetCursor(&arrow);
}

int	FindBBClick(int i, Point pt)
{
	Rect	tmpRect;	/* for IMap kludge	*/

	switch (i) {
		case TxOut:
		case TSer:
			if (PtInBodyViewRect(i, pt))
				return(i);
		case IMap:
		case IPhs:
			tmpRect = w[i].GR;
			++tmpRect.right;
			++tmpRect.bottom;
			if (PtInRect(pt, &tmpRect))
				return(i);
		case BifD:
		case BAttr:
			if (FindSpecialClick(i, pt))
				return(i);
	}
	return(NoWindow);
}


	/* ---------- Update routine ---------- */

HandleUpdate(WindowPtr wp, int myHitNumber)
{
	int		theRef;
	Rect	tmpRect;

	theRef = WindowIndex(wp);
	if (theRef == NoWindow) {
		SetPort(wp);
			/*	Next 2 lines are a kludge (for dialog boxes put in front of reg.
			 *	windows, when calling HandleAllUpdates()--
			 *	otherwise, would be in an infinite loop).
			 */
		BeginUpdate(wp);
		EndUpdate(wp);
		return;
	}

	SetPort(wp);
	BeginUpdate(wp);
		/* Clear window, draw grow icon, etc. */
	NoClippingRect(wp);
	EraseRect(&wp->portRect); DrawGrowIcon(wp);
	DrawControls(wp);
	StdDocWindowClipRect(wp);

	switch (theRef) {
		case Trajectory:		RefreshTraj();			break;
		case PhaseSpace:		RefreshPS();			break;
		case TextOutput:
		case TimeSeries:		UpdateText(theRef, A, freq, damp);
														break;
		case ImpactMap:			RefreshIM(myHitNumber);	break;
		case ImpactPhases:		RefreshIP(myHitNumber);	break;
		case Animation:			RefreshAnimation();		break;
		case BifDiagram:		RefreshBD();			break;
		case BasinsAttraction:	RefreshBA();			break;
	}
	EndUpdate(wp);
}

MakePicture(int windex, GrafPtr gp, Rect *pDstRect)
{
	GrafPtr		tmpgp;
	RgnHandle	MaskRgn = NewRgn();
	Point		pt = {0, 0};
	Rect		tmpRect;

	CopyRgn(GrayRgn, MaskRgn);
	GetPort(&tmpgp);
	SetPort((GrafPtr) w[windex].Peek);
	GlobalToLocal(&pt);
	OffsetRgn(MaskRgn, pt.h, pt.v);
	SetPort(tmpgp);

	if (!SectRect(pDstRect, &(**MaskRgn).rgnBBox, &tmpRect)) ;

/*
	ClipRect(pDstRect);
	CopyBits(&w[windex].Peek->port.portBits, &gp->portBits,
			pDstRect, pDstRect, srcCopy, 0L);
*/

	if ((windex == Animation) && (Table.top > 0)) {
		ClipRect(pDstRect);
		DrawAnimationFrame();
		DrawAnimation(TRUE);
	}
	else {
		ClipRect(&tmpRect);
		CopyBits(&w[windex].Peek->port.portBits, &gp->portBits,
				&tmpRect, &tmpRect, srcCopy, MaskRgn);
	}

	DisposeRgn(MaskRgn);
}

HandlePICTCopy(int windex)
{
	PicHandle	pic;
	long		theError;
	Rect		DstRect;
	RgnHandle	OldClipRgn;

	SetPort(w[windex].Peek);
	OldClipRgn = NewRgn();
	GetClip(OldClipRgn);

	if ((windex == Animation) && (Table.top > 0)) {
		DstRect = Table;
		DstRect.top = Ball.top;
		DstRect.bottom = Frame.bottom;
		DstRect.right += AnimaHOffset;
	}
	else
		GetStdDocWindowClipRect(&DstRect, (WindowPtr) w[windex].Peek);
	pic = OpenPicture(&DstRect);
	 MakePicture(windex, (WindowPtr) w[windex].Peek, &DstRect);
	ClosePicture();
	SetClip(OldClipRgn);
	DisposeRgn(OldClipRgn);

	if ( (theError = ZeroScrap()) != noErr)
		return;
	HLock(pic);
	theError = PutScrap((long) (**pic).picSize, 'PICT', (Ptr) *pic);
	HUnlock(pic);
	KillPicture(pic);
}

HandleCopy(int windex)
{
	switch (windex) {
		case NoWindow:
						return;
		case TextOutput:
		case TimeSeries:
						HandleTEXTCopy(windex);
						break;
		default:
						HandlePICTCopy(windex);
	}
}


	/* ---------- General graphics routines ---------- */

NewGraphics(Boolean ClearSolutions, Boolean NeedToResetFileName)
{
	NewBBWindows(ClearSolutions);
	if (ClearSolutions) {
		NewTextEdit();
		w[TxOut].WindowCleared = w[TSer].WindowCleared = TRUE;
		if (NeedToResetFileName) {
			if (ProgramStatus < /*was <=--don't know why*/ BIFURCATION_DIAGRAM)
				SetTextTitle(FALSE, TxOut);
			else
				SetANilFileName(TxOut);
		}
		SetANilFileName(TSer);
		NeedToSave = NeedToSaveTS = FALSE;
	}
	NewContGraphics();
	NewImpactGraphics();
}

Boolean InitGraphics()
{
	pTxOutWindowCleared = &w[TxOut].WindowCleared;
	pTSerWindowCleared = &w[TSer].WindowCleared;

	InitBBWindows();
	if (!InitBBOffPorts())
		return(FALSE);

	InitContGraphics();
	SetPhaseRangeGlobals(MinPhaseInt);

/*
	InitBBWindows();
	InitBBOffPorts(TABLE_WIDTH);
*/

	MakeSineWave();
	InitImpactGraphics();
	InitWholeScreenPSCoords();
	SetPSCoords();

	if (LoadDashedPatterns()) ;
	MagnifyingGlass = GetCursor(MagnifyingGlassCursorID);

	return(TRUE);
}

CreateWindowRecords()
{
	pw = (Windows *) NewPtr(sizeof(Windows));
}

HaltGraphics()
{
	HaltImpactGraphics();
	HaltBBWindows();
}