/*
 *	bbImpactGraphics.c
 *
 *	Graphics routines for Bouncing Balls impact windows,
 *	those which display information about impacts, as opposed to
 *	the continuous windows, those that in theory run continuously.
 *
 *	Routines for continuous windows are stored in bbContGraphics.c.
 *
 *	These windows are:
 *			Text Output	(most routines are in bbText.c)
 *			Impact Map
 *			Impact Phases
 *
 *		-TA 11/8/90
 */

#include "bbPrototypes.h"
#include "bbGraphics.h"
#include "bbMenus.h"

				/*	Global external vars from bb.c	*/
extern	pIterationVars		pCurrentVars;
extern	hBBPreferences		hBBPrefs;
								/*	From bb.c		*/
extern	Boolean				PrintingOnly;
								/*	From bbGraphics.c	*/
extern	Windows				*pw;
extern	CursHandle			watch;
								/*	From bbWindows.c; screen size	*/
extern	int					MaxScreenY;
								/*	From bbMenus.c; for magnification	*/
extern	MenuHandle			ViewMenuH;
								/*	From bbBifDiag.c, for AddGraphicSolution()	*/
extern	Boolean				*pBDGraphInAllWindows, *pBAGraphInAllWindows;


	double	AutoVMax;				/* VMax shared with Phase Space */
	Str255	VMaxString, NegVMaxString;
	double	VMin;
	Str255	VMinString;

	Rect	LatestImpactRect, IPLatestImpactRect;
	Rect	IMapBox, IPhsBox;
	int		StartHitNumber;	/* first hit to show in impact map ... */

	double	MinPhase, MaxPhase;		/* min, max phases to display on impact map ... */
	Str255	MinPhaseString, MaxPhaseString;
	long	CurrentIPOffsetCount;

	DoubleRect	FullIMapRange, NewIMapRange;
	DoubleRect	FullIPhsRange, NewIPhsRange;
	BooleanRect	NewUsesFullIMap = {TRUE, TRUE, TRUE, TRUE},
				UsesFullIMap = {TRUE, TRUE, TRUE, TRUE},
				NewUsesFullIPhs = {TRUE, TRUE, TRUE, TRUE},
				UsesFullIPhs = {TRUE, TRUE, TRUE, TRUE};

	int			IMXFieldWidth, IMYFieldWidth, IPXFieldWidth, IPYFieldWidth;

			/*	-------------------------------------------------	*/
			/*	---------- Window-Independent Routines ----------	*/
			/*	-------------------------------------------------	*/

/* For Sound */
double GetVelRatio(double theVelocity)
{
	if (theVelocity > VMax) theVelocity = VMax;
	else if (theVelocity < VMin) theVelocity = VMin;
	return( (theVelocity - VMin)/(VMax - VMin) );
}

SetVMax(Boolean RecalcAuto)
/*	Call whenever A, omega, damp, or ScreenNumAmplitudes change.
 *	Used by Impact Map, Phase Space, and 3-D Phase Space.
 */
{
	double	tVMax, nVMax;
	double	oVMax = VMax;

	if (RecalcAuto) {
		/* Calculate tVmax (fit on screen) = sqrt(2*g*A*ScreenNumAmplitudes)
		 * (Tyler's to-fit-with-NumAmplitudes algorithm), then
		 * nVmax = A omega * (1+damp)/(1-damp) ( * 2 omega/g dimensionless(--so?) )
		 * (Nick's physical "trapping" algorithm), then
		 * if nVMax > tVMax, VMax = nVMax
		 * Vmin = -A*omega
		 * (Tyler's is computer-related and user-changeable, Nick's is physical)
		 */
		tVMax = sqrt(2.0*G*NON_ZERO_A*ScreenNumAmplitudes);
		if (damp < 1)
			nVMax = A*Omega * (1.0+damp) / (1.0-damp);
		else nVMax = tVMax;
		AutoVMax = MAX(nVMax, tVMax);
	}
	if (UseAutoVMax)
		VMax = AutoVMax;

	VMin = -A*Omega;
	myFToStr(VMax, CP(VMaxString),
			StdIMFieldWidth, StdIMFieldWidth-IMPrecDiff);
	myFToStr(VMin, CP(VMinString),
			StdIMFieldWidth, StdIMFieldWidth-IMPrecDiff);
	sEquate(CP(NegVMaxString), CP(VMaxString));
	sInsertC(CP(NegVMaxString), '-', 1);

	w[IMap].DrawFromScratch = TRUE;
	InitFullIMapRange();
	if (w[IMap].Peek != NIL)
/*
		if ( (oVMax != VMax) || (!w[IMap].WindowCleared) )
*/
			StdForceUpdate((WindowPtr) w[IMap].Peek);
	/*	Reinitialize PSpace coords, clear so far	*/
	InitPSScales();
	if (w[PSpace].Back != NOPORT)
		ClearPhaseSpace();
}

GetNewVMax()
{
	double	oVMax = VMax;

	if (!DoSetVMaxDialog(&VMax, &UseAutoVMax))	return;
	if (VMax == oVMax)	return;

	SetVMax(FALSE);
	InitImpactMapScales();
}

SetImpactMapStart(int theHitNumber)
{
	StartHitNumber = theHitNumber;
	w[IMap].DrawFromScratch = TRUE;
	StdForceUpdate((WindowPtr) w[IMap].Peek);
	w[IPhs].DrawFromScratch = TRUE;
	StdForceUpdate((WindowPtr) w[IPhs].Peek);
}

			/*	Routines involved in setting different phase ranges	*/
double GetOffsetPhase(double del)
{
/*
	if (MinPhase != 0.0) {
*/
		del -= MinPhase;
		if (del>1.0+AlmostZero)		/*	used to be >=	*/
			del -= 1.0;
		if (del<0.0)
			del += 1.0;
/*
	}
*/
	return(del);
}

double GetUnOffsetPhase(double del)
{
	if (MinPhase != 0.0)
		del += MinPhase;
	return(del);
}

SetPhaseRangeGlobals(int theMinPhaseInt)
{
	int		i;
	double	tmpMinX, tmpMaxX;
	double	tmpMinY, tmpMaxY;
	Boolean	NeedToInitIP;

	NewIMapRange[xmin_i] = GetUnOffsetPhase(NewIMapRange[xmin_i]);
	NewIMapRange[xmax_i] = GetUnOffsetPhase(NewIMapRange[xmax_i]);
	for (i=ymin_i; i<ItemsInRect; ++i)
		NewIPhsRange[i] = GetUnOffsetPhase(NewIPhsRange[i]);

	MinPhaseInt = theMinPhaseInt;
	MinPhase = ((double) MinPhaseInt)/PHASE_INT_RANGE;
	MaxPhase = MinPhase + 1.0;
	SetPhaseString(MinPhaseInt, CP(MinPhaseString));
	SetPhaseString(MinPhaseInt+100, CP(MaxPhaseString));

	if ((w[IMap].Magnifying) && (w[IMap].GraphRange[xmax_i] -
			w[IMap].GraphRange[xmin_i] < 1.0 - AlmostZero) ) {
		tmpMinX = GetOffsetPhase(NewIMapRange[xmin_i]);
		tmpMaxX = GetOffsetPhase(NewIMapRange[xmax_i]);
		w[IMap].GraphRange[xmin_i] = NewIMapRange[xmin_i] = MIN(tmpMinX, tmpMaxX);
		w[IMap].GraphRange[xmax_i] = NewIMapRange[xmax_i] = MAX(tmpMinX, tmpMaxX);
		UsesFullIMap[xmin_i] = NewUsesFullIMap[xmin_i] =
				(w[IMap].GraphRange[xmin_i] < 0.0 + AlmostZero);
		UsesFullIMap[xmax_i] = NewUsesFullIMap[xmax_i] =
				(w[IMap].GraphRange[xmax_i] > 1.0 - AlmostZero);
		w[IMap].xGraphAroundZero = (w[IMap].GraphRange[xmin_i] <= 0.0);
		if (tmpMinX > tmpMaxX)
			InitImpactMapScales();
	}
	else {
		w[IMap].GraphRange[xmin_i] = NewIMapRange[xmin_i] = 0.0;
		w[IMap].GraphRange[xmax_i] = NewIMapRange[xmax_i] = 1.0;
	}

	NeedToInitIP = FALSE;
	if ((w[IPhs].Magnifying) && (w[IPhs].GraphRange[xmax_i] -
			w[IPhs].GraphRange[xmin_i] < 1.0 - AlmostZero) ) {
		tmpMinX = GetOffsetPhase(NewIPhsRange[xmin_i]);
		tmpMaxX = GetOffsetPhase(NewIPhsRange[xmax_i]);
		w[IPhs].GraphRange[xmin_i] = NewIPhsRange[xmin_i] = MIN(tmpMinX, tmpMaxX);
		w[IPhs].GraphRange[xmax_i] = NewIPhsRange[xmax_i] = MAX(tmpMinX, tmpMaxX);
		UsesFullIPhs[xmin_i] = NewUsesFullIPhs[xmin_i] =
				(w[IPhs].GraphRange[xmin_i] < 0.0 + AlmostZero);
		UsesFullIPhs[xmax_i] = NewUsesFullIPhs[xmax_i] =
				(w[IPhs].GraphRange[xmax_i] > 1.0 - AlmostZero);
		w[IPhs].xGraphAroundZero = (w[IPhs].GraphRange[xmin_i] <= 0.0);
		if (tmpMinX > tmpMaxX)
			NeedToInitIP = TRUE;
	}
	else {
		w[IPhs].GraphRange[xmin_i] = NewIPhsRange[xmin_i] = 0.0;
		w[IPhs].GraphRange[xmax_i] = NewIPhsRange[xmax_i] = 1.0;
	}
	if ((w[IPhs].Magnifying) && (w[IPhs].GraphRange[ymax_i] -
			w[IPhs].GraphRange[ymin_i] < 1.0 - AlmostZero) ) {
		tmpMinY = GetOffsetPhase(NewIPhsRange[ymin_i]);
		tmpMaxY = GetOffsetPhase(NewIPhsRange[ymax_i]);
		w[IPhs].GraphRange[ymin_i] = NewIPhsRange[ymin_i] = MIN(tmpMinY, tmpMaxY);
		w[IPhs].GraphRange[ymax_i] = NewIPhsRange[ymax_i] = MAX(tmpMinY, tmpMaxY);
		UsesFullIPhs[ymin_i] = NewUsesFullIPhs[ymin_i] =
				(w[IPhs].GraphRange[ymin_i] < 0.0 + AlmostZero);
		UsesFullIPhs[ymax_i] = NewUsesFullIPhs[ymax_i] =
				(w[IPhs].GraphRange[ymax_i] > 1.0 - AlmostZero);
		w[IPhs].yGraphAroundZero = (w[IPhs].GraphRange[ymin_i] <= 0.0);
		if (tmpMinX > tmpMaxX)
			NeedToInitIP = TRUE;
	}
	else {
		w[IPhs].GraphRange[ymin_i] = NewIPhsRange[ymin_i] = 0.0;
		w[IPhs].GraphRange[ymax_i] = NewIPhsRange[ymax_i] = 1.0;
	}
	if (NeedToInitIP)
		InitImpactPhaseScales();

/*
	w[IMap].GraphRange[xmin_i] = NewIMapRange[xmin_i] =
			GetOffsetPhase(NewIMapRange[xmin_i]);
	w[IMap].GraphRange[xmax_i] = NewIMapRange[xmax_i] =
			GetOffsetPhase(NewIMapRange[xmax_i]);
	for (i=ymin_i; i<ItemsInRect; ++i)
		w[IPhs].GraphRange[i] = NewIPhsRange[i] =
				GetOffsetPhase(NewIPhsRange[i]);
*/
}

GetPhaseRange()
{
	int	NewMinPhaseInt = MinPhaseInt;

	if (!DoPhaseRangeDialog(&NewMinPhaseInt, 0))	return;
	if (NewMinPhaseInt == MinPhaseInt)				return;

	SetPhaseRangeGlobals(NewMinPhaseInt);
	w[IMap].DrawFromScratch = TRUE;
	StdForceUpdate((WindowPtr) w[IMap].Peek);
	w[IPhs].DrawFromScratch = TRUE;
	StdForceUpdate((WindowPtr) w[IPhs].Peek);
}


			/*	-----------------------------------------	*/
			/*	---------- Impact Map Routines ----------	*/
			/*	-----------------------------------------	*/

		/*	Routines related to impact map magnification	*/
/*
 *	NOTE: FullIMapRange phases are actual, those of w[IMap].GraphRange
 *	and NewIMapRange (except when in dialog) offset.
 *	Same below for Impact Phases graph ranges (not implemented yet).
 *		-TA 1/24/91
 */

InitFullIMapRange()
{
	int	i;

	FullIMapRange[ymin_i] = VMin;
	FullIMapRange[xmin_i] = MinPhase;
	FullIMapRange[ymax_i] = VMax;
	FullIMapRange[xmax_i] = MaxPhase;

			/*	x-axis (Phase)	*/
	for (i=xmin_i; i<ItemsInRect; i += 2) {
		if (UsesFullIMap[i])
			w[IMap].GraphRange[i] = GetOffsetPhase(FullIMapRange[i]);
		if (NewUsesFullIMap[i])
				NewIMapRange[i] = GetOffsetPhase(FullIMapRange[i]);
	}

		/*	y-axis (Velocity) (this done on both axes in quadratic map)	*/
	i = ymin_i;
	if (UsesFullIMap[i]) {
		if (UsesFullIMap[i+2] || (w[IMap].GraphRange[i+2] > FullIMapRange[i]))
			w[IMap].GraphRange[i] = FullIMapRange[i];
		else
			UsesFullIMap[i] = FALSE;
	}
	if (NewUsesFullIMap[i]) {
		if (NewUsesFullIMap[i+2] || (NewIMapRange[i+2] > FullIMapRange[i]))
			NewIMapRange[i] = FullIMapRange[i];
		else
			NewUsesFullIMap[i] = FALSE;
	}

	i = ymax_i;
	if (UsesFullIMap[i]) {
		if (UsesFullIMap[i-2] || (w[IMap].GraphRange[i-2] < FullIMapRange[i]))
			w[IMap].GraphRange[i] = FullIMapRange[i];
		else
			UsesFullIMap[i] = FALSE;
	}
	if (NewUsesFullIMap[i]) {
		if (NewUsesFullIMap[i-2]  || (NewIMapRange[i-2] < FullIMapRange[i]))
			NewIMapRange[i] = FullIMapRange[i];
		else
			NewUsesFullIMap[i] = FALSE;
	}

/*
	for (i=ymin_i; i<ItemsInRect; ++i) {
		if (UsesFullIMap[i]) {
			if ( (i==xmin_i) || (i==xmax_i) )
				w[IMap].GraphRange[i] = GetOffsetPhase(FullIMapRange[i]);
			else
				w[IMap].GraphRange[i] = FullIMapRange[i];
		}
		if (NewUsesFullIMap[i]) {
			if ( (i==xmin_i) || (i==xmax_i) )
				NewIMapRange[i] = GetOffsetPhase(FullIMapRange[i]);
			else
				NewIMapRange[i] = FullIMapRange[i];
		}
	}
*/
}

SetFullIMapRange()
/*	Never use--was used in "Use Full Impact Map" menu command, BB 1.5a	*/
{
	int	i;

	InitFullIMapRange();

	for (i=ymin_i; i<ItemsInRect; ++i) {
		NewUsesFullIMap[i] = UsesFullIMap[i] = TRUE;
		if ( (i==xmin_i) || (i==xmax_i) )
			w[IMap].GraphRange[i] = NewIMapRange[i] =
					GetOffsetPhase(FullIMapRange[i]);
		else
			w[IMap].GraphRange[i] = NewIMapRange[i] = FullIMapRange[i];
	}
	w[IMap].Magnifying = FALSE;
	w[IMap].xGraphAroundZero = w[IMap].yGraphAroundZero = TRUE;
	InitImpactMapScales();
}

GetIMMagnification(Boolean SkipDialog)
{
	int			i;
	double		min, max;
	Boolean		GoodOp;

	InitFullIMapRange();
	NewIMapRange[xmin_i] = GetUnOffsetPhase(NewIMapRange[xmin_i]);
	NewIMapRange[xmax_i] = GetUnOffsetPhase(NewIMapRange[xmax_i]);

		/*	Round to RelPlaces (2) relative decimal places	*/
	Round2Doubles(NewIMapRange[xmin_i], NewIMapRange[xmax_i], RelPlaces, &min, &max);
	if (NewIMapRange[xmin_i] > FullIMapRange[xmin_i])
		NewIMapRange[xmin_i] = min;
	if (NewIMapRange[xmax_i] < FullIMapRange[xmax_i])
		NewIMapRange[xmax_i] = max;
	Round2Doubles(NewIMapRange[ymin_i], NewIMapRange[ymax_i], RelPlaces, &min, &max);
	if (NewIMapRange[ymin_i] > FullIMapRange[ymin_i])
		NewIMapRange[ymin_i] = min;
	if (NewIMapRange[ymax_i] < FullIMapRange[ymax_i])
		NewIMapRange[ymax_i] = max;

/*	BB 2.01 change
	if (!SkipDialog)
*/
	if ( (!SkipDialog) || (NewIMapRange[xmin_i] == NewIMapRange[xmax_i]) ||
			(NewIMapRange[ymin_i] == NewIMapRange[ymax_i]) )
		GoodOp = DoMagnifyImpactMapDialog(&NewIMapRange, &FullIMapRange,
				&NewUsesFullIMap);
	else
		GoodOp = TRUE;

	if (GoodOp) {
		NewIMapRange[xmin_i] = GetOffsetPhase(NewIMapRange[xmin_i]);
		NewIMapRange[xmax_i] = GetOffsetPhase(NewIMapRange[xmax_i]);
		for (i=ymin_i; i<ItemsInRect; ++i)
			UsesFullIMap[i] = NewUsesFullIMap[i];

		if (	(!UsesFullIMap[ymin_i]) || (!UsesFullIMap[xmin_i]) ||
				(!UsesFullIMap[ymax_i]) || (!UsesFullIMap[xmax_i]) ) {
			w[IMap].Magnifying = TRUE;
/* BB 2.01 deletion, 5/24/92 -- moved out of this conditional statement
			w[IMap].xGraphAroundZero =
				(NewIMapRange[xmin_i]*NewIMapRange[xmax_i])<=0;
			w[IMap].yGraphAroundZero =
				(NewIMapRange[ymin_i]*NewIMapRange[ymax_i])<=0;
*/
		}
/*	BB 2.01 addition, 5/24/92
 */
		else
			w[IMap].Magnifying = FALSE;
		w[IMap].xGraphAroundZero =
			(NewIMapRange[xmin_i]*NewIMapRange[xmax_i])<=0;
		w[IMap].yGraphAroundZero =
			(NewIMapRange[ymin_i]*NewIMapRange[ymax_i])<=0;
/*	end of change
 */
		if (	(w[IMap].GraphRange[ymin_i] != NewIMapRange[ymin_i]) ||
				(w[IMap].GraphRange[xmin_i] != NewIMapRange[xmin_i]) ||
				(w[IMap].GraphRange[ymax_i] != NewIMapRange[ymax_i]) ||
				(w[IMap].GraphRange[xmax_i] != NewIMapRange[xmax_i]) ) {
			for (i=ymin_i; i<ItemsInRect; ++i)
				w[IMap].GraphRange[i] = NewIMapRange[i];
			InitImpactMapScales();
		}
	}
	else
		for (i=ymin_i; i<ItemsInRect; ++i) {
			NewIMapRange[i] = w[IMap].GraphRange[i];
			NewUsesFullIMap[i] = UsesFullIMap[i];
		}
}

Boolean	TrackIMapRect(Point pt, double *pv0, double *pdelta)
/* Gets initial conditions or other from Impact Map mouse clicks */
{
	int	*pi1, *pi2, *pi3, i, index;

	if ( TrackDashedRect(pt.h, pt.v, &w[IMap].GR, &IMapBox, TRUE) ) {
		pi1 = (int *) &IMapBox;
		pi2 = (int *) &w[IMap].GR;
		pi3 = (int *) &w[IMap].Origin;
		for (i=ymin_i; i<ItemsInRect; ++i) {
			if (i%2==0)	index = (i+2)%4;
			else		index = i;
			if ( (pi1[i] == pi2[i]) && UsesFullIMap[index] ) {
				NewUsesFullIMap[index] = TRUE;
				NewIMapRange[index] = FullIMapRange[index];
			}
			else {
				NewUsesFullIMap[index] = FALSE;
				if (i%2==0)
					NewIMapRange[index] =
						yCoordToFloat(IMap, pi3[i%2] - pi1[i]);
				else
					NewIMapRange[index] =
						xCoordToFloat(IMap, pi1[i] - pi3[i%2]);
			}
		}
		return(TRUE);
	}
	*pv0 = yCoordToFloat(IMap, w[IMap].Origin.v - pt.v);
	*pdelta = GetUnOffsetPhase(xCoordToFloat(IMap, pt.h - w[IMap].Origin.h));
	return(FALSE);
}

InitImpactMapScales()
{
	Rect	tmpRect;
	double	range;
	int		i;

	SetPort(w[IMap].Peek);
	w[IMap].DrawFromScratch = TRUE;
	StdForceUpdate((GrafPtr) w[IMap].Peek);

		/*	Calculate how much space needed for axis labels	*/
	if (w[IMap].Magnifying) {
		range = w[IMap].GraphRange[ymax_i]-w[IMap].GraphRange[ymin_i];
		for(i=-1;i<2;++i) {
			if (range > 1.0) break;
			range *= 10;
		}
		IMYFieldWidth = StdIMFieldWidth + MAX(i,0);

		range = w[IMap].GraphRange[xmax_i]-w[IMap].GraphRange[xmin_i];
		for(i=-2;i<4;++i) {
			if (range > 1.0) break;
			range *= 10;
		}
		IMXFieldWidth = StdIMFieldWidth + MAX(i,0);
	}
	else
		IMYFieldWidth = IMXFieldWidth = StdIMFieldWidth;

	tmpRect = w[IMap].Peek->port.portRect;
	tmpRect.left += (StdMargin + (IMYFieldWidth+3)*CharWidth('0'));
	tmpRect.top	+= IM_TopMargin;
	tmpRect.right -= (ScrollBarWidth + StdMargin +
			IMXFieldWidth * CharWidth('0') / 2);
	tmpRect.bottom -= (ScrollBarWidth + StdMargin + TextLineHeight + TextHeight);
	w[IMap].GR = tmpRect;

	range = w[IMap].GraphRange[xmax_i] - w[IMap].GraphRange[xmin_i];
	if (range == 0.0) range = 1.0;
	w[IMap].xScale = (double) (tmpRect.right - tmpRect.left) / range;
				/*	Denominator (range) 1.0 if no magnification	*/

	w[IMap].yScale = (tmpRect.bottom - tmpRect.top) / /* (VMax-VMin) */
			(w[IMap].GraphRange[ymax_i] - w[IMap].GraphRange[ymin_i]);
	if (w[IMap].yGraphAroundZero)
		w[IMap].Origin.v = w[IMap].GR.top +
				RoundToInt(w[IMap].GraphRange[ymax_i]*w[IMap].yScale);
	else
		w[IMap].Origin.v = w[IMap].GR.bottom;
	w[IMap].Origin.h = w[IMap].GR.left;
	w[IMap].BackOrigin = w[IMap].Origin;
	SetRect(&w[IMap].PR,
			w[IMap].GR.left - w[IMap].Origin.h,
			w[IMap].Origin.v - w[IMap].GR.bottom,
			w[IMap].GR.right - w[IMap].Origin.h + 1,
			w[IMap].Origin.v - w[IMap].GR.top + 1);
			/* +1 to account for pixel hanging below and to left of point */

/*
	tmpRect = w[IMap].GR;
*/
	tmpRect.top = tmpRect.bottom+1;
	tmpRect.bottom = w[IMap].Peek->port.portRect.bottom - (ScrollBarWidth-1);
	InsetRect(&tmpRect, -1, 0);
	ResizeControl(w[IMap].HS[0], &tmpRect);

	tmpRect = w[IMap].GR;
	tmpRect.right = tmpRect.left-1;
	tmpRect.left = 0;
	InsetRect(&tmpRect, 0, -1);
/*
	tmpRect.left -= MAX(StringWidth(VMaxString), StringWidth(VMinString))+3;
	tmpRect.top -= TextHeight/2;
	tmpRect.bottom += TextHeight/2+1;
*/
	ResizeControl(w[IMap].HS[1], &tmpRect);
}

PlotInImpactMap(double del, double vel, Boolean Highlighting)
{
	Point	pt;

	w[IMap].WindowCleared = FALSE;

	SetPort(w[IMap].Peek);
	if (Highlighting)
		if (LatestImpactRect.left != 0) {
			 InvertOval(&LatestImpactRect);
			SetPort(w[IMap].Back);
			 InvertOval(&LatestImpactRect);
			SetPort(w[IMap].Peek);
		}
	pt.h = xFloatToCoord(IMap, GetOffsetPhase(del));
	pt.v = yFloatToCoord(IMap, vel);
			/*	Next statement: if rounded would be MaxPhase, make MinPhase;
			 *	if delete this statement, enlarge drawn rect surrounding
			 *	GR by one pixel on the right.
			 */
	if (pt.h == /* was >= */ w[IMap].GR.right-w[IMap].Origin.h)
		if (	(!w[IMap].Magnifying) ||
				( (w[IMap].GraphRange[xmin_i] == MinPhase) &&
				(w[IMap].GraphRange[xmax_i] == MaxPhase) ) )
			pt.h = w[IMap].GR.left-w[IMap].Origin.h;
	if ( !PtInRect(pt, &w[IMap].PR) ) {
		if (Highlighting)
			OffsetRect(&LatestImpactRect,-LatestImpactRect.left,
					-LatestImpactRect.top);
		return;
	}
	BothMoveTo(IMap, pt.h, pt.v);
	SetPort(w[IMap].Peek);	Line(0, 0);
	SetPort(w[IMap].Back);	Line(0, 0);
	SetPort(w[IMap].Peek);
	if (Highlighting) {
			/* next statement from WindowMoveTo */
		OffsetRect(&LatestImpactRect, -LatestImpactRect.left + pt.h +
			w[IMap].Origin.h - 2,
			-LatestImpactRect.top + w[IMap].Origin.v - pt.v - 2);
		 InvertOval(&LatestImpactRect);
		SetPort(w[IMap].Back);
		 InvertOval(&LatestImpactRect);
		SetPort(w[IMap].Peek);
	}
}

RefreshIM(int myHitNumber)
{
	int		i, x;
	char	*s;
	Rect	tmpRect;
	Str255	LocalVMinString, LocalVMaxString;
	Str255	LocalMinPhaseString, LocalMaxPhaseString;

	int		OverflowOnRight, MaxPStrLen;	/* intermediate vars	*/

	if (!w[IMap].DrawFromScratch) {
		SetPort(w[IMap].Peek);
		myCopyBits(w[IMap].Back, (GrafPtr) w[IMap].Peek,
			&w[IMap].Peek->port.portRect, 0,0,0,0, srcCopy, (RgnHandle) NIL);
		return;
	}

	w[IMap].WindowCleared = TRUE;

	w[IMap].DrawFromScratch = FALSE;
	SetPort(w[IMap].Back);
	EraseRect(&w[IMap].Back->portRect);

	tmpRect = w[IMap].GR;
	InsetRect(&tmpRect, -1, -1);
	/*	Use this if don't adjust near-max phases (see PlotInImpactMap()) 
		++tmpRect.right;
	*/
	FrameRect(&tmpRect);

	/*	Get endpoint strings; increase precision if magnification dictates */
	if (w[IMap].Magnifying) {
		myFToStrNoPad(w[IMap].GraphRange[ymin_i], CP(LocalVMinString),
				IMYFieldWidth, IMYFieldWidth - IMPrecDiff);
		myFToStrNoPad(w[IMap].GraphRange[ymax_i], CP(LocalVMaxString),
				IMYFieldWidth, IMYFieldWidth - IMPrecDiff);
		myFToStrNoPad(GetUnOffsetPhase(w[IMap].GraphRange[xmin_i]),
				CP(LocalMinPhaseString),
				IMXFieldWidth, IMXFieldWidth - IMPrecDiff);
		myFToStrNoPad(GetUnOffsetPhase(w[IMap].GraphRange[xmax_i]),
				CP(LocalMaxPhaseString),
				IMXFieldWidth, IMXFieldWidth - IMPrecDiff);
	}
	else {
		sEquate(CP(LocalVMinString), CP(VMinString));
		sEquate(CP(LocalVMaxString), CP(VMaxString));
		sEquate(CP(LocalMinPhaseString), CP(MinPhaseString));
		sEquate(CP(LocalMaxPhaseString), CP(MaxPhaseString));
	}

		/* Draw Horizontal Hash Marks */
	for(i=0;i<4;++i) {
		x = xFloatToCoord(IMap, GetOffsetPhase(i/4.0));
		if ( (x<=0) || (x>=w[IMap].GR.right-w[IMap].GR.left-1) )
			continue;
		WindowMoveTo(IMap, x, w[IMap].Origin.v-w[IMap].GR.bottom+HashSize+1);
/*
		MoveTo( w[IMap].GR.left + i * (w[IMap].GR.right-w[IMap].GR.left)/4,
								w[IMap].GR.bottom - HashSize );
*/
		Line(0, HashSize);
		Move(0, -(w[IMap].GR.bottom - w[IMap].GR.top - 1) );
		Line(0, HashSize);
		if (i==0)
			if ( (x > StringWidth(CP(LocalMinPhaseString))/2+4) &&
					(x < w[IMap].GR.right - w[IMap].GR.left -
					StringWidth(CP(LocalMaxPhaseString))/2-4) ) {
				WindowMoveTo(IMap, x, w[IMap].Origin.v-w[IMap].GR.bottom);
				Move(-CharWidth('0')/2+1, TextMargin + TextHeight-1);
				DrawChar('0');
			}
	}

		/* Label velocity (vertical) axis */
/*	Not enough that around zero--endpoint must not be zero
	if (w[IMap].yGraphAroundZero) {
*/
	if (w[IMap].GraphRange[ymin_i] * w[IMap].GraphRange[ymax_i] < 0) {
		WindowMoveTo(IMap, 0, 0);
		Line(w[IMap].GR.right - w[IMap].GR.left, 0);
	}
	WindowMoveTo(IMap, 0, w[IMap].Origin.v - w[IMap].GR.top);
		Move(-TextMargin - StringWidth(&LocalVMaxString), TextHeight/2);
		DrawString(&LocalVMaxString);
	WindowMoveTo(IMap, 0, w[IMap].Origin.v - w[IMap].GR.bottom);
		Move(-TextMargin - StringWidth(&LocalVMinString), TextHeight/2-1);
		DrawString(&LocalVMinString);
	if (w[IMap].yGraphAroundZero)
		if ( (w[IMap].Origin.v - w[IMap].GR.top > TextHeight) &&
				(w[IMap].GR.bottom - w[IMap].Origin.v > TextHeight) ) {
			WindowMoveTo(IMap, 0, 0);
				Move(-TextMargin - CharWidth('0'), TextHeight/2);
				DrawChar('0');
		}

	s = "\PVelocity";
	SetRect(&tmpRect, StdMargin+2,
			(w[IMap].GR.top+w[IMap].GR.bottom-((int)s[0])*TextLineHeight)/2,
			StdMargin + 12/*CharWidth('V') + 1*/, MaxScreenY);
	tmpRect.top = MAX(tmpRect.top, w[IMap].GR.top);
	TextBox(&(s[1]), (long) (s[0]), &tmpRect, teJustCenter);

		/* Label phase (horizontal) axis */
	MoveTo(w[IMap].GR.left, w[IMap].GR.bottom);
	Move(-StringWidth(CP(LocalMinPhaseString))/2, TextMargin + TextHeight-1);
	DrawString(CP(LocalMinPhaseString));
	MoveTo(w[IMap].GR.right, w[IMap].GR.bottom);
	MaxPStrLen = StringWidth(CP(LocalMaxPhaseString));
	OverflowOnRight = MAX(MaxPStrLen/2 - ImpactRightMargin, 0);
	Move(-MaxPStrLen/2 - OverflowOnRight, TextMargin + TextHeight-1);
	DrawString(CP(LocalMaxPhaseString));
/*
	if (w[IMap].xGraphAroundZero) {		(* does not need to be TRUE	*)
		x=xFloatToCoord(IMap, GetOffsetPhase(0.0));
		if ( (x > StringWidth(CP(LocalMinPhaseString))/2+4) &&
				(x < w[IMap].GR.right - w[IMap].GR.left -
				StringWidth(CP(LocalMaxPhaseString))/2-4) ) {
			WindowMoveTo(IMap, x, w[IMap].Origin.v-w[IMap].GR.bottom);
			Move(-CharWidth('0')/2+1, TextMargin + TextHeight-1);
			DrawChar('0');
		}
	}
*/

	s = "\PPhase";
	MoveTo( (w[IMap].GR.right + w[IMap].GR.left)/2, w[IMap].GR.bottom);
	Move(-StringWidth(s)/2, TextMargin + TextLineHeight + TextHeight - 3);
	DrawString(s);

	SetPort(w[IMap].Peek);
	myCopyBits(w[IMap].Back, (GrafPtr) w[IMap].Peek,
			&w[IMap].Peek->port.portRect, 0,0,0,0, srcCopy, (RgnHandle) NIL);

	OffsetRect(&LatestImpactRect,-LatestImpactRect.left,-LatestImpactRect.top);
	if (myHitNumber >= StartHitNumber) {
		if (myHitNumber - StartHitNumber > MANY_IMPACTS)
			StartWaiting();
		for (i=StartHitNumber;i<=myHitNumber-1;++i)
			PlotInImpactMap(GetSolutionInfo(i, GET_PHASE),
					GetSolutionInfo(i, GET_VELOCITY), FALSE);
		if (myHitNumber - StartHitNumber > MANY_IMPACTS)
			StopWaiting();
		PlotInImpactMap(GetSolutionInfo(i, GET_PHASE),
				GetSolutionInfo(i, GET_VELOCITY), TRUE);
	}
}


	/*	--------------------------------------------	*/
	/*	---------- Impact Phases Routines ----------	*/
	/*	--------------------------------------------	*/


GetImpactPhasesOffset()
{
	int	oImpactPhasesOffset = ImpactPhasesOffset;

	if (DoImpactPhasesOffsetDialog(&ImpactPhasesOffset))
		if (ImpactPhasesOffset != oImpactPhasesOffset) {
			w[IPhs].DrawFromScratch = TRUE;
			StdForceUpdate((GrafPtr) w[IPhs].Peek);
		}
}

	/*	Routines related to Impact Phases magnification	*/

InitFullIPhsRange()
{
	int	i;

	FullIPhsRange[ymin_i] = MinPhase;
	FullIPhsRange[xmin_i] = MinPhase;
	FullIPhsRange[ymax_i] = MaxPhase;
	FullIPhsRange[xmax_i] = MaxPhase;
	for (i=ymin_i; i<ItemsInRect; ++i) {
		if (UsesFullIPhs[i])
			w[IPhs].GraphRange[i] = GetOffsetPhase(FullIPhsRange[i]);
		if (NewUsesFullIPhs[i])
			NewIPhsRange[i] = GetOffsetPhase(FullIPhsRange[i]);
	}
}

SetFullIPhsRange()
{
	int	i;

	InitFullIPhsRange();

	for (i=ymin_i; i<ItemsInRect; ++i) {
		NewUsesFullIPhs[i] = UsesFullIPhs[i] = TRUE;
		w[IPhs].GraphRange[i] = NewIPhsRange[i] =
					GetOffsetPhase(FullIPhsRange[i]);
	}
	w[IPhs].Magnifying = FALSE;
	w[IPhs].xGraphAroundZero = w[IPhs].yGraphAroundZero = TRUE;
	InitImpactPhaseScales();
}

GetIPMagnification(Boolean SkipDialog)
{
	int		i;
	double	min, max;
	Boolean	GoodOp;

	InitFullIPhsRange();
	for (i=ymin_i; i<ItemsInRect; ++i)
		NewIPhsRange[i] = GetUnOffsetPhase(NewIPhsRange[i]);

	Round2Doubles(NewIPhsRange[xmin_i], NewIPhsRange[xmax_i], RelPlaces, &min, &max);
	if (NewIPhsRange[xmin_i] > FullIPhsRange[xmin_i])
		NewIPhsRange[xmin_i] = min;
	if (NewIPhsRange[xmax_i] < FullIPhsRange[xmax_i])
		NewIPhsRange[xmax_i] = max;
	Round2Doubles(NewIPhsRange[ymin_i], NewIPhsRange[ymax_i], RelPlaces, &min, &max);
	if (NewIPhsRange[ymin_i] > FullIPhsRange[ymin_i])
		NewIPhsRange[ymin_i] = min;
	if (NewIPhsRange[ymax_i] < FullIPhsRange[ymax_i])
		NewIPhsRange[ymax_i] = max;

/*	BB 2.01 change
	if (!SkipDialog)
*/
	if ( (!SkipDialog) || (NewIPhsRange[xmin_i] == NewIPhsRange[xmax_i]) ||
			(NewIPhsRange[ymin_i] == NewIPhsRange[ymax_i]) )
		GoodOp = DoMagnifyImpactPhasesDialog(&NewIPhsRange, &FullIPhsRange,
				&NewUsesFullIPhs, ImpactPhasesOffset);
	else
		GoodOp = TRUE;

	if (GoodOp) {
		for (i=ymin_i; i<ItemsInRect; ++i)
			NewIPhsRange[i] = GetOffsetPhase(NewIPhsRange[i]);
		for (i=ymin_i; i<ItemsInRect; ++i)
			UsesFullIPhs[i] = NewUsesFullIPhs[i];

		if (	(!UsesFullIPhs[ymin_i]) || (!UsesFullIPhs[xmin_i]) ||
				(!UsesFullIPhs[ymax_i]) || (!UsesFullIPhs[xmax_i]) ) {
			w[IPhs].Magnifying = TRUE;
/* BB 2.01 deletion, 5/24/92 -- moved out of this conditional statement
			w[IPhs].xGraphAroundZero =
				(NewIPhsRange[xmin_i]*NewIPhsRange[xmax_i])<=0;
			w[IPhs].yGraphAroundZero =
				(NewIPhsRange[ymin_i]*NewIPhsRange[ymax_i])<=0;
*/
		}
/*	BB 2.01 addition, 5/24/92
 */
		else
			w[IPhs].Magnifying = FALSE;
		w[IPhs].xGraphAroundZero =
			(NewIPhsRange[xmin_i]*NewIPhsRange[xmax_i])<=0;
		w[IPhs].yGraphAroundZero =
			(NewIPhsRange[ymin_i]*NewIPhsRange[ymax_i])<=0;
/*	end of addition
 */
		if (	(w[IPhs].GraphRange[ymin_i] != NewIPhsRange[ymin_i]) ||
				(w[IPhs].GraphRange[xmin_i] != NewIPhsRange[xmin_i]) ||
				(w[IPhs].GraphRange[ymax_i] != NewIPhsRange[ymax_i]) ||
				(w[IPhs].GraphRange[xmax_i] != NewIPhsRange[xmax_i]) ) {
			for (i=ymin_i; i<ItemsInRect; ++i)
				w[IPhs].GraphRange[i] = NewIPhsRange[i];
			InitImpactPhaseScales();
		}
	}
	else
		for (i=ymin_i; i<ItemsInRect; ++i) {
			NewIPhsRange[i] = w[IPhs].GraphRange[i];
			NewUsesFullIPhs[i] = UsesFullIPhs[i];
		}
}

Boolean	TrackIPhsRect(Point pt)
/* Gets Impact Phases magnification from mouse presses */
{
	int	*pi1, *pi2, *pi3, i, index;

	if ( TrackDashedRect(pt.h, pt.v, &w[IPhs].GR, &IPhsBox, TRUE) ) {
		pi1 = (int *) &IPhsBox;
		pi2 = (int *) &w[IPhs].GR;
		pi3 = (int *) &w[IPhs].Origin;
		for (i=ymin_i; i<ItemsInRect; ++i) {
			if (i%2==0)	index = (i+2)%4;
			else		index = i;
			if ( (pi1[i] == pi2[i]) && UsesFullIPhs[index] ) {
				NewUsesFullIPhs[index] = TRUE;
				NewIPhsRange[index] = FullIPhsRange[index];
			}
			else {
				NewUsesFullIPhs[index] = FALSE;
				if (i%2==0)
					NewIPhsRange[index] =
						yCoordToFloat(IPhs, pi3[i%2] - pi1[i]);
				else
					NewIPhsRange[index] =
						xCoordToFloat(IPhs, pi1[i] - pi3[i%2]);
			}
		}
		return(TRUE);
	}
	return(FALSE);
}

InitImpactPhaseScales()
{
	Rect	tmpRect;
	double	range;
	int		i;

	SetPort(w[IPhs].Peek);
	w[IPhs].DrawFromScratch = TRUE;
	StdForceUpdate((WindowPtr) w[IPhs].Peek);

		/*	Calculate space needed for axis labels	*/
	if (w[IPhs].Magnifying) {
		range = w[IPhs].GraphRange[ymax_i]-w[IPhs].GraphRange[ymin_i];
		for(i=-2;i<4;++i) {
			if (range > 1.0) break;
			range *= 10;
		}
		IPYFieldWidth = StdIMFieldWidth + MAX(i,0);

		range = w[IPhs].GraphRange[xmax_i]-w[IPhs].GraphRange[xmin_i];
		for(i=-2;i<4;++i) {
			if (range > 1.0) break;
			range *= 10;
		}
		IPXFieldWidth = StdIMFieldWidth + MAX(i,0);
	}
	else
		IPYFieldWidth = IPXFieldWidth = StdIMFieldWidth;

	tmpRect = w[IPhs].Peek->port.portRect;
	tmpRect.left += (StdMargin + (IPYFieldWidth+3)*CharWidth('0'));
	tmpRect.top	+= IM_TopMargin;
	tmpRect.right -= (ScrollBarWidth + StdMargin +
			IPXFieldWidth * CharWidth('0') / 2);
	tmpRect.bottom -= (ScrollBarWidth + StdMargin + TextLineHeight + TextHeight);
	w[IPhs].GR = tmpRect;

	range = w[IPhs].GraphRange[xmax_i] - w[IPhs].GraphRange[xmin_i];
	if (range == 0.0) range = 1.0;
	w[IPhs].xScale = (double) (tmpRect.right - tmpRect.left) / range;

	range = w[IPhs].GraphRange[ymax_i] - w[IPhs].GraphRange[ymin_i];
	if (range == 0.0) range = 1.0;
	w[IPhs].yScale = (double) (tmpRect.bottom - tmpRect.top) / range;
				/*	Denominators (range) 1.0 if no magnification	*/

	w[IPhs].Origin.v = w[IPhs].GR.bottom;
	w[IPhs].Origin.h = w[IPhs].GR.left;
	w[IPhs].BackOrigin.v = w[IPhs].GR.bottom;
	w[IPhs].BackOrigin.h = w[IPhs].GR.left;
	SetRect(&w[IPhs].PR,
			w[IPhs].GR.left - w[IPhs].Origin.h,
			w[IPhs].Origin.v - w[IPhs].GR.bottom,
			w[IPhs].GR.right - w[IPhs].Origin.h + 1,
			w[IPhs].Origin.v - w[IPhs].GR.top + 1);
			/* +1 to account for pixel hanging below and to left of point */

	tmpRect.top = tmpRect.bottom+1;
	tmpRect.bottom = w[IPhs].Peek->port.portRect.bottom - (ScrollBarWidth-1);
	InsetRect(&tmpRect, -1, 0);
	ResizeControl(w[IPhs].HS[0], &tmpRect);


	tmpRect = w[IPhs].GR;
	tmpRect.right = tmpRect.left-1;
	tmpRect.left = 0;
	InsetRect(&tmpRect, 0, -1);
/*
	tmpRect.left -= MAX(StringWidth(VMaxString), StringWidth(VMinString))+3;
	tmpRect.top -= TextHeight/2;
	tmpRect.bottom += TextHeight/2+1;
*/
	ResizeControl(w[IPhs].HS[1], &tmpRect);
}

PlotImpactPhases(double del, double time, int myHitNumber, Boolean Highlighting)
{
	Point	pt;
	double	PrevOffsetDel;
	double	OffsetDel = GetOffsetPhase(del);
	int		index;

	w[IPhs].WindowCleared = FALSE;

	SetPort(w[IPhs].Peek);
	if (Highlighting)
		if (IPLatestImpactRect.left != 0) {
			 InvertOval(&IPLatestImpactRect);
			SetPort(w[IPhs].Back);
			 InvertOval(&IPLatestImpactRect);
			SetPort(w[IPhs].Peek);
		}
	if (time == 0.0)
		CurrentIPOffsetCount = -1;
	if (++CurrentIPOffsetCount >= ImpactPhasesOffset) {
		index = myHitNumber - ImpactPhasesOffset;
		PrevOffsetDel = GetOffsetPhase(GetSolutionInfo(index, GET_PHASE));
		pt.h = xFloatToCoord(IPhs, PrevOffsetDel);
		pt.v = yFloatToCoord(IPhs, OffsetDel);
				/*	Next (2) statement(s): if rounded would be MaxPhase, make MinPhase;
				 *	if delete this statement, enlarge drawn rect surrounding
				 *	GR by one pixel on the right (and top).
				 */
		if (pt.h == /* was >= */ w[IPhs].GR.right-w[IPhs].Origin.h)
			if (	(!w[IPhs].Magnifying) ||
					( (w[IPhs].GraphRange[xmin_i] == MinPhase) &&
					(w[IPhs].GraphRange[xmax_i] == MaxPhase) ) )
				pt.h = w[IPhs].GR.left-w[IPhs].Origin.h;
		if (pt.v >= /* was >= */ w[IPhs].Origin.v - w[IPhs].GR.top)
			if (	(!w[IPhs].Magnifying) ||
					( (w[IPhs].GraphRange[ymin_i] == MinPhase) &&
					(w[IPhs].GraphRange[ymax_i] == MaxPhase) ) )
				pt.v = w[IPhs].Origin.v - w[IPhs].GR.bottom;
		if ( !PtInRect(pt, &w[IPhs].PR) ) {
			if (Highlighting)
				OffsetRect(&IPLatestImpactRect,-IPLatestImpactRect.left,
						-IPLatestImpactRect.top);
			return;
		}
		BothMoveTo(IPhs, pt.h, pt.v);
		SetPort(w[IPhs].Peek);	Line(0, 0);
		SetPort(w[IPhs].Back);	Line(0, 0);
		SetPort(w[IPhs].Peek);
		if (Highlighting) {
				/* next statement from WindowMoveTo */
			OffsetRect(&IPLatestImpactRect, -IPLatestImpactRect.left + pt.h +
				w[IPhs].Origin.h - 2,
				-IPLatestImpactRect.top + w[IPhs].Origin.v - pt.v - 2);
			 InvertOval(&IPLatestImpactRect);
			SetPort(w[IPhs].Back);
			 InvertOval(&IPLatestImpactRect);
			SetPort(w[IPhs].Peek);
		}
	}
	else
		OffsetRect(&IPLatestImpactRect,
				-IPLatestImpactRect.left, -IPLatestImpactRect.top);
}

RefreshIP(int myHitNumber)
{
	int		i, j;
	int		x, y;
	char	*s;
	Str255	s1;
	Rect	tmpRect;
	Str255	LocalXMinString, LocalXMaxString;
	Str255	LocalYMinString, LocalYMaxString;

	int		OverflowOnRight, MaxPStrLen;	/* intermediate vars	*/

	if (!w[IPhs].DrawFromScratch) {
		SetPort(w[IPhs].Peek);
		myCopyBits(w[IPhs].Back, (GrafPtr) w[IPhs].Peek,
			&w[IPhs].Peek->port.portRect, 0,0,0,0, srcCopy, (RgnHandle) NIL);
		return;
	}

	w[IPhs].WindowCleared = TRUE;

	w[IPhs].DrawFromScratch = FALSE;
	SetPort(w[IPhs].Back);
	EraseRect(&w[IPhs].Back->portRect);

	tmpRect = w[IPhs].GR;
/*
	InsetRect(&tmpRect, -1, 0);
	tmpRect.bottom += 2;
*/
	InsetRect(&tmpRect, -1, -1);
	/*	Use this if don't adjust near-max phases (see PlotInImpactMap()) 
		++tmpRect.right;
	*/
	FrameRect(&tmpRect);

		/* Draw Hash marks */
	for(i=0;i<4;++i) {
		x = xFloatToCoord(IPhs, GetOffsetPhase(i/4.0));
		if ( (x<=0) || (x>=w[IPhs].GR.right-w[IPhs].GR.left-1) )
			continue;
		WindowMoveTo(IPhs, x, w[IPhs].Origin.v-w[IPhs].GR.bottom+HashSize+1);
		Line(0, HashSize);
		Move(0, -(w[IPhs].GR.bottom - w[IPhs].GR.top - 1) );
		Line(0, HashSize);
	}
	for(i=0;i<4;++i) {
		y = yFloatToCoord(IPhs, GetOffsetPhase(i/4.0));
		if ( (y<=0) || (y>=w[IPhs].GR.bottom-w[IPhs].GR.top-1) )
			continue;
		WindowMoveTo(IPhs, w[IPhs].GR.left-w[IPhs].Origin.h+HashSize, y);
		Line(-HashSize, 0);
		Move(w[IPhs].GR.right - w[IPhs].GR.left - 1, 0);
		Line(-HashSize, 0);
	}

		/*	Get endpoint strings; increase precision if magnification dictates */
	if (w[IPhs].Magnifying) {
		myFToStrNoPad(GetUnOffsetPhase(w[IPhs].GraphRange[ymin_i]),
				CP(LocalYMinString),
				IPYFieldWidth, IPYFieldWidth-IMPrecDiff);
		myFToStrNoPad(GetUnOffsetPhase(w[IPhs].GraphRange[ymax_i]),
				CP(LocalYMaxString),
				IPYFieldWidth, IPYFieldWidth-IMPrecDiff);
		myFToStrNoPad(GetUnOffsetPhase(w[IPhs].GraphRange[xmin_i]),
				CP(LocalXMinString),
				IPXFieldWidth, IPXFieldWidth-IMPrecDiff);
		myFToStrNoPad(GetUnOffsetPhase(w[IPhs].GraphRange[xmax_i]),
				CP(LocalXMaxString),
				IPXFieldWidth, IPXFieldWidth-IMPrecDiff);
	}
	else {
		sEquate(CP(LocalYMinString), CP(MinPhaseString));
		sEquate(CP(LocalYMaxString), CP(MaxPhaseString));
		sEquate(CP(LocalXMinString), CP(MinPhaseString));
		sEquate(CP(LocalXMaxString), CP(MaxPhaseString));
	}

		/* Label vertical (phase n+1) axis */
	WindowMoveTo(IPhs, 0, w[IPhs].Origin.v - w[IPhs].GR.top);
		Move(-TextMargin - StringWidth(&LocalYMaxString), TextHeight/2);
		DrawString(&LocalYMaxString);
	WindowMoveTo(IPhs, 0, w[IPhs].Origin.v - w[IPhs].GR.bottom);
		Move(-TextMargin - StringWidth(&LocalYMinString), TextHeight/2-1);
		DrawString(&LocalYMinString);
	y=yFloatToCoord(IPhs, GetOffsetPhase(0.0));
	if ( (y > TextHeight) && (y <
			w[IPhs].GR.bottom - w[IPhs].GR.top - TextHeight+1) ) {
		WindowMoveTo(IPhs, w[IPhs].GR.left-w[IPhs].Origin.h, y);
		Move(-TextMargin - CharWidth('0'), TextHeight/2);
		DrawChar('0');
	}

	sprintf(CP(s1), "Phasen+%d", ImpactPhasesOffset);
	CtoPstr(CP(s1));
	SetRect(&tmpRect, StdMargin+2,
			(w[IPhs].GR.top+w[IPhs].GR.bottom-((int)s1[0])*TextLineHeight)/2,
			StdMargin + 12/*CharWidth('V') + 1*/, MaxScreenY);
	tmpRect.top = MAX(tmpRect.top, w[IPhs].GR.top);
	TextBox(&(s1[1]), (long) (s1[0]), &tmpRect, teJustCenter);

		/* Label phase (horizontal) axis */
/*
	MoveTo(w[IPhs].GR.left, w[IPhs].GR.bottom);
	Move(-StringWidth(CP(LocalXMinString))/2, TextMargin + TextHeight-1);
	DrawString(CP(LocalXMinString));
	MoveTo(w[IPhs].GR.right, w[IPhs].GR.bottom);
	Move(-StringWidth(CP(LocalXMaxString))/2, TextMargin + TextHeight-1);
	DrawString(CP(LocalXMaxString));
	x=xFloatToCoord(IPhs, GetOffsetPhase(0.0));
	if ( (x > StringWidth("\P-0.00")/2+4) && (x <
			w[IPhs].GR.right - w[IPhs].GR.left - StringWidth("\P0.00")/2-4) ) {
		WindowMoveTo(IPhs, x, w[IPhs].Origin.v-w[IPhs].GR.bottom);
		Move(-CharWidth('0')/2+1, TextMargin + TextHeight-1);
		DrawChar('0');
	}
*/
	MoveTo(w[IPhs].GR.left, w[IPhs].GR.bottom);
	Move(-StringWidth(CP(LocalXMinString))/2, TextMargin + TextHeight-1);
	DrawString(CP(LocalXMinString));
	MoveTo(w[IPhs].GR.right, w[IPhs].GR.bottom);
	MaxPStrLen = StringWidth(CP(LocalXMaxString));
	OverflowOnRight = MAX(MaxPStrLen/2 - ImpactRightMargin, 0);
	Move(-MaxPStrLen/2 - OverflowOnRight, TextMargin + TextHeight-1);
	DrawString(CP(LocalXMaxString));

	x=xFloatToCoord(IPhs, GetOffsetPhase(0.0));
	if ( (x > StringWidth(CP(LocalXMinString))/2+4) &&
			(x < w[IPhs].GR.right - w[IPhs].GR.left -
			StringWidth(CP(LocalXMaxString))/2-4) ) {
		WindowMoveTo(IPhs, x, w[IPhs].Origin.v-w[IPhs].GR.bottom);
		Move(-CharWidth('0')/2+1, TextMargin + TextHeight-1);
		DrawChar('0');
	}

	s = "\PPhasen";
	MoveTo( (w[IPhs].GR.right + w[IPhs].GR.left)/2, w[IPhs].GR.bottom);
	Move(-StringWidth(s)/2, TextMargin + TextLineHeight + TextHeight - 3);
	DrawString(s);

	SetPort(w[IPhs].Peek);
	myCopyBits(w[IPhs].Back, (GrafPtr) w[IPhs].Peek,
			&w[IPhs].Peek->port.portRect, 0,0,0,0, srcCopy, (RgnHandle) NIL);

	OffsetRect(&IPLatestImpactRect,-IPLatestImpactRect.left,-IPLatestImpactRect.top);
	if (myHitNumber >= StartHitNumber) {
		if (myHitNumber - StartHitNumber > MANY_IMPACTS)
			StartWaiting();
		i=StartHitNumber;
		j=0;
		while(TRUE) {
			if (i<0)
				break;
			if (GetSolutionInfo(i, GET_TIME) == 0.0)
				break;
			if (++j >= ImpactPhasesOffset)
				break;
			--i;
		}
		i = StartHitNumber + (ImpactPhasesOffset-j);
		CurrentIPOffsetCount = ImpactPhasesOffset - (i-myHitNumber);

		if (i>myHitNumber) return;
/*
		if ( (**(mySolutions[SI1(i)]))[SI2(i)].time != 0.0 )
			PlotImpactPhases((**(mySolutions[SI1(i-1)]))[SI2(i-1)].del,
					0.0, i, FALSE);
*/
		for (;i<=myHitNumber-1;++i)
			PlotImpactPhases(GetSolutionInfo(i, GET_PHASE),
					GetSolutionInfo(i, GET_TIME), i, FALSE);
		if (myHitNumber - StartHitNumber > MANY_IMPACTS)
			StopWaiting();
		PlotImpactPhases(GetSolutionInfo(i, GET_PHASE),
				GetSolutionInfo(i, GET_TIME), i, TRUE);
	}
}


		/*	-------------------------------------------------------	*/
		/*	------           Inter-Window Routines           ------	*/
		/*	------ (incl. AddGraphicSolution, Bottom Line) ------	*/
		/*	-------------------------------------------------------	*/

Boolean	AddGraphicSolution(hitnum, thePhase, velocity, time, StickyState, PS)
	int		hitnum;
	double	thePhase, velocity, time;
	int		StickyState;
	int		PS;		/*	ProgramStatus	*/
{
	Boolean	WriteOK;

/*
	switch (PS) {
		case BIFURCATION_DIAGRAM:
					if (!*pBDGraphInAllWindows)
						return(TRUE);
					break;
		case BASINS_OF_ATTRACTION:
					if (!*pBAGraphInAllWindows)
						return(TRUE);
					break;
		default:	break;
	}
*/
	if ( (PS == BIFURCATION_DIAGRAM) && (!*pBDGraphInAllWindows) )
		return(TRUE);
	if ( (PS == BASINS_OF_ATTRACTION) && (!*pBAGraphInAllWindows) )
		return(TRUE);

	WriteOK = WriteOutput(hitnum, thePhase, velocity, time, StickyState, PS);
	if (!PrintingOnly) {
		if (WindowOpen(IMap))
			PlotInImpactMap(thePhase, velocity, TRUE);
		if (WindowOpen(IPhs))
			PlotImpactPhases(thePhase, time, hitnum, TRUE);
	}
	return(WriteOK);
}

NewImpactGraphics()
{
	StartHitNumber = 0;
}

InitImpactGraphics()
{
	SetRect(&LatestImpactRect,0,0,5,5);
	SetRect(&IPLatestImpactRect,0,0,5,5);
	InitBBTextEdit(A, freq, damp);
	StartHitNumber = 0;
	InitFullIMapRange();
	InitFullIPhsRange();
}

HaltImpactGraphics()
{
	HaltBBTextEdit();
}
