/*	bbIteration.c
 *
 *	Handles Bouncing Ball's iteration
 *
 *	-TA 7/23/90
 */

#include	"bbPrototypes.h"

		/*	Returned by GetFileHeader, used in StartOpeningFile	*/
enum {
		GOOD_HEADER,
		TIME_SERIES_HEADER,
		BAD_HEADER
};

				/*	Global external vars from bb.c	*/
extern	IterationVars		NormalVars, *pCurrentVars;
extern	hBBPreferences		hBBPrefs;
								/*	External variables from bb.c	*/
extern	SolutionsHandles	mySolutions;
extern	int					ProgramStatus, LastProgramStatus;
extern	int					theEventMask, UsualEventMask;

								/*	From bbMenus.c	*/
extern	Boolean				NeedToSave;

								/*	From bbContGraphics.c	*/
extern	Boolean				RedrawSine;
extern	Boolean				NeedToResetTSVars;

								/*	From bbText.c	*/
extern	TEHandle			HeaderTE;

								/*	From bbBifDiag.c, for OKToGraph()	*/
extern	Boolean				*pBDGraphInAllWindows, *pBAGraphInAllWindows;
								/*	From bbSpecial.c, for InitNextIteration()	*/
extern	Boolean				NeedNextSpecialValue;
								/*	From bbGraphics.c, for opening	*/
extern	CursHandle			watch;


			/*	Global variables shared with bbGraphics.c	*/
Str255	CurrentLine, StuckLine;

			/* Iteration variables */
Str255	NextLine;

double	LastT;


		/*	----------------------------------------------------------	*/
		/*	----------	Solution Method-Invariant Routines	----------	*/
		/*	----------------------------------------------------------	*/

SetHitNumber(int NewHitNumber)
{
	HitNumber = NewHitNumber;
	HitInd1 = SI1(HitNumber);
	HitInd2 = SI2(HitNumber);
}

Boolean	HandleSticking(int theEjection)
	/*	Now only called during NORMAL_ITERATION -TA 8/9/90	*/
	/*	And BIFURCATION_DIAGRAM, BASINS_OF_ATTRACTION -TA 2/27/91	*/
{
	switch (theEjection) {
		case NoEjection:
				StuckStatus = STUCK_FOREVER;
				if (ProgramStatus >= BIFURCATION_DIAGRAM)
					StartStickingMessage(STUCK_FOREVER);
				return(FALSE);
/*
				return( (ProgramStatus != NORMAL_ITERATION) &&
						(ProgramStatus <= BIFURCATION_DIAGRAM) );
*/
		case InAreaOfEjection:
				return(TRUE);	/* damp == 0, but not stuck here */
		case NormalEjection:
/*
				HandleAllUpdates();
*/
				++StuckCount;
				StuckStatus = STUCK;
				return(TRUE);
	}
}

Boolean SetItoJ(Boolean NeedPhase)
	/*	Used in NORMAL_ITERATION and PLAYBACK	*/
{
	xi = xj;	vi = vj;	ti = tj;
	if (NeedPhase)
		phase = (freq==0.0) ? delta : fmod(ti/Period + delta, 1.0);
	ti_Continuing = ti;

	if (StuckStatus == STUCK) {
		StopStickingMessage();
		StuckStatus = NOT_STUCK;
		return(FALSE);
	}
	else
		return(TRUE);
}

SetInitConditions()
	/*	Used in PLAYBACK and OPENING	*/
{
	v0 = vj;
	if ( delta != phase ) {
		delta = phase;
		RedrawSine = TRUE;
	}
	SetTrajTimeOffset(HitTime);
		/*		Brought from Initialize_Iteration_Globals -TA 4/29/91	*/
	StuckCount = 0;
	StuckStatus = NOT_STUCK;
/*	I had deleted HitNumber>0 test; reinserted it because of
 *	double updating when start opening or playback.
 *	Did I delete it for some reason other than generality?
 *		-TA 12/26/90
 *	Need for v0, delta in Trajectory window; Animation window (?);
 *	other continuous windows (?)
 *		-TA 12/26/90
 */
/*
	if (HitNumber > 0) {
		NewGraphics(FALSE, FALSE);
		HandleAllUpdates();
	}
*/
	if (HitNumber > 0)
		NewGraphics(FALSE, FALSE);
	else
		NewContGraphics();
	HandleAllUpdates();
}

		/*	----------------------------------------------------------	*/
		/*	----------	First Solution Method: Iteration	----------	*/
		/*	----------------------------------------------------------	*/

Boolean InitNextIteration()
{
	int i;

	if ( SetItoJ(TRUE) ) {
		if ( !FindNStep(&ndt, ti, xi, vi) ) {
			if ( !HandleSticking( FindEjection(&ndt,&dt,&tj,&xj,&vj) ) )
				return(FALSE);
/* changed test in MAIN(), to check for StuckStatus
				return(FALSE);
*/
		}
		else dt = ndt;
	}

	LastT = 0.0;
	t = dt; tau = ti + t;
		/*	If no root on interval messages appear, put Nick's kludge here */
		/*	(Moved Nick's kludge to FindNStep	-TA 6/12/91) */
	
	ElapsedTime += dt;
	return(TRUE);
}

Boolean	Initialize_Iteration_Globals(Boolean BeginningIteration,
		Boolean ForceV0)
{
	if ( v0 < vTable(0.0) ) {
		if (ForceV0)
			v0 = vTable(0.0) * 1.01;
		else {
			ERROR("\PError: Initial velocity less than table velocity");
			return(FALSE);
		}
	}
	if (!BeginningIteration) return(TRUE);

	StuckCount = 0;
	StuckStatus = NOT_STUCK;
	ElapsedTime = HitTime = 0.0;
	SetTrajTimeOffset(HitTime);

	tau = tj = 0.0;
	xj = xTable(tau);
	vj = v0;

/*
	return(InitNextIteration());
*/
	return(TRUE);	/* InitNextIteration done by StartNextIteration()	*/
}

Boolean Reset_Iteration_Globals()
	/* for "Continuing Iteration" */
{
	ElapsedTime = HitTime = (**(mySolutions[HitInd1]))[HitInd2].time;
	tj = ti_Continuing;
	vj = (**(mySolutions[HitInd1]))[HitInd2].vel;
	xj = xTable(tj);

	return(InitNextIteration());
}

int IterateSolution()
{
static	int	IterNum;	/*	eventually get rid of	*/
	double	distance, theRoot;

	if (StuckStatus == STUCK) {
		StartStickingMessage(StuckCount);
		HitTime = tj;
		ElapsedTime = HitTime;
		return(FOUND_SOLUTION);
	}

	if (ElapsedTime > tstop)
		{	ERROR("\PTime Expired");	return(BAD_SOLUTION);	}

	LastT = t;	/*	Needed because sometimes (t+dt)-dt != t,
				 *	because of precision	-TA 6/12/91	*/
	t += dt;	tau = ti + t;
	ElapsedTime += dt;
++IterNum;
	if ( impact(distance = FindDistance(tau, t, xi, vi)) ) {
IterNum=0;
		if (!root(&theRoot, LastT, t, xi, vi, ti)) {
			if (ProgramStatus >= BIFURCATION_DIAGRAM) {
				ERROR("\PRoot finding error; Bouncing Ball will assume sticking solution");
				return(ROOT_FINDING_ERROR);
			}
			else {
				ERROR("\PRoot finding error: no zero on interval");
				return(BAD_SOLUTION);
			}
		}
		t = theRoot;
		tj = ti + t;
			/* HitTime does not get reset when Trajectory screen filled */
		ElapsedTime = HitTime = tj;
		xj = xTable(tj);
		vj = SetVj(tj, t, vi);
		return(FOUND_SOLUTION);
	}
	else { /* no impact */
		tdt = FindTStep(distance, tau, t, xi, vi);
		if (tdt>ndt) dt = tdt; else dt = ndt;
		return(NO_SOLUTION_YET);
	}
}

		/*	----------------------------------------------------------	*/
		/*	----------	Second Solution Method: Playback	----------	*/
		/*	----------------------------------------------------------	*/

Boolean InitPlayback()
{
	int		tmp_RealHitNumber, tmp_HitNumber = HitNumber;
	Boolean	oNeedToSave = NeedToSave;

	pCurrentVars = &NormalVars;
	tmp_RealHitNumber = MAX(RealHitNumber, HitNumber);
	RETURN_IF_FALSE(tmp_RealHitNumber>=0);
	SetHitNumber(-1);
	NewGraphics(TRUE, FALSE);
	NeedToSave = oNeedToSave;	/*	set to FALSE by NewGraphics()	*/
	if ( !StartIterating(PLAYBACK) ) {	/* in for menu handling, other */
		SetHitNumber(tmp_HitNumber);
		return(FALSE);
	}
	StuckCount = 0;
	RealHitNumber = tmp_RealHitNumber;
/*
	ProgramStatus = PLAYBACK;
*/
	ti_Continuing = 0.0;
	return(TRUE);
}

GetPlaybackSolution()
{
	if (StuckStatus == STUCK)
		StartStickingMessage(StuckCount);

	SetHitNumber(++HitNumber);

	vj		= (**(mySolutions[HitInd1]))[HitInd2].vel;
	phase	= (**(mySolutions[HitInd1]))[HitInd2].del;
	HitTime	= (**(mySolutions[HitInd1]))[HitInd2].time;
	if ( HitTime == 0.0 ) {
		StuckStatus = NOT_STUCK;
		SetInitConditions();
	}
	tj = HitTime;
	xj = xTable(tj);
}

Boolean CheckNextPlayback()
{
	if ( SetItoJ(FALSE) ) ;

	StuckStatus = (**(mySolutions[HitInd1]))[HitInd2].stuck;
/*
	if (StuckStatus != NOT_STUCK)
		RETURN_IF_FALSE( HandleSticking(StuckStatus) );
			(* never FALSE, 8/9/90 *)
*/
	if (StuckStatus == STUCK) ++StuckCount;
	return( HitNumber < RealHitNumber );
}

		/*	------------------------------------------------------	*/
		/*	----------	Third Solution Method: Opening	----------	*/
		/*	------------------------------------------------------	*/

StopOpeningFile(Boolean OpenOK)
{
	theEventMask = UsualEventMask;
	StopWaiting();

	CloseOpenFile();
	if (!OpenOK) {
		SetTextTitle(FALSE, TxOut);
		if (HitNumber < 0)
			NewGraphics(FALSE, FALSE);
/*
		(* else *)
			NeedToSave = TRUE;
*/
	}
	else
		NeedToSave = FALSE;
	if (HitNumber >= 0) {
		ti_Continuing = HitTime;
		if (!GraphWhileOpening) {
			SetTrajTimeOffset(HitTime);
			NeedToResetTSVars = TRUE;
		}
	}
}

Boolean	GetFileParam(double *pf)
{
	Str255	theLine;
	int		ItemsRead;

	RETURN_IF_FALSE(BBGetFileLine(CP(theLine)));
	PtoCstr(CP(theLine));
	ItemsRead = sscanf(CP(theLine), "%*s %lf", pf);
	CtoPstr(CP(theLine));
	RETURN_IF_FALSE(ItemsRead == 1);
/*
	AppendText(CP(theLine), HeaderTE);
*/
	return(TRUE);
}

int	GetFileHeader()
{
	Str255	theLine;

	if ( !BBGetFileLine(CP(theLine)) )		return(BAD_HEADER);
	if ( !CheckFirstLine(CP(theLine)) )	return(TIME_SERIES_HEADER);

	if (!GetFileParam(&A))		return(BAD_HEADER);
	if (!GetFileParam(&freq))	return(BAD_HEADER);
	if (!GetFileParam(&damp))	return(BAD_HEADER);

	if ( !BBGetFileLine(CP(theLine)) )		return(BAD_HEADER);
/*
	AppendText(CP(theLine), HeaderTE);
*/
	SetHeaders(A, freq, damp);

	return(GOOD_HEADER);
}

Boolean StartOpeningFile()
{
	double	oA=A, oFreq=freq, oDamp=damp;
	int		HeaderStatus;

	pCurrentVars = &NormalVars;
	ResetSolutions();
	NewGraphics(TRUE, FALSE);
	InitHeaderTEs();
	StopIterating(ProgramStatus >= BIFURCATION_DIAGRAM);
	NeedToSave = FALSE;

	if (!StartingMemCheck()) {
		ERROR("\PNot enough memory to open file");
		SetTextTitle(FALSE, TxOut);
		SetHeaders(A, freq, damp);
		return(FALSE);
	}
	if (!OpenFile()) {
		ERROR("\POpen operation was not successful");
		SetTextTitle(FALSE, TxOut);
		SetHeaders(A, freq, damp);
		return(FALSE);
	}
	StartWaiting();
	if ( (HeaderStatus=GetFileHeader()) != GOOD_HEADER) {
		StopOpeningFile(FALSE);
		if (HeaderStatus == TIME_SERIES_HEADER)
			ERROR("\PError: Cannot open Bouncing Ball Time Series files");
		else if (HeaderStatus == BAD_HEADER)
			ERROR("\PError: Not a Bouncing Ball file");
		ResetDelimiter();
		A = oA;	freq = oFreq;	damp = oDamp;
		SetHeaders(A, freq, damp);
		return(FALSE);
	}

	NeedToSave = TRUE;
/*	deleted 4/29/91v0, delta not set yet.
	RETURN_IF_FALSE(Initialize_Iteration_Globals(TRUE, FALSE));
	RETURN_IF_FALSE(InitNextIteration());
*/

	SetValuesFromTableParams(oA, oFreq, oDamp);

	SetTextTitle(TRUE, TxOut);

	if (!BBGetFileLine(CP(NextLine))) {
		NextLine[0]=0;
		StopOpeningFile(TRUE);
		return(FALSE);
	}

	ActivateTextScrolls(FALSE);
	LastProgramStatus = ProgramStatus = OPENING;
	theEventMask = keyDownMask | updateMask | app4Mask;
	return(TRUE);
}

Boolean	GetFloats(char *Line, double *pphase, double *pvel, double *ptime)
{
	int		ItemsRead;

	PtoCstr(Line);
	ItemsRead = sscanf(Line, "%*ld %lf %lf %lf", pphase, pvel, ptime);
	CtoPstr(Line);
	return(ItemsRead == 3);
}

Boolean CheckForStuckOpen()
{
	int	OldStuckStatus;

	OldStuckStatus = NewStuckStatus;
	if (StrFind("\Pable", CP(NextLine)))
		NewStuckStatus = STUCK;
	else if (StrFind("\Porever", CP(NextLine)))
		NewStuckStatus = STUCK_FOREVER;
	else
		NewStuckStatus = NOT_STUCK;

	if (NewStuckStatus != NOT_STUCK) {
		sEquate(CP(StuckLine), CP(NextLine));
		if (!BBGetFileLine(CP(NextLine))) {
			NextLine[0] = 0;
			if (NewStuckStatus == STUCK) {
				NewStuckStatus = OldStuckStatus;
				return(FALSE);
			}
		}
		else if (NewStuckStatus == STUCK)
			++StuckCount;
	}
	return(TRUE);
}

int GetOpenSolution()
{
	double	tmpPhase, tmpVj, tmpHitTime;
	Boolean	GoodSolution;

	sEquate(CP(CurrentLine), CP(NextLine));
	GoodSolution = GetFloats(CP(CurrentLine), &tmpPhase, &tmpVj, &tmpHitTime);

	if (GoodSolution) {
		if (!BBGetFileLine(CP(NextLine))) NextLine[0] = 0;
		if (!CheckForStuckOpen())
			GoodSolution = FALSE;
	}

	if (!GoodSolution) {
		ERROR("\PUnable to open entire file");
		StopOpeningFile(FALSE);
		return(BAD_SOLUTION);
	}

	if ( GraphWhileOpening && (StuckStatus == STUCK) )
		StartStickingMessage(StuckCount);

	phase = tmpPhase;
	vj = tmpVj;
	HitTime = tmpHitTime;
	if (GraphWhileOpening) {
		tj = HitTime;
		xj = xTable(tj);
	}
	if ( HitTime == 0.0 )
		SetInitConditions();
	return(FOUND_SOLUTION);
}

Boolean FinishedOpening()
{
	return( NextLine[0] == 0 );
}

Boolean CheckNextOpening()
{
	Boolean	Successful = TRUE;

	if ( FinishedOpening() ) {
		Successful = FALSE;
		StopOpeningFile(TRUE);
	}

	SetHitNumber(++HitNumber);
	if (GraphWhileOpening) {
		if (SetItoJ(FALSE)) ;
	}
	else
		vi = vj;
	StuckStatus = NewStuckStatus;

	return(Successful);
}

		/*	------------------------------------------------------	*/
		/*	----------	Inter-Solution Method Routines	----------	*/
		/*	------------------------------------------------------	*/

int GetNextSolution(int PS /* ProgramStatus */ )
{
	switch (PS) {
		case NORMAL_ITERATION:
		case BIFURCATION_DIAGRAM:
		case BASINS_OF_ATTRACTION:
					return( IterateSolution() );
		case PLAYBACK:
					GetPlaybackSolution();
					return(FOUND_SOLUTION);
		case OPENING:
					return( GetOpenSolution() );
	}
}

Boolean StartNextSolution(int PS, Boolean *OKCont)
	/*	Iteration will be stopped here unless there's a problem	*/
{
	switch (PS) {
		case NORMAL_ITERATION:
		case BIFURCATION_DIAGRAM:
		case BASINS_OF_ATTRACTION:
					SetHitNumber(++HitNumber);
/*
					if (HitTime > 0.0)
*/
						if (!InitNextIteration()) {
							*OKCont = FALSE;
							if (PS == NORMAL_ITERATION)
								return(FALSE);
						}
/*
					if (PS == BIFURCATION_DIAGRAM)			return(GraphBD());
					else if (PS == BASINS_OF_ATTRACTION)	return(TRUE);
					else return(TRUE);
*/
					return(TRUE);
		case PLAYBACK:
					*OKCont = HitNumber >= 0;
					return( CheckNextPlayback() );
		case OPENING:
					*OKCont = HitNumber >= 0;
					return( CheckNextOpening() );
	}
}

Boolean OKtoGraph(int PS)
{
	switch (PS) {
		case NORMAL_ITERATION:		return(TRUE);
		case PLAYBACK:				return(HitTime != 0.0);
		case OPENING:				return(GraphWhileOpening && (HitTime != 0.0));
		case BIFURCATION_DIAGRAM:	return(*pBDGraphInAllWindows);
		case BASINS_OF_ATTRACTION:	return(*pBAGraphInAllWindows);
	}
}
