/*
 *	bbMenus.c
 *
 *	Handles Bouncing Ball's menu requirements
 *
 *	-TA 7/20/90
 */

#include <stdlib.h>
#include <MenuUtils.h>
#include "bbPrototypes.h"
#include "bbMenus.h"
#include "bbGraphics.h"		/*	For special window opening only	*/

#define	DOT					'.'


			/*	External variables from bb.c, mainly for Control menu handling	*/
extern		IterationVars	NormalVars, *pCurrentVars;
extern		hBBPreferences	hBBPrefs;
extern		int				ProgramStatus, IterationStatus;
extern		Boolean			ClearSolutions, PAUSING, OKToContinue;
extern		Boolean			OKToContinueBD, OKToContinueBA;
extern		int				CurrentModifiers;	/* for AdjustBBMenus	*/
extern		Boolean			SingleStepOn;

			/*	External variables from bbSpecial.c, for Special menu handling	*/
extern		Boolean			DrawBD, DrawBA;

			/*	External variables from bb.c for MultiFinder management	*/
extern		Boolean MultiFinderRunning;
extern		int		BackgroundSleep;

			/*	External variables from bbGraphics.c for special window opening	*/
extern		Windows	*pw;

MenuHandle	AppleMenuH, FileMenuH, EditMenuH;
MenuHandle	ControlMenuH, ParametersMenuH, ViewMenuH, SpecialMenuH;
MenuHandle	SimpleSoundMenuH, AdvancedSoundMenuH, SoundMenuH;
MenuHandle	WindowsMenuH;

Boolean		NeedToSave, NeedToSaveTS, NeedToSaveBD, NeedToSaveBA;


		/* ----------------------------------------- */
		/* ---------- Menu Initialization ---------- */
		/* ----------------------------------------- */

QuitPrematurely()
		/* due to resource fork problem */
{
		SysBeep(10); SysBeep(10); SysBeep(10);
		exit(1);
}

InitBBMenus(Boolean UsingColor)
{
	if (!StdAddMenu(AppleMenuID,		&AppleMenuH))		QuitPrematurely();
	if (!StdAddMenu(FileMenuID,			&FileMenuH))		QuitPrematurely();
	if (!StdAddMenu(EditMenuID,			&EditMenuH))		QuitPrematurely();
	if (!StdAddMenu(ControlMenuID,		&ControlMenuH))		QuitPrematurely();
/*
	if (!StdAddMenu(ParametersMenuID,	&ParametersMenuH))	QuitPrematurely();
*/
	if (!StdAddMenu(ViewMenuID,			&ViewMenuH))		QuitPrematurely();
	if (!StdAddMenu(SpecialMenuID,		&SpecialMenuH))		QuitPrematurely();

	if (!UseAdvancedSound) {
		if (!StdAddMenu(SimpleSoundMenuID,	&SimpleSoundMenuH))	QuitPrematurely();
		AdvancedSoundMenuH = GetMenu(AdvancedSoundMenuID);
		if (AdvancedSoundMenuH == NIL) {
			ERROR("\PUnable to create menus");
			exit(1);
		}
		SoundMenuH = SimpleSoundMenuH;
	}
	else {
		if (!StdAddMenu(AdvancedSoundMenuID,	&AdvancedSoundMenuH))	QuitPrematurely();
		SimpleSoundMenuH = GetMenu(SimpleSoundMenuID);
		if (SimpleSoundMenuH == NIL) {
			ERROR("\PUnable to create menus");
			exit(1);
		}
		SoundMenuH = AdvancedSoundMenuH;
	}

	if (!StdAddMenu(WindowsMenuID,		&WindowsMenuH))		QuitPrematurely();


	AddResMenu(AppleMenuH, 'DRVR');
		/*	*** Moved to later--windows not created yet ***
	AddWindowsToMenu();
		*/
		/*	Done in ControlMenuStop from StopIterating from _INIT
	DisableItem(ControlMenuH, PauseContinueItem);
		 */

	CheckItem(EditMenuH, SuppressSaveItem, SuppressSaveWarnings);
	CheckItem(EditMenuH, AutoCommandKeyItem, AutoCommandKey);
	if (MultiFinderRunning) {
		AppendMenu(EditMenuH, "\P(-");
		AppendMenu(EditMenuH, "\PBackground Operation.../G");
	}

	CheckItem(ViewMenuH, ScrollTrajectoryItem, ScrollTrajectory);
	CheckItem(ViewMenuH, FasterPhaseSpaceItem, FasterPhaseSpace);
	CheckItem(ViewMenuH, UseTimeSeriesItem, UseTimeSeries);

	if (!UsingColor) {
		DisableItem(SpecialMenuH, UseColorBDItem);
		DisableItem(SpecialMenuH, UseColorBAItem);
		DisableItem(SpecialMenuH, ColorsItem);
	}

	CheckItem(SimpleSoundMenuH, PhaseTonesItem, PhaseTonesOn);
	CheckItem(SimpleSoundMenuH, PhaseTonesItem+1, VelTonesOn);
	CheckItem(AdvancedSoundMenuH, PhaseTonesItem, PhaseTonesOn);
	CheckItem(AdvancedSoundMenuH, PhaseTonesItem+1, VelTonesOn);
	CheckItem(AdvancedSoundMenuH, OneOctaveItem, NumOctaves == 1);
	CheckItem(AdvancedSoundMenuH, ThreeOctavesItem, NumOctaves == 3);
	CheckItem(AdvancedSoundMenuH, MajorMode, CurrentMode == MajorMode);
	CheckItem(AdvancedSoundMenuH, MinorMode, CurrentMode == MinorMode);
	CheckItem(AdvancedSoundMenuH, ChromaticMode, CurrentMode == ChromaticMode);
	CheckItem(AdvancedSoundMenuH, ContinuousMode, CurrentMode == ContinuousMode);

/*
	DisableItem(ViewMenuH, UseFullImpactMapItem);
	DisableItem(ViewMenuH, UseFullImpactPhasesItem);
*/

	DrawMenuBar();
	HiliteMenu(0);

	NeedToSave = NeedToSaveTS = NeedToSaveBD = NeedToSaveBA = FALSE;
}

		/* --------------------------------------------------------- */
		/* ---------- Specific Menu-Manipulation Routines ---------- */
		/* --------------------------------------------------------- */

AdjustBBMenus()
{
	WindowPeek	wp = (WindowPeek) FrontWindow();
	Boolean		DA = IS_SYS_WINDOW(wp);
	
	ToggleItemEnabled(EditMenuH, UndoItem, DA);
	ToggleItemEnabled(EditMenuH, CutItem, DA);
/*	Don't adjust Copy
	ToggleItemEnabled(EditMenuH, CopyItem, DA);
*/
	ToggleItemEnabled(EditMenuH, PasteItem, DA);
	ToggleItemEnabled(EditMenuH, ClearItem, DA);

	ToggleItemEnabled(ViewMenuH, ClearPhaseSpaceItem, WindowOpen(PhaseSpace));
	ToggleItemEnabled(ViewMenuH, ClearTimeSeriesItem, AnyTSText());

	if (OPTION_DOWN)
		SetItem(WindowsMenuH, CloseWindowItem, "\PClose All Windows");
	else
		SetItem(WindowsMenuH, CloseWindowItem, "\PClose Window");

	ToggleItemEnabled(WindowsMenuH, CloseWindowItem, wp != NIL);
	while (TRUE) {
		if ( (wp == NIL) || (!wp->visible) ) {
			ToggleItemEnabled(WindowsMenuH, ArrangeWindowsItem, FALSE);
			break;
		}
		if (WindowIndex((WindowPtr) wp) != NoWindow) {
			ToggleItemEnabled(WindowsMenuH, ArrangeWindowsItem, TRUE);
			break;
		}
		wp = wp->nextWindow;
	}

/*
	ToggleItemEnabled(ViewMenuH, UseFullImpactMapItem, IsMagnified(ImpactMap));
	ToggleItemEnabled(ViewMenuH, UseFullImpactPhasesItem, IsMagnified(ImpactPhases));
*/

/*
	ToggleItemEnabled(SpecialMenuH, StartBifDiagItem, DrawBD);
	ToggleItemEnabled(SpecialMenuH, PauseBifDiagItem, DrawBD);
	ToggleItemEnabled(SpecialMenuH, ClearBifDiagItem, DrawBD);
	ToggleItemEnabled(SpecialMenuH, StartBasinsAttrItem, DrawBA);
	ToggleItemEnabled(SpecialMenuH, PauseBasinsAttrItem, DrawBA);
	ToggleItemEnabled(SpecialMenuH, ClearBasinsAttrItem, DrawBA);
*/
}

CheckUseTimeSeries(Boolean theUseTimeSeries)
{
	CheckItem(ViewMenuH, UseTimeSeriesItem, UseTimeSeries);
}

ControlMenuStart(int PS)
{
	switch (PS) {
		case BIFURCATION_DIAGRAM:
/*
				SetItem(SpecialMenuH, StartBifDiagItem, "\PStop Bifurcation Diagram");
*/
				SetItem(SpecialMenuH, StopContBifDiagItem, "\PStop Bifurcation Diagram");
				EnableItem(SpecialMenuH, StopContBifDiagItem);
				break;
		case BASINS_OF_ATTRACTION:
/*
				SetItem(SpecialMenuH, StartBasinsAttrItem, "\PStop Basins of Attraction");
*/
				SetItem(SpecialMenuH, StopContBasinsAttrItem, "\PStop Basins of Attraction");
				EnableItem(SpecialMenuH, StopContBasinsAttrItem);
				break;
		default:
/*
				SetItem(ControlMenuH, BeginIterationItem, "\PStop Iteration");
*/
				SetItem(ControlMenuH, StopContinueItem, "\PStop Iteration");
				EnableItem(ControlMenuH, StopContinueItem);
				DisableItem(ControlMenuH, DoOneImpactItem);
				break;
	}
}

ControlMenuStop(int PS, Boolean OKToCont)
{
	switch (PS) {
		case BIFURCATION_DIAGRAM:
/*
				SetItem(SpecialMenuH, StartBifDiagItem, "\PStart Bifurcation Diagram");
*/
				SetItem(SpecialMenuH, StopContBifDiagItem, (OKToCont) ?
						"\PContinue Bifurcation Diagram" : "\PStop Bifurcation Diagram");
				ToggleItemEnabled(SpecialMenuH, StopContBifDiagItem, OKToCont);
				break;
		case BASINS_OF_ATTRACTION:
				SetItem(SpecialMenuH, StartBasinsAttrItem, "\PStart Basins of Attraction");
				SetItem(SpecialMenuH, StopContBasinsAttrItem, (OKToCont) ?
						"\PContinue Basins of Attraction" : "\PStop Basins of Attraction");
				ToggleItemEnabled(SpecialMenuH, StopContBasinsAttrItem, OKToCont);
				break;
		default:
				SetItem(ControlMenuH,BeginIterationItem,"\PBegin Iteration");
				SetItem(ControlMenuH, StopContinueItem, (OKToCont) ?
						"\PContinue Iteration" : "\PStop Iteration");
				ToggleItemEnabled(ControlMenuH, StopContinueItem, OKToCont);
/*
				ToggleItemEnabled(ControlMenuH, DoOneImpactItem, OKToCont);
*/
				ToggleItemEnabled(ControlMenuH, DoOneImpactItem, TRUE);
				break;
	}
}

		/* ----------------------------------- */
		/* ---------- Menu Handling ---------- */
		/* ----------------------------------- */

HandleAppleMenu(int menuItem)
{
	switch (menuItem) {
		case 1:
			DoAboutBBDialog();
			break;
		default:
			HandleDeskAcc(AppleMenuH, 0, menuItem);
			break;
	}	/* switch menuItem from case AppleMenuID */
}

Boolean DoSave()
{
	Boolean		GoodOp, HadToSaveAs;
	int			theWindex;

	theWindex = DecideWhichSaveWindow();
			/* GoodOp used differently here from elsewhere*/
	GoodOp = TRUE;
	if (!SaveFileExists()) {
		GoodOp = SaveAs();
		HandleAllUpdates();
		HadToSaveAs = TRUE;
	}
	else HadToSaveAs = FALSE;
	if (GoodOp) {
		if (SaveFile()) {
			if (HadToSaveAs && (theWindex == TxOut))
				if (!SpecialGraphing())
					SetTextTitle(TRUE, TxOut);
				else
					SetANilFileName(TxOut);
			switch (theWindex) {
				case TxOut:	NeedToSave = FALSE;		break;
				case TSer:	NeedToSaveTS = FALSE;	break;
				case BifD:	NeedToSaveBD = FALSE;	break;
				case BAttr:	NeedToSaveBA = FALSE;	break;
			}
		}
		else {
			if (HadToSaveAs)
				SetANilFileName(theWindex);
			GoodOp = FALSE;
		}
	}
	return(GoodOp);
}

Boolean CheckOneSave(Boolean Quitting, int windex)
{
	Boolean		WantsToSave = FALSE, NTS;
	Boolean		GoodOp;
	WindowPtr	wp;
	int			FrontWindex;

	switch (windex) {
		case TSer:	NTS = NeedToSaveTS;	break;
		case BifD:	NTS = NeedToSaveBD;	break;
		case BAttr:	NTS = NeedToSaveBA;	break;
		default:	NTS = NeedToSave;
					windex = TxOut;
					break;
	}

	if (NTS) {
		if (Quitting) {
			if (!DoQuitWarningDialog(&WantsToSave, windex))
				return(FALSE);
		}
		else if (!DoSaveWarningDialog(&WantsToSave, windex))
			return(FALSE);
		if (WantsToSave) {
			wp = FrontWindow();
			SelectWindowIndex(windex);
			GoodOp = DoSave();
			if (wp != NIL) {
				FrontWindex = WindowIndex(wp);
				if (FrontWindex != NoWindow)
					SelectWindowIndex(FrontWindex);
				else
					SelectWindow(wp);
			}
			return(GoodOp);
		}
	}

	return(TRUE);
}

Boolean CheckNormalSaves(Boolean Quitting)
{
	if (!CheckOneSave(Quitting, TxOut))	return(FALSE);
	if (!CheckOneSave(Quitting, TSer))	return(FALSE);

	return(TRUE);
}

Boolean CheckAllSaves(Boolean Quitting)
{
	if (!CheckOneSave(Quitting, TxOut))	return(FALSE);
	if (!CheckOneSave(Quitting, TSer))	return(FALSE);
	if (!CheckOneSave(Quitting, BifD))	return(FALSE);
	if (!CheckOneSave(Quitting, BAttr))	return(FALSE);

	return(TRUE);
}

HandleFileMenu(int menuItem)
{
	int			theWindex;

	switch (menuItem) {
		case NewItem:
						if (!CheckAllSaves(FALSE))	return;
						ResetSolutions();
						ProgramStatus = BIFURCATION_DIAGRAM;	StopIterating(FALSE);
						ProgramStatus = BASINS_OF_ATTRACTION;	StopIterating(FALSE);
						ProgramStatus = NORMAL_ITERATION;		StopIterating(FALSE);
						NewGraphics(TRUE, TRUE);
						NewSpecial(BifD);
						NewSpecial(BAttr);
						NeedToSaveBD = NeedToSaveBA = FALSE;
/* put in NewGraphics(TRUE)
						SetTextTitle(FALSE, TxOut);
						NeedToSave = NeedToSaveTS = FALSE;
*/
/*
						NeedToSaveBD = NeedToSaveBA = FALSE;
*/
						break;
		case OpenItem:
						if (GetOpenFileName(&theWindex)) {
							if (!CheckOneSave(FALSE, theWindex))
								return;

/*	2.01 addition
 */
							SetOpenFileName(theWindex);
/*	End, 2.01 addition
 */

							if (theWindex == TxOut)
								if (!CheckOneSave(FALSE, TSer))
									return;
							if ( (theWindex == BifD) || (theWindex == BAttr) ) {
								if (OpenFile())
									OpenSpecialFile(theWindex);
							}
							else
								if (StartOpeningFile())
									IterationStatus = ITERATING;
						}
						break;
		case SaveItem:
						if (!DoSave()) ;
						break;
		case SaveAsItem:
						theWindex = DecideWhichSaveWindow();
						if ( !SaveAs() )		return;
						HandleAllUpdates();
						if ( !SaveFile() )
							ResetFileName();
						else switch (theWindex) {
							case TxOut:	if (!SpecialGraphing())
											SetTextTitle(TRUE, TxOut);
										else
											SetANilFileName(TxOut);
										NeedToSave = FALSE;		break;
							case TSer:	NeedToSaveTS = FALSE;	break;
							case BifD:	NeedToSaveBD = FALSE;	break;
							case BAttr:	NeedToSaveBA = FALSE;	break;
						}
						break;
		case SaveDefaultsItem:
						SaveDefaults();
						break;
		case PageSetupItem:
						DoPageSetup();
						break;
		case PrintItem:
						PrintFrontWindow();
						break;
		case QuitItem:
						if (CheckAllSaves(TRUE))
							_HALT();
						break;
	}
}

HandleEditMenu(int menuItem)
{
	WindowPeek	wp = (WindowPeek) FrontWindow();
	int			windex;

	if (IS_SYS_WINDOW(wp) && (menuItem<=ClearItem))
		HandleSysEdit(menuItem);
	else {
		windex = WindowIndex((WindowPtr) wp);
		switch (menuItem) {
			case CopyItem:	HandleCopy(windex);		break;
			case SuppressSaveItem:
					SuppressSaveWarnings = !SuppressSaveWarnings;
					CheckItem(EditMenuH, menuItem, SuppressSaveWarnings);
					break;
			case AutoCommandKeyItem:
					AutoCommandKey = !AutoCommandKey;
					CheckItem(EditMenuH, menuItem, AutoCommandKey);
					break;
			case BackgroundOperationItem:
					if (DoBackgroundDialog(&RunInBackground,
							&FasterBackground, &UseSoundInBackground))
						BackgroundSleep =
								(FasterBackground) ? DAY_DREAM : CATNAP;
					break;
			default:
					break;
		}
	}
}

HandleControlMenu(int menuItem)
{
	Boolean	GoodValues;

	switch (menuItem) {
		case BeginIterationItem:
/*
			if ( (ProgramStatus == WAITING) ||
					(ProgramStatus >= BIFURCATION_DIAGRAM) ) {
*/
				pCurrentVars = &NormalVars;
				GoodValues = Initialize_Iteration_Globals(TRUE, FALSE);
				NewGraphics(ClearSolutions, TRUE);
				if (ClearSolutions)
					ResetSolutions();
				if (!GoodValues) return;
				StartIterating(NORMAL_ITERATION);
				NeedToSave = TRUE;
/*
			}
			else {
				if (ProgramStatus == PLAYBACK) {
					(* SetHitNumber(RealHitNumber) *);
					NeedToSave = TRUE;
				}
				StopIterating(HitNumber>=0); return;
			}
*/
			break;
		case StopContinueItem:
			if (OKToContinue) {
				if (!Reset_Iteration_Globals()) {
					StopIterating(FALSE);
					return;
				}
				if (StartIterating(NORMAL_ITERATION)) {
					IterationStatus = ITERATING;
					NeedToSave = TRUE;
				}
			}
			else {
				PAUSING = !PAUSING;
				ActivateTextScrolls(PAUSING);
				if (PAUSING) {
					SetItem(ControlMenuH, menuItem, "\PContinue Iteration");
					EnableItem(ControlMenuH, DoOneImpactItem);
				}
				else {
					SetItem(ControlMenuH, menuItem, "\PStop Iteration");
					DisableItem(ControlMenuH, DoOneImpactItem);
				}
			}
			break;
		case DoOneImpactItem:
			SingleStepOn = TRUE;
			if ( (ProgramStatus != NORMAL_ITERATION) && (!OKToContinue) )
				HandleControlMenu(BeginIterationItem);
			else
				HandleControlMenu(StopContinueItem);
			break;
		case PlaybackItem:
			if (InitPlayback())
				IterationStatus = ITERATING;
			break;
		case InitialConditionsItem:
			if (GetInitialConditions(TRUE))
				BeginIfAppropriate();
			break;
		case TableParametersItem:
			if (GetTableParameters())
				BeginIfAppropriate();
			break;
		case GetInitFromLastItem:
			GetInitialFromLast();
			if (GetInitialConditions(FALSE))
				BeginIfAppropriate();
			break;
		case IterationItem:
			GetIterationParams();
			break;
/*
		case RunInBackgroundItem:
			if (!RunInBackground)
				if ( !DoRunInBackgroundDLog(&BackgroundSleep,
						&SuppressBackgroundSound) )
					break;
			RunInBackground = !RunInBackground;
			CheckItem(ControlMenuH, RunInBackgroundItem, RunInBackground);
			break;
*/
	}
}

/*
HandleParametersMenu(int menuItem)
{
	switch (menuItem) {
		case InitialConditionsItem:
				if (GetInitialConditions(TRUE))
					BeginIfAppropriate();
				break;
		case TableParametersItem:
				if (GetTableParameters())
					BeginIfAppropriate();
				break;
		case GetInitFromLastItem:
				GetInitialFromLast();
				if (GetInitialConditions(FALSE))
					BeginIfAppropriate();
				break;
		case IterationItem:
				GetIterationParams();
				break;
	}
}
*/

HandleViewMenu(int menuItem)
{
	switch (menuItem) {
		case PhaseRangeItem:
								GetPhaseRange();
								break;
		case MaxVelocityItem:
								GetNewVMax();
								break;
		case SimulationSpeedItem:
								GetSimulationSpeed();
								break;
		case TrajectoryScalingItem:
								GetTrajScalingParams();
								break;
		case ScrollTrajectoryItem:
								SwitchScrollTrajectory();
								break;
		case ClearImpactsItem:
								SetImpactMapStart(HitNumber+1);
								break;
		case ShowAllImpactsItem:
								SetImpactMapStart(0);
								break;
		case ImpactPhasesOffsetItem:
								GetImpactPhasesOffset();
								break;
		case MagnifyImpactMapItem:
								GetIMMagnification(FALSE);
								break;
/*
		case UseFullImpactMapItem:
								SetFullIMapRange();
								break;
*/
		case MagnifyImpactPhasesItem:
								GetIPMagnification(FALSE);
								break;
/*
		case UseFullImpactPhasesItem:
								SetFullIPhsRange();
								break;
*/
		case FasterPhaseSpaceItem:
								CheckItem(ViewMenuH, menuItem, ToggleFasterPhaseSpace());
								break;
		case ClearPhaseSpaceItem:
								ClearPhaseSpace();
								break;
		case UseTimeSeriesItem:
								ToggleUseTimeSeries(TRUE);
								break;
		case TimeSeriesItem:
								GetTimeSeriesInfo();
								break;
		case ClearTimeSeriesItem:
								ClearTimeSeries();
								break;
	}
}

HandleSpecialMenu(int menuItem)
{
	switch (menuItem) {
/*
		case StartBifDiagItem:
						if ( (ProgramStatus != BIFURCATION_DIAGRAM) ) {
							if (!CheckOneSave(FALSE, BifD))
								return;
							StartBifurcationDiagram();
						}
						else
							StopIterating(TRUE);
						break;
		case PauseBifDiagItem:
						if (OKToContinueBD)
							ContinueBifurcationDiagram();
						else {
							PAUSING = !PAUSING;
							ActivateTextScrolls(PAUSING);
							if (PAUSING)
								SetItem(SpecialMenuH, menuItem, "\PContinue Bifurcation Diagram");
							else SetItem(SpecialMenuH, menuItem, "\PPause Bifurcation Diagram");
						}
						break;
*/
		case StartBifDiagItem:
						if (!CheckOneSave(FALSE, BifD))
							return;
						StartBifurcationDiagram();
						break;
		case StopContBifDiagItem:
						if (OKToContinueBD)
							ContinueBifurcationDiagram();
						else {
							PAUSING = !PAUSING;
							ActivateTextScrolls(PAUSING);
							if (PAUSING)
								SetItem(SpecialMenuH, menuItem, "\PContinue Bifurcation Diagram");
							else {
								if (!WindowOpen(BifD))
									if (DoShowWindowDialog(BifD))
										BBOpenWindow((WindowPtr) w[BifD].Peek);
								SetItem(SpecialMenuH, menuItem, "\PStop Bifurcation Diagram");
							}
						}
						break;
		case BifurcationDiagramItem:
						GetBifDiagInfo(FALSE);
						break;
		case BifDiagPhaseRangeItem:
						GetSpecialPhaseRange(BifD);
						break;
		case ClearBifDiagItem:
/*
						if (DoSpecialClearDialog("\Pbifurcation")) {
*/
						if (CheckOneSave(FALSE, BifD)) {
							ClearBifurcationDiagram();
							if (ProgramStatus == BIFURCATION_DIAGRAM)
								StopIterating(FALSE);
						}
						break;
		case GraphInAllWindowsBDItem:
						SwitchGraphInAll(BifDiagram);
						break;
		case UseColorBDItem:
						SwitchInColor(BifDiagram);
						break;
/*
		case StartBasinsAttrItem:
						if ( (ProgramStatus != BASINS_OF_ATTRACTION) ) {
							if (!CheckOneSave(FALSE, BAttr))
								return;
							StartBasinsOfAttraction();
						}
						else
							StopIterating(TRUE);
						break;
		case PauseBasinsAttrItem:
						if (OKToContinueBA)
							ContinueBasinsOfAttraction();
						else {
							PAUSING = !PAUSING;
							ActivateTextScrolls(PAUSING);
							if (PAUSING)
								SetItem(SpecialMenuH, menuItem, "\PContinue Basins Of Attraction");
							else SetItem(SpecialMenuH, menuItem, "\PPause Basins Of Attraction");
						}
						break;
*/
		case StartBasinsAttrItem:
						if (!CheckOneSave(FALSE, BAttr))
							return;
						StartBasinsOfAttraction();
						break;
		case StopContBasinsAttrItem:
						if (OKToContinueBA)
							ContinueBasinsOfAttraction();
						else {
							PAUSING = !PAUSING;
							ActivateTextScrolls(PAUSING);
							if (PAUSING)
								SetItem(SpecialMenuH, menuItem, "\PContinue Basins Of Attraction");
							else {
								if (!WindowOpen(BAttr))
									if (DoShowWindowDialog(BAttr))
										BBOpenWindow((WindowPtr) w[BAttr].Peek);
								SetItem(SpecialMenuH, menuItem, "\PStop Basins Of Attraction");
							}
						}
						break;
		case BasinsOfAttractionItem:
						GetBasAttrInfo(FALSE);
						break;
		case BasinsAttrPhaseRangeItem:
						GetSpecialPhaseRange(BAttr);
						break;
		case ClearBasinsAttrItem:
/*
						if (DoSpecialClearDialog("\Pbasins of attraction")) {
*/
						if (CheckOneSave(FALSE, BAttr)) {
							ClearBasinsOfAttraction();
							if (ProgramStatus == BASINS_OF_ATTRACTION)
								StopIterating(FALSE);
						}
						break;
		case GraphInAllWindowsBAItem:
						SwitchGraphInAll(BasinsAttraction);
						break;
		case UseColorBAItem:
						SwitchInColor(BasinsAttraction);
						break;
		case PatternsItem:
						GetSolutionPatterns();
						break;
		case ColorsItem:
						GetSolutionColors();
						break;
		case PrintPatternListItem:
						PrintPatternList();
						break;
	}
}

SwitchSoundMenu()
/*	Attempts to fix Jeremiah's highlighting bug:
 *	1. BB 1.5a HiliteMenu at end -- didn't work
 *	2. BB 2.0a HiliteMenu at beginning -- worked!
 */
{
	UseAdvancedSound = !UseAdvancedSound;
HiliteMenu(0);
	if (UseAdvancedSound) {
		DeleteMenu(SimpleSoundMenuID);
		SoundMenuH = AdvancedSoundMenuH;
		InsertMenu(AdvancedSoundMenuH, WindowsMenuID);
	}
	else {
		DeleteMenu(AdvancedSoundMenuID);
		SoundMenuH = SimpleSoundMenuH;
		InsertMenu(SimpleSoundMenuH, WindowsMenuID);
	}
/*
	DrawMenuBar();
*/
/*	HiliteMenu(0) here only did not work
	HiliteMenu(0);
*/
/*
	DrawMenuBar();
*/
}

HandleSoundMenu(int menuItem)
{
	switch (menuItem) {
		case PhaseTonesItem:
		case VelocityTonesItem:
			FlipTonesOn(menuItem);
			break;
		case OneOctaveItem:		/* or AdvancedOptionsItem	*/
			if (!UseAdvancedSound) {
				SwitchSoundMenu();
				break;
			}
			SetNumOctaves(1);
			CheckItem(SoundMenuH, menuItem, TRUE);
			CheckItem(SoundMenuH, menuItem+1, FALSE);
			break;
		case ThreeOctavesItem:
			SetNumOctaves(3);
			CheckItem(SoundMenuH, menuItem, TRUE);
			CheckItem(SoundMenuH, menuItem-1, FALSE);
			break;
		case MajorMode:
		case MinorMode:
		case ChromaticMode:
		case ContinuousMode:
			SetToneMode(menuItem);
			break;
		case SimpleOptionsItem:
			SwitchSoundMenu();
			break;
	} /* switch menuItem from SoundMenuID */
}

HandleWindowsMenu(int menuItem)
{
	switch (menuItem) {
/*
		case EnlargeWindowItem:
				if (WindowIndex(FrontWindow()) != NoWindow)
					MakeLargeWindow(FrontWindow());
				break;
*/
		case CloseWindowItem:
				if (OPTION_DOWN)	BBCloseAllWindows();
				else				if (BBCloseFrontWindow()) ;
				break;
		case ArrangeWindowsItem:
				ArrangeWindows();
				break;
		case DefaultConfigItem:
				DefaultWindowConfiguration();
				break;
		default:
				SelectWindowIndex(menuItem-WindowSeparatorItem-1);
				break;
	}
}

_MENU(int menuID, int menuItem)
{
	switch (menuID) {
		case AppleMenuID:		HandleAppleMenu(menuItem);			break;
		case FileMenuID:		HandleFileMenu(menuItem);			break;
		case EditMenuID:		HandleEditMenu(menuItem);			break;
		case ControlMenuID:		HandleControlMenu(menuItem);		break;
/*
		case ParametersMenuID:	HandleParametersMenu(menuItem);		break;
*/
		case ViewMenuID:		HandleViewMenu(menuItem);			break;
		case SpecialMenuID:		HandleSpecialMenu(menuItem);		break;
		case SimpleSoundMenuID:
		case AdvancedSoundMenuID:
								HandleSoundMenu(menuItem);			break;
		case WindowsMenuID:		HandleWindowsMenu(menuItem);		break;
	}
}

HandleMenu(long theMenu)
{
	if (HiWord(theMenu) != 0)
		_MENU(HiWord(theMenu), LoWord(theMenu));
}

HandleKeyDown(long EventMessage, int Modifiers)
{
	char	c;
	long	MenuSelection;

	if (!AutoCommandKey) {
		if ( !(c = IsCommandKey(EventMessage, Modifiers)) )	return;
	}
	else
		c = LoWord(EventMessage & charCodeMask);

	if (ProgramStatus == OPENING) {
		if (c == DOT) {
			StopOpeningFile(FALSE);
			StopIterating(HitNumber>=0);
		}
		return;
	}
	MenuSelection = MenuKey(c);
	if (HiWord(MenuSelection) == 0) {
		switch (c) {
			case DOT:
				if (PAUSING)	break;
				switch (ProgramStatus) {
					case WAITING:
						break;
					case BIFURCATION_DIAGRAM:
						MenuSelection = ( ((long) SpecialMenuID) << 16 ) +
								StopContBifDiagItem;
						break;
					case BASINS_OF_ATTRACTION:
						MenuSelection = ( ((long) SpecialMenuID) << 16 ) +
								StopContBasinsAttrItem;
						break;
					default:
						MenuSelection = ( ((long) ControlMenuID) << 16 ) +
								StopContinueItem;
						break;
				}
				break;
/*
			case 'g':
				if (ProgramStatus != WAITING) break;
				MenuSelection = ( ((long) ControlMenuID) << 16 ) + BeginIterationItem;
				break;
*/
		}
	}

	HandleMenu(MenuSelection);
}

/*
_KEY(int c, int mods)
{
	if (!(mods & cmdKey)) return;
	HandleMenu(MenuKey(c));
}
*/
