/*
 *	bbPhysics.c
 *
 *	Physics-related routines for Bouncing Ball
 *
 */

#include "bbPrototypes.h"

#define	MaxIterations	1024

	/* Remaining constants all used to determine dt */
#define	MinDivisor		2					/* Rascal was eff. 10 */
#define	StepsPerCycle	32					/* Rascal was 128 */
#define	Sqrt3Over2		(1.73205080757/2.0)

				/*	Global external vars from bb.c	*/
extern	pIterationVars		pCurrentVars;

double	Tolerance;

SetTolerance(int theAccuracy)
{
	Tolerance = pow(10, -theAccuracy);
}

double GetStopTime(long cycles)
{
	return(Period * cycles);
}

SetFrequencyParams()
{
	Period = (freq==0) ? DEFAULT_PERIOD : 1.0/freq;
	Omega = 2 * PI * freq;
}

/*
double xTableFixed(double tt);
double xTableFixed(double tt)
{
	double	thePhase;
	Fixed	phase_fxd;
	int		sign = 1;

	thePhase = fmod(Omega*tt+2.0*PI*delta, 2.0*PI);
	if (thePhase > PI) {
		sign = -1;
		thePhase = 2.0*PI-thePhase;
	}
	if (thePhase > PI/2.0)	thePhase = PI-thePhase;
	phase_fxd = X2Fix(&thePhase);
	return( A * (sign*Frac2X(FracSin(phase_fxd)) + 1.0) );
}
*/

double xTable(double tt)
	/* table position; s(t)=A (sin(omega t+2 delta)+1) */
{
	return( A * (sin(Omega*tt+2.0*PI*delta) + 1.0) );
}

double vTable(double tt)
	/* table velocity; w(t)=A omega cos(omega t+2 delta) */
{
	return( A * Omega * cos(Omega*tt+2.0*PI*delta) );
}

double aTable(double tt)
	/* table acceleration; a(t)=-A omega^2 sin(omega t+2 delta) */
{
	return( -A*Omega*Omega * sin(Omega*tt+2.0*PI*delta) );
}

double xBall(tt, Xi, Vi) /* ball position, Xi+Vi tt-1/2 g tt^2 */
	double	tt, Xi, Vi;
{
	return( Xi + Vi*tt - 1.0/2.0*G*tt*tt );
}

double vBall(double tt, double Vi)	 /* ball velocity, Vi-g tt */
{
	return( Vi - G*tt );
}

double tFromXTable(double x, Boolean InPrimaryRange)
{
	double	tt;

	tt = (asin(x/A - 1.0) - 2.0*PI*delta ) / Omega;
	if (!InPrimaryRange)	tt += Period/2.0;
	tt = fmod(tt+Period+Period, Period);
	return(tt);
}

double FindDistance(Tau, tt, Xi, Vi)	/* distance between ball and wall */
	double	Tau, tt, Xi, Vi;
{
	return( xBall(tt,Xi,Vi) - xTable(Tau) );
}

Boolean	impact(double distance)
{
	return( distance <= 0.0 );
}

Boolean	FindNStep(p_step, Ti, Xi, Vi)
	double	*p_step, Ti, Xi, Vi;
	/*	Find iterative step size using *N*ick Tufillaro's method */
{
	double	tmpVDiff, tmpADiff, tmax;
	int		i;
	double	Tau, T;

	*p_step = Period/StepsPerCycle;
	if ( (tmpVDiff = vBall(0.0,Vi)-vTable(Ti)) <= 0.0 ) return(FALSE);
/*
	if (tmpVDiff < AssumedZero)
		if (tmpADiff <= 0.0) return(FALSE);
*/
	if ( (tmpADiff = aTable(Ti) + G) != 0.0)
		tmax = fabs(2.0*tmpVDiff/tmpADiff/MinDivisor);
	else	tmax = *p_step;
	if (tmax < Period*Tolerance) return(FALSE);
	*p_step = MIN(tmax, *p_step);

		/* If no root on interval messages appear, put Nick's kludge here */
		/*	(Moved from InitNextIteration()	-TA 6/12/91)	*/
		/*	here it is */
/*	StuckStatus would be stuck only if return(FALSE) above
	if (StuckStatus == NOT_STUCK)
*/
		T = *p_step;	Tau = Ti + T;
		for(i=0; impact(FindDistance(Tau, T, Xi, Vi)); ++i) {
			*p_step /= 10.0;
			if (*p_step < Period*Tolerance)
				return(FALSE);
			T = *p_step; Tau = Ti + T;
/*	Here keep going until *p_step < Period*Tolerance
			if (i>10) {
				ERROR("\PError: Can't calculate dt");
				if (ProgramStatus >= BIFURCATION_DIAGRAM)
					NeedNextSpecialValue = TRUE;
				return(FALSE);
			}
*/
		}

	return(TRUE);
}

double	FindTStep(distance, Tau, tt, Xi, Vi)
	double	distance, Tau, tt, Xi, Vi;
	/*
	 *	As of 3/90, use several methods.
	 *
	 *	Method 0) only used if A or freq == 0 (no table motion); picks
	 *	step to reflect ball across parabola, then moves halfway to impact.
	 *
	 *	Method 1) if ball above max table height*1.01, figures out step
	 *	such that ball will be just (1%) above max table height, on its
	 *	way down.
	 *
	 *	Method 2) step = (-b + sqrt(b^2+2 g distance))/g,
	 *	b = g tt - Vi + Aw
	 *	If table moved up from time tt at velocity Aw (max. table vel.),
	 *	would meet ball at tt + step
	 *
	 *	Method 3) uses a quadratic approximation of sin, one for each
	 *	quadrant, to calculate approximate next solution.
	 *	If would reach beyond end of quadrant, shifts and does
	 *	calculations from there.
	 *	The quadratic approximations are, by quadrant:
	 *		1)	(0phi<PI/2)		sin(phi) ->PI/2*phi	- phi^2/2
	 *		2)	(PI/2phi<PI)		sin(phi) -> 	  1	- phi^2/PI
	 *		3)	(PIphi<3 PI/2)		sin(phi) ->    -phi	+ phi^2/PI
	 *		4)	(3 PI/2phi<2 PI)	sin(phi) -> 	-1	+ phi^2/2
	 *	These four were then rewritten about a phi0 in the respective ranges.
	 *
	 *	Method 4) For what would normally be very small step size,
	 *	restricts iteration to one segment (1/18 currently) of a cycle, and
	 *	uses, instead of max table velocity Aw above, max table velocity
	 *	on the interval in question.  Returns step at which table would
	 *	meet ball or step to end of interval, whichever is smaller.
	 *	(3/91: Subject to minimum of 1/32 (currently) Period--Safe???)
	 */
	/*	Finds iteration step size using *T*yler Abbott's method */
{
	double	b, aq, bq, discriminant;
	double	phase0to1, LocalPhase, MaxTableVel;
	int		sign;
	int		PhaseArea;
	int		Quadrants = 4, NumSmallDivisions = 18, NumTinyDivisions = 32;
	double	Step1, Step2, Step3, Step4, theStep, MaxStep, tmpStep;
	double	t3, d3; /* time and distance for method 3 */

		/*	---------- Method 0 ----------	*/

	if ( (A==0.0) || (freq==0.0) )
		if ( vBall(tt, Vi) > 0 )
			return( 2.0*Vi/G - tt - tt + tt/2.0 );
				/*	reflects across parabola, moves halfway to next hit */
				/*	Should I return 0 if vBall < 0?	*/

		/*	---------- Method 1 ----------	*/
	if ( xBall(tt, Xi, Vi) > 2*A*1.01 ) {
		Step1 = (Vi + sqrt(Vi*Vi - 2.0*G*(2*A*1.01 - Xi)))/G - tt;
		return(Step1);
	}

		/*	---------- Method 2 ----------	*/

	phase0to1 = fmod(Tau+delta*Period, Period)/Period;

		/*	The first number below is where Method 3's slope is greater
		 *	than Method 2's, for phi < PI/2 - 1. The second is the first
		 *	point on the curve of Method 3's approximation a straight
		 *	line going out at Method 2's slope would eventually
		 *	re-intersect Method 3's curve;
		 *	phi = 1 - sqrt(PI^2/2 - 2 PI + 2)
		 *	The numbers have been normalized to 0 to 1.
		 */
	if ( (phase0to1 < 0.0909) || (phase0to1 > 0.7806) ) {
		b = G*tt - Vi + A*Omega;
		Step2 = (-b + sqrt(b*b + 2.0*G*distance))/G;
	}
	else
		Step2 = 0.0;
/*			Removed when allowed Method 3 to continue into next quadrant.
	if (Step2 > Period/Quadrants)
		return(Step2);
*/

		/*	---------- Method 3 ----------	*/

	LocalPhase = fmod( phase0to1, (1.0)/((double) Quadrants) ) *2.0*PI;
	PhaseArea = (int) (phase0to1 * Quadrants);
	Step3 = 0.0;	t3 = tt;		d3 = distance;
	for(;;) {
		aq = -G/2.0;
		bq = vBall(t3, Vi);
		switch(PhaseArea) {
			case 0:
			case 4:								/*	Should never be 4 */
					aq += A*Omega*Omega/2.0;
					bq -= A*Omega*(PI/2.0 - LocalPhase);
					break;
			case 1:
					aq += A*Omega*Omega/PI;
					bq += 2.0*A*Omega*LocalPhase/PI;
					break;
			case 2:
					aq -= A*Omega*Omega/PI;
					bq += A*Omega*(1.0 - 2.0*LocalPhase/PI);
					break;
			case 3:
					aq -= A*Omega*Omega/2.0;
					bq -= A*Omega*LocalPhase;
					break;
		}
		discriminant = sqrt(bq*bq - 4.0*aq*d3);
		if (discriminant > 0.0) {
			if (aq>0)	sign = ( (-bq - discriminant) > 0) ? -1 :  1;
			else		sign = ( (-bq + discriminant) < 0) ?  1 : -1;
			tmpStep = (-bq + sign*discriminant) / (2.0*aq);
		}
		else
			tmpStep == 0.0;	/* really beyond MaxStep */
		MaxStep = ( (1.0)/((double) Quadrants) - (LocalPhase/2.0/PI) +
				AlmostZero /* to assure makes next quadrant */ )*Period;
		if ( (tmpStep > MaxStep) || (discriminant == 0.0) ) {
			Step3 += MaxStep;	t3 += MaxStep;
			d3 = aq*MaxStep*MaxStep + bq*MaxStep + d3;
			LocalPhase = 0.0;
			PhaseArea = (PhaseArea + 1) % Quadrants;
			continue;
		}
		else {
			Step3 += tmpStep;
			break;
		}
	}

	theStep = MAX(Step2, Step3);
	if (theStep > Period/NumSmallDivisions)
		return(theStep);

		/*	---------- Method 4 ----------	*/

		/*	Probably in a near-sticking situation;make more divisions	*/

		/*	Cheating here; not absolutely safe, i.e., could overshoot	*/
		/*	bounds, but should be OK. -TA 7/24/90 						*/
/*
	NumSmallDivisions = (int) (Period/ndt) / 20;
	if (NumSmallDivisions == 0) return(ndt);
*/
	PhaseArea = (int) (phase0to1 * NumSmallDivisions);
	b = G*tt - Vi +
			A*Omega*cos((PhaseArea+(PhaseArea<NumSmallDivisions/2)?1.0:0.0)/
			((double) NumSmallDivisions) * 2*PI );
	Step4 = (-b + sqrt(b*b + 2.0*G*distance))/G;
	MaxStep = ((PhaseArea+1.0)/((double) NumSmallDivisions) -
			phase0to1+AlmostZero) * Period;
	Step4 = MIN(Step4, MaxStep);
	return( MAX(Period/NumTinyDivisions, MAX(Step4, theStep)) );
}

/*
int	MaxCount = 0;
*/

Boolean	root(p_TheRoot, a, b, Xi, Vi, Ti)
	double *p_TheRoot, a, b, Xi, Vi, Ti;
{
	double	m;
	double	da, db, dm;
	int		count = 0;
	Boolean	wasA;

	da = FindDistance(Ti+a,a,Xi,Vi);
	db = FindDistance(Ti+b,b,Xi,Vi);

	while(TRUE) {
		if (++count > MaxIterations)
			return(FALSE); /* put in error string later? */
		if(da*db > 0.0)
			return(FALSE); /* Should not happen after through loop once? */

/*
if (count > MaxCount)	MaxCount = count;
*/
		m = a + ( da*(b-a) / ( fabs(da) + fabs(db) ) );
			/* used to be m = (a+b)/2; now does straight-line interpolation */
		dm = FindDistance(Ti+m,m,Xi,Vi);

		if ( fabs(dm) < NON_ZERO_A*Tolerance )
			{	*p_TheRoot = m; return(TRUE);	}
		else {
			if (da*dm<0.0)	{	b = m; db = dm;	}
			else	{	a = m; da = dm;	}
		}
				/*	Probably approaching asymptotically, so go in 1/10,	*/
				/*	then cut distance in half (m = (a+b)/2)				*/
		if (count >= 4) {
			wasA = (m==a);
			if (wasA)	m = ((9.0*a) + b) / 10.0;
			else		m = ((9.0*b) + a) / 10.0;
			dm = FindDistance(Ti+m,m,Xi,Vi);
/*
++count;
*/
			if ( fabs(dm) < NON_ZERO_A*Tolerance )
				{	*p_TheRoot = m; return(TRUE);	}
			else {
				if (da*dm<0.0)	{	b = m; db = dm;	}
				else	{	a = m; da = dm;	}
			}

			if (wasA && (b==m))		continue;
			if (!wasA && (a==m))	continue;

			m = (a + b) / 2;
			dm = FindDistance(Ti+m,m,Xi,Vi);
/*
++count;
*/
			if ( fabs(dm) < NON_ZERO_A*Tolerance )
				{	*p_TheRoot = m; return(TRUE);	}
			else {
				if (da*dm<0.0)	{	b = m; db = dm;	}
				else	{	a = m; da = dm;	}
			}
		}
	} /* while */
}

double SetVj(double Tj, double tt, double Vi)
{
	return( (1+damp) * vTable(Tj) - damp * ((A==0.0) ? -Vi : vBall(tt, Vi)) );
}

int	FindEjection(p_ndt, p_dt, p_tj, p_xj, p_vj)
	double	*p_ndt, *p_dt, *p_tj, *p_xj, *p_vj;
{
	double	MaxAccel, t_alpha, t_beta, t_ph;

	if ( (MaxAccel = A*Omega*Omega) <= G ) return(NoEjection); /* stuck forever */

	t_alpha = asin(G/MaxAccel) / Omega;
	t_beta = Period/2.0 - t_alpha;
	*p_ndt = ( (t_alpha<t_beta) ? (t_beta-t_alpha) : Period ) / StepsPerCycle;
	t_ph = fmod(*p_tj+delta*Period, Period);

	if ( !( (t_ph>t_alpha) && (t_ph<t_beta) ) ) {
		*p_tj += ( (t_ph>t_alpha) ? Period : 0.0 ) + (t_alpha - t_ph);
		*p_xj = xTable(*p_tj);
		*p_vj = vTable(*p_tj);
		*p_dt = t_beta - t_alpha;
		return(NormalEjection);
	}
	else {
		*p_dt = t_beta - t_ph;
		return(InAreaOfEjection);
	}
		/* only alpha (damping) = 0, hits in ejection area
		 * will be ejected immediately */
}
