/*	bbSpecial.c
 *
 *	Routines for Bifurcation Diagram, Basin of Attraction
 *
 *		-TA 2/8/91
*/

#include <OffPort.h>
#include <MenuUtils.h>
#include "bbPrototypes.h"
#include "bbMenus.h"
#include "bbGraphics.h"
#include "bbSpecial.h"

#define	SolutionPatsName	"\PEight Grays"

				/*	From bb.c	*/
extern	IterationVars	NormalVars, *pCurrentVars;
extern	hBBPreferences	hBBPrefs;

extern	int				ProgramStatus, PAUSING;
extern	Boolean			UsingColorMac;

				/*	From bbGraphics.c	*/
extern	Windows			*pw;

				/*	From bbMenus.c	*/
extern	MenuHandle		SpecialMenuH, WindowsMenuH;

				/*	From bbBifDiag.c	*/
extern	BifDiagRecord		BD_Rec;
extern	Boolean				DrawBD;
extern	int					**BD_Points;
extern	Boolean				NeedToColorAfter;
extern	Str255				BD_Spaces;

				/*	From bbBasinAttr.c	*/
extern	BasAttrRecord		BA_Rec;
extern	Boolean				DrawBA;
extern	Point				**BA_Points;
extern	Str255				BA_Spaces;

				/*	From bbPhysics.c	*/
extern	double				Tolerance;

				/*	Shared with bbIteration.c, for OKToGraph(),
				 *	and bb.c, for StartIterating()	*/
Boolean				*pBDGraphInAllWindows, *pBAGraphInAllWindows;
				/*	Shared with bb.c, for SetDefaults()	*/
Boolean				*pBDUseColor, *pBAUseColor;
				/*	Shared with bbIteration.c, for InitNextIteration()	*/
Boolean				NeedNextSpecialValue;
				/*	Shared with bbBifDiag.c, bbBasinAttr.c	*/
Boolean				NeedToTurnOnGraphing;

Pattern				SolutionPats[NumSolutionTypes];
RGBColor			SolutionColors[NumSolutionTypes];
long				ColorLongs[NumSolutionTypes];



			/*	--------------------------------------------------	*/
			/*	----------	Special Utility Routines	----------	*/
			/*	--------------------------------------------------	*/

SeeIfStopSpecial(int windex)
{
	if (windex < BifDiagram)					return;

	if (PAUSING)								return;

	switch (windex) {
		case BifD:
			if (ProgramStatus != BIFURCATION_DIAGRAM)	return;
			if (DoSpecialCloseStopDialog("\Pbifurcation"))
				HandleSpecialMenu(StopContBifDiagItem);
			break;
		case BAttr:
			if (ProgramStatus != BASINS_OF_ATTRACTION)	return;
			if (DoSpecialCloseStopDialog("\Pbasins of attraction"))
				HandleSpecialMenu(StopContBasinsAttrItem);
			break;
	}
}

Boolean SpecialGraphing()
/*	Returns TRUE only if special program status and graphing in all windows	*/
{
	switch (ProgramStatus) {
		case BIFURCATION_DIAGRAM:	return(*pBDGraphInAllWindows);
		case BASINS_OF_ATTRACTION:	return(*pBAGraphInAllWindows);
		default:					return(FALSE);
	}
}

int	GetSpecialTimes(int theTimes, int DefaultTimes)
{
	return ((theTimes != AutoTimes) ? theTimes : DefaultTimes);
}

SpecialMenuEnable(int windex, Boolean TurnOn, Boolean PauseOK)
{
	int		MenuOffset;
	Boolean	TurnOnPhaseRange;

	MenuOffset = ((windex==BifD) ? StartBifDiagItem : StartBasinsAttrItem ) -
			StartBifDiagItem;
	ToggleItemEnabled(SpecialMenuH, MenuOffset + StartBifDiagItem, TurnOn);
	ToggleItemEnabled(SpecialMenuH, MenuOffset + ClearBifDiagItem, TurnOn);
	ToggleItemEnabled(SpecialMenuH, MenuOffset + StopContBifDiagItem,
			TurnOn && PauseOK);

	TurnOnPhaseRange = TurnOn;
	if ( (windex == BifD) && (!BD_Rec.AllPhases) )
		TurnOnPhaseRange = FALSE;
	if ( (windex == BAttr) && (!BA_Rec.AllPhases) )
		TurnOnPhaseRange = FALSE;
	ToggleItemEnabled(SpecialMenuH, MenuOffset + BifDiagPhaseRangeItem,
			TurnOnPhaseRange);
}

CheckGraphInAll(int windex)
{
	if (windex == BifD)
		CheckItem(SpecialMenuH, GraphInAllWindowsBDItem,
			BD_Rec.GraphInAllWindows);
	if (windex == BAttr)
		CheckItem(SpecialMenuH, GraphInAllWindowsBAItem,
			BA_Rec.GraphInAllWindows);
}

SwitchGraphInAll(int windex)
{
	Boolean	*pGIAW;
	int		AppropriatePS;

	switch (windex) {
		case BifD:	pGIAW = &BD_Rec.GraphInAllWindows;
					AppropriatePS = BIFURCATION_DIAGRAM;
					break;
		case BAttr:	pGIAW = &BA_Rec.GraphInAllWindows;
					AppropriatePS = BASINS_OF_ATTRACTION;
					break;
	}

	if (*pGIAW) {
		*pGIAW = FALSE;
		CheckGraphInAll(windex);
		if (ProgramStatus == AppropriatePS) {
			SetHeaders(NormalVars._A, NormalVars._freq, NormalVars._damp);
			NewGraphics(TRUE, FALSE);
			ActivateTextScrolls(TRUE);
			SetTextTitle(FALSE, TxOut);
		}
	}
	else {
		if (ProgramStatus == AppropriatePS) {
			DoGraphInAllAlert(windex);
			NeedToTurnOnGraphing = TRUE;
		}
		else {
			*pGIAW = TRUE;
			CheckGraphInAll(windex);
		}
	}
}

CheckSpecialColor(int windex)
{
	if (windex == BifD)
		CheckItem(SpecialMenuH, UseColorBDItem, BD_Rec.InColor);
	if (windex == BAttr)
		CheckItem(SpecialMenuH, UseColorBAItem, BA_Rec.InColor);
}

SwitchInColor(int windex)
{
	Boolean	*pInColor;
	int		AppropriatePS;

	switch (windex) {
		case BifD:	pInColor = &BD_Rec.InColor;
					AppropriatePS = BIFURCATION_DIAGRAM;
					break;
		case BAttr:	pInColor = &BA_Rec.InColor;
					AppropriatePS = BASINS_OF_ATTRACTION;
					break;
	}

	*pInColor = !(*pInColor);
	CheckSpecialColor(windex);
	StdForceUpdate((WindowPtr) w[windex].Peek);
	if (*pInColor) {
		if ((windex==BifD) && (ProgramStatus == AppropriatePS))
			NeedToColorAfter = TRUE;
	}
	else
		if (windex==BAttr)
			DrawBAComposite();
}

SetSpecialTitle(Boolean FileExists, int windex)
{
	char	*pfname;
	Str255	oTitle;

	if (FileExists)
		pfname = GetSaveFName(windex);
	else {
		switch (windex) {
			case BifD:	pfname = "\PBifurcation Diagram";
						break;
			case BAttr:	pfname = "\PBasins of Attraction";
						break;
		}
		SetANilFileName(windex);
	}

	SetItem(WindowsMenuH, WindowSeparatorItem + windex + 1, pfname);
	GetWTitle((WindowPtr) w[windex].Peek, CP(oTitle));
	if (sComp(pfname, CP(oTitle)) != 0)
		SetWTitle(w[windex].Peek, pfname);
}

StringAddVar(char *s, double f, char *label, Boolean MorePrecision)
{
	Str255	fstring;
	int		fw = SpecialFieldWidth + ((MorePrecision) ? 1 : 0);

	sAppend(s, label);
	PadFtoStr(f, CP(fstring), fw, (fw-1), TRUE);
	sAppend(s, CP(fstring));
}

StringAddVarLn(char *s, double f, char *label, Boolean MorePrecision)
{
	Str255	fstring;

	StringAddVar(s, f, label, MorePrecision);
	sInsertC(s, ReturnChar, label[0]+1);
}

Boolean CreateSpecialOffPort(GrafPtr *pgp, Rect *pr)
{
	Rect	tr;

	tr = *pr;
	OffsetRect(&tr, -tr.left, -tr.top);
	if (*pgp != NOPORT)
		DisposeOffPort(*pgp);
	*pgp = NewOffPort(&tr);
	if (*pgp == NIL) {
		*pgp = NOPORT;
		return(FALSE);
	}
	SetPort(*pgp);
	SetOrigin(pr->left, pr->top);
	ClipRect(&((*pgp)->portRect));
	EraseRect(&((*pgp)->portRect));

	return(TRUE);
}

GetSpecialPhaseRange(int windex)
{
	int	otmpMinPhaseInt, tmpMinPhaseInt, *pMinPhaseInt;

	switch (windex) {
		case BifD:
			otmpMinPhaseInt = tmpMinPhaseInt =
					RoundToInt(BD_Rec.MinPhase * PHASE_INT_RANGE);
			if (DoPhaseRangeDialog(&tmpMinPhaseInt, windex))
				if (tmpMinPhaseInt != otmpMinPhaseInt)
					ChangeBDPhaseRange(((double) tmpMinPhaseInt)/PHASE_INT_RANGE);
			break;
		case BAttr:
			otmpMinPhaseInt = tmpMinPhaseInt =
					RoundToInt(BA_Rec.MinPhase * PHASE_INT_RANGE);
			if (DoPhaseRangeDialog(&tmpMinPhaseInt, windex))
				if (tmpMinPhaseInt != otmpMinPhaseInt)
					ChangeBAPhaseRange(((double) tmpMinPhaseInt)/PHASE_INT_RANGE);
			break;
	}
}

HandleSpecialHotSpot(int windex, int WhichHotSpot)
{
	switch (windex) {
		case BifDiagram:
			if (WhichHotSpot == 1) {
				if (TestItemEnabled(SpecialMenuH, BifDiagPhaseRangeItem)) {
					GetSpecialPhaseRange(windex);
					break;
				}
				else
					GetBifDiagInfo(FALSE);
			}
			else
				GetBifDiagInfo(FALSE);
			break;
		case BasinsAttraction:
			if (WhichHotSpot == 0) {
				if (TestItemEnabled(SpecialMenuH, BasinsAttrPhaseRangeItem)) {
					GetSpecialPhaseRange(windex);
					break;
				}
				else
					GetBasAttrInfo(FALSE);
			}
			else
				GetBasAttrInfo(FALSE);
			break;
	}
}


#define	ProbableTolerance			1e-2
#define	AlternatingTolerance		3e-3
#define	MonotonicTolerance			5e-5
#define	SomewhatRegularTolerance	1e-7
#define	AnyOldTolerance				1e-9

Boolean	TestAdjacents(double y0, double y1, double y2, double y3,
		int per, int k, int InfoType, Boolean IsAlternating)
{
	int		AdjacentK;
	double	AdjacentY0, AdjacentY1;
	Boolean	AdjacentsOK;
	double	AdjacentY2, AdjacentY3;
	double	tmpY1, tmpAY1;

	AdjacentK = k+per/2;
	AdjacentY0 = GetSolutionInfo(HitNumber - 0*per - AdjacentK, InfoType);
	AdjacentY1 = GetSolutionInfo(HitNumber - 1*per - AdjacentK, InfoType);
	AdjacentY2 = GetSolutionInfo(HitNumber - 2*per - AdjacentK, InfoType);
	AdjacentY3 = GetSolutionInfo(HitNumber - 3*per - AdjacentK, InfoType);
	if (IsAlternating) {
		AdjacentsOK = TRUE;
		tmpY1 = MIN(y1, y2);
		tmpAY1 = MIN(AdjacentY1, AdjacentY2);
		AdjacentsOK = AdjacentsOK &&
				( MAX(tmpY1, tmpAY1) <= MAX(y0, AdjacentY0) );
		tmpY1 = MAX(y1, y2);
		tmpAY1 = MAX(AdjacentY1, AdjacentY2);
		AdjacentsOK = AdjacentsOK &&
				(MIN(tmpY1, tmpAY1) >= MIN(y0, AdjacentY0));

		tmpY1 = MIN(y2, y3);
		tmpAY1 = MIN(AdjacentY2, AdjacentY3);
		AdjacentsOK = AdjacentsOK &&
				( MAX(tmpY1, tmpAY1) <= MAX(y1, AdjacentY1) );
		tmpY1 = MAX(y2, y3);
		tmpAY1 = MAX(AdjacentY2, AdjacentY3);
		AdjacentsOK = AdjacentsOK &&
				(MIN(tmpY1, tmpAY1) >= MIN(y1, AdjacentY1));
/*
		AdjacentsOK = (MAX(y1, MIN(AdjacentY1, AdjacentY2)) <= MAX(y0, AdjacentY0)) &&
			(MIN(y1, AdjacentY1) >= MIN(y0, AdjacentY0));
*/
	}
	else
		AdjacentsOK = (MAX(y1, AdjacentY1) <= MAX(y0, AdjacentY0)) &&
			(MIN(y1, AdjacentY1) >= MIN(y0, AdjacentY0));
	return(AdjacentsOK);
}

Boolean TestPeriodicity(int per, Boolean *pWillBePeriodic,
		Boolean *pProbablyPeriodic)
	/*	y0-3 are the last 4 phasees then velocities, 0 most recent.
	 *	Periodic if:
	 *	1.	a.	y0 between y1 & y2, & y1 between y2 & y3; and
	 *		b.	diff, y0-y2, < diff, y1-y3 (deleted); and
	 *		c.	diff, y0-y1, < AlternatingTolerance; and
	 *		d.	diffs, y1-y3 and y0-y2, < AlternatingTolerance; or
	 *	2.	a.	series y0-y3 monotonic; and
	 *		b.	successive differences shrinking;
	 *		c.	diff, y0-y1, < MonotonicTolerance (if not,
	 *			sets *pWillBePeriodic), subject to (3c); or
	 *	3.	a.	successive differences shrinking;
	 *		b.	diff, y2-y3, very small; and
	 *		c.	if even periodicity, adjacent branches do not
	 *			approach one another.
	 *	4.	a.	diff, y0-y1, y1-y2, y2-y3, extremely small; and
	 *		b.	if even periodicity, adjacent branches do not
	 *			approach one another.
	 *	NOTE: For test 3b, I feel like there should be an additional
	 *	test, i.e., ok to approach one another as long as not converging
	 *	to one another (in practice, converging much more slowly than
	 *	the points in each branch). For instance, as far as I know,
	 *	though I haven't seen it, it could approach a period two solution
	 *	monotonically from the *outside* of the solution.
	 *	ALSO: Problems with phase wrap-arounds.
	 *												-TA 3/1-2/91, 5/20/91
	 */
{
	int		k, j;
	int		InfoType;
	double	y0, y1, y2, y3;
	double	d0, d1, d2;
	double	scale;
	Boolean	IsPeriodic;
	Boolean	tmpWillBePeriodic, MaybeWillBePeriodic = TRUE;
	Boolean	tmpProbablyPeriodic, MaybeProbablyPeriodic = TRUE;
	Boolean	ThisOnePeriodic = TRUE;
	Boolean	NeedWillBe = !*pWillBePeriodic;
	Boolean	NeedProbable = !*pProbablyPeriodic;

	if (GetSolutionInfo(HitNumber, GET_TIME) -
			GetSolutionInfo(HitNumber-per, GET_TIME) < Period/2.0)
		return(FALSE);		/*	sticking; this is a loose test	*/

	for (k=0; k<per; ++k) {
		for (j=0; j<=1; ++j) {
			IsPeriodic = FALSE;
			InfoType = (j==0) ? GET_PHASE : GET_VELOCITY;
			tmpWillBePeriodic = FALSE;
			tmpProbablyPeriodic = FALSE;
			y0 = GetSolutionInfo(HitNumber - 0*per - k, InfoType);
			y1 = GetSolutionInfo(HitNumber - 1*per - k, InfoType);
			y2 = GetSolutionInfo(HitNumber - 2*per - k, InfoType);
			y3 = GetSolutionInfo(HitNumber - 3*per - k, InfoType);
			d0 = y0-y1;
			d1 = y1-y2;
			d2 = y2-y3;
			scale = (j==0) ? 1.0 : MAX(y0,1);
	
				/*	Alternating Convergence	*/
			if (BETWEEN(y0, y1, y2))
				if (BETWEEN(y1, y2, y3))
/* unnecessary
					if ( fabs(y3-y1) > fabs(y2-y0) )
*/
						if (fabs(d2)/scale < AlternatingTolerance) {
							if (fabs(y3-y1)/scale < AlternatingTolerance)
								if (fabs(y2-y0)/scale < AlternatingTolerance)
/*
									IsPeriodic = TRUE;
*/
									if ((per%2 == 0) && (k<per/2) ) {
										if (TestAdjacents(y0, y1, y2, y3, per, k, InfoType, TRUE)) {
											IsPeriodic = TRUE;
										}
									}
									else
										IsPeriodic = TRUE;
						}
						else tmpProbablyPeriodic = TRUE;
	
				/*	Monotonic Convergence	*/
			if (!IsPeriodic)
				if ( (d0*d1>=0) && (d1*d2>=0) )
					if ( (fabs(d0) < fabs(d1)) &&
							(fabs(d1) < fabs(d2)) ) {
/*
						if ((per%2 == 0) && (k>=per/2) ) {
*/
						if ((per%2 == 0) && (k<per/2) ) {
							if (TestAdjacents(y0, y1, y2, y3, per, k, InfoType, FALSE)) {
								tmpWillBePeriodic = TRUE;
								IsPeriodic = (fabs(d2)/scale < MonotonicTolerance);
							}
						}
						else {
							tmpWillBePeriodic = TRUE;
							IsPeriodic = (fabs(d2)/scale < MonotonicTolerance);
						}
					}
					else if ( fabs(d0-d1) < fabs(d2-d1) ) {
						if ((per%2 == 0) && (k<per/2) ) {
							if (TestAdjacents(y0, y1, y2, y3, per, k, InfoType, FALSE)) {
								tmpWillBePeriodic = TRUE;
							}
						}
						else {
							tmpWillBePeriodic = TRUE;
						}
					}
	
				/*	Somewhat Regular Convergence	*/
			if (!IsPeriodic)
				if ( (fabs(d0) < fabs(d1)) && (fabs(d1) < fabs(d2)) )
					if ( (fabs(d2/scale) < SomewhatRegularTolerance) ) {	/*asymptotic--needed?*/
						if ((per%2 == 0) && (k<per/2) ) {
							if (TestAdjacents(y0, y1, y2, y3, per, k, InfoType, FALSE))
								IsPeriodic = TRUE;
						}
						else
							IsPeriodic = TRUE;
					}

				/*	Any Old Convergence	*/
			if (!IsPeriodic)
				if ( (fabs(d0/scale) < AnyOldTolerance) &&
						(fabs(d1/scale) < AnyOldTolerance) &&
						(fabs(d2/scale) < AnyOldTolerance) ) {	/*asymptotic--needed?*/
					if ((per%2 == 0) && (k<per/2) ) {
						if (TestAdjacents(y0, y1, y2, y3, per, k, InfoType, FALSE))
							IsPeriodic = TRUE;
					}
					else
						IsPeriodic = TRUE;
				}

			/*	Probably Periodic?	*/
			if ( (!IsPeriodic) && (!tmpWillBePeriodic) &&
					(!tmpProbablyPeriodic) )
				if ( (fabs(d0/scale) < ProbableTolerance) &&
						(fabs(d1/scale) < ProbableTolerance) &&
						(fabs(d2/scale) < ProbableTolerance) )
					tmpProbablyPeriodic = TRUE;
	
			MaybeProbablyPeriodic = ( MaybeProbablyPeriodic &&
					(tmpProbablyPeriodic || tmpWillBePeriodic || IsPeriodic) );
			MaybeWillBePeriodic = ( MaybeWillBePeriodic &&
					(tmpWillBePeriodic || IsPeriodic) );
			ThisOnePeriodic = (ThisOnePeriodic && IsPeriodic);
			if (!ThisOnePeriodic && (!MaybeWillBePeriodic || !NeedWillBe) &&
					(!MaybeProbablyPeriodic || !NeedProbable)) {
				*pWillBePeriodic = MaybeWillBePeriodic;
				*pProbablyPeriodic = MaybeProbablyPeriodic;
				return(FALSE);
			}
		}
	}
	*pProbablyPeriodic = MaybeProbablyPeriodic;
	*pWillBePeriodic = MaybeWillBePeriodic;
	return(ThisOnePeriodic);
}


	/*	File operations	*/

PicHandle MakeSpecialPicture(GrafPtr gp)
{
	PicHandle	pic;
	Rect		DstRect = gp->portRect;
	RgnHandle	OldClipRgn;

	SetPort(gp);
	OldClipRgn = NewRgn();
	GetClip(OldClipRgn);

	pic = OpenPicture(&DstRect);
		ClipRect(&gp->portRect);
		CopyBits(&gp->portBits, &gp->portBits, &DstRect, &DstRect, 0L, srcCopy);
	ClosePicture();
	SetClip(OldClipRgn);
	DisposeRgn(OldClipRgn);

	return(pic);
}

Boolean SaveSpecialFile(int theWindex)
{
	Boolean	GoodOp;

	switch (theWindex) {
		case BifD:	GoodOp = SaveBDFile();	break;
		case BAttr:	GoodOp = SaveBAFile();	break;
	}
	SetSpecialTitle(GoodOp, theWindex);
	return(GoodOp);
}

Boolean OpenSpecialFile(int theWindex)
{
	Boolean	GoodOp;

	switch (theWindex) {
		case BifD:	GoodOp = OpenBDFile();	break;
		case BAttr:	GoodOp = OpenBAFile();	break;
	}
	CloseOpenFile();
/*
	SetSpecialTitle(GoodOp, theWindex);
*/
	return(GoodOp);
}


			/*	--------------------------------------------------	*/
			/*	----------	General-Purpose Routines	----------	*/
			/*	--------------------------------------------------	*/

		/*	----------	Solution Patterns, Colors Routines	----------	*/

Boolean LoadSolutionPatterns()
{
	Handle	PH;
	int		i;

	if ( (PH = GetNamedResource(PatternListResType, SolutionPatsName)) == NIL ) {
		for(i = 0; i < NumSolutionTypes; ++i)
			BlockEquate(CP(SolutionPats[i]), CP(ltGray), sizeof(Pattern));
		BlockEquate(CP(SolutionPats[0]), CP(black), sizeof(Pattern));
		BlockEquate(CP(SolutionPats[0]), CP(dkGray), sizeof(Pattern));
		BlockEquate(CP(SolutionPats[0]), CP(gray), sizeof(Pattern));
		BlockEquate(CP(SolutionPats[NumSolutionTypes-1]), CP(white), sizeof(Pattern));
		return(FALSE);
	}
	HLock(PH);
	BlockEquate(CP(SolutionPats), *PH+2, NumSolutionTypes*sizeof(Pattern));
	HUnlock(PH);
	ReleaseResource(PH);
	return(TRUE);
}

GetSolutionPatterns()
{
	SolutionInfo	tsi;
	int				i;

	for(i=0; i<NumSolutionTypes; ++i)
		tsi[i] = psi[i];
	if (DoSolutionInfoDialog(FALSE)) {
		for(i=0; i<NumSolutionTypes; ++i)
			if(tsi[i] != psi[i])
				break;
		if (i<NumSolutionTypes) {
			if (DrawBA && !BA_Rec.InColor) {
				DrawBAComposite();
				StdForceUpdate((WindowPtr) w[BAttr].Peek);
			}
		}
	}
	else
		for(i=0; i<NumSolutionTypes; ++i)
			psi[i] = tsi[i];
}

SetSpecialColor(int i, Boolean UseRGBValues)
{
	GrafPtr	gp;

	if (UseRGBValues)
		RGBForeColor(&SolutionColors[i]);
	else {
		GetPort(&gp);
		gp->fgColor = ColorLongs[i];
	}
}

InitSpecialColors()
{
	int 		i=0;
	GrafPtr		curPort;
	WindowPtr	wp;
	Rect		r = {50, 50, 60, 60};

	GetPort(&curPort);
	wp = NewWindow(0L, &r, "\p", false, noGrowDocProc, -1L, true, 0L);

	ColorLongs[i++] = blackColor;	ColorLongs[i++] = magentaColor;
	ColorLongs[i++] = redColor;		ColorLongs[i++] = yellowColor;
	ColorLongs[i++] = greenColor;	ColorLongs[i++] = cyanColor;
	ColorLongs[i++] = blueColor;	ColorLongs[i++] = whiteColor;

	SetPort(wp);
	for(i=0; i<NumSolutionTypes; ++i) {
		wp->fgColor = ColorLongs[i];
		GetForeColor(&SolutionColors[i]);
	}
/*
	wp->fgColor = blackColor;	GetForeColor(&SolutionColors[i++]);
	wp->fgColor = magentaColor;	GetForeColor(&SolutionColors[i++]);
	wp->fgColor = redColor;		GetForeColor(&SolutionColors[i++]);
	wp->fgColor = yellowColor;	GetForeColor(&SolutionColors[i++]);
	wp->fgColor = greenColor;	GetForeColor(&SolutionColors[i++]);
	wp->fgColor = cyanColor;	GetForeColor(&SolutionColors[i++]);
	wp->fgColor = blueColor;	GetForeColor(&SolutionColors[i++]);
	wp->fgColor = whiteColor;	GetForeColor(&SolutionColors[i++]);
*/

	SetPort(curPort);
	DisposeWindow(wp);	/*	?	*/
}

GetSolutionColors()
{
	SolutionInfo	tsi;
	int				i;

	for(i=0; i<NumSolutionTypes; ++i)
		tsi[i] = csi[i];
	if (DoSolutionInfoDialog(TRUE)) {
		for(i=0; i<NumSolutionTypes; ++i)
			if(tsi[i] != csi[i])
				break;
		if (i<NumSolutionTypes) {
			if (DrawBA && BA_Rec.InColor)
				StdForceUpdate((WindowPtr) w[BAttr].Peek);
			if (DrawBD && BD_Rec.InColor)
				StdForceUpdate((WindowPtr) w[BifD].Peek);
		}
	}
	else
		for(i=0; i<NumSolutionTypes; ++i)
			csi[i] = tsi[i];
}

SetSpecialControls(int windex)
{
	Rect	WholeRect, IntersectRect, NullRect = {0,0,0,0};
	Rect	*pLocalGR = (windex==BifD) ? &BD_Rec.LocalGR : &BA_Rec.LocalGR;
	int		hOffset = GetCtlValue(w[windex].BSB);
	int		vOffset = GetCtlValue(w[windex].RSB);

	GetStdDocWindowClipRect(&WholeRect, (WindowPtr) w[windex].Peek);
	if ( ((windex==BifD) && (!DrawBD)) || ((windex==BAttr) && (!DrawBA)) ) {
		ResizeControl(w[windex].HS[0], &WholeRect);
		ResizeControl(w[windex].HS[1], &NullRect);
		*pLocalGR = NullRect;
		return;
	}

	if (windex==BifD)	IntersectRect = BD_Rec.HS0_Rect;
	else				IntersectRect = BA_Rec.HS0_Rect;
	OffsetRect(&IntersectRect, -hOffset, -vOffset);
	IntersectRect.top += 1;
	IntersectRect.right += 1;
	SectRect(&WholeRect, &IntersectRect, &IntersectRect);
	ResizeControl(w[windex].HS[0], &IntersectRect);

	if (windex==BifD)	IntersectRect = BD_Rec.HS1_Rect;
	else				IntersectRect = BA_Rec.HS1_Rect;
	OffsetRect(&IntersectRect, -hOffset, -vOffset);
	IntersectRect.bottom += 1;
	SectRect(&WholeRect, &IntersectRect, &IntersectRect);
	ResizeControl(w[windex].HS[1], &IntersectRect);

	IntersectRect = w[windex].GR;
	OffsetRect(&IntersectRect, -hOffset, -vOffset);
	SectRect(&WholeRect, &IntersectRect, &IntersectRect);
/*
	if (windex==BifD)
		IntersectRect.right += 1;
*/
/*
	IntersectRect.bottom += 1;
*/
	*pLocalGR = IntersectRect;
/*
	tmpRect.bottom = w[BifD].Peek->port.portRect.bottom - (ScrollBarWidth-1);
*/
}

Boolean FindSpecialClick(int windex, Point pt)
{
	Rect	tmpRect;

	if (windex==BifD)
		tmpRect = BD_Rec.LocalGR;
	else
		tmpRect = BA_Rec.LocalGR;
	++tmpRect.right;
	++tmpRect.bottom;
	return(PtInRect(pt, &tmpRect));
}

	/*	Scrolling Routines	*/

SetSpecialScrolling(int windex)
{
	Rect	MaxRect;
	int		theMax;

	switch (windex) {
		case BifD:	if (!DrawBD)
						MaxRect = w[windex].Peek->port.portRect;
					else
						MaxRect = BD_Rec.WindowSize;
					OffsetRect(&MaxRect, -MaxRect.left, -MaxRect.top);
					break;
		case BAttr:	if (!DrawBA)
						MaxRect = w[windex].Peek->port.portRect;
					else
						MaxRect = BA_Rec.WindowSize;
					OffsetRect(&MaxRect, -MaxRect.left, -MaxRect.top);
					break;
	}

	theMax = MAX(MaxRect.bottom - w[windex].Peek->port.portRect.bottom, 0);
	SetCtlValue(w[windex].RSB, MIN(GetCtlValue(w[windex].RSB), theMax));
	SetCtlMax(w[windex].RSB, theMax);

	theMax = MAX(MaxRect.right - w[windex].Peek->port.portRect.right, 0);
	SetCtlValue(w[windex].BSB, MIN(GetCtlValue(w[windex].BSB), theMax));
	SetCtlMax(w[windex].BSB, theMax);

	SetSpecialControls(windex);
	if (WindowIndex(FrontWindow()) == windex)
		BBHiliteControls(windex, TRUE);
}

ScrollCopyBackRect(int windex, Rect *pr, Boolean ToWindow)
{
	GrafPtr	gp;
	Rect	tr;
	int		hOffset;
	int		vOffset;

	if (ToWindow) {
		hOffset = GetCtlValue(w[windex].BSB);
		vOffset = GetCtlValue(w[windex].RSB);
		SetPort(w[windex].Peek);
	}
	else {
		hOffset = 0;
		vOffset = 0;
	}
	GetPort(&gp);
	myCopyBits(w[windex].Back, gp,
			pr, hOffset, vOffset, 0, 0, srcCopy, (RgnHandle) NIL);
	tr = w[windex].GR;
	OffsetRect(&tr, -hOffset, -vOffset);
	if (SectRect(&tr, pr, &tr))
		switch (windex) {
			case BifD:	CopyBDGraph(&tr, hOffset, vOffset);	break;
			case BAttr:	CopyBAGraph(&tr, hOffset, vOffset);	break;
		}
}

ScrollCopyBackToWindow(int windex)
{
	ScrollCopyBackRect(windex, &w[windex].Peek->port.portRect, TRUE);
}

CopySpecialBackToPrint(int windex)
{
	ScrollCopyBackRect(windex, &w[windex].Back->portRect, FALSE);
}

NewSpecial(int windex)
{
	int	i;

	switch (windex) {
		case BifD:
					if (BD_Rec.Graph != NOPORT) {
						SetPort(BD_Rec.Graph);
						EraseRect(&BD_Rec.Graph->portRect);
					}
					if (BD_Rec.CurrentXInt >= 0)
						SetSpecialTitle(FALSE, windex);
					break;
		case BAttr:
					for(i=0; i<NumBAOffPorts; ++i)
						if (BA_Rec.Graphs[i] != NOPORT) {
							SetPort(BA_Rec.Graphs[i]);
							EraseRect(&BA_Rec.Graphs[i]->portRect);
						}
					if (BA_Rec.CurrentXInt >= 0)
						SetSpecialTitle(FALSE, windex);
					break;
	}
	if (!w[windex].WindowCleared)
		StdForceUpdate((WindowPtr) w[windex].Peek);
	w[windex].WindowCleared = TRUE;
}

InitSpecial()
{
	int	i;

	NeedNextSpecialValue = FALSE;

	DrawBD = FALSE;
	w[BifD].DrawFromScratch = TRUE;
	w[BifD].Magnifying = TRUE;
	pBDGraphInAllWindows = &BD_Rec.GraphInAllWindows;
	pBDUseColor = &BD_Rec.InColor;
	BD_Rec.GraphInAllWindows = DefaultBDGIAW;
	BD_Rec.InColor = DefaultBDUseColor;
	CheckGraphInAll(BifD);
	CheckSpecialColor(BifD);
	BD_Rec.Graph = NOPORT;
	BD_Rec.InitConds = NIL;
	BD_Rec.SolutionTypes = NIL;
	BD_Rec.Reserved1 = BD_Rec.Reserved2 = BD_Rec.Reserved3 =
			BD_Rec.Reserved4 = 0L;
	BD_Points = NIL;

	DrawBA = FALSE;
	w[BAttr].DrawFromScratch = TRUE;
	w[BAttr].Magnifying = TRUE;
	pBAGraphInAllWindows = &BA_Rec.GraphInAllWindows;
	pBAUseColor = &BA_Rec.InColor;
	BA_Rec.GraphInAllWindows = DefaultBAGIAW;
	BA_Rec.InColor = DefaultBAUseColor;
	CheckGraphInAll(BAttr);
	CheckSpecialColor(BAttr);
	for(i=0; i<NumBAOffPorts; ++i)
		BA_Rec.Graphs[i] = NOPORT;
	BA_Rec.Reserved1 = BA_Rec.Reserved2 = BA_Rec.Reserved3 =
			BA_Rec.Reserved4 = 0L;
	BA_Points = NIL;

	SpecialMenuEnable(BifD, FALSE, FALSE);
	SpecialMenuEnable(BAttr, FALSE, FALSE);

	SetSpecialControls(BifD);
	SetSpecialControls(BAttr);

	if (!LoadSolutionPatterns())
		ERROR("\PCould not load solution patterns. Will use smaller set.");

	if (UsingColorMac)
		InitSpecialColors();

/*
	for(i=0; i<NumSolutionTypes; ++i) {
		psi[i] = csi[i] = i;
	}
*/

	sEquate(CP(BD_Spaces), "\P   ");
	sEquate(CP(BA_Spaces), CP(BD_Spaces));

}
