/*	bbSpecialDialog.c
 *
 *	Routines which handle Bouncing Ball's "special" Dialog Boxes,
 *	those for Bifurcation Diagram and Basin of Attraction.
 *
 *		-TA, 2/91
 *
 */

#include <ctype.h>
#include <limits.h>
#include <DialogUtils.h>

#include "bbPrototypes.h"
#include "bbGraphics.h"		/*	For "bbSpecial.h" only	*/
#include "bbSpecial.h"

enum {
		GraphInAllAlertID = 371,
		SolutionInfoDLogID = 481,
		BifurcationDiagramDLogID,
		BasinsOfAttractionDLogID,
		BifDiagColorDLogID,
		BasAttrColorDLogID,
		SpecialWarningDLogID,
		SpecialOpenContinueDLogID,
		ClearSpecialDLogID,
		ReplaceSolutionInfoDLogID,
		SpecialCloseStopDLogID
};

enum {
		BD_XAxis = 9,
		BD_YAxis = BD_XAxis + 3,
		BD_FirstET = BD_YAxis + 3,
		BD_FirstXAxis = BD_FirstET,
		BD_FirstYAxis = BD_FirstXAxis + 2,
		BD_OtherTableParam = BD_FirstYAxis + 2,
		BD_InitIter = BD_OtherTableParam + 4,
		BD_ShowIter,
		BD_GraphInAll = 25,
		BD_WindowSize = BD_GraphInAll + 2,
		BD_TableIncString = BD_WindowSize + 2,
		BD_AutoInitial = BD_TableIncString + 4,
		BD_AutoShow,
		BD_StatOtherTableParam = BD_AutoShow + 2,
		BD_Restart = 45,
		BD_InColor,
		BD_Colors
};

enum {
		BA_FirstET = 9,
		BA_FirstPhase = BA_FirstET,
		BA_FirstVelocity = BA_FirstPhase + 2,
		BA_TableParams = BA_FirstVelocity + 2,
		BA_MaxIter = BA_TableParams + 3,
		BA_IncSize,
		BA_WindowSize = 28,
		BA_Faster = BA_WindowSize + 2,
		BA_GraphInAll,
		BA_Restart,
		BA_InColor,
		BA_ColorsOrPatterns,
		BA_AutoMax
};

#define	DLogPrecision			6
#define UserOutlineItemNum		3
#define	MagIMapFirstET			6
#define	DNST					(NumSolutionTypes + 1)

				/*	Global external vars from bb.c	*/
extern	pIterationVars	pCurrentVars;
extern	hBBPreferences	hBBPrefs;
extern	int				CurrentModifiers;
extern	Boolean			UsingColorMac;
extern	Boolean			GettingParamFiles;

				/*	From bbImpactGraphics.c	*/
extern	double			MinPhase, MaxPhase, VMin;

				/*	From bbBifDiag.c	*/
extern	pBifDiagRecord	pbd;

				/*	From bbBasAttr.c	*/
extern	pBasAttrRecord	pba;

				/*	From bbSpecial.c	*/
extern	Pattern			SolutionPats[NumSolutionTypes];
extern	RGBColor		SolutionColors[NumSolutionTypes];

				/*	From bbWindows.c; for DoSpecialCloseStopDialog	*/
extern	Boolean			ClosingAllWindows;


			/*	For Bifurcation Diagram	*/
Str255			AmplitudeStatString, FrequencyStatString;
Str255			AmplitudeIncString, FrequencyIncString;
Str255			AmplitudeString, FrequencyString;
Str255			AutoInitString, AutoString;
Str255			XAxisStrings[ItemsInRect], YAxisStrings[ItemsInRect];
/*
Str255			GraphRangeStrings[ItemsInRect];
*/

			/*	For Basins of Attraction	*/
/*
Str255			AutoMaxString;
*/
Str255			ColorString, PatternString;
DoubleRect		*pNGR;

			/*	For both	*/
Boolean			ChangedDamping;
double			OldDamping;

			/*	For Solution Info (Colors, Patterns)	*/
SolutionInfo	*ptsi;


		/*	------------------------------------------------------	*/
		/*	----------	General Purpose-Type Routines	----------	*/
		/*	------------------------------------------------------	*/

Boolean	DoSpecialWarningDialog(char *ps)
{
	DialogPtr	dp;

	if (SHIFT_DOWN)	return(TRUE);

	dp = GetNewDialog(SpecialWarningDLogID, NIL, -1L);
	ParamText(ps, "\P","\P","\P");
	if ( HandleDialog(dp, NO_EDIT_TEXT, Cancel,
			UserOutlineItemNum, NIL, NIL, NIL) == OK )
		DialogReturn(dp, TRUE)
	else
		DialogReturn(dp, FALSE)
}

Boolean	DoSpecialOpenContinueDialog(char *ps)
{
	DialogPtr	dp;

	dp = GetNewDialog(SpecialOpenContinueDLogID, NIL, -1L);

	ParamText(ps, "\P","\P","\P");
	if ( HandleDialog(dp, NO_EDIT_TEXT, Cancel,
			UserOutlineItemNum, NIL, NIL, NIL) == OK )
		DialogReturn(dp, TRUE)
	else
		DialogReturn(dp, FALSE)
}

Boolean	DoSpecialCloseStopDialog(char *ps)
{
	DialogPtr	dp;

	if (SHIFT_DOWN)		return(TRUE);
	if (!ClosingAllWindows)
		if (OPTION_DOWN)	return(FALSE);

	dp = GetNewDialog(SpecialCloseStopDLogID, NIL, -1L);

	ParamText(ps, "\P","\P","\P");
	if ( HandleDialog(dp, NO_EDIT_TEXT, Cancel,
			UserOutlineItemNum, NIL, NIL, NIL) == OK )
		DialogReturn(dp, TRUE)
	else
		DialogReturn(dp, FALSE)
}

Boolean	DoSpecialClearDialog(char *ps)
{
	DialogPtr	dp;

	if (SHIFT_DOWN)	return(TRUE);

	dp = GetNewDialog(ClearSpecialDLogID, NIL, -1L);
	ParamText(ps, "\P","\P","\P");
	if ( HandleDialog(dp, NO_EDIT_TEXT, Cancel,
			UserOutlineItemNum, NIL, NIL, NIL) != OK )
		DialogReturn(dp, TRUE)
	else
		DialogReturn(dp, FALSE)
}

Boolean ReplaceSolutionInfoDialog(char *ColOrPat, char *SpecialType)
{
	DialogPtr	dp;

	if (!GettingParamFiles) {
		if (SHIFT_DOWN)	return(TRUE);
		if (OPTION_DOWN) return(FALSE);
	}

	dp = GetNewDialog(ReplaceSolutionInfoDLogID, NIL, -1L);
	ParamText(ColOrPat, SpecialType,"\P","\P");
	if ( HandleDialog(dp, NO_EDIT_TEXT, Cancel,
			UserOutlineItemNum, NIL, NIL, NIL) == OK )
		DialogReturn(dp, TRUE)
	else
		DialogReturn(dp, FALSE)
}

DoGraphInAllAlert(int windex)
{
	char	*s0, *s1;

	if (SHIFT_DOWN)	return;

	switch (windex) {
		case BifDiagram:
					s0 = (pbd->XAxisIsAmplitude) ? "\Pthe amplitude" : "\Pthe frequency";
					s1 = "\Ps";
					break;
		case BasinsAttraction:
					s0 = "\Pthe initial conditions";
					s1 = "\P";
					break;
	}
	ParamText(s0, s1, "\P", "\P");
	if (!NoteAlert(GraphInAllAlertID,NIL)) SysBeep(30);
}


		/*	------------------------------------------------------	*/
		/*	----------	Bifurcation Diagram Routines	----------	*/
		/*	------------------------------------------------------	*/

Boolean BifDiagDLogItemFilter(DialogPtr dp, int itemHit)
{
	Str255	s;
	int		i;
	int		index;
	int		NewMin, NewMax, OldMin, OldMax;
	char	*StatStr, *IncStr, *OldStr, *NewStr;
	double	NewDamping;

/*
	if ((ChangedDamping) && (itemHit != BD_OtherTableParam+1) ) {
		ChangedDamping = FALSE;
		if ( myGetItemFloat(dp, BD_OtherTableParam+1, &NewDamping) )
			if (NewDamping != OldDamping) {
				pbd->DefaultIITimes = GetDefaultInitIterTimes(NewDamping);
				LongToString((long) pbd->DefaultIITimes, CP(AutoInitString));
				if (myGetItemCtlValue(dp, BD_AutoInitial) > 0)
					mySetItemText(dp, BD_InitIter, CP(AutoInitString));
				OldDamping = NewDamping;
			}
	}
*/

	switch (itemHit) {
		case BD_XAxis:
		case BD_XAxis+1:
				if (itemHit == BD_XAxis) {
					if (pbd->XAxisIsAmplitude)	break;
					OldMin = ymin_i;	OldMax = ymax_i;
					NewMin = xmin_i;	NewMax = xmax_i;
					StatStr = CP(FrequencyStatString);
					IncStr = CP(AmplitudeIncString);
					OldStr = CP(AmplitudeString);
					NewStr = CP(FrequencyString);
				}
				else {
					if (!pbd->XAxisIsAmplitude)	break;
					OldMin = xmin_i;	OldMax = xmax_i;
					NewMin = ymin_i;	NewMax = ymax_i;
					StatStr = CP(AmplitudeStatString);
					IncStr = CP(FrequencyIncString);
					OldStr = CP(FrequencyString);
					NewStr = CP(AmplitudeString);
				}
				pbd->XAxisIsAmplitude = !pbd->XAxisIsAmplitude;
				i=BD_FirstXAxis;
				myGetItemText(dp, i, CP(XAxisStrings[OldMin]));
				mySetItemText(dp, i, CP(XAxisStrings[NewMin]));
				myGetItemText(dp, ++i, CP(XAxisStrings[OldMax]));
				mySetItemText(dp, i, CP(XAxisStrings[NewMax]));
				mySelItemText(dp, --i);
				mySetItemText(dp, BD_StatOtherTableParam, StatStr);
/*
				mySetItemText(dp, BD_TableIncString, IncStr);
*/
				myGetItemText(dp, BD_FirstET+4, OldStr);
				mySetItemText(dp, BD_FirstET+4, NewStr);
				break;
		case BD_YAxis:
		case BD_YAxis+1:
				if (itemHit == BD_YAxis) {
					if (pbd->YAxisIsPhase)	break;
					OldMin = ymin_i;	OldMax = ymax_i;
					NewMin = xmin_i;	NewMax = xmax_i;
				}
				else {
					if (!pbd->YAxisIsPhase)	break;
					OldMin = xmin_i;	OldMax = xmax_i;
					NewMin = ymin_i;	NewMax = ymax_i;
				}
				pbd->YAxisIsPhase = !pbd->YAxisIsPhase;
				i=BD_FirstYAxis;
				myGetItemText(dp, i, CP(YAxisStrings[OldMin]));
				mySetItemText(dp, i, CP(YAxisStrings[NewMin]));
				myGetItemText(dp, ++i, CP(YAxisStrings[OldMax]));
				mySetItemText(dp, i, CP(YAxisStrings[NewMax]));
				mySelItemText(dp, --i);
				break;
		case BD_OtherTableParam+1:	/*	damping	*/
				ChangedDamping = TRUE;
				break;
		case BD_InitIter:
				if (myGetItemCtlValue(dp, BD_AutoInitial) > 0) {
					myGetItemText(dp, itemHit, CP(s));
					if (sUpperComp(CP(s), CP(AutoString)) != 0)
						mySetItemCtlValue(dp, BD_AutoInitial, FALSE);
				}
				else {
					myGetItemText(dp, itemHit, CP(s));
					if (sUpperComp(CP(s), CP(AutoString)) == 0)
						mySetItemCtlValue(dp, BD_AutoInitial, TRUE);
				}
				break;
		case BD_ShowIter:
				if (myGetItemCtlValue(dp, BD_AutoShow) > 0) {
					myGetItemText(dp, itemHit, CP(s));
					if (sUpperComp(CP(s), CP(AutoString)) != 0)
						mySetItemCtlValue(dp, BD_AutoShow, FALSE);
				}
				else {
					myGetItemText(dp, itemHit, CP(s));
					if (sUpperComp(CP(s), CP(AutoString)) == 0)
						mySetItemCtlValue(dp, BD_AutoShow, TRUE);
				}
				break;
		case BD_AutoInitial:
				if (myGetItemCtlValue(dp, itemHit) == 0) {
					mySetItemText(dp, BD_InitIter, CP(AutoString));
					mySelItemText(dp, BD_InitIter);
				}
				else {
					mySelItemText(dp, BD_InitIter);
					return(TRUE);
				}
				break;
		case BD_AutoShow:
				if (myGetItemCtlValue(dp, itemHit) == 0) {
					mySetItemText(dp, BD_ShowIter, CP(AutoString));
					mySelItemText(dp, BD_ShowIter);
				}
				else {
					mySelItemText(dp, BD_ShowIter);
					return(TRUE);
				}
				break;
		case BD_InColor:
				if (myGetItemCtlValue(dp, itemHit) == 0)
					myEnableCtlItem(dp, BD_Colors);
				else
					myDisableCtlItem(dp, BD_Colors);
				break;
		case BD_Colors:
				GetSolutionColors();
				break;
	}
	return(FALSE);
}

Boolean CheckBifurcationDiagramDialog(DialogPtr dp, int itemNo)
{
	Boolean		NumbersOK = TRUE;
	int			i;
	double		tmpF;
	long		tmpL;
	Str255		ErrorStrings[BD_InitIter-BD_FirstET];
	int			X_Min, X_Max, Y_Min, Y_Max;
	DoubleRect	theGR;

	if (itemNo != OK) return(TRUE);

	i=0;
	sEquate(CP(ErrorStrings[i++]), "\PStarting X-Axis Value");
	sEquate(CP(ErrorStrings[i++]), "\PEnding X-Axis Value");
	sEquate(CP(ErrorStrings[i++]), "\PY-Axis Minimum");
	sEquate(CP(ErrorStrings[i++]), "\PY-Axis Maximum");
	if (pbd->XAxisIsAmplitude)
		sEquate(CP(ErrorStrings[i++]), "\PFrequency");
	else
		sEquate(CP(ErrorStrings[i++]), "\PAmplitude");
	sEquate(CP(ErrorStrings[i++]), "\PDamping Coefficient");
	sEquate(CP(ErrorStrings[i++]), "\PInitial Velocity");
	sEquate(CP(ErrorStrings[i++]), "\PInitial Phase");

	for(i=BD_FirstET; i<BD_InitIter; ++i) {
		if ( !myGetItemFloat(dp, i, &tmpF) ) {
			FixBadValAlert(CP(ErrorStrings[i-BD_FirstET]));
			SelectBadField(dp, i, &NumbersOK);
			return(FALSE);
		}
		switch (i) {
			case BD_FirstXAxis:
					if (tmpF < 0.0) {
						if (pbd->XAxisIsAmplitude)
							ERROR("\PAmplitude must be non-negative.");
						else
							ERROR("\PFrequency must be non-negative.");
						SelectBadField(dp, i, &NumbersOK);
						return(FALSE);
					}
					theGR[xmin_i] = tmpF;
					break;
			case BD_FirstXAxis+1:
/*	Allow to go backwards??? (not <=)	*/
					if (tmpF < 0.0) {
						if (pbd->XAxisIsAmplitude)
							ERROR("\PAmplitude must be non-negative.");
						else
							ERROR("\PFrequency must be non-negative.");
						SelectBadField(dp, i, &NumbersOK);
						return(FALSE);
					}
					if (tmpF == theGR[xmin_i]) {
						ERROR("\PEnding x-axis value may not equal starting value.");
						SelectBadField(dp, i, &NumbersOK);
						return(FALSE);
					}
					theGR[xmax_i] = tmpF;
					break;
			case BD_FirstYAxis:
					if (pbd->YAxisIsPhase)
/*
						if ( (tmpF<pbd->MinPhase) || (tmpF>pbd->MaxPhase) ) {
*/
						if ( (tmpF<-1.0) || (tmpF>1.0) ) {
							ERROR("\PPhase must be between -1 and 1.");
							SelectBadField(dp, i, &NumbersOK);
							return(FALSE);
						}
					theGR[ymin_i] = tmpF;
					break;
			case BD_FirstYAxis+1:
					if (tmpF <= theGR[ymin_i]) {
						ERROR("\PY-axis maximum must be greater than y-axis minimum.");
						SelectBadField(dp, i, &NumbersOK);
						return(FALSE);
					}
					if (pbd->YAxisIsPhase) {
/*
						if ( (tmpF<pbd->MinPhase) || (tmpF>pbd->MaxPhase) ) {
*/
						if ( (tmpF<-1.0) || (tmpF>1.0) ) {
							ERROR("\PPhase must be between -1 and 1.");
							SelectBadField(dp, i, &NumbersOK);
							return(FALSE);
						}
						if (tmpF-theGR[xmin_i] > 1) {
							ERROR("\PPhase range must be less than or equal to 1.");
							SelectBadField(dp, i, &NumbersOK);
							return(FALSE);
						}
					}
					theGR[ymax_i] = tmpF;
					break;
			case BD_OtherTableParam:
					if (tmpF < 0.0) {
						if (pbd->XAxisIsAmplitude)
							ERROR("\PFrequency must be non-negative.");
						else
							ERROR("\PAmplitude must be non-negative.");
						SelectBadField(dp, i, &NumbersOK);
						return(FALSE);
					}
					if (pbd->XAxisIsAmplitude)
						pbd->Vars._freq = tmpF;
					else
						pbd->Vars._A = tmpF;
					break;
			case BD_OtherTableParam+1:
					pbd->Vars._damp = tmpF;
					if (ChangedDamping) {
						ChangedDamping = FALSE;
						if (tmpF != OldDamping) {
							pbd->DefaultIITimes = GetDefaultInitIterTimes(tmpF);
							LongToString((long) pbd->DefaultIITimes, CP(AutoInitString));
							if (myGetItemCtlValue(dp, BD_AutoInitial) > 0)
								mySetItemText(dp, BD_InitIter, CP(AutoInitString));
							OldDamping = tmpF;
						}
					}
					break;
			case BD_OtherTableParam+2:
					pbd->v00 = tmpF;
					break;
			case BD_OtherTableParam+3:
					pbd->delta0 = tmpF;
					break;
		}
	}

	if (myGetItemCtlValue(dp, BD_AutoInitial) > 0)
		pbd->InitIterTimes = AutoTimes;
	else
		if ( myGetLongParamItem(dp, i, &tmpL,
				"\PNumber of Initial Iterations", 1L, (long) INT_MAX) )
			pbd->InitIterTimes = (int) tmpL;
		else
			SelectBadField(dp, i, &NumbersOK);

	++i;
	if (myGetItemCtlValue(dp, BD_AutoShow) > 0)
		pbd->ShowTimes = AutoTimes;
	else
		if ( myGetLongParamItem(dp, i, &tmpL,
				"\PNumber of Iterations to Show", 1L, (long) INT_MAX) )
			pbd->ShowTimes = (int) tmpL;
		else
			SelectBadField(dp, i, &NumbersOK);

	X_Min = (pbd->XAxisIsAmplitude) ? xmin_i : ymin_i;
	X_Max = (pbd->XAxisIsAmplitude) ? xmax_i : ymax_i;
	Y_Min = (pbd->YAxisIsPhase) ? xmin_i : ymin_i;
	Y_Max = (pbd->YAxisIsPhase) ? xmax_i : ymax_i;
	pbd->XAxis[X_Min] = theGR[xmin_i];
	pbd->XAxis[X_Max] = theGR[xmax_i];
	pbd->YAxis[Y_Min] = theGR[ymin_i];
	pbd->YAxis[Y_Max] = theGR[ymax_i];

	return(NumbersOK);
}

Boolean DoBifurcationDiagramDialog(Boolean *StartIt)
{
	DialogPtr	dp;
	int			i, index;
	BooleanRect	tmpUsesFull;
	DoubleRect	tmpActual;
	Str255		ErrorStrings[ItemsInRect];
	double		tmpF;
	Str255		s;

	OldDamping = pba->Vars._damp;

	sEquate(CP(AmplitudeStatString), "\PAmplitude:");
	sEquate(CP(FrequencyStatString), "\PFrequency:");
	sEquate(CP(AutoString), "\PAuto");
	sEquate(CP(AmplitudeIncString), "\PWith each amplitude value, iterate");
	sEquate(CP(FrequencyIncString), "\PWith each frequency value, iterate");

	for(i=0; i<ItemsInRect; ++i) {
		myFToStrNoPad(pbd->XAxis[i], CP(XAxisStrings[i]), 0, DLogPrecision);
		myFToStrNoPad(pbd->YAxis[i], CP(YAxisStrings[i]), 0, DLogPrecision);
	}
	myFToStrNoPad(pbd->Vars._A, CP(AmplitudeString), 0, DLogPrecision);
	myFToStrNoPad(pbd->Vars._freq, CP(FrequencyString), 0, DLogPrecision);

	dp =GetNewDialog(BifurcationDiagramDLogID, NIL, -1L);

	i = BD_FirstET;

	if (pbd->XAxisIsAmplitude) {
		mySetItemCtlValue(dp, BD_XAxis, TRUE);
		mySetItemText(dp, BD_StatOtherTableParam, CP(FrequencyStatString));
/*
		mySetItemText(dp, BD_TableIncString, CP(AmplitudeIncString));
*/
		mySetItemText(dp, i++, CP(XAxisStrings[xmin_i]));
		mySetItemText(dp, i++, CP(XAxisStrings[xmax_i]));
	}
	else {
		mySetItemCtlValue(dp, BD_XAxis+1, TRUE);
		mySetItemText(dp, BD_StatOtherTableParam, CP(AmplitudeStatString));
/*
		mySetItemText(dp, BD_TableIncString, CP(FrequencyIncString));
*/
		mySetItemText(dp, i++, CP(XAxisStrings[ymin_i]));
		mySetItemText(dp, i++, CP(XAxisStrings[ymax_i]));
	}

	if (pbd->YAxisIsPhase) {
		mySetItemCtlValue(dp, BD_YAxis, TRUE);
		mySetItemText(dp, i++, CP(YAxisStrings[xmin_i]));
		mySetItemText(dp, i++, CP(YAxisStrings[xmax_i]));
	}
	else {
		mySetItemCtlValue(dp, BD_YAxis+1, TRUE);
		mySetItemText(dp, i++, CP(YAxisStrings[ymin_i]));
		mySetItemText(dp, i++, CP(YAxisStrings[ymax_i]));
	}

	mySetItemText(dp, i++, (pbd->XAxisIsAmplitude) ?
			CP(FrequencyString) : CP(AmplitudeString));
	mySetItemFloat(dp, i++, pbd->Vars._damp, DLogPrecision);
	mySetItemFloat(dp, i++, pbd->v00, DLogPrecision);
	mySetItemFloat(dp, i++, pbd->delta0, DLogPrecision);

/*
	LongToString((long) pbd->DefaultIITimes, CP(AutoInitString));
*/
	if (pbd->InitIterTimes == AutoTimes) {
		mySetItemText(dp, i++, CP(AutoString));
		mySetItemCtlValue(dp, BD_AutoInitial, TRUE);
	}
	else
		mySetItemLong(dp, i++, (long) pbd->InitIterTimes);

	if (pbd->ShowTimes == AutoTimes) {
		mySetItemText(dp, i++, CP(AutoString));
		mySetItemCtlValue(dp, BD_AutoShow, TRUE);
	}
	else
		mySetItemLong(dp, i++, (long) pbd->ShowTimes);

	if (pbd->GraphInAllWindows)
		mySetItemCtlValue(dp, BD_GraphInAll, TRUE);
	mySetItemCtlValue(dp, BD_WindowSize + ((pbd->FullWindow) ? 1 : 0), TRUE);
	mySetItemCtlValue(dp, BD_Restart, TRUE);
	mySetItemCtlValue(dp, BD_InColor, pbd->InColor);
	if (!UsingColorMac)
		myDisableCtlItem(dp, BD_InColor);
	if (pbd->InColor)
		myEnableCtlItem(dp, BD_Colors);
	else
		myDisableCtlItem(dp, BD_Colors);

	ChangedDamping = FALSE;
	if (HandleDialog(dp, BD_FirstET,  Cancel, UserOutlineItemNum, NIL,
			&BifDiagDLogItemFilter, &CheckBifurcationDiagramDialog) == OK) {
		
		pbd->GraphInAllWindows = myGetItemCtlValue(dp, BD_GraphInAll) > 0;
		pbd->FullWindow = myGetItemCtlValue(dp, BD_WindowSize+1) > 0;
		*StartIt = myGetItemCtlValue(dp, BD_Restart) > 0;
		pbd->InColor = myGetItemCtlValue(dp, BD_InColor) > 0;

		DialogReturn(dp, TRUE);
	}
	else DialogReturn(dp, FALSE);
}



		/*	------------------------------------------------------	*/
		/*	----------	Basins of Attraction Routines	----------	*/
		/*	------------------------------------------------------	*/


Boolean BasAttrDLogItemFilter(DialogPtr dp, int itemHit)
{
	double	NewDamping;
	Str255	s;

/*
	if ((ChangedDamping) && (itemHit != BA_TableParams+2) ) {
		ChangedDamping = FALSE;
		if ( myGetItemFloat(dp, BA_TableParams+2, &NewDamping) )
			if (NewDamping != OldDamping) {
				pba->DefaultMaxTimes = GetDefaultMaxTimes(NewDamping);
				LongToString((long) pba->DefaultMaxTimes, CP(AutoMaxString));
				if (myGetItemCtlValue(dp, BA_AutoMax) > 0)
					mySetItemText(dp, BA_MaxIter, CP(AutoMaxString));
				OldDamping = NewDamping;
			}
	}
*/

	switch (itemHit) {
		case BA_TableParams+2:	/*	damping	*/
				ChangedDamping = TRUE;
				break;
		case BA_MaxIter:
				if (myGetItemCtlValue(dp, BA_AutoMax) > 0) {
					myGetItemText(dp, itemHit, CP(s));
					if (sUpperComp(CP(s), CP(AutoString)) != 0)
						mySetItemCtlValue(dp, BA_AutoMax, FALSE);
				}
				else {
					myGetItemText(dp, itemHit, CP(s));
					if (sUpperComp(CP(s), CP(AutoString)) == 0)
						mySetItemCtlValue(dp, BA_AutoMax, TRUE);
				}
				break;
		case BA_AutoMax:
				if (myGetItemCtlValue(dp, itemHit) == 0) {
					mySetItemText(dp, BA_MaxIter, CP(AutoString));
					mySelItemText(dp, BA_MaxIter);
				}
				break;
		case BA_InColor:
				if (myGetItemCtlValue(dp, itemHit) == 0)
					mySetItemText(dp, BA_ColorsOrPatterns, CP(ColorString));
				else
					mySetItemText(dp, BA_ColorsOrPatterns, CP(PatternString));
				break;
		case BA_ColorsOrPatterns:
				if (myGetItemCtlValue(dp, BA_InColor) > 0)
					GetSolutionColors();
				else
					GetSolutionPatterns();
				break;
	}
	return(FALSE);
}

Boolean CheckBasinsOfAttractionDialog(DialogPtr dp, int itemNo)
{
	Boolean		NumbersOK = TRUE;
	int			i;
	double		tmpF;
	long		tmpL;
	Str255		ErrorStrings[BA_MaxIter-BA_FirstET];

	if (itemNo != OK) return(TRUE);

	i=0;
	sEquate(CP(ErrorStrings[i++]), "\PStarting Phase Value");
	sEquate(CP(ErrorStrings[i++]), "\PEnding Phase Value");
	sEquate(CP(ErrorStrings[i++]), "\PStarting Velocity Value");
	sEquate(CP(ErrorStrings[i++]), "\PEnding Velocity Value");
	sEquate(CP(ErrorStrings[i++]), "\PAmplitude");
	sEquate(CP(ErrorStrings[i++]), "\PFrequency");
	sEquate(CP(ErrorStrings[i++]), "\PDamping Coefficient");

	for(i=BA_FirstET; i<BA_MaxIter; ++i) {
		if ( !myGetItemFloat(dp, i, &tmpF) ) {
			FixBadValAlert(CP(ErrorStrings[i-BA_FirstET]));
			SelectBadField(dp, i, &NumbersOK);
			return(FALSE);
		}
		switch (i) {
			case BA_FirstPhase:
					if ( (tmpF<-1.0) || (tmpF>1.0) ) {
						ERROR("\PPhase must be between -1 and 1.");
						SelectBadField(dp, i, &NumbersOK);
						return(FALSE);
					}
					(*pNGR)[xmin_i] = tmpF;
					break;
			case BA_FirstPhase+1:
					if (tmpF <= (*pNGR)[xmin_i]) {
						ERROR("\PEnding phase must be greater than starting phase.");
						SelectBadField(dp, i, &NumbersOK);
						return(FALSE);
					}
					if ( (tmpF<-1.0) || (tmpF>1.0) ) {
						ERROR("\PPhase must be between -1 and 1.");
						SelectBadField(dp, i, &NumbersOK);
						return(FALSE);
					}
					if (tmpF-(*pNGR)[xmin_i] > 1) {
						ERROR("\PPhase range must be less than or equal to 1.");
						SelectBadField(dp, i, &NumbersOK);
						return(FALSE);
					}
					(*pNGR)[xmax_i] = tmpF;
					break;
			case BA_FirstVelocity:
					(*pNGR)[ymin_i] = tmpF;
					break;
			case BA_FirstVelocity+1:
					if (tmpF <= (*pNGR)[ymin_i]) {
						ERROR("\PEnding velocity must be greater than starting velocity.");
						SelectBadField(dp, i, &NumbersOK);
						return(FALSE);
					}
					(*pNGR)[ymax_i] = tmpF;
					break;
			case BA_TableParams:
					if (tmpF < 0.0) {
						ERROR("\PAmplitude must be non-negative.");
						SelectBadField(dp, i, &NumbersOK);
						return(FALSE);
					}
					pba->Vars._A = tmpF;
					break;
			case BA_TableParams+1:
					if (tmpF < 0.0) {
						ERROR("\PFrequency must be non-negative.");
						SelectBadField(dp, i, &NumbersOK);
						return(FALSE);
					}
					pba->Vars._freq = tmpF;
					break;
			case BA_TableParams+2:
					pba->Vars._damp = tmpF;
/*
					if (ChangedDamping) {
						ChangedDamping = FALSE;
						if (tmpF != OldDamping) {
							pba->DefaultMaxTimes = GetDefaultMaxTimes(tmpF);
							LongToString((long) pba->DefaultMaxTimes, CP(AutoMaxString));
							if (myGetItemCtlValue(dp, BA_AutoMax) > 0)
								mySetItemText(dp, BA_MaxIter, CP(AutoMaxString));
							OldDamping = tmpF;
						}
					}
*/
					break;
		}
	}

	if (myGetItemCtlValue(dp, BA_AutoMax) > 0)
		pba->MaxTimes = AutoTimes;
	else
		if ( myGetLongParamItem(dp, i, &tmpL,
				"\PMaximum Number of Iterations", 1L, (long) INT_MAX) )
			pba->MaxTimes = (int) tmpL;
		else
			SelectBadField(dp, i, &NumbersOK);

	++i;
	if ( myGetLongParamItem(dp, i, &tmpL,
			"\PIncrement Size", MIN_StepSize, MAX_StepSize) )
		pba->StepSize = (int) tmpL;
	else
		SelectBadField(dp, i, &NumbersOK);

	return(NumbersOK);
}

Boolean DoBasinsOfAttractionDialog(Boolean *StartIt, DoubleRect *pNewGraphRange)
{
	DialogPtr	dp;
	int			i, index;
	BooleanRect	tmpUsesFull;
	DoubleRect	tmpActual;
	double		tmpF;
	Str255		s;

	pNGR = pNewGraphRange;
	OldDamping = pba->Vars._damp;

	sEquate(CP(ColorString), "\PColors...");
	sEquate(CP(PatternString), "\PPatterns...");
	sEquate(CP(AutoString), "\PAuto");

	dp =GetNewDialog(BasinsOfAttractionDLogID, NIL, -1L);

	i = BA_FirstET;

	mySetItemFloat(dp, i++, (*pNGR)[xmin_i], DLogPrecision);
	mySetItemFloat(dp, i++, (*pNGR)[xmax_i], DLogPrecision);
	mySetItemFloat(dp, i++, (*pNGR)[ymin_i], DLogPrecision);
	mySetItemFloat(dp, i++, (*pNGR)[ymax_i], DLogPrecision);

	mySetItemFloat(dp, i++, pba->Vars._A, DLogPrecision);
	mySetItemFloat(dp, i++, pba->Vars._freq, DLogPrecision);
	mySetItemFloat(dp, i++, pba->Vars._damp, DLogPrecision);

/*
	LongToString((long) pba->DefaultMaxTimes, CP(AutoMaxString));
*/
	if (pba->MaxTimes == AutoTimes) {
		mySetItemText(dp, i++, CP(AutoString));
		mySetItemCtlValue(dp, BA_AutoMax, TRUE);
	}
	else
		mySetItemLong(dp, i++, (long) pba->MaxTimes);

	mySetItemLong(dp, i++, (long) pba->StepSize);

	mySetItemCtlValue(dp, BA_Faster, pba->Faster);
	mySetItemCtlValue(dp, BA_GraphInAll, pba->GraphInAllWindows);
	mySetItemCtlValue(dp, BA_WindowSize + ((pba->FullWindow) ? 1 : 0), TRUE);
	mySetItemCtlValue(dp, BA_Restart, TRUE);
	mySetItemCtlValue(dp, BA_InColor, pba->InColor);
	if (!UsingColorMac)
		myDisableCtlItem(dp, BA_InColor);
	if (pba->InColor)
		mySetItemText(dp, BA_ColorsOrPatterns, CP(ColorString));
	else
		mySetItemText(dp, BA_ColorsOrPatterns, CP(PatternString));

	ChangedDamping = FALSE;
	if (HandleDialog(dp, BA_FirstET,  Cancel, UserOutlineItemNum, NIL,
			&BasAttrDLogItemFilter, &CheckBasinsOfAttractionDialog) == OK) {
		
		pba->Faster = myGetItemCtlValue(dp, BA_Faster) > 0;
		pba->GraphInAllWindows = myGetItemCtlValue(dp, BA_GraphInAll) > 0;
		pba->FullWindow = myGetItemCtlValue(dp, BA_WindowSize+1) > 0;
		*StartIt = myGetItemCtlValue(dp, BA_Restart) > 0;
		pba->InColor = myGetItemCtlValue(dp, BA_InColor) > 0;

		DialogReturn(dp, TRUE);
	}
	else DialogReturn(dp, FALSE);
}


		/*	--------------------------------------------------	*/
		/*	----------	Solution Patterns Routines	----------	*/
		/*	--------------------------------------------------	*/

pascal void DrawSolColsRoutine(WindowPtr wp, int itemNo)
{
	int			itemType;
	Handle		item;
	Rect		box;

	GetDItem((DialogPtr) wp, itemNo, &itemType, &item, &box);
	SetPort(wp);
	SetSpecialColor(itemNo / DNST - 2, FALSE);
	PaintRect(&box);
	SetSpecialColor(BlackColorIndex, FALSE);
	FrameRect(&box);
}

pascal void DrawSolPatsRoutine(WindowPtr wp, int itemNo)
{
	int			itemType;
	Handle		item;
	Rect		box;

	GetDItem((DialogPtr) wp, itemNo, &itemType, &item, &box);
	FillRect(&box, &SolutionPats[itemNo / DNST - 2]);
	FrameRect(&box);
}

Boolean SolutionItemFilterProc(DialogPtr dp, int itemHit)
{
	int				newPatIndex, oldPatIndex;
	int				newSolIndex, oldSolIndex;
	int				i;

	if ( (itemHit < DNST*2) || (itemHit % DNST == NumSolutionTypes) )
		return(FALSE);

	newPatIndex = itemHit / DNST - 2;
	newSolIndex = itemHit % DNST;
	oldPatIndex = (*ptsi)[newSolIndex];
	if (newPatIndex == oldPatIndex)	return(TRUE);

	for(i=0; i<NumSolutionTypes; ++i)
		if ((*ptsi)[i] == newPatIndex) {
			oldSolIndex = i;
			break;
		}
	(*ptsi)[newSolIndex] = newPatIndex;
	(*ptsi)[oldSolIndex] = oldPatIndex;
	mySetItemCtlValue(dp, (newPatIndex+2) * DNST + newSolIndex, TRUE);
	mySetItemCtlValue(dp, (newPatIndex+2) * DNST + oldSolIndex, FALSE);
	mySetItemCtlValue(dp, (oldPatIndex+2) * DNST + oldSolIndex, TRUE);
	mySetItemCtlValue(dp, (oldPatIndex+2) * DNST + newSolIndex, FALSE);

	return(TRUE);
}

DialogPtr GetSolutionInfoDialog()
{
	return(GetNewDialog(SolutionInfoDLogID, NIL, -1L));
}

Boolean DoSolutionInfoDialog(Boolean DoingColor)
{
	DialogPtr	dp;
	int			i, j;
	int			itemType;
	Handle		item;
	Rect		box;

	ptsi = (DoingColor) ? &csi : &psi;

	dp = GetNewDialog(SolutionInfoDLogID, NIL, -1L);
	if (DoingColor)	ParamText("\PCOLORS", "\P", "\P", "\P");
	else			ParamText("\PPATTERNS", "\P", "\P", "\P");
	for(i=0; i<NumSolutionTypes; ++i)
		mySetItemCtlValue(dp, ((*ptsi)[i]+2)*DNST + i, TRUE);
	for(i =  3*DNST - 1; i <= DNST * (DNST+1) - 1; i += DNST) {
		GetDItem(dp, i, &itemType, &item, &box);
		if (DoingColor)
			SetDItem(dp, i, itemType, (Handle) &DrawSolColsRoutine, &box);
		else
			SetDItem(dp, i, itemType, (Handle) &DrawSolPatsRoutine, &box);
	}
	if (HandleDialog(dp, NO_EDIT_TEXT, Cancel,
			UserOutlineItemNum, NIL, &SolutionItemFilterProc, NIL) == OK) {
/*	Does as goes...
		for(i=0; i<NumSolutionTypes; ++i)
			for(j=0; j<NumSolutionTypes; ++j)
				if (myGetItemCtlValue(dp, (i+2)*DNST + j)) {
					(*ptsi)[j] = i;
					break;
				}
*/
		DialogReturn(dp, TRUE)
	}
	else
		DialogReturn(dp, FALSE)
}
