/*
 *	bbText.c
 *
 *	Handles Bouncing Ball's TextEdit chores.
 *
 *	-TA 6/26/90
 */

#include "bbPrototypes.h"
#include "bbGraphics.h"

enum {
		Ignore_WholePlaces,
		Default_WholePlaces,
		Vel_WholePlaces
};

#define	LEFT_TEXTMARGIN			8
#define MaxTELength				32500
#define	No_NumLines_Yet			-1
#define	OVERFLOW_MSG_1			"\P     *  *  *  *  *  *  *  *  *"
#define	OVERFLOW_MSG_2			"\P    Previous Solutions Currently"
#define	OVERFLOW_MSG_3			"\P    Unavailable--refer to manual"
#define	OVERFLOW_MSG_4			OVERFLOW_MSG_1
#define	TxOutHeaderLine			"\P  Hit   Phase    Velocity    Time"
#define	TSerHeaderLine			"\P  Time    Velocity  Position"

#define	BodyTE					(Texts[TextIndex].body)
#define	HeaderTE				(Texts[TextIndex].header)
#define	TheOF					(Texts[TextIndex].OF)
#define	NumLines				(Texts[TextIndex].NumLines)
#define	nthLine					(Texts[TextIndex].nthLine)
#define	OldNumLines				(Texts[TextIndex].OldNumLines)
#define	MaxLinesOnScreen		(Texts[TextIndex].MaxLinesOnScreen)
#define	HeaderLine				(Texts[TextIndex].HeaderLine)
#define	TextPort				(Texts[TextIndex].port)

#define	STORE_TextIndex			OldIndex = TextIndex
#define	RESTORE_TextIndex		TextIndex = OldIndex

					/*	From bb.c	*/
extern	hBBPreferences		hBBPrefs;
					/*	From bbWindows.c */
extern	int		MaxScreenY;
					/*	From bbIteration.cUsed in WriteOutput only	*/
extern	Str255	CurrentLine, StuckLine;
					/*	From bbFiles.cUsed in InitTimeSeries only	*/
extern	Boolean	GoodTSSaveFName;
					/*	From bbMenus.c	*/
extern	Boolean	NeedToSave, NeedToSaveTS;
					/*	From bbGraphics.c	*/
extern	Boolean	*pTxOutWindowCleared, *pTSerWindowCleared;

	BBText	Texts[NumTexts];
	int		TextIndex;
	int		OldIndex;
	Boolean	TextScrollingOn;
	Boolean	NeedToSetTxOutHeader, NeedToSetTSerHeader;


		/*	-------	General info., &c., routines	-------	*/

SetTextPorts(GrafPtr gp1, GrafPtr gp2)
{
	Texts[TxOutTextIndex].port = gp1;
	Texts[TSerTextIndex].port = gp2;
}

int GetTextIndex(int windex)
{
	switch (windex)	{
		case TextOutput:	return(TxOutTextIndex);
		case TimeSeries:	return(TSerTextIndex);
		default:			return(NO_TEXT_INDEX);
	}
}

int GetWindexFromTindex(int tindex)
{
	switch (tindex)	{
		case TxOutTextIndex:	return(TextOutput);
		case TSerTextIndex:		return(TimeSeries);
		default:				return(NoWindow);
	}
}

BBTextPtr GetTextFromTindex(int theTextIndex)
{
	return( &(Texts[theTextIndex]) );
}

TEHandle GetBBTEHandle(int windex, int WhichText)
{
	int	theIndex;

	if ( (theIndex = GetTextIndex(windex)) == NO_TEXT_INDEX )
		return(NIL);
	else switch (WhichText) {
		case HEADER_TEXT:	return(Texts[theIndex].header);
		case BODY_TEXT:		return(Texts[theIndex].body);
		default:			return(NIL);
	}
}

Boolean AnyTSText()
{
	TEHandle	teh = GetBBTEHandle(TimeSeries, BODY_TEXT);
	return( (teh==NIL) ? FALSE : ((**teh).teLength != 0) );
}


ScrollText(TEHandle teh, int LinesToScroll)
{
	TEScroll(0, (int) (LinesToScroll*(**teh).lineHeight), teh);
}

ScrollWindowText(int windex, int LinesToScroll)
{
	switch (windex) {
		case TextOutput:
		case TimeSeries:
							ScrollText(BodyTE, LinesToScroll);	break;
	}
}

Scroll_If_Needed()
{
	if (nthLine > NumLines) {
		ScrollText(BodyTE, -(nthLine-NumLines) );
/*
		IncTextScrollValue(TxOut, nthLine-NumLines);
*/
		nthLine = NumLines;
	}
}

ScrollBodyTE()
{
	int	NumScrollLines;
	int	WindIndex = GetWindexFromTindex(TextIndex);

	if (BodyTE == NIL) return;
	if (NumLines == OldNumLines) return;
	if (OldNumLines == No_NumLines_Yet) {	/*	Initializing BodyTE	*/
		OldNumLines = NumLines;
		return;
	}

	if (NumLines < OldNumLines) {
		if (!TextScrollingOn)
			Scroll_If_Needed();
	}
	else if (NumLines > OldNumLines) {
		if ( (!TextScrollingOn) && (nthLine == OldNumLines) ) {
			NumScrollLines = MIN(NumLines, (**BodyTE).nLines) - nthLine;
			ScrollText(BodyTE, NumScrollLines);
			nthLine += NumScrollLines;
		}
		else if ( (TextScrollingOn) && ((NumScrollLines =
				(NumLines-OldNumLines)-TextScrollValueFromBottom(WindIndex)) > 0) ) {
			NumScrollLines = MIN(NumScrollLines, GetTextScrollValue(WindIndex));
			ScrollText(BodyTE, NumScrollLines);
			IncTextScrollValue(WindIndex, -NumScrollLines);
		}
	}
	OldNumLines = NumLines;
}

SetNumLines(int windex)
	/*	Call whenever text window size changes or BodyTE inited	*/
{
	STORE_TextIndex;
	TextIndex = GetTextIndex(windex);

	if (BodyTE == NIL) return;	/* not created yet */
	NumLines = ( TextPort->portRect.bottom - (**BodyTE).viewRect.top -
			(ScrollBarWidth+BottomTextMargin) ) / (**BodyTE).lineHeight /* +1 is kludge */;
	(**BodyTE).viewRect.bottom = (**BodyTE).viewRect.top +
				(int) (NumLines * (**BodyTE).lineHeight);
	(**HeaderTE).viewRect.right = (**BodyTE).viewRect.right =
			TextPort->portRect.right - (ScrollBarWidth-1);
	ScrollBodyTE();
	SetTextScrollPage(windex, NumLines);
	if (!TextScrollingOn) return;
	SetTextScrollMax(windex, (**BodyTE).nLines - NumLines, FALSE);

	RESTORE_TextIndex;
}

UpdateText(int windex, double Ampl, double Freq, double Damp)
{
	STORE_TextIndex;
	TextIndex = GetTextIndex(windex);
	switch (windex) {
		case TimeSeries:
						if (HeaderTE == NIL)	break;
						if (NeedToSetTSerHeader) {
							SetHeader(windex, Ampl, Freq, Damp);
							NeedToSetTSerHeader = FALSE;
						}
						else
							TEUpdate(&TextPort->portRect, HeaderTE);
						TEUpdate(&TextPort->portRect, BodyTE);
						break;
		case TextOutput:
						if (NeedToSetTxOutHeader) {
							SetHeader(windex, Ampl, Freq, Damp);
							NeedToSetTxOutHeader = FALSE;
						}
						else
							TEUpdate(&TextPort->portRect, HeaderTE);
						TEUpdate(&TextPort->portRect, BodyTE);
						break;
	}
	RESTORE_TextIndex;
}

HandleTextActivation(int windex, int ActivateFlag)
{
	TEHandle	teh = GetBBTEHandle(windex, BODY_TEXT);

	if (ActivateFlag)
		TextIndex = GetTextIndex(windex);
	if (teh != NIL) {
		if ((ActivateFlag) && (TextScrollingOn))
			TEActivate(teh);
		else
			TEDeactivate(teh);
	}
}

ActivateATextScroll(int windex)
{
	if (GetTextScrollMax(windex) > 0)	return;

	STORE_TextIndex;
	TextIndex = GetTextIndex(windex);

	SetTextScrollMax(windex, (**BodyTE).nLines - NumLines, TRUE);
	TEActivate(BodyTE);

	RESTORE_TextIndex;
}

DeactivateATextScroll(int windex)
{
	STORE_TextIndex;
	TextIndex = GetTextIndex(windex);

	if (TextScrollingOn) {
		ScrollText(BodyTE, -TextScrollValueFromBottom(windex));
		nthLine = MIN((**BodyTE).nLines, NumLines);
	}
	SetTextScrollMax(windex, 0, TRUE);
	TESetSelect((**BodyTE).teLength, (**BodyTE).teLength, BodyTE);
	TEDeactivate(BodyTE);

	RESTORE_TextIndex;
}

ActivateTextScrolls(Boolean Activate_It)
{
	int	windex;

	STORE_TextIndex;

	for(TextIndex = TxOutTextIndex; TextIndex < NumTexts; ++TextIndex) {
		if (BodyTE == NIL)	continue;
		if ( (TextIndex == TSerTextIndex) && (!UseTimeSeries) )	continue;
		windex = GetWindexFromTindex(TextIndex);
		if (Activate_It)
			ActivateATextScroll(windex);
		else
			DeactivateATextScroll(windex);
		HiliteTextScrollBar(windex,
			Activate_It && (WindowIndex(FrontWindow()) == windex));
	}
	TextScrollingOn = Activate_It;
	RESTORE_TextIndex;
}

PtInBodyViewRect(int windex, Point pt)
{
	TEHandle	teh = GetBBTEHandle(windex, BODY_TEXT);
	Rect		tmpRect = (**teh).viewRect;

	tmpRect.left = 0;
	return( PtInRect(pt, &tmpRect) );
}

HandleTextSelection(int windex, Point pt, int mods)
{
	TEHandle	teh;
	int			i;

	if (!TextScrollingOn) return;

	teh = GetBBTEHandle(windex, BODY_TEXT);
	TEClick(pt, (mods & shiftKey) > 0, teh);
	if ( (**teh).selStart == (**teh).selEnd ) {
			/* Before version 1.42a, single click selected line
			 * 		-TA 1/20/91
			 */
			TEDeactivate(teh);	/*	Kludge to make vertical caret go away,	*/
			TEActivate(teh);
	}
}

HandleTEXTCopy(int windex)
{
	TEHandle	teh = GetBBTEHandle(windex, BODY_TEXT);
	int junk;
	PScrapStuff	pss = InfoScrap();

	if ( (**teh).selStart != (**teh).selEnd ) {
		TECopy(teh);
		if ( ZeroScrap() == noErr )
			if ((junk = TEToScrap()) != noErr) ;
	}
}

InitHeaderTE()
{
	Rect	destRect, viewRect;
	int		theTop = (TextIndex == TSerTextIndex) ? TS_BodyTETop : BodyTETop;

	SetPort(TextPort);
	SetRect(&destRect, LEFT_TEXTMARGIN, 2, 400, theTop);
	viewRect = destRect;
	if (HeaderTE != NIL) TEDispose(HeaderTE);
	HeaderTE = TENew(&destRect, &viewRect);
}

InitHeaderTEs()
{
	STORE_TextIndex;

	for(TextIndex = TxOutTextIndex; TextIndex < NumTexts; ++TextIndex)
		InitHeaderTE();

	RESTORE_TextIndex;
}

AppendVar(char *s, double d, TEHandle teh, int WholePlaces)
{
	AppendText(s, teh);
	if (WholePlaces > Ignore_WholePlaces)
		AppendDecFloat(d, TxOutFieldWidth, TxOutPrecision, WholePlaces,
				TRUE, TRUE, teh);
	else
		AppendPadFloat(d, TxOutFieldWidth, TxOutPrecision, TRUE, teh);
}

AppendVarLn(char *s, double d, TEHandle teh, int WholePlaces)
{
/*
	AppendVar(s, d, teh, WholePlaces);
*/
	AppendText(s, teh);
	if (WholePlaces > Ignore_WholePlaces)
		AppendDecFloat(d, TxOutFieldWidth, TxOutPrecision, WholePlaces,
				TRUE, FALSE, teh);	/*	FALSE only diff, AppendVar call	*/
	else
		AppendPadFloat(d, TxOutFieldWidth, TxOutPrecision, TRUE, teh);
	AppendTextLn("\P", teh);
}

SetHeader(int windex, double Ampl, double Freq, double Damp)
{
	Str255	s;
	int		i;

	STORE_TextIndex;
	TextIndex = GetTextIndex(windex);

	SetPort(TextPort);
	InitHeaderTE();
	if (windex == TimeSeries) {
		AppendText("\PSample Rate: ", HeaderTE);
		LongToString( (long) TS_SampleRate, CP(s));
		for (i=s[0];i<=5;++i)
			AppendChar(' ', HeaderTE);
		AppendTextLn(CP(s), HeaderTE);
	}
	AppendVarLn("\PAmplitude: ", Ampl, HeaderTE, Ignore_WholePlaces);
	AppendVarLn("\PFrequency: ", Freq, HeaderTE, Ignore_WholePlaces);
	AppendVarLn("\PDamping:   ", Damp, HeaderTE, Ignore_WholePlaces);
	AppendTextLn(HeaderLine, HeaderTE);

	RESTORE_TextIndex;
}

SetHeaders(double Ampl, double Freq, double Damp)
{
	int	windex;

	STORE_TextIndex;

	for(TextIndex = TxOutTextIndex; TextIndex < NumTexts; ++TextIndex) {
/*
		if ( (TextIndex == TSerTextIndex) && !UseTimeSeries )	continue;
*/
		SetHeader(GetWindexFromTindex(TextIndex), Ampl, Freq, Damp);
	}

	RESTORE_TextIndex;
}

ResetBodyOFs()
{
	int	i=-1;

	while (++i < MaxBodyOFs) {
		if (TheOF.hands[i] != NIL) {
			DisposHandle(TheOF.hands[i]);
			TheOF.hands[i] = (Handle) NIL;
		}
		else break;
	}
	TheOF.size = 0L;
	TheOF.index = 0L;
	TheOF.ind1 = TheOF.index / BodyOFIncrement;
	TheOF.ind2 = TheOF.index % BodyOFIncrement;
	TheOF.num = 0;
}

InitBodyOFs()
{
	int	i;

	for(i=0;i<MaxBodyOFs;++i)
		TheOF.hands[i] = NIL;
	ResetBodyOFs();
}

ClikLoopScroll(TEHandle teh, int LinesToScroll)
{
	ScrollText(teh, -LinesToScroll);
	IncTextScrollValue(GetWindexFromTindex(TextIndex), LinesToScroll);
}

pascal Boolean PasClikLoop()
{
	Point	pt;
	int		windex = GetWindexFromTindex(TextIndex);

	if (!StillDown()) return(FALSE);

	GetMouse(&pt);
	if ( !PtInRect(pt, &(**BodyTE).viewRect) ) {
		if (pt.v < (**BodyTE).viewRect.top)
			if (GetTextScrollValue(windex) > 0)
				ClikLoopScroll(BodyTE, -1);
		if (pt.v > (**BodyTE).viewRect.bottom)
			if (TextScrollValueFromBottom(windex) > 0)
				ClikLoopScroll(BodyTE, 1);
	}
	return(TRUE);
}

InitBodyTE(int windex)
{
	Rect	destRect, viewRect;
	int		theTop = ( (windex==TSer) ? TS_BodyTETop : BodyTETop );
	Boolean	theWindowCleared;

	theWindowCleared = ( (windex==TSer) ? *pTSerWindowCleared :
			*pTxOutWindowCleared );

	if (windex == TxOut)	NeedToSave = FALSE;
	if (windex == TSer)		NeedToSaveTS = FALSE;

	STORE_TextIndex;
	TextIndex = GetTextIndex(windex);

	if (!theWindowCleared)
		StdForceUpdate((WindowPtr) Texts[TextIndex].port);
/*
	StdForceUpdate((WindowPtr) Texts[TextIndex].port);
*/
	if (TextIndex == TSerTextIndex)
		GoodTSSaveFName = FALSE;

	SetPort(TextPort);
	SetRect(&destRect, LEFT_TEXTMARGIN, theTop, 400, theTop+TextLineHeight);
	SetRect(&viewRect, destRect.left, destRect.top, destRect.right,
		TextPort->portRect.bottom - (ScrollBarWidth + StdMargin) );
	if (BodyTE != NIL) TEDispose(BodyTE);
	BodyTE = TENew(&destRect, &viewRect);
	SetClikLoop((ProcPtr) &PasClikLoop, BodyTE);
	OldNumLines = No_NumLines_Yet;
	SetNumLines(windex);
	MaxLinesOnScreen = MaxScreenY / (**BodyTE).lineHeight + 1;

	nthLine = 0;

	ResetBodyOFs();

	RESTORE_TextIndex;
}

InitBodyTEs()
/*	unused?	*/
{
	InitBodyTE(TextOutput);
	if (Texts[TSerTextIndex].body != NIL)
		InitBodyTE(TimeSeries);
}

Boolean	CheckBodyTESize()
{
	int		LastDelLine, NumDelChars, i, BeginningLines;
	long	CharsToCopy, CharsCopied;

	if ((**BodyTE).teLength < MaxTELength) return(TRUE);

	BeginningLines = (TheOF.num > 0) ? OVERFLOW_MSG_LINES : 0;
	(**BodyTE).selStart = (**BodyTE).lineStarts[BeginningLines];
	LastDelLine = (**BodyTE).nLines-MaxLinesOnScreen;
	(**BodyTE).selEnd = (**BodyTE).lineStarts[LastDelLine];

	NumDelChars = (**BodyTE).selEnd - (**BodyTE).selStart;
	if ((TheOF.index + NumDelChars) > TheOF.size) {
		if (!CheckMem(0L)) return(FALSE);
		TheOF.hands[TheOF.num] = NewHandle(BodyOFIncrement);
		if ((long) TheOF.hands[TheOF.num] == NIL)
			return(FALSE);
		TheOF.size += BodyOFIncrement;
		++TheOF.num;
	}

	CharsCopied = 0L;
	for (i=TheOF.ind1; i<TheOF.num && NumDelChars>0; ++i) {
		CharsToCopy = MIN(BodyOFIncrement - TheOF.ind2, NumDelChars);
		BlockEquate(*TheOF.hands[TheOF.ind1]+TheOF.ind2,
			*(**BodyTE).hText+(**BodyTE).selStart+CharsCopied,
			(long) CharsToCopy );
		CharsCopied += CharsToCopy;
		NumDelChars -= CharsToCopy;
		TheOF.index += CharsToCopy;
		TheOF.ind1 = TheOF.index / BodyOFIncrement;
		TheOF.ind2 = TheOF.index % BodyOFIncrement;
	}

	TEDelete(BodyTE);
	if (TheOF.num == 1) {
		TESetSelect(0, 0, BodyTE);
		AppendTextLn(OVERFLOW_MSG_1, BodyTE);
		AppendTextLn(OVERFLOW_MSG_2, BodyTE);
		AppendTextLn(OVERFLOW_MSG_3, BodyTE);
		AppendTextLn(OVERFLOW_MSG_4, BodyTE);
	}
	TESetSelect((**BodyTE).teLength, (**BodyTE).teLength, BodyTE);
	ScrollText(BodyTE, LastDelLine-OVERFLOW_MSG_LINES);

	return(CheckMem(0L));
}

Boolean CheckBothBodyTESizes()
{
	STORE_TextIndex;

	for(TextIndex = TxOutTextIndex; TextIndex < NumTexts; ++TextIndex) {
		if (BodyTE == NIL)	continue;
		if (!CheckBodyTESize())	return(FALSE);
	}

	RESTORE_TextIndex;
	return(TRUE);
}

Boolean	WriteOutput(hitnum, thePhase, velocity, time, StickyState, PS)
	int		hitnum;
	double	thePhase, velocity, time;
	int		StickyState;
	int		PS;		/*	ProgramStatus	*/
{
	int		i;
	Str255	s;
	char	*ps = CP(s);
	Boolean	BodySizeOK;

	*pTxOutWindowCleared = FALSE;
/*
	SetPort(w[TxOut].Peek);
*/
	STORE_TextIndex;
	TextIndex = TxOutTextIndex;

	if ( (PS == OPENING) && (DelimiterIndex == SPACE_Delim) ) {
		AppendText(CP(CurrentLine), BodyTE);
		if (StickyState != NOT_STUCK)
			{ AppendText(CP(StuckLine), BodyTE); ++nthLine; }
	}
	else {
		LongToString((long) (hitnum), ps);
		for(i=s[0];i<5;++i) sInsertC(ps, ' ', 1);
		AppendText(ps, BodyTE);

		AppendVar("\P  ", thePhase, BodyTE, Default_WholePlaces);
		AppendVar("\P  ", velocity, BodyTE, Vel_WholePlaces);
		AppendVarLn("\P  ", time, BodyTE, Default_WholePlaces);

		if (StickyState != NOT_STUCK) {
			++nthLine;
			if (StickyState == STUCK) AppendTextLn("\PBall stuck to table", BodyTE);
			else {
				Scroll_If_Needed();
				if (PS < BIFURCATION_DIAGRAM)
					ERROR("\PBall stuck forever");
				AppendTextLn("\PBall stuck forever", BodyTE);
			}
		}
	}

	++nthLine;
	Scroll_If_Needed();

	BodySizeOK = CheckBodyTESize();
	RESTORE_TextIndex;
	return(BodySizeOK);
}

Boolean	WriteTS(time, vel, pos)
	double	time, vel, pos;
{
	Boolean	BodySizeOK;

/*
	SetPort(w[TxOut].Peek);
*/
	STORE_TextIndex;
	TextIndex = TSerTextIndex;

	AppendVar("\P", time, BodyTE, Default_WholePlaces);
	AppendVar("\P  ", vel, BodyTE, Vel_WholePlaces);
	AppendVarLn("\P  ", pos, BodyTE, Default_WholePlaces);

	++nthLine;
	Scroll_If_Needed();

	BodySizeOK = CheckBodyTESize();
	RESTORE_TextIndex;
	return(BodySizeOK);
}

NewTextEdit()
	/*	Called from NewGraphics(TRUE) only	*/
{
	InitBodyTE(TextOutput);
	InitBodyTE(TimeSeries);
}

InitTimeSeries(double Ampl, double Freq, double Damp, Boolean NeedHeader)
{
	InitBodyTE(TimeSeries);
	if (NeedHeader)
		SetHeader(TimeSeries, Ampl, Freq, Damp);
}

InitBBTextEdit(double Ampl, double Freq, double Damp)
{
	TextIndex = TxOutTextIndex;
	HeaderLine = TxOutHeaderLine;
	BodyTE = NIL;
	HeaderTE = NIL;
	InitBodyOFs();
	InitBodyTE(TextOutput);
	SetHeader(TextOutput, Ampl, Freq, Damp);
	NeedToSetTxOutHeader = FALSE;

	TextIndex = TSerTextIndex;
	HeaderLine = TSerHeaderLine;
	BodyTE = NIL;
	HeaderTE = NIL;
	InitBodyOFs();
	InitBodyTE(TimeSeries);
	SetHeader(TimeSeries, Ampl, Freq, Damp);
	NeedToSetTSerHeader = FALSE;

	TextIndex = TxOutTextIndex;
}

DisposeTextEdit(int windex)
{
	STORE_TextIndex;
	TextIndex = GetTextIndex(windex);
	if (HeaderTE != NIL)	TEDispose(HeaderTE);
	if (BodyTE != NIL)		TEDispose(BodyTE);
	HeaderTE = BodyTE = NIL;
	RESTORE_TextIndex;
}

HaltBBTextEdit()
{
	DisposeTextEdit(TextOutput);
	DisposeTextEdit(TimeSeries);
}
