/*	qm.c -- Quadratic Map
 *
 *	Iterates the logistic map, x_n+1 = lambda * x_n * (1 - x_n).
 *
 *	Tyler Abbott and Nick Tufillaro
 *	1991
 */

#include <stdlib.h>
#include <MultiFinder.h>
#include <MFUtils.h>
#include "qmPrototypes.h"
#include "qmMenus.h"
#include "qmGraphics.h"	/*	For GetInitialConditions() only	*/
#include "qmDebug.h"

			/*	These are used in _MAIN only, in	*/
			/*	calls to GraphFrom and AddSolution	*/
/*
extern	double			ti, tj, vi, xi, phase;
*/

		/*	External variables from qmSpecial.c	*/
extern	Boolean			*pBDGraphInAllWindows, *pBDUseColor;
		/*	External variables from qmFiles.c	*/
extern	int				OpenWindex;
		/*	External variables from qmMenus.c	*/
extern	Boolean			NeedToSave;
		/*	External variables from qmGraphics.c	*/
extern	CursHandle		watch;
extern	Windows			w;		/*	For GetInitialConditions() only	*/

			/*	Global variables used throughout program	*/
IterationVars		NormalVars, *pCurrentVars;
hQMPreferences		hQMPrefs;

			/*	Global variable used in qmIteration.c, qmGraphics.c, special */
SolutionsHandles	mySolutions;

			/*	Global variables used in qmIteration.c, qmMenus.c, elsewhere */
int		ProgramStatus;
int		theEventMask, UsualEventMask;
Boolean	PrintingOnly;
int		CurrentModifiers;
Boolean	SingleStepOn;
Boolean	UserWaiting;

double	nX0;	/* to be next x0 */

int		LastProgramStatus;
int		IterationStatus;
Boolean	ClearSolutions;
Boolean	RestartIteration;
Boolean	PAUSING, OKToContinue;
Boolean	OKToContinueBD = FALSE;
Boolean	MemoryFull;
Boolean	SuppressErrorMessages;
long	LastTickCount;

int		SolutionMemSize;	/* # solutions memory allocated for */
int		NumSolutionArrays;	/* # solution arrays memory ""  "" */

Boolean MultiFinderRunning;	/*	for MultiFinder	*/
Boolean	InForeground, HasWaitNextEvent;
long	BackgroundSleep;
		/*	If in background, breaks in the middle of _MAIN;		*/
		/*	NeedToFinishIteration tells if need to finish on resume	*/
Boolean	NeedToFinishIteration;

Boolean	UsingColorMac;
Boolean	System7OrLater;

Boolean	GettingParamFiles;
Boolean	GotATextFile, GotABDFile;
int		OpenedParamFiles;

long	ParamTest = 0L;

Boolean	WasADefaultsFile;


	/*	--------------------------------------------------	*/
	/*	----------	Memory-Checking Routines	----------	*/
	/*	--------------------------------------------------	*/

Boolean CheckMem(Size ExtraMemory)
/*	Require at least (MinMemory + ExtraMemory) free bytes, "breathing space."
 *			-TA 3/23/90
 *	ExtraMemory 0L except in "Special" functions
 *			-TA 3/5/91
 */
{
	Size	grow, MemAvail;

	MemAvail = MaxMem(&grow);
	return(MemAvail > (MinMemory + ExtraMemory));
}

Boolean StartingMemCheck()
{
	if (HitNumber >= SolutionMemSize-1) {
		if (!AddSolutionArray()) {
			MemoryFull = TRUE;
			return(FALSE);
		}
	}
	if (!CheckBodyTESizes()) {
		MemoryFull = TRUE;
		return(FALSE);
	}
	if (!CheckMem(0L)) {
		MemoryFull = TRUE;
		return(FALSE);
	}

	MemoryFull = FALSE;
	return(TRUE);
}

	/*	--------------------------------------------------	*/
	/*	----------	Iteration Control Routines	----------	*/
	/*	--------------------------------------------------	*/

GetIterationSpeed()
{
	int		tmpTime = MAX_RealIterationTime - RealIterationTime;

	if (DoIterationSpeedDialog(&tmpTime)) ;
		RealIterationTime = MAX_RealIterationTime - tmpTime;
}

Boolean StartIterating(int New_PS)
{
	Boolean	ContinueOK;

	if (HitNumber >= MaxSolutions) {
		ERROR("\PMaximum number of solutions reached");
		return(FALSE);
	}
	if (!StartingMemCheck())	{
		ERROR("\PSorrynot enough memory to perform requested action");
		return(FALSE);
	}

		/*	If already done StopIterating(), ProgramStatus will be WAITING	*/
	if (ProgramStatus != WAITING) {
		switch (ProgramStatus) {
			case BIFURCATION_DIAGRAM:	ContinueOK = TRUE;	/* ? */		break;
			default:					ContinueOK = HitNumber >= 0;	break;
		}
		/*	Need variable to show in middle of BifD?	*/
		StopIterating(ContinueOK);
	}
	if ( (New_PS < BIFURCATION_DIAGRAM) &&
			(LastProgramStatus >= BIFURCATION_DIAGRAM) ) {
		SetValuesFromSystemParams(0.0);
		SetHeaders(lambda);
		NewGraphics(TRUE, TRUE);
	}

	LastProgramStatus = ProgramStatus = New_PS;
	IterationStatus = WRITING;
	ControlMenuStart(ProgramStatus);
	PAUSING = FALSE;
	switch (ProgramStatus) {
		case BIFURCATION_DIAGRAM:	OKToContinueBD = FALSE;
									if (*pBDGraphInAllWindows) {
										ActivateTextScrolls(FALSE);
										SetTextTitle(TRUE, BifD);
									}
									break;
		default:					OKToContinue = FALSE;
									ActivateTextScrolls(FALSE);
									break;
	}
	RealHitNumber = -1;
	return(TRUE);
}

StopIterating(Boolean ContinueOK)
{
	Boolean *pOKToContinue;

	switch (ProgramStatus) {
		case BIFURCATION_DIAGRAM:
									pOKToContinue = &OKToContinueBD;
									pCurrentVars = &NormalVars;
									if (*pBDGraphInAllWindows) {
										NewGraphics(TRUE, FALSE);
										SetHeaders(lambda);
										SetTextTitle(TRUE, TxOut);
											/*	zeroed out in StartIterating()	*/
									}
									break;
		default:
									pOKToContinue = &OKToContinue;
									break;
	}
	*pOKToContinue = ContinueOK;
	ControlMenuStop(ProgramStatus, *pOKToContinue);
	ProgramStatus = WAITING;
	IterationStatus = NOT_ITERATING;
	PAUSING = FALSE;
	ActivateTextScrolls(TRUE);
}

BeginIfAppropriate()
{
	if (RestartIteration)
		StartIterating(NORMAL_ITERATION);
	else if (ProgramStatus == NORMAL_ITERATION)
		StopIterating(FALSE);
}

	/*	--------------------------------------------------	*/
	/*	----------	Solution-Handling Routines	----------	*/
	/*	--------------------------------------------------	*/

double GetSolution(int HitNum)
{
	return((**(mySolutions[SI1(HitNum)]))[SI2(HitNum)].x_n);
}

int GetSolutionHit(int HitNum)
{
	return((**(mySolutions[SI1(HitNum)]))[SI2(HitNum)].curHN);
}

ResetSolutions()
{
	int	i=0;

	SetHitNumber(-1);
	RealHitNumber = -1;
	while (++i < MaxSolutionArrays) {
		if (mySolutions[i] != NIL) {
			DisposHandle(mySolutions[i]);
			mySolutions[i] = NIL;
		}
		else break;
	}
	SolutionMemSize = SolutionIncrement;
	NumSolutionArrays = 1;
}

Boolean AddSolutionArray()
{
	mySolutions[NumSolutionArrays] =
			(SolutionsHandle) NewHandle( (sizeof(Solutions)) );
	if ((long) mySolutions[NumSolutionArrays] == NIL) return(FALSE);
	else {
		++NumSolutionArrays;
		SolutionMemSize += SolutionIncrement;
	}

	return(CheckMem(0L));
}

AddSolution(double x_n)
{
	Boolean	WriteOK;

	WriteOK = AddGraphicSolution(HitNumber, CurHitNumber, x_n, ProgramStatus);

	if (  (!PrintingOnly) &&
			( (InForeground) || (UseSoundInBackground) )  ) {
		if ( CurHitNumber >= Embedding )
			MakeTone(x_n, GetSolution(HitNumber-Embedding));
		else MakeTone(x_n, NO_TONE);
	}

	if (!WriteOK) {
		ERROR("\PNot enough memory to store more solutions");
		MemoryFull = TRUE;
		if (ProgramStatus == OPENING) 
			StopOpeningFile(FALSE);
		StopIterating(HitNumber>=0);
	}

	if (ProgramStatus == PLAYBACK) return;

	if ( (HitNumber >= SolutionMemSize-1) && (!MemoryFull) ) {
		if (!AddSolutionArray()) {
			MemoryFull = TRUE;
			if (!( (ProgramStatus == OPENING) && (FinishedOpening()) )) {
				ERROR("\PNot enough memory to store more solutions");
				if (ProgramStatus == OPENING)
					StopOpeningFile(FALSE);
				StopIterating(HitNumber>=0);
			}
		}
	}

	(**(mySolutions[HitInd1]))[HitInd2].x_n		= x_n;
	(**(mySolutions[HitInd1]))[HitInd2].curHN	= CurHitNumber;

	if (HitNumber >= MaxSolutions) {
		ERROR("\PMaximum number of solutions reached");
		StopIterating(FALSE);
		return;
	}
}

	/*	----------------------------------------------	*/
	/*	----------	Initializing Routines	----------	*/
	/*	----------------------------------------------	*/

SetDefaultPrefs()
{
	int	i;

/*!!*/	DefaultX0 = DEFAULT_X0;
/*!!*/	DefaultLambda = DEFAULT_LAMBDA;
/*!!*/	DefaultBDGIAW = FALSE;
/*!!*/	DefaultBDUseColor = UsingColorMac;

/*!!*/	GraphWhileOpening = FALSE;
/*!!*/	DelimiterIndex = SPACE_Delim;
/*!!*/	SuppressSaveWarnings = FALSE;
/*!!*/	AutoCommandKey = FALSE;
/*!!*/	RunInBackground = TRUE;
/*!!*/	FasterBackground = FALSE;
/*!!*/	UseSoundInBackground = FALSE;
/*!!*/	MaxIterations = DEFAULT_MAXITERATIONS;
/*!!*/	XMin = 0.0;
/*!!*/	XMax = 0.0;
/*!!*/	UseAutoXMin = TRUE;
/*!!*/	UseAutoXMax = TRUE;
/*!!*/	RealIterationTime = 0;
/*!!*/	NLHeight = 4;
/*!!*/	PSParams.NumberLine = PSParams.Parabola = PSParams.Line45 =
			PSParams.IterationLines = PSParams.Arrows =
			PSParams.Box = PSParams.SameSizeAxes = FALSE;
/*!!*/	Embedding = DEFAULT_Embedding;
/*!!*/	EPSParams.Polynomial2n = EPSParams.Line45 = EPSParams.Box =
			EPSParams.SameSizeAxes = FALSE;
/*!!*/	OrbitDotSize = 1;
/*!!*/	OrbitSpacing = 0;
/*!!*/	ScrollOrbit = FALSE;
/*!!*/	XnTonesOn = TRUE;
/*!!*/	XnkTonesOn = FALSE;
/*!!*/	CurrentMode = ContinuousMode;

	for(i=0; i<NumSolutionTypes; ++i)
/*!!*/		csi[i] = i;
}

SetDefaults()
{
	pIterationVars	piv = pCurrentVars;

	pCurrentVars = &NormalVars;

	DefaultX0 = x0;
	DefaultLambda = lambda;

	DefaultBDGIAW = *pBDGraphInAllWindows;
	DefaultBDUseColor = *pBDUseColor;

	pCurrentVars = piv;
}

SaveDefaults()
{
	SetDefaultWindowConfiguration();
	SetDefaults();
	if (!SaveDefaultFile()) {
		ERROR("\PError: Could not save defaults file");
		return;
	}
}

/*
SetDefaultParams()
{
	v0	=	DefaultV0;	delta	=	DefaultDelta;
	A	=	DefaultA;	freq	=	DefaultFreq;	damp = DefaultDamp;
	SetFrequencyParams();
	SpecialAccuracy = MAX(Accuracy, MIN_SPECIAL_ACCURACY);
	tstop = GetStopTime(Cycles);
}
*/

SetDefaultParams()
{
	x0		= DefaultX0;
	lambda	=	DefaultLambda;
/*
	MaxIterations	= DEFAULT_MAXITERATIONS;
*/
}

DetermineIfColor()
{
	GDHandle	theGD;
	int			ScreenDepth;

	if ( TrapAvailable(GetMainDeviceTrapID, ToolTrap, TRUE) ) {
		theGD = GetMainDevice();
		ScreenDepth = (**(**theGD).gdPMap).pixelSize;
		UsingColorMac = (ScreenDepth >= 4);
	}
	else
		UsingColorMac = FALSE;
}

InitMultiFinderGlobals()
{
	EventRecord	anEvent;

	InForeground = TRUE;
/*
	RunInBackground = TRUE;
*/
	MultiFinderRunning = MultiFinderOn();
	HasWaitNextEvent = WNEIsImplemented();
	BackgroundSleep = (FasterBackground) ? DAY_DREAM : CATNAP;
/*
	BackgroundSleep = (DAY_DREAM+CATNAP)/2;
*/
/*
	UseSoundInBackground = FALSE;
*/

	if (MultiFinderRunning)			/* Brings program to top MF layer */
		if (HasWaitNextEvent) WaitNextEvent(0, &anEvent, 0L, NIL);
		else GetNextEvent(0, &anEvent);
}

_INIT()
{
	int		i;
	Boolean	GotPrefs;

	if (!CheckMem(MinMemToStart-MinMemory)) {
		ERROR("\PSorrynot enough memory to run Quadratic Map");
		exit(1);
	}

	DILoad();

	InitMultiFinderGlobals();

	UserWaiting = FALSE;
	watch = GetCursor(watchCursor);
	StartWaiting();

	System7OrLater = (SystemVersion() >> 8) >= 7;
	DetermineIfColor();

	InitQMFiles();
	WasADefaultsFile = OpenDefaults(&GotPrefs);
	if (GotPrefs) {
		DefaultBDUseColor = DefaultBDUseColor && UsingColorMac;
	}
	else {
		ResrvMem(sizeof(QMPreferences));
		hQMPrefs = (hQMPreferences) NewHandle(sizeof(QMPreferences));
		HLock((Handle) hQMPrefs);	/*	almost always locked	*/
		SetDefaultPrefs();
	}

	pCurrentVars = &NormalVars;
	LastProgramStatus = ProgramStatus = WAITING;
	SuppressErrorMessages = FALSE;
	SetDefaultParams();

	InitQMMenus(UsingColorMac);
	InitSounds();
	if (!InitGraphics()) {
		StopWaiting();
		ERROR("\PSorrynot enough memory to run Quadratic Map");
		exit(1);
	}

	AddWindowsToMenu();
	InitQMFiles();

	StopIterating(FALSE);		/* Takes care of global vars */

	mySolutions[0] = (SolutionsHandle) NewHandle( (sizeof(Solutions)) );
	for(i=1;i<MaxSolutionArrays;++i)
		mySolutions[i] = NIL;
	ResetSolutions();

	ClearSolutions = FALSE;
	RestartIteration = TRUE;
	SingleStepOn = FALSE;

	InitSpecial();

	if (MemoryFull = !CheckMem(0L)) {
		StopWaiting();
		ERROR("\PSorrynot enough memory to run Quadratic Map");
		_HALT();
	}

	theEventMask = UsualEventMask = everyEvent;

	PrintingOnly = GettingParamFiles =  FALSE;
	GotATextFile = GotABDFile = FALSE;
	OpenedParamFiles = 0;
	switch (QMGetParamFiles()) {
		case appPrint:
						PrintingOnly = TRUE;
		case appOpen:
						theEventMask = UsualEventMask = 0;
						GettingParamFiles = TRUE;
						break;
	}

	InitialWindowConfiguration(PrintingOnly);
	StopWaiting();
}

	/*	--------------------------------------------------	*/
	/*	----------	Parameter-Setting Routines	----------	*/
	/*	--------------------------------------------------	*/

GetInitialFromLast()
{
	if (HitNumber >= 0)
		nX0 = (**(mySolutions[HitInd1]))[HitInd2].x_n;
}

Boolean	GetInitialConditions(Boolean DefaultSame)
{
	Boolean			GoodValues = TRUE;
	pIterationVars	piv = pCurrentVars;
	Boolean			GoodOp;

	pCurrentVars = &NormalVars;
	if (DefaultSame)	nX0 = x0;

	RestartIteration = TRUE;
	if ( DefaultSame || !SHIFT_DOWN )
		GoodOp = DoInitCondDialog(&nX0, &ClearSolutions, &RestartIteration);
	else
		GoodOp = TRUE;

	if (GoodOp) {
		x0 = nX0;
		StdForceUpdate((WindowPtr) w[Orbit].Peek);
		if ( (ProgramStatus < BIFURCATION_DIAGRAM) || (RestartIteration) )
			NewGraphics(ClearSolutions, TRUE);
		if (ClearSolutions)
			ResetSolutions();
		if (RestartIteration) {
			GoodValues = Initialize_Iteration_Globals(RestartIteration);
			if (!GoodValues) {
				StopIterating(FALSE);
				return(FALSE);
			}
		}
		else
			pCurrentVars = piv;
		return(TRUE);
	}
	else {
		pCurrentVars = piv;
		return(FALSE);
	}
}

SetValuesFromSystemParams(double oLambda)
{
	Boolean	ChangeMade;
	int		i;

	if (lambda != oLambda) {
/*
		SetVMax(TRUE);
*/
			/*	Added 10/24/91, wasn't refreshing when no data -TA	*/
		for (i=0;i<NumNormalWindows;++i)
			w[i].WindowCleared = FALSE;

		SetAutoGraphRange(lambda, FALSE);
/*
		InitPSScales();
		InitEPSScales();
		SetOrbitCoords();
*/
	}
}

Boolean GetSystemParameters()
{
	double			oLambda;
	Boolean			GoodValues = TRUE, GoodOp;
	pIterationVars	piv = pCurrentVars;

	pCurrentVars = &NormalVars;
	oLambda = lambda;
	RestartIteration = TRUE;
	GoodOp = DoPhysSysDialog(&RestartIteration);
	if (GoodOp) {
		if (!CheckNormalSaves(FALSE)) {
			GoodOp = FALSE;
			lambda = oLambda;
		}
	}

	if (GoodOp) {
		pCurrentVars = &NormalVars;
		SetValuesFromSystemParams(oLambda);
/*
		if ( (oLambda==0.0)^(lambda==0.0) )
			RedrawSine = TRUE;
*/
		if (!SpecialGraphing())
			SetHeaders(lambda);

		if (RestartIteration)
			GoodValues = Initialize_Iteration_Globals(RestartIteration);
		if ( (ProgramStatus < BIFURCATION_DIAGRAM) || (RestartIteration) ) {
			NewGraphics(TRUE, TRUE);
			ResetSolutions();
		}
		else
			pCurrentVars = piv;
		if (!GoodValues) {
			StopIterating(FALSE);
			return(FALSE);
		}
		return(TRUE);
	}
	else {
		pCurrentVars = piv;
		return(FALSE);
	}
}

GetMaxIterations()
{
	DoMaxIterationsDialog(&MaxIterations);
}


	/*	--------------------------------------------------	*/
	/*	----------	Iteration-Looping Routines	----------	*/
	/*	--------------------------------------------------	*/

AdvanceIterationStatus()
{
	if (IterationStatus > NOT_ITERATING)
		IterationStatus = (IterationStatus + 1) % IterationStatuses;
}

Boolean Handle_ITERATING(int PS)
{
	switch (GetNextSolution(PS)) {
		case BAD_SOLUTION:
							StopIterating((PS==OPENING) && (HitNumber>=0));
							return(FALSE);
		case NO_SOLUTION_YET:
							return(FALSE);
		case FOUND_SOLUTION:
							AdvanceIterationStatus();
							return(TRUE);
	}
}

Handle_GRAPHING(int PS)
{
	if ( OKtoGraph(PS) ) {
/*
		if (CurHitNumber > 0)
			if (CurHitNumber == 1)	GraphFrom(xj, HitNumber);
*/
/*
		GraphFrom(HitNumber+1, CurHitNumber+1, xj);
*/
	}
	AdvanceIterationStatus();
}

Boolean Handle_WRITING(int PS, Boolean *pOKCont)
{
	Boolean	TimeToStop;
static long timer1, timer2;

	*pOKCont = TRUE;
	TimeToStop = !StartNextSolution(PS, pOKCont);

/*
	GraphFrom(HitNumber, xj);
*/
	AddSolution(xj);
timer2=TickCount()-timer1;
timer1 = TickCount();

	if (PS == NORMAL_ITERATION)
		NeedToSave = TRUE;
	if (PS == BIFURCATION_DIAGRAM)
		GraphBD();

	if (CurHitNumber == 0)
		LastTickCount = TickCount();

	AdvanceIterationStatus();
	return(!TimeToStop);
}

Boolean Handle_META_ITERATING(int PS, Boolean *pOKCont, Boolean *pNeedToWait)
{
	Boolean	TimeToStop = FALSE, ShouldAdvance = TRUE;
	Boolean	CheckTime = TRUE;

	if (ProgramStatus == BIFURCATION_DIAGRAM) {
		if (!GetNextBD()) {
			*pOKCont = FALSE;
			TimeToStop = TRUE;
		}
		if (!*pBDGraphInAllWindows)
			CheckTime = FALSE;
	}

	*pNeedToWait = FALSE;
	if ( CheckTime && (CurHitNumber >= 0) )
		if (TickCount() < LastTickCount + RealIterationTime*WAIT_TICK_MULTIPLIER) {
			*pNeedToWait = TRUE;
			return(TRUE);
		}
	LastTickCount = TickCount();

	if (SingleStepOn) {
/*
		TimeToStop = TRUE;
*/
		HandleControlMenu(StopContinueItem);
		SingleStepOn = FALSE;
	}

	if (ShouldAdvance)
		AdvanceIterationStatus();
	return(!TimeToStop);
}

HandleParamFiles()
{
	if (ProgramStatus == OPENING)
		return;

	if ((PrintingOnly) && (OpenedParamFiles > 0) ) {
		StdForceUpdate(FrontWindow());
		PrintFrontWindow();
	}
	for(;;) {
		OpenWindex = GetNextParamFile();
		if (OpenWindex == NoWindow) {
			if (PrintingOnly)
				_HALT();
			else {
				GettingParamFiles = FALSE;
				theEventMask = UsualEventMask = everyEvent;
				return;
			}
		}
		switch (OpenWindex) {
			case TxOut:
					if ((GotATextFile) && (!PrintingOnly))	return;
					SelectWindowIndex(OpenWindex);
					if (StartOpeningFile()) {
						GotATextFile = TRUE;
						++OpenedParamFiles;
						IterationStatus = ITERATING;
						return;
					}
					break;
			case BifD:
					if ((GotABDFile) && (!PrintingOnly))	return;
					SelectWindowIndex(OpenWindex);
					if (OpenFile())
						if (OpenSpecialFile(OpenWindex)) {
							GotABDFile = TRUE;
							++OpenedParamFiles;
							HandleAllUpdates();
							return;
						}
					break;
		}
	}

}

_MAIN()
{
	Boolean	TimeToStop = FALSE, OKCont = FALSE;
	Boolean	NeedToWait;

	if ( (!InForeground) && (!RunInBackground) ) return;

	if (GettingParamFiles)
		HandleParamFiles();

	if (PAUSING) return;
	if (ProgramStatus == WAITING) return;

	switch (IterationStatus) {
		case ITERATING:
						if (!Handle_ITERATING(ProgramStatus))
							return;
						if (!InForeground) {
							NeedToFinishIteration = TRUE;
							return;
						}
						/*	normally don't returnwhen get solution,
						 *	always do full cycle */
		case GRAPHING:
						Handle_GRAPHING(ProgramStatus);
						if (!InForeground) return;
		case WRITING:
						TimeToStop = !Handle_WRITING(ProgramStatus, &OKCont);
						if (!InForeground) return;
		case META_ITERATING:
						if (!Handle_META_ITERATING(ProgramStatus,
								&OKCont, &NeedToWait))
							TimeToStop = TRUE;
						if (!InForeground)
							NeedToFinishIteration = FALSE;
						if (NeedToWait)					return;
						if (TimeToStop)
							{	StopIterating(OKCont);	return;	}
						break;
	}
}


	/*	--------------------------------------------------	*/
	/*	----------	Event-Handling Routines	----------	*/
	/*	--------------------------------------------------	*/

HandleContent(WindowPtr wp, Point pt, int mods)
{
	int		windex;
	int		WhichHotSpot;

	if ( (windex=WindowIndex(wp)) == NoWindow) return;

	SetPort(wp);		GlobalToLocal(&pt);

	if( HandleControl(windex, pt, &WhichHotSpot) ) {
		if (WhichHotSpot != NO_HotSpot)
			switch (windex) {
				case Orbit:
								if (WhichHotSpot == 0) {
									if (GetInitialConditions(TRUE))
										BeginIfAppropriate();
								}
								else if (WhichHotSpot == 1)
									GetGraphRange();
								break;
				case TextOutput:
								if (GetSystemParameters())
									BeginIfAppropriate();
								break;
				case PhaseSpace:
								if (WhichHotSpot == 0)
									HandleLeftPSsHS(windex);
								else if (WhichHotSpot == 1)
									GetPhaseSpaceInfo();
								break;
				case EmbeddedPS:
								if (WhichHotSpot == 0)
									HandleLeftPSsHS(windex);
								else if (WhichHotSpot == 1)
									GetEmbeddedPSInfo();
								break;
				case BifDiagram:
								if (WhichHotSpot == 0)
									GetBifDiagInfo(FALSE);
								else if (WhichHotSpot == 1)
									GetBifDiagInfo(FALSE);
								break;
			}
		return;
	}

	switch ( FindQMClick(windex, pt) ) {
		case TextOutput:
						HandleTextSelection(windex, pt, mods);
						break;
		case Orbit:
						while(StillDown()) ;
						GetOrbitInfo();
						break;
		case PhaseSpace:
						if ( TrackPSRect(pt) )
							GetPSMagnification(SHIFT_DOWN);
						else
							if (CheckPSClick(pt, &nX0))
								if (GetInitialConditions(FALSE))
									BeginIfAppropriate();
						break;
		case EmbeddedPS:
						if ( TrackEPSRect(pt) )
							GetEPSMagnification(SHIFT_DOWN);
						else
							CheckEPSClick(pt);
						break;
		case BifDiagram:
						if ( TrackBDRect(pt) )
							GetBifDiagInfo(TRUE);
						else
							GetBifDiagInfo(FALSE);
						break;
	}

}

HandleMouseDown(EventRecord *pev)
{
	WindowPtr	WhichWindow;
	int			WindowPart;

	WindowPart = FindWindow(pev->where, &WhichWindow);
	if (WindowIndex(WhichWindow) != NoWindow)
		if (WhichWindow != FrontWindow())
			if (WindowPart != inDrag)
				{	SelectWindow(WhichWindow);	return;	}
	switch (WindowPart) {
		case inSysWindow:
						SystemClick(pev, WhichWindow);
						break;
		case inMenuBar:
						AdjustQMMenus();
						HandleMenu(MenuSelect(pev->where));
						break;
		case inGrow:
						QMGrowWindow(pev->where, WhichWindow);
						break;
		case inDrag:
						QMDragWindow(pev->where, WhichWindow, pev->modifiers);
						break;
		case inGoAway:
						QMGoAwayWindow(pev->where, WhichWindow);
						break;
		case inZoomIn:
		case inZoomOut:
						QMZoomWindow(pev->where, WhichWindow, WindowPart);
						break;
		case inDesk:
						break;
 		case inContent:
 						HandleContent(WhichWindow, pev->where, pev->modifiers);
 						break;
	}
}

CheckVars_HandleUpdate(WindowPtr wp)
/*	Sets pCurrentVars to appropriate value for given window.
 *	By the way, since for BifD and BAttr, all updating is done by
 *	copying offports, the WindowIndex test is not really necessary
 */
{
	pIterationVars	piv = pCurrentVars;

	if ( (ProgramStatus == BIFURCATION_DIAGRAM) && !*pBDGraphInAllWindows )
		if (WindowIndex(wp) != BifDiagram)
			pCurrentVars = &NormalVars;

	HandleUpdate(wp, HitNumber);
	pCurrentVars = piv;
}

HandleApp4Event(EventRecord *pEvent)
{
	WindowPtr	wp;

	if (App4Selector(pEvent) == SUSPEND_RESUME_SELECTOR) {
		wp = FrontWindow();
		if (SuspResIsSuspend(pEvent->message)) {
			InForeground = FALSE;
			NeedToFinishIteration = FALSE;
			if (WindowIndex(wp) != NoWindow)
				QMHandleActivate(wp, FALSE);
		}
		else {
			InForeground = TRUE;
			SetCursor(&arrow);
			if (WindowIndex(wp) != NoWindow)
				QMHandleActivate(wp, TRUE);
			if (NeedToFinishIteration)
				_MAIN();
		}
	}
}

_EVENT(EventRecord *theEvent)
{
	switch (theEvent->what) {
		case mouseDown:
			CurrentModifiers = theEvent->modifiers;
			HandleMouseDown(theEvent);
			break;
		case keyDown:
			CurrentModifiers = theEvent->modifiers;
			HandleKeyDown(theEvent->message, theEvent->modifiers);
			break;
		case updateEvt:
			CheckVars_HandleUpdate((WindowPtr) theEvent->message);
			break;
		case diskEvt:
			if (StdHandleDiskInsertion(theEvent->message) != noErr)
				;
			break;
		case activateEvt:
			if (WindowIndex((WindowPtr) theEvent->message) == NoWindow)
				return;
			else QMHandleActivate((WindowPtr) theEvent->message, theEvent->modifiers & activeFlag);
			break;
		case app4Evt:
			HandleApp4Event(theEvent);
			break;
	}
}

Boolean NextEvent(int eventMask, EventRecord *theEvent)
{
	int	MFSleep;

	if (HasWaitNextEvent) {
		if ( PAUSING || (ProgramStatus == WAITING) )
			MFSleep = ALL_NIGHT;
		else if ( !InForeground )
			MFSleep = (RunInBackground) ? BackgroundSleep : ALL_NIGHT;
		else
			MFSleep = DAY_DREAM;
		return( WaitNextEvent(eventMask, theEvent, MFSleep, NIL) );
	}
	else
		return( GetNextEvent(eventMask, theEvent) );
}

HandleAllActivates()
{
	EventRecord	theEvent;

	for(;;)
		if (GetNextEvent(activMask, &theEvent)) _EVENT(&theEvent);
		else break;
}

HandleAllUpdates()
{
	EventRecord	theEvent;

	for(;;)
		if (GetNextEvent(updateMask, &theEvent)) _EVENT(&theEvent);
		else break;
}

HandleEvent()
{
	Boolean		ok;
	EventRecord	theEvent;

	HiliteMenu(0);
	if (InForeground) {
		if (UserWaiting)	SetCursor(*watch);
		else				SetCursor(&arrow);
	}
	if (!HasWaitNextEvent)
		SystemTask ();		/* Handle desk accessories */
	
	if (NextEvent(theEventMask, &theEvent))
		_EVENT(&theEvent);
	else
		_MAIN();
}

	/*	------------------------------------------	*/
	/*	----------	Main Loop; Quitting	----------	*/
	/*	------------------------------------------	*/

main()
{
	InitMacintosh();
	InitPrinting();
	_INIT();
	
	for (;;)
		HandleEvent();
}

_HALT()
{
	ResetSolutions();
	HaltGraphics();
	HaltPrinting();
	DIUnload();
	exit(1);
}
