/*	qm.h
 *
 *	Quadratic Map constants, data structures
 */

#include <Color.h>		/*	Currently for qmWindows.c only -TA 10/29/90	*/
#include <GeneralMacros.h>
#include "qmIteration.h"

#define	GetMainDeviceTrapID	0xAA2A	/*	Used in qm.c, qmWindows.c	*/

#define AssumedZero			1e-20
#define	AlmostZero			1e-10

#define	G					981.0	/* Acceleration of Gravity, CGS */

#define	CP(s)				((char *) &(s))
#define	RETURN_IF_FALSE(a)	if (!(a)) return(FALSE)

#define	SHIFT_DOWN			((CurrentModifiers & shiftKey) > 0)
#define	OPTION_DOWN			((CurrentModifiers & optionKey) > 0)

#define	DEFAULT_X0				0.5
#define	DEFAULT_LAMBDA			3
#define	MAX_ITERATIONS			32767L
#define	DEFAULT_MAXITERATIONS	MAX_ITERATIONS
/*
#define	NON_ZERO_A			( (A==0.0) ? DEFAULT_A : A )
*/

#define	MAX_RealIterationTime	100
#define	MAX_WAIT_PERIOD			5	/*	seconds	*/
#define	TICKS_PER_SECOND		60
#define	WAIT_TICK_MULTIPLIER	(MAX_WAIT_PERIOD*TICKS_PER_SECOND/MAX_RealIterationTime)

	/* MinMemory: 64K + 4K "buffer"--needed, I believe, for TextEdit */
#define	MinMemory			68 * 1024L
#define	MinMemToStart		128 * 1024L		/* at least 128K to start qm */

#define	SolutionIncrement	512
#define	MaxSolutions		32767
#define MaxSolutionArrays	MaxSolutions/SolutionIncrement + 1	/* 64 */
#define	SI1(i)				((i)/SolutionIncrement)
#define	SI2(i)				((i)%SolutionIncrement)

	/* For "Body Overflow," qmText.c (and qmFiles.c) */
#define	BodyOFIncrement		32768L
#define	OVERFLOW_MSG_LINES	4
#define	CharsPerSolution	22L		/* really 20 */
#define	MaxBodyOFs			(MaxSolutions * CharsPerSolution / BodyOFIncrement + 1)

#define	NO_ANSWER			-1
#define	NO_TONE				-1.0

		/*	Embedding constants; for qmImpactGraphics, qmDialog	*/
#define	MIN_Embedding		1
#define	MAX_Embedding		1024
#define	DEFAULT_Embedding	2

		/*	Delimiter Description Length, for qmFiles, qmDialog	*/
#define	DelimDescLength		32
#define	DelimStr(index)		CP(DelimiterStrings[index])

		/*	WaitNextEvent sleep values (in ticks), for MultiFinder */
#define	ALL_NIGHT	300
#define	CATNAP		10
#define	DAY_DREAM	0

#define	NO_HotSpot	-1


		/*	"Program Statuses"	*/
enum {
		WAITING,
		NORMAL_ITERATION,
		OPENING,
		PLAYBACK,
		BIFURCATION_DIAGRAM
};

		/*	"Iteration Statuses"	*/
enum {
		NOT_ITERATING = -1,
		ITERATING,
		GRAPHING,
		WRITING,
		META_ITERATING,
		IterationStatuses /* sp? */
};

		/*	Windows	*/
enum {
		NoWindow = -1,
		PhaseSpace,			PSpace = PhaseSpace,
		TextOutput,			TxOut = TextOutput,
		Orbit,
		EmbeddedPS,			EmbPS = EmbeddedPS,
		NumDefaultWindows,
		NumNormalWindows = NumDefaultWindows,

		BifDiagram = NumNormalWindows,
							BifD = BifDiagram,
		NumWindows
};

		/*	Returned by GetNextSolution() (qmIteration.c)	*/
enum {
		BAD_SOLUTION = -1,
		NO_SOLUTION_YET,
		FOUND_SOLUTION
};

		/*	Used in GetQMTEHandle (qmText.c), also called in qmPrint.c	*/
enum {
		HEADER_TEXT,
		BODY_TEXT
};

		/*	Used in qmText.c, also called in qmFiles.c	*/
enum {
		NO_TEXT_INDEX = -1,
		TxOutTextIndex,
		NumTexts
};

		/*	Delimiter Indicesused in qmFiles.c, qmDialog.c	*/
enum {
		FIRST_Delim,
		SPACE_Delim = FIRST_Delim,
		TAB_Delim,
		COMMA_Delim,
		CR_Delim,
		LAST_Delim,
		NUM_Delims = LAST_Delim
};

enum {
	NoSolutionType = -1,
	UnboundedSolution,
	Period1Solution,
	Period2Solution,
	Period3Solution,
	Period4Solution,
	Period8Solution,
	Period16Solution,
	OtherSolution,
	NumSolutionTypes
};

typedef	char	myFileNameType[64];

typedef struct {
	int		curHN;
	double	x_n;
}		Solution;

	/*	Use an array of handles because I can't seem to access more than
	 *	32767 bytes in an array of structures pointed to by one handle.
	 *	32767 bytes corresponds to about 1260 solutionsI was getting
	 *	a crash every time I went over 1024 solutions (the memory was
	 *	added in increments of 512 solutions).  Perhaps Think C's
	 *	maximum array size, static or dynamic, is 32767 bytes.
	 *	I could, however, access more than 32767 characters pointed to
	 *	by a handlerefer to BodyOverflow use in CheckBodyTESize() in
	 *	qmGraphics.c.  For this, I think the limit is 8 MB (per Inside Mac);
	 *	that's the maximum amount the NewHandle() command can allocate.
	 *		-TA 3/23/90
	 */

typedef	Solution			Solutions[SolutionIncrement];
typedef	Solutions			**SolutionsHandle;
typedef	SolutionsHandle		SolutionsHandles[MaxSolutionArrays];

typedef	Handle				BodyOFHandles[MaxBodyOFs];
typedef	struct {
			BodyOFHandles	hands;
			long			index, size;
			int				num;
			int				ind1, ind2;
		}	BodyOF;
typedef	struct {
			TEHandle	header;
			TEHandle	body;
			BodyOF		OF;
			int			NumLines, nthLine;
			int			OldNumLines;
			int			MaxLinesOnScreen;	/* Max. number of text lines on screen */
			char		*HeaderLine;
			GrafPtr		port;
		}	QMText, *QMTextPtr;

typedef	struct {				/*	Used in qmWindows, qmFiles	*/
			int		windex;
			Rect	wr;
		}	WindowLocation;

typedef	struct {
			int				num;
			WindowLocation	ws[];
		}	WindowLocs, **hWindowLocs;

	/*	For magnification	*/
enum {
	ymin_i,
	xmin_i,
	ymax_i,
	xmax_i,
	ItemsInRect
};
enum {
	top, left, bottom, right
};
typedef double	DoubleRect[ItemsInRect];	/*	Used in graphics, qmDialog	*/
typedef Boolean	BooleanRect[ItemsInRect];

	/*	For Phase Space, Embedded Phase Space	*/
typedef struct {
	Boolean	NumberLine, Parabola, Line45, IterationLines, Arrows, Box;
	Boolean	SameSizeAxes;
}	PhaseSpaceParams;

typedef struct {
	Boolean	Polynomial2n, Line45, Box;
	Boolean	SameSizeAxes;
}	EmbeddedPSParams;

											/*	Used in special windows	*/
typedef	int				SolutionInfo[NumSolutionTypes];
typedef	SolutionInfo	*pSolutionInfo;
