/*	qmDialog.c
 *
 *	Routines which handle Quadratic Map's Dialog Boxes
 *
 *		-TA, 7/87...
 *
 */

#include <ctype.h>
#include <limits.h>
#include <DialogUtils.h>

#include "qmPrototypes.h"

enum {
		GeneralAlertID			= 351,
		BadParamAlertID,
		dlogdummy1,
		OutOfRangeAlertID,
		AllBadValsAlertID,
		FixBadValAlertID,
		FixBadIntValAlertID,
		AboutQMDLogID			= 400,
		PrintingDLogID,
		DelimiterDLogID,
		QuitWarningDLogID,
		SaveWarningDLogID,
		BackgroundDLogID,
		InitCondDLogID			= 452,
		PhysSysDLogID,
		GraphRangeDLogID,
		OverflowDLogID,
		dlogdummy2,
		dlogdummy3,
		PhaseSpaceDLogID,
		MaxIterationsDLogID,
		EmbeddedPSDLogID,
		MagnifyPhaseSpacesDLogID,
		OrbitDLogID,
		OldIterSpeedDLogID,
		IterationSpeedDLogID,
		ShowWindowDLogID
};

#define	ScrollValBarCDEFID		128
#define	PopUpCDEFID				129
#define	DelimPopUpMenuID		128

#define	DLogPrecision			10
#define UserOutlineItemNum		3
#define	FirstEditTextItem		6
#define	ETItem1					FirstEditTextItem
#define	MinOverflowMsgTime		120L		/* ticks, = 2 seconds */
#define	DelimItemOffset			4
#define	IterSpeedItem			5
#define	MagPSFirstET			6
#define	MagPSNoMag				17
#define	GR_UseAutoXMinItem		9
#define	GR_UseAutoXMaxItem		10
#define	PSs_FirstTextItem		6
#define	PSs_FirstCheckBox		10
#define	MIN_NLHeight			1
#define	MAX_NLHeight			16

#define	O_FirstET				5		/*	Orbit ...	*/
#define	MIN_DotSize				1
#define	MAX_DotSize				8
#define	MIN_HorSpace			0
#define	MAX_HorSpace			8


				/*	Global external vars from qm.c	*/
extern	pIterationVars		pCurrentVars;
extern	hQMPreferences		hQMPrefs;
extern	int					CurrentModifiers;
extern	Boolean				SuppressErrorMessages;
extern	Boolean				System7OrLater;

DialogPtr		PrintingDialog;
DialogPtr		OverflowDialog;
DialogPtr		IterSpeedDialog;

double			IC_x0;							/*	Initial Conditions	*/
double			SP_lambda;						/*	System Parameters	*/
double			GR_XMin, GR_XMax;				/*	Graph Range	*/
long			MI_MaxIterations;				/*	Max. Iterations	*/
int				IS_RealIterTime;				/*	Iteration Speed	*/
ControlHandle	SVBch;
BooleanRect		MPS_UsesFull;					/*	Magnify Phase Space ...	*/
DoubleRect		MPS_Actual, *pMPS_Full;
int				PS_NumberLineHeight;			/*	Phase Space ...	*/
int				EPS_Embedding;					/*	Embedded Phase Space ...	*/
int				O_DotSize, O_HorSpace;			/*	Orbit ...	*/

long			OverflowMessageStartTime;
Boolean			OverflowOn = FALSE;
int				DelimButtonItemNum, DelimTextItemNum;
int				StdFileOutlineItemNum;
Boolean			FirstAccessStdFileDLog;
Str255			GraphRangeStrings[2];
Str255			MagRangeStrings[ItemsInRect];


SelectBadField(DialogPtr dp, int index, Boolean *PreviousAllRight)
{
	if (*PreviousAllRight)
		mySelItemText(dp, index);
	*PreviousAllRight = FALSE;
}

	/* ------------------------------------------------------ */
	/* -------- Actual dialog/alert calling routines -------- */
	/* ------------------------------------------------------ */

BadParamAlert(char *s)
{
	ParamText(s,"\P","\P","\P");
	if (!CautionAlert(BadParamAlertID,NIL)) SysBeep(30);
}

AllBadValsAlert(char *s)
{
	ParamText(s,"\P","\P","\P");
	if (!CautionAlert(AllBadValsAlertID,NIL)) SysBeep(30);
}

FixBadValAlert(char *s)
{
	ParamText(s,"\P","\P","\P");
	if (!CautionAlert(FixBadValAlertID,NIL)) SysBeep(30);
}

FixBadIntValAlert(char *s)
{
	ParamText(s,"\P","\P","\P");
	if (!CautionAlert(FixBadIntValAlertID,NIL)) SysBeep(30);
}

ErrorAlert(char *s1, char *s2, char *s3, char *s4)
{
	ParamText(s1,s2,s3,s4);
	if (!StopAlert(GeneralAlertID,NIL))
		SysBeep(30);
}

ERROR(char *s)
{
	if (!SuppressErrorMessages)
		ErrorAlert(s,"\P","\P","\P");
}

OutOfRangeAlert(char *s, long minL, long maxL)
{
	Str255	s1, s2;

	LongToString(minL, CP(s1));
	LongToString(maxL, CP(s2));
	ParamText(s, s1, s2, "\P");
	if (!CautionAlert(OutOfRangeAlertID,NIL)) SysBeep(30);
}

Boolean myGetLongParamItem(dp, itemNo, pl, s, minL, maxL)
	DialogPtr	dp;
	int			itemNo;
	long		*pl, minL, maxL;
	char		*s;
	/* Returns long from item itemNo, subject to contraints minL & maxL */
{
	double	d = 0.0;

	if (!myGetItemLong(dp, itemNo, pl)) {
		if (myGetItemFloat(dp, itemNo, &d))
			{	FixBadIntValAlert(s); return(FALSE);	}
		else
			{	FixBadValAlert(s); return(FALSE);		}
	}
/*
	if (!myGetItemLong(dp, itemNo, pl)) {
		myGetItemText(dp, itemNo, CP(itemStr));
		if (IsFloat(CP(itemStr)))	FixBadIntValAlert(s);
		else						FixBadValAlert(s);
		return(FALSE);
	}
*/
	if (!myGetItemFloat(dp, itemNo, &d)) ;
	if (fabs(d-*pl) > AlmostZero)
		{	FixBadIntValAlert(s); return(FALSE);	}
	if ( !( (*pl >= minL) && (*pl <= maxL) ) )
		{ OutOfRangeAlert(s, minL, maxL); return(FALSE);}
	return(TRUE);
}

DoAboutQMDialog()
{
	DialogPtr	dp;

	dp = GetNewDialog(AboutQMDLogID, NIL, -1L);
	if (HandleDialog(dp, NO_SELECT_ITEM,
			Cancel, UserOutlineItemNum, NIL, NIL, NIL) == OK) ;
	DisposDialog(dp);
}

Boolean CheckInitCondDialog(DialogPtr dp, int itemNo)
{
	Boolean	NumbersOK = TRUE;
	int		i = -1;

	if (itemNo != OK) return(TRUE);

	++i;
	if ( !myGetItemFloat(dp, ETItem1+i+i, &IC_x0) ) {
		FixBadValAlert("\PInitial Condition");
		SelectBadField(dp, ETItem1+i+i, &NumbersOK);
	}

	return(NumbersOK);
}

Boolean DoInitCondDialog(nX0, myClearSolutions, RestartIteration)
	double	*nX0;
	Boolean	*myClearSolutions, *RestartIteration;
{
	DialogPtr	dp;
	int			ClearSolChkBox = 7, RestartChkBox = 8; /* "constants" */
	int			i;

	dp = GetNewDialog(InitCondDLogID, NIL, -1L);

	i = -1;
	++i; mySetItemFloat(dp, ETItem1+i+i,*nX0,DLogPrecision);
	mySetItemCtlValue(dp, ClearSolChkBox,(int) *myClearSolutions);
	mySetItemCtlValue(dp, RestartChkBox, (int) *RestartIteration);

	if (HandleDialog(dp, ETItem1, Cancel, UserOutlineItemNum,
			NIL, NIL, &CheckInitCondDialog) == OK) {
		*nX0 = IC_x0;
		*myClearSolutions = myGetItemCtlValue(dp,ClearSolChkBox);
		*RestartIteration = myGetItemCtlValue(dp,RestartChkBox);
		DialogReturn(dp, TRUE);
	}
	else DialogReturn(dp, FALSE);
}

Boolean CheckPhysSysDialog(DialogPtr dp, int itemNo)
{
	Boolean	NumbersOK = TRUE;
	int		i = -1;

	if (itemNo != OK) return(TRUE);

	++i;
	if ( !myGetItemFloat(dp, ETItem1+i+i, &SP_lambda) ) {
		FixBadValAlert("\Plambda");
		SelectBadField(dp, ETItem1+i+i, &NumbersOK);
	}
	return(NumbersOK);
}

Boolean DoPhysSysDialog(Boolean *RestartIteration)
{
	DialogPtr	dp;
	int			RestartChkBox = 7; /* "constant" */
	int			i;

	dp = GetNewDialog(PhysSysDLogID, NIL, -1L);

	i = -1;
	++i; mySetItemFloat(dp, ETItem1+i+i, lambda, DLogPrecision);
	mySetItemCtlValue(dp, RestartChkBox, (int) *RestartIteration);

	if (HandleDialog(dp, ETItem1, Cancel, UserOutlineItemNum,
			NIL, NIL, &CheckPhysSysDialog) == OK) {
		lambda = SP_lambda;
		*RestartIteration = myGetItemCtlValue(dp,RestartChkBox);
		DialogReturn(dp, TRUE);
	}
	else DialogReturn(dp, FALSE);
}

Boolean CheckMaxIterationsDialog(DialogPtr dp, int itemNo)
{
	int		i = -1;
	long	tmpL;
	Boolean	NumbersOK = TRUE;

	if (itemNo != OK) return(TRUE);

	++i;
	if ( myGetLongParamItem(dp, ETItem1+i+i, &tmpL,
			"\PMaximum Iterations", 0L, MAX_ITERATIONS) )
		MI_MaxIterations = tmpL;
	else
		SelectBadField(dp, ETItem1+i+i, &NumbersOK);

	return(NumbersOK);
}

Boolean DoMaxIterationsDialog(long *pMaxIterations)
{
	DialogPtr	dp;
	int			i;

	dp = GetNewDialog(MaxIterationsDLogID, NIL, -1L);

	i = -1;
	++i; mySetItemLong(dp, ETItem1+i+i, *pMaxIterations);

	if (HandleDialog(dp, ETItem1, Cancel, UserOutlineItemNum,
			NIL, NIL, &CheckMaxIterationsDialog) == OK) {
		*pMaxIterations = MI_MaxIterations;
		DialogReturn(dp, TRUE);
	}
	else DialogReturn(dp, FALSE);
}

Boolean GraphRangeDLogItemFilter(DialogPtr dp, int itemHit)
{
	Str255	s;
	int		index;

	switch (itemHit) {
		case ETItem1:
		case ETItem1+2:
				index = (itemHit - ETItem1) / 2;
				if (myGetItemCtlValue(dp, GR_UseAutoXMinItem+index)) {
					myGetItemText(dp, itemHit, CP(s));
					if ( sComp(CP(GraphRangeStrings[index]), CP(s)) != 0 )
						mySetItemCtlValue(dp, GR_UseAutoXMinItem+index, FALSE);
				}
				break;
		case GR_UseAutoXMinItem:
		case GR_UseAutoXMaxItem:
				index = itemHit - GR_UseAutoXMinItem;
				if (!myGetItemCtlValue(dp, itemHit)) {
					mySetItemText(dp, ETItem1+2*index,
							CP(GraphRangeStrings[index]));
					mySelItemText(dp, ETItem1+2*index);
				}
				break;
	}
	return(FALSE);
}

Boolean CheckGraphRangeDialog(DialogPtr dp, int itemNo)
{
	int		i = -1;
	Boolean	NumbersOK = TRUE;

	if (itemNo != OK) return(TRUE);

	++i;
	if ( !myGetItemFloat(dp, ETItem1+i+i, &GR_XMin) ) {
		FixBadValAlert("\Pminimum x value");
		SelectBadField(dp, ETItem1+i+i, &NumbersOK);
	}
	++i;
	if ( !myGetItemFloat(dp, ETItem1+i+i, &GR_XMax) ) {
		FixBadValAlert("\Pmaximum x value");
		SelectBadField(dp, ETItem1+i+i, &NumbersOK);
	}
	if (GR_XMax <= GR_XMin) {
		ERROR("\PMaximum value must be greater than minimum.");
		SelectBadField(dp, ETItem1+i+i, &NumbersOK);
	}

	return(NumbersOK);
}

Boolean DoGraphRangeDialog(double *pXMin, double *pXMax,
		double AutoXMin, double AutoXMax,
		Boolean *pUseAutoXMin, Boolean *pUseAutoXMax)
{
	DialogPtr	dp;
	int			i;

	dp = GetNewDialog(GraphRangeDLogID, NIL, -1L);

	i = -1;
	myFToStrNoPad(AutoXMin, CP(GraphRangeStrings[0]), 0, DLogPrecision);
	myFToStrNoPad(AutoXMax, CP(GraphRangeStrings[1]), 0, DLogPrecision);

	i = -1;
	++i;
	if (*pUseAutoXMin) {
		mySetItemCtlValue(dp, GR_UseAutoXMinItem, TRUE);
		mySetItemText(dp, ETItem1+i+i, CP(GraphRangeStrings[0]));
	}
	else
		mySetItemFloat(dp, ETItem1+i+i, *pXMin, DLogPrecision);

	++i;
	if (*pUseAutoXMax) {
		mySetItemCtlValue(dp, GR_UseAutoXMaxItem, TRUE);
		mySetItemText(dp, ETItem1+i+i, CP(GraphRangeStrings[1]));
	}
	else
		mySetItemFloat(dp, ETItem1+i+i, *pXMax, DLogPrecision);

	if (HandleDialog(dp, ETItem1, Cancel, UserOutlineItemNum, NIL,
			&GraphRangeDLogItemFilter, &CheckGraphRangeDialog) == OK) {
		*pUseAutoXMin = myGetItemCtlValue(dp, GR_UseAutoXMinItem);
		*pUseAutoXMax = myGetItemCtlValue(dp, GR_UseAutoXMaxItem);
		*pXMin = (*pUseAutoXMin) ? AutoXMin : GR_XMin;
		*pXMax = (*pUseAutoXMax) ? AutoXMax : GR_XMax;
		DialogReturn(dp, TRUE);
	}
	else DialogReturn(dp, FALSE);
}

StartOverflowMessage(Boolean positive)
{
	char *s;

	s = (positive) ? "\Ppositive" : "\Pnegative";
	ParamText("\PIteration will go to ", s, "\P infinity.", "\P");
	OverflowDialog = GetNewDialog(OverflowDLogID, NIL, -1L);
	ShowWindow(OverflowDialog);
	DrawDialog(OverflowDialog);
	OverflowMessageStartTime = TickCount();
	OverflowOn = TRUE;
}

StopOverflowMessage()
{
/*
	while ( TickCount() < (OverflowMessageStartTime+MinOverflowMsgTime) )
		;
*/
	if ( TickCount() >= (OverflowMessageStartTime+MinOverflowMsgTime) )
	HideWindow(OverflowDialog);
	DisposDialog(OverflowDialog);
	OverflowDialog = NIL;
	OverflowOn = FALSE;
	return;
}

Boolean IterSpeedDLogEventFilter(DialogPtr dp, EventRecord *pTheEvent,
		int *pItemHit, Boolean *pHowCKFShouldReturn)
{
	char	theChar;
	long	rc;
	int		ScrollAmount, NewValue;

	switch (pTheEvent->what) {
		case keyDown: 
		case autoKey:
			theChar = BitAnd(pTheEvent->message, charCodeMask);
			rc = GetCRefCon(SVBch);
			switch (theChar) {
				case Left_ArrowKey:		ScrollAmount = -LoWord(rc);	break;
				case Right_ArrowKey:	ScrollAmount =  LoWord(rc);	break;
				case Up_ArrowKey:		ScrollAmount =  HiWord(rc);	break;
				case Down_ArrowKey:		ScrollAmount = -HiWord(rc);	break;
				default:				return(FALSE);
			}
			NewValue = GetCtlValue(SVBch) + ScrollAmount;
			if (NewValue > GetCtlMax(SVBch))	NewValue = GetCtlMax(SVBch);
			if (NewValue < GetCtlMin(SVBch))	NewValue = GetCtlMin(SVBch);
			SetCtlValue(SVBch, NewValue);
			*pHowCKFShouldReturn = TRUE;
			return(TRUE);
	}
	return(FALSE);
}

Boolean DoIterationSpeedDialog(int *pRealIterTime)
{
						/*	next two lines are "constants"	*/
	int				ScrollValBarItem = 5;
	long			PageScrollAmount = 10L, ButtonScrollAmount = 1L;
	int				type;
	Handle			item;
	Rect			box;

	IterSpeedDialog =GetNewDialog(IterationSpeedDLogID, NIL, -1L);
	GetDItem(IterSpeedDialog, ScrollValBarItem, &type, &item, &box);
	if (type & ctrlItem) {
		DisposeControl((ControlHandle) item);
		SVBch = NewControl(IterSpeedDialog, &box, "\P", TRUE,
				*pRealIterTime, 0, MAX_RealIterationTime,
				ScrollValBarCDEFID<<4, 0L /* (long) &SetDialogPhases */ );
		SetCRefCon(SVBch, (PageScrollAmount << 16) | (ButtonScrollAmount) );
		SetDItem(IterSpeedDialog, ScrollValBarItem, ctrlItem+resCtrl, (Handle) SVBch, &box);
	}
	if (HandleDialog(IterSpeedDialog, NO_SELECT_ITEM,  Cancel,
			UserOutlineItemNum, &IterSpeedDLogEventFilter, NIL, NIL) == OK) {
		*pRealIterTime = GetCtlValue(SVBch);
		DialogReturn(IterSpeedDialog, TRUE);
	}
	else DialogReturn(IterSpeedDialog, FALSE);
}

/*
Boolean CheckIterationSpeedDialog(DialogPtr dp, int itemNo)
{
	Boolean	GoodValue = TRUE;

	if (itemNo != OK) return(TRUE);

	if ( !myGetItemFloat(dp, IterSpeedItem, &IS_RealIterTime) ) {
		FixBadValAlert("\Piteration speed");
		SelectBadField(dp, IterSpeedItem, &GoodValue);
	}
	else if (IS_RealIterTime < 0.0) {
		ERROR("\PIteration speed must be greater than or equal to 0.");
		SelectBadField(dp, IterSpeedItem, &GoodValue);
	}
	return(GoodValue);
}

Boolean DoIterationSpeedDialog(double *pRealIterTime)
{
	DialogPtr	dp;

	dp =GetNewDialog(IterationSpeedDLogID, NIL, -1L);

/*
	mySetItemFloat(dp, IterSpeedItem, *pRealIterTime, 3);
*)

	if (HandleDialog(dp, IterSpeedItem,  Cancel, UserOutlineItemNum,
			NIL, NIL, &CheckIterationSpeedDialog) == OK) {
		*pRealIterTime = IS_RealIterTime;
		DialogReturn(dp, TRUE);
	}
	else DialogReturn(dp, FALSE);
}
*/

Boolean MagPSDLogItemFilter(DialogPtr dp, int itemHit)
{
	Str255	s;
	int		index, i;

	switch (itemHit) {
		case MagPSFirstET:
		case MagPSFirstET+3:
		case MagPSFirstET+6:
		case MagPSFirstET+9:
				index = (itemHit - MagPSFirstET) / 3;
				index = (index%2)*2 + (1-index/2);	/*	0->1, 1->3, 2->0, 3->2	*/
				if (myGetItemCtlValue(dp, itemHit+1)) {
					myGetItemText(dp, itemHit, CP(s));
					if ( sComp(CP(MagRangeStrings[index]),
							CP(s)) != 0 )
						mySetItemCtlValue(dp, itemHit+1, FALSE);
				}
				break;
		case MagPSFirstET+1:
		case MagPSFirstET+4:
		case MagPSFirstET+7:
		case MagPSFirstET+10:
				index = (itemHit - MagPSFirstET - 1) / 3;
				index = (index%2)*2 + (1-index/2);	/*	0->1, 1->3, 2->0, 3->2	*/
				if (!myGetItemCtlValue(dp, itemHit)) {
					mySetItemText(dp, itemHit-1,
							CP(MagRangeStrings[index]));
					mySelItemText(dp, itemHit-1);
				}
				break;
		case MagPSNoMag:
				for(i=0; i <= 3; ++i) {
					index = (i%2)*2 + (1-i/2);	/*	0->1, 1->3, 2->0, 3->2	*/
					mySetItemText(dp, MagPSFirstET+3*i,
							CP(MagRangeStrings[index]));
					mySetItemCtlValue(dp, MagPSFirstET+3*i+1, TRUE);
				}
				mySelItemText(dp, MagPSFirstET);
				break;
	}
	return(FALSE);
}

Boolean CheckMagnifyPhaseSpacesDialog(DialogPtr dp, int itemNo)
{
	int		i, j, index;
	Str255	ErrorStrings[ItemsInRect];
	Boolean	NumbersOK = TRUE;
	double	tmpF;

	if (itemNo != OK) return(TRUE);

	i=0;
	sEquate(CP(ErrorStrings[i++]), "\Pminimum y-axis value");
	sEquate(CP(ErrorStrings[i++]), "\Pminimum x-axis value");
	sEquate(CP(ErrorStrings[i++]), "\Pmaximum y-axis value");
	sEquate(CP(ErrorStrings[i++]), "\Pmaximum x-axis value");

	for(i=ymin_i;i<ItemsInRect;++i) {
/*		(inverse)
		j = (1-i%2)*2 + i/2;		(*	0->2, 1->0, 2->3, 3->1	*)
*/
		j = (i%2)*2 + (1-i/2);
		index = MagPSFirstET+3*i;
		MPS_UsesFull[j] = myGetItemCtlValue(dp, index+1);
		if (!MPS_UsesFull[j]) {
			if ( !myGetItemFloat(dp, index, &tmpF) ) {
				FixBadValAlert(CP(ErrorStrings[j]));
				SelectBadField(dp, index, &NumbersOK);
				return(FALSE);
			}
			switch (j) {
				case ymin_i:
					break;
				case xmin_i:
					break;
				case ymax_i:
					if (tmpF <= MPS_Actual[ymin_i]) {
						ERROR("\PMaximum y-axis value must be greater than minimum.");
						SelectBadField(dp, index, &NumbersOK);
						return(FALSE);
					}
					break;
				case xmax_i:
					if (tmpF <= MPS_Actual[xmin_i]) {
						ERROR("\PMaximum x-axis value must be greater than minimum.");
						SelectBadField(dp, index, &NumbersOK);
						return(FALSE);
					}
					break;
			}
			MPS_Actual[j] = tmpF;
		}
		else {
			MPS_Actual[j] = (*pMPS_Full)[j];
			if ( (j == xmax_i) || (j == ymax_i) ) {
				if (MPS_Actual[j-2] >= MPS_Actual[j]) {
					if (j == xmax_i)
						ERROR("\PMinimum x-axis value must be less than maximum.");
					else if (j == ymax_i)
						ERROR("\PMinimum y-axis value must be less than maximum.");
					SelectBadField(dp, index - 3, &NumbersOK);
					return(FALSE);
				}
			}
		}
	}
	return(NumbersOK);
}

Boolean DoMagnifyPhaseSpacesDialog(DoubleRect *Actual, DoubleRect *Full,
		BooleanRect *UsesFull, Boolean IsPhaseSpace)
{
	DialogPtr	dp;
	int			i, index;
	char		*ps;

	pMPS_Full = Full;

	dp =GetNewDialog(MagnifyPhaseSpacesDLogID, NIL, -1L);
	ps = (IsPhaseSpace) ? "\P        " : "\PEMBEDDED";
	ParamText(ps, "\P", "\P", "\P");

	for(i=ymin_i;i<ItemsInRect;++i) {
		MPS_Actual[i] = (*Actual)[i];
		index = (1-i%2)*2 + i/2;		/*	0->2, 1->0, 2->3, 3->1	*/
		index = MagPSFirstET+3*index;
		myFToStrNoPad((*Full)[i], CP(MagRangeStrings[i]),
				0, DLogPrecision);
		if ((*UsesFull)[i])
			mySetItemText(dp, index, CP(MagRangeStrings[i]));
		else
			mySetItemFloat(dp, index, MPS_Actual[i], DLogPrecision);
		mySetItemCtlValue(dp, index+1, (*UsesFull)[i]);
	}

	if (HandleDialog(dp, MagPSFirstET,  Cancel, UserOutlineItemNum,
			NIL, &MagPSDLogItemFilter, &CheckMagnifyPhaseSpacesDialog) == OK) {
		for(i=ymin_i;i<ItemsInRect;++i) {
			(*Actual)[i] = MPS_Actual[i];
			(*UsesFull)[i] = MPS_UsesFull[i];
		}
		DialogReturn(dp, TRUE);
	}
	else DialogReturn(dp, FALSE);
}

Boolean PhaseSpaceDialogFilter(DialogPtr dp, int itemNo)
{
	if (itemNo == PSs_FirstCheckBox+4) {
		if (myGetItemCtlValue(dp, PSs_FirstCheckBox+4)) {
			myDisableCtlItem(dp, PSs_FirstCheckBox+5);
			myDisableCtlItem(dp, PSs_FirstCheckBox+6);
		}
		else {
			myEnableCtlItem(dp, PSs_FirstCheckBox+5);
			myEnableCtlItem(dp, PSs_FirstCheckBox+6);
		}
	}
	return(FALSE);
}

Boolean CheckPhaseSpaceDialog(DialogPtr dp, int itemNo)
{
	long	tmpL;
	Boolean	GoodValue = TRUE;

	if (itemNo != OK) return(TRUE);

	if ( myGetLongParamItem(dp, PSs_FirstTextItem, &tmpL,
			"\PNumber Line Height",
			(long) MIN_NLHeight, (long) MAX_NLHeight) )
		PS_NumberLineHeight = (int) tmpL;
	else
		SelectBadField(dp, PSs_FirstTextItem, &GoodValue);

	return(GoodValue);
}

Boolean DoPhaseSpaceDialog(int *pNLHeight, PhaseSpaceParams *ppsi,
		Boolean *pShowAllIterLines)
{
	DialogPtr	dp;
	int			i;

	dp = GetNewDialog(PhaseSpaceDLogID, NIL, -1L);

	mySetItemLong(dp, PSs_FirstTextItem, (long) *pNLHeight);
	i= PSs_FirstCheckBox;
	mySetItemCtlValue(dp, i++, ppsi->NumberLine);
	mySetItemCtlValue(dp, i++, ppsi->Parabola);
	mySetItemCtlValue(dp, i++, ppsi->Line45);
	mySetItemCtlValue(dp, i++, ppsi->Box);
	mySetItemCtlValue(dp, i++, ppsi->IterationLines);
	mySetItemCtlValue(dp, i++, *pShowAllIterLines);
	mySetItemCtlValue(dp, i++, ppsi->Arrows);
	mySetItemCtlValue(dp, i++, ppsi->SameSizeAxes);
	if (!ppsi->IterationLines) {
		myDisableCtlItem(dp, PSs_FirstCheckBox+5);
		myDisableCtlItem(dp, PSs_FirstCheckBox+6);
	}

	if (HandleDialog(dp, PSs_FirstTextItem, Cancel, UserOutlineItemNum,
			NIL, &PhaseSpaceDialogFilter, &CheckPhaseSpaceDialog) == OK) {
		*pNLHeight = PS_NumberLineHeight;
		i= PSs_FirstCheckBox;
		ppsi->NumberLine		= myGetItemCtlValue(dp, i++);
		ppsi->Parabola			= myGetItemCtlValue(dp, i++);
		ppsi->Line45			= myGetItemCtlValue(dp, i++);
		ppsi->Box				= myGetItemCtlValue(dp, i++);
		ppsi->IterationLines	= myGetItemCtlValue(dp, i++);
		*pShowAllIterLines		= myGetItemCtlValue(dp, i++);
		ppsi->Arrows			= myGetItemCtlValue(dp, i++);
		ppsi->SameSizeAxes		= myGetItemCtlValue(dp, i++);
		DialogReturn(dp, TRUE);
	}
	else DialogReturn(dp, FALSE);
}

Boolean CheckEmbeddedPSDialog(DialogPtr dp, int itemNo)
{
	long	tmpL;
	Boolean	GoodValue = TRUE;

	if (itemNo != OK) return(TRUE);

	if ( myGetLongParamItem(dp, PSs_FirstTextItem, &tmpL,
			"\PEmbedding",
			(long) MIN_Embedding, (long) MAX_Embedding) )
		EPS_Embedding = (int) tmpL;
	else
		SelectBadField(dp, PSs_FirstTextItem, &GoodValue);

	return(GoodValue);
}

Boolean DoEmbeddedPSDialog(int *pEmbedding, EmbeddedPSParams *pepsi)
{
	DialogPtr	dp;
	int			i;

	dp = GetNewDialog(EmbeddedPSDLogID, NIL, -1L);

	mySetItemLong(dp, PSs_FirstTextItem, (long) *pEmbedding);
	i= PSs_FirstCheckBox;
	mySetItemCtlValue(dp, i++, pepsi->Polynomial2n);
	mySetItemCtlValue(dp, i++, pepsi->Line45);
	mySetItemCtlValue(dp, i++, pepsi->Box);
	mySetItemCtlValue(dp, i++, pepsi->SameSizeAxes);

	if (HandleDialog(dp, PSs_FirstTextItem, Cancel, UserOutlineItemNum,
			NIL, NIL, &CheckEmbeddedPSDialog) == OK) {
		*pEmbedding = EPS_Embedding;
		i= PSs_FirstCheckBox;
		pepsi->Polynomial2n		= myGetItemCtlValue(dp, i++);
		pepsi->Line45			= myGetItemCtlValue(dp, i++);
		pepsi->Box				= myGetItemCtlValue(dp, i++);
		pepsi->SameSizeAxes		= myGetItemCtlValue(dp, i++);
		DialogReturn(dp, TRUE);
	}
	else DialogReturn(dp, FALSE);
}

Boolean CheckOrbitDialog(DialogPtr dp, int itemNo)
{
	long	tmpL;
	Boolean	NumbersOK = TRUE;

	if (itemNo != OK) return(TRUE);

	if ( myGetLongParamItem(dp, O_FirstET, &tmpL,
			"\PDot Size", (long) MIN_DotSize, (long) MAX_DotSize) )
		O_DotSize = (int) tmpL;
	else
		SelectBadField(dp, O_FirstET, &NumbersOK);

	if ( myGetLongParamItem(dp, O_FirstET+1, &tmpL,
			"\PHorizontal Spacing", (long) MIN_HorSpace, (long) MAX_HorSpace) )
		O_HorSpace = (int) tmpL;
	else
		SelectBadField(dp, O_FirstET+1, &NumbersOK);

	return(NumbersOK);
}

Boolean DoOrbitDialog(int *pDotSize, int *pHorSpace)
{
	DialogPtr	dp;

	dp = GetNewDialog(OrbitDLogID, NIL, -1L);

	mySetItemLong(dp, O_FirstET, (long) *pDotSize);
	mySetItemLong(dp, O_FirstET+1, (long) *pHorSpace);

	if (HandleDialog(dp, O_FirstET, Cancel, UserOutlineItemNum,
			NIL, NIL, &CheckOrbitDialog) == OK) {
		*pDotSize = O_DotSize;
		*pHorSpace = O_HorSpace;
		DialogReturn(dp, TRUE);
	}
	else DialogReturn(dp, FALSE);
}

Boolean DoShowWindowDialog(int windex)
{
	DialogPtr	dp;
	char		*p;

	if (SHIFT_DOWN)		return(TRUE);
	if (OPTION_DOWN)	return(FALSE);

	switch (windex) {
		case BifD:	p = "\PBifurcation Diagram";	break;
	}
	ParamText(p, "\P", "\P", "\P");

	dp = GetNewDialog(ShowWindowDLogID, NIL, -1L);
	if (HandleDialog(dp, NO_EDIT_TEXT,
			Cancel, UserOutlineItemNum, NIL, NIL, NIL) == OK)
		DialogReturn(dp, TRUE)
	else
		DialogReturn(dp, FALSE)
}



		/*	----------------------------------------------	*/
		/*	----------	System-related dialogs	----------	*/
		/*	----------------------------------------------	*/

StartPrintingDialog()
{
	PrintingDialog = GetNewDialog(PrintingDLogID, NIL, -1L);
	ShowWindow(PrintingDialog);
	DrawDialog(PrintingDialog);
}

Boolean HandlePrintingDialog()
{
	EventRecord	theEvent;
	int			itemHit;
	DialogPtr	theDialog;

	if (GetNextEvent((mDownMask | keyDownMask), &theEvent)) {
		if ( (theEvent.modifiers & cmdKey) && 
				((char) (theEvent.message & charCodeMask) == '.') )
			return(FALSE);
		if (IsDialogEvent(&theEvent))
			if (DialogSelect(&theEvent, &theDialog, &itemHit))
				if (theDialog == PrintingDialog)
					if (itemHit == 1)
						return(FALSE);
	}
	return(TRUE);
}

StopPrintingDialog()
{
	HideWindow(PrintingDialog);
	DisposDialog(PrintingDialog);
}

Boolean DoQuitWarningDialog(Boolean *pWantsToSave, int windex)
{
	DialogPtr	dp;
	Str255		TextWindowTitle;
	int			CloseItem;
	char		*ps;

	if (SuppressSaveWarnings) {
		*pWantsToSave = FALSE;
		return(TRUE);
	}

	if (SHIFT_DOWN || OPTION_DOWN) {
		*pWantsToSave = SHIFT_DOWN;
		return(TRUE);
	}

	dp = GetNewDialog(QuitWarningDLogID, NIL, -1L);
	switch (windex) {
		case BifDiagram:
				ps = GetSaveFName(BifD);
				if (ps[0] == 0)
					ParamText("\PBifurcation Diagram", "\P", "\P", "\P");
				else
					ParamText("\PBifurcation Diagram", "\P ", ps, "\P");
				break;
		default:
				GetTextTitle(CP(TextWindowTitle));
				ParamText("\P", CP(TextWindowTitle), "\P", "\P");
				break;
	}
	CloseItem = HandleDialog(dp, NO_EDIT_TEXT,
			Cancel+1, UserOutlineItemNum+1, NIL, NIL, NIL);
	switch (CloseItem) {
		case OK:
		case OK+1:
			*pWantsToSave = (CloseItem == OK);
			DialogReturn(dp, TRUE)
		case OK+2:
			DialogReturn(dp, FALSE)
	}
}

Boolean DoSaveWarningDialog(Boolean *pWantsToSave, int windex)
{
	DialogPtr	dp;
	char		*p;
	int			CloseItem;

	if (SuppressSaveWarnings) {
		*pWantsToSave = FALSE;
		return(TRUE);
	}

	if (SHIFT_DOWN || OPTION_DOWN) {
		*pWantsToSave = SHIFT_DOWN;
		return(TRUE);
	}

	dp = GetNewDialog(SaveWarningDLogID, NIL, -1L);
	switch (windex) {
		case BifDiagram:		p = "\Pbifurcation diagram ";	break;
		default:				p = "\P";				break;
	}
	ParamText(p, "\P", "\P", "\P");

	CloseItem = HandleDialog(dp, NO_EDIT_TEXT,
			Cancel+1, UserOutlineItemNum+1, NIL, NIL, NIL);
	switch (CloseItem) {
		case OK:
		case OK+1:
			*pWantsToSave = (CloseItem == OK);
			DialogReturn(dp, TRUE)
		case OK+2:
			DialogReturn(dp, FALSE)
	}
}

Boolean DoBackgroundDialog(Boolean *pRunInBackground,
		Boolean *pRunQuickly, Boolean *pUseSoundInBG)
{
	DialogPtr	dp;
	int			FirstCheckBox = 11;	/*	"constant"	*/
	int			i;

	dp = GetNewDialog(BackgroundDLogID, NIL, -1L);
	ParamText("\PQuadratic Map", "\P", "\P", "\P");

	i = FirstCheckBox;
	mySetItemCtlValue(dp, i++, *pRunInBackground);
	mySetItemCtlValue(dp, i++, !*pRunQuickly);
	mySetItemCtlValue(dp, i++, *pRunQuickly);
	mySetItemCtlValue(dp, i++, *pUseSoundInBG);

	if (HandleDialog(dp, NO_EDIT_TEXT,
			Cancel+1, UserOutlineItemNum, NIL, NIL, NIL) == OK) {
		i = FirstCheckBox;
		*pRunInBackground = myGetItemCtlValue(dp, i++);
		i++;
		*pRunQuickly = myGetItemCtlValue(dp, i++);
		*pUseSoundInBG = myGetItemCtlValue(dp, i++);
		DialogReturn(dp, TRUE);
	}
	else
		DialogReturn(dp, FALSE);
}

	/*	----------------------------------------------	*/
	/*	------	Standard File Package Routines,	------	*/
	/*	------	 customized for Quadratic Map	------	*/
	/*	----------------------------------------------	*/

pascal int MakeStdFileOutline(int item, DialogPtr StdFileDialog)
{
	if (FirstAccessStdFileDLog) {		/*	(item == -1)?	*/
		if (!System7OrLater)
			MakeUserItemOutline(StdFileDialog, StdFileOutlineItemNum);
		FirstAccessStdFileDLog = FALSE;
	}
}

pascal int DelimiterDLog(int itemNo, DialogPtr SaveAsDialog)
{
	int				type;
	Handle			item;
	Rect			box;
	ControlHandle	PUMch;

	if (FirstAccessStdFileDLog) {
		if (!System7OrLater)
			MakeUserItemOutline(SaveAsDialog, StdFileOutlineItemNum);
		FirstAccessStdFileDLog = FALSE;

		GetDItem(SaveAsDialog, DelimButtonItemNum, &type, &item, &box);
		if (type & ctrlItem) {
			DisposeControl((ControlHandle) item);
			PUMch = NewControl(SaveAsDialog, &box, "\PDelimiter:", TRUE,
					0, 0, 0, PopUpCDEFID << 4,
					(((long) DelimPopUpMenuID) << 16) | (DelimiterIndex+1));
			box = (**PUMch).contrlRect;
			SetDItem(SaveAsDialog, DelimButtonItemNum, ctrlItem+resCtrl,
					(Handle) PUMch, &box);
		}
	}
	if (itemNo == DelimButtonItemNum) {
		GetDItem(SaveAsDialog, DelimButtonItemNum, &type, &item, &box);
		DelimiterIndex = GetCtlValue((ControlHandle) item) - 1;
	}
	return(itemNo);
}

/*	From DialogUtils.c	*/
extern	Point	StdFilePoint;

Boolean QMSaveAsDialog(char *prompt, int *vRef, char *fname)
{
	SFReply		mySFReply;
	DialogTHndl	dth;
	GrafPtr		CurPort;
	int			theFont, theSize;
	Rect		DelimButtonRect = {0,0,0,0};
	Rect		OutlineRect = {0,0,0,0};
	int			OldDelimiterIndex = DelimiterIndex;

	if ((dth = (DialogTHndl) GetFreshResource('DLOG', putDlgID)) == 0L) {
		ERROR("\PSystem Error: Unable to create Save As... dialog box");
		return(FALSE);
	}

	if (!System7OrLater)
		if ((StdFileOutlineItemNum = AddDItem(putDlgID, TRUE, userItem,
				&OutlineRect, "\P")) == 0 ) {
			ERROR("\PSystem Error: Unable to create Save As... dialog box");
			return(FALSE);
		}
	
	(**dth).boundsRect.bottom += 25;
	GetPort(&CurPort);
	theFont = CurPort->txFont;	theSize = CurPort->txSize;
	TextFont(systemFont);		TextSize(12);
	OffsetRect(&DelimButtonRect, 14, (**dth).boundsRect.bottom - 25);
	TextFont(theFont);			TextSize(theSize);

	if ((DelimButtonItemNum = AddDItem(putDlgID, FALSE, ctrlItem+radCtrl,
			&DelimButtonRect, "\P")) == 0 ) {
		ERROR("\PSystem Error: Unable to create Save As... dialog box");
		return(FALSE);
	}

	FirstAccessStdFileDLog = TRUE;
	SFPPutFile(StdFilePoint, prompt, fname, &DelimiterDLog, &mySFReply,
			putDlgID, &CommandKeyFilter);
	ReleaseDialogResources(dth);
	if (!mySFReply.good) {
		DelimiterIndex = OldDelimiterIndex;
		return(FALSE);
	}
	*vRef = mySFReply.vRefNum;
	sEquate(fname, CP(mySFReply.fName));
	return(TRUE);
}

Boolean QMSpecialSaveAsDialog(char *prompt, int *vRef, char *fname)
{
	SFReply		mySFReply;
	DialogTHndl	dth;
	Rect		OutlineRect = {0,0,0,0};

	if ((dth = (DialogTHndl) GetFreshResource('DLOG', putDlgID)) == 0L) {
		ERROR("\PSystem Error: Unable to create Save As... dialog box");
		return(FALSE);
	}

	if (!System7OrLater)
		if ((StdFileOutlineItemNum = AddDItem(putDlgID, TRUE, userItem,
				&OutlineRect, "\P")) == 0 ) {
			ERROR("\PSystem Error: Unable to create Save As... dialog box");
			return(FALSE);
		}

	if (!System7OrLater) {
		FirstAccessStdFileDLog = TRUE;
		SFPPutFile(StdFilePoint, prompt, fname, &MakeStdFileOutline, &mySFReply,
				putDlgID, &CommandKeyFilter);
	}
	else
		SFPPutFile(StdFilePoint, prompt, fname, NIL, &mySFReply,
				putDlgID, &CommandKeyFilter);
	ReleaseDialogResources(dth);
	if (!mySFReply.good)
		return(FALSE);
	*vRef = mySFReply.vRefNum;
	sEquate(fname, CP(mySFReply.fName));
	return(TRUE);
}

Boolean QMOpenDialog(OSType *OpenFileTypes, int NumOSTypes, int *vRef,
		char *fname, OSType *fType)
{
	SFReply		mySFReply;
	SFTypeList	mySFTypeList;
	int			i;
	DialogTHndl	dth;
	GrafPtr		CurPort;
	int			theFont, theSize;
	Rect		OutlineRect  = {0,0,0,0};

	if ((dth = (DialogTHndl) GetFreshResource('DLOG', getDlgID)) == 0L) {
		ERROR("\PSystem Error: Unable to create Open... dialog box");
		return(FALSE);
	}

	if (!System7OrLater)
		if ((StdFileOutlineItemNum = AddDItem(getDlgID, TRUE, userItem,
				&OutlineRect, "\P")) == 0 ) {
			ERROR("\PSystem Error: Unable to create Open... dialog box");
			return(FALSE);
		}

	for(i=0; i<NumOSTypes; ++i)
		mySFTypeList[i] = *OpenFileTypes++;
/*
	if (TRUE)
		SFPGetFile(StdFilePoint, "\P", 0L, NumOSTypes, &mySFTypeList, &GWO_Check,
				&mySFReply, getDlgID, &CommandKeyFilter);
	else
*/

	SetCheckCharIfNoCommand(FALSE);
	if (!System7OrLater) {
		FirstAccessStdFileDLog = TRUE;
		SFPGetFile(StdFilePoint, "\P", 0L,NumOSTypes, &mySFTypeList, &MakeStdFileOutline,
				&mySFReply, getDlgID, &CommandKeyFilter);
	}
	else
		SFPGetFile(StdFilePoint, "\P", 0L,NumOSTypes, &mySFTypeList, NIL,
				&mySFReply, getDlgID, &CommandKeyFilter);
	SetCheckCharIfNoCommand(TRUE);
	ReleaseDialogResources(dth);
	if (!mySFReply.good)
		return(FALSE);

	*vRef = mySFReply.vRefNum;
	sEquate(fname, CP(mySFReply.fName));
	*fType = mySFReply.fType;
	return(TRUE);
}
