/*	qmGraphUtils.c
 *
 *	Utilities used mainly by qmGraphics, to handle
 *	Quadratic Map's graphics requirements.
 *
 *	-TA
 */

#include <limits.h>
#include "qmPrototypes.h"
#include "qmGraphics.h"

#define	LabelFieldWidth		6
#define	LabelPrecision		(LabelFieldWidth-2)
#define	DashedLinePatsName	"\PDashed Line"
#define	MAX_INT_LONG		((long) INT_MAX)
#define	SqrtMaxPosInt		181		/* "sqrt" 32767 (really 181.01657...) */

				/*	Global external vars from qm.c	*/
extern	pIterationVars		pCurrentVars;

	/*	External variables from qmGraphics.c */
extern	CursHandle	MagnifyingGlass;

	/*	For fixed point operations	*/
Fixed	PSs_xscale_fxd, PSs_yscale_fxd, PSs_xoffset_fxd, PSs_yoffset_fxd, PSs_lambda_fxd;

			/*	To force ThePatterns to line up on even-byte boundary	*/
long	KludgeForPatterns[2*NumDashedLinePatterns];
Pattern	*ThePatterns = (Pattern *) &KludgeForPatterns;

Boolean	UseDashedLine;


WriteVar(char *s, double f)
{
	DrawString(s);
	DrawFloat(f, LabelFieldWidth, LabelPrecision, TRUE);
	DrawString("\P	");
}

DrawXn(char *s)
{
	if (RealFont(monaco, 12)) {
		TextSize(12);
		DrawChar('x');
	}
	else
		DrawChar('X');
	TextSize(9);
	Move(0, 3);
	DrawString(s);
}


	/*	---------------------------------------------------- */
	/*	-------	Fixed-point routines for creating	-------- */
	/*	-------		parabolas, 2^n polynomials		-------- */
	/*	---------------------------------------------------- */

InitFixedQMVars(double theLambda, double xScale, double yScale,
	DoubleRect *pgr, Boolean xGA0, Boolean yGA0)
{
	PSs_xscale_fxd = X2Fix(&xScale);
	PSs_yscale_fxd = X2Fix(&yScale);
	PSs_lambda_fxd = X2Fix(&theLambda);
	if (xGA0)	PSs_xoffset_fxd = 0L;
	else		PSs_xoffset_fxd = ((*pgr)[xmax_i] > 0) ?
						X2Fix(&(*pgr)[xmin_i]) : X2Fix(&(*pgr)[xmax_i]);

	if (yGA0)	PSs_yoffset_fxd = 0L;
	else		PSs_yoffset_fxd = ((*pgr)[ymax_i] > 0) ?
						X2Fix(&(*pgr)[ymin_i]) : X2Fix(&(*pgr)[ymax_i]);
}

Fixed QMFixed(Fixed l_fxd, Fixed x_fxd)
{
	return(FixMul(l_fxd, FixMul(x_fxd,(1L<<16)-x_fxd)));
}

int FixedQM(int x, int n)
{
	Fixed	tmpfxd;
	int		tmpi, i;

	tmpfxd = FixDiv(Long2Fix((long) x), PSs_xscale_fxd) + PSs_xoffset_fxd;
	for(i=1; i<=n; ++i)
		tmpfxd = QMFixed(PSs_lambda_fxd, tmpfxd);
	tmpfxd = FixMul(tmpfxd-PSs_yoffset_fxd, PSs_yscale_fxd);
	tmpi = FixRound(tmpfxd);
	return(tmpi);
}


	/* ----------------------------------------- */
	/* -------	Axis-drawing Routines	-------- */
	/* ----------------------------------------- */

DrawHAxis(Boolean ShowZero, double theXMin, double theXMax,
		Rect *pgr, Point Origin, double scale, double HashDist,
		int HashNum, Boolean ExactHashes, int Hash1Size, int Hash2Size,
		int fw, int spacing, int extra_spacing, int text_spacing,
		int axis_spacing)
{
	int			minCoord, maxCoord;
	double		ScaledMax, Multiplier, MinMultRange, HashMeasurement;
	int			ScaledMaxInt, HashInc, theHashSize;
	int			HashCoord, minWriteCoord, maxWriteCoord;
	int			min0WriteCoord, max0WriteCoord, otherWriteCoord;
	Boolean		HashPos, HashNeg;
	int			HashSizePos, HashSizeNeg;
	RgnHandle	tmpClipRgn;
	Str255		s;
	int			swidth;
	GrafPtr		gp;
	int			tSize;
	int			sign;
	double		maxabs, minabs;
	int			*pMinWC, *pMaxWC, *pChangeWC;
	int			i;
/*	1.01 addition
 */
	double		oHashMeasurement;
/*	end, 1.01 addition
 */

	GetPort(&gp);
	tSize = gp->txSize;
	minCoord =  pgr->left;
	maxCoord = pgr->right-1;

	MoveTo(minCoord, Origin.v);
	Line(maxCoord - minCoord, 0);

	if (HashDist == 0.0) {
		ScaledMax = theXMax - theXMin;
		MinMultRange = (ExactHashes) ? HashNum*HashNum : HashNum;
		while (ScaledMax > 10.0*MinMultRange)	ScaledMax /= 10.0;
		while (ScaledMax < MinMultRange)	ScaledMax *= 10.0;
		Multiplier = (theXMax - theXMin) / ScaledMax;
		ScaledMaxInt = RoundToInt(ScaledMax);
		if (ExactHashes)	HashInc = ScaledMaxInt / (2*HashNum) + 1;
		else				HashInc = ScaledMaxInt / HashNum;
		HashDist = HashInc*Multiplier;
	}

	HashPos = (Origin.v > pgr->top);
	HashNeg = (Origin.v+1 < pgr->bottom);
	theHashSize = (HashPos && HashNeg) ? Hash2Size : Hash1Size;
	if (HashPos)	HashSizePos = MIN(theHashSize, Origin.v - pgr->top);
	if (HashNeg)	HashSizeNeg = MIN(theHashSize, pgr->bottom-Origin.v-1);
	if (HashPos)
		otherWriteCoord = Origin.v +
				( (HashNeg) ? HashSizeNeg : extra_spacing ) +
				tSize + spacing;
	else
		otherWriteCoord = Origin.v - spacing - extra_spacing;
/*		Only if clip below before hashes
	tmpClipRgn = NewRgn();
	GetClip(tmpClipRgn);
*/

	myFToStrNoPad(theXMin, CP(s), fw, fw-1);
	MoveTo(minCoord, otherWriteCoord);
	DrawJustString(teJustCenter, CP(s));
	minWriteCoord = minCoord + StringWidth(CP(s))/2 + text_spacing;

	myFToStrNoPad(theXMax, CP(s), fw, fw-1);
	MoveTo(maxCoord, otherWriteCoord);
	DrawJustString(teJustCenter, CP(s));
	maxWriteCoord = maxCoord - StringWidth(CP(s))/2 - text_spacing;

	if ( (ShowZero) && (theXMax*theXMin < 0) && !(HashPos && HashNeg) &&
			( (Origin.h - CharWidth('0')/2 > minWriteCoord) &&
			(Origin.h+CharWidth('0')/2+CharWidth('0')%2 < maxWriteCoord) ) ) {
		MoveTo(Origin.h, otherWriteCoord);
		Move(-CharWidth('0')/2, 0);
		DrawChar('0');
		max0WriteCoord = Origin.h - CharWidth('0')/2 - text_spacing;
		min0WriteCoord = Origin.h + CharWidth('0')/2 + text_spacing;
	}
	else if ( (theXMax*theXMin < 0) && (HashPos && HashNeg) ) {
		max0WriteCoord = Origin.h - axis_spacing;
		min0WriteCoord = minWriteCoord;
	}
	else {
		max0WriteCoord = maxWriteCoord;
		min0WriteCoord = minWriteCoord;
	}

		/*	Label negative part of axisboth together	*/
	for(i=0; i<=2; ++i) {	/*	i = 0 negative, i = 1 positive	*/
		sign = (i==0) ? -1 : 1;
		HashDist = sign * fabs(HashDist);
		maxabs = (i==0) ? theXMin : theXMax;
		minabs = (i==0) ? theXMax : theXMin;
		pMinWC = (i==0) ? &minWriteCoord : &min0WriteCoord;
		pMaxWC = (i==0) ? &max0WriteCoord : &maxWriteCoord;
		pChangeWC = (i==0) ? pMaxWC : pMinWC;
		if (maxabs * sign <= 0.0) continue;

		if (minabs * sign <= 0.0)
			HashMeasurement = HashDist;
		else
/* 1.01 change, 5/26/92
			HashMeasurement = ((long) (minabs/HashDist) + 1) * HashDist;
*/
			HashMeasurement = (floor(minabs/HashDist) + 1) * HashDist;
		while(sign*(HashMeasurement-maxabs) < -(theXMax-theXMin)*1e-10) {
			myFToStrNoPad(HashMeasurement, CP(s), fw, fw-1);
			swidth = StringWidth(CP(s));
			if (theXMin * theXMax <= 0)
				HashCoord = Origin.h + RoundToInt(HashMeasurement*scale);
			else {
/*
				HashCoord = pgr->left +
						RoundToInt((HashMeasurement-theXMin)*scale);
*/
				if (theXMax > 0)
					HashCoord = Origin.h +
							RoundToInt((HashMeasurement-theXMin)*scale);
				else
					HashCoord = Origin.h +
							RoundToInt((HashMeasurement-theXMax)*scale);
			}

				/*	Need to do ClipRect(pgr) before hashes,
				 *	SetClip(tmpClipRgn) after?	*/
			MoveTo(HashCoord, Origin.v);
			if (HashPos)	Line(0, -HashSizePos);
			MoveTo(HashCoord, Origin.v);
			if (HashNeg)	Line(0, HashSizeNeg);

			if ( (HashCoord - swidth/2 > *pMinWC) &&
					(HashCoord + swidth/2 + swidth%2 < *pMaxWC) ) {
				MoveTo(HashCoord, otherWriteCoord);
				DrawJustString(teJustCenter, CP(s));
				*pChangeWC = HashCoord + sign*(swidth/2 + text_spacing);
			}
/* 1.01 change (below)--oHashMeasurement
			HashMeasurement += HashDist;
 */
			oHashMeasurement = HashMeasurement;
			HashMeasurement += HashDist;
			if (oHashMeasurement == HashMeasurement)
					/* HashDist less than one significant bit of HashMeasurement	*/
				break;
		}
	}

		/*	Label positive part of axis	*/
/*
	if (theXMax > 0.0) {
		if (theXMin <= 0.0)
			HashMeasurement = HashDist;
		else
			HashMeasurement = ((long) (theXMin/HashDist) + 1) * HashDist;
		while(HashMeasurement-theXMax < -(theXMax-theXMin)*1e-10) {
			myFToStrNoPad(HashMeasurement, CP(s), fw, fw-1);
			if (theXMin * theXMax <= 0)
				HashCoord = Origin.h + RoundToInt(HashMeasurement*scale);
			else
				HashCoord = pgr->left +
						RoundToInt((HashMeasurement-theXMin)*scale);
			MoveTo(HashCoord, Origin.v);
			if (HashPos)	Line(0, -HashSizePos);
			MoveTo(HashCoord, Origin.v);
			if (HashNeg)	Line(0, HashSizeNeg);
			if ( (HashCoord - StringWidth(CP(s))/2 > min0WriteCoord) &&
					(HashCoord + StringWidth(CP(s))/2 < maxWriteCoord) ) {
				MoveTo(HashCoord, otherWriteCoord);
				DrawJustString(teJustCenter, CP(s));
				min0WriteCoord = HashCoord + StringWidth(CP(s))/2 + text_spacing;
			}
			HashMeasurement += HashDist;
		}
	}
*/

}

DrawVAxis(Boolean ShowZero, double theXMin, double theXMax,
		Rect *pgr, Point Origin, double scale, double HashDist,
		int HashNum, Boolean ExactHashes, int Hash1Size, int Hash2Size,
		int fw, int spacing, int extra_spacing, int text_spacing,
		int axis_spacing)
{
	int			minCoord, maxCoord;
	double		ScaledMax, Multiplier, MinMultRange, HashMeasurement;
	int			ScaledMaxInt, HashInc, theHashSize;
	int			HashCoord, minWriteCoord, maxWriteCoord;
	int			min0WriteCoord, max0WriteCoord, otherWriteCoord;
	Boolean		HashPos, HashNeg;
	int			HashSizePos, HashSizeNeg;
	RgnHandle	tmpClipRgn;
	Str255		s;
	int			swidth;
	GrafPtr		gp;
	int			tSize;
	int			sign;
	double		maxabs, minabs;
	int			*pMinWC, *pMaxWC, *pChangeWC;
	int			i;
	int			theJust;
/*	1.01 addition
 */
	double		oHashMeasurement;
/*	end, 1.01 addition
 */

	GetPort(&gp);
	tSize = gp->txSize;
	minCoord = pgr->bottom-1;
	maxCoord = pgr->top;

	MoveTo(Origin.h, minCoord);
	Line(0, maxCoord - minCoord);

	if (HashDist == 0.0) {
		ScaledMax = theXMax - theXMin;
		MinMultRange = (ExactHashes) ? HashNum*HashNum : HashNum;
		while (ScaledMax > 10.0*MinMultRange)	ScaledMax /= 10.0;
		while (ScaledMax < MinMultRange)	ScaledMax *= 10.0;
		Multiplier = (theXMax - theXMin) / ScaledMax;
		ScaledMaxInt = RoundToInt(ScaledMax);
		if (ExactHashes)	HashInc = ScaledMaxInt / (2*HashNum) + 1;
		else				HashInc = ScaledMaxInt / HashNum;
		HashDist = HashInc*Multiplier;
	}

	HashPos = (Origin.h+1 < pgr->right);
	HashNeg = (Origin.h > pgr->left);
	theHashSize = (HashPos && HashNeg) ? Hash2Size : Hash1Size;
	if (HashPos)	HashSizePos = MIN(theHashSize, pgr->right-Origin.h-1);
	if (HashNeg)	HashSizeNeg = MIN(theHashSize, Origin.h - pgr->left);
	if (HashPos)
		otherWriteCoord = Origin.h -
				( (HashNeg) ? HashSizeNeg : extra_spacing ) - spacing;
	else
		otherWriteCoord = Origin.h + spacing + extra_spacing;
	theJust = (HashPos) ? teJustRight : teJustLeft;

	myFToStrNoPad(theXMin, CP(s), fw, fw-1);
	MoveTo(otherWriteCoord, minCoord);
	Move(0, tSize/2);
	DrawJustString(theJust, CP(s));
	minWriteCoord = minCoord - tSize/2 - tSize%2 - text_spacing;

	myFToStrNoPad(theXMax, CP(s), fw, fw-1);
	MoveTo(otherWriteCoord, maxCoord);
	Move(0, tSize/2);
	DrawJustString(theJust, CP(s));
	maxWriteCoord = maxCoord + tSize/2 + tSize%2 + text_spacing;

	if ( (ShowZero) && (theXMax*theXMin < 0) && !(HashPos && HashNeg) &&
			( (Origin.v + tSize/2 < minWriteCoord) &&
			(Origin.v - tSize/2 - tSize%2 > maxWriteCoord) ) ) {
		MoveTo(otherWriteCoord, Origin.v);
		Move(0, tSize/2);
		DrawJustString(theJust, "\P0");
		max0WriteCoord = Origin.v + tSize/2 + tSize%2 + text_spacing;
		min0WriteCoord = Origin.v - tSize/2 - tSize%2 - text_spacing;
	}
	else if ( (theXMax*theXMin < 0) && (HashPos && HashNeg) ) {
		max0WriteCoord = Origin.v + axis_spacing;
		min0WriteCoord = minWriteCoord;
	}
	else {
		max0WriteCoord = maxWriteCoord;
		min0WriteCoord = minWriteCoord;
	}

		/*	Label negative part of axisboth together	*/
	for(i=0; i<=2; ++i) {	/*	i = 0 negative, i = 1 positive	*/
		sign = (i==0) ? -1 : 1;
		HashDist = sign * fabs(HashDist);
		maxabs = (i==0) ? theXMin : theXMax;
		minabs = (i==0) ? theXMax : theXMin;
		pMinWC = (i==0) ? &minWriteCoord : &min0WriteCoord;
		pMaxWC = (i==0) ? &max0WriteCoord : &maxWriteCoord;
		pChangeWC = (i==0) ? pMaxWC : pMinWC;
		if (maxabs * sign <= 0.0) continue;

		if (minabs * sign <= 0.0)
			HashMeasurement = HashDist;
		else
/* 1.01 change, 5/26/92
			HashMeasurement = ((long) (minabs/HashDist) + 1) * HashDist;
*/
			HashMeasurement = (floor(minabs/HashDist) + 1) * HashDist;
		while(sign*(HashMeasurement-maxabs) < -(theXMax-theXMin)*1e-10) {
			myFToStrNoPad(HashMeasurement, CP(s), fw, fw-1);
			swidth = StringWidth(CP(s));
			if (theXMin * theXMax <= 0)
				HashCoord = Origin.v - RoundToInt(HashMeasurement*scale);
			else {
/*
				HashCoord = pgr->bottom -
						RoundToInt((HashMeasurement-theXMin)*scale);
*/
				if (theXMax > 0)
					HashCoord = Origin.v -
							RoundToInt((HashMeasurement-theXMin)*scale);
				else
					HashCoord = Origin.v -
							RoundToInt((HashMeasurement-theXMax)*scale);
			}

				/*	Need to do ClipRect(pgr) before hashes,
				 *	SetClip(tmpClipRgn) after?	*/
			MoveTo(Origin.h, HashCoord);
			if (HashPos)	Line(HashSizePos, 0);
			MoveTo(Origin.h, HashCoord);
			if (HashNeg)	Line(-HashSizeNeg, 0);

			if ( (HashCoord + tSize/2 < *pMinWC) &&
					(HashCoord - tSize/2 - tSize%2 > *pMaxWC) ) {
				MoveTo(otherWriteCoord, HashCoord);
				Move(0, tSize/2);
				DrawJustString(theJust, CP(s));
				*pChangeWC = HashCoord - sign*(tSize/2 + tSize%2 + text_spacing);
			}
/* 1.01 change (below)--oHashMeasurement
			HashMeasurement += HashDist;
 */
			oHashMeasurement = HashMeasurement;
			HashMeasurement += HashDist;
			if (oHashMeasurement == HashMeasurement)
					/* HashDist less than one significant bit of HashMeasurement	*/
				break;
		}
	}

}

LabelOrbitVAxis(double theXMin, double theXMax,
		Rect *pgr, Point Origin, int HAxisPos, double scale, double HashDist,
		int HashNum, int fw, int spacing, int extra_spacing,
		int text_spacing, int ods, Boolean DoZeroHash, RgnHandle GraphClipRgn)
{
	int			minCoord, maxCoord;
	double		ScaledMax, Multiplier, MinMultRange, HashMeasurement;
	int			ScaledMaxInt, HashInc, theHashSize;
	int			HashCoord, minWriteCoord, maxWriteCoord;
	int			min0WriteCoord, max0WriteCoord, otherWriteCoord;
	RgnHandle	tmpClipRgn;
	Str255		s;
	int			swidth;
	GrafPtr		gp;
	int			tSize;
	int			sign;
	double		maxabs, minabs;
	int			*pMinWC, *pMaxWC, *pChangeWC;
	int			i;
	int			theJust;
	Rect		tmpRect;
	RgnHandle	tmpRgn;
	double		RealScale;
	int			RealOriginV = (HAxisPos == -1) ? Origin.v : HAxisPos;
/*	1.01 addition
 */
	double		oHashMeasurement;
	Boolean		NoHashes = FALSE;
/*	end, 1.01 addition
 */

	RealScale = (pgr->bottom - pgr->top-1) / (theXMax - theXMin);

	GetPort(&gp);
	tSize = gp->txSize;
	minCoord = pgr->bottom-1;
	maxCoord = pgr->top;

	if (HashDist == 0.0) {

/*	1.01 addition
 */
		if (HashNum == 0)
			NoHashes = TRUE;
		else {
/*	end, 1.01 addition
 */

			ScaledMax = theXMax - theXMin;
			MinMultRange = HashNum;
			while (ScaledMax > 10.0*MinMultRange)	ScaledMax /= 10.0;
			while (ScaledMax < MinMultRange)	ScaledMax *= 10.0;
			Multiplier = (theXMax - theXMin) / ScaledMax;
			ScaledMaxInt = RoundToInt(ScaledMax);
			HashInc = ScaledMaxInt / HashNum;
			HashDist = HashInc*Multiplier;

/*	1.01 addition
 */
		}
/*	end, 1.01 addition
 */

	}

	theHashSize = HashSize;
	otherWriteCoord = Origin.h - extra_spacing - spacing;
	theJust = teJustRight;

	myFToStrNoPad(theXMin, CP(s), fw, fw-1);
	MoveTo(otherWriteCoord, minCoord);
	Move(0, tSize/2);
	DrawJustString(theJust, CP(s));
	minWriteCoord = minCoord - tSize/2 - tSize%2 - text_spacing;

	myFToStrNoPad(theXMax, CP(s), fw, fw-1);
	MoveTo(otherWriteCoord, maxCoord);
	Move(0, tSize/2);
	DrawJustString(theJust, CP(s));
	maxWriteCoord = maxCoord + tSize/2 + tSize%2 + text_spacing;

	if ( (theXMax*theXMin < 0) &&
			( (RealOriginV + tSize/2 < minWriteCoord) &&
			(RealOriginV - tSize/2 - tSize%2 > maxWriteCoord) ) ) {
		if (DoZeroHash) {
			MoveTo(Origin.h, RealOriginV);
			Line(theHashSize, 0);
		}
		MoveTo(otherWriteCoord, RealOriginV);
		Move(0, tSize/2);
		DrawJustString(theJust, "\P0");
		max0WriteCoord = RealOriginV + tSize/2 + tSize%2 + text_spacing;
		min0WriteCoord = RealOriginV - tSize/2 - tSize%2 - text_spacing;
	}
	else {
		max0WriteCoord = maxWriteCoord;
		min0WriteCoord = minWriteCoord;
	}

/*	1.01 addition
 */
	if (NoHashes)
		return;
/*	end, 1.01 addition
 */

		/*	Label negative part of axisboth together	*/
	for(i=0; i<=2; ++i) {	/*	i = 0 negative, i = 1 positive	*/
		sign = (i==0) ? -1 : 1;
		HashDist = sign * fabs(HashDist);
		maxabs = (i==0) ? theXMin : theXMax;
		minabs = (i==0) ? theXMax : theXMin;
		pMinWC = (i==0) ? &minWriteCoord : &min0WriteCoord;
		pMaxWC = (i==0) ? &max0WriteCoord : &maxWriteCoord;
		pChangeWC = (i==0) ? pMaxWC : pMinWC;
		if (maxabs * sign <= 0.0) continue;

		if (minabs * sign <= 0.0)
			HashMeasurement = HashDist;
		else
/* 1.01 change, 5/26/92
			HashMeasurement = ((long) (minabs/HashDist) + 1) * HashDist;
*/
			HashMeasurement = (floor(minabs/HashDist) + 1) * HashDist;
		while(sign*(HashMeasurement-maxabs) < -(theXMax-theXMin)*1e-10) {
			myFToStrNoPad(HashMeasurement, CP(s), fw, fw-1);
			swidth = StringWidth(CP(s));
/*
			if (theXMin * theXMax <= 0)
				HashCoord = RealOriginV -
						RoundToInt(HashMeasurement*scale) * ods;
			else {
				if (theXMax > 0)
					HashCoord = RealOriginV -
							RoundToInt((HashMeasurement-theXMin)*scale) * ods;
				else
					HashCoord = RealOriginV -
							RoundToInt((HashMeasurement-theXMax)*scale) * ods;
			}
*/
			if (theXMin * theXMax <= 0)
				HashCoord = RealOriginV-RoundToInt(HashMeasurement*RealScale);
			else {
				if (theXMax > 0)
					HashCoord = RealOriginV -
							RoundToInt((HashMeasurement-theXMin)*RealScale);
				else
					HashCoord = RealOriginV -
							RoundToInt((HashMeasurement-theXMax)*RealScale);
			}

				/*	Need to do ClipRect(pgr) before hashes,
				 *	SetClip(tmpClipRgn) after?	*/
			MoveTo(Origin.h, HashCoord);
			Line(theHashSize, 0);

			if (GraphClipRgn != NIL) {
				SetRect(&tmpRect, Origin.h, HashCoord,
						Origin.h + theHashSize+1, HashCoord + 1);
				tmpRgn = NewRgn();
				RectRgn(tmpRgn, &tmpRect);
				DiffRgn(GraphClipRgn, tmpRgn, GraphClipRgn);
				DisposeRgn(tmpRgn);
			}

			if ( (HashCoord + tSize/2 < *pMinWC) &&
					(HashCoord - tSize/2 - tSize%2 > *pMaxWC) ) {
				MoveTo(otherWriteCoord, HashCoord);
				Move(0, tSize/2);
				DrawJustString(theJust, CP(s));
				*pChangeWC = HashCoord - sign*(tSize/2 + tSize%2 + text_spacing);
			}
/* 1.01 change (below)--oHashMeasurement
			HashMeasurement += HashDist;
 */
			oHashMeasurement = HashMeasurement;
			HashMeasurement += HashDist;
			if (oHashMeasurement == HashMeasurement)
					/* HashDist less than one significant bit of HashMeasurement	*/
				break;
		}
	}

}


	/* ----------------------------------------------------------------- */
	/* -------		  Routines Relating to Tracking in			-------- */
	/* ------- Phase Space windows, with selection rectangle	-------- */
	/* ----------------------------------------------------------------- */

Boolean LoadDashedPatterns()
{
	Handle	PH;
	int		i;

	if ( (PH = GetNamedResource('PAT#', DashedLinePatsName)) == NIL ) {
		for(i = 0; i < NumDashedLinePatterns; ++i)
			BlockEquate(CP(ThePatterns[i]), CP(gray), sizeof(Pattern));
		return(UseDashedLine = FALSE);
	}
	HLock(PH);
	BlockEquate(CP(ThePatterns[0]), *PH+2, NumDashedLinePatterns*sizeof(Pattern));
	HUnlock(PH);
	ReleaseResource(PH);
	return(UseDashedLine = TRUE);
}

Boolean TrackDashedRect(int h, int v, Rect *p_DragRect, Rect *p_Outline,
		Boolean ForMagnification)
{
	long		l, PosLong, DashedLineWaitTicks = 7L;
	Point		oldpt, pt;
	PenState	PS;
	Boolean		MovedSome;
	Rect		tmpDR;

		/*	kludge because PinRect doesn't seem to let point equal
		 *	bottom or right of *p_DragRect
		 *		-TA 1/22/91
		 */
	tmpDR = *p_DragRect;
	++tmpDR.right;
	++tmpDR.bottom;

	GetPenState(&PS);
	PenMode(patXor);
	SetRect(p_Outline, h, v, h, v);
	oldpt.h = -1; oldpt.v = -1;
	WaitTicks(DashedLineWaitTicks);
	MovedSome = FALSE;
	for(l=-1L;StillDown();) {
		++l;
		GetMouse(&pt);
		if ( (pt.h==oldpt.h) && (pt.v==oldpt.v) )
			PenPat(&ThePatterns[l%4+8]);
		else {
			if (l>0) {
				PenPat(&ThePatterns[(l-1)%8]);
				FrameRect(p_Outline);
			}
			PenPat(&ThePatterns[l%8]);
			PosLong = PinRect(&tmpDR, pt);
			pt.h = LoWord(PosLong);	pt.v = HiWord(PosLong);
			if (pt.h>=h)	{	p_Outline->left = h;	p_Outline->right = pt.h;	}
			else			{	p_Outline->left = pt.h;	p_Outline->right = h;		}
			if (pt.v>=v)	{	p_Outline->top = v;		p_Outline->bottom = pt.v;	}
			else			{	p_Outline->top = pt.v;	p_Outline->bottom = v;		}

			if (!MovedSome)
				if ( (pt.h-h)*(pt.h-h) + (pt.v-v)*(pt.v-v) > 2 ) {
					MovedSome = TRUE;
					if (ForMagnification)
						SetCursor(*MagnifyingGlass);
				}
		}
		FrameRect(p_Outline);
		oldpt.h = pt.h; oldpt.v = pt.v;
		WaitTicks(DashedLineWaitTicks);
	}
	if (l>=0) {
		PenPat(&ThePatterns[l%8]);
		FrameRect(p_Outline);
	}
	PenMode(PS.pnMode);
	PenPat(&PS.pnPat);
	if (MovedSome && ForMagnification)
		SetCursor(&arrow);
	if (l>=0)	return(MovedSome);
	else		return(FALSE);
}
