/*
 *	qmGraphics.c
 *	Graphics routines for Quadratic Map
 */

#include "qmPrototypes.h"
#include "qmGraphics.h"

				/*	From qm.c, for NewGraphics(TRUE, TRUE)	*/
extern	int		ProgramStatus;
				/*	From qm.c, for "user waiting" (watch cursor, ticking?)	*/
extern	Boolean	UserWaiting;
extern	Boolean	InForeground;
					/*	For TxOut updating	*/
extern	IterationVars		*pCurrentVars;
				/*	From qmMenus.c, for NewGraphics(TRUE, )	*/
extern	Boolean	NeedToSave;
				/*	From qmImpactGraphics, for FindQMClick()	*/
extern	Rect	NLRect;

	Windows		w;

	CursHandle	watch, MagnifyingGlass;

	Boolean		*pTxOutWindowCleared;


StartWaiting()
{
	UserWaiting = TRUE;
	HLock((Handle) watch);
	if (InForeground)
		SetCursor(*watch);
}

StopWaiting()
{
	UserWaiting = FALSE;
	HUnlock((Handle) watch);
	if (InForeground)
		SetCursor(&arrow);
}

int	FindQMClick(int i, Point pt)
{
	Rect	tmpRect;	/* for IMap kludge	*/

	switch (i) {
		case TxOut:
			if (PtInBodyViewRect(i, pt))
				return(i);
		case PSpace:
			if (PtInRect(pt, &NLRect))
				return(i);
		case EmbPS:
		case Orbit:
			tmpRect = w[i].GR;
/*
			++tmpRect.right;
			++tmpRect.bottom;
*/
			if (PtInRect(pt, &tmpRect))
				return(i);
		case BifD:
			if (FindSpecialClick(i, pt))
				return(i);
	}
	return(NoWindow);
}


	/* ---------- Update routine ---------- */

HandleUpdate(WindowPtr wp, int myHitNumber)
{
	int		theRef;
	Rect	tmpRect;

	theRef = WindowIndex(wp);
	if (theRef == NoWindow) {
		SetPort(wp);
			/*	Next 2 lines are a kludge (for dialog boxes put in front of reg.
			 *	windows, when calling HandleAllUpdates()--
			 *	otherwise, would be in an infinite loop).
			 */
		BeginUpdate(wp);
		EndUpdate(wp);
		return;
	}

	SetPort(wp);
	BeginUpdate(wp);
		/* Clear window, draw grow icon, etc. */
	NoClippingRect(wp);
	EraseRect(&wp->portRect); DrawGrowIcon(wp);
	DrawControls(wp);
	StdDocWindowClipRect(wp);

	switch (theRef) {
		case Orbit:				RefreshOrbit();				break;
		case TextOutput:		UpdateText(theRef, lambda);	break;
		case PhaseSpace:		RefreshPS(myHitNumber);		break;
		case EmbeddedPS:		RefreshEPS(myHitNumber);	break;
		case BifDiagram:		RefreshBD();				break;
	}
	EndUpdate(wp);
}

MakePicture(int windex, GrafPtr gp, Rect *pDstRect)
{
	GrafPtr		tmpgp;
	RgnHandle	MaskRgn = NewRgn();
	Point		pt = {0, 0};
	Rect		tmpRect;

	CopyRgn(GrayRgn, MaskRgn);
	GetPort(&tmpgp);
	SetPort((GrafPtr) w[windex].Peek);
	GlobalToLocal(&pt);
	OffsetRgn(MaskRgn, pt.h, pt.v);
	SetPort(tmpgp);

	if (!SectRect(pDstRect, &(**MaskRgn).rgnBBox, &tmpRect)) ;

/*
	ClipRect(pDstRect);
	CopyBits(&w[windex].Peek->port.portBits, &gp->portBits,
			pDstRect, pDstRect, srcCopy, 0L);
*/
	ClipRect(&tmpRect);
	CopyBits(&w[windex].Peek->port.portBits, &gp->portBits,
			&tmpRect, &tmpRect, srcCopy, MaskRgn);

	DisposeRgn(MaskRgn);
}

HandlePICTCopy(int windex)
{
	PicHandle	pic;
	long		theError;
	Rect		DstRect;
	RgnHandle	OldClipRgn;

	SetPort(w[windex].Peek);
	OldClipRgn = NewRgn();
	GetClip(OldClipRgn);

	GetStdDocWindowClipRect(&DstRect, (WindowPtr) w[windex].Peek);
	pic = OpenPicture(&DstRect);
	 MakePicture(windex, (WindowPtr) w[windex].Peek, &DstRect);
	ClosePicture();
	SetClip(OldClipRgn);
	DisposeRgn(OldClipRgn);

	if ( (theError = ZeroScrap()) != noErr)
		return;
	HLock(pic);
	theError = PutScrap((long) (**pic).picSize, 'PICT', (Ptr) *pic);
	HUnlock(pic);
	KillPicture(pic);
}

HandleCopy(int windex)
{
	switch (windex) {
		case NoWindow:								return;
		case TextOutput:	HandleTEXTCopy(windex);	break;
		default:			HandlePICTCopy(windex);	break;
	}
}


	/* ---------- General graphics routines ---------- */

NewGraphics(Boolean ClearSolutions, Boolean NeedToResetFileName)
{
	if (ClearSolutions) {
		SetOrbitOffset(0);
		NewQMWindows(ClearSolutions);
	}
	if (ClearSolutions) {
		NewTextEdit();
		w[TxOut].WindowCleared = TRUE;
		if (NeedToResetFileName) {
			if (ProgramStatus < /*was <=--don't know why*/ BIFURCATION_DIAGRAM)
				SetTextTitle(FALSE, TxOut);
			else
				SetANilFileName(TxOut);
		}
		NeedToSave = FALSE;
	}
/*
	NewContGraphics();
*/
	NewImpactGraphics(ClearSolutions);
}

Boolean InitGraphics()
{
	pTxOutWindowCleared = &w[TxOut].WindowCleared;
	InitQMWindows();
	if (!InitQMOffPorts())
		return(FALSE);

/*
	InitContGraphics();
*/

	InitImpactGraphics();

	if (LoadDashedPatterns()) ;
	MagnifyingGlass = GetCursor(MagnifyingGlassCursorID);

	return(TRUE);
}

HaltGraphics()
{
	HaltImpactGraphics();
	HaltQMWindows();
}