/*
 *	qmImpactGraphics.c
 *
 *	Graphics routines for Quadratic Maps impact windows,
 *	which *all* of Quadratic Map's normal windows are.
 *	(Bouncing Ball has both impact and continuous windows.)
 *
 *		-TA 4/22/91
 */

#include <limits.h>
#include "qmPrototypes.h"
#include "qmGraphics.h"
#include "qmMenus.h"
#include <MenuUtils.h>

#define	ARROW_SIZE		3
#define	MUCH_EMBEDDING	7
#define	FIXED_ACCURACY	(1.0/(double) INT_MAX)
#define	CLICK_SLOP		10L

				/*	Global external vars from qm.c	*/
extern	pIterationVars		pCurrentVars;
extern	hQMPreferences		hQMPrefs;
								/*	From qm.c		*/
extern	Boolean				PrintingOnly;
								/*	From qmGraphics.c	*/
extern	Windows				w;
extern	CursHandle			watch;
								/*	From qmWindows.c; screen size	*/
extern	int					MaxScreenY;
extern	int					RightWindowLimit, BottomWindowLimit;
								/*	From qmMenus.c; for magnification	*/
extern	MenuHandle			ViewMenuH;
								/*	From qmBifDiag.c, for AddGraphicSolution()	*/
extern	Boolean				*pBDGraphInAllWindows;


	Rect	PSLatestImpactRect, EPSLatestImpactRect;
	Rect	PSBox, EPSBox;
	int		StartHitNumber;	/* first hit to show in Phase Spaces ... */
	int		StartILHitNumber;

	double	AutoXMin, AutoXMax;
	int		MaxFixedEmbedding;
	Rect	NLRect, NLImpactRect;
	long	CurrentEPSOffsetCount;
	long	CurrentPSOffsetCount;

	DoubleRect	FullPSRange, NewPSRange;
	DoubleRect	FullEPSRange, NewEPSRange;
	BooleanRect	NewUsesFullPS = {TRUE, TRUE, TRUE, TRUE},
				UsesFullPS = {TRUE, TRUE, TRUE, TRUE},
				NewUsesFullEPS = {TRUE, TRUE, TRUE, TRUE},
				UsesFullEPS = {TRUE, TRUE, TRUE, TRUE};

	int			PSXFieldWidth, PSYFieldWidth;
	int			EPSXFieldWidth, EPSYFieldWidth;
	int			OYFieldWidth;

	int			OrbitOffset, Orbit_nOffset, BackOrbit_nOffset;
	int			Max_n_Orbit, BackMax_n_Orbit;
	int			NStart;
	int			XStop;
	Rect		OrbitClipRect;
	Boolean		RestartOrbitGraph;
	Boolean		RefreshingWhileGraphing;
	RgnHandle	OrbitGraphClipRgn;
	int			OrbitHAxisPos;


			/*	-------------------------------------------------	*/
			/*	---------- Window-Independent Routines ----------	*/
			/*	-------------------------------------------------	*/

SetPhaseSpacesStart(int theHitNumber)
{
	StartHitNumber = theHitNumber;
	w[PSpace].DrawFromScratch = TRUE;
	StdForceUpdate((WindowPtr) w[PSpace].Peek);
	w[EmbPS].DrawFromScratch = TRUE;
	StdForceUpdate((WindowPtr) w[EmbPS].Peek);
}

SetIterLinesStart(int theHitNumber, Boolean NeedRefresh)
{
	StartILHitNumber = theHitNumber;
	HandleIterLineMenus();
	if (NeedRefresh) {
		w[PSpace].DrawFromScratch = TRUE;
		StdForceUpdate((WindowPtr) w[PSpace].Peek);
	}
}


		/*	------------------------------------------------	*/
		/*	---------- Orbit Window Routines ---------------	*/
		/*	------------------------------------------------	*/

GetOrbitInfo()
{
	int	oOrbitDotSize = OrbitDotSize, oOrbitSpacing = OrbitSpacing;

	if (DoOrbitDialog(&OrbitDotSize, &OrbitSpacing)) {
		if ((oOrbitDotSize!=OrbitDotSize) || (oOrbitSpacing!=OrbitSpacing))	{	
			SetOrbitCoords();
			w[Orbit].DrawFromScratch = TRUE;
			StdForceUpdate((GrafPtr) w[Orbit].Peek);
		}
	}
}

SwitchScrollOrbit()
{
	ScrollOrbit = !ScrollOrbit;
	CheckItem(ViewMenuH, ScrollOrbitItem, ScrollOrbit);
	if (!ScrollOrbit) {
		OrbitOffset += Orbit_nOffset;
		BackOrbit_nOffset = Orbit_nOffset = 0;
	}
	return(ScrollOrbit);
}

ForceRefreshOrbit()
{
RgnHandle	tmpRgn;
SetPort(w[Orbit].Peek);
tmpRgn = NewRgn();
GetClip(tmpRgn);
SetClip(OrbitGraphClipRgn);
EraseRect(&w[Orbit].Peek->port.portRect);
SetClip(tmpRgn);
return;
/*
	StdForceUpdate((WindowPtr) w[Orbit].Peek);
	HandleUpdate((WindowPtr) w[Orbit].Peek, 0);
*/
	SetPort(w[Orbit].Peek);
	EraseRect(&w[Orbit].Peek->port.portRect);
	RefreshOrbit();
}

SetOrbitOffset(int theHitNumber)
{
	OrbitOffset = theHitNumber;
			/*	what's left was ( phase - delta + ((phase>=delta) ? 0.0 : 1.0) ) * Period;	*/
			/*	or fmod(theHitTime, Period)	*/
	if (ScrollOrbit)
		BackOrbit_nOffset = Orbit_nOffset = theHitNumber;

	if (theHitNumber == 0)
		if (RealIterationTime == 0.0)
			StartDetachedTickDelay();
}

SetWholeScreenOrbitCoords()
{
	double	range;
	int		i;

	range = w[Orbit].GraphRange[ymax_i] - w[Orbit].GraphRange[ymin_i];
	for(i=-1;i<2;++i) {
		if (range > 1.0) break;
		range *= 10;
	}
	OYFieldWidth = StdPSFieldWidth + MAX(i,0);

	w[Orbit].BackLimits.left = 2*StdMargin + OYFieldWidth*CharWidth('0');
	/*	For multiple screens, use MaxCurScreenX, Y, where these are the
	 *	corresponding maxes for the main device?		-TA 10/29/90
	 */
		/*	w[Orbit].BackLimits.right (used to be ScreenMaxXCoord)
		 *	will always allow each period the same # of pixels
		 */
	w[Orbit].BackLimits.right = RightWindowLimit -
			Orbit_RightMargin - w[Orbit].BackLimits.left;
	BackMax_n_Orbit = w[Orbit].BackLimits.right / (OrbitDotSize+OrbitSpacing);
	if (BackMax_n_Orbit*(OrbitDotSize+OrbitSpacing) + OrbitDotSize <=
			w[Orbit].BackLimits.right)
		++BackMax_n_Orbit;
/*
	--BackMax_n_Orbit;
*/
	w[Orbit].BackLimits.right = w[Orbit].BackLimits.left +
			(BackMax_n_Orbit-1)*(OrbitDotSize+OrbitSpacing) + OrbitDotSize;

	w[Orbit].BackLimits.bottom = BottomWindowLimit-Orbit_BottomMargin;

	w[Orbit].BackOrigin.h = w[Orbit].BackLimits.left;
}

SetOrbitCoords()
{
	int		MaxWMYC, WindowMaxXCoord, WindowMaxYCoord;
	Rect	HotSpotRect;

	w[Orbit].GraphRange[ymin_i] = XMin;
	w[Orbit].GraphRange[ymax_i] = XMax;

	SetWholeScreenOrbitCoords();

	SetPort(w[Orbit].Peek);
	w[Orbit].DrawFromScratch = TRUE;
	StdForceUpdate((GrafPtr) w[Orbit].Peek);
	WindowMaxXCoord = w[Orbit].Peek->port.portRect.right -
		Orbit_RightMargin - w[Orbit].BackLimits.left;
	Max_n_Orbit = WindowMaxXCoord / (OrbitDotSize+OrbitSpacing);
	if (Max_n_Orbit*(OrbitDotSize+OrbitSpacing) + OrbitDotSize <=
			WindowMaxXCoord)
		++Max_n_Orbit;
/*
	--Max_n_Orbit;
*/

	w[Orbit].GraphRange[xmin_i] = 0;
	w[Orbit].GraphRange[xmax_i] = Max_n_Orbit-1;

	WindowMaxXCoord =
			(Max_n_Orbit-1)*(OrbitDotSize+OrbitSpacing) + OrbitDotSize;

	MaxWMYC = WindowMaxYCoord = w[Orbit].Peek->port.portRect.bottom -
		Orbit_BottomMargin - Orbit_TopMargin;
	WindowMaxYCoord = (WindowMaxYCoord/OrbitDotSize) * OrbitDotSize;

	SetRect(&w[Orbit].GR, w[Orbit].BackLimits.left, 0,
			w[Orbit].BackLimits.left + WindowMaxXCoord, 0);
	w[Orbit].BackLimits.top = w[Orbit].GR.top =
			Orbit_TopMargin + (MaxWMYC - WindowMaxYCoord)/2;
	w[Orbit].BackLimits.bottom = w[Orbit].GR.bottom =
			w[Orbit].GR.top + WindowMaxYCoord;
	OrbitClipRect = w[Orbit].GR;

	w[Orbit].xScale = 1;
	w[Orbit].yScale = (w[Orbit].GR.bottom - w[Orbit].GR.top-1) / OrbitDotSize /
			(w[Orbit].GraphRange[ymax_i] - w[Orbit].GraphRange[ymin_i]);

	w[Orbit].xGraphAroundZero = TRUE;
	w[Orbit].yGraphAroundZero =
			(w[Orbit].GraphRange[ymax_i]*w[Orbit].GraphRange[ymin_i]<=0.0);

	w[Orbit].Origin.h = w[Orbit].GR.left;
/*
	if (w[Orbit].GraphRange[ymin_i]*w[Orbit].GraphRange[ymax_i] >= 0.0) {
		if (w[Orbit].GraphRange[ymax_i] > 0)
			w[Orbit].Origin.v = w[Orbit].GR.bottom - OrbitDotSize;
		else
			w[Orbit].Origin.v = w[Orbit].GR.top;
	}
	else
*/

	if (w[Orbit].yGraphAroundZero)
		w[Orbit].Origin.v = w[Orbit].GR.top +
				RoundToInt(w[Orbit].GraphRange[ymax_i]*w[Orbit].yScale) *
				OrbitDotSize;
	else
		if (w[Orbit].GraphRange[ymax_i] > 0)
			w[Orbit].Origin.v = w[Orbit].GR.bottom-OrbitDotSize;
		else
			w[Orbit].Origin.v = w[Orbit].GR.top;

/*	added 9/26/91	*/
	OrbitHAxisPos = w[Orbit].Origin.v;
	if (w[Orbit].GraphRange[ymin_i] * w[Orbit].GraphRange[ymax_i] < 0)
		OrbitHAxisPos += OrbitDotSize/2;

	w[Orbit].BackOrigin.v = w[Orbit].Origin.v;

/*	Needed?
	SetRect(&w[Orbit].PR,
			w[Orbit].GR.left - w[Orbit].Origin.h,
			w[Orbit].Origin.v - w[Orbit].GR.bottom,
			w[Orbit].GR.right - w[Orbit].Origin.h + 1,
			w[Orbit].Origin.v - w[Orbit].GR.top + 1);
*/
			/* +1 to account for pixel hanging below and to left of point */

	HotSpotRect = w[Orbit].Peek->port.portRect;
	HotSpotRect.top = w[Orbit].GR.bottom /* changed 9/26/91 + 1 */;
	HotSpotRect.right	-= ScrollBarWidth-1;
	HotSpotRect.bottom	-= ScrollBarWidth-1;
	ResizeControl(w[Orbit].HS[0], &HotSpotRect);

	HotSpotRect.bottom = HotSpotRect.top;
	HotSpotRect.top = w[Orbit].GR.top;
	HotSpotRect.right = w[Orbit].Origin.h - 1;
	ResizeControl(w[Orbit].HS[1], &HotSpotRect);

/*
	++w[Orbit].GR.right;
*/
}

DrawOrbitBackground()
{
	int			XWriteCoord, YWriteCoord;
	int			i, yhash;
	double		HashDist;
	int			HashNum;
	Rect		tmpRect;
	RgnHandle	tmpRgn;
	int			ClipHAxisPos;

	RectRgn(OrbitGraphClipRgn, &w[Orbit].GR);

	SetPort(w[Orbit].Peek);
		/* Draw Axes */
	if (w[Orbit].GraphRange[ymin_i]*w[Orbit].GraphRange[ymax_i] >= 0.0) {
/* change 9/26/91
		if (w[Orbit].GraphRange[ymax_i] > 0.0)
			MoveTo(w[Orbit].GR.left-1, w[Orbit].GR.bottom);
		else
			MoveTo(w[Orbit].GR.left-1, w[Orbit].GR.top-1);
*/
/*
		if (w[Orbit].GraphRange[ymax_i] > 0.0)
			MoveTo(w[Orbit].GR.left-1, w[Orbit].GR.bottom-1);
		else
			MoveTo(w[Orbit].GR.left-1, w[Orbit].GR.top);
*/
		if (w[Orbit].GraphRange[ymax_i] > 0.0)
			ClipHAxisPos = w[Orbit].GR.bottom-1;
		else
			ClipHAxisPos = w[Orbit].GR.top;
		MoveTo(w[Orbit].GR.left-1, ClipHAxisPos);

/*	Next lines added 11/8/91...	*/
/*
		SetRect(&tmpRect, w[Orbit].GR.left, OrbitHAxisPos,
				w[Orbit].GR.right, OrbitHAxisPos + 1);
*/
		SetRect(&tmpRect, w[Orbit].GR.left, ClipHAxisPos,
				w[Orbit].GR.right, ClipHAxisPos + 1);
		tmpRgn = NewRgn();
		RectRgn(tmpRgn, &tmpRect);
		DiffRgn(OrbitGraphClipRgn, tmpRgn, OrbitGraphClipRgn);
		DisposeRgn(tmpRgn);
/*	End of 11/8/91 addition	*/
	}
	else {
/* change 9/26/91
		MoveTo(w[Orbit].GR.left, w[Orbit].Origin.v);
*/
		MoveTo(w[Orbit].GR.left, OrbitHAxisPos);

/* change 9/26/91
		SetRect(&tmpRect, w[Orbit].GR.left, w[Orbit].Origin.v,
				w[Orbit].GR.right, w[Orbit].Origin.v + 1);
*/
		SetRect(&tmpRect, w[Orbit].GR.left, OrbitHAxisPos,
				w[Orbit].GR.right, OrbitHAxisPos + 1);
		tmpRgn = NewRgn();
		RectRgn(tmpRgn, &tmpRect);
		DiffRgn(OrbitGraphClipRgn, tmpRgn, OrbitGraphClipRgn);
		DisposeRgn(tmpRgn);
	}
	Line(w[Orbit].GR.right - w[Orbit].GR.left, 0);

/*
	Move(HorAxisLabelHOffset, 0);
	if (RealFont(monaco, 12)) {
		TextSize(12);
		Move(0, TextHeight/2-TextHeight%2 + 1);
		DrawChar('n');
		TextSize(9);
	}
	else {
		Move(0, TextHeight/2-TextHeight%2);
		DrawChar('n');
	}
*/
	Move(HorAxisLabelHOffset, TextHeight/2-TextHeight%2);
	DrawChar('n');

	MoveTo(w[Orbit].GR.left-1, w[Orbit].GR.top);
	Line(0, w[Orbit].GR.bottom - w[Orbit].GR.top-1);

/*
	WindowMoveTo(Orbit,w[Orbit].GR.right - w[Orbit].GR.left, 0);
*/
/*
	WindowLineTo(Orbit,0,0);
	WindowLineTo(Orbit,0,w[Orbit].GR.bottom - w[Orbit].GR.top);
*/
/*
	WindowLineTo(Orbit,-1,0);
	WindowLineTo(Orbit,-1,w[Orbit].GR.bottom - w[Orbit].GR.top);
*/

		/*	Label Axes	*/

/*
	WindowMoveTo(Orbit,w[Orbit].GR.right - w[Orbit].Origin.h, 0);
	Move(HorAxisLabelHOffset, TextHeight/2-TextHeight%2);
	DrawChar('n');
*/
	WindowMoveTo(Orbit,0,w[Orbit].Origin.v - w[Orbit].GR.top);

	Move(0, -VerAxisLabelVOffset);
	if (RealFont(monaco, 12)) {
		TextSize(12);
		Move(-CharWidth('x')/2-CharWidth('x')%2, 0);
		DrawChar('x');
		TextSize(9);
	}
	else {
		Move(-CharWidth('X')/2-CharWidth('X')%2, 0);
		DrawChar('X');
	}

/*
	Move(-CharWidth('x')/2, -VerAxisLabelVOffset);
	DrawChar('x');
*/

		/*	Draw position axis (vertical) hash marks	*/

	HashDist = 0.0;
	HashNum = 0;
	HashNum = (w[Orbit].GR.bottom - w[Orbit].GR.top) /
			(TextHeight*2) / ((OrbitDotSize+1)/2) + 1;
	if ( (w[Orbit].GraphRange[ymin_i] == AutoXMin) &&
			(w[Orbit].GraphRange[ymax_i] == AutoXMax) ) {
		if ( (AutoXMax - AutoXMin)/0.1 > HashNum )
			HashDist = 0.25;
		else
			HashDist = 0.1;
	}
	else
		HashNum = MIN(HashNum, 10);

	LabelOrbitVAxis(w[Orbit].GraphRange[ymin_i], w[Orbit].GraphRange[ymax_i],
			&w[Orbit].GR, w[Orbit].Origin, OrbitHAxisPos, w[Orbit].yScale, HashDist,
			HashNum, OYFieldWidth, 2, 3, 3, OrbitDotSize, FALSE, OrbitGraphClipRgn);
/*
	for (i=2; i<=ScreenNumAmplitudes; i+=2) {
		yhash = yFloatToCoord(Orbit, i*A);
		if (yhash > w[Orbit].Origin.v - w[Orbit].GR.top)
			break;
		WindowMoveTo(Orbit, 0, yhash);
		Line(HashSize,0);
	}
	NumOrbitAmplitudeHashes = i/2 - 1;
*/

		/* Write out initial conditions */
	XWriteCoord = MAX((w[Orbit].GR.right + w[Orbit].GR.left - 66), 0)/2;
	YWriteCoord = w[Orbit].GR.bottom+(Orbit_BottomMargin-ScrollBarWidth)/2 +
						TextHeight/2 + TextHeight%2;
	MoveTo(XWriteCoord, YWriteCoord);

	WriteVar("\PX0: ", x0);
}

RefreshOrbit()
{
	int	i;

	DrawOrbitBackground();
/*
	ClipRect(&w[Orbit].Peek->port.portRect);
*/
/*
	if (RestartOrbitGraph) {
*/
	if (!w[Orbit].DrawFromScratch) {
		myCopyBits(w[Orbit].Back, (GrafPtr) w[Orbit].Peek, &w[Orbit].GR,
			(Orbit_nOffset-BackOrbit_nOffset) * (OrbitDotSize+OrbitSpacing),
			w[Orbit].BackOrigin.v - w[Orbit].Origin.v,
			0, 0, srcOr, NIL );
		return;
	}

	w[Orbit].WindowCleared = TRUE;

	w[Orbit].DrawFromScratch = FALSE;
	SetPort(w[Orbit].Back);
	EraseRect(&w[Orbit].Back->portRect);

RefreshingWhileGraphing = FALSE;
if (RefreshingWhileGraphing) {
	RefreshingWhileGraphing = FALSE;
	return;
}
	for(i=OrbitOffset+Orbit_nOffset; i<HitNumber; ++i)
		PlotInOrbit(i-OrbitOffset, yFloatToCoord(Orbit, GetSolution(i)));
	if ( !RefreshingWhileGraphing && (HitNumber>=0) )
		PlotInOrbit(i-OrbitOffset, yFloatToCoord(Orbit, GetSolution(i)));
	RefreshingWhileGraphing = FALSE;
}

	/*	Not Currently Used -TA 5/6/91	*/
InitOrbitGraphing(int yint)
{
	int			ShiftAmount, ScrollAmount;
	RgnHandle	updateRgn;

		/* if window has been shrunk beyond cursor since last graph... */
	if (NStart-Orbit_nOffset > Max_n_Orbit) {
		if (ScrollOrbit) {
			ScrollAmount = NStart - Orbit_nOffset - Max_n_Orbit;
			Orbit_nOffset += ScrollAmount;

			SetPort(w[Orbit].Peek);
			updateRgn = NewRgn();
			ScrollRect(&w[Orbit].GR, -ScrollAmount, 0, updateRgn);
			myCopyBits(w[Orbit].Back, (GrafPtr) w[Orbit].Peek,
					&(**updateRgn).rgnBBox,
					Orbit_nOffset-BackOrbit_nOffset,
					w[Orbit].BackOrigin.v-w[Orbit].Origin.v,
					0,0, srcCopy, NIL);
			DisposeRgn(updateRgn);
		}
		else {
			ShiftAmount = Max_n_Orbit - (NStart-Orbit_nOffset);
	
			SetPort(w[Orbit].Back);
			updateRgn = NewRgn();
			ScrollRect(&w[Orbit].Back->portRect, -(int) ShiftAmount,
					0, updateRgn);
			DisposeRgn(updateRgn);
			ForceRefreshOrbit();
			/*	Had to put back in because SetContGraphingClipping before,
			 *	was after	*/
		 	SetPort(w[Orbit].Peek);
			ClipRect(&OrbitClipRect);
	
			NStart -= ShiftAmount;
			XStop -= ShiftAmount;
			OrbitOffset += ShiftAmount;
		}
	}

	SetPort(w[Orbit].Peek);
	WindowMoveTo(Orbit, NStart-Orbit_nOffset, yint);
	SetPort(w[Orbit].Back);
	BackMoveTo(Orbit, NStart-BackOrbit_nOffset, yint);
}

PlotInOrbit(int xint, int yint)
{
	RgnHandle	updateRgn, tmpRgn;
	Rect		tmpRect;
	int			Orbit_xint, Back_xint;
	int			i, yhash;
	Point		pt;
	int			ScrollAmt = OrbitDotSize + OrbitSpacing;
	RgnHandle	tmpRgn2;

	w[Orbit].WindowCleared = FALSE;

			/* if beyond edge of graph */
		/*	BackOrbit_nOffset always 0, except temporarily here--delete?	*/
	if ( ScrollOrbit &&
			(xint-BackOrbit_nOffset > BackMax_n_Orbit+Max_n_Orbit) ) {
		BackOrbit_nOffset += BackMax_n_Orbit;

		OrbitOffset += BackMax_n_Orbit;
		xint -= BackOrbit_nOffset;
		Orbit_nOffset -= BackOrbit_nOffset;

		SetPort(w[Orbit].Back);
		updateRgn = NewRgn();
		ScrollRect(&w[Orbit].Back->portRect, -BackOrbit_nOffset*ScrollAmt,
				0, updateRgn);
		DisposeRgn(updateRgn);
		Move(-BackOrbit_nOffset*ScrollAmt, 0);
		BackOrbit_nOffset = 0;
/*
		BackOrbit_nOffset = ++BackOrbit_nOffset % PeriodWidth;
		if (BackOrbit_nOffset == 0) {
			OrbitOffset += Period;
			NStart -= PeriodWidth;
			XStop -= PeriodWidth;
			xint -= PeriodWidth;
			Orbit_nOffset -= PeriodWidth;
		}
		SetPort(w[Orbit].Back);
		updateRgn = NewRgn();
		ScrollRect(&w[Orbit].BackLimits, -1, 0, updateRgn);
		DisposeRgn(updateRgn);
		Move(-1, 0);
*/
	}
	if (xint-Orbit_nOffset >= Max_n_Orbit) {
		if (ScrollOrbit) {
			++Orbit_nOffset;
			SetPort(w[Orbit].Peek);
			updateRgn = NewRgn();
			tmpRgn2 = NewRgn();
			GetClip(tmpRgn2);
			SetClip(OrbitGraphClipRgn);
			ScrollRect(&w[Orbit].GR, -ScrollAmt, 0, updateRgn);
	
			if ((**updateRgn).rgnBBox.left < w[Orbit].GR.right-ScrollAmt) {
				tmpRect = w[Orbit].GR;
				tmpRect.left = tmpRect.right-ScrollAmt;
				tmpRgn = NewRgn();
				RectRgn(tmpRgn, &tmpRect);
				DiffRgn(updateRgn, tmpRgn, updateRgn);
				DisposeRgn(tmpRgn);
						/* should not be 0,0,0,0	*/
				myCopyBits(w[Orbit].Back, (GrafPtr) w[Orbit].Peek,
						&(**updateRgn).rgnBBox,
						(Orbit_nOffset-BackOrbit_nOffset)*ScrollAmt,
						w[Orbit].BackOrigin.v-w[Orbit].Origin.v,
						0,0, srcCopy, NIL);
			}
			SetClip(tmpRgn2);
			DisposeRgn(tmpRgn2);
			DisposeRgn(updateRgn);
			Move(-ScrollAmt, 0);
		}
		else {
			w[Orbit].DrawFromScratch = TRUE;
			OrbitOffset += Max_n_Orbit;
/*
			RefreshingWhileGraphing = TRUE;
*/
				/*	Not sure this next step works right . . .
				 *	(now skips updating mechanism, *unlike* BB,
				 *	because could cause recursive updating,
				 *	which mal-affects EndUpdate())
				 */
			ForceRefreshOrbit();
		 	SetPort(w[Orbit].Peek);
			ClipRect(&OrbitClipRect);

			xint -= Max_n_Orbit;
			BothMoveTo(Orbit, xint-1, yint);
			NStart -= Max_n_Orbit;
			XStop -= Max_n_Orbit;
/*
			OrbitOffset += (WindowNumPeriods*Period);
*/
		}
	}

	Orbit_xint = xint - Orbit_nOffset;
	Back_xint = xint - BackOrbit_nOffset;

	SetPort(w[Orbit].Peek);
/*
	WindowMoveTo(Orbit, Orbit_xint, yint);
	Line(0, 0);
*/
	MoveTo(w[Orbit].Origin.h + Orbit_xint*(OrbitDotSize+OrbitSpacing),
			w[Orbit].Origin.v - yint*OrbitDotSize);
	PenSize(OrbitDotSize, OrbitDotSize);
	Line(0, 0);
	PenSize(1, 1);

	SetPort(w[Orbit].Back);
/*
	BackMoveTo(Orbit, Orbit_xint, yint);
	Line(0, 0);
*/
	MoveTo(w[Orbit].Origin.h + Back_xint*(OrbitDotSize+OrbitSpacing),
			w[Orbit].Origin.v - yint*OrbitDotSize);
	PenSize(OrbitDotSize, OrbitDotSize);
	Line(0, 0);
	PenSize(1, 1);
}


			/*	-----------------------------------------	*/
			/*	---------- Phase Space Routines ----------	*/
			/*	-----------------------------------------	*/

HandleIterLineMenus()
{
	if (PSParams.IterationLines) {
		ToggleItemEnabled(ViewMenuH, ClearIterLinesItem, TRUE);
		ToggleItemEnabled(ViewMenuH, ShowAllIterLinesItem,
				(StartILHitNumber > 1) );
	}
	else {
		ToggleItemEnabled(ViewMenuH, ClearIterLinesItem, FALSE);
		ToggleItemEnabled(ViewMenuH, ShowAllIterLinesItem, FALSE);
	}
}

GetPhaseSpaceInfo()
{
	int		oNLH = NLHeight;
	Boolean	oNLOn = PSParams.NumberLine, oSSA = PSParams.SameSizeAxes;
	Boolean	oILOn = PSParams.IterationLines;
	Boolean	ShowAllIterLines, oSAIL;

	oSAIL = ShowAllIterLines = (PSParams.IterationLines) &&
		(StartILHitNumber == 0);
	if (DoPhaseSpaceDialog(&NLHeight, &PSParams, &ShowAllIterLines)) {
		if ( (oNLOn != PSParams.NumberLine) ||
				(oNLOn && (oNLH != NLHeight)) ||
				(oSSA != PSParams.SameSizeAxes) )
			InitPSScales();
		if (PSParams.IterationLines) {
			if ( (!oILOn) || (oSAIL != ShowAllIterLines) )
				StartILHitNumber = (ShowAllIterLines) ? 0 : HitNumber+1;
		}
		HandleIterLineMenus();
		w[PSpace].DrawFromScratch = TRUE;
		StdForceUpdate((GrafPtr) w[PSpace].Peek);
	}
}

		/*	Routines related to phase space magnification	*/

InitFullPSRange()
{
	int	i;

	FullPSRange[ymin_i] = XMin;
	FullPSRange[xmin_i] = XMin;
	FullPSRange[ymax_i] = XMax;
	FullPSRange[xmax_i] = XMax;

	for (i=ymin_i; i<=xmin_i; ++i) {		/*	ymin_i, xmin_i	*/
		if (UsesFullPS[i]) {
			if (UsesFullPS[i+2] || (w[PSpace].GraphRange[i+2] > FullPSRange[i]))
				w[PSpace].GraphRange[i] = FullPSRange[i];
			else
				UsesFullPS[i] = FALSE;
		}
		if (NewUsesFullPS[i]) {
			if (NewUsesFullPS[i+2] || (NewPSRange[i+2] > FullPSRange[i]))
				NewPSRange[i] = FullPSRange[i];
			else
				NewUsesFullPS[i] = FALSE;
		}
	}
	for (i=ymax_i; i<=xmax_i; ++i) {		/*	ymax_i, xmax_i	*/
		if (UsesFullPS[i]) {
			if (UsesFullPS[i-2] || (w[PSpace].GraphRange[i-2] < FullPSRange[i]))
				w[PSpace].GraphRange[i] = FullPSRange[i];
			else
				UsesFullPS[i] = FALSE;
		}
		if (NewUsesFullPS[i]) {
			if (NewUsesFullPS[i-2]  || (NewPSRange[i-2] < FullPSRange[i]))
				NewPSRange[i] = FullPSRange[i];
			else
				NewUsesFullPS[i] = FALSE;
		}
	}
/*
	for (i=ymin_i; i<ItemsInRect; ++i) {
		if (UsesFullPS[i]) {
			w[PSpace].GraphRange[i] = FullPSRange[i];
		}
		if (NewUsesFullPS[i]) {
			NewPSRange[i] = FullPSRange[i];
		}
	}
*/
}

SetFullPSRange()
{
	int	i;

	InitFullPSRange();

	for (i=ymin_i; i<ItemsInRect; ++i) {
		NewUsesFullPS[i] = UsesFullPS[i] = TRUE;
		w[PSpace].GraphRange[i] = NewPSRange[i] = FullPSRange[i];
	}
	w[PSpace].Magnifying = FALSE;
	InitPSScales();
}

GetPSMagnification(Boolean SkipDialog)
{
	int			i;
	double		min, max;
	Boolean		GoodOp;

	InitFullPSRange();

		/*	Round to RelPlaces (2) relative decimal places	*/
	Round2Doubles(NewPSRange[xmin_i], NewPSRange[xmax_i], RelPlaces, &min, &max);
	if (NewPSRange[xmin_i] > FullPSRange[xmin_i])
		NewPSRange[xmin_i] = min;
	if (NewPSRange[xmax_i] < FullPSRange[xmax_i])
		NewPSRange[xmax_i] = max;
	Round2Doubles(NewPSRange[ymin_i], NewPSRange[ymax_i], RelPlaces, &min, &max);
	if (NewPSRange[ymin_i] > FullPSRange[ymin_i])
		NewPSRange[ymin_i] = min;
	if (NewPSRange[ymax_i] < FullPSRange[ymax_i])
		NewPSRange[ymax_i] = max;

/* 1.01 change
	if (!SkipDialog)
 */
	if ( (!SkipDialog) || (NewPSRange[xmin_i] == NewPSRange[xmax_i]) ||
			(NewPSRange[ymin_i] == NewPSRange[ymax_i]) )
		GoodOp = DoMagnifyPhaseSpacesDialog(&NewPSRange, &FullPSRange,
				&NewUsesFullPS, TRUE);
	else
		GoodOp = TRUE;

	if (GoodOp) {
		for (i=ymin_i; i<ItemsInRect; ++i)
			UsesFullPS[i] = NewUsesFullPS[i];

		if (	(!UsesFullPS[ymin_i]) || (!UsesFullPS[xmin_i]) ||
				(!UsesFullPS[ymax_i]) || (!UsesFullPS[xmax_i]) ) {
			w[PSpace].Magnifying = TRUE;
		}
/*	QM 1.01 addition, 5/24/92
 */
		else
			w[PSpace].Magnifying = FALSE;
/*	end of change
 */
		if (	(w[PSpace].GraphRange[ymin_i] != NewPSRange[ymin_i]) ||
				(w[PSpace].GraphRange[xmin_i] != NewPSRange[xmin_i]) ||
				(w[PSpace].GraphRange[ymax_i] != NewPSRange[ymax_i]) ||
				(w[PSpace].GraphRange[xmax_i] != NewPSRange[xmax_i]) ) {
			for (i=ymin_i; i<ItemsInRect; ++i)
				w[PSpace].GraphRange[i] = NewPSRange[i];
			InitPSScales();
		}
	}
	else
		for (i=ymin_i; i<ItemsInRect; ++i) {
			NewPSRange[i] = w[PSpace].GraphRange[i];
			NewUsesFullPS[i] = UsesFullPS[i];
		}
}

CheckPSClick(Point pt, double *px0)
{
	Point	tpt;
	Boolean	HitBox = FALSE;
	double	f;

	if (PSParams.NumberLine)
		if (PtInRect(pt, &NLRect)) {
			*px0 = xCoordToFloat(PSpace, pt.h - w[PSpace].Origin.h);
			return(TRUE);
		}

	tpt.h = w[PSpace].Origin.h + xFloatToCoord(PSpace, 1.0);
	f = yCoordToFloat(PSpace, w[PSpace].Origin.v - pt.v);
	f = MIN(f, 1.0);	f = MAX(f, 0.0);
	tpt.v = w[PSpace].Origin.v - yFloatToCoord(PSpace, f);
	if (PtSqrDist(pt, tpt) <= CLICK_SLOP)
		HitBox = TRUE;
	if (!HitBox) {
		tpt.v = w[PSpace].Origin.v - yFloatToCoord(PSpace, 1.0);
		f = xCoordToFloat(PSpace, pt.h - w[PSpace].Origin.h);
		f = MIN(f, 1.0);	f = MAX(f, 0.0);
		tpt.h = w[PSpace].Origin.h + xFloatToCoord(PSpace, f);
		if (PtSqrDist(pt, tpt) <= CLICK_SLOP)
			HitBox = TRUE;
	}
	if (HitBox) {
		PSParams.Box = !PSParams.Box;
		w[PSpace].DrawFromScratch = TRUE;
		StdForceUpdate((WindowPtr) w[PSpace].Peek);
		return(FALSE);
	}

	for (tpt.h=pt.h-3; tpt.h<=pt.h+3; ++tpt.h) {
		if (tpt.h<w[PSpace].GR.left)	continue;
		if (tpt.h>w[PSpace].GR.right)	break;
		tpt.v = w[PSpace].Origin.v - yFloatToCoord(PSpace,
				xCoordToFloat(PSpace, tpt.h-w[PSpace].Origin.h));
		if (PtSqrDist(pt, tpt) <= CLICK_SLOP) {
			PSParams.Line45 = !PSParams.Line45;
			w[PSpace].DrawFromScratch = TRUE;
			StdForceUpdate((WindowPtr) w[PSpace].Peek);
			return(FALSE);
		}
	}
	for (tpt.h=pt.h-3; tpt.h<=pt.h+3; ++tpt.h) {
		if (tpt.h<w[PSpace].GR.left)	continue;
		if (tpt.h>w[PSpace].GR.right)	break;
		tpt.v = w[PSpace].Origin.v - yFloatToCoord(PSpace,
				QuadraticMap(lambda,
				xCoordToFloat(PSpace, tpt.h-w[PSpace].Origin.h)));
		if (PtSqrDist(pt, tpt) <= CLICK_SLOP) {
			PSParams.Parabola = !PSParams.Parabola;
			w[PSpace].DrawFromScratch = TRUE;
			StdForceUpdate((WindowPtr) w[PSpace].Peek);
			return(FALSE);
		}
	}
	*px0 = xCoordToFloat(PSpace, pt.h - w[PSpace].Origin.h);
	return(TRUE);
}

Boolean	TrackPSRect(Point pt)
/* Gets initial conditions or other from Impact Map mouse clicks */
{
	int		*pi1, *pi2, *pi3, i, index;
	Rect	tr;

	if (!PtInRect(pt, &w[PSpace].GR)) {
		while (StillDown()) ;
		return(FALSE);
	}

	if ( TrackDashedRect(pt.h, pt.v, &w[PSpace].GR, &PSBox, TRUE) ) {
		if (PSBox.right > PSBox.left)	--PSBox.right;
		if (PSBox.bottom > PSBox.top)	--PSBox.bottom;
		tr = w[PSpace].GR;
		--tr.right;	--tr.bottom;
		pi1 = (int *) &PSBox;
		pi2 = (int *) &tr;
		pi3 = (int *) &w[PSpace].Origin;
		for (i=ymin_i; i<ItemsInRect; ++i) {
			if (i%2==0)	index = (i+2)%4;
			else		index = i;
			if ( (pi1[i] == pi2[i]) && UsesFullPS[index] ) {
				NewUsesFullPS[index] = TRUE;
				NewPSRange[index] = FullPSRange[index];
			}
			else {
				NewUsesFullPS[index] = FALSE;
				if (i%2==0)
					NewPSRange[index] =
						yCoordToFloat(PSpace, pi3[i%2] - pi1[i]);
				else
					NewPSRange[index] =
						xCoordToFloat(PSpace, pi1[i] - pi3[i%2]);
			}
		}
		return(TRUE);
	}
	/*	Really check for what aspect clickingregions?	*/
	return(FALSE);
}

InitPSScales()
{
	Rect	tmpRect;
	double	range;
	double	tmpxmin = w[PSpace].GraphRange[xmin_i];
	double	tmpxmax = w[PSpace].GraphRange[xmax_i];
	double	tmpymin = w[PSpace].GraphRange[ymin_i];
	double	tmpymax = w[PSpace].GraphRange[ymax_i];
	int		i;
	int		labelOffset, originCoord, bottomOffset;

	SetPort(w[PSpace].Peek);
	w[PSpace].DrawFromScratch = TRUE;
	StdForceUpdate((GrafPtr) w[PSpace].Peek);

		/*	Calculate how much space needed for axis labels	*/
/*
	if (w[PSpace].Magnifying) {
*/
		range = tmpymax-tmpymin;
		for(i=-1;i<2;++i) {
			if (range > 1.0) break;
			range *= 10;
		}
		PSYFieldWidth = StdPSFieldWidth + MAX(i,0);

		range = tmpxmax - tmpxmin;
		for(i=-2;i<4;++i) {
			if (range > 1.0) break;
			range *= 10;
		}
		PSXFieldWidth = StdPSFieldWidth + MAX(i,0);
/*
	}
	else
		PSYFieldWidth = PSXFieldWidth = StdPSFieldWidth;
*/

	tmpRect = w[PSpace].Peek->port.portRect;
/*
	tmpRect.top	+= PS_TopMargin;
	tmpRect.bottom -= (ScrollBarWidth + StdMargin + TextLineHeight + TextHeight);
	tmpRect.left += (StdMargin + (PSYFieldWidth+1)*CharWidth('0'));
	tmpRect.right -= (ScrollBarWidth + 2*StdMargin +
			StringWidth("\PXn"));
*/
	tmpRect.top += 2*StdMargin;
	tmpRect.bottom -= (ScrollBarWidth + StdMargin);
	if (PSParams.NumberLine) {
		tmpRect.bottom -= (NLHeight + 2 + StdMargin);
		NLRect.top = tmpRect.bottom + StdMargin +1;
		NLRect.bottom = NLRect.top + NLHeight;
		SetRect(&NLImpactRect, 0, NLRect.top-1, 3, NLRect.bottom+1);
	}
	if (tmpymax <= 0.0) {
		tmpRect.top += HashSize + TextHeight;
		tmpRect.bottom -= (3 + TextHeight + StdMargin);
	}
	else {
		tmpRect.top += TextHeight + 3;
		labelOffset = TextHeight + StdMargin + HashSize;
		bottomOffset = labelOffset;
		originCoord = RoundToInt( -MIN(tmpymin,0.0) /
				(tmpymax-tmpymin) * (tmpRect.bottom-tmpRect.top) ) - HashSize;
		originCoord = MAX(originCoord, 0);
		bottomOffset -= MIN(labelOffset, originCoord);
		bottomOffset = MAX(bottomOffset, TextHeight/2);
		tmpRect.bottom -= bottomOffset;
	}

	tmpRect.left += StdMargin;
	tmpRect.right -= (ScrollBarWidth + 2*StdMargin);
	labelOffset = PSYFieldWidth * CharWidth('0');
	if (tmpxmax <= 0.0) {
		tmpRect.left += StdMargin + StringWidth("\PXn");
		tmpRect.right -= labelOffset;
	}
	else {
		tmpRect.right -= StringWidth("\PXn");
		tmpRect.left += StdMargin + labelOffset;
		originCoord = RoundToInt( -MIN(tmpxmin,0.0) /
				(tmpxmax-tmpxmin) * (tmpRect.right-tmpRect.left) ) - HashSize;
		originCoord = MAX(originCoord, 0);
		tmpRect.left -= MIN(labelOffset, originCoord);
		tmpRect.left = MAX(tmpRect.left,
				StdMargin + PSXFieldWidth*CharWidth('0')/2);
	}

	if (PSParams.SameSizeAxes) {
		i = MIN(tmpRect.right-tmpRect.left, tmpRect.bottom-tmpRect.top);
		if (i<tmpRect.right-tmpRect.left) {
			i = (tmpRect.right-tmpRect.left) - i;
			tmpRect.left += i/2;
			tmpRect.right -= i/2 + i%2;
		}
		else if (i<tmpRect.bottom-tmpRect.top) {
			i = (tmpRect.bottom-tmpRect.top) - i;
			tmpRect.top += i/2;
			tmpRect.bottom -= i/2 + i%2;
		}
	}
	w[PSpace].GR = tmpRect;

	if (PSParams.NumberLine) {
		NLRect.left = tmpRect.left;
		NLRect.right = tmpRect.right;
	}

	w[PSpace].xScale = (double) (tmpRect.right - tmpRect.left - 1) /
			(tmpxmax - tmpxmin);
	w[PSpace].yScale = (tmpRect.bottom - tmpRect.top - 1) /
			(tmpymax - tmpymin);

	w[PSpace].xGraphAroundZero = (w[PSpace].GraphRange[xmin_i] *
		w[PSpace].GraphRange[xmax_i] <= 0);
	w[PSpace].yGraphAroundZero = (w[PSpace].GraphRange[ymin_i] *
		w[PSpace].GraphRange[ymax_i] <= 0);

	if (w[PSpace].xGraphAroundZero)
		w[PSpace].Origin.h = w[PSpace].GR.right - 1 -
				RoundToInt(tmpxmax*w[PSpace].xScale);
					/* go from right so will round down to smaller value */
	else
		if (w[PSpace].GraphRange[xmax_i] > 0)
			w[PSpace].Origin.h = w[PSpace].GR.left;
		else
			w[PSpace].Origin.h = w[PSpace].GR.right-1;
	if (w[PSpace].yGraphAroundZero)
		w[PSpace].Origin.v = w[PSpace].GR.top +
				RoundToInt(tmpymax*w[PSpace].yScale);
	else
		if (w[PSpace].GraphRange[ymax_i] > 0)
			w[PSpace].Origin.v = w[PSpace].GR.bottom-1;
		else
			w[PSpace].Origin.v = w[PSpace].GR.top;
/*
		w[PSpace].Origin.v = w[PSpace].GR.bottom;
*/

	w[PSpace].BackOrigin = w[PSpace].Origin;
	SetRect(&w[PSpace].PR,
			w[PSpace].GR.left - w[PSpace].Origin.h,
			w[PSpace].Origin.v - w[PSpace].GR.bottom + 1,
			w[PSpace].GR.right - w[PSpace].Origin.h,
			w[PSpace].Origin.v - w[PSpace].GR.top + 1);
/*
	SetRect(&w[PSpace].PR,
			w[PSpace].GR.left - w[PSpace].Origin.h,
			w[PSpace].Origin.v - w[PSpace].GR.bottom,
			w[PSpace].GR.right - w[PSpace].Origin.h,
			w[PSpace].Origin.v - w[PSpace].GR.top);
*/
/*
			w[PSpace].GR.right - w[PSpace].Origin.h + 1,
			w[PSpace].Origin.v - w[PSpace].GR.top + 1);
*/
			/* +1 to account for pixel hanging below and to left of point */

	tmpRect.right = tmpRect.left;
	tmpRect.left = 0;
/*
	tmpRect.left -= MAX(StringWidth(VMaxString), StringWidth(VMinString))+3;
	tmpRect.top -= TextHeight/2;
	tmpRect.bottom += TextHeight/2+1;
*/
	ResizeControl(w[PSpace].HS[0], &tmpRect);

	tmpRect = w[PSpace].GR;
	tmpRect.top = tmpRect.bottom;
	if (PSParams.NumberLine)
		tmpRect.bottom = NLRect.top - 1;
	else
		tmpRect.bottom = w[PSpace].Peek->port.portRect.bottom - (ScrollBarWidth-1);
	ResizeControl(w[PSpace].HS[1], &tmpRect);
}

Boolean UseIterLines(int myHN)
{
	return( PSParams.IterationLines && (myHN >= StartILHitNumber) );
}

PlotInPhaseSpace(double x_j, int myHN, int myCHN, Boolean Highlighting)
{
			double		x_k, x_i;
	static	Point		pt;
			Point		opt = pt, pt2, pt3, pt4;
			int			NLCoord;
			PenState	ps;
			RgnHandle	tmpRgn1, tmpRgn2;
			Rect		tr1, tr2, tr3;
			int			ArrowCoord, sign;

	w[PSpace].WindowCleared = FALSE;

	SetPort(w[PSpace].Peek);
	GetPenState(&ps);

	if (Highlighting) {
		if (NLImpactRect.left != 0) {
			 InvertRect(&NLImpactRect);
			SetPort(w[PSpace].Back);
			 InvertRect(&NLImpactRect);
			SetPort(w[PSpace].Peek);
		}
		if (PSLatestImpactRect.left != 0) {
			 InvertOval(&PSLatestImpactRect);
			SetPort(w[PSpace].Back);
			 InvertOval(&PSLatestImpactRect);
			SetPort(w[PSpace].Peek);
		}
	}

	if (PSParams.NumberLine) {
		NLCoord = xFloatToCoord(PSpace, x_j);
		BothMoveTo(PSpace, NLCoord, w[PSpace].Origin.v - NLRect.top);
		SetPort(w[PSpace].Peek);	Line(0, NLHeight);
		SetPort(w[PSpace].Back);	Line(0, NLHeight);
		SetPort(w[PSpace].Peek);
		if (Highlighting) {
			OffsetRect(&NLImpactRect, -NLImpactRect.left + NLCoord +
				w[PSpace].Origin.h - 1, 0);
			 InvertRect(&NLImpactRect);
			SetPort(w[PSpace].Back);
			 InvertRect(&NLImpactRect);
			SetPort(w[PSpace].Peek);
		}
	}

	if (myCHN == 0)
		CurrentPSOffsetCount = -1;
	if (++CurrentPSOffsetCount < 1) {
		OffsetRect(&PSLatestImpactRect,
				-PSLatestImpactRect.left, -PSLatestImpactRect.top);
		return;
	}

	x_i = GetSolution(myHN-1);
	if (CurrentPSOffsetCount > 1)
		x_k = GetSolution(myHN-2);
			/*	May need to chart from (x_k, x_i) *to* (x_i, x_j)	*/
	pt.h = xFloatToCoord(PSpace, x_i);
	pt.v = yFloatToCoord(PSpace, x_j);

	if ( (CurrentPSOffsetCount > 1) && (UseIterLines(myHN)) ) {
		SetPort(w[PSpace].Peek);	tmpRgn1 = NewRgn();
			GetClip(tmpRgn1);		ClipRect(&w[PSpace].GR);
		SetPort(w[PSpace].Back);	tmpRgn2 = NewRgn();
			GetClip(tmpRgn2);		ClipRect(&w[PSpace].GR);
		pt2.v = opt.v;
		pt2.h = pt.h;
		BothMoveTo(PSpace, opt.h, opt.v);
		SetPort(w[PSpace].Peek);
		Line(pt2.h-opt.h, -(pt2.v-opt.v));
		SetPort(w[PSpace].Back);
		Line(pt2.h-opt.h, -(pt2.v-opt.v));
		if (PSParams.Arrows) {
			tr1 = w[PSpace].GR;	--tr1.right; --tr1.bottom;
			OffsetRect(&tr1, -w[PSpace].Origin.h, 0);
			tr1.top = w[PSpace].Origin.v - tr1.bottom;
			tr1.bottom = w[PSpace].Origin.v - w[PSpace].GR.top;
			SetRect(&tr2, MIN(opt.h,pt2.h),opt.v, MAX(opt.h,pt2.h),opt.v+1);
			if (!SectRect(&tr1, &tr2, &tr3)) ;
			if (ABS(tr3.right-tr3.left) >= ARROW_SIZE) {
				sign = (pt2.h>opt.h) ? -1 : 1;
				ArrowCoord = sign*ARROW_SIZE;
				BothMoveTo(PSpace, (tr3.right+tr3.left-sign)/2, opt.v);
				SetPort(w[PSpace].Peek);
				 Move(-ArrowCoord/2, 0);
				 Line(ArrowCoord, ARROW_SIZE);
				 Move(-ArrowCoord, -ARROW_SIZE);
				 Line(ArrowCoord, -ARROW_SIZE);
				SetPort(w[PSpace].Back);
				 Move(-ArrowCoord/2, 0);
				 Line(ArrowCoord, ARROW_SIZE);
				 Move(-ArrowCoord, -ARROW_SIZE);
				 Line(ArrowCoord, -ARROW_SIZE);
			}
		}
		BothMoveTo(PSpace, pt2.h, pt2.v);
		SetPort(w[PSpace].Peek);
		Line(pt.h-pt2.h, -(pt.v-pt2.v));
		SetPort(w[PSpace].Back);
		Line(pt.h-pt2.h, -(pt.v-pt2.v));
		if (PSParams.Arrows) {
			SetRect(&tr2, pt.h, MIN(pt.v,pt2.v), pt.h+1, MAX(pt.v,pt2.v));
			if (!SectRect(&tr1, &tr2, &tr3)) ;
			if (ABS(tr3.bottom-tr3.top) >= ARROW_SIZE) {
				sign = (pt.v>pt2.v) ? 1 : -1;
				ArrowCoord = sign * ARROW_SIZE;
				BothMoveTo(PSpace, pt.h, (tr3.bottom+tr3.top + sign)/2);
				SetPort(w[PSpace].Peek);
				 Move(0, -ArrowCoord/2);
				 Line(ARROW_SIZE, ArrowCoord);
				 Move(-ARROW_SIZE, -ArrowCoord);
				 Line(-ARROW_SIZE, ArrowCoord);
				SetPort(w[PSpace].Back);
				 Move(0, -ArrowCoord/2);
				 Line(ARROW_SIZE, ArrowCoord);
				 Move(-ARROW_SIZE, -ArrowCoord);
				 Line(-ARROW_SIZE, ArrowCoord);
			}
		}
		SetPort(w[PSpace].Peek);	SetClip(tmpRgn1);	DisposeRgn(tmpRgn1);
		SetPort(w[PSpace].Back);	SetClip(tmpRgn2);	DisposeRgn(tmpRgn2);
	}

	if ( !PtInRect(pt, &w[PSpace].PR) ) {
		if (Highlighting)
			OffsetRect(&PSLatestImpactRect,-PSLatestImpactRect.left,
					-PSLatestImpactRect.top);
		return;
	}

	if (PSParams.Parabola && !UseIterLines(myHN)) {
		SetPort(w[PSpace].Peek);	PenMode(patBic);
		SetPort(w[PSpace].Back);	PenMode(patBic);
		SetPort(w[PSpace].Peek);
	}
	BothMoveTo(PSpace, pt.h, pt.v);
	SetPort(w[PSpace].Peek);	Line(0, 0);
	SetPort(w[PSpace].Back);	Line(0, 0);
	SetPort(w[PSpace].Peek);
	if (PSParams.Parabola && !UseIterLines(myHN)) {
		SetPort(w[PSpace].Peek);	PenMode(ps.pnMode);
		SetPort(w[PSpace].Back);	PenMode(ps.pnMode);
		SetPort(w[PSpace].Peek);
	}

	if (Highlighting) {
			/* next statement from WindowMoveTo */
		OffsetRect(&PSLatestImpactRect, -PSLatestImpactRect.left + pt.h +
			w[PSpace].Origin.h - 2,
			-PSLatestImpactRect.top + w[PSpace].Origin.v - pt.v - 2);
		 InvertOval(&PSLatestImpactRect);
		SetPort(w[PSpace].Back);
		 InvertOval(&PSLatestImpactRect);
		SetPort(w[PSpace].Peek);
	}
}

RefreshPS(int myHitNumber)
{
	int			i, x, y;
	Rect		tmpRect;
	double		HashDist;
	int			HashNum;
	double		tmpXMin = w[PSpace].GraphRange[xmin_i];
	double		tmpXMax = w[PSpace].GraphRange[xmax_i];
	double		tmpYMin = w[PSpace].GraphRange[ymin_i];
	double		tmpYMax = w[PSpace].GraphRange[ymax_i];
	RgnHandle	tmpClipRgn;

	int		OverflowOnRight, MaxPStrLen;	/* intermediate vars	*/

	if (!w[PSpace].DrawFromScratch) {
		SetPort(w[PSpace].Peek);
		myCopyBits(w[PSpace].Back, (GrafPtr) w[PSpace].Peek,
			&w[PSpace].Peek->port.portRect, 0,0,0,0, srcCopy, (RgnHandle) NIL);
		return;
	}

	w[PSpace].WindowCleared = TRUE;

	w[PSpace].DrawFromScratch = FALSE;
	SetPort(w[PSpace].Back);
	EraseRect(&w[PSpace].Back->portRect);

	/*	Draw Axes	*/
	HashDist = 0.0;
	HashNum = 0;
/*	(same below)
	if ( (UseAutoXMin && UseAutoXMax) )
		if (UsesFullPS[xmin_i] && UsesFullPS[xmax_i])
*/
	HashNum = (w[PSpace].GR.right - w[PSpace].GR.left) /
			(PSXFieldWidth*CharWidth('0')+3) + 1;
	if ( (w[PSpace].GraphRange[xmin_i] == AutoXMin) &&
			(w[PSpace].GraphRange[xmax_i] == AutoXMax) ) {
		if ( (AutoXMax - AutoXMin)/0.1 > HashNum )
			HashDist = 0.25;
		else
			HashDist = 0.1;
	}
	else
		HashNum = MIN(HashNum, 10);

	DrawHAxis(TRUE, w[PSpace].GraphRange[xmin_i],
			w[PSpace].GraphRange[xmax_i],
			&w[PSpace].GR, w[PSpace].Origin, w[PSpace].xScale,
			HashDist, HashNum, FALSE, HashSize, HashSize-1, PSXFieldWidth,
			2, 3, 3, 0);

	HashDist = 0.0;
	HashNum = 0;
	HashNum = (w[PSpace].GR.bottom - w[PSpace].GR.top) /
			(PSXFieldWidth*CharWidth('0')+3) + 1;
	if ( (w[PSpace].GraphRange[ymin_i] == AutoXMin) &&
			(w[PSpace].GraphRange[ymax_i] == AutoXMax) ) {
		if ( (AutoXMax - AutoXMin)/0.1 > HashNum )
			HashDist = 0.25;
		else
			HashDist = 0.1;
	}
	else
		HashNum = MIN(HashNum, 10);

	DrawVAxis(TRUE, w[PSpace].GraphRange[ymin_i],
			w[PSpace].GraphRange[ymax_i],
			&w[PSpace].GR, w[PSpace].Origin, w[PSpace].yScale,
			HashDist, HashNum, FALSE, HashSize, HashSize-1, PSYFieldWidth,
			2, 3, 3, HashSize+2+TextHeight);

	if (w[PSpace].GraphRange[xmax_i] <= 0.0) {
		MoveTo(w[PSpace].GR.left, w[PSpace].Origin.v);
		Move(-StdMargin - StringWidth("\PXn"), 0);
	}
	else {
		MoveTo(w[PSpace].GR.right, w[PSpace].Origin.v);
		Move(StdMargin, 0);
	}
	Move(0, TextHeight/2 - TextHeight%2);
	DrawXn("\Pn");

	if (w[PSpace].GraphRange[ymax_i] <= 0.0) {
		MoveTo(w[PSpace].Origin.h, w[PSpace].GR.bottom);
		Move(0, StdMargin + TextHeight);
	}
	else {
		MoveTo(w[PSpace].Origin.h, w[PSpace].GR.top);
		Move(0, -StdMargin-3);
	}
	Move(-StringWidth("\PXn+1")/2, 0);
	DrawXn("\Pn+1");

	if (PSParams.NumberLine) {
		tmpRect = NLRect;
		InsetRect(&tmpRect, -1, -1);
		FrameRect(&tmpRect);
	}

	tmpClipRgn = NewRgn();
	GetClip(tmpClipRgn);
	ClipRect(&w[PSpace].GR);

	if (PSParams.Parabola) {
		InitFixedQMVars(lambda, w[PSpace].xScale, w[PSpace].yScale,
				&w[PSpace].GraphRange, w[PSpace].xGraphAroundZero,
				w[PSpace].yGraphAroundZero);
		for (x=w[PSpace].GR.left-w[PSpace].Origin.h;
				x<w[PSpace].GR.right-w[PSpace].Origin.h; ++x) {
/*
			y = yFloatToCoord(PSpace,
					QuadraticMap(lambda, xCoordToFloat(PSpace, x)));
*/
			y = FixedQM(x, 1);
			if (x==w[PSpace].GR.left-w[PSpace].Origin.h)
				BackMoveTo(PSpace, x, y);
			else
				BackLineTo(PSpace, x, y);
		}
	}

	if (PSParams.Line45)
		if ( (tmpXMax > tmpYMin) && (tmpXMin < tmpYMax) ) {
			x = xFloatToCoord(PSpace, MAX(tmpXMin, tmpYMin));
			y = yFloatToCoord(PSpace, MAX(tmpXMin, tmpYMin));
			BackMoveTo(PSpace, x, y);
			x = xFloatToCoord(PSpace, MIN(tmpXMax, tmpYMax));
			y = yFloatToCoord(PSpace, MIN(tmpXMax, tmpYMax));
			BackLineTo(PSpace, x, y);
		}

	if (PSParams.Box) {
		if ( ((tmpXMax-1.0) * (tmpXMin-1.0) <= 0.0) &&
				(tmpYMin<1.0) && (tmpYMax>0.0) ) {
			x = xFloatToCoord(PSpace, 1.0);
			y = yFloatToCoord(PSpace, MAX(0.0, tmpYMin));
			BackMoveTo(PSpace, x, y);
			y = yFloatToCoord(PSpace, MIN(1.0, tmpYMax));
			BackLineTo(PSpace, x, y);
		}
		if ( ((tmpYMax-1.0) * (tmpYMin-1.0) <= 0.0) &&
				(tmpXMin<1.0) && (tmpXMax>0.0) ) {
			y = yFloatToCoord(PSpace, 1.0);
			x = xFloatToCoord(PSpace, MAX(0.0, tmpXMin));
			BackMoveTo(PSpace, x, y);
			x = xFloatToCoord(PSpace, MIN(1.0, tmpXMax));
			BackLineTo(PSpace, x, y);
		}
	}

	SetClip(tmpClipRgn);
	DisposeRgn(tmpClipRgn);

	SetPort(w[PSpace].Peek);
	myCopyBits(w[PSpace].Back, (GrafPtr) w[PSpace].Peek,
			&w[PSpace].Peek->port.portRect, 0,0,0,0, srcCopy, (RgnHandle) NIL);

	OffsetRect(&PSLatestImpactRect,-PSLatestImpactRect.left,-PSLatestImpactRect.top);
	OffsetRect(&NLImpactRect,-NLImpactRect.left, 0);
	if (myHitNumber >= StartHitNumber) {
		if (myHitNumber - StartHitNumber > MANY_ITERATIONS)
			StartWaiting();
		for (i=StartHitNumber;i<=myHitNumber-1;++i)
			PlotInPhaseSpace(GetSolution(i), i, GetSolutionHit(i), FALSE);
		if (myHitNumber - StartHitNumber > MANY_ITERATIONS)
			StopWaiting();
		PlotInPhaseSpace(GetSolution(i), i, GetSolutionHit(i), TRUE);
	}
}


	/*	---------------------------------------------------	*/
	/*	---------- Embedded Phase Space Routines ----------	*/
	/*	---------------------------------------------------	*/


GetEmbeddedPSInfo()
{
	Boolean	oSSA = EPSParams.SameSizeAxes;

	if (DoEmbeddedPSDialog(&Embedding, &EPSParams)) {
		if (oSSA != EPSParams.SameSizeAxes)
			InitEPSScales();
		w[EmbPS].DrawFromScratch = TRUE;
		StdForceUpdate((GrafPtr) w[EmbPS].Peek);
	}
}

	/*	Routines related to Impact Phases magnification	*/

InitFullEPSRange()
{
	int	i;

	FullEPSRange[ymin_i] = XMin;
	FullEPSRange[xmin_i] = XMin;
	FullEPSRange[ymax_i] = XMax;
	FullEPSRange[xmax_i] = XMax;

	for (i=ymin_i; i<=xmin_i; ++i) {		/*	ymin_i, xmin_i	*/
		if (UsesFullEPS[i]) {
			if (UsesFullEPS[i+2] || (w[EmbPS].GraphRange[i+2] > XMin))
				w[EmbPS].GraphRange[i] = FullEPSRange[i];
			else
				UsesFullEPS[i] = FALSE;
		}
		if (NewUsesFullEPS[i]) {
			if (NewUsesFullEPS[i+2] || (NewEPSRange[i+2] > XMin))
				NewEPSRange[i] = FullEPSRange[i];
			else
				NewUsesFullEPS[i] = FALSE;
		}
	}
	for (i=ymax_i; i<=xmax_i; ++i) {		/*	ymax_i, xmax_i	*/
		if (UsesFullEPS[i]) {
			if (UsesFullEPS[i-2] || (w[EmbPS].GraphRange[i-2] < XMax))
				w[EmbPS].GraphRange[i] = FullEPSRange[i];
			else
				UsesFullEPS[i] = FALSE;
		}
		if (NewUsesFullEPS[i]) {
			if (NewUsesFullEPS[i-2] || (NewEPSRange[i-2] < XMax))
				NewEPSRange[i] = FullEPSRange[i];
			else
				NewUsesFullEPS[i] = FALSE;
		}
	}
}

SetFullEPSRange()
{
	int	i;

	InitFullEPSRange();

	for (i=ymin_i; i<ItemsInRect; ++i) {
		NewUsesFullEPS[i] = UsesFullEPS[i] = TRUE;
		w[EmbPS].GraphRange[i] = NewEPSRange[i] = FullEPSRange[i];
	}
	w[EmbPS].Magnifying = FALSE;
	InitEPSScales();
}

GetEPSMagnification(Boolean SkipDialog)
{
	int		i;
	double	min, max;
	Boolean	GoodOp;

	InitFullEPSRange();

	Round2Doubles(NewEPSRange[xmin_i], NewEPSRange[xmax_i], RelPlaces, &min, &max);
	if (NewEPSRange[xmin_i] > FullEPSRange[xmin_i])
		NewEPSRange[xmin_i] = min;
	if (NewEPSRange[xmax_i] < FullEPSRange[xmax_i])
		NewEPSRange[xmax_i] = max;
	Round2Doubles(NewEPSRange[ymin_i], NewEPSRange[ymax_i], RelPlaces, &min, &max);
	if (NewEPSRange[ymin_i] > FullEPSRange[ymin_i])
		NewEPSRange[ymin_i] = min;
	if (NewEPSRange[ymax_i] < FullEPSRange[ymax_i])
		NewEPSRange[ymax_i] = max;

/* 1.01 change
	if (!SkipDialog)
 */
	if ( (!SkipDialog) || (NewEPSRange[xmin_i] == NewEPSRange[xmax_i]) ||
			(NewEPSRange[ymin_i] == NewEPSRange[ymax_i]) )
		GoodOp = DoMagnifyPhaseSpacesDialog(&NewEPSRange, &FullEPSRange,
				&NewUsesFullEPS, FALSE);
	else
		GoodOp = TRUE;

	if (GoodOp) {
		for (i=ymin_i; i<ItemsInRect; ++i)
			UsesFullEPS[i] = NewUsesFullEPS[i];

		if (	(!UsesFullEPS[ymin_i]) || (!UsesFullEPS[xmin_i]) ||
				(!UsesFullEPS[ymax_i]) || (!UsesFullEPS[xmax_i]) ) {
			w[EmbPS].Magnifying = TRUE;
		}
/*	QM 1.01 addition, 5/24/92
 */
		else
			w[EmbPS].Magnifying = FALSE;
/*	end of change
 */
		if (	(w[EmbPS].GraphRange[ymin_i] != NewEPSRange[ymin_i]) ||
				(w[EmbPS].GraphRange[xmin_i] != NewEPSRange[xmin_i]) ||
				(w[EmbPS].GraphRange[ymax_i] != NewEPSRange[ymax_i]) ||
				(w[EmbPS].GraphRange[xmax_i] != NewEPSRange[xmax_i]) ) {
			for (i=ymin_i; i<ItemsInRect; ++i)
				w[EmbPS].GraphRange[i] = NewEPSRange[i];
			InitEPSScales();
		}
	}
	else
		for (i=ymin_i; i<ItemsInRect; ++i) {
			NewEPSRange[i] = w[EmbPS].GraphRange[i];
			NewUsesFullEPS[i] = UsesFullEPS[i];
		}
}

CheckEPSClick(Point pt)
{
	Point	tpt;
	Boolean	HitBox = FALSE;
	double	f;

	tpt.h = w[EmbPS].Origin.h + xFloatToCoord(EmbPS, 1.0);
	f = yCoordToFloat(EmbPS, w[EmbPS].Origin.v - pt.v);
	f = MIN(f, 1.0);	f = MAX(f, 0.0);
	tpt.v = w[EmbPS].Origin.v - yFloatToCoord(EmbPS, f);
	if (PtSqrDist(pt, tpt) <= CLICK_SLOP)
		HitBox = TRUE;
	if (!HitBox) {
		tpt.v = w[EmbPS].Origin.v - yFloatToCoord(EmbPS, 1.0);
		f = xCoordToFloat(EmbPS, pt.h - w[EmbPS].Origin.h);
		f = MIN(f, 1.0);	f = MAX(f, 0.0);
		tpt.h = w[EmbPS].Origin.h + xFloatToCoord(EmbPS, f);
		if (PtSqrDist(pt, tpt) <= CLICK_SLOP)
			HitBox = TRUE;
	}
	if (HitBox) {
		EPSParams.Box = !EPSParams.Box;
		w[EmbPS].DrawFromScratch = TRUE;
		StdForceUpdate((WindowPtr) w[EmbPS].Peek);
		return;
	}

	for (tpt.h=pt.h-3; tpt.h<=pt.h+3; ++tpt.h) {
		if (tpt.h<w[EmbPS].GR.left)	continue;
		if (tpt.h>w[EmbPS].GR.right)	break;
		tpt.v = w[EmbPS].Origin.v - yFloatToCoord(EmbPS,
				xCoordToFloat(EmbPS, tpt.h-w[EmbPS].Origin.h));
		if (PtSqrDist(pt, tpt) <= CLICK_SLOP) {
			EPSParams.Line45 = !EPSParams.Line45;
			w[EmbPS].DrawFromScratch = TRUE;
			StdForceUpdate((WindowPtr) w[EmbPS].Peek);
			return;
		}
	}
	for (tpt.h=pt.h-3; tpt.h<=pt.h+3; ++tpt.h) {
		if (tpt.h<w[EmbPS].GR.left)	continue;
		if (tpt.h>w[EmbPS].GR.right)	break;
		tpt.v = w[EmbPS].Origin.v - yFloatToCoord(EmbPS,
				nQuadMap(Embedding, lambda,
				xCoordToFloat(EmbPS, tpt.h-w[EmbPS].Origin.h)));
		if (PtSqrDist(pt, tpt) <= CLICK_SLOP) {
			EPSParams.Polynomial2n = !EPSParams.Polynomial2n;
			w[EmbPS].DrawFromScratch = TRUE;
			StdForceUpdate((WindowPtr) w[EmbPS].Peek);
			return;
		}
	}
}

Boolean	TrackEPSRect(Point pt)
/* Gets EPS magnification from mouse presses */
{
	int		*pi1, *pi2, *pi3, i, index;
	Rect	tr;

	if ( TrackDashedRect(pt.h, pt.v, &w[EmbPS].GR, &EPSBox, TRUE) ) {
		if (EPSBox.right > EPSBox.left)	--EPSBox.right;
		if (EPSBox.bottom > EPSBox.top)	--EPSBox.bottom;
		tr = w[EmbPS].GR;
		--tr.right;	--tr.bottom;
		pi1 = (int *) &EPSBox;
		pi2 = (int *) &tr;
		pi3 = (int *) &w[EmbPS].Origin;
		for (i=ymin_i; i<ItemsInRect; ++i) {
			if (i%2==0)	index = (i+2)%4;
			else		index = i;
			if ( (pi1[i] == pi2[i]) && UsesFullEPS[index] ) {
				NewUsesFullEPS[index] = TRUE;
				NewEPSRange[index] = FullEPSRange[index];
			}
			else {
				NewUsesFullEPS[index] = FALSE;
				if (i%2==0)
					NewEPSRange[index] =
						yCoordToFloat(EmbPS, pi3[i%2] - pi1[i]);
				else
					NewEPSRange[index] =
						xCoordToFloat(EmbPS, pi1[i] - pi3[i%2]);
			}
		}
		return(TRUE);
	}
	/*	Really check for what aspect clickingregions?	*/
	return(FALSE);
}

InitEPSScales()
{
	Rect	tmpRect;
	double	range;
	double	tmpxmin = w[EmbPS].GraphRange[xmin_i];
	double	tmpxmax = w[EmbPS].GraphRange[xmax_i];
	double	tmpymin = w[EmbPS].GraphRange[ymin_i];
	double	tmpymax = w[EmbPS].GraphRange[ymax_i];
	int		i;
	int		labelOffset, originCoord, bottomOffset;

	SetPort(w[EmbPS].Peek);
	w[EmbPS].DrawFromScratch = TRUE;
	StdForceUpdate((WindowPtr) w[EmbPS].Peek);

		/*	Calculate space needed for axis labels	*/
/*
	if (w[EmbPS].Magnifying) {
*/
		range = tmpymax-tmpymin;
		for(i=-1;i<2;++i) {
			if (range > 1.0) break;
			range *= 10;
		}
		EPSYFieldWidth = StdPSFieldWidth + MAX(i,0);

		range = tmpxmax - tmpxmin;
		for(i=-1;i<2;++i) {
			if (range > 1.0) break;
			range *= 10;
		}
		EPSXFieldWidth = StdPSFieldWidth + MAX(i,0);
/*
	}
	else
		EPSYFieldWidth = EPSXFieldWidth = StdPSFieldWidth;
*/

	tmpRect = w[EmbPS].Peek->port.portRect;
/*
	tmpRect.left += (StdMargin + (EPSYFieldWidth+3)*CharWidth('0'));
	tmpRect.top	+= PS_TopMargin;
	tmpRect.right -= (ScrollBarWidth + StdMargin +
			EPSXFieldWidth * CharWidth('0') / 2);
	tmpRect.bottom -= (ScrollBarWidth + StdMargin + TextLineHeight + TextHeight);
*/
	tmpRect.top += 2*StdMargin;
	tmpRect.bottom -= (ScrollBarWidth + StdMargin);
	if (tmpymax <= 0.0) {
		tmpRect.top += HashSize + TextHeight;
		tmpRect.bottom -= (3 + TextHeight + StdMargin);
	}
	else {
		tmpRect.top += TextHeight + 3;
		labelOffset = TextHeight + StdMargin + HashSize;
		bottomOffset = labelOffset;
		originCoord = RoundToInt( -MIN(tmpymin,0.0) /
				(tmpymax-tmpymin) * (tmpRect.bottom-tmpRect.top) ) - HashSize;
		originCoord = MAX(originCoord, 0);
		bottomOffset -= MIN(labelOffset, originCoord);
		bottomOffset = MAX(bottomOffset, TextHeight/2);
		tmpRect.bottom -= bottomOffset;
	}

	tmpRect.left += StdMargin;
	tmpRect.right -= (ScrollBarWidth + 2*StdMargin);
	labelOffset = EPSYFieldWidth * CharWidth('0');
	if (tmpxmax <= 0.0) {
		tmpRect.left += StdMargin + StringWidth("\PXn");
		tmpRect.right -= labelOffset;
	}
	else {
		tmpRect.right -= StringWidth("\PXn");
		tmpRect.left += StdMargin + labelOffset;
		originCoord = RoundToInt( -MIN(tmpxmin,0.0) /
				(tmpxmax-tmpxmin) * (tmpRect.right-tmpRect.left) ) - HashSize;
		originCoord = MAX(originCoord, 0);
		tmpRect.left -= MIN(labelOffset, originCoord);
		tmpRect.left = MAX(tmpRect.left,
				StdMargin + EPSXFieldWidth*CharWidth('0')/2);
	}

	if (EPSParams.SameSizeAxes) {
		i = MIN(tmpRect.right-tmpRect.left, tmpRect.bottom-tmpRect.top);
		if (i<tmpRect.right-tmpRect.left) {
			i = (tmpRect.right-tmpRect.left) - i;
			tmpRect.left += i/2;
			tmpRect.right -= i/2 + i%2;
		}
		else if (i<tmpRect.bottom-tmpRect.top) {
			i = (tmpRect.bottom-tmpRect.top) - i;
			tmpRect.top += i/2;
			tmpRect.bottom -= i/2 + i%2;
		}
	}
	w[EmbPS].GR = tmpRect;

	w[EmbPS].xScale = (double) (tmpRect.right - tmpRect.left - 1) /
			(tmpxmax - tmpxmin);
	w[EmbPS].yScale = (tmpRect.bottom - tmpRect.top - 1) /
			(tmpymax - tmpymin);

	MaxFixedEmbedding = (int) ( log(1/FIXED_ACCURACY/2.0/w[EmbPS].yScale) /
			log(1.0*fabs(lambda)) );

	w[EmbPS].xGraphAroundZero = (w[EmbPS].GraphRange[xmin_i] *
		w[EmbPS].GraphRange[xmax_i] <= 0);
	w[EmbPS].yGraphAroundZero = (w[EmbPS].GraphRange[ymin_i] *
		w[EmbPS].GraphRange[ymax_i] <= 0);

	if (w[EmbPS].xGraphAroundZero)
		w[EmbPS].Origin.h = w[EmbPS].GR.right - 1 -
				RoundToInt(tmpxmax*w[EmbPS].xScale);
	else
		if (w[EmbPS].GraphRange[xmax_i] > 0)
			w[EmbPS].Origin.h = w[EmbPS].GR.left;
		else
			w[EmbPS].Origin.h = w[EmbPS].GR.right-1;
	if (w[EmbPS].yGraphAroundZero)
		w[EmbPS].Origin.v = w[EmbPS].GR.top +
				RoundToInt(tmpymax*w[EmbPS].yScale);
	else
		if (w[EmbPS].GraphRange[ymax_i] > 0)
			w[EmbPS].Origin.v = w[EmbPS].GR.bottom-1;
		else
			w[EmbPS].Origin.v = w[EmbPS].GR.top;
/*
		w[EmbPS].Origin.v = w[EmbPS].GR.bottom;
*/

	w[EmbPS].BackOrigin = w[EmbPS].Origin;
	SetRect(&w[EmbPS].PR,
			w[EmbPS].GR.left - w[EmbPS].Origin.h,
			w[EmbPS].Origin.v - w[EmbPS].GR.bottom + 1,
			w[EmbPS].GR.right - w[EmbPS].Origin.h,
			w[EmbPS].Origin.v - w[EmbPS].GR.top + 1);
/*
	SetRect(&w[EmbPS].PR,
			w[EmbPS].GR.left - w[EmbPS].Origin.h,
			w[EmbPS].Origin.v - w[EmbPS].GR.bottom,
			w[EmbPS].GR.right - w[EmbPS].Origin.h + 1,
			w[EmbPS].Origin.v - w[EmbPS].GR.top + 1);
*/
			/* +1 to account for pixel hanging below and to left of point */

	tmpRect.right = tmpRect.left;
	tmpRect.left = 0;
/*
	tmpRect.left -= MAX(StringWidth(VMaxString), StringWidth(VMinString))+3;
	tmpRect.top -= TextHeight/2;
	tmpRect.bottom += TextHeight/2+1;
*/
	ResizeControl(w[EmbPS].HS[0], &tmpRect);


	tmpRect = w[EmbPS].GR;
	tmpRect.top = tmpRect.bottom;
	tmpRect.bottom = w[EmbPS].Peek->port.portRect.bottom - (ScrollBarWidth-1);
	ResizeControl(w[EmbPS].HS[1], &tmpRect);
}

PlotInEmbeddedPS(double x_n, int myHN, int myCHN, Boolean Highlighting)
{
	Point		pt;
	double		x_nk;
	PenState	ps;

	w[EmbPS].WindowCleared = FALSE;

	SetPort(w[EmbPS].Peek);
	GetPenState(&ps);

	if (Highlighting)
		if (EPSLatestImpactRect.left != 0) {
			 InvertOval(&EPSLatestImpactRect);
			SetPort(w[EmbPS].Back);
			 InvertOval(&EPSLatestImpactRect);
			SetPort(w[EmbPS].Peek);
		}
	if (myCHN == 0)
		CurrentEPSOffsetCount = -1;
	if (++CurrentEPSOffsetCount < Embedding) {
		OffsetRect(&EPSLatestImpactRect,
				-EPSLatestImpactRect.left, -EPSLatestImpactRect.top);
		return;
	}

	x_nk = GetSolution(myHN - Embedding);
	pt.h = xFloatToCoord(EmbPS, x_nk);
	pt.v = yFloatToCoord(EmbPS, x_n);

	if ( !PtInRect(pt, &w[EmbPS].PR) ) {
		if (Highlighting)
			OffsetRect(&EPSLatestImpactRect,-EPSLatestImpactRect.left,
					-EPSLatestImpactRect.top);
		return;
	}

	if (EPSParams.Polynomial2n) {
		SetPort(w[EmbPS].Peek);	PenMode(patBic);
		SetPort(w[EmbPS].Back);	PenMode(patBic);
		SetPort(w[EmbPS].Peek);
	}
	BothMoveTo(EmbPS, pt.h, pt.v);
	SetPort(w[EmbPS].Peek);	Line(0, 0);
	SetPort(w[EmbPS].Back);	Line(0, 0);
	SetPort(w[EmbPS].Peek);
	if (EPSParams.Polynomial2n) {
		SetPort(w[EmbPS].Peek);	PenMode(ps.pnMode);
		SetPort(w[EmbPS].Back);	PenMode(ps.pnMode);
		SetPort(w[EmbPS].Peek);
	}

	if (Highlighting) {
			/* next statement from WindowMoveTo */
		OffsetRect(&EPSLatestImpactRect, -EPSLatestImpactRect.left + pt.h +
			w[EmbPS].Origin.h - 2,
			-EPSLatestImpactRect.top + w[EmbPS].Origin.v - pt.v - 2);
		 InvertOval(&EPSLatestImpactRect);
		SetPort(w[EmbPS].Back);
		 InvertOval(&EPSLatestImpactRect);
		SetPort(w[EmbPS].Peek);
	}
}

RefreshEPS(int myHitNumber)
{
	int			i, j, x, y;
	Str255		s;
	Rect		tmpRect;
	double		HashDist;
	int			HashNum;
	double		tmpXMin = w[EmbPS].GraphRange[xmin_i];
	double		tmpXMax = w[EmbPS].GraphRange[xmax_i];
	double		tmpYMin = w[EmbPS].GraphRange[ymin_i];
	double		tmpYMax = w[EmbPS].GraphRange[ymax_i];
	RgnHandle	tmpClipRgn;

	int		OverflowOnRight, MaxPStrLen;	/* intermediate vars	*/

	if (!w[EmbPS].DrawFromScratch) {
		SetPort(w[EmbPS].Peek);
		myCopyBits(w[EmbPS].Back, (GrafPtr) w[EmbPS].Peek,
			&w[EmbPS].Peek->port.portRect, 0,0,0,0, srcCopy, (RgnHandle) NIL);
		return;
	}

	w[EmbPS].WindowCleared = TRUE;

	w[EmbPS].DrawFromScratch = FALSE;
	SetPort(w[EmbPS].Back);
	EraseRect(&w[EmbPS].Back->portRect);

	/*	Draw Axes	*/
	HashDist = 0.0;
	HashNum = 0;
/*	(same below)
	if ( (UseAutoXMin && UseAutoXMax) )
		if (UsesFullPS[xmin_i] && UsesFullPS[xmax_i])
*/
	HashNum = (w[EmbPS].GR.right - w[EmbPS].GR.left) /
			(PSXFieldWidth*CharWidth('0')+3) + 1;
	if ( (w[EmbPS].GraphRange[xmin_i] == AutoXMin) &&
			(w[EmbPS].GraphRange[xmax_i] == AutoXMax) ) {
		if ( (AutoXMax - AutoXMin)/0.1 > HashNum )
			HashDist = 0.25;
		else
			HashDist = 0.1;
	}
	else
		HashNum = MIN(HashNum, 10);
	DrawHAxis(TRUE, w[EmbPS].GraphRange[xmin_i],
			w[EmbPS].GraphRange[xmax_i],
			&w[EmbPS].GR, w[EmbPS].Origin, w[EmbPS].xScale,
			HashDist, HashNum, FALSE, HashSize, HashSize-1, PSXFieldWidth,
			2, 3, 3, 0);



	HashDist = 0.0;
	HashNum = 0;
	HashNum = (w[EmbPS].GR.bottom - w[EmbPS].GR.top) /
			(PSXFieldWidth*CharWidth('0')+3) + 1;
	if ( (w[EmbPS].GraphRange[ymin_i] == AutoXMin) &&
			(w[EmbPS].GraphRange[ymax_i] == AutoXMax) ) {
		if ( (AutoXMax - AutoXMin)/0.1 > HashNum )
			HashDist = 0.25;
		else
			HashDist = 0.1;
	}
	else
		HashNum = MIN(HashNum, 10);
	DrawVAxis(TRUE, w[EmbPS].GraphRange[ymin_i],
			w[EmbPS].GraphRange[ymax_i],
			&w[EmbPS].GR, w[EmbPS].Origin, w[EmbPS].yScale,
			HashDist, HashNum, FALSE, HashSize, HashSize-1, PSYFieldWidth,
			2, 3, 3, HashSize+2+TextHeight);

	if (w[EmbPS].GraphRange[xmax_i] <= 0.0) {
		MoveTo(w[EmbPS].GR.left, w[EmbPS].Origin.v);
		Move(-StdMargin - StringWidth("\PXn"), 0);
	}
	else {
		MoveTo(w[EmbPS].GR.right, w[EmbPS].Origin.v);
		Move(StdMargin, 0);
	}
	Move(0, TextHeight/2 - TextHeight%2);
	DrawXn("\Pn");

	if (w[EmbPS].GraphRange[ymax_i] <= 0.0) {
		MoveTo(w[EmbPS].Origin.h, w[EmbPS].GR.bottom);
		Move(0, StdMargin + TextHeight);
	}
	else {
		MoveTo(w[EmbPS].Origin.h, w[EmbPS].GR.top);
		Move(0, -StdMargin-3);
	}
	sprintf(CP(s), "n+%d", Embedding);
	CtoPstr(CP(s));
	Move(-StringWidth("\PXn+1")/2, 0);
	DrawXn(CP(s));

	tmpClipRgn = NewRgn();
	GetClip(tmpClipRgn);
	ClipRect(&w[EmbPS].GR);

	if (EPSParams.Polynomial2n) {
/*
		if (Embedding >= MUCH_EMBEDDING) {
*/
		if (Embedding > MaxFixedEmbedding)
			StartWaiting();
		else
			InitFixedQMVars(lambda, w[EmbPS].xScale, w[EmbPS].yScale,
					&w[EmbPS].GraphRange, w[EmbPS].xGraphAroundZero,
					w[EmbPS].yGraphAroundZero);
		for (x=w[EmbPS].GR.left-w[EmbPS].Origin.h;
				x<w[EmbPS].GR.right-w[EmbPS].Origin.h; ++x) {
			if (Embedding <= MaxFixedEmbedding)
				y = FixedQM(x, Embedding);
			else
				y = yFloatToCoord(EmbPS,
						nQuadMap(Embedding, lambda, xCoordToFloat(EmbPS, x)));
/*
			y = yFloatToCoord(EmbPS,
					nQuadMap(Embedding, lambda, xCoordToFloat(EmbPS, x)));
*/
			if (x==w[EmbPS].GR.left-w[EmbPS].Origin.h)
				BackMoveTo(EmbPS, x, y);
			else
				BackLineTo(EmbPS, x, y);
		}
		if (Embedding > MaxFixedEmbedding)
			StopWaiting();
	}

	if (EPSParams.Line45)
		if ( (tmpXMax > tmpYMin) && (tmpXMin < tmpYMax) ) {
			x = xFloatToCoord(EmbPS, MAX(tmpXMin, tmpYMin));
			y = yFloatToCoord(EmbPS, MAX(tmpXMin, tmpYMin));
			BackMoveTo(EmbPS, x, y);
			x = xFloatToCoord(EmbPS, MIN(tmpXMax, tmpYMax));
			y = yFloatToCoord(EmbPS, MIN(tmpXMax, tmpYMax));
			BackLineTo(EmbPS, x, y);
		}

	if (EPSParams.Box) {
		if ( ((tmpXMax-1.0) * (tmpXMin-1.0) <= 0.0) &&
				(tmpYMin<1.0) && (tmpYMax>0.0) ) {
			x = xFloatToCoord(EmbPS, 1.0);
			y = yFloatToCoord(EmbPS, MAX(0.0, tmpYMin));
			BackMoveTo(EmbPS, x, y);
			y = yFloatToCoord(EmbPS, MIN(1.0, tmpYMax));
			BackLineTo(EmbPS, x, y);
		}
		if ( ((tmpYMax-1.0) * (tmpYMin-1.0) <= 0.0) &&
				(tmpXMin<1.0) && (tmpXMax>0.0) ) {
			y = yFloatToCoord(EmbPS, 1.0);
			x = xFloatToCoord(EmbPS, MAX(0.0, tmpXMin));
			BackMoveTo(EmbPS, x, y);
			x = xFloatToCoord(EmbPS, MIN(1.0, tmpXMax));
			BackLineTo(EmbPS, x, y);
		}
	}

	SetClip(tmpClipRgn);
	DisposeRgn(tmpClipRgn);

	SetPort(w[EmbPS].Peek);
	myCopyBits(w[EmbPS].Back, (GrafPtr) w[EmbPS].Peek,
			&w[EmbPS].Peek->port.portRect, 0,0,0,0, srcCopy, (RgnHandle) NIL);

	OffsetRect(&EPSLatestImpactRect,-EPSLatestImpactRect.left,-EPSLatestImpactRect.top);
	if (myHitNumber >= StartHitNumber) {
		if (myHitNumber - StartHitNumber > MANY_ITERATIONS)
			StartWaiting();
		i=StartHitNumber;
		j=0;
		while(TRUE) {
			if (i<0)
				break;
			if (GetSolutionHit(i) == 0)
				break;
			if (++j >= Embedding)
				break;
			--i;
		}
		i = StartHitNumber + (Embedding-j);
		CurrentEPSOffsetCount = Embedding - (i-myHitNumber);

		if (i>myHitNumber) return;

		for (;i<=myHitNumber-1;++i)
			PlotInEmbeddedPS(GetSolution(i), i, GetSolutionHit(i), FALSE);
		if (myHitNumber - StartHitNumber > MANY_ITERATIONS)
			StopWaiting();
		PlotInEmbeddedPS(GetSolution(i), i, GetSolutionHit(i), TRUE);
	}
}


		/*	-------------------------------------------------------	*/
		/*	------           Inter-Window Routines           ------	*/
		/*	------ (incl. AddGraphicSolution, Bottom Line) ------	*/
		/*	-------------------------------------------------------	*/

SetAutoGraphRange(double theLambda, Boolean AtStartup)
{
	double	oAutoXMin = AutoXMin, oAutoXMax = AutoXMax;

	AutoXMax = MAX(1.0, fabs(theLambda)/4.0);
	AutoXMin = (theLambda < 0) ? -AutoXMax : 0.0;

	if (AtStartup) {
		if (UseAutoXMin)	XMin = AutoXMin;
		if (UseAutoXMax)	XMax = AutoXMax;
		return;
	}

	if ( (AutoXMin != oAutoXMin) ||
				(AutoXMax != oAutoXMax) ) {
		if (UseAutoXMin)	XMin = AutoXMin;
		if (UseAutoXMax)	XMax = AutoXMax;
		SetOrbitCoords();
		InitFullPSRange();
		InitPSScales();
		InitFullEPSRange();
		InitEPSScales();
		NewGraphics(FALSE, FALSE);
	}
}

GetGraphRange()
{
	double	oXMin = XMin, oXMax = XMax;

	if ( DoGraphRangeDialog(&XMin, &XMax, AutoXMin, AutoXMax,
			&UseAutoXMin, &UseAutoXMax) ) {
		if ( (XMin != oXMin) ||
					(XMax != oXMax) ) {
/*
			SetWholeScreenOrbitCoords();
*/
			SetOrbitCoords();
/*
			SetVMax(TRUE);
*/
			InitFullPSRange();
			InitPSScales();
			InitFullEPSRange();
			InitEPSScales();
			NewGraphics(FALSE, FALSE);
		}
	}
}

HandleLeftPSsHS(int windex)
{
	Boolean Magnifying = FALSE;
	Boolean	*UseAuto;
	int		i;

	UseAuto = (Boolean *) ((windex == PSpace) ? &UsesFullPS : &UsesFullEPS);
	for (i=0; i<ItemsInRect; ++i)
		if (!UseAuto[i]) {
			Magnifying = TRUE;
			break;
		}

	if (Magnifying) {
		if (windex == PSpace)
			GetPSMagnification(FALSE);
		else
			GetEPSMagnification(FALSE);
	}
	else
		GetGraphRange();
}

GraphFrom(int hitnum, int curhitnum, double x_n)
{
	if (WindowOpen(PSpace))
		PlotInPhaseSpace(x_n, hitnum, curhitnum, TRUE);
}

Boolean	AddGraphicSolution(int hitnum, int curhitnum, double x_n, int PS)
{
	Boolean	WriteOK;

/*
	switch (PS) {
		case BIFURCATION_DIAGRAM:
					if (!*pBDGraphInAllWindows)
						return(TRUE);
					break;
		case BASINS_OF_ATTRACTION:
					if (!*pBAGraphInAllWindows)
						return(TRUE);
					break;
		default:	break;
	}
*/
	if ( (PS == BIFURCATION_DIAGRAM) && (!*pBDGraphInAllWindows) )
		return(TRUE);

	WriteOK = WriteOutput(curhitnum, x_n, PS);
	if (!PrintingOnly) {
		if (WindowOpen(PSpace))
			PlotInPhaseSpace(x_n, hitnum, curhitnum, TRUE);
		if (WindowOpen(EmbPS))
			PlotInEmbeddedPS(x_n, hitnum, curhitnum, TRUE);
		if (WindowOpen(Orbit))
			PlotInOrbit(hitnum-OrbitOffset, yFloatToCoord(Orbit, x_n));
	}
	if (x_n == NEG_INFINITY)
		if (PS < BIFURCATION_DIAGRAM)
			ERROR("\PSolution unbounded");
	return(WriteOK);
}

NewImpactGraphics(Boolean ClearSolutions)
{
	StartHitNumber = 0;
	SetIterLinesStart(0, FALSE);
/*
	OffsetRect(&NLImpactRect, -NLImpactRect.left, 0);
*/
}

InitImpactGraphics()
{
/*
AutoXMin = 0.0; AutoXMax = 1.0;
UseAutoXMin = UseAutoXMax = TRUE;
XMin = AutoXMin; XMax = AutoXMax;
*/
/*
OrbitDotSize = 1; OrbitSpacing = 0;
NLHeight = 4;
*/
	SetAutoGraphRange(lambda, TRUE);
	StartILHitNumber = 0;
	SetRect(&NLImpactRect,0,0,0,0);
	RefreshingWhileGraphing = FALSE;
	OrbitGraphClipRgn = NewRgn();

/*
	PSParams.NumberLine = PSParams.Parabola = PSParams.Line45 =
			PSParams.IterationLines = PSParams.Arrows =
			PSParams.Box = PSParams.SameSizeAxes = FALSE;
	EPSParams.Polynomial2n = EPSParams.Line45 = EPSParams.Box =
			EPSParams.SameSizeAxes = FALSE;
*/

	SetRect(&PSLatestImpactRect,0,0,5,5);
	SetRect(&EPSLatestImpactRect,0,0,5,5);
	InitQMTextEdit(lambda);
	StartHitNumber = 0;
/*
	UseAutoXMin = UseAutoXMax = TRUE;
	Embedding = DEFAULT_Embedding;
*/
	InitFullPSRange();
	InitFullEPSRange();

/*
	SetWholeScreenOrbitCoords();
*/
	SetOrbitCoords();
	HandleIterLineMenus();
}

HaltImpactGraphics()
{
	HaltQMTextEdit();
}
