/*
 *	qmMenus.c
 *
 *	Handles Quadratic Map's menu requirements
 *
 *	-TA 4/23/91
 */

#include <stdlib.h>
#include <MenuUtils.h>
#include "qmPrototypes.h"
#include "qmMenus.h"
#include "qmGraphics.h"		/*	For special window opening only	*/

#define	DOT					'.'


			/*	External variables from qm.c, mainly for Control menu handling	*/
extern		IterationVars	NormalVars, *pCurrentVars;
extern		hQMPreferences	hQMPrefs;
extern		int				ProgramStatus, IterationStatus;
extern		Boolean			ClearSolutions, PAUSING, OKToContinue;
extern		Boolean			OKToContinueBD;
extern		int				CurrentModifiers;	/* for AdjustQMMenus	*/
extern		Boolean			SingleStepOn;

			/*	External variables from qmSpecial.c, for Special menu handling	*/
extern		Boolean			DrawBD;

			/*	External variables from qm.c for MultiFinder management	*/
extern		Boolean MultiFinderRunning;
extern		long	BackgroundSleep;

			/*	External variables from qmGraphics.c for special window opening	*/
extern		Windows	w;


MenuHandle	AppleMenuH, FileMenuH, EditMenuH;
MenuHandle	ControlMenuH, ParametersMenuH, ViewMenuH, SpecialMenuH;
MenuHandle	SimpleSoundMenuH, AdvancedSoundMenuH, SoundMenuH;
MenuHandle	WindowsMenuH;

Boolean		NeedToSave, NeedToSaveBD;


		/* ----------------------------------------- */
		/* ---------- Menu Initialization ---------- */
		/* ----------------------------------------- */

QuitPrematurely()
		/* due to resource fork problem */
{
		SysBeep(10); SysBeep(10); SysBeep(10);
		exit(1);
}

InitQMMenus(Boolean UsingColor)
{
	if (!StdAddMenu(AppleMenuID,		&AppleMenuH))		QuitPrematurely();
	if (!StdAddMenu(FileMenuID,			&FileMenuH))		QuitPrematurely();
	if (!StdAddMenu(EditMenuID,			&EditMenuH))		QuitPrematurely();
	if (!StdAddMenu(ControlMenuID,		&ControlMenuH))		QuitPrematurely();
/*
	if (!StdAddMenu(ParametersMenuID,	&ParametersMenuH))	QuitPrematurely();
*/
	if (!StdAddMenu(ViewMenuID,			&ViewMenuH))		QuitPrematurely();
	if (!StdAddMenu(SpecialMenuID,		&SpecialMenuH))		QuitPrematurely();
	if (!StdAddMenu(SoundMenuID,	&SoundMenuH))			QuitPrematurely();
/*
	if (!StdAddMenu(SimpleSoundMenuID,	&SimpleSoundMenuH))	QuitPrematurely();
*/
	if (!StdAddMenu(WindowsMenuID,		&WindowsMenuH))		QuitPrematurely();

/*
	AdvancedSoundMenuH = GetMenu(AdvancedSoundMenuID);
	if (AdvancedSoundMenuH == NIL) {
		ERROR("\PUnable to create menus");
		exit(1);
	}
	UseAdvancedSound = FALSE;
	SoundMenuH = SimpleSoundMenuH;
*/
/*
	SuppressSaveWarnings = FALSE;
	AutoCommandKey = FALSE;
*/

	AddResMenu(AppleMenuH, 'DRVR');
		/*	*** Moved to later--windows not created yet ***
	AddWindowsToMenu();
		*/
		/*	Done in ControlMenuStop from StopIterating from _INIT
	DisableItem(ControlMenuH, StopContinueItem);
		 */

	CheckItem(EditMenuH, SuppressSaveItem, SuppressSaveWarnings);
	CheckItem(EditMenuH, AutoCommandKeyItem, AutoCommandKey);
	if (MultiFinderRunning) {
		AppendMenu(EditMenuH, "\P(-");
		AppendMenu(EditMenuH, "\PBackground Operation/G");
/*
		FasterBackground = FALSE;
*/
	}

	CheckItem(ViewMenuH, ScrollOrbitItem, ScrollOrbit);

	if (!UsingColor) {
		DisableItem(SpecialMenuH, UseColorBDItem);
		DisableItem(SpecialMenuH, ColorsItem);
	}

/*
	CheckItem(SimpleSoundMenuH, XnTonesItem, TRUE);
	CheckItem(AdvancedSoundMenuH, XnTonesItem, TRUE);
	CheckItem(AdvancedSoundMenuH, ContinuousMode, TRUE);
*/
/*
	CheckItem(SoundMenuH, XnTonesItem, TRUE);
	CheckItem(SoundMenuH, ContinuousMode, TRUE);
*/
	CheckItem(SoundMenuH, XnTonesItem, XnTonesOn);
	CheckItem(SoundMenuH, XnTonesItem+1, XnkTonesOn);
	CheckItem(SoundMenuH, MajorMode, CurrentMode == MajorMode);
	CheckItem(SoundMenuH, MinorMode, CurrentMode == MinorMode);
	CheckItem(SoundMenuH, ChromaticMode, CurrentMode == ChromaticMode);
	CheckItem(SoundMenuH, ContinuousMode, CurrentMode == ContinuousMode);

/*
	DisableItem(ViewMenuH, UseFullPhaseSpaceItem);
	DisableItem(ViewMenuH, UseFullEmbeddedPSItem);
*/

	DrawMenuBar();
	HiliteMenu(0);

	NeedToSave = NeedToSaveBD = FALSE;
}

		/* --------------------------------------------------------- */
		/* ---------- Specific Menu-Manipulation Routines ---------- */
		/* --------------------------------------------------------- */

AdjustQMMenus()
{
	WindowPeek	wp = (WindowPeek) FrontWindow();
	Boolean		DA = IS_SYS_WINDOW(wp);
	
	ToggleItemEnabled(EditMenuH, UndoItem, DA);
	ToggleItemEnabled(EditMenuH, CutItem, DA);
/*	Don't adjust Copy
	ToggleItemEnabled(EditMenuH, CopyItem, DA);
*/
	ToggleItemEnabled(EditMenuH, PasteItem, DA);
	ToggleItemEnabled(EditMenuH, ClearItem, DA);

	if (OPTION_DOWN)
		SetItem(WindowsMenuH, CloseWindowItem, "\PClose All Windows");
	else
		SetItem(WindowsMenuH, CloseWindowItem, "\PClose Window");

	ToggleItemEnabled(WindowsMenuH, CloseWindowItem, wp != NIL);
	while (TRUE) {
		if ( (wp == NIL) || (!wp->visible) ) {
			ToggleItemEnabled(WindowsMenuH, ArrangeWindowsItem, FALSE);
			break;
		}
		if (WindowIndex((WindowPtr) wp) != NoWindow) {
			ToggleItemEnabled(WindowsMenuH, ArrangeWindowsItem, TRUE);
			break;
		}
		wp = wp->nextWindow;
	}

/*
	ToggleItemEnabled(ViewMenuH, UseFullPhaseSpaceItem, IsMagnified(PhaseSpace));
	ToggleItemEnabled(ViewMenuH, UseFullEmbeddedPSItem, IsMagnified(EmbeddedPS));
*/

/*
	ToggleItemEnabled(SpecialMenuH, StartBifDiagItem, DrawBD);
	ToggleItemEnabled(SpecialMenuH, PauseBifDiagItem, DrawBD);
	ToggleItemEnabled(SpecialMenuH, ClearBifDiagItem, DrawBD);
*/
}

ControlMenuStart(int PS)
{
	switch (PS) {
		case BIFURCATION_DIAGRAM:
				SetItem(SpecialMenuH, StopContinueBifDiagItem, "\PStop Bifurcation Diagram");
				EnableItem(SpecialMenuH, StopContinueBifDiagItem);
				break;
		default:
				SetItem(ControlMenuH, StopContinueItem, "\PStop Iteration");
				EnableItem(ControlMenuH, StopContinueItem);
				DisableItem(ControlMenuH, DoOneIterationItem);
				break;
	}
}

ControlMenuStop(int PS, Boolean OKToCont)
{
	switch (PS) {
		case BIFURCATION_DIAGRAM:
				SetItem(SpecialMenuH, StopContinueBifDiagItem, (OKToCont) ?
						"\PContinue Bifurcation Diagram" : "\PStop Bifurcation Diagram");
				ToggleItemEnabled(SpecialMenuH, StopContinueBifDiagItem, OKToCont);
				break;
		default:
				SetItem(ControlMenuH, StopContinueItem, (OKToCont) ?
						"\PContinue Iteration" : "\PStop Iteration");
				ToggleItemEnabled(ControlMenuH, StopContinueItem, OKToCont);
/*
				ToggleItemEnabled(ControlMenuH, DoOneIterationItem, OKToCont);
*/
				ToggleItemEnabled(ControlMenuH, DoOneIterationItem, TRUE);
				break;
	}
}

		/* ----------------------------------- */
		/* ---------- Menu Handling ---------- */
		/* ----------------------------------- */

HandleAppleMenu(int menuItem)
{
	switch (menuItem) {
		case 1:
			DoAboutQMDialog();
			break;
		default:
			HandleDeskAcc(AppleMenuH, 0, menuItem);
			break;
	}	/* switch menuItem from case AppleMenuID */
}

Boolean DoSave()
{
	Boolean		GoodOp, HadToSaveAs;
	int			theWindex;

	theWindex = DecideWhichSaveWindow();
			/* GoodOp used differently here from elsewhere*/
	GoodOp = TRUE;
	if (!SaveFileExists()) {
		GoodOp = SaveAs();
		HandleAllUpdates();
		HadToSaveAs = TRUE;
	}
	else HadToSaveAs = FALSE;
	if (GoodOp) {
		if (SaveFile()) {
			if (HadToSaveAs && (theWindex == TxOut))
				if (!SpecialGraphing())
					SetTextTitle(TRUE, TxOut);
				else
					SetANilFileName(TxOut);
			switch (theWindex) {
				case TxOut:	NeedToSave = FALSE;		break;
				case BifD:	NeedToSaveBD = FALSE;	break;
			}
		}
		else {
			if (HadToSaveAs)
				SetANilFileName(theWindex);
			GoodOp = FALSE;
		}
	}
	return(GoodOp);
}

Boolean CheckOneSave(Boolean Quitting, int windex)
{
	Boolean		WantsToSave = FALSE, NTS;
	Boolean		GoodOp;
	WindowPtr	wp;
	int			FrontWindex;

	switch (windex) {
		case BifD:	NTS = NeedToSaveBD;	break;
		default:	NTS = NeedToSave;
					windex = TxOut;
					break;
	}

	if (NTS) {
		if (Quitting) {
			if (!DoQuitWarningDialog(&WantsToSave, windex))
				return(FALSE);
		}
		else if (!DoSaveWarningDialog(&WantsToSave, windex))
			return(FALSE);
		if (WantsToSave) {
			wp = FrontWindow();
			SelectWindowIndex(windex);
			GoodOp = DoSave();
			if (wp != NIL) {
				FrontWindex = WindowIndex(wp);
				if (FrontWindex != NoWindow)
					SelectWindowIndex(FrontWindex);
				else
					SelectWindow(wp);
			}
			return(GoodOp);
		}
	}

	return(TRUE);
}

Boolean CheckNormalSaves(Boolean Quitting)
{
	if (!CheckOneSave(Quitting, TxOut))	return(FALSE);

	return(TRUE);
}

Boolean CheckAllSaves(Boolean Quitting)
{
	if (!CheckOneSave(Quitting, TxOut))	return(FALSE);
	if (!CheckOneSave(Quitting, BifD))	return(FALSE);

	return(TRUE);
}

HandleFileMenu(int menuItem)
{
	int			theWindex;

	switch (menuItem) {
		case NewItem:
						if (!CheckAllSaves(FALSE))	return;
						ResetSolutions();
						ProgramStatus = BIFURCATION_DIAGRAM;	StopIterating(FALSE);
						ProgramStatus = NORMAL_ITERATION;		StopIterating(FALSE);
						NewGraphics(TRUE, TRUE);
						NewSpecial(BifD);
						NeedToSaveBD = FALSE;
/* put in NewGraphics(TRUE)
						SetTextTitle(FALSE, TxOut);
						NeedToSave = FALSE;
*/
/*
						NeedToSaveBD = NeedToSaveBA = FALSE;
*/
						break;
		case OpenItem:
						if (GetOpenFileName(&theWindex)) {
							if (!CheckOneSave(FALSE, theWindex))
								return;
/*	1.01 addition
 */
							SetOpenFileName(theWindex);
/*	end, 1.01 addition
 */
							if (theWindex == BifD) {
								if (OpenFile())
									OpenSpecialFile(theWindex);
							}
							else
								if (StartOpeningFile())
									IterationStatus = ITERATING;
						}
						break;
		case SaveItem:
						if (!DoSave()) ;
						break;
		case SaveAsItem:
						theWindex = DecideWhichSaveWindow();
						if ( !SaveAs() )		return;
						HandleAllUpdates();
						if ( !SaveFile() )
							ResetFileName();
						else switch (theWindex) {
							case TxOut:	if (!SpecialGraphing())
											SetTextTitle(TRUE, TxOut);
										else
											SetANilFileName(TxOut);
										NeedToSave = FALSE;		break;
							case BifD:	NeedToSaveBD = FALSE;	break;
						}
						break;
		case SaveDefaultsItem:
						SaveDefaults();
						break;
		case PageSetupItem:
						DoPageSetup();
						break;
		case PrintItem:
						PrintFrontWindow();
						break;
		case QuitItem:
						if (CheckAllSaves(TRUE))
							_HALT();
						break;
	}
}

HandleEditMenu(int menuItem)
{
	WindowPeek	wp = (WindowPeek) FrontWindow();
	int			windex;

	if (IS_SYS_WINDOW(wp) && (menuItem<=ClearItem))
		HandleSysEdit(menuItem);
	else {
		windex = WindowIndex((WindowPtr) wp);
		switch (menuItem) {
			case CopyItem:	HandleCopy(windex);		break;
			case SuppressSaveItem:
					SuppressSaveWarnings = !SuppressSaveWarnings;
					CheckItem(EditMenuH, menuItem, SuppressSaveWarnings);
					break;
			case AutoCommandKeyItem:
					AutoCommandKey = !AutoCommandKey;
					CheckItem(EditMenuH, menuItem, AutoCommandKey);
					break;
			case BackgroundOperationItem:
					if (DoBackgroundDialog(&RunInBackground,
							&FasterBackground, &UseSoundInBackground))
						BackgroundSleep =
								(FasterBackground) ? DAY_DREAM : CATNAP;
		}
	}
}

HandleControlMenu(int menuItem)
{
	Boolean	GoodValues;

	switch (menuItem) {
		case BeginIterationItem:
/*
			if ( (ProgramStatus == WAITING) ||
					(ProgramStatus >= BIFURCATION_DIAGRAM) ) {
*/
				pCurrentVars = &NormalVars;
				GoodValues = Initialize_Iteration_Globals(TRUE);
				NewGraphics(ClearSolutions, TRUE);
				if (ClearSolutions)
					ResetSolutions();
				if (!GoodValues) return;
				StartIterating(NORMAL_ITERATION);
				NeedToSave = TRUE;
/*
			}
			else {
				PAUSING = TRUE;
				ActivateTextScrolls(PAUSING);
/*
				if (ProgramStatus == PLAYBACK) {
					/* SetHitNumber(RealHitNumber) *);
					NeedToSave = TRUE;
				}
				StopIterating(HitNumber>=0); return;
*)
			}
*/
			break;
		case StopContinueItem:
/*
			if (PAUSING && (ProgramStatus < BIFURCATION_DIAGRAM)) {
				PAUSING = !PAUSING;
				ActivateTextScrolls(PAUSING);
				break;
			}
*/
			if (OKToContinue) {
				if (!Reset_Iteration_Globals()) {
					StopIterating(FALSE);
					return;
				}
				if (StartIterating(NORMAL_ITERATION)) {
					IterationStatus = ITERATING;
					NeedToSave = TRUE;
				}
			}
			else {
				PAUSING = !PAUSING;
				ActivateTextScrolls(PAUSING);
				if (PAUSING) {
					SetItem(ControlMenuH, menuItem, "\PContinue Iteration");
					EnableItem(ControlMenuH, DoOneIterationItem);
				}
				else {
					SetItem(ControlMenuH, menuItem, "\PStop Iteration");
					DisableItem(ControlMenuH, DoOneIterationItem);
				}
			}
			break;
		case DoOneIterationItem:
			SingleStepOn = TRUE;
			if ( (ProgramStatus != NORMAL_ITERATION) && (!OKToContinue) )
				HandleControlMenu(BeginIterationItem);
			else
				HandleControlMenu(StopContinueItem);
			break;
		case PlaybackItem:
			if (InitPlayback())
				IterationStatus = ITERATING;
			break;
		case InitialConditionsItem:
			if (GetInitialConditions(TRUE))
				BeginIfAppropriate();
			break;
		case SystemParametersItem:
			if (GetSystemParameters())
				BeginIfAppropriate();
			break;
		case GetInitFromLastItem:
			GetInitialFromLast();
			if (GetInitialConditions(FALSE))
				BeginIfAppropriate();
			break;
		case MaxIterationsItem:
			GetMaxIterations();
			break;
/*
		case RunInBackgroundItem:
			RunInBackground = !RunInBackground;
			CheckItem(ControlMenuH, RunInBackgroundItem, RunInBackground);
			break;
*/
	}
}

HandleViewMenu(int menuItem)
{
	switch (menuItem) {
		case GraphRangeItem:
								GetGraphRange();
								break;
		case IterationSpeedItem:
								GetIterationSpeed();
								break;
		case PSViewItem:
								GetPhaseSpaceInfo();
								break;
		case EPSViewItem:
								GetEmbeddedPSInfo();
								break;
		case OrbitViewItem:
								GetOrbitInfo();
								break;
		case ScrollOrbitItem:
								SwitchScrollOrbit();
								break;
		case ClearIterationsItem:
								SetPhaseSpacesStart(HitNumber+1);
								break;
		case ShowAllIterationsItem:
								SetPhaseSpacesStart(0);
								break;
		case ClearIterLinesItem:
								SetIterLinesStart(HitNumber+1, TRUE);
								break;
		case ShowAllIterLinesItem:
								SetIterLinesStart(0, TRUE);
								break;
		case MagnifyPhaseSpaceItem:
								GetPSMagnification(FALSE);
								break;
/*
		case UseFullPhaseSpaceItem:
								SetFullPSRange();
								break;
*/
		case MagnifyEmbeddedPSItem:
								GetEPSMagnification(FALSE);
								break;
/*
		case UseFullEmbeddedPSItem:
								SetFullEPSRange();
								break;
*/
	}
}

HandleSpecialMenu(int menuItem)
{
	switch (menuItem) {
/*
		case StartBifDiagItem:
						if ( (ProgramStatus != BIFURCATION_DIAGRAM) ) {
							if (!CheckOneSave(FALSE, BifD))
								return;
							StartBifurcationDiagram();
						}
						else
							StopIterating(TRUE);
						break;
		case StopContinueBifDiagItem:
						if (OKToContinueBD)
							ContinueBifurcationDiagram();
						else {
							PAUSING = !PAUSING;
							ActivateTextScrolls(PAUSING);
							if (PAUSING)
								SetItem(SpecialMenuH, menuItem, "\PContinue Bifurcation Diagram");
							else SetItem(SpecialMenuH, menuItem, "\PPause Bifurcation Diagram");
						}
						break;
*/
		case StartBifDiagItem:
						if (!CheckOneSave(FALSE, BifD))
							return;
						StartBifurcationDiagram();
						break;
		case StopContinueBifDiagItem:
						if (OKToContinueBD)
							ContinueBifurcationDiagram();
						else {
							PAUSING = !PAUSING;
							ActivateTextScrolls(PAUSING);
							if (PAUSING)
								SetItem(SpecialMenuH, menuItem, "\PContinue Bifurcation Diagram");
							else {
								if (!WindowOpen(BifD))
									if (DoShowWindowDialog(BifD))
										QMOpenWindow((WindowPtr) w[BifD].Peek);
								SetItem(SpecialMenuH, menuItem, "\PStop Bifurcation Diagram");
							}
						}
						break;
		case BifurcationDiagramItem:
						GetBifDiagInfo(FALSE);
						break;
		case ClearBifDiagItem:
/*
						if (DoSpecialClearDialog("\Pbifurcation")) {
*/
						if (CheckOneSave(FALSE, BifD)) {
							ClearBifurcationDiagram();
							if (ProgramStatus == BIFURCATION_DIAGRAM)
								StopIterating(FALSE);
						}
						break;
		case GraphInAllWindowsBDItem:
						SwitchGraphInAll(BifDiagram);
						break;
		case UseColorBDItem:
						SwitchInColor(BifDiagram);
						break;
		case ColorsItem:
						GetSolutionColors();
						break;
	}
}

SwitchSoundMenu()
{
/*
	UseAdvancedSound = !UseAdvancedSound;
HiliteMenu(0);
	if (UseAdvancedSound) {
		DeleteMenu(SimpleSoundMenuID);
		SoundMenuH = AdvancedSoundMenuH;
		InsertMenu(AdvancedSoundMenuH, WindowsMenuID);
	}
	else {
		DeleteMenu(AdvancedSoundMenuID);
		SoundMenuH = SimpleSoundMenuH;
		InsertMenu(SimpleSoundMenuH, WindowsMenuID);
	}
/*
	DrawMenuBar();
*)
/*	HiliteMenu(0) here only did not work
	HiliteMenu(0);
*)
/*
	DrawMenuBar();
*)
*/
}

HandleSoundMenu(int menuItem)
{
/*
	if ( (menuItem == AdvancedOptionsItem) && (!UseAdvancedSound) ) {
		SwitchSoundMenu();
		return;
	}
*/
	switch (menuItem) {
		case XnTonesItem:
		case XnkTonesItem:
			FlipTonesOn(menuItem);
			break;
		case MajorMode:
		case MinorMode:
		case ChromaticMode:
		case ContinuousMode:
			SetToneMode(menuItem);
			break;
/*
		case SimpleOptionsItem:
			SwitchSoundMenu();
			break;
*/
	}
}

HandleWindowsMenu(int menuItem)
{
	WindowPeek	wp;
	int			kind;
	
	switch (menuItem) {
/*
		case EnlargeWindowItem:
				if (WindowIndex(FrontWindow()) != NoWindow)
					MakeLargeWindow(FrontWindow());
				break;
*/
		case CloseWindowItem:
				if (OPTION_DOWN)	QMCloseAllWindows();
				else				if (QMCloseFrontWindow()) ;
				break;
		case ArrangeWindowsItem:
				ArrangeWindows();
				break;
		case DefaultConfigItem:
				DefaultWindowConfiguration();
				break;
		default:
				SelectWindowIndex(menuItem-WindowSeparatorItem-1);
				break;
	}
}

_MENU(int menuID, int menuItem)
{
	switch (menuID) {
		case AppleMenuID:		HandleAppleMenu(menuItem);			break;
		case FileMenuID:		HandleFileMenu(menuItem);			break;
		case EditMenuID:		HandleEditMenu(menuItem);			break;
		case ControlMenuID:		HandleControlMenu(menuItem);		break;
/*
		case ParametersMenuID:	HandleParametersMenu(menuItem);		break;
*/
		case ViewMenuID:		HandleViewMenu(menuItem);			break;
		case SpecialMenuID:		HandleSpecialMenu(menuItem);		break;
/*
		case SimpleSoundMenuID:
		case AdvancedSoundMenuID:
*/
		case SoundMenuID:
								HandleSoundMenu(menuItem);			break;
		case WindowsMenuID:		HandleWindowsMenu(menuItem);		break;
	}
}

HandleMenu(long theMenu)
{
	if (HiWord(theMenu) != 0)
		_MENU(HiWord(theMenu), LoWord(theMenu));
}

HandleKeyDown(long EventMessage, int Modifiers)
{
	char	c;
	long	MenuSelection;

	if (!AutoCommandKey) {
		if ( !(c = IsCommandKey(EventMessage, Modifiers)) )	return;
	}
	else
		c = LoWord(EventMessage & charCodeMask);

	if (ProgramStatus == OPENING) {
		if (c == DOT) {
			StopOpeningFile(FALSE);
			StopIterating(HitNumber>=0);
		}
		return;
	}
	MenuSelection = MenuKey(c);
	if (HiWord(MenuSelection) == 0) {
		switch (c) {
			case DOT:
				if (PAUSING)	break;
				switch (ProgramStatus) {
					case WAITING:
						break;
					case BIFURCATION_DIAGRAM:
						MenuSelection = ( ((long) SpecialMenuID) << 16 ) +
								StopContinueBifDiagItem;
						break;
					default:
						MenuSelection = ( ((long) ControlMenuID) << 16 ) +
								StopContinueItem;
						break;
				}
				break;
/*
			case 'g':
				if (ProgramStatus != WAITING) break;
				MenuSelection = ( ((long) ControlMenuID) << 16 ) + BeginIterationItem;
				break;
*/
		}
	}

	HandleMenu(MenuSelection);
}

/*
_KEY(int c, int mods)
{
	if (!(mods & cmdKey)) return;
	HandleMenu(MenuKey(c));
}
*/
