/*	qmSpecial.c
 *
 *	Routines for Bifurcation Diagram, Basin of Attraction
 *
 *		-TA 4/22/91
*/

#include <OffPort.h>
#include <MenuUtils.h>
#include "qmPrototypes.h"
#include "qmMenus.h"
#include "qmGraphics.h"
#include "qmSpecial.h"

				/*	From qm.c	*/
extern	IterationVars	NormalVars, *pCurrentVars;
extern	hQMPreferences	hQMPrefs;

extern	int				ProgramStatus;
extern	Boolean			PAUSING;
extern	Boolean			UsingColorMac;

				/*	From qmGraphics.c	*/
extern	Windows			w;

				/*	From qmMenus.c	*/
extern	MenuHandle		SpecialMenuH, WindowsMenuH;

				/*	From qmBifDiag.c	*/
extern	BifDiagRecord		BD_Rec;
extern	Boolean				DrawBD;
extern	int					**BD_Points;
extern	Boolean				NeedToColorAfter;

				/*	Shared with qmIteration.c, for OKToGraph(),
				 *	and qm.c, for StartIterating()	*/
Boolean				*pBDGraphInAllWindows, *pBDUseColor;
				/*	Shared with qmIteration.c, for InitNextIteration()	*/
Boolean				NeedNextSpecialValue;
				/*	Shared with qmBifDiag.c	*/
Boolean				NeedToTurnOnGraphing;

RGBColor			SolutionColors[NumSolutionTypes];
long				ColorLongs[NumSolutionTypes];


			/*	--------------------------------------------------	*/
			/*	----------	Special Utility Routines	----------	*/
			/*	--------------------------------------------------	*/

SeeIfStopSpecial(int windex)
{
	if (windex != BifDiagram)					return;

	if (ProgramStatus != BIFURCATION_DIAGRAM)	return;
	if (PAUSING)								return;

	if (DoSpecialCloseStopDialog("\Pbifurcation"))
		HandleSpecialMenu(StopContinueBifDiagItem);
}

Boolean SpecialGraphing()
/*	Returns TRUE only if special program status and graphing in all windows	*/
{
	switch (ProgramStatus) {
		case BIFURCATION_DIAGRAM:	return(*pBDGraphInAllWindows);
		default:					return(FALSE);
	}
}

int	GetSpecialTimes(int theTimes, int DefaultTimes)
{
	return ((theTimes != AutoTimes) ? theTimes : DefaultTimes);
}

SpecialMenuEnable(int windex, Boolean TurnOn, Boolean PauseOK)
{
	int		MenuOffset;

	MenuOffset = 0;
	ToggleItemEnabled(SpecialMenuH, MenuOffset + StartBifDiagItem, TurnOn);
	ToggleItemEnabled(SpecialMenuH, MenuOffset + ClearBifDiagItem, TurnOn);
	ToggleItemEnabled(SpecialMenuH, MenuOffset + StopContinueBifDiagItem,
			TurnOn && PauseOK);
}

CheckGraphInAll(int windex)
{
	if (windex == BifD)
		CheckItem(SpecialMenuH, GraphInAllWindowsBDItem,
			BD_Rec.GraphInAllWindows);
}

SwitchGraphInAll(int windex)
{
	Boolean	*pGIAW;
	int		AppropriatePS;

	switch (windex) {
		case BifD:	pGIAW = &BD_Rec.GraphInAllWindows;
					AppropriatePS = BIFURCATION_DIAGRAM;
					break;
	}

	if (*pGIAW) {
		*pGIAW = FALSE;
		CheckGraphInAll(windex);
		if (ProgramStatus == AppropriatePS) {
			SetHeaders(NormalVars._lambda);
			NewGraphics(TRUE, FALSE);
			ActivateTextScrolls(TRUE);
			SetTextTitle(FALSE, TxOut);
		}
	}
	else {
		if (ProgramStatus == AppropriatePS) {
			DoGraphInAllAlert(windex);
			NeedToTurnOnGraphing = TRUE;
		}
		else {
			*pGIAW = TRUE;
			CheckGraphInAll(windex);
		}
	}
}

CheckSpecialColor(int windex)
{
	if (windex == BifD)
		CheckItem(SpecialMenuH, UseColorBDItem, BD_Rec.InColor);
}

SwitchInColor(int windex)
{
	Boolean	*pInColor;
	int		AppropriatePS;

	switch (windex) {
		case BifD:	pInColor = &BD_Rec.InColor;
					AppropriatePS = BIFURCATION_DIAGRAM;
					break;
	}

	*pInColor = !(*pInColor);
	CheckSpecialColor(windex);
	StdForceUpdate((WindowPtr) w[windex].Peek);
	if (*pInColor) {
		if ((windex==BifD) && (ProgramStatus == AppropriatePS))
			NeedToColorAfter = TRUE;
	}
}

SetSpecialTitle(Boolean FileExists, int windex)
{
	char	*pfname;

	if (FileExists)
		pfname = GetSaveFName(windex);
	else {
		switch (windex) {
			case BifD:	pfname = "\PBifurcation Diagram";
						break;
		}
		SetANilFileName(windex);
	}

	SetItem(WindowsMenuH, WindowSeparatorItem + windex + 1, pfname);
	SetWTitle(w[windex].Peek, pfname);
}

StringAddVar(char *s, double f, char *label, Boolean MorePrecision,
	Boolean ToPad)
{
	Str255	fstring;
	int		fw = SpecialFieldWidth + ((MorePrecision) ? 1 : 0);

	sAppend(s, label);
	if (ToPad)
		PadFtoStr(f, CP(fstring), fw, (fw-1), TRUE);	/*	TRUE in BB	*/
	else
		myFToStrNoPad(f, CP(fstring), fw, (fw-1));
	sAppend(s, CP(fstring));
}

StringAddVarLn(char *s, double f, char *label, Boolean MorePrecision,
		Boolean ToPad)
{
	Str255	fstring;

	StringAddVar(s, f, label, MorePrecision, ToPad);
	sInsertC(s, ReturnChar, label[0]+1);
}

Boolean CreateSpecialOffPort(GrafPtr *pgp, Rect *pr)
{
	Rect	tr;

	tr = *pr;
	OffsetRect(&tr, -tr.left, -tr.top);
	if (*pgp != NOPORT)
		DisposeOffPort(*pgp);
	*pgp = NewOffPort(&tr);
	if (*pgp == NIL) {
		*pgp = NOPORT;
		return(FALSE);
	}
	SetPort(*pgp);
	SetOrigin(pr->left, pr->top);
	ClipRect(&((*pgp)->portRect));
	EraseRect(&((*pgp)->portRect));

	return(TRUE);
}

#define	ProbableTolerance		1e-2
#define	AlternatingTolerance	1e-3
#define	MonotonicTolerance		1e-5
#define	Tolerance				1e-10

Boolean	TestAdjacents(double y0, double y1, int per, int k)
{
	int		AdjacentK;
	double	AdjacentY0, AdjacentY1;

	AdjacentK = k+per/2;
	AdjacentY0 = GetSolution(HitNumber - 0*per - AdjacentK);
	AdjacentY1 = GetSolution(HitNumber - 1*per - AdjacentK);
	if ( (MAX(y1, AdjacentY1) <= MAX(y0, AdjacentY0)) &&
			(MIN(y1, AdjacentY1) >= MIN(y0, AdjacentY0)) )
		return(TRUE);
	else
		return(FALSE);
}

Boolean TestPeriodicity(int per, Boolean *pWillBePeriodic,
		Boolean *pProbablyPeriodic)
	/*	y0-3 are the last 4 x's, 0 most recent.
	 *	Periodic if:
	 *	1.	a.	y0 between y1 & y2, & y1 between y2 & y3; and
	 *		b.	diff, y0-y2, < diff, y1-y3; and
	 *		c.	diff, y0-y1, < AlternatingTolerance; and
	 *		d.	diff, y1-y3, < AlternatingTolerance; or
	 *	2.	a.	series y0-y3 monotonic; and
	 *		b.	successive differences shrinking;
	 *		c.	diff, y0-y1, < MonotonicTolerance (if not,
	 *			sets *pWillBePeriodic), subject to (3b); or
	 *	3.	a.	diff, y0-y1, y1-y2, y2-y3, very small; and
	 *		b.	if even periodicity, adjacent branches do not
	 *			approach one another.
	 *	NOTE: For test 3b, I feel like there should be an additional
	 *	test, i.e., ok to approach one another as long as not converging
	 *	to one another (in practice, converging much more slowly than
	 *	the points in each branch).
	 *												-TA 3/1-2/91, 5/20/91
	 */
{
	int		k;
	double	y0, y1, y2, y3;
	double	d0, d1, d2;
	Boolean	IsPeriodic;
	Boolean	tmpWillBePeriodic, MaybeWillBePeriodic = TRUE;
	Boolean	tmpProbablyPeriodic, MaybeProbablyPeriodic = TRUE;
	Boolean	ThisOnePeriodic = TRUE;
	Boolean	NeedWillBe = !*pWillBePeriodic;
	Boolean	NeedProbable = !*pProbablyPeriodic;

	for (k=0; k<per; ++k) {
		IsPeriodic = FALSE;
		tmpWillBePeriodic = FALSE;
		tmpProbablyPeriodic = FALSE;
		y0 = GetSolution(HitNumber - 0*per - k);
		y1 = GetSolution(HitNumber - 1*per - k);
		y2 = GetSolution(HitNumber - 2*per - k);
		y3 = GetSolution(HitNumber - 3*per - k);
		d0 = y0-y1;
		d1 = y1-y2;
		d2 = y2-y3;

			/*	Alternating Convergence	*/
		if (BETWEEN(y0, y1, y2))
			if (BETWEEN(y1, y2, y3))
				if ( fabs(y3-y1) > fabs(y2-y0) )
					if (fabs(d2)/1 < AlternatingTolerance) {
						if (fabs(y3-y1)/1 < AlternatingTolerance)
							IsPeriodic = TRUE;
					}
					else tmpProbablyPeriodic = TRUE;

			/*	Monotonic Convergence	*/
		if (!IsPeriodic)
			if ( (d0*d1>=0) && (d1*d2>=0) )
				if ( (fabs(d0) < fabs(d1)) &&
						(fabs(d1) < fabs(d2)) ) {
/*
					if ((per%2 == 0) && (k>=per/2) ) {
*/
					if ((per%2 == 0) && (k<per/2) ) {
						if (TestAdjacents(y0, y1, per, k)) {
							tmpWillBePeriodic = TRUE;
							IsPeriodic = (fabs(d2)/1 < MonotonicTolerance);
						}
					}
					else {
						tmpWillBePeriodic = TRUE;
						IsPeriodic = (fabs(d2)/1 < MonotonicTolerance);
					}
				}
				else if ( fabs(d0-d1) < fabs(d2-d1) ) {	/*	2nd diff.	*/
					if ((per%2 == 0) && (k<per/2) ) {
						if (TestAdjacents(y0, y1, per, k)) {
							tmpWillBePeriodic = TRUE;
						}
					}
					else {
						tmpWillBePeriodic = TRUE;
					}
				}

			/*	Any Old Convergence	*/
		if (!IsPeriodic)
			if ( (fabs(d0/1) < Tolerance) && (fabs(d1/1) < Tolerance) &&
					(fabs(d2/1) < Tolerance) ) {	/*asymptotic--needed?*/
				if ((per%2 == 0) && (k<per/2) ) {
					if (TestAdjacents(y0, y1, per, k))
						IsPeriodic = TRUE;
				}
				else
					IsPeriodic = TRUE;
			}

			/*	Probably Periodic?	*/
		if ( (!IsPeriodic) && (!tmpWillBePeriodic) &&
				(!tmpProbablyPeriodic) )
			if ( (fabs(d0/1) < ProbableTolerance) &&
					(fabs(d1/1) < ProbableTolerance) &&
					(fabs(d2/1) < ProbableTolerance) )
				tmpProbablyPeriodic = TRUE;

		MaybeProbablyPeriodic = ( MaybeProbablyPeriodic &&
				(tmpProbablyPeriodic || tmpWillBePeriodic || IsPeriodic) );
		MaybeWillBePeriodic = ( MaybeWillBePeriodic &&
				(tmpWillBePeriodic || IsPeriodic) );
		ThisOnePeriodic = (ThisOnePeriodic && IsPeriodic);
		if (!ThisOnePeriodic && (!MaybeWillBePeriodic || !NeedWillBe) &&
				(!MaybeProbablyPeriodic || !NeedProbable)) {
			*pWillBePeriodic = MaybeWillBePeriodic;
			*pProbablyPeriodic = MaybeProbablyPeriodic;
			return(FALSE);
		}
	}

	*pProbablyPeriodic = MaybeProbablyPeriodic;
	*pWillBePeriodic = MaybeWillBePeriodic;
	return(ThisOnePeriodic);
/*
	return(TRUE);
*/
}


	/*	File operations	*/

PicHandle MakeSpecialPicture(GrafPtr gp)
{
	PicHandle	pic;
	Rect		DstRect = gp->portRect;
	RgnHandle	OldClipRgn;

	SetPort(gp);
	OldClipRgn = NewRgn();
	GetClip(OldClipRgn);

	pic = OpenPicture(&DstRect);
		ClipRect(&gp->portRect);
		CopyBits(&gp->portBits, &gp->portBits, &DstRect, &DstRect, 0L, srcCopy);
	ClosePicture();
	SetClip(OldClipRgn);
	DisposeRgn(OldClipRgn);

	return(pic);
}

Boolean SaveSpecialFile(int theWindex)
{
	Boolean	GoodOp;

	switch (theWindex) {
		case BifD:	GoodOp = SaveBDFile();	break;
	}
	SetSpecialTitle(GoodOp, theWindex);
	return(GoodOp);
}

Boolean OpenSpecialFile(int theWindex)
{
	Boolean	GoodOp;

	switch (theWindex) {
		case BifD:	GoodOp = OpenBDFile();	break;
	}
	CloseOpenFile();
/*
	SetSpecialTitle(GoodOp, theWindex);
*/
	return(GoodOp);
}


			/*	--------------------------------------------------	*/
			/*	----------	General-Purpose Routines	----------	*/
			/*	--------------------------------------------------	*/

		/*	----------	Solution Patterns, Colors Routines	----------	*/

SetSpecialColor(int i, Boolean UseRGBValues)
{
	GrafPtr	gp;

	if (UseRGBValues)
		RGBForeColor(&SolutionColors[i]);
	else {
		GetPort(&gp);
		gp->fgColor = ColorLongs[i];
	}
}

InitSpecialColors()
{
	int 		i=0;
	GrafPtr		curPort;
	WindowPtr	wp;
	Rect		r = {50, 50, 60, 60};

	GetPort(&curPort);
	wp = NewWindow(0L, &r, "\p", false, noGrowDocProc, -1L, true, 0L);

	ColorLongs[i++] = blackColor;	ColorLongs[i++] = magentaColor;
	ColorLongs[i++] = redColor;		ColorLongs[i++] = yellowColor;
	ColorLongs[i++] = greenColor;	ColorLongs[i++] = cyanColor;
	ColorLongs[i++] = blueColor;	ColorLongs[i++] = whiteColor;

	SetPort(wp);
	for(i=0; i<NumSolutionTypes; ++i) {
		wp->fgColor = ColorLongs[i];
		GetForeColor(&SolutionColors[i]);
	}

	SetPort(curPort);
	DisposeWindow(wp);
}

GetSolutionColors()
{
	SolutionInfo	tsi;
	int				i;

	for(i=0; i<NumSolutionTypes; ++i)
		tsi[i] = csi[i];
	if (DoSolutionInfoDialog()) {
		for(i=0; i<NumSolutionTypes; ++i)
			if(tsi[i] != csi[i])
				break;
		if (i<NumSolutionTypes) {
			if (DrawBD && BD_Rec.InColor)
				StdForceUpdate((WindowPtr) w[BifD].Peek);
		}
	}
	else
		for(i=0; i<NumSolutionTypes; ++i)
			csi[i] = tsi[i];
}

SetSpecialControls(int windex)
{
	Rect	WholeRect, IntersectRect, NullRect = {0,0,0,0};
	Rect	*pLocalGR = &BD_Rec.LocalGR;
	int		hOffset = GetCtlValue(w[windex].BSB);
	int		vOffset = GetCtlValue(w[windex].RSB);

	GetStdDocWindowClipRect(&WholeRect, (WindowPtr) w[windex].Peek);
	if ((windex==BifD) && (!DrawBD)) {
		ResizeControl(w[windex].HS[0], &WholeRect);
		ResizeControl(w[windex].HS[1], &NullRect);
		*pLocalGR = NullRect;
		return;
	}

	if (windex==BifD)	IntersectRect = BD_Rec.HS0_Rect;
	OffsetRect(&IntersectRect, -hOffset, -vOffset);
	IntersectRect.bottom += 1;
	SectRect(&WholeRect, &IntersectRect, &IntersectRect);
	ResizeControl(w[windex].HS[0], &IntersectRect);

	if (windex==BifD)	IntersectRect = BD_Rec.HS1_Rect;
	OffsetRect(&IntersectRect, -hOffset, -vOffset);
	IntersectRect.top += 1;
	IntersectRect.right += 1;
	SectRect(&WholeRect, &IntersectRect, &IntersectRect);
	ResizeControl(w[windex].HS[1], &IntersectRect);

	IntersectRect = w[windex].GR;
	OffsetRect(&IntersectRect, -hOffset, -vOffset);
	SectRect(&WholeRect, &IntersectRect, &IntersectRect);
/*
	if (windex==BifD)
		IntersectRect.right += 1;
*/
/*
	IntersectRect.bottom += 1;
*/
	*pLocalGR = IntersectRect;
/*
	tmpRect.bottom = w[BifD].Peek->port.portRect.bottom - (ScrollBarWidth-1);
*/
}

Boolean FindSpecialClick(int windex, Point pt)
{
	Rect	tmpRect;

	if (windex==BifD)
		tmpRect = BD_Rec.LocalGR;
	++tmpRect.right;
	++tmpRect.bottom;
	return(PtInRect(pt, &tmpRect));
}

	/*	Scrolling Routines	*/

SetSpecialScrolling(int windex)
{
	Rect	MaxRect;
	int		theMax;

	switch (windex) {
		case BifD:	if (!DrawBD)
						MaxRect = w[windex].Peek->port.portRect;
					else
						MaxRect = BD_Rec.WindowSize;
					OffsetRect(&MaxRect, -MaxRect.left, -MaxRect.top);
					break;
	}

	theMax = MAX(MaxRect.bottom - w[windex].Peek->port.portRect.bottom, 0);
	SetCtlValue(w[windex].RSB, MIN(GetCtlValue(w[windex].RSB), theMax));
	SetCtlMax(w[windex].RSB, theMax);

	theMax = MAX(MaxRect.right - w[windex].Peek->port.portRect.right, 0);
	SetCtlValue(w[windex].BSB, MIN(GetCtlValue(w[windex].BSB), theMax));
	SetCtlMax(w[windex].BSB, theMax);

	SetSpecialControls(windex);
	if (WindowIndex(FrontWindow()) == windex)
		QMHiliteControls(windex, TRUE);
}

ScrollCopyBackRect(int windex, Rect *pr, Boolean ToWindow)
{
	GrafPtr	gp;
	Rect	tr;
	int		hOffset;
	int		vOffset;

	if (ToWindow) {
		hOffset = GetCtlValue(w[windex].BSB);
		vOffset = GetCtlValue(w[windex].RSB);
		SetPort(w[windex].Peek);
	}
	else {
		hOffset = 0;
		vOffset = 0;
	}
	GetPort(&gp);
	myCopyBits(w[windex].Back, gp,
			pr, hOffset, vOffset, 0, 0, srcCopy, (RgnHandle) NIL);
	tr = w[windex].GR;
	OffsetRect(&tr, -hOffset, -vOffset);
	if (SectRect(&tr, pr, &tr))
		switch (windex) {
			case BifD:	CopyBDGraph(&tr, hOffset, vOffset);	break;
		}
}

ScrollCopyBackToWindow(int windex)
{
	ScrollCopyBackRect(windex, &w[windex].Peek->port.portRect, TRUE);
}

CopySpecialBackToPrint(int windex)
{
	ScrollCopyBackRect(windex, &w[windex].Back->portRect, FALSE);
}

NewSpecial(int windex)
{
	int	i;

	switch (windex) {
		case BifD:
					if (BD_Rec.Graph != NOPORT) {
						SetPort(BD_Rec.Graph);
						EraseRect(&BD_Rec.Graph->portRect);
					}
					if (BD_Rec.CurrentXInt >= 0)
						SetSpecialTitle(FALSE, windex);
					break;
	}
/*
	StdForceUpdate((WindowPtr) w[windex].Peek);
*/
	if (!w[windex].WindowCleared)
		StdForceUpdate((WindowPtr) w[windex].Peek);
	w[windex].WindowCleared = TRUE;
}

InitSpecial()
{
	int	i;

	NeedNextSpecialValue = FALSE;

	DrawBD = FALSE;
	w[BifD].DrawFromScratch = TRUE;
	w[BifD].Magnifying = TRUE;
	pBDGraphInAllWindows = &BD_Rec.GraphInAllWindows;
	pBDUseColor = &BD_Rec.InColor;
	BD_Rec.GraphInAllWindows = DefaultBDGIAW;
	CheckGraphInAll(BifD);
	BD_Rec.InColor = DefaultBDUseColor;
	CheckSpecialColor(BifD);
	BD_Rec.Graph = NOPORT;
	BD_Rec.InitConds = NIL;
	BD_Rec.SolutionTypes = NIL;
	BD_Rec.Reserved1 = BD_Rec.Reserved2 = BD_Rec.Reserved3 =
			BD_Rec.Reserved4 = 0L;
	BD_Points = NIL;

	SpecialMenuEnable(BifD, FALSE, FALSE);

	SetSpecialControls(BifD);

	if (UsingColorMac)
		InitSpecialColors();

/*
	for(i=0; i<NumSolutionTypes; ++i) {
		csi[i] = i;
	}
*/
}
