/*	qmSpecialDialog.c
 *
 *	Routines which handle Quadratic Map's "special" Dialog Boxes,
 *	those for Bifurcation Diagram and Basin of Attraction.
 *
 *		-TA, 4/22/91
 *
 */

#include <ctype.h>
#include <limits.h>
#include <DialogUtils.h>

#include "qmPrototypes.h"
#include "qmSpecial.h"

enum {
		GraphInAllAlertID = 371,
		SolutionInfoDLogID = 481,
		BifurcationDiagramDLogID,
		spdlogdummy1,
		spdlogdummy2,
		spdlogdummy3,
		SpecialWarningDLogID,
		SpecialOpenContinueDLogID,
		ClearSpecialDLogID,
		ReplaceSolutionInfoDLogID,
		SpecialCloseStopDLogID
};

enum {
		BD_FirstET = 9,
		BD_FirstXAxis = BD_FirstET,
		BD_FirstYAxis = BD_FirstXAxis + 2,
		BD_InitCond = BD_FirstYAxis + 2,
		BD_InitIter,
		BD_ShowIter,
		BD_WindowSize = 17,
		BD_GraphInAll = BD_WindowSize + 2,
		BD_Restart,
		BD_InColor,
		BD_Colors,
		BD_AutoInitial,
		BD_AutoShow
};

#define	DLogPrecision			8
#define UserOutlineItemNum		3
#define	DNST					(NumSolutionTypes + 1)

				/*	Global external vars from qm.c	*/
extern	pIterationVars	pCurrentVars;
extern	hQMPreferences	hQMPrefs;
extern	int				CurrentModifiers;
extern	Boolean			UsingColorMac;
extern	Boolean			GettingParamFiles;

				/*	From qmBifDiag.c	*/
extern	pBifDiagRecord	pbd;

				/*	From qmSpecial.c	*/
extern	RGBColor		SolutionColors[NumSolutionTypes];

				/*	From qmWindows.c; for DoSpecialCloseStopDialog	*/
extern	Boolean			ClosingAllWindows;

			/*	For Bifurcation Diagram	*/
Str255			AutoInitString, AutoString;

			/*	For Solution Info (Colors, Patterns)	*/
SolutionInfo	*ptsi;


		/*	------------------------------------------------------	*/
		/*	----------	General Purpose-Type Routines	----------	*/
		/*	------------------------------------------------------	*/

Boolean	DoSpecialWarningDialog(char *ps)
{
	DialogPtr	dp;

	if (SHIFT_DOWN)	return(TRUE);

	dp = GetNewDialog(SpecialWarningDLogID, NIL, -1L);
	ParamText(ps, "\P","\P","\P");
	if ( HandleDialog(dp, NO_EDIT_TEXT, Cancel,
			UserOutlineItemNum, NIL, NIL, NIL) == OK )
		DialogReturn(dp, TRUE)
	else
		DialogReturn(dp, FALSE)
}

Boolean	DoSpecialOpenContinueDialog(char *ps)
{
	DialogPtr	dp;

	dp = GetNewDialog(SpecialOpenContinueDLogID, NIL, -1L);

	ParamText(ps, "\P","\P","\P");
	if ( HandleDialog(dp, NO_EDIT_TEXT, Cancel,
			UserOutlineItemNum, NIL, NIL, NIL) == OK )
		DialogReturn(dp, TRUE)
	else
		DialogReturn(dp, FALSE)
}

Boolean	DoSpecialCloseStopDialog(char *ps)
{
	DialogPtr	dp;

	if (SHIFT_DOWN)		return(TRUE);
	if (!ClosingAllWindows)
		if (OPTION_DOWN)	return(FALSE);

	dp = GetNewDialog(SpecialCloseStopDLogID, NIL, -1L);

	ParamText(ps, "\P","\P","\P");
	if ( HandleDialog(dp, NO_EDIT_TEXT, Cancel,
			UserOutlineItemNum, NIL, NIL, NIL) == OK )
		DialogReturn(dp, TRUE)
	else
		DialogReturn(dp, FALSE)
}

Boolean	DoSpecialClearDialog(char *ps)
{
	DialogPtr	dp;

	if (SHIFT_DOWN)	return(TRUE);

	dp = GetNewDialog(ClearSpecialDLogID, NIL, -1L);
	ParamText(ps, "\P","\P","\P");
	if ( HandleDialog(dp, NO_EDIT_TEXT, Cancel,
			UserOutlineItemNum, NIL, NIL, NIL) != OK )
		DialogReturn(dp, TRUE)
	else
		DialogReturn(dp, FALSE)
}

Boolean ReplaceSolutionInfoDialog(char *ColOrPat, char *SpecialType)
{
	DialogPtr	dp;

	if (!GettingParamFiles) {
		if (SHIFT_DOWN)	return(TRUE);
		if (OPTION_DOWN) return(FALSE);
	}

	dp = GetNewDialog(ReplaceSolutionInfoDLogID, NIL, -1L);
	ParamText(ColOrPat, SpecialType,"\P","\P");
	if ( HandleDialog(dp, NO_EDIT_TEXT, Cancel,
			UserOutlineItemNum, NIL, NIL, NIL) == OK )
		DialogReturn(dp, TRUE)
	else
		DialogReturn(dp, FALSE)
}

DoGraphInAllAlert(int windex)
{
	char	*s0, *s1;

	if (SHIFT_DOWN)	return;

	switch (windex) {
		case BifDiagram:
					s0 = "\Plambda";
					s1 = "\Ps";
					break;
	}
	ParamText(s0, s1, "\P", "\P");
	if (!NoteAlert(GraphInAllAlertID,NIL)) SysBeep(30);
}


		/*	------------------------------------------------------	*/
		/*	----------	Bifurcation Diagram Routines	----------	*/
		/*	------------------------------------------------------	*/

Boolean BifDiagDLogItemFilter(DialogPtr dp, int itemHit)
{
	Str255	s;
	int		i;
	int		index;
	int		NewMin, NewMax, OldMin, OldMax;
	char	*StatStr, *OldStr, *NewStr;
	double	NewDamping;

	switch (itemHit) {
		case BD_InitIter:
				if (myGetItemCtlValue(dp, BD_AutoInitial) > 0) {
					myGetItemText(dp, itemHit, CP(s));
					if (sUpperComp(CP(s), CP(AutoString)) != 0)
						mySetItemCtlValue(dp, BD_AutoInitial, FALSE);
				}
				else {
					myGetItemText(dp, itemHit, CP(s));
					if (sUpperComp(CP(s), CP(AutoString)) == 0)
						mySetItemCtlValue(dp, BD_AutoInitial, TRUE);
				}
				break;
		case BD_ShowIter:
				if (myGetItemCtlValue(dp, BD_AutoShow) > 0) {
					myGetItemText(dp, itemHit, CP(s));
					if (sUpperComp(CP(s), CP(AutoString)) != 0)
						mySetItemCtlValue(dp, BD_AutoShow, FALSE);
				}
				else {
					myGetItemText(dp, itemHit, CP(s));
					if (sUpperComp(CP(s), CP(AutoString)) == 0)
						mySetItemCtlValue(dp, BD_AutoShow, TRUE);
				}
				break;
		case BD_AutoInitial:
				if (myGetItemCtlValue(dp, itemHit) == 0) {
					mySetItemText(dp, BD_InitIter, CP(AutoString));
					mySelItemText(dp, BD_InitIter);
				}
				else {
					mySelItemText(dp, BD_InitIter);
					return(TRUE);
				}
				break;
		case BD_AutoShow:
				if (myGetItemCtlValue(dp, itemHit) == 0) {
					mySetItemText(dp, BD_ShowIter, CP(AutoString));
					mySelItemText(dp, BD_ShowIter);
				}
				else {
					mySelItemText(dp, BD_ShowIter);
					return(TRUE);
				}
				break;
		case BD_InColor:
				if (myGetItemCtlValue(dp, itemHit) == 0)
					myEnableCtlItem(dp, BD_Colors);
				else
					myDisableCtlItem(dp, BD_Colors);
				break;
		case BD_Colors:
				GetSolutionColors();
				break;
	}
	return(FALSE);
}

Boolean CheckBifurcationDiagramDialog(DialogPtr dp, int itemNo)
{
	Boolean		NumbersOK = TRUE;
	int			i;
	double		tmpF;
	long		tmpL;
	Str255		ErrorStrings[BD_InitIter-BD_FirstET];

	if (itemNo != OK) return(TRUE);

	i=0;
	sEquate(CP(ErrorStrings[i++]), "\PStarting Lambda Value");
	sEquate(CP(ErrorStrings[i++]), "\PEnding Lambda Value");
	sEquate(CP(ErrorStrings[i++]), "\PY-Axis Minimum");
	sEquate(CP(ErrorStrings[i++]), "\PY-Axis Maximum");
	sEquate(CP(ErrorStrings[i++]), "\PInitial Condition");

	for(i=BD_FirstET; i<BD_InitIter; ++i) {
		if ( !myGetItemFloat(dp, i, &tmpF) ) {
			FixBadValAlert(CP(ErrorStrings[i-BD_FirstET]));
			SelectBadField(dp, i, &NumbersOK);
			return(FALSE);
		}
		switch (i) {
			case BD_FirstXAxis:
					pbd->GraphRange[xmin_i] = tmpF;
					break;
			case BD_FirstXAxis+1:
/*	Allow to go backwards??? (not <=)	*/
					if (tmpF == pbd->GraphRange[xmin_i]) {
						ERROR("\PEnding lambda may not equal starting lambda.");
						SelectBadField(dp, i, &NumbersOK);
						return(FALSE);
					}
					pbd->GraphRange[xmax_i] = tmpF;
					break;
			case BD_FirstYAxis:
					pbd->GraphRange[ymin_i] = tmpF;
					break;
			case BD_FirstYAxis+1:
					if (tmpF <= pbd->GraphRange[ymin_i]) {
						ERROR("\PY-axis maximum must be greater than y-axis minimum.");
						SelectBadField(dp, i, &NumbersOK);
						return(FALSE);
					}
					pbd->GraphRange[ymax_i] = tmpF;
					break;
			case BD_InitCond:
					pbd->x00 = tmpF;
					break;
		}
	}

	if (myGetItemCtlValue(dp, BD_AutoInitial) > 0)
		pbd->InitIterTimes = AutoTimes;
	else
		if ( myGetLongParamItem(dp, i, &tmpL,
				"\PNumber of Initial Iterations", 1L, (long) INT_MAX) )
			pbd->InitIterTimes = (int) tmpL;
		else
			SelectBadField(dp, i, &NumbersOK);

	++i;
	if (myGetItemCtlValue(dp, BD_AutoShow) > 0)
		pbd->ShowTimes = AutoTimes;
	else
		if ( myGetLongParamItem(dp, i, &tmpL,
				"\PNumber of Iterations to Show", 1L, (long) INT_MAX) )
			pbd->ShowTimes = (int) tmpL;
		else
			SelectBadField(dp, i, &NumbersOK);

	return(NumbersOK);
}

Boolean DoBifurcationDiagramDialog(Boolean *StartIt)
{
	DialogPtr	dp;
	int			i, index;
	BooleanRect	tmpUsesFull;
	DoubleRect	tmpActual;
	Str255		ErrorStrings[ItemsInRect];
	double		tmpF;
	Str255		s;

	sEquate(CP(AutoString), "\PAuto");

	dp =GetNewDialog(BifurcationDiagramDLogID, NIL, -1L);

	i = BD_FirstET;

	mySetItemFloat(dp, i++, pbd->GraphRange[xmin_i], DLogPrecision);
	mySetItemFloat(dp, i++, pbd->GraphRange[xmax_i], DLogPrecision);
	mySetItemFloat(dp, i++, pbd->GraphRange[ymin_i], DLogPrecision);
	mySetItemFloat(dp, i++, pbd->GraphRange[ymax_i], DLogPrecision);
	mySetItemFloat(dp, i++, pbd->x00, DLogPrecision);

	if (pbd->InitIterTimes == AutoTimes) {
		mySetItemText(dp, i++, CP(AutoString));
		mySetItemCtlValue(dp, BD_AutoInitial, TRUE);
	}
	else
		mySetItemLong(dp, i++, (long) pbd->InitIterTimes);

	if (pbd->ShowTimes == AutoTimes) {
		mySetItemText(dp, i++, CP(AutoString));
		mySetItemCtlValue(dp, BD_AutoShow, TRUE);
	}
	else
		mySetItemLong(dp, i++, (long) pbd->ShowTimes);

	if (pbd->GraphInAllWindows)
		mySetItemCtlValue(dp, BD_GraphInAll, TRUE);
	mySetItemCtlValue(dp, BD_WindowSize + ((pbd->FullWindow) ? 1 : 0), TRUE);
	mySetItemCtlValue(dp, BD_Restart, TRUE);
	mySetItemCtlValue(dp, BD_InColor, pbd->InColor);
	if (!UsingColorMac)
		myDisableCtlItem(dp, BD_InColor);
	if (pbd->InColor)
		myEnableCtlItem(dp, BD_Colors);
	else
		myDisableCtlItem(dp, BD_Colors);

	if (HandleDialog(dp, BD_FirstET,  Cancel, UserOutlineItemNum, NIL,
			&BifDiagDLogItemFilter, &CheckBifurcationDiagramDialog) == OK) {
		
		pbd->GraphInAllWindows = myGetItemCtlValue(dp, BD_GraphInAll) > 0;
		pbd->FullWindow = myGetItemCtlValue(dp, BD_WindowSize+1) > 0;
		*StartIt = myGetItemCtlValue(dp, BD_Restart) > 0;
		pbd->InColor = myGetItemCtlValue(dp, BD_InColor) > 0;

		DialogReturn(dp, TRUE);
	}
	else DialogReturn(dp, FALSE);
}


		/*	--------------------------------------------------	*/
		/*	----------	Solution Patterns Routines	----------	*/
		/*	--------------------------------------------------	*/

pascal void DrawSolColsRoutine(WindowPtr wp, int itemNo)
{
	int			itemType;
	Handle		item;
	Rect		box;

	GetDItem((DialogPtr) wp, itemNo, &itemType, &item, &box);
	SetPort(wp);
	SetSpecialColor(itemNo / DNST - 2, FALSE);
	PaintRect(&box);
	SetSpecialColor(BlackColorIndex, FALSE);
	FrameRect(&box);
}

Boolean SolutionItemFilterProc(DialogPtr dp, int itemHit)
{
	int				newPatIndex, oldPatIndex;
	int				newSolIndex, oldSolIndex;
	int				i;

	if ( (itemHit < DNST*2) || (itemHit % DNST == NumSolutionTypes) )
		return(FALSE);

	newPatIndex = itemHit / DNST - 2;
	newSolIndex = itemHit % DNST;
	oldPatIndex = (*ptsi)[newSolIndex];
	if (newPatIndex == oldPatIndex)	return(TRUE);

	for(i=0; i<NumSolutionTypes; ++i)
		if ((*ptsi)[i] == newPatIndex) {
			oldSolIndex = i;
			break;
		}
	(*ptsi)[newSolIndex] = newPatIndex;
	(*ptsi)[oldSolIndex] = oldPatIndex;
	mySetItemCtlValue(dp, (newPatIndex+2) * DNST + newSolIndex, TRUE);
	mySetItemCtlValue(dp, (newPatIndex+2) * DNST + oldSolIndex, FALSE);
	mySetItemCtlValue(dp, (oldPatIndex+2) * DNST + oldSolIndex, TRUE);
	mySetItemCtlValue(dp, (oldPatIndex+2) * DNST + newSolIndex, FALSE);

	return(TRUE);
}

DialogPtr GetSolutionInfoDialog()
{
	return(GetNewDialog(SolutionInfoDLogID, NIL, -1L));
}

Boolean DoSolutionInfoDialog()
{
	DialogPtr	dp;
	int			i, j;
	int			itemType;
	Handle		item;
	Rect		box;

	ptsi = &csi;

	dp = GetNewDialog(SolutionInfoDLogID, NIL, -1L);
	ParamText("\PCOLORS", "\P", "\P", "\P");
	for(i=0; i<NumSolutionTypes; ++i)
		mySetItemCtlValue(dp, ((*ptsi)[i]+2)*DNST + i, TRUE);
	for(i =  3*DNST - 1; i <= DNST * (DNST+1) - 1; i += DNST) {
		GetDItem(dp, i, &itemType, &item, &box);
		SetDItem(dp, i, itemType, (Handle) &DrawSolColsRoutine, &box);
	}
	if (HandleDialog(dp, NO_EDIT_TEXT, Cancel,
			UserOutlineItemNum, NIL, &SolutionItemFilterProc, NIL) == OK) {
		DialogReturn(dp, TRUE)
	}
	else
		DialogReturn(dp, FALSE)
}
