/*	qmWindows.c
 *
 *	Handles Quadratic Map's windows
 *
 *		-TA 8/14/90
 */

#include <OffPort.h>
#include "qmPrototypes.h"
#include "qmGraphics.h"
#include "qmSpecial.h"
#include "qmMenus.h"

#define	WindowTitleResID		128
#define	DragMargin				4

typedef struct {
			int		windex;
			Point	wpos;
		}	WindowArrangement;

	/*	External variables from qmGraphics.c */
extern	Windows		w;

	/*	External variables from qm.c */
extern	int			CurrentModifiers;	
extern	Boolean		UsingColorMac;
extern	Boolean		WasADefaultsFile;

	/*	From qmContGraphics.c */
extern	GrafPtr		AnimationParts;

	/*	From qmMenus.c	*/
extern	MenuHandle	WindowsMenuH;

	/*	From qmPrint.c	*/
extern	GrafPtr		PrintOffPort;

	/*	From qmBifDiag.c--for ArrangeWindows()	*/
extern	Boolean			DrawBD;
extern	BifDiagRecord	BD_Rec;

	int		MinScreenX, MaxScreenX, MinScreenY, MaxScreenY;	/* Screen Size */
	int		MinMainScreenX, MinMainScreenY, MaxMainScreenX, MaxMainScreenY;
	Rect	LWR;							/* "LargeWindowRect"zoomed out */
	int		RightWindowLimit, BottomWindowLimit;
	int		StdWindowWidth, StdWindowHeight;

	WindowArrangement	WAs[NumNormalWindows];

	WindowLocation		wd[NumWindows], CurWD[NumWindows];
	int					DefaultWindowNum, CurWindowNum;

	Boolean				ClosingAllWindows = FALSE;


int	WindowIndex(WindowPtr wp)
{
	int	i;
	for(i=0;i<NumWindows;++i) if (wp == (WindowPtr) w[i].Peek) return(i);
	return(NoWindow);
}

Boolean WindowOpen(int windex)
{
	return(w[windex].Peek->visible);
}

Boolean IsMagnified(int windex)
{
	return(w[windex].Magnifying);
}

Size WindowLocsSize(int n)
{
	return(sizeof(int) + n*sizeof(WindowLocation));
}


	/* In following, argument i represents the window index */

		/*	In WindowMove/LineTo, ClipRgn handles out-of-region behavior */
WindowMoveTo(int i, int x, int y) /* Using Sized but not offset coords */
{
	MoveTo(x + w[i].Origin.h, w[i].Origin.v - y);
}

WindowLineTo(int i, int x, int y)
{
	LineTo(x + w[i].Origin.h, w[i].Origin.v - y);
}

/*	Back (used to be Screen) MoveTo and LineTo used to draw to
 *	Trajectory and Phase Space Window OffPorts (w[].Back).
 */

BackMoveTo(int i, int x, int y)	/* Using Sized but not offset coords */
{
/*
	if (x<0) x=0;		if (x>MaxXCoord-Traj_LeftMargin) x=MaxXCoord-Traj_LeftMargin;
	if (y<0) y=0;		if (y>MaxYCoord-Traj_TopMargin) y=MaxYCoord-Traj_TopMargin;
*/
	MoveTo(x + w[i].BackOrigin.h, w[i].BackOrigin.v - y);
}

BackLineTo(int i, int x, int y)
{
/*
	if (	(x<w[i].BackLimits.left-w[i].BackOrigin.h) ||
			(x>w[i].BackLimits.right -w[i].BackOrigin.h) ||
			(y<w[i].BackOrigin.v-w[i].BackLimits.bottom) ||
			(y>w[i].BackOrigin.v-w[i].BackLimits.top)	)
		MoveTo(x + w[i].BackOrigin.h, w[i].BackOrigin.v - y);
	else
*/
		LineTo(x + w[i].BackOrigin.h, w[i].BackOrigin.v - y);
}

BothMoveTo(int i, int x, int y)
{
	SetPort(w[i].Peek);	 WindowMoveTo(i, x, y);
	SetPort(w[i].Back);	 BackMoveTo(i, x, y);
}

BothLineTo(int i, int x, int y)
{
	SetPort(w[i].Peek);	 WindowLineTo(i, x, y);
	SetPort(w[i].Back);	 BackLineTo(i, x, y);
}


	/* ------------------------------------------------- */
	/* ---------- Coordinate scaling routines ---------- */
	/* ---- (argument i represents the window index) --- */
	/* ------------------------------------------------- */

int	xFloatToCoord(int i, double f)
{
/*
	if (!w[i].Magnifying)
		return( RoundToInt(f*w[i].xScale) ); (* trunc? *)
	else {
*/

		/*	Not sure why assymetry in EffectiveZero below	-TA 5/1/91	*/
	if (w[i].xScale >= 0) {
		if (f<w[i].GraphRange[xmin_i])
			return(-MaxScreenX);
		else if (f>w[i].GraphRange[xmax_i])
			return(MaxScreenX);
	}
	else {
		if (f-w[i].GraphRange[xmax_i]<-EffectiveZero)
			return(MaxScreenX);
		else if (f-w[i].GraphRange[xmin_i]>EffectiveZero)
			return(-MaxScreenX);
	}
	if (w[i].xGraphAroundZero)
		return( RoundToInt(f*w[i].xScale) ); /* trunc? */
	else
		if ( (MAX(w[i].GraphRange[xmax_i], w[i].GraphRange[xmin_i]) > 0.0) ||
				(i==BifD) )
			return( RoundToInt((f-w[i].GraphRange[xmin_i])*w[i].xScale) ); /* trunc? */
		else
			return( RoundToInt((f-w[i].GraphRange[xmax_i])*w[i].xScale) ); /* trunc? */
}

int	yFloatToCoord(int i, double f)
{
/*
	if (!w[i].Magnifying)
		return( RoundToInt(f*w[i].yScale) ); (* trunc? *)
	else {
*/

	if (w[i].yScale >= 0) {
		if (f-w[i].GraphRange[ymin_i]<-EffectiveZero)
			return(-MaxScreenY);
		else if (f-w[i].GraphRange[ymax_i]>EffectiveZero)
			return(MaxScreenY);
	}
	else {
		if (f<w[i].GraphRange[ymax_i])
			return(MaxScreenY);
		else if (f>w[i].GraphRange[ymin_i])
			return(-MaxScreenY);
	}
	if (w[i].yGraphAroundZero)
		return( RoundToInt(f*w[i].yScale) ); /* trunc? */
	else
		if (MAX(w[i].GraphRange[ymax_i], w[i].GraphRange[ymin_i]) > 0.0)
			return( RoundToInt((f-w[i].GraphRange[ymin_i])*w[i].yScale) ); /* trunc? */
		else
			return( RoundToInt((f-w[i].GraphRange[ymax_i])*w[i].yScale) ); /* trunc? */
/*
	if (w[i].yGraphAroundZero)
		return( RoundToInt(f*w[i].yScale) ); (* trunc? *)
	else
		return( RoundToInt((f-w[i].GraphRange[ymin_i])*w[i].yScale) ); (* trunc? *)
*/
}

double	xCoordToFloat(int i, int Coord)
{
	if (w[i].xGraphAroundZero)
		return( (double) Coord/w[i].xScale );
	else
/*
		if (w[i].GraphRange[xmax_i] > 0)
*/
		if ( (MAX(w[i].GraphRange[xmax_i], w[i].GraphRange[xmin_i]) > 0.0) ||
				(i==BifD) )
			return( (double) Coord/w[i].xScale + w[i].GraphRange[xmin_i] );
		else
			return( (double) Coord/w[i].xScale + w[i].GraphRange[xmax_i] );
}

double	yCoordToFloat(int i, int Coord)
{
	if (w[i].yGraphAroundZero)
		return( (double) Coord/w[i].yScale );
	else
		if (MAX(w[i].GraphRange[ymax_i], w[i].GraphRange[ymin_i]) > 0.0)
			return( (double) Coord/w[i].yScale + w[i].GraphRange[ymin_i] );
		else
			return( (double) Coord/w[i].yScale + w[i].GraphRange[ymax_i] );
}

Fixed xFloatToFixed(int i, double f)
/*	Need to process magnification?	*/
{
	double tmpF = f*w[i].xScale;
	return( X2Fix(&tmpF) );
}

Fixed yFloatToFixed(int i, double f)
/*	Need to process magnification?	*/
{
	double tmpF = f*w[i].yScale;
	return( X2Fix(&tmpF) );
}


	/* ----------------------------------------------------------------- */
	/* ---- Nearly-generic window manipulation (Event) sub-routines ---- */
	/* ----------------------------------------------------------------- */

QMForceUpdate(WindowPtr wp)
/* Called whenever window size changes */
{
	int		windex;
	Rect	tmpRect;

	StdForceUpdate(wp);

	switch (windex = WindowIndex(wp)) {
		case Orbit:
			SetOrbitCoords();	break;
		case PSpace:
			InitPSScales();		break;
		case EmbPS:
			InitEPSScales();	break;
		case TxOut:
			tmpRect = (**w[windex].HS[0]).contrlRect;
			tmpRect.right =
					w[windex].Peek->port.portRect.right - (ScrollBarWidth-1);
			ResizeControl(w[windex].HS[0], &tmpRect);
			SetNumLines(windex);		break;
		case BifD:
			SetSpecialScrolling(windex);break;
	}

	ResetScrollBars(windex);

}

QMSizeWindow(WindowPtr wp, int h, int v)
{
	SizeWindow(wp, h, v, TRUE);
	QMForceUpdate(wp);
}

QMOpenWindow(WindowPtr wp)
{
	ShowWindow(wp);
	SelectWindow(wp);
	SetItemMark(WindowsMenuH, WindowIndex(wp) + WindowSeparatorItem + 1, diamondMark);
}

QMCloseWindow(WindowPtr wp)
{
	int	i = WindowIndex(wp);

	HideWindow(wp);
	CheckItem(WindowsMenuH, WindowIndex(wp) + WindowSeparatorItem + 1, FALSE);
	if (i == BifDiagram)
		SeeIfStopSpecial(i);
}

Boolean QMCloseFrontWindow()
{
	WindowPeek	wp;
	int			kind;
	
	wp = (WindowPeek) FrontWindow();
	kind = (wp) ? wp->windowKind : 0;
	if (kind < 0) CloseDeskAcc(kind);
	else if (kind != 0)
		QMCloseWindow((WindowPtr) wp);
	return(kind != 0);
}

QMCloseAllWindows()
{
	ClosingAllWindows = TRUE;

	while (QMCloseFrontWindow())
		;

	ClosingAllWindows = FALSE;
}

QMZoomWindow(Point Where, WindowPtr WhichWindow, int Part)
{
	if (TrackBox(WhichWindow, Where, Part)) {
		StdZoomWindow(WhichWindow, Part, FALSE);
		QMForceUpdate(WhichWindow);
	}
}

MakeLargeWindow(WindowPtr wp)
{
	StdZoomWindow(wp, inZoomOut, FALSE);
	QMForceUpdate(wp);
}

QMGrowWindow(Point pt, WindowPtr WhichWindow)
{
	int		windex;
	Rect	SizeRect;
	long	lresult;
	int		rightlimit, bottomlimit;

	switch (windex = WindowIndex(WhichWindow)) {
		case PSpace:
		case EmbPS:
		case Orbit:
			SetRect(&SizeRect, 80, 80,
				RightWindowLimit+1, BottomWindowLimit+1);
/*
			SetRect(&SizeRect, StdMargin + StdMargin + ScrollBarWidth + 4,
				StdMargin + StdMargin + ScrollBarWidth + 4,
				RightWindowLimit+1, BottomWindowLimit+1);
			SetRect(&SizeRect, Orbit_LeftMargin + Orbit_RightMargin + 4,
				Orbit_TopMargin + Orbit_BottomMargin + 4,
				RightWindowLimit+1, BottomWindowLimit+1);
*/
			break;
		case TxOut:
			SetRect(&SizeRect, ScrollBarWidth + 45,
				BodyTETop+TextLineHeight+ScrollBarWidth+BottomTextMargin,
				RightWindowLimit+1, BottomWindowLimit+1);
			break;
		case BifD:
			SetPort(w[windex].Back);	/*	For CharWidth()	*/
			SetRect(&SizeRect,
				ScrollBarWidth + 2*StdMargin +
				(2*StdPSFieldWidth+3) * CharWidth('0') + BD_MinSize,
				PS_TopMargin + BD_BottomMargin + BD_MinSize,
				RightWindowLimit+1, BottomWindowLimit+1);
			break;
		default:
			SetRect(&SizeRect, ScrollBarWidth, ScrollBarWidth,
				RightWindowLimit+1, BottomWindowLimit+1);
			break;
	}
	lresult = GrowWindow(WhichWindow, pt, &SizeRect);
	QMSizeWindow((WindowPtr) WhichWindow, LoWord(lresult), HiWord(lresult));
}

QMDragWindow(Point pt, WindowPtr WhichWindow, int mods)
{
	Rect	boundsRect;

	if (!(mods & cmdKey))	SelectWindow(WhichWindow);

	SetRect(&boundsRect, MinScreenX, MinScreenY, MaxScreenX, MaxScreenY);
	InsetRect(&boundsRect, DragMargin, DragMargin);
	DragWindow(WhichWindow, pt, &boundsRect);
}

QMGoAwayWindow(Point pt, WindowPtr WhichWindow)
{
	if (TrackGoAway(WhichWindow, pt)) {
		if (OPTION_DOWN)	QMCloseAllWindows();
		else				QMCloseWindow(WhichWindow);
	}
}

QMHandleActivate(WindowPtr wp, int ActivateFlag)
{
	int			windex;
	TEHandle	teh = NIL;

	windex = WindowIndex(wp);

	HandleTextActivation(windex, ActivateFlag);

	if ( (w[windex].RSB == NIL) && (w[windex].BSB == NIL) ) {
		StdHandleActivate(wp, ActivateFlag);
		return;
	}

	if (!(ActivateFlag & 1))
		EraseGrowIcon(wp);
	SetPort(wp); ClipRect(&wp->portRect);
	DrawGrowIcon(wp);
	QMHiliteControls(windex, ActivateFlag & 1);

		/* unhighlight text if necessary */
	StdDocWindowClipRect(wp);
}


	/* ----------------------------------------------------- */
	/* ----	Main Window Routineswere in qmGraphics.c	---- */
	/* ----------------------------------------------------- */

SelectWindowIndex(int i)
{
	if (!WindowOpen(i))
		switch (i) {
			case PSpace:
			case EmbPS:
			case Orbit:	w[i].DrawFromScratch = TRUE;	break;
		}
	QMOpenWindow((WindowPtr) w[i].Peek);
}

SetOpenWindows()
{
	int			i;
	WindowPeek	wp;

	/*	Check that normal windows not bigger than their bitmaps	*/
	for(i=CurWindowNum-1;i>=0;--i) {
		if (CurWD[i].windex >= NumNormalWindows)
			continue;
		if (CurWD[i].wr.bottom - CurWD[i].wr.top > BottomWindowLimit)
			CurWD[i].wr.bottom = CurWD[i].wr.top + BottomWindowLimit;
		if (CurWD[i].wr.right - CurWD[i].wr.left > RightWindowLimit)
			CurWD[i].wr.right = CurWD[i].wr.left + RightWindowLimit;
	}

	/*	Verify all windows appear on screen (created on larger screen?)	*/
	/*	Should really see if title bar region is in GrayRgn?	*/
	for(i=CurWindowNum-1;i>=0;--i) {
		if (CurWD[i].wr.top < MinScreenY + DragMargin)
			OffsetRect(&CurWD[i].wr, 0, LWR.top - CurWD[i].wr.top);
		if (CurWD[i].wr.top > MaxScreenY)
			OffsetRect(&CurWD[i].wr, 0, LWR.top - CurWD[i].wr.top);
		if (CurWD[i].wr.right < MinScreenX + DragMargin)
			OffsetRect(&CurWD[i].wr, LWR.left - CurWD[i].wr.left, 0);
		if (CurWD[i].wr.left > MaxScreenX - DragMargin)
			OffsetRect(&CurWD[i].wr, LWR.right - CurWD[i].wr.right, 0);
	}

	for(i=CurWindowNum-1;i>=0;--i) {
		QMSizeWindow((WindowPtr) w[CurWD[i].windex].Peek,
				CurWD[i].wr.right-CurWD[i].wr.left, CurWD[i].wr.bottom-CurWD[i].wr.top);
		MoveWindow(w[CurWD[i].windex].Peek, CurWD[i].wr.left, CurWD[i].wr.top, FALSE);
		SelectWindowIndex(CurWD[i].windex);
/*
		HandleAllUpdates();
*/
	}
}

DefaultWindowConfiguration()
{
	int			i;
	WindowPeek	wp;

	for(i=DefaultWindowNum-1;i>=0;--i) {
		QMSizeWindow((WindowPtr) w[wd[i].windex].Peek,
				wd[i].wr.right-wd[i].wr.left, wd[i].wr.bottom-wd[i].wr.top);
		MoveWindow(w[wd[i].windex].Peek, wd[i].wr.left, wd[i].wr.top, FALSE);
		SelectWindowIndex(wd[i].windex);
	}
	if (DefaultWindowNum > 0)
		wp = w[wd[DefaultWindowNum-1].windex].Peek;
	else
		for(wp=(WindowPeek) FrontWindow(); wp != NIL;) {
			if (wp->visible)
				if ( WindowIndex((WindowPtr) wp) != NoWindow )
					break;
			wp = wp->nextWindow;
		}

	for(wp = wp->nextWindow; wp != NIL;) {
		if (wp->visible)
			if ( WindowIndex((WindowPtr) wp) != NoWindow )
				QMCloseWindow((WindowPtr) wp);
		wp = wp->nextWindow;
	}

}

Boolean GetDefaultWindowConfiguration(Boolean AtStartup)
/*	Returns TRUE if sets default either by file or by default default.
 *	AtStartup TRUE implies that there is no current default, so uses
 *	default default if no file.
 *		-TA 1/21/91
 */
{
	int	i;

/*
	if (!OpenDefaults()) {
*/
	if (!WasADefaultsFile) {
		if (!AtStartup) return(FALSE);
		for(i=0;i<NumDefaultWindows;++i) {
			wd[i].windex = i;
			wd[i].wr = w[i].Peek->port.portRect;
			OffsetRect(&wd[i].wr, -w[i].Peek->port.portBits.bounds.left,
					-w[i].Peek->port.portBits.bounds.top);	
				/*	wd[i].wr now equals what will become
				 *	(**w[i].Peek->contRgn).rgnBBox; the contRgn, however,
				 *	is {0,0,0,0} when the window is invisible, as it is now
				 *		-TA 12/29/90
				 */
		}
		DefaultWindowNum = i;
	}
	return(TRUE);
}

InitialWindowConfiguration(Boolean PrintingOnly)
{
	int		i;

	for(i=0;i<NumNormalWindows;++i) {
		QMSizeWindow((WindowPtr) w[i].Peek, StdWindowWidth, StdWindowHeight);
		MoveWindow(w[i].Peek, WAs[i].wpos.h, WAs[i].wpos.v, FALSE);
	}
	if (PrintingOnly) {
		return;
	}
	if (GetDefaultWindowConfiguration(TRUE)) ;
	CurrentModifiers = GetCurrentModifiers();
	DefaultWindowConfiguration();
}

SetCurWindows(int windex)
{
	WindowPeek	wp;
	int			i;

	if (windex == BifD) {
		CurWindowNum = 1;
		CurWD[0].windex = windex;
		wp = w[windex].Peek;
		CurWD[0].wr = (**wp->contRgn).rgnBBox;
		return;
	}

	CurWindowNum = 0;
	wp = (WindowPeek) FrontWindow();
	while(wp != NIL) {
		if (wp->visible)
			if ( (i=WindowIndex((WindowPtr) wp)) != NoWindow )
				if (i != BifD) {
					CurWD[CurWindowNum].windex = i;
					CurWD[CurWindowNum].wr = (**wp->contRgn).rgnBBox;
					++CurWindowNum;
				}
		wp = wp->nextWindow;
	}
}

SetDefaultWindowConfiguration()
{
	WindowPeek	wp = (WindowPeek) FrontWindow();
	int			i;

	DefaultWindowNum = 0;
	while(wp != NIL) {
		if (wp->visible)
			if ( (i=WindowIndex((WindowPtr) wp)) != NoWindow ) {
				wd[DefaultWindowNum].windex = i;
				wd[DefaultWindowNum].wr = (**wp->contRgn).rgnBBox;
				++DefaultWindowNum;
			}
		wp = wp->nextWindow;
	}
}

AddWindowsToMenu()
{
	int		i;
	Str255	s;

	for(i=0;i<NumWindows;++i) {
		GetWTitle(w[i].Peek, CP(s));
		AppendMenu(WindowsMenuH, CP(s));
	}

}

SetTextTitle(Boolean FileExists, int windex)
{
	char	*pfname;

	switch (windex) {
		case BifD:	pfname = "\P(Bifurcation Diagram Data)";
					SetANilFileName(TxOut);
					break;
		default:		/*	(TxOut; TxOut test below is relic)	*/
					pfname = GetSaveFName(windex);
					if ( (!FileExists) && (windex == TxOut) )
						SetANilFileName(TxOut);
					if ( (!FileExists) || (pfname[0] == 0) )
						pfname = "\PUntitled";
					break;
	}

	SetItem(WindowsMenuH, TextWindowMenuItem, pfname);
	SetWTitle(w[TxOut].Peek, pfname);
}

GetTextTitle(char *theTitle)
{
	sEquate(theTitle, (char *) *(w[TxOut].Peek->titleHandle));
}

int SelectOrderIndex(int i)
{
	switch (i) {
		case 0:
		case 1:	return(i);
		case 2:
		case 3:	return(i+3);
		case 4:
		case 5:
		case 6:	return(i-2);
	}
}

ArrangeWindows()
{
	int		i, j;
	int		windex, orderIndex;
	Point	SpecialPt;

		/*	Set non-default positions for non-default windows	*/
	for(i=NumDefaultWindows;i<NumNormalWindows;++i)
		if (WindowOpen(i))
			if (WAs[NumNormalWindows-i-1].windex != i)
				for(j=0;j<NumNormalWindows;++j) {
					if (WAs[j].windex == NoWindow) {
						WAs[j].windex = i;
						break;
					}
				}

		/*	Size (to smaller of screen/WindowSize),
		 *	move, select back windows (non-"normal")
		 */
	for(i=NumWindows-1; i>=NumNormalWindows; --i)
		if (WindowOpen(i)) {
			if (i == BifD) {
				if (DrawBD) {
					SpecialPt.h = MIN(LWR.right - LWR.left,
							BD_Rec.WindowSize.right - BD_Rec.WindowSize.left);
					SpecialPt.v = MIN(LWR.bottom - LWR.top,
							BD_Rec.WindowSize.bottom - BD_Rec.WindowSize.top);
				}
				else {
					SpecialPt.h = LWR.right - LWR.left;
					SpecialPt.v = LWR.bottom - LWR.top;
				}
			}
			QMSizeWindow((WindowPtr) w[i].Peek, SpecialPt.h, SpecialPt.v);
			MoveWindow(w[i].Peek, LWR.left, LWR.top, FALSE);
			SelectWindow(w[i].Peek);
		}

		/*	Select "normal" windows	*/
	for(i=NumNormalWindows-1;i>=0;--i) {
		if (!WindowOpen(i))
			continue;
		QMSizeWindow((WindowPtr) w[i].Peek, StdWindowWidth, StdWindowHeight);
		MoveWindow(w[i].Peek, WAs[i].wpos.h, WAs[i].wpos.v, FALSE);
		SelectWindow(w[i].Peek);
	}
}

InitWindowArrangement()
{
	int	i;

	for(i=0;i<NumNormalWindows;++i)
		WAs[i].windex = NoWindow;
	WAs[0].wpos.h = LWR.left;
	WAs[0].wpos.v = LWR.top;
	WAs[1].wpos.h = LWR.right-StdWindowWidth;
	WAs[1].wpos.v = LWR.top;
	WAs[2].wpos.h = LWR.right-StdWindowWidth;
	WAs[2].wpos.v = LWR.bottom-StdWindowHeight;
	WAs[3].wpos.h = LWR.left;
	WAs[3].wpos.v = LWR.bottom-StdWindowHeight;
}

NewQMWindow(int windex)
{
	StdForceUpdate((WindowPtr) w[windex].Peek);

	if (w[windex].Back != NOPORT) {
		SetPort(w[windex].Back);
		EraseRect(&w[windex].Back->portRect);
	}
}

NewQMWindows(Boolean ClearSolutions)
{
	int	i;

	for (i=0;i<NumNormalWindows;++i) {
		if (w[i].WindowCleared) continue;
		if ( (i == PSpace) || (i == EmbPS) || (i == Orbit) )
			StdForceUpdate((WindowPtr) w[i].Peek);
		else
			NewQMWindow(i);
	}

	if (ClearSolutions) {
		w[PSpace].DrawFromScratch = TRUE;
		w[EmbPS].DrawFromScratch = TRUE;
		w[Orbit].DrawFromScratch = TRUE;
	}
}

Boolean InitQMOffPorts()
{
	int		i;
	Rect	tmpRect;

	for(i=0;i<NumWindows;++i)
		w[i].Back = NOPORT;

	RightWindowLimit = ( (MaxScreenX-MinScreenX) -
			(MaxMainScreenX-MinMainScreenX) ) +
			LWR.right - LWR.left;
	BottomWindowLimit =( (MaxScreenY-MinScreenY) -
			(MaxMainScreenY-MinMainScreenY) ) +
			LWR.bottom - LWR.top;

	SetRect(&tmpRect, 0, 0, RightWindowLimit, BottomWindowLimit);
	if ( (PrintOffPort		= NewOffPort(&tmpRect)) == NIL)		return(FALSE);
	tmpRect.right *= 2;
	if ( (w[Orbit].Back		= NewOffPort(&tmpRect)) == NIL)		return(FALSE);
	tmpRect.right /= 2;
	if ( (w[PSpace].Back	= NewOffPort(&tmpRect)) == NIL)		return(FALSE);
	if ( (w[EmbPS].Back		= NewOffPort(&tmpRect)) == NIL)		return(FALSE);
	if ( (w[BifD].Back		= NewOffPort(&tmpRect)) == NIL)		return(FALSE);

	for(i=0;i<NumWindows;++i)
		if (w[i].Back != NOPORT) {
			SetPort(w[i].Back);	
			SetStdTextAttributes();
		}

	return(TRUE);
}

InitQMWindows()
{
	GDHandle	MainScreen;
	Rect		CurScreenRect;
	RgnHandle	CurScreenRgn;

	WindowPeek	tmpWPeek;
	int			i;
	Rect		tmpRect;
	Str255		s;

	MinScreenX = (**GrayRgn).rgnBBox.left;
	MaxScreenX = (**GrayRgn).rgnBBox.right;
	MinScreenY = (**GrayRgn).rgnBBox.top;
	MaxScreenY = (**GrayRgn).rgnBBox.bottom;

	if ( TrapAvailable(GetMainDeviceTrapID, ToolTrap, TRUE) ) {
		MainScreen = GetMainDevice();
/*
		CurScreenRect = (**MainScreen).gdRect;
*/
		CurScreenRgn = NewRgn();
		RectRgn(CurScreenRgn, &(**MainScreen).gdRect);
		SectRgn(CurScreenRgn, GrayRgn, CurScreenRgn);
		CurScreenRect = (**CurScreenRgn).rgnBBox;
	}
	else
		CurScreenRect = (**GrayRgn).rgnBBox;
	MinMainScreenX = CurScreenRect.left;
	MinMainScreenY = CurScreenRect.top;
	MaxMainScreenX = CurScreenRect.right;
	MaxMainScreenY = CurScreenRect.bottom;

	SetRect(&tmpRect,0,0,0,0);
	tmpWPeek = (WindowPeek) NewWindow(NIL, &tmpRect, "\P", FALSE /* visible */,
		documentProc+8 /* +8 for zoom */, NIL, TRUE /* goAwayFlag */, NIL);
	LWR = (**(WStateData**) ((*tmpWPeek).dataHandle)).stdState;
	DisposeWindow(tmpWPeek);

		/* Set default window sizes */
	StdWindowWidth = (MaxMainScreenX - 4*LWR_MARGIN-2)/2;
	StdWindowHeight = (MaxMainScreenY - MBarHeight - (3+4+3) )/2 - TitleBarWidth;

	for (i=0;i<NumWindows;++i) {
		if ( (i<NumNormalWindows) || !UsingColorMac )
			w[i].Peek = (WindowPeek) NewWindow(NIL, &LWR, "\P", FALSE /* visible */,
				documentProc+8 /* +8 for zoom */, NIL, TRUE /* goAwayFlag */, NIL);
		else
			w[i].Peek = (WindowPeek) NewCWindow(NIL, &LWR, "\P", FALSE /* visible */,
				documentProc+8 /* +8 for zoom */, NIL, TRUE /* goAwayFlag */, NIL);
		SetPort(w[i].Peek);	
		SetStdTextAttributes();
		ClipRect(&tmpRect);
	
		w[i].Magnifying = FALSE;
		w[i].xGraphAroundZero = w[i].yGraphAroundZero = TRUE;
		w[i].WindowCleared = TRUE;
		w[i].ReservedBoolean = FALSE;
		w[i].Reserved1 = w[i].Reserved2 = w[i].Reserved3 =
				w[i].Reserved4 = 0L;
	}

	SetTextPorts((GrafPtr) w[TxOut].Peek);

	for(i=0;i<NumWindows;++i) {
		GetIndString(&s, WindowTitleResID, i+1);
		SetWTitle(w[i].Peek, &s);
	}
	SetTextTitle(FALSE, TxOut);

	InitQMControls();
	InitWindowArrangement();
/*
	InitialWindowConfiguration();
*/
}

HaltQMWindows()
{
	int	i;

	DisposeOffPort(PrintOffPort);
	for (i=0;i<NumWindows;++i) {
		if (w[i].Back != NOPORT)
			DisposeOffPort(w[i].Back);
		KillControls((WindowPtr) w[i].Peek);
		DisposeWindow(w[i].Peek);
	}
}
