# This file is a group of files which have been specially packaged
# by bundle, a program on page 98 of Kernighan & Pike.  You can read
# the file as it is and if you have a UNIX system you can execute it
# as a shell script and it will unbundle itself into separate files.
# To unbundle, remove everything above this line and then execute
# the file with sh.
\n#==========================
#= Here is file Makefile =
#==========================\n
echo Makefile 1>&2
cat >Makefile <<'End of Makefile'
# makefile for ode
# this version of the makefile should work for most systems
# for gcc you might want to try the -traditional flag
# for other compilers you may want to add optimization flags.

OBJ = gram.o expr.o sym.o misc.o ext.o prt.o \
	am.o ama.o rk.o rka.o eu.o float.o stperr.o

SRC = am.c ama.c eu.c expr.c ext.c float.c fps.h \
	gram.y lex.l makefile misc.c num.h ode.h prt.c \
	rk.c rka.c stperr.c sym.c

ode:	$(OBJ)
	$(CC) -o ode $(OBJ) -loln -lm

$(OBJ):	ode.h
float.o: fps.h
gram.o:	lex.c gram.c
rka.o ama.o:	num.h

lex.c:	lex.l
	lex lex.l
	mv lex.yy.c lex.c

gram.c:	gram.y
	yacc gram.y
	mv y.tab.c gram.c

clean:
	-rm $(OBJ) y.tab.c lex.c lex.yy.c gram.c

print:
	pr $(SRC) | lpr
End of Makefile
\n#==========================
#= Here is file Ode.manpage =
#==========================\n
echo Ode.manpage 1>&2
cat >Ode.manpage <<'End of Ode.manpage'



ODE(1)                   USER COMMANDS                     ODE(1)



NAME
     ode - numerical solution of ordinary differential equations

SYNOPSIS
     ode [ options ... ]

DESCRIPTION
     Ode is a language designed for  solving  the  initial  value
     problem  for  a  family of first-order ordinary differential
     equations.  Three distinct  methods  are  available:  Runge-
     Kutta-Fehlberg  (default),  Adams-Moulton  and  Euler.   The
     Adams-Moulton and Runge-Kutta routines  are  available  with
     adaptive step size.

     Ode reads its program from the standard input.  An ode  pro-
     gram  -- or simply an 'ode' -- consists of a list of expres-
     sions for  evaluating  the  derivatives  of  functions,  and
     optional control statements.

     The command-line flags are:

     -f file
          Read file before reading the standard input.

     -A [step], -R [step], -E [step]
          Use the Adams-Moulton, Runge-Kutta-Fehlberg,  or  Euler
          method,  respectively.  Only one of these may be given.
          The optional step causes the method to be applied  with
          a constant step size.

     -h hmin [hmax]
          Indicates a lower and an optional upper  bound  on  the
          step  size.   The  default  lower  bound is the machine
          limit.

     -r rmax [rmin]
          Sets a  ceiling  for  the  relative  single-step  error
          (default  is  1e-8).   The adaptive methods monitor the
          error to  adjust  the  step  size.   If  the  limit  is
          exceeded  in  any variable, the step is abandoned.  The
          value rmin can be used to force the step size to grow.

     -e emax [emin]
          Similarly for the absolute single-step error.

     -s
          Suppresses the ceiling on single-step error so that the
          program continues if it reaches these limits.

     -p prec
          Values are printed to prec significant figures.




Sun Release 4.1     Last change: Reed College                   1






ODE(1)                   USER COMMANDS                     ODE(1)



     -t
          A title line names the columns.

DIAGNOSTICS
     Mostly self-explanatory.  The biggest exception  is  'syntax
     error',  meaning  there  is  a  grammatical error.  Language
     error messages are of the form

          ode: nnn: message...

     where 'nnn' is the  location  of  the  line  containing  the
     error.   When  the -f flag is used, the phrase '(file)' fol-
     lows the 'nnn' for errors encountered inside the file.  Sub-
     sequently,  when ode begins reading the standard input, line
     numbers start over from one.

     Run-time errors elicit a message describing the problem  and
     the solution is abandoned.

SEE ALSO
     graph(1)
     "Ode User's Manual"

BUGS
     No effort is made to  recover  successfully  from  syntactic
     errors  in  the input.  However, there is a meager effort to
     resynchronize so more than one error can  be  found  in  one
     scan.
























Sun Release 4.1     Last change: Reed College                   2






End of Ode.manpage
\n#==========================
#= Here is file Ode.manual =
#==========================\n
echo Ode.manual 1>&2
cat >Ode.manual <<'End of Ode.manual'









     Ode - Numerical Solution of Ordinary Differential
                         Equations


                      Nick Tufillaro*
                       Graham A. Ross

                        Reed College
                   Portland, Oregon 97202



                          ABSTRACT

          Ode solves the initial value  problem  for  a
     family  of first-order differential equations when
     provided with  an  explicit  expression  for  each
     equation.   Ode parses a set of equations, initial
     conditions, and control statements and  then  pro-
     vides  an efficient numerical solution.  Ode makes
     the initial value problem  easy  to  express;  for
     example, the ode program,

             # an ode to Euler
             y  = 1
             y' = y
             print y from 1
             step 0, 1

     prints 2.718282.

     The Ode User's Manual contains a guide to applying
     ode and a discussion of its design and implementa-
     tion.



1 January 1982

      Copyright Nicholas B. Tufillaro, 1982-1994.
      All rights reserved.




_______________________________________________________
* The first  author's  support  from  National  Science
Foundation  CAUSE  Grant No. SER 78-06405 is gratefully
acknowledged.

















     Ode - Numerical Solution of Ordinary Differential
                         Equations


                      Nick Tufillaro*
                       Graham A. Ross

                        Reed College
                   Portland, Oregon 97202



1.  Introduction

     Ode renders a numerical solution to the  initial  value
problem  for many families of first order differential equa-
tions.  It is a programming  language  which  resembles  the
mathematical  language so that the problem posed by a system
is easy  to  state,  thereby  making  a  numerical  solution
readily  available.  Ode solves higher-order systems since a
simple procedure converts an  nth'  order  equation  into  n
first  order  equations  (see  Section 3.2).  Three distinct
methods are  implemented  at  present:  Runge-Kutta-Fehlberg
(default),  Adams-Moulton  and Euler.  The Adams-Moulton and
Runge-Kutta routines are available with adaptive step size.

     The Ode User's Manual falls into two parts.  The  first
is   a  tutorial  which  begins  with  simple  examples  and
progresses to show how ode is used with other programs, not-
ably  graph(1)  and plot(1).  Examples also demonstrate that
ode cannot  be  used  blindly!   Correct  results  from  ode
presume  that the user is aware of certain information about
the particular system of equations, for instance  the  loca-
tion  of  any  singularities.   A discussion of interpreting
results and the rudiments of error analysis are also found.

     The second part begins at Section 5; it contains a con-
cise  statement  of  the linguistic and numerical aspects of
ode, and a discussion  of  its  design  and  implementation.
This  is the definitive statement of ode, with the exception
of the code itself.  This second  part  also  contains  some
information  which would be useful in maintaining or modify-
ing ode.

1.1.  What is the numerical solution of an ODE?

     First a blizzard of definitions.  A time honored defin-
ition  says a differential equation is an equation involving
an unknown function and  its  derivatives.   A  differential
equation is ordinary if the unknown function depends only on
one independent variable, often denoted 't'.  The  order  of









                           - 2 -


the  differential equation is the order of the highest-order
derivative in the equation.  One speaks of a family or  sys-
tem  of  equations  when more than one equation is involved.
If the equations are dependent on one another they are  said
to  be  coupled.   A solution is any function satisfying the
equations.  An initial value problem is present  when  there
exist  subsidiary conditions on the unknown function and its
derivatives, all of which are given at the same value of the
independent  variable.  This "initial condition" specifies a
unique solution.  Questions about the existence and  unique-
ness  of  a  solution,  along  with further terminology, are
found in any introductory text (cf.  Birkhoff/Rota,  Chapter
1).

     In practical problems the solution  of  a  differential
equation is usually not known in terms of familiar functions
or it is, as the buzzword goes, intractable.  Hence the need
for a numerical solution.

     A numerical method for a system of  ordinary  differen-
tial  equations produces approximate solutions at particular
points using only the operations  of  arithmetic  and  func-
tional evaluations of the given equations.  Ode solves first
order initial value problems of the form:

      x' = f(t,x,y,...,z)
      y' = g(t,x,y,...,z)
         .
         .
         .
      z' = h(t,x,y,...,z)

given the initial values for each dependent variable at  the
initial value of the independent variable,

      x(a) = b
      y(a) = c
           .
           .
           .
      z(a) = d
         t = a

where a,b,c,...,d are constants.

     In the equations shown above f, g, ..., and  h  can  be
expressed  using  the  usual operators (+,-,*,/,^) and func-
tions listed in section 8.3.  In order to solve a system  of
equations  with ode, it must be possible to solve explicitly
for each derivative.  Ode can't solve an equation  in  which
the  first  derivative is implicitly defined.  For instance,
ode can't solve the system,

      y' + y' = 1.

     All numerical methods involve  the  calculation  of  an









                           - 3 -


approximate  solution  at discrete intervals of the indepen-
dent variable 't', where the difference between any two suc-
cessive steps (usually denoted 'h') may be constant or adap-
tive.  In general, as 'h'  decreases  the  solution  becomes
more  accurate.   Again,  the solution rendered by ode is an
approximation to the actual solution.

     Reflection upon the geometric fact expressed by a  sin-
gle  first  order  differential  equation  gives one insight
about how a method works.  The set of  all  solution  curves
expresses a vector field or flow.  A numerical solution sim-
ply samples the direction of  the  flow  in  small  discrete
steps  and  moves accordingly.  This is the heart of Euler's
method.

2.  Simple examples

     Let's begin with simple problems that  will  illustrate
the essential elements of composing an ode while not getting
bogged down in too many details.  The  entire  specification
for  the  language  may be found in Section 5; look there if
these examples are too elementary.

2.1.  Euler

     The first ode to compose is for  the  simplest  example
imaginable, to wit,

      y'(t) = y(t)

with the initial condition,

      y(0) = 1

The solution to this differential  equation  is  a  function
with  the property that the first derivative of the function
is equal to the function  itself.   But  that's  easy,  it's
just,

      y(t) = e^t,

and in particular,

      y(1) = e^1 = 2.718282

which is Euler's constant to seven places.

     A similar result is obtained with ode by  the  sequence
of instructions

        ode
        y' = y
        y  = 1
        print t, y
        step 0, 1

Two columns of numbers will appear.  The first is the  value
of the independent variable 't', and the second is the value









                           - 4 -


of the dependent variable 'y'.  The last line should be

        1 2.718282.

After this output ode will wait  for  further  instructions.
Entering

        step 1, 0

should bring you back to where you started,  more  or  less.
To  exit ode type a line containing only a '.', and then tap
return.  Your shell prompt should appear.   Ode  also  exits
when it sees an end-of-file on its command stream, which you
can send from the terminal by typing ctrl-d.

     Exercise 2.1.  Run this ode again, except leave out one
or  more  of  the lines.  What error messages do you get? If
the  program  works,  what  are  the  default  values.   For
instance, what does this program do?

        ode
        y = 1
        y' = y
        .

Don't forget the period and then tap return.

     Each line of the ode is pretty  much  self-explanatory.
The line

        ode

starts up ode, which is now  ready  to  receive  information
about the problem.  The line

        step 0, 1

is the step statement.  It sets the beginning and the end of
the  domain  for  the independent variable and causes ode to
perform the numerical step.  If it's missing, ode does noth-
ing  --  but  at least it does it right!  The first value of
the step statement (0) and the assignment statement

        y  = 1

are equivalent to the  initial  condition  y(0)  =  1.   The
statements  y'  = y and y = 1 are very different; y = 1 sets
the initial condition for y, which must  always  be  a  con-
stant,  whereas  y' = y defines a function.  Whenever a step
statement is encountered the  program  tries  to  produce  a
solution.   The  print statement tells what values are to be
printed and in what order; for instance,

        print y, y'









                           - 5 -


prints the value of the function y and its derivative.

     Since our solution of this problem is numerical  it  is
only  an approximation.  Unlike the exact solution, one must
step from the value of the initial condition of the indepen-
dent variable to the point desired.  The actual step, as you
no doubt noticed, takes  place  by  means  of  a  number  of
smaller  steps;  the step size, often denoted 'h', is chosen
to minimize the error.  The result is an  approximate  solu-
tion,  an  integral curve, on the interval given by the step
statement.

     Let's make a file containing the following:

        # an ode to Euler
        y  = 1
        y' = y
        print t, y, y'

Name this file euler.  The '#'  is  the  comment  statement,
which  may  appear  anywhere in an ode.  Everything from its
appearance to the end of the line is  ignored.   To  execute
the program say

        ode -f euler
        step 0, 1

and the file euler is read into ode.  Ode then waits for any
further statements.  Notice that

        ode < euler

is not equivalent to

        ode -f euler.

The former causes ode to get all its  input  from  the  file
euler,  while  the  latter  allows subsequent input from the
terminal.  (See "Introduction to the UNIX  Shell"  for  more
information on the subject of input/output redirection.)

     Exercise 2.2.  Ode can take command-line flags to  tell
it  to  do  something special.  The '-f' flag must always be
followed by a file name; it is a request to  read  the  con-
tents of the file into ode.  Use ode with the '-t' flag,

        ode -t

What does this do? How about "ode -p 10"?  "ode -r 1e-12"?

2.2.  Cosine

     Consider the differential equation and  initial  condi-
tions,









                           - 6 -



      y''(t) = -y(t)
       y'(0) = 0
        y(0) = 1

The solution is

      y(t) = cos(t)

     In order to  solve  this  problem  using  ode  we  must
express  this second order equation as two first order equa-
tions.  Toward this end  perform  the  following  change  of
variables. Let,

      yprime' = -y.
           y' = yprime

These two equations are the  same  as  the  single  equation
above since,

      yprime' = y'' = -y.

The reduction of an 'n-th' order problem to 'n' first  order
problems is a standard trick in this game which you will use
often.  It may be helpful to think of y' as velocity.   Make
a file containing

        # cosine : y''(t) = -y(t), y(0) = 1, y'(0) = 0
        cosine' = -sine
        sine' = cosine
        cosine  = 1
        sine  = 0
        print t, cosine

Name the file cosine.  Now type 

        ode -f cosine | graph | plot 
        step 0, 2*PI
        .

And sit tight for a few seconds.  A graph  of  the  solution
curve  will  soon  appear.   Documentation  for graph(1) and
plot(1) is found in the UNIX Programmer's Manual, Volume  1.
The '(1)' refers to the section number.

     Exercise 2.3.   Use  ode  to  solve  y''(t) = -y(t)  with
y(0) = 0, y'(0) = 1.

     Exercise 2.4.  While using a graphics terminal try:

        ode -f cosine | graph -m 0 | plot
        step 0, 2*PI
        .

Look up the '-m' flag in graph.  What does it  do?   Do  you
see how the step size changes?










                           - 7 -


     In the next section we will deal a bit more with how to
reduce  an 'n-th' order problem and with how to make a graph
of the solution curve.

     Ode knows a number of functions,  to  wit,  abs,  sqrt,
sin,  cos, atan, exp, ln, sinh, cosh, and tanh.  As you see,
'PI' is a synonym.  The names of the functions  listed
above  are reserved, so, for instance, we couldn't have used
'cos' and 'sin' for our variables in the previous example.

     Exercise 2.5.  Investigate the differential equation

      y'(t) = cos(y)

with ode for different initial conditions.

What  is   the   actual   solution?    Do   the   same   for
     
      y'(t) = cos(ty).

     Exercise 2.6.  Write down  eighteen  first  order  dif-
ferential equations for the infamous 'three body problem' of
physics and find 'stable' configurations using ode.

3.  Advanced Topics

     Section 3 highlights some of the finer points  of  ode.
However  this  section  along with the previous one does not
cover all of ode's capabilities.  Read Section 5  after  you
have gained some experience with ode to find out about them.

     Use a graphics terminal for the following exercises. 
Or "xterm" which is capable of tektronix's 4104 emulation.

3.1.  Use of Graph and Plot

     Look  up  and  read  the  documentation  for  graph(1),
plot(1),  and spline(1).  Graph and plot are general purpose
programs of great pliancy.  Graph indulges in copious flags.
Proper  use  of  them  allows one to display multiple graphs
with labels on a single screen.  We will focus on a  few  of
the features that often arise while using graph with ode.

     Graph has a nasty peculiarity that you should look  out
for:  it  cannot  graph  more  than about 2000 points.  This
defect may be circumvented by a careful  use  of  the  print
statement.  The general structure of the print statement is

        print <pr-list> [every <const>] [from <const>].

The bracket notation [], means that the enclosed  statements
are optional.  Up till now we have not used the every state-
ment and the from statement.  The '<pr-list>'  is  a  comma-
separated  list  of  variables  that occurs in the ode.  For
instance in










                           - 8 -



        print t, y, y' every 5 from 1

the <pr-list> is <t, y, y'>.  The statements 'every  5'  and
'from  1' do what you hopefully suspect.  In this case every
fifth value is printed and the printing will begin when  the
independent  variable  't',  is  equal to 1.  The use of the
statements every and from to dodge the 2000 point  limit  is
obvious.   The from statement is also handy where we want to
view only the last part of the solution curve.

     Exercise 3.1 Use the cosine file of Section 2.  Try:

        ode -f cosine | graph | plot
        print t, cosine from 2*PI
        step 0, 4*PI
        .

The default for the every statement is 'every 1'.  Next try:

        ode -f cosine | graph | plot
        print t, cosine every 3 from 2*PI
        step 0, 10*PI
        .

Try other print statements.  How big does the interval  have
to  be  until graph silently stops plotting points?  Can you
elude the limit with the every statement?

     A valuable graph flag is '-m'.  It stands for mode.  In
particular,

        graph -m 0

means print only points rather than connecting them to  form
lines.   It  is  possible  to draw multiple graphs with this
option.

     Exercise 3.2 Try:

        ode -f cosine | graph -m 0 | plot
        step 0, 2*PI
        print t, sine
        step 2*PI, 0
        .

The ability to draw multiple graphs is big-league.   It  can
be  used to construct a phase plane portrait and to check on
the stability of the solution.  More  about  this  a  little
later.  An ode may contain any number of print statements so
that you can draw multiple graphs.  The  most  recent  print
statement is used with each step statement.

     Exercise 3.3.  A Strange Attractor.  An  eloquent  plot









                           - 9 -


is given by the Lorentz system.

        # Lorentz model, a third order system
        # illustration of a "strange attractor"
        # interesting cases are r = 26, x = z = 0,
        #       y = 1, 2.5 < t < 30
        # and r = 17, x = z = 0,
        #       y = 1, 1 < t < 50, print from 15
        # Poincare plot is x vs. y
        # use with ode -f lorentz | graph -m 0 | plot
        r  = 26
        x' = -3 * ( x - y)
        y' = -x * z + r * x - y
        z' = x * y - z
        x = 0
        y = 1
        z = 0
        print x, y

In this example 'r' is a symbolic constant.  Put this ode in
a file named lorentz and construct the suggested plots.

3.2.  Solution of Higher Order Systems

     A system of differential equations  of  various  orders
can be transformed into a larger system of first order equa-
tions by systematically  introducing  new  variables.   This
process  of reduction is not unique, in general it is possi-
ble to achieve the reduction  in  many  ways;  nevertheless,
there  is  a systematic way to go about it.  Let's transform
the initial value problem

      x''' = cos(x) + sin(x') - ex'' + t^2
      x''(0) = 13
       x'(0) = 7
        x(0) = 3

into a form suitable for solution by ode.

  Old Var.   New Var.   Initial Value       Diff. Eq.
  t          t                0                t'=1
  x          x                3                x'=y
  x'         y                7                y'=z
  x''        z               13         z'=cosx+siny-e+t^2

     Exercise  3.4.   Solve  this  system  using  ode.   The
expression  't^2'  is valid for ode so you don't need to say
't*t'; in general ode knows 't^u', to raise t  to  the  u-th
power.

     The first order equations are  often  significant.   In
physical  problems  they  may  be  Hamilton's  equations  of
motion.








                           - 10 -


     Exercise 3.5.  Reduction is not unique.  Newton's equa-
tion of motion for a simple pendulum is

      m l^2 a'' + m g l sin(a) = 0
where 'l' is the length of the string and 'a' is the  angle.
Reduce  this  to  two  first  order  equations by the method
illustrated above.  Do this before proceeding  to  the  next
paragraph if you know what's good for you.

     The solution is

(NOTE: DUE TO FORMATING DIFFICULITIES IN ASCII CONVERSION
THERE MAY BE SOME ERRORS IN THE MATHEMATICAL SECTIONS OF 
THE TEXT. NBT: 4 SEPTEMBER 1994. A PROPERLY TYPESET MANUAL
MAY BE SENT TO YOU BY DIRECTLY CONTACTING NICK: nbt@reed.edu).

      a' = w
      w' = l - g sin(a)

What is the  physi-
cal meaning of both sets of equations?  How are they related
(what's C)?  Chose reasonable initial conditions  and  solve
both sets of equations with ode.

3.3.  Phase Plane Construction

     The construction of a phase plane illustrates  how  ode
may  be used in an interactive fashion.  The idea is to draw
a set of integral curves from different  initial  conditions
in the quickest manner possible.

     If we have two first order equations  then  the  family
of all solution curves is called the 'phase portrait'; it is
of paramount interest in the qualitative theory of differen-
tial equations and also possesses an aesthetic appeal.

     In practice the phase portrait  is  drawn  from  a  few
well-chosen solution curves.  Picking a good set of solution
curves may take some experimentation.  The construction   of
a  phase  portrait is illustrated in an example from popula-
tion biology.

     Exercise  3.6.   A  predator-prey  problem   based   on
Volterra's  model.  In a lake there are two species of fish:
A (the prey) who  live  by  eating  a  plentiful  supply  of
plants,  and  B  (the  predator) who eat A.  Let x(t) be the
population of A and y(t) the population of B.  A crude model
for the interaction of A and B yields the equations of popu-
lation:

      x' = x(a-by)
      y' = y(cx-d)

for a, b, c, d > 0 and (0,0) = (b/d,a/c).









                           - 11 -


It can be shown that if the solution curves are closed,  the
fluctuations  of  x(t)  and  y(t), starting from any initial
population (greater than zero), are  periodic.   To  draw  a
phase portrait try:

        ode | graph -m 0 | plot
        # an example from population biology
        # construction of a phase portrait for
        # Predator-prey equations of Volterra and Lotka
        x' = (a - b*y) * x
        y' = (c*x - d) * y
        a = 1; b = 1; c = 1; d = 1;
        print x, y
        x = 1
        y = 2
        step 0, 10
        x = 1
        y = 3
        step 0, 10
        x = 1
        y = 4
        step 0, 10
        x = 1
        y = 5
        step 0, 10
        .

The plot  will  not  begin  until  the  terminating  dot  is
entered.

3.4.  Obscurata

     There are a few frills that  can  come  in  handy  when
things  aren't  going quite right.  For example, The examine
statement can be  used  to  discover  pertinent  information
about  any  variable in a system.  It's described in section
8.2

     The print statement can be used to print out more  than
just  the  value of a variable.  If the name of the variable
is followed by a "'", the  derivative  of  the  variable  is
printed instead.  Following the variable name with "~", "!",
or "?" prints the accumulated error (not implemented in  the
current  version of ode), the absolute single-step error, or
the relative single-step error, respectively.

4.  Error

     Be warned, this section is  incomplete.   Entire  books
exist  on any single subject mentioned (e.g., floating point
error).  The goals of this section  are  modest;  first,  to
introduce  the  most basic notions of error analysis as they
relate to ode; second, to steer you around a few of the more
obvious  pitfalls.   Look  through a numerical analysis book









                           - 12 -


before beginning this section (cf. Atkinson).

4.1.  Basics

     A cornucopia of definitions whose relevance  will  soon
be  apparent.   The error of greatest concern is the differ-
ence between the actual solution and the  approximate  solu-
tion;  this is termed the global or accumulated error, since
the error is built up during each step.   Unfortunately,  an
approximation of this error is usually not available without
knowledge of the actual solution or some hard core numerical
analysis.   Thus  there  arise  a  number  of  more workable
notions of error.  The local or single  step  error  is  the
difference  between  the actual solution and the approximate
solution at any single  step,  assuming  the  value  at  the
beginning of the step is correct.

     The relative error is the error divided by the approxi-
mate  value.   Why  not the actual value?  Well, this is not
known, so it would be a useless definition.  Ode, if left to
itself,  will  track the relative single step error in order
to pick the step size.  By default, it maintains an accuracy
of  eight  digits  on any single step.  This accuracy can be
adjusted with the '-r' flag.

     The sources of error are generally twofold.  The  first
is  due to the limits of machine computation.  All computers
work  with  floating-point  numbers,  which  are  not   real
numbers, but only an approximation to real numbers.  You can
see the difference between the two by using the '-p 17' flag
which will print 17 significant digits.  This floating point
error is propagated through each step.  However, all  calcu-
lations  made  by  ode are done to double precision, so this
tends to be a small source of error.

     The second source of error is often called the theoret-
ical  truncation  error;  it  is  the difference between the
actual solution and the approximate solution due  solely  to
the numerical method.  At the root of many numerical methods
is an infinite series; for ordinary differential  equations,
it  is  a Taylor expansion.  Since the computer can not take
all the terms in an infinite series, a numerical method uses
a truncated series; hence the origin of the term.  The accu-
mulated error is a sum of the theoretical  truncation  error
and  the  floating  point  error,  although  in practice the
floating point error is seldom included.  Some authors  also
term  this  the  actual  truncation  error.  The single step
error for ode consists only of  the  theoretical  truncation
error.

     No term in numerical analysis is more  democratic  than
stability.   We  will say a numerical method is stable (when
applied to  a  particular  system)  if  the  relative  error
decreases as the step size decreases.  Both the '-R' and the









                           - 13 -


'-A' algorithms are stable (based more on empirical  results
than theoretical) for a wide class of systems.

4.2.  Pitfalls

     Here are some common sources of trouble  to  watch  out
for.

4.2.1.  Singularities

     Look up the definitions  of  ordinary  point,  singular
point,  regular singular point, and irregular singular point
in your favorite differential equations text.  If you  don't
have  a  favorite,  try  mine, Birkhoff and Rota, Chapter 9.
Ode should not be asked  to  step  over  singularities.   It
should not be applied to singularities except possibly at an
endpoint.  Always attempt to locate  and  classify  all  the
singularities for a system.

4.2.2.  Ill-Posed  Problems

     Be careful when chosing the domain (step interval)  for
a  problem.   You must know that the system and the solution
have a well-defined value on the domain.  The solution  must
also  be  real.  Ode cannot pole-vault complex singularities
from the real line.  Whenever you get weird  results,  e.g.,
the  step  size suddenly shrinks to the machine limit or the
solution suddenly blows up, check to make sure  the  problem
is  not  ill-posed.  Keep your eyes peeled for those complex
poles.

     As an example of an  undefined  solution  consider  the
innocent looking system:

      y' = y^2
    y(1) = -1

The solution for t > 0 is

      y(t) = t-1

With this problem you must avoid any step statement which is
in  the  neighborhood of t = 0.  Try your luck with the step
statement

        step 1, -1

for this system.  How does ode react?

     As a second example consider a system  which  is  unde-
fined for y = 0:

  

The general solution is

    









                           - 14 -


and if y(2) = 2 then

      
Clearly any step statement which  includes  negative  values
will  produce  bogus  results.  Also ode will not be able to
evaluate the system at any value for which y =  0.   Try  to
solve  this system for some bogus intervals using ode.  When
using a constant step size be careful not to to 'step  over'
such  trouble  areas.  Ode, with an adaptive step size, will
often spot such problems, but not always.  Caveat emptor.

4.2.3.  Critical points

     An autonomous system is one that does not  include  the
independent variable explicitly in the derivatives.  A crit-
ical point for a system is a point such that each derivative
of the family is exactly zero.  For instance the system

      y' = 2x
      x' = 2y

has only one critical point at (x,y) = (0,0).

     A critical point is sometimes referred to as a  stagna-
tion  point.   This  is because a system at a critical point
will stay there forever, although a system near  a  critical
point can undergo more violent action.

     Exercise 4.1.  Solve the system above  using  ode  with
the  initial condition x(0) = y(0) = 0.  The solution should
be a single point.  Now do the same  with  points  near  the
critical point.  What happens?

     Exercise 4.2.  Find all  the  critical  points  of  the
Lorentz system mentioned in exercise 3.3.

     Always locate all  the  critical  points  of  a  system
before attempting a solution with ode.

     Critical points are classified (e.g., equilibrium, vor-
tex,  unstable,  stable,  etc.) and this knowledge may be of
use; to find out more about this look in a book dealing with
the   qualitative  theory  of  differential  equations  (cf.
Birkhoff/Rota, Chapter 6).

4.2.4.  Unsuitable Numerical Method

     Lastly, if the results are bad, it just might  be  that
the  numerical  method  is not suitable to the problem.  For
instance, ode at present has no numerical method which  han-
dles the so-called 'stiff' problems very well.

     Exercise 4.3.  Examine the 'stiff system':











                           - 15 -




for 0 < t < 10.  The exact solution is

Solve with ode using various methods, step sizes, and  rela-
tive  single step error tolerances.  Compare with the actual
solution.

4.3.  Checks on Error

     This section provides some rough  and  ready  heuristic
checks  on  the accuracy of the solution.  This is what they
really do.

4.3.1.  Examine stability of solution curves: do  they  con-
verge?

     Use of ode for any serious study should include a check
on  stability.   Always  check how changing the step size or
the tolerance (with the  '-r'  flag)  affects  the  solution
curve.   As  the  step size decreases, the curve should con-
verge.  If it does not then either  the  step  size  is  not
small  enough  or  the  numerical  method is unsuitable with
respect to the problem.  So apply the following check.

     If using an adaptive step size, superimpose  the  solu-
tion curves for various tolerances (e.g., -r 1e-9, -r 1e-11,
-r 1e-13, ...).  If the curves converge then the problem  is
to all appearances stable and your accuracy sufficient.

     If employing a constant step size, then perform a simi-
lar analysis by halving the step.

     Exercise 4.4.   A  Q.C.D.  equation,  courtesy  of  Ray
Mayer.  Make a file named qcd containing:

        # an equation arising in quantum theory
        f'   = fp
        fp'  = -f*g^2
        g'   = gp
        gp'  = g*f^2
        f  = 0
        fp = -1
        g  = 1
        gp = -1
        print t, f
        step 0, 5

Next make a file named stability containing:












                           - 16 -



        : sserr is relative single step error limit
        for sserr
        do
           ode -r $sserr < qcd
        done | graph -m 0 | plot

The file stability is a shell command file.  It is a program
that  will  superimpose  curves of different tolerances.  To
execute the program type:

        sh stability 1 1e-1 1e-2 1e-3 1e-4 1e-5

and a graph for each tolerance will be drawn.

4.3.2.  Check constants of the system: are they constant?

     Many  systems  will  have  invariant  quantities.   For
instance,  if  the problem is a conservative physical system
then the total energy (according to the  latest  theory)  is
constant.   Knowledge  about the qualitative behavior of any
variable can be used to check the quality of the solution.

4.3.3.  Check a family of solution curves: do they diverge?

     A rough idea of how error is propagated is obtained  by
viewing  a  family  of solution curves about the solution in
question.  If they diverge sharply - that is, if  two  solu-
tions  start  out  very  close  and end far apart - then the
quality of the solution may be dubious.  On the other  hand,
if  the curves do not sharply diverge then any error will in
all likelihood not increase by more than an order of  magni-
tude.  Such problems are sometimes called well-conditioned.

5.  Using Ode

     This section describes how to  use  ode  on  UNIX.   It
describes  the diagnostic messages produced and the command-
line arguments.

5.1.  Command Line Arguments

     The prototypical ode command line is:
     ode [-f file] [-ARE [step]] [-h hmin [hmax]]  [-r  rmax
     [rmin]] [-e emax [emin]] [-s] [-p prec] [-t]

The following text makes reference to this model  in  naming
the arguments on the command line.

     Ode reads its program from the standard input.  A piece
of an ode can be stored in a file.  If the -f flag is given,
then the file is read before the standard input.

     Only one of the -A, -R, or -E flags may be given.  Each
flag specifies a different numerical method.  If an optional









                           - 17 -


step value is present then a constant step size is employed,
instead of an adaptive one.

     If -E is specified, then  a  'quick  and  dirty'  Euler
scheme  is  used.   The default step value for Euler runs is
0.1.  Not recommended for serious applications.

     If -R is specified,  then  a  fifth-order  Runge-Kutta-
Fehlberg  algorithm  is  employed with an adaptive step size
(default).  When a constant step size and no error  analysis
are  requested,  then  a  classical fourth-order Runge-Kutta
scheme is used.

     If -A is specified with  or  without  a  step,  then  a
fourth-order Adams-Moulton predictor-corrector is used.  The
Runge-Kutta-Fehlberg algorithm is used to  get  past  values
when needed.

     The -h flag indicates a lower  and  an  optional  upper
bound  on the step size.  The numerical routine will not let
the step size go below the value specified  for  hmin.   The
default is to allow the step to shrink to the machine limit.
The value hmax, if given, specifies the maximum value of the
step size.  It is useful in preventing the numerical routine
from skipping quickly over an interesting region.

     The -r  argument  sets  a  ceiling  for  the  'relative
single-step  error'.   The  -e  flag  is  for  the 'absolute
single-step error'.   The  relative  single-step  error  may
never  exceed  rmax (default is 10^-9) in any dependent vari-
able.  If  one  does,  the  solution  is  abandoned  and  an
appropriate error message is printed.  In a similar fashion,
an optional lower bound may be given  to  suggest  when  the
step  size  should be doubled (default is rmax/1000).  Solu-
tions using an adaptive step size  monitor  the  single-step
error  to  adjust  the  step  size.  Thus, a decrease in the
single-step error  will  cause  smaller  step  sizes  to  be
picked.

     The -s flag suppresses the ceiling on single-step error
so  that  the  program continues if it reaches these limits.
This can result in large numerical errors.

     The -p argument causes numeric results  to  be  printed
with  prec  significant figures.  If this flag is given, all
output is in scientific notation.

     The -t argument  prints  a  title,  naming  the  output
columns.   With  this  argument, the default print format is
scientific notation.

5.2.  Diagnostics

     Most messages from the  language  processor  are  self-









                           - 18 -


explanatory.  The biggest exception is 'syntax error', indi-
cating a grammatical error.  Language error messages are  of
the form

        ode: nnn: message...

where 'nnn' is the  location  of  the  line  containing  the
error.   When  the -f flag is used, the phrase '(file)' fol-
lows the 'nnn' for errors encountered inside the file.  Sub-
sequently,  when ode begins reading the standard input, line
numbers start over from one.

     An arithmetic exception arises when an illegal calcula-
tion  is  attempted.  All versions of ode recognize the fol-
lowing exceptional conditions: square  root  of  a  negative
number,  logarithm  of  a  non-positive number, and negative
number to a non-integer power.  In addition, the PDP-11  and
VAX  versions of ode discriminate among the hardware faults,
giving: division by zero; floating point overflow (a  finite
number becomes indistinguishable from infinity), and, in the
case of the PDP-11, floating point underflow (a number which
should not vanish becomes indistinguishable from zero) where
simpler  machines  report  simply  "arithmetic   exception".
Because of a quirk in the VAX trap architecture it turns out
to be  impractical  to  diagnose  floating-point  underflow.
When a calculation underflows, no diagnostic is produced and
the result is zero.

     Run-time errors include conditions in  which  an  error
ceiling is exceeded, as well as arithmetic exceptions.  When
one occurs, a message describing the problem is emitted  and
the  solution  is  abandoned.   If appropriate, an effort is
made to indicate which dynamic variable was being calculated
at the time of the exception.  A typical message is

        ode: real number too large while calculating y'

which indicates that some number involved  with  calculating
the  derivative  of  y exceeded the machine's size limit for
floating-point numbers (about 1.7x10^38 on the PDP-11 and the
VAX).

6.  Limits

6.1.  Space Limitations

     The data space used in solving an ode consists of  both
fixed  and  dynamic  regions.  The fixed space is used effi-
ciently regardless of the complexity of  the  problem  being
solved.   The  dynamic  space contains the symbol table, the
expression lists, and several other  data  structures  whose
sizes  vary.   The space for these structures comes out of a
common pool; they play against each other.










                           - 19 -


     Every identifier in an ode  requires  230  bytes;  this
space  can never be reclaimed.  Each operator in an equation
costs 14 bytes.  For the sake of speed, some  operators  are
expanded  into  a  sequence  of  simpler operations, each of
which requires 14 bytes.  In particular, some common  powers
are  reduced  to  sequences of square roots, squares, cubes,
and inversions.  The extent of these transformations can  be
seen  with  the help of the examine statement.  The operator
space for an equation is reclaimed whenever a  new  equation
is defined for that particular dynamic variable.  Each lexi-
cal token in the most complex statement costs 10 bytes,  but
this  space  is  reclaimed  during the scan.  Thus, the most
complex statement circumscribes this requirement.  Each ele-
ment  in a print statement costs 6 bytes.  Upon execution of
the next print statement this space is reclaimed.  The total
space  available  for  these  dynamic  structures amounts to
about 45,000 bytes on the PDP-11.  The limit is much  higher
on the VAX.

     In addition to these absolute  space  limitations,  the
maximum  depth  of  the  operand  stack  used for evaluating
derivatives is thirty cells.  This limit  could  conceivably
be  exceeded  for a very complex differential equation.  The
solution is to simplify the equation if possible, and  where
not, to recompile ode with a larger operand stack.

6.2.  Speed Limitations.

     The time required to solve an ode depends  on  a  great
many  factors.  A few of them are: number of equations, com-
plexity of equations (number of operators and nature of  the
operators),  and  number  of steps taken (a very complicated
function of the difficulty of solution, unless constant step
sizes  are  used).  The most effective way to gauge the time
required for solution of a system is to  clock  a  short  or
imprecise run of the problem and reason as follows: the time
required to take two steps is roughly  twice  that  required
for  one;  and there is a relationship between the number of
steps required and the relative error ceiling chosen.   That
relationship  depends  on  the method being used, the diffi-
culty of solution, and perhaps on the magnitude of the error
ceiling  itself.   A few carefully planned short runs can be
used to determine this relationship,  enabling  a  long  but
imprecise run to be used as an aid in projecting the cost of
a more precise run over the same region.  Lastly, if a great
deal of data is printed -- especially on a terminal -- it is
likely that more  wall-clock  time  is  spent  printing  the
results than solving the problem numerically.

7.  Internal Design

7.1.  Language Implementation Details

     The input language is implemented with a Yacc-generated









                           - 20 -


parser  and  a Lex-generated lexical analyzer.  The benefits
of this approach are great.  The basic recognizer  was  com-
pleted  in  about  two  days.   And  adding  comments to the
language required that  exactly  one  line  of  the  lexical
analyzer source code be changed.

     Differential equations are compiled into  linked  lists
of  stack  operations,  one  list per dynamic variable.  The
function evaluator (eval)  traverses  one  of  these  lists,
applying  the indicated operations to a small operand stack.
At the completion of the traverse, the new derivative is  on
top of the stack.  The routine pops and returns this result.
Another routine (field) is called by the numerical  routines
whenever  a  re-evaluation  of the gradient is required.  It
calls eval for each dynamic variable in turn.

     Changes to the language involve  modifications  to  the
yacc  grammar,  the  lex  rules, and/or the semantics stored
with the grammar and rules.  The steps required to add a new
built-in  function, for example, are (aside from writing the
function itself):

(1)  Invent a new terminal  symbol  name  for  the  reserved
     word.

(2)  Add the new reserved word to the lex rules.

(3)  Add the new  productions  to  the  right-hand  side  of
     "expr"  in the grammar, together with whatever semantic
     action is required (using the included macros if possi-
     ble).

(4)  Define a new value for exper and add the  stack  code
     to eval.

(5)  Run make.

7.2.  Numerical Methods

     The numerical and language procedures were designed  in
accordance  with the Jesuit maxim.  The major elements mani-
pulated by both are the  symbol  table,  and  the  functions
field  and  printq.   Addition  of  a  new numerical routine
requires that the numerical routine be written, and  perhaps
that  some  fields inside the symbol table entry be expanded
or added.  None of these changes affect the  performance  or
compromise  the integrity of the parser, although they might
change the space-cost of  a  dynamic  variable.   Similarly,
modifications  to  the  lexical analyzer or the grammar only
affect the parse tables and related  data  structures  which
have no effect on the numerical routines.

     At each step the numerical routine  fills  each  symbol
table  entry  with  relevant  values.   Then a call to field









                           - 21 -


evaluates every differential equation and puts the result in
the  symbol  table  element  syrime.   The  evaluator gets
current values for the dynamics  from  the  field  syalue.
Both syalue and syrime are in some sense temporary vari-
ables.  All correct results eventually find their  way  into
syal[0]  and  syri[0];  these  are  never touched by the
language processor.  Thus the  real  interface  between  the
language  and  numerical  procedures  is merely syalue and
syrime.  It is important to maintain meaningful values  in
syalue  and  syrime,  since  a  floating-point exception
might disrupt the control flow at any moment.  It is  recom-
mended that the pointer fsp be used in traversing the symbol
table, so that arithmetic exceptions  can  be  diagnosed  as
accurately as possible.

     Fourth and fifth order methods were chosen as the  best
compromise  between  speed  and accuracy.  The adaptive step
size has proved a boon in both areas.

7.3.  Details of Error and Fault Recovery

     Command-line errors  are  diagnosed  at  initialization
time and are generally fatal.

     Errors in the input stream  are  diagnosed  during  the
lexical  scan  or at parse time.  They elicit an appropriate
message.  The error recovery method used in  the  parser  is
the  vanilla  Yacc  scheme.   (see  S.  Johnson,  "Yacc: Yet
Another Compiler-Compiler", 1978, pp. 16-18).

     Run-time errors are detected by the function evaluator,
the numerical routines, or the floating-point unit hardware.
In each case, a message is printed describing the error  and
ode  returns  to  a known state awaiting input.  The current
step is abandoned, but all variable values are left reflect-
ing  the  state of the machine just prior to the fault.  All
of the numerical routines scan the  symbol  table  computing
values  for  each  dynamic  variable, using a common pointer
known to  the  error  recovery  routines  named  fsp.   This
pointer  provides a way of identifying the offending dynamic
variable when a trap  occurs.   A  small  assembly  language
assist  allows convenient setup and polling of the floating-
point unit status words on the PDP-11.  The same function is
performed on the VAX in C.

8.  The Ode Input Language

     This section provides a concise summary  of  the  input
language to ode.

8.1.  Formal Grammar

     The following is a formal description of the grammar of
ode's  input language.  The notation is BNF, for Backus-Naur









                           - 22 -


form (cf. Wirth).  Non-terminal symbols in the  grammar  are
enclosed in angle brackets.  Terminal tokens are in all cap-
ital letters and are defined below.  Bare words and  symbols
stand for themselves.

<program>    ::=        ... empty ...
               |  <program> <statement>


<statement>  ::=  SEP
               |  IDENTIFIER = <const> SEP
               |  IDENTIFIER ' = <expression> SEP
               |  print <printlist> <optevery> <optfrom> SEP
               |  step <const> , <const> , <const> SEP
               |  step <const> , <const> SEP
               |  examine IDENTIFIER SEP


<printlist>  ::=  <printitem>
               |  <printlist> , <printitem>


<printitem>  ::=  IDENTIFIER
               |  IDENTIFIER '
               |  IDENTIFIER ?
               |  IDENTIFIER !
               |  IDENTIFIER ~


<optevery>   ::=        ... empty ...
               |  every <const>


<optfrom>    ::=        ... empty ...
               |  from <const>


<const>      ::=  <expression>


<expression> ::=  ( <expression> )
               |  <expression> + <expression>
               |  <expression> - <expression>
               |  <expression> * <expression>
               |  <expression> / <expression>
               |  <expression> ^ <expression>
               |  FUNCTION ( <expression> )
               |  - <expression>
               |  NUMBER
               |  IDENTIFIER













                           - 23 -


8.1.1.  Operator Precedences

The grammar above is ambiguous.  This table  summarizes  the
precedences  and associativities of operators within expres-
sions.  Precedences decrease from top to bottom.

         Class        Operators   Associativity
         
         Exponential          ^           right
         Multiplicative      * /          left
         Additive            + -          left


8.2.  Statement Semantics

     There are five kinds of statements:

        IDENTIFIER ' = <expression>

This  defines  a  first-order  differential  equation.   The
derivative  of  the IDENTIFIER is given by the <expression>.
If a dynamic variable doesn't appear on the left side  of  a
statement  of  this  form,  its  derivative is assumed to be
zero.  That is, it is a symbolic constant.

        IDENTIFIER = <const>

This sets the initial value of the IDENTIFIER to the current
value of the <expression>.  Dynamic variables which have not
been initialized are set to zero.

        step <const> , <const>
        step <const> , <const> , <const>

These statements describe the behavior  of  the  independent
variable  and  cause  the  numerical routine to be executed.
The first <const>  is  the  independent  variable's  initial
value.   The second is its final value.  The third is a step
size; if given, it overrides the one given  on  the  command
line  (see  Section 5).  Usually, the step size isn't speci-
fied and it varies as the  computation  proceeds.   However,
the  step  always terminates with each variable equal to the
correct final value.  Another  step  statement  may  follow,
after  changing  any  equation, initial condition, or state-
ment.

        print <printlist> [ every <const> ] [ from <const> ]

This statement controls the content  and  frequency  of  the
numerical  output.   A <printlist> is a comma-separated list
of IDENTIFIERs, where any of the IDENTIFIERs might  be  fol-
lowed  by an apostrophe, denoting the derivative, or a ques-
tion mark, denoting the relative single step  error,  or  an
exclamation  mark,  denoting the absolute single step error,
or a tilde, denoting the accumulated error (not  implemented







                           - 24 -


at  present).  The specified values are printed in the order
they are found.  Both the every clause and the  from  clause
are  optional.   If  the every clause is present, a printing
occurs every <const> iterations of the numerical  algorithm.
The default is to print on every iteration (i.e. 'every 1').
The first and last values are always printed.  If  the  from
clause  is  present  it  means  to  begin  printing when the
independent  variable  reaches  or  exceeds  <const>.    The
default  is  to  print starting with the initial conditions.
If there is no print statement, then the  independent  vari-
able  and  all  dependent  variables which have differential
equations associated with them are printed.  The independent
variable is printed first; the dependent variables follow in
the order their equations were given.

        examine IDENTIFIER

A table of interesting information about the named  variable
is  printed  on  the  standard output when this statement is
executed.  For example, if the statement

        examine y

were encountered after execution of the example given in the
abstract of this paper, the output would be:

        "y" is a dynamic variable
        value:2.718282
        prime:2.718282
        sserr:1.121662e-09
        aberr:3.245638e-09
        acerr:0
        code:   push "y"

The phrase dynamic variable simply means  that  there  is  a
differential  equation  describing  the  behavior of y.  The
numeric fields are:

value   Current value of the variable.

prime   Current derivative of the variable.

sserr   Relative single-step error for the last step taken.

aberr   Absolute single-step error for the last step taken.

acerr   Total error accumulated during the most recent  step
        statement.  Not implemented at present.

     The "code" section lists the stack operations  required
to  compute the derivative of "y" (somewhat reminiscent of a
reverse Polish calculator).  The  information  may  come  in
handy  in  discovering  whether  the precedences in the dif-
ferential equation statement were interpreted  correctly  or









                           - 25 -


in  determining  the  time- or space-expense of a particular
calculation.  "Push y" means to load y's value on the stack,
which  is  all  that's required to compute its derivative in
this case.

8.3.  Lexical Issues and Definition of Terminal Tokens

     In the input language, upper-  and  lower-case  letters
are  distinct.   Comments  begin  with the character '#' and
continue to the end of the line.  Long lines may be  contin-
ued  onto  a  second  line  by  ending the first line with a
backslash  ('\').   The  combination  backslash-newline   is
equivalent to a space.  Spaces or tabs are required whenever
they are needed to separate identifiers, numbers,  and  key-
words  from  one  another.   Except  as separators, they are
ignored.

     The terminal tokens referred to by the grammar in  Sub-
Section 1 are described here.

FUNCTION
     One of the words: abs, atan, cos, cosh, exp,  ln,  sin,
     sinh,  sqrt,  tanh.   These  do  just what you probably
     expect.  All angles are expressed in radians.  The atan
     function gives a value between -J/2 and J/2.

IDENTIFIER
     A sequence of alphanumeric characters starting with  an
     alphabetic  character.   The first eight characters are
     significant.  Upper- and lower-case  letters  are  dis-
     tinct.   In identifiers, the underscore (') is alpha-
     betic.  Function names and keywords cannot be  used  as
     identifiers, nor can 'PI'.

NUMBER
     A non-empty sequence of digits  possibly  containing  a
     decimal point and possibly followed by an exponent.  An
     exponent is 'e' or  'E',  followed  by  an  (optionally
     signed)  one- or two-digit number.  All numbers and all
     parts of numbers are radix ten.  A number  cannot  con-
     tain  any  white  space.   The  special  word 'PI' is a
     number equal to not two.

SEP
     A semi-colon or a (non-escaped) new-line.

8.4.  The Independent Variable

     The independent variable is identifiable  as  the  only
variable which never appears on the left side of a differen-
tial equation or an initial value assignment.  If  there  is
more than one such variable, an error message is printed and
the step is abandoned.  If there  is  none,  an  independent
variable   is   invented   and  given  the  impossible  name









                           - 26 -


'(indep)'.

9.  References

Atkinson, K., An Introduction to  Numerical  Analysis,  John
Wiley,  New York, 1978.  Pages 380-384 contain a nice review
of literature on the numerical  solution  of  ordinary  dif-
ferential equations.

Birkhoff, G., Rota,  G.,  Ordinary  Differential  Equations,
Ginn and Company, New York, 1962.

Wirth,  N.,  Algorithms  +  Data  Structures   =   Programs,
Prentice-Hall, Englewood Cliffs, N.J., 1976.

Nicholas B. Tufillaro, Tyler Abbott, Jeremiah Reilly, An ex-
perimental approach to nonlinear dynamics and chaos, Addison-
Wesley, 1992.

10.  Appendix: Odies but Goodies


# chemical kinetics simulation
# A + 2B <==> C --> D
# rate constants are:
#       kf : A + 2B --> C
#       kb : C --> A + 2B
#       kd : C --> D

a' = kb*c - kf*a*b^2
b' = kb*c - kf*a*b^2
c' = kf*a*b^2 - kb*c - kd*c
d' = kd*c

c = 0
d = 0
a = 0.1
b = 1

kf = 1
kb = 1
kd = 1
print t,c every 10
step 0,10




















                           - 27 -



# Two similar pendula coupled with a horizontal spring.
# illustration of "beating"
# if the motion is undamped then the equations of motion
# are approximated by

xone' = vxone
xtwo' = vxtwo

vxone' = -g*xone/lone + k/mone * ( xtwo - xone)
vxtwo' = -g*xtwo/ltwo - k/ltwo * ( xtwo - xone)

k = 1
g = 9.8
lone = 5
ltwo = 5
mone = 1
mtwo = 1

xone = 0.0
xtwo = 0.3
vxone = 0
vxtwo = 0

print t, xone


# bead sliding on a smooth circular wire of radius a:
# there is viscous damping so the bead should home in
# on the equilibrium point.
# equation of motion is

g = 9.8 # acceleration due to gravity
a = 1   # radius of hoop
w = 10  # angular velocity of hoop
b = 1   # damping coefficient

the' = vthe
vthe' = (w^2)*sin(the)*cos(the) - (g/a)*sin(the) - b * vthe
the = 0.1
vthe = 0

print t, the


# nonlinear system with a limit cycle

yone' = yone + ytwo - yone*(yone^2+ytwo^2)
ytwo' = -yone + ytwo - ytwo*(yone^2+ytwo^2)

yone = 1
ytwo = 2
print yone, ytwo










                           - 28 -



# A disk dynamo simulation for the earth's magnetic field.
# an attempt to simulate the dramatic switch in the polarity
# every eon or so.
# cf. M. Steele, B.A. thesis, Reed College, 1981 (Physics)

A = 1
V = 1
Q = 14.625
S = 5

w' = Q - z * y - V * w
y' = S * ( A * z - y)
z' = w * y - z

w = 1
y = 1
z = 1

print t, z
step 0, 10


# A damped driven harmonic oscillator.
# y'' = -k/m * y - R/m * y' + cos(w*t)
# if R^2 > 4km, overdamped
# if R^2 = 4km, critically damped
# if R^2 < 4km, damped
# this equation reduces to two first order equations:

y'      = vy
vy'     = -k/m * y - R/m * vy + cos(w*t)
y       = 1
vy      = 0
k       = 1
m       = 1
R       = 0.5
w       = 2*PI

print t, y























                           - 29 -



# coupled oscillator : two pendula simulation
m = 1
M = 1.0625

a    = 0.5
adot = 0
l    = 10
ldot = 0
ldot' = (m*l*adot^2 - M*9.8 + m*9.8*cos(a))/(M + m)
l'    = ldot
adot' = -(9.8*sin(a) + 2*adot*ldot) / l
a'    = adot

print l, ldot every so often


# A rumor spreads through a closed population of
# constant size N+1.
# At time t the total population can be classified
# into three categories:
#       x persons who are ignorant of the rumor;
#       y persons who are actively spreading the rumor;
#       z persons who have heard the rumor but have
#       stopped spreading it;
# If two persons who are spreading the rumor meet
# then they stop spreading it.  The contact rate
# between any two categories is constant, u.
#       Show that the equations
#               x' = -u * x * y,
#               y' = u * ( x*y - y*(y - 1) - y*z)
# give a deterministic model of the problem.
# Find the equation of the phase paths and sketch
# the phase diagram.  Show that, when initially
# y = 1 and x = N, the number of people who
# ultimately never hear the rumor is s, where
#       2N + 1 -2s + N log( s/N) = 0

x' = -u * x * y
y' = u * ( x*y - y*(y-1) - y*( 100 + 1 - y - x))

y = 1
x = 100

u = 1

print t, x
















                           - 30 -



# planetary orbit simulation
# two suns situated at (0,0) and (-5,0)
# one planet starting out at (1,0)
# x and y are positions
# vx and vy are velocities

vx' = -x/((x^2+y^2)^(3/2)) -(x+5)/(((x+5)^2+y^2)^(3/2))
vy' = -y/((x^2+y^2)^(3/2)) -y/(((x+5)^2+y^2)^(3/2))
y' = vy
x' = vx

x = 1
y = 0

print x,y every 10

# these values seem to give a nice orbit:
# vx = 0
# vy = 1.142

# a more exciting result can be obtained from:
vx = 0
vy = 1.165




































End of Ode.manual
\n#==========================
#= Here is file READ_ME =
#==========================\n
echo READ_ME 1>&2
cat >READ_ME <<'End of READ_ME'
To compile on a Unix machine just run "make".
If this is a shell archive, you may need to
delete any mail headers and then: "sh file.sh".
More details can be found in the enclosed 
man page (Ode.manpage) or the User's Manual
(Ode.manual).

Program Files:

ode - numerical solution of differential equations

"make" to make an ode
"make clean" to clean up the mess
"make print" for a source listing
"make cp" to put ode in /usr/reed

am.c            Adams-Moulton with fixed step size
ama.c           Same with adaptive step size
eu.c            Euler quick-and-dirty method
expr.c          Expression evaluator
ext.c           External initializations
float.c         FPP setup and trap handler
fps.h           Floating point status definitions
gram.y          Grammar
lex.l           Scanner rules
mch.s           Machine support for PDP-11/70 version (FPU stuff)
misc.c          Mainline, I/O, etc.
num.h           Constants for the numerical routines
ode.h           Miscellaneous declarations and constants
prt.c           Print queue management
rk.c            Runge-Kutta with fixed step size
rka.c           Same with adaptive step size
stperr.c        Error handling for adaptive step routines
sym.c           Symbol table manager


Other Files:

x.NAME.example  Example files
Ode.manpage 	ASCII Man(1) Page
Ode.manual	ASCII Ode Manual (30 Pages). Some errors due to
		ASCII conversion in equation typesetting.
Makefile	File for making ode.

ode.sh          Shell archive of all the files here. To unbundle make
                a directory called "ode". 
		cd ode
		cp ode.sh to ode
		sh ode.sh
                ps. don't forget to remove any mail header file info before 
                    extracting

This code is Copyrighted by Nicholas B. Tufillaro, 1982-1994.
All rights reserved. It may not be used for any commercial
purpose with out the permission of the author.

Copyright Nicholas B. Tufillaro 1982-1994. All rights reserved.
                Internet: nbt@reed.edu

End of READ_ME
\n#==========================
#= Here is file am.c =
#==========================\n
echo am.c 1>&2
cat >am.c <<'End of am.c'
#include "ode.h"

#define PASTVAL        (3)     /* previous values, val[0] is current value */

/*
 * Adams-Moulton with constant step size
 * Copyright Nicholas B. Tufillaro, 1982-1994. All rights reserved.
 */
am()
{
        double t;
        double halfstep = HALF * tstep;
        double sconst = tstep / 24.0; /* step constant */
        double onesixth = 1.0 / 6.0;

        /* Runge-Kutta start up */
        for ( it = 0, t = tstart; it <= PASTVAL && !STOPR; t = tstart + ( ++it) * tstep) {
                symtab->sy_value = symtab->sy_val[0] = t;
                field();
                for( fsp = dqueue; fsp != NULL; fsp = fsp->sy_link) {
                        register j;
                        for ( j = it; j > 0; j--) {
                                fsp->sy_val[j] = fsp->sy_val[j-1];
                                fsp->sy_pri[j] = fsp->sy_pri[j-1];
                        }
                        fsp->sy_pri[0] = fsp->sy_prime;
                        fsp->sy_val[0] = fsp->sy_value;
                }
                /* output */
                printq();
                if ( it == PASTVAL)
                        break;  /* start-up complete */
                for ( fsp=dqueue; fsp !=NULL; fsp = fsp->sy_link)   {
                        fsp->sy_k[0] = tstep * fsp->sy_prime;
                        fsp->sy_value = fsp->sy_val[0] + HALF * fsp->sy_k[0];
                }
                symtab->sy_value = t + halfstep;
                field();
                for ( fsp=dqueue; fsp !=NULL; fsp = fsp->sy_link)   {
                        fsp->sy_k[1] = tstep * fsp->sy_prime;
                        fsp->sy_value = fsp->sy_val[0] + HALF * fsp->sy_k[1];
                }
                symtab->sy_value = t + halfstep;
                field();
                for ( fsp=dqueue; fsp !=NULL; fsp = fsp->sy_link)   {
                        fsp->sy_k[2] = tstep * fsp->sy_prime;
                        fsp->sy_value = fsp->sy_val[0] + fsp->sy_k[2];
                }
                symtab->sy_value = t + tstep;
                field();
                for ( fsp=dqueue; fsp !=NULL; fsp = fsp->sy_link)
                        fsp->sy_k[3] = tstep * fsp->sy_prime;
                for ( fsp=dqueue; fsp !=NULL; fsp = fsp->sy_link) {
                        fsp->sy_value = fsp->sy_val[0] + onesixth *
                                        ( fsp->sy_k[0]
                                        + TWO * fsp->sy_k[1]
                                        + TWO * fsp->sy_k[2]
                                        + fsp->sy_k[3]);
                }
        }
        /* predictor - corrector */
        while ( !STOPA)
        {
                /* Adams-Bashforth predictor */
                for ( fsp = dqueue; fsp != NULL ; fsp = fsp->sy_link) {
                        fsp->sy_value = fsp->sy_val[0] + ( sconst) *
                                        (55 * fsp->sy_pri[0]
                                        -59 * fsp->sy_pri[1]
                                        +37 * fsp->sy_pri[2]
                                        -9  * fsp->sy_pri[3]);
                }
                symtab->sy_val[0] = symtab->sy_value =  t = tstart + ( ++it) * tstep;
                field();
                /* Adams-Moulton corrector */
                for ( fsp = dqueue; fsp != NULL; fsp = fsp->sy_link) {
                        fsp->sy_value = fsp->sy_val[0] + ( sconst) *
                                        (9  * fsp->sy_prime
                                        +19 * fsp->sy_pri[0]
                                        -5  * fsp->sy_pri[1]
                                        +     fsp->sy_pri[2]);
                }
                field();
                /* cycle indices */
                for ( fsp = dqueue; fsp != NULL; fsp = fsp->sy_link) {
                        register j;
                        for ( j = PASTVAL; j > 0; j--) {
                                fsp->sy_val[j] = fsp->sy_val[j-1];
                                fsp->sy_pri[j] = fsp->sy_pri[j-1];
                        }
                        fsp->sy_val[0] = fsp->sy_value;
                        fsp->sy_pri[0] = fsp->sy_prime;
                }
                /* output */
                printq();
        }
}
End of am.c
\n#==========================
#= Here is file ama.c =
#==========================\n
echo ama.c 1>&2
cat >ama.c <<'End of ama.c'
#include "ode.h"
#include "num.h"

#define PASTVAL        ( 3)    /* past values, val[0] is current value */
#define T_LT_TSTOP (tstep>0 ? t<tstop:t>tstop)
#define NEARSTOP       (tstep > 0 ? \
                        t+0.9375*tstep > tstop && t+0.0625*tstep < tstop : \
                        t+0.9375*tstep < tstop && t+0.0625*tstep > tstop)

/*
 * Adams-Moulton with adaptive step size
 * Copyright Nicholas B. Tufillaro, 1982-1994. All rights reserved.
 */
ama()
{
        register gdval; /* good value to print ? */
        register overtime;
        long startit;
        double prevstep;
        double t;

        overtime = 1;
        gdval = YES;    startit = 0;    t = tstart;
top:
        /* fifth-order Runge-Kutta start-up */
        it = startit;
        while(it<=startit+PASTVAL&&(T_LT_TSTOP || overtime--)) {
                symtab->sy_value = t;
                field();
                if ( gdval) {
                        for ( fsp = symtab; fsp != NULL; fsp = fsp->sy_link) {
                                register j;
                                for ( j = it-startit; j > 0; j--) {
                                        fsp->sy_val[j] = fsp->sy_val[j-1];
                                        fsp->sy_pri[j] = fsp->sy_pri[j-1];
                                }
                                fsp->sy_val[0] = fsp->sy_value;
                                fsp->sy_pri[0] = fsp->sy_prime;
                        }
                        printq();       /* output */
                        if ( it == startit+PASTVAL)
                                break;  /* start-up complete */
                }
                if (tstep * (t+tstep-tstop) > 0)
                        tstep = tstop - t;
                for ( fsp = dqueue; fsp != NULL; fsp = fsp->sy_link) {
                        fsp->sy_k[0] = tstep * fsp->sy_prime;
                        fsp->sy_value = fsp->sy_val[0] + C20 * fsp->sy_k[0];
                }
                symtab->sy_value = t + C2t * tstep;
                field();
                for ( fsp = dqueue; fsp != NULL; fsp = fsp->sy_link) {
                        fsp->sy_k[1] = tstep * fsp->sy_prime;
                        fsp->sy_value = fsp->sy_val[0] + C30 * fsp->sy_k[0]
                            + C31 * fsp->sy_k[1];
                }
                symtab->sy_value = t + C3t * tstep;
                field();
                for ( fsp = dqueue; fsp != NULL; fsp = fsp->sy_link) {
                        fsp->sy_k[2] = tstep * fsp->sy_prime;
                        fsp->sy_value = fsp->sy_val[0] + C40 * fsp->sy_k[0]
                            + C41 * fsp->sy_k[1]
                            + C42 * fsp->sy_k[2];
                }
                symtab->sy_value = t + C4t * tstep;
                field();
                for ( fsp = dqueue; fsp != NULL; fsp = fsp->sy_link) {
                        fsp->sy_k[3] = tstep * fsp->sy_prime;
                        fsp->sy_value = fsp->sy_val[0] + C50 * fsp->sy_k[0]
                            + C51 * fsp->sy_k[1]
                            + C52 * fsp->sy_k[2]
                            + C53 * fsp->sy_k[3];
                }
                symtab->sy_value = t + tstep;
                field();
                for ( fsp = dqueue; fsp != NULL; fsp = fsp->sy_link) {
                        fsp->sy_k[4] = tstep * fsp->sy_prime;
                        fsp->sy_value = fsp->sy_val[0] + C60 * fsp->sy_k[0]
                            + C61 * fsp->sy_k[1]
                            + C62 * fsp->sy_k[2]
                            + C63 * fsp->sy_k[3]
                            + C64 * fsp->sy_k[4];
                }
                symtab->sy_value = t + C6t * tstep;
                field();
                for ( fsp = dqueue; fsp != NULL; fsp = fsp->sy_link)
                        fsp->sy_k[5] = tstep * fsp->sy_prime;
                for ( fsp = dqueue; fsp != NULL; fsp = fsp->sy_link) {
                        fsp->sy_predi = fsp->sy_val[0] + A0 * fsp->sy_k[0]
                            + A2 * fsp->sy_k[2]
                            + A3 * fsp->sy_k[3]
                            + A4 * fsp->sy_k[4];
                        fsp->sy_value = fsp->sy_val[0] + B0 * fsp->sy_k[0]
                            + B2 * fsp->sy_k[2]
                            + B3 * fsp->sy_k[3]
                            + B4 * fsp->sy_k[4]
                            + B5 * fsp->sy_k[5];
                        if ( fsp->sy_value != 0.0)
                                fsp->sy_sserr = fabs( 1.0 - fsp->sy_predi /
                                fsp->sy_value);
                        fsp->sy_aberr = fabs( fsp->sy_value - fsp->sy_predi);
                }
                if ( !conflag && T_LT_TSTOP) {
                        maxerr();
                        if ( hierror()) { 
                                tstep *= 0.5;
                                for ( fsp = symtab; fsp != NULL; fsp = fsp->sy_link)
                                        fsp->sy_value = fsp->sy_val[0];
                                gdval = NO;
                                continue;
                        }
                        else if ( lowerror() && prevstep != tstep) {
                                prevstep = tstep; /* prevent infinite loops */
                                tstep *= 2.0;
                                for ( fsp = dqueue; fsp != NULL; fsp = fsp->sy_link)
                                        fsp->sy_value = fsp->sy_val[0];
                                gdval = NO;
                                continue;
                        }
                }
                gdval = YES;
                ++it;
                t += tstep; /* the round off error is gross */
        }
        /* predictor - corrector */
        while ( T_LT_TSTOP ) {
                /* Adams-Bashforth predictor */
                if(tstep*(t+tstep-tstop) > 0){
                        startit = it;
                        goto top;
                }
                for ( fsp = dqueue; fsp != NULL ; fsp = fsp->sy_link) {
                        fsp->sy_predi = fsp->sy_value
                            = fsp->sy_val[0] + tstep/24.0 *
                            (55 * fsp->sy_pri[0]
                            -59 * fsp->sy_pri[1]
                            +37 * fsp->sy_pri[2]
                            -9  * fsp->sy_pri[3]);
                }
                ++it;
                symtab->sy_value = t += tstep;
                /* the round off error is gross */
                field();
                /* Adams-Moulton corrector */
                for ( fsp = dqueue; fsp != NULL; fsp = fsp->sy_link) {
                        fsp->sy_value = fsp->sy_val[0] + tstep/24.0 *
                            (9  * fsp->sy_prime
                            +19 * fsp->sy_pri[0]
                            -5  * fsp->sy_pri[1]
                            +     fsp->sy_pri[2]);
                        if ( fsp->sy_value != 0.0)
                                fsp->sy_sserr = ECONST *
                                    fabs( 1.0 - fsp->sy_predi / fsp->sy_value);
                        fsp->sy_aberr = ECONST *
                            fabs( fsp->sy_value - fsp->sy_predi);
                        fsp->sy_value += ECONST * ( fsp->sy_predi - fsp->sy_value);
                }
                if ( !conflag) {
                        maxerr();
                        if ( hierror()) {
                                tstep *= 0.5;
                                t = symtab->sy_val[0];
                                for ( fsp = symtab; fsp != NULL; fsp = fsp->sy_link) {
                                        fsp->sy_value = fsp->sy_val[0];
                                        fsp->sy_prime = fsp->sy_pri[0];
                                }
                                startit = --it;
                                gdval = NO;
                                goto top;
                        }
                        else if ( lowerror()) {
                                        tstep *= 2.0;
                                        t = symtab->sy_val[0];
                                        for ( fsp = symtab; fsp != NULL; fsp = fsp->sy_link) {
                                                fsp->sy_value = fsp->sy_val[0];
                                                fsp->sy_prime = fsp->sy_pri[0];
                                        }
                                        startit = --it;
                                        gdval = NO;
                                        goto top;
                        }
                }
                field();
                /* cycle indices */
                for ( fsp = symtab; fsp != NULL; fsp = fsp->sy_link) {
                        register j;
                        for ( j = PASTVAL; j > 0; j--) {
                                fsp->sy_val[j] = fsp->sy_val[j-1];
                                fsp->sy_pri[j] = fsp->sy_pri[j-1];
                        }
                        fsp->sy_val[0] = fsp->sy_value;
                        fsp->sy_pri[0] = fsp->sy_prime;
                }
                /* output */
                printq();
        }
}
End of ama.c
\n#==========================
#= Here is file eu.c =
#==========================\n
echo eu.c 1>&2
cat >eu.c <<'End of eu.c'
/*
 * Euler's method: quick and dirty.
 * Copyright Nicholas B. Tufillaro, 1982-1994. All rights reserved.
 */

#include "ode.h"

eu()
{
        double t;

        for ( it = 0, t = tstart; !STOPR; t = tstart + ( ++it) * tstep) {
                symtab->sy_value = symtab->sy_val[0] = t;
                field();
                for( fsp = dqueue; fsp != NULL; fsp = fsp->sy_link) {
                        fsp->sy_val[0] = fsp->sy_value;
                        fsp->sy_pri[0] = fsp->sy_prime;
                }
                printq();
                for ( fsp = dqueue; fsp != NULL; fsp = fsp->sy_link)
                        fsp->sy_value = fsp->sy_val[0] + tstep * fsp->sy_prime;
        }
}

/*
 * if tstart == tstop, i.e., step nowhere, then just print current values
 */
trivial()
{
        symtab->sy_value = symtab->sy_val[0] = tstart;
        field();
        for( fsp = dqueue; fsp != NULL; fsp = fsp->sy_link) {
                fsp->sy_val[0] = fsp->sy_value;
                fsp->sy_pri[0] = fsp->sy_prime;
        }
        printq();
}
End of eu.c
\n#==========================
#= Here is file expr.c =
#==========================\n
echo expr.c 1>&2
cat >expr.c <<'End of expr.c'
/*
 * expression tree evaluator
 * and expression space
 * management routines
 * Copyright Nicholas B. Tufillaro, 1982-1994. All rights reserved.
 */

#include "ode.h"

double
eval(ep)
register struct expr *ep;
{
        register double *sp;
        double tmp;

        for (sp = &stack[NSTACK]; ep!=NULL; ep=ep->ex_next) {
#ifdef DEBUG
                if (sp > &stack[NSTACK])
                        panic("expression stack underflow\n");
#endif
                if (sp <= stack)
                        panic("stack overflow -- bump NSTACK\n");
                switch (ep->ex_oper) {
                case E_CONST:
                        *--sp = ep->ex_value;
                        break;
                case E_IDENT:
                        *--sp = ep->ex_sym->sy_value;
                        break;
                case E_PLUS:
                        tmp = *sp++;
                        *sp += tmp;
                        break;
                case E_MINUS:
                        tmp = *sp++;
                        *sp -= tmp;
                        break;
                case E_MULT:
                        tmp = *sp++;
                        *sp *= tmp;
                        break;
                case E_DIV:
                        tmp = *sp++;
                        *sp /= tmp;
                        break;
                case E_POWER:
                        tmp = *sp++;
                        if (tmp!=(int)tmp && *sp<0)
                                rterror("neg. number to non-integer power");
                        *sp = pow(*sp,tmp);
                        break;
                case E_SQAR:
                        *sp *= *sp;
                        break;
                case E_CUBE:
                        *sp *= *sp * *sp;
                        break;
                case E_INV:
                        *sp = 1. / *sp;
                        break;
                case E_SQRT:
                        if (*sp < 0)
                                rterror("sqrt of a neg. number");
                        *sp = sqrt(*sp);
                        break;
                case E_SIN:
                        *sp = sin(*sp);
                        break;
                case E_COS:
                        *sp = cos(*sp);
                        break;
                case E_ATAN:
                        *sp = atan(*sp);
                        break;
                case E_ABS:
                        if (*sp < 0)
                                *sp = -*sp;
                        break;
                case E_EXP:
                        *sp = exp(*sp);
                        break;
                case E_LOG:
                        if (*sp <= 0)
                                rterror("ln of non-pos. number");
                        *sp = log(*sp);
                        break;
                case E_SINH:
                        *sp = sinh(*sp);
                        break;
                case E_COSH:
                        *sp = cosh(*sp);
                        break;
                case E_TANH:
                        *sp = tanh(*sp);
                        break;
                case E_NEG:
                        *sp = -*sp;
                        break;
                default:
                        panic("bad ex_oper (%d) in eval\n",ep->ex_oper);
                }
        }
        return *sp;
}

struct expr *
ealloc()
{
        register struct expr *ep;

        ep = (struct expr *)malloc(sizeof(struct expr));
        if (ep == NULL)
                fatal("out of core\n");

        ep->ex_next = NULL;
        ep->ex_sym = NULL;
        ep->ex_oper = E_NOOP;
        return ep;
}

efree(ep)
struct expr *ep;
{
        if (ep==NULL || ep==&exprzero || ep==&exprone)
                return;
        efree(ep->ex_next);
        free((char *)ep);
}
End of expr.c
\n#==========================
#= Here is file ext.c =
#==========================\n
echo ext.c 1>&2
cat >ext.c <<'End of ext.c'
/*
 * external initializations for ode
 * Copyright Nicholas B. Tufillaro, 1982-1994. All rights reserved.
 */

#include "ode.h"

int	prec	= -1;
long	it	= 0;
double	hmin	= HMIN;
double	hmax	= HMAX;
double	ssmin	= 1e-11;
double	ssmax	= 1e-8;
double	abmin	= 1e-36;
double	abmax	= 1e36;
double	acmax	= 1e36;
struct	sym	*symtab = NULL;
struct	sym	*fsp	= NULL;
struct	sym	*dqueue	= NULL;
struct	prt	*pqueue = NULL;
struct	expr	exprzero = {
	E_CONST,
	0.,
	NULL,
	NULL,
};
struct	expr	exprone = {
	E_CONST,
	1.,
	NULL,
	NULL,
};
End of ext.c
\n#==========================
#= Here is file float.c =
#==========================\n
echo float.c 1>&2
cat >float.c <<'End of float.c'
/*
 * routines to enable and disable FPU traps
 * floating point errors come here
 * Copyright Nicholas B. Tufillaro, 1982-1994. All rights reserved.
 */

#include "ode.h"
#include <signal.h>
#include "fps.h"

#ifdef VAX_4BSD
fptrap(signo,param,xx,pc,psl)
{
        char *e;

        if (signo != SIGFPE)
                goto other;
        sigsys(SIGFPE,fptrap);
        switch (param) {

        case FE_OVRF:
                e = "real number too large";
                break;

        case FE_DIV0:
                e = "division by zero";
                break;

        default:
        other:
                e = "unknown arithmetic error";
        }
        rterror(e);
}

#else
#ifdef V7_1170
static fbits;
fptrap()
{
        struct fpstat fe;
        char *e;

        stst(&fe);
        signal(SIGFPE,fptrap);
        switch (fe.fps_er) {
        case FE_DIV0:
                e = "division by zero";
                break;
        case FE_OVRF:
                e = "real number too large";
                break;
        case FE_UNDF:
                e = "real number too small";
                break;
        default:
                e = "unknown floating point error";
        }
        rterror(e);
}
#else
fptrap()
{
        rterror("arithmetic exception");
}
#endif
#endif

setflt()
{
#ifdef VAX_4BSD
        if ((int)sigsys(SIGFPE,SIG_IGN) != (int)SIG_IGN)
                sigsys(SIGFPE,fptrap);
#else
        if (signal(SIGFPE,SIG_IGN) != SIG_IGN)
                signal(SIGFPE,fptrap);
#ifdef V7_1170
        /*
         * set fp unit to trap on
         *   underflow
         *   overflow
         *   integer conversion error
         * as well as divide by zero
         */
        fbits = stfps();
        ldfps(fbits|FS_FIU|FS_FIV|FS_FIC);
#endif
#endif
}

resetflt()
{
#ifdef VAX_4BSD
        if ((int)sigsys(SIGFPE,SIG_IGN) != (int)SIG_IGN)
                signal(SIGFPE,SIG_DFL);
#else
        if (signal(SIGFPE,SIG_IGN) != SIG_IGN)
                signal(SIGFPE,SIG_DFL);
#ifdef V7_1170
        ldfps(fbits);
#endif
#endif
}
End of float.c
\n#==========================
#= Here is file fps.h =
#==========================\n
echo fps.h 1>&2
cat >fps.h <<'End of fps.h'
#ifdef V7_1170

/*
 * structure filled in by stst
 * Copyright Nicholas B. Tufillaro, 1982-1994. All rights reserved.
 */
struct fpstat {
        short   fps_er;
        short   fps_pc;
};

/*
 * bits in the FPS word
 */
#define FS_FER  0100000 /* error occurred */
#define FS_FID  040000  /* interrupt disable */
#define FS_FIUV 04000   /* interrupt on -0 */
#define FS_FIU  02000   /* interrupt on underflow */
#define FS_FIV  01000   /* interrupt on overflow */
#define FS_FIC  0400    /* interrupt on conversion error */
#define FS_FD   0200    /* double precision mode */
#define FS_FL   0100    /* long integer mode */
#define FS_FT   040     /* chop mode */
#define FS_FMM  020     /* maint. mode */
#define FS_FN   010     /* last result was <0 */
#define FS_FZ   04      /* last result was ==0 */
#define FS_FV   02      /* last op caused expon. overflow */
#define FS_FC   01      /* last op caused carry (conv. only) */

/*
 * fp errors (values for FEC)
 */
#define FE_OP   2       /* flt op code error */
#define FE_DIV0 4       /* flt divide by 0 */
#define FE_CONV 6       /* conv error */
#define FE_OVRF 8       /* flt overflow */
#define FE_UNDF 10      /* flt underflow */
#define FE_MIN0 12      /* -0 */
#define FE_TRAP 14      /* maint. trap */

#endif

#ifdef VAX_4BSD

/*
 * values for param in fptrap (float.c)
 */
#define FE_OVRF 3
#define FE_DIV0 4

#endif
End of fps.h
\n#==========================
#= Here is file gram.y =
#==========================\n
echo gram.y 1>&2
cat >gram.y <<'End of gram.y'
%{
/*
 * Grammar for ode:
 * Most things are self-explanatory.
 * When you're done with a lexptr-type object
 * you should free it with lfree.  They are
 * used for passing constants around while parsing
 * (computing the value of) a cexpr.  The macros
 * for evaluating operators and functions are the
 * most important thing to be familiar with before
 * toying with the semantics.
 * Copyright Nicholas B. Tufillaro, 1982-1994. All rights reserved.
 */
#include "ode.h"
/*
 * value is true iff both operands pass ONECON
 */
#define TWOCON(x,y) (ONECON(x) && ONECON(y))
/*
 * Value must be a (struct expr *).  Returns true if its arg,
 * when evaluated, would put a constant on the stack
 */
#define ONECON(x) (x->ex_oper==E_CONST && x->ex_next==NULL)
/*
 * performs ordinary binary arithmetic, when
 * there are two constants in an expr.
 */
#define COMBINE(x,y,r,op) {x->ex_value op= y->ex_value; efree(y); r = x;}
/*
 * generates stack code for binary arithmetic,
 * as for a dyadic operator in an expression
 */
#define BINARY(x,y,r,op) {register struct expr *ep=ealloc();\
                ep->ex_oper = op;\
                concat(y,ep);\
                concat(r=x,y);}
/*
 * performs ordinary unary arithmetic, when
 * there is a constant in an expr.
 * "-" seems to work as a monadic operator.
 */
#define CONFUNC(x,r,f) {x->ex_value = f(x->ex_value); r = x;}
/*
 * compile stack code for a unary operation,
 * as for a monadic operator in an expression.
 */
#define UNARY(oprnd,r,op) {register struct expr *ep=ealloc();\
                ep->ex_oper = op;\
                concat(r=oprnd,ep);}
/*
 * performs binary arithmetic in a cexpr
 */
#define CEXOP(x,y,r,op) {x->lx_u.lxu_value op= y->lx_u.lxu_value;\
                lfree(y);\
                r = x;}
/*
 * performs unary arithmetic in a cexpr
 */
#define CEXFUNC(x,r,f) {x->lx_u.lxu_value = f(x->lx_u.lxu_value);r=x;}

/*
 * A hook for future upgrades in error reporting
 */
static  char    *errmess = NULL;

int     erritem;
%}
%union {
        struct  lex     *lexptr;
        struct  expr    *exprptr;
        struct  prt     *prtptr;
        int     simple;
}
%token <lexptr> NUMBER IDENT SEP
%token SQRT EXP LOG SIN COS ATAN SINH COSH TANH ABS
%token EVERY FROM PRINT STEP EXAM
%start program
%type <simple> prttag
%type <lexptr> cexpr
%type <exprptr> expr
%type <prtptr> prtitem
%nonassoc '='
%left '+' '-'
%left '*' '/'
%right '^'
%right UMINUS
%%
program         : stat
                | program stat
                ;

stat            : SEP
                        { lfree($1); }
                | IDENT '=' expr SEP
                        {
                        register struct sym *sp;
                        sp = lookup($1->lx_u.lxu_name);
                        sp->sy_value = eval($3);
                        sp->sy_flags |= SF_INIT;
                        lfree($1);
                        efree($3);
                        lfree($4);
                        }
                | error SEP
                        {
                        if (errmess == NULL)
                                errmess = "syntax error";
                        error("%s: %s\n",lineno($2->lx_lino),errmess);
                        errmess = NULL;
                        lfree($2);
                        yyerrok;
                        yyclearin;
                        }
                | IDENT '\'' '=' expr SEP
                        {
                        register struct sym *sp;
                        register struct prt *pp, *qp;
                        sp = lookup($1->lx_u.lxu_name);
                        efree(sp->sy_expr);
                        sp->sy_expr = $4;
                        sp->sy_flags |= SF_ISEQN;
                        if (!sawprint) {
                                for (pp=pqueue; pp!=NULL; pp=pp->pr_link)
                                        if (pp->pr_sym == sp)
                                                goto found;
                                pp = palloc();
                                pp->pr_sym = sp;
                                if (pqueue == NULL)
                                        pqueue = pp;
                                else {
                                        for (qp=pqueue; qp->pr_link!=NULL; )
                                                qp = qp->pr_link;
                                        qp->pr_link = pp;
                                }
                        }
                found:
                        lfree($1);
                        lfree($5);
                        }
                | PRINT prtlist optevery optfrom SEP
                        {
                        sawprint = 1;
                        prerr = erritem;
                        erritem = 0;
                        lfree($5);
                        }
                | STEP cexpr ',' cexpr SEP
                        {
                        lfree($5);
                        tstart = $2->lx_u.lxu_value;
                        lfree($2);
                        tstop = $4->lx_u.lxu_value;
                        lfree($4);
                        if (!conflag)
                                startstep();
                        solve();
                        sawstep = 1;
                        }
                | STEP cexpr ',' cexpr ',' cexpr SEP
                        {
                        int savstep, savconf;
                        lfree($7);
                        tstart = $2->lx_u.lxu_value;
                        lfree($2);
                        tstop = $4->lx_u.lxu_value;
                        lfree($4);
                        savstep = tstep;
                        tstep = $6->lx_u.lxu_value;
                        lfree($6);
                        savconf = conflag;
                        conflag = 1;
                        solve();
                        tstep = savstep;
                        conflag = savconf;
                        sawstep = 1;
                        }
                | EXAM IDENT SEP
                        {
                        struct sym *sp;

                        lfree($3);
                        sp = lookup($2->lx_u.lxu_name);
                        lfree($2);
                        printf("\"%.*s\" is ",NAMMAX,sp->sy_name);
                        switch (sp->sy_flags & SF_DEPV) {
                        case SF_DEPV:
                        case SF_ISEQN:
                                printf("a dynamic variable\n");
                                break;
                        case SF_INIT:
                                printf("an initialized constant\n");
                                break;
                        case 0:
                                printf("an uninitialized constant\n");
                                break;
                        default:
                                panic("impossible (%d) in EXAM action\n",
                                        sp->sy_flags);
                        }
                        printf("value:");
                        prval(sp->sy_value);
                        printf("\nprime:");
                        prval(sp->sy_prime);
                        printf("\nsserr:");
                        prval(sp->sy_sserr);
                        printf("\naberr:");
                        prval(sp->sy_aberr);
                        printf("\nacerr:");
                        prval(sp->sy_acerr);
                        putchar('\n');
                        prexq(sp->sy_expr);
                        fflush(stdout);
                        }
                ;

prtlist         : prtitem
                        {
                        pfree(pqueue);
                        pqueue = $1;
                        }
                | prtlist ',' prtitem
                        {
                        register struct prt *pp;
                        for (pp=pqueue; pp->pr_link!=NULL; pp=pp->pr_link)
                                ;
                        pp->pr_link = $3;
                        }
                ;

prtitem         : IDENT prttag
                        {
                        register struct prt *pp;
                        pp = palloc();
                        pp->pr_sym = lookup($1->lx_u.lxu_name);
                        pp->pr_which = $2;
                        lfree($1);
                        $$ = pp;
                        }
                ;

prttag          : /* empty */
                        { $$ = P_VALUE; }
                | '\''
                        { $$ = P_PRIME; }
                | '~'
                        {
                        $$ = P_ACERR;
                        erritem = 1;
                        }
                | '!'
                        {
                        $$ = P_ABERR;
                        erritem = 1;
                        }
                | '?'
                        {
                        $$ = P_SSERR;
                        erritem = 1;
                        }
                ;

optevery        : /* empty */
                        { sawevery = 0; }
                | EVERY cexpr
                        {
                        sawevery = 1;
                        tevery = $2->lx_u.lxu_value;
                        lfree($2);
                        }
                ;

optfrom         : /* empty */
                        { sawfrom = 0; }
                | FROM cexpr
                        {
                        sawfrom = 1;
                        tfrom = $2->lx_u.lxu_value;
                        lfree($2);
                        }
                ;

cexpr           : '(' cexpr ')'
                        {
                        $$ = $2;
                        }
                | cexpr '+' cexpr
                        {
                        CEXOP($1,$3,$$,+)
                        }
                | cexpr '-' cexpr
                        {
                        CEXOP($1,$3,$$,-)
                        }
                | cexpr '*' cexpr
                        {
                        CEXOP($1,$3,$$,*)
                        }
                | cexpr '/' cexpr
                        {
                        CEXOP($1,$3,$$,/)
                        }
                | cexpr '^' cexpr
                        {
                        $1->lx_u.lxu_value =
                                pow($1->lx_u.lxu_value,$3->lx_u.lxu_value);
                        lfree($3);
                        $$ = $1;
                        }
                | SQRT '(' cexpr ')'
                        {
                        CEXFUNC($3,$$,sqrt)
                        }
                | ABS '(' cexpr ')'
                        {
                        if ($3->lx_u.lxu_value < 0)
                                $3->lx_u.lxu_value = -($3->lx_u.lxu_value);
                        $$ = $3;
                        }
                | EXP '(' cexpr ')'
                        {
                        CEXFUNC($3,$$,exp)
                        }
                | LOG '(' cexpr ')'
                        {
                        CEXFUNC($3,$$,log)
                        }
                | SIN '(' cexpr ')'
                        {
                        CEXFUNC($3,$$,sin)
                        }
                | COS '(' cexpr ')'
                        {
                        CEXFUNC($3,$$,cos)
                        }
                | ATAN '(' cexpr ')'
                        {
                        CEXFUNC($3,$$,atan)
                        }
                | SINH '(' cexpr ')'
                        {
                        CEXFUNC($3,$$,sinh)
                        }
                | COSH '(' cexpr ')'
                        {
                        CEXFUNC($3,$$,cosh)
                        }
                | TANH '(' cexpr ')'
                        {
                        CEXFUNC($3,$$,tanh)
                        }
                | '-' cexpr     %prec UMINUS
                        {
                        CEXFUNC($2,$$,-)
                        }
                | NUMBER
                        { $$ = $1; }
                ;

expr            : '(' expr ')'
                        { $$ = $2; }
                | expr '+' expr
                        {
                        if (TWOCON($1,$3))
                                COMBINE($1,$3,$$,+)
                        else
                                BINARY($1,$3,$$,E_PLUS);
                        }
                | expr '-' expr
                        {
                        if (TWOCON($1,$3))
                                COMBINE($1,$3,$$,-)
                        else
                                BINARY($1,$3,$$,E_MINUS);
                        }
                | expr '*' expr
                        {
                        if (TWOCON($1,$3))
                                COMBINE($1,$3,$$,*)
                        else
                                BINARY($1,$3,$$,E_MULT);
                        }
                | expr '/' expr
                        {
                        if (TWOCON($1,$3))
                                COMBINE($1,$3,$$,/)
                        else if (ONECON($3) && $3->ex_value!=0.) {
                                /* division by constant */
                                $3->ex_value = 1./$3->ex_value;
                                BINARY($1,$3,$$,E_MULT);
                        } else
                                BINARY($1,$3,$$,E_DIV);
                        }
                | expr '^' expr
                        {
                        double f;

                        if (TWOCON($1,$3)) {
                                /* case const ^ const */
                                $1->ex_value = pow($1->ex_value,$3->ex_value);
                                efree($3);
                        } else if (ONECON($1)) {
                                if ($1->ex_value == 1.) {
                                        /* case 1 ^ x */
                                        efree($3);
                                        $$ = $1;
                                } else
                                        goto other;
                        } else if (!ONECON($3))
                                goto other;
                        else {
                                register invert = 0;
                                f = $3->ex_value;
                                if (f < 0.) {
                                        /*
                                         * negative exponent means
                                         * to append an invert cmd
                                         */
                                        f = -f;
                                        invert = 1;
                                }
                                if (f == 2.) {
                                        /* case x ^ 2 */
                                        $3->ex_oper = E_SQAR;
                                        concat($1,$3);
                                        $$ = $1;
                                } else if (f == 3.) {
                                        /* case x ^ 3 */
                                        $3->ex_oper = E_CUBE;
                                        concat($1,$3);
                                        $$ = $1;
                                } else if (f == 0.5) {
                                        /* case x ^ .5 */
                                        $3->ex_oper = E_SQRT;
                                        concat($1,$3);
                                        $$ = $1;
                                } else if (f == 1.5) {
                                        /* case x ^ 1.5 */
                                        $3->ex_oper = E_CUBE;
                                        BINARY($1,$3,$$,E_SQRT);
                                } else if (f == 1.) {
                                        /* case x ^ 1 */
                                        efree($3);
                                        $$ = $1;
                                } else if (f == 0.) {
                                        /* case x ^ 0 */
                                        efree($1);
                                        $3->ex_value = 1.;
                                        $$ = $3;
                                } else {
                        other:
                                        /* default */
                                        invert = 0;
                                        BINARY($1,$3,$$,E_POWER);
                                }
                                if (invert)
                                        UNARY($$,$$,E_INV)
                        }
                        }
                | SQRT '(' expr ')'
                        {
                        if (ONECON($3))
                                CONFUNC($3,$$,sqrt)
                        else
                                UNARY($3,$$,E_SQRT);
                        }
                | ABS '(' expr ')'
                        {
                        if (ONECON($3)) {
                                if ($3->ex_value < 0)
                                        $3->ex_value = -($3->ex_value);
                                $$ = $3;
                        } else
                                UNARY($3,$$,E_ABS);
                        }
                | EXP '(' expr ')'
                        {
                        if (ONECON($3))
                                CONFUNC($3,$$,exp)
                        else
                                UNARY($3,$$,E_EXP);
                        }
                | LOG '(' expr ')'
                        {
                        if (ONECON($3))
                                CONFUNC($3,$$,log)
                        else
                                UNARY($3,$$,E_LOG);
                        }
                | SIN '(' expr ')'
                        {
                        if (ONECON($3))
                                CONFUNC($3,$$,sin)
                        else
                                UNARY($3,$$,E_SIN);
                        }
                | COS '(' expr ')'
                        {
                        if (ONECON($3))
                                CONFUNC($3,$$,cos)
                        else
                                UNARY($3,$$,E_COS);
                        }
                | ATAN '(' expr ')'
                        {
                        if (ONECON($3))
                                CONFUNC($3,$$,atan)
                        else
                                UNARY($3,$$,E_ATAN);
                        }
                | SINH '(' expr ')'
                        {
                        if (ONECON($3))
                                CONFUNC($3,$$,sinh)
                        else
                                UNARY($3,$$,E_SINH);
                        }
                | COSH '(' expr ')'
                        {
                        if (ONECON($3))
                                CONFUNC($3,$$,cosh)
                        else
                                UNARY($3,$$,E_COSH);
                        }
                | TANH '(' expr ')'
                        {
                        if (ONECON($3))
                                CONFUNC($3,$$,tanh)
                        else
                                UNARY($3,$$,E_TANH);
                        }
                | '-' expr      %prec UMINUS
                        {
                        if (ONECON($2))
                                CONFUNC($2,$$,-)
                        else
                                UNARY($2,$$,E_NEG);
                        }
                | NUMBER
                        {
                        $$ = ealloc();
                        $$->ex_oper = E_CONST;
                        $$->ex_value = $1->lx_u.lxu_value;
                        lfree($1);
                        }
                | IDENT
                        {
                        $$ = ealloc();
                        $$->ex_oper = E_IDENT;
                        $$->ex_sym = lookup($1->lx_u.lxu_name);
                        lfree($1);
                        }
                ;
%%
#include "lex.c"
yyerror(s)
char *s;
{}

/*
 * tack two queues of stack code together
 * e1 is connected on the tail of e0
 * There is no good way to test for circular
 * lists, hence the silly count.
 */
concat(e0,e1)
register struct expr *e0, *e1;
{
        register count;

        if (e0==NULL || e1==NULL) {
                panic("NULL expression queue\n");
        }
        for (count = 0; e0->ex_next != NULL; e0 = e0->ex_next)
                if (++count > 10000) {
                        panic("circular expression queue\n");
                }
        e0->ex_next = e1;
}

prexq(ep)
register struct expr *ep;
{
        char *s;

        printf(" code:");
        if (ep == NULL)
                putchar('\n');
        for (; ep != NULL; ep = ep->ex_next) {
                switch (ep->ex_oper) {
                case E_LOG: s = "ln"; break;
                case E_ABS: s = "abs"; break;
                case E_COS: s = "cos"; break;
                case E_EXP: s = "exp"; break;
                case E_SIN: s = "sin"; break;
                case E_DIV: s = "divide"; break;
                case E_INV: s = "invert"; break;
                case E_NEG: s = "negate"; break;
                case E_PLUS: s = "add"; break;
                case E_ATAN: s = "atan"; break;
                case E_COSH: s = "cosh"; break;
                case E_CUBE: s = "cube"; break;
                case E_SINH: s = "sinh"; break;
                case E_SQRT: s = "sqrt"; break;
                case E_TANH: s = "tanh"; break;
                case E_SQAR: s = "square"; break;
                case E_MULT: s = "multiply"; break;
                case E_POWER: s = "power"; break;
                case E_MINUS: s = "subtract"; break;
                case E_CONST:
                        printf("\tpush ");
                        prval(ep->ex_value);
                        putchar('\n');
                        s = NULL;
                        break;
                case E_IDENT:
                        printf("\tpush \"%.*s\"\n",
                                NAMMAX,ep->ex_sym->sy_name);
                        s = NULL;
                        break;
                default: s = "unknown!";
                }
                if (s != NULL)
                        printf("\t%s\n",s);
        }
}
End of gram.y
\n#==========================
#= Here is file lex.l =
#==========================\n
echo lex.l 1>&2
cat >lex.l <<'End of lex.l'
%C
%{
/*
 * input is redefined as a real routine below
 * Copyright Nicholas B. Tufillaro, 1982-1994. All rights reserved.
 */
#undef input
double atof();
static	curline	=	1;
%}
Dig		[0-9]
Expo		([eE][-+]?{Dig}{Dig}?)
%%
\#.*$		;/* nb: comments can't be continued */
\\\n		{
			curline++;
		}
[ \t]		;
[;\n]		{
			yylval.lexptr = lalloc();
			yylval.lexptr->lx_lino = curline;
			if (*yytext == '\n')
				curline++;
			return SEP;
		}
PI		{
			yylval.lexptr = lalloc();
			yylval.lexptr->lx_u.lxu_value =
				3.14159265358979323846;
			return NUMBER;
		}
^\.$		{
			/* lonely . isn't EOF in a -f file */
			if (infile == NULL)
				return EOF;
			REJECT;
		}
every		{ return EVERY; }
from		{ return FROM; }
print		{ return PRINT; }
step		{ return STEP; }
examine		{ return EXAM; }
abs		{ return ABS; }
sqrt		{ return SQRT; }
exp		{ return EXP; }
ln		{ return LOG; }
sin		{ return SIN; }
cos		{ return COS; }
atan		{ return ATAN; }
sinh		{ return SINH; }
cosh		{ return COSH; }
tanh		{ return TANH; }
[_a-zA-Z][_a-zA-Z0-9]*		{
			yylval.lexptr = lalloc();
			strncpy(yylval.lexptr->lx_u.lxu_name,yytext,NAMMAX);
			return IDENT;
		}
{Dig}+{Expo}?		|
{Dig}+"."{Dig}*{Expo}?	|
{Dig}*"."{Dig}+{Expo}?	{
			/*
			 * restrictions on numbers:
			 * radix 10 only
			 * a bare Expo isn't enough (matches IDENT)
			 */
			yylval.lexptr = lalloc();
			yylval.lexptr->lx_u.lxu_value = atof(yytext);
			return NUMBER;
		}
[-+*/^()',=?~!]	{
			/*
			 * accept as few as possible so the error
			 * message can be a smart one
			 */
			return yytext[0];
		}
.		{
			if (*yytext > '~' || *yytext < ' ')
				error("%s: bad char '\\%02o'\n",
					lineno(curline), *yytext&0377);
			else
				error("%s: bad char '%c'\n",
					lineno(curline), *yytext);
			return *yytext;
		}
%%
/*
 * space management for the lexer
 */
struct lex *
lalloc()
{
	register struct lex *lp;

	lp = (struct lex *)malloc(sizeof(struct lex));
	if (lp == NULL)
		fatal("out of core\n");
	lp->lx_u.lxu_value = 0.0;
	lp->lx_lino = 0;
	return lp;
}
lfree(lp)
register struct lex *lp;
{
	if (lp != NULL)
		free((char *)lp);
}

/*
 * Handle input files
 * This routine manages the transition from file to standard
 * input when the -f flag is given.
 * All the yy nonsense is here to appease lex.  Note the
 * functional similarity to lex's "input" macro, found in
 * any normal yy.lex.c file.
 */
input()
{
	if (yysptr > yysbuf)
		yytchar = *--yysptr;
	else if (infile != NULL) {
		yytchar = getc(infile);
		if (yytchar == EOF) {
			infile = NULL;
			curline = 1;
			goto standard;
		}
	} else
standard:
		yytchar = getchar();
	if (yytchar == EOF)
		return 0;
	if (yytchar == '\n')
		yylineno++;
	return yytchar;
}

/*
 * make a line number in
 * a standard format
 */
char *
lineno(n)
{
	static char text[16];

	sprintf(text,"%d%s",n,infile==NULL?"":"(file)");
	return text;
}
End of lex.l
\n#==========================
#= Here is file mch.s =
#==========================\n
echo mch.s 1>&2
cat >mch.s <<'End of mch.s'
/ routines for accessing the floating trap bits on
/ the PDP-11.  stst is for finding the error number.
.globl _stfps,_ldfps,_stst
stst	= 170300^tst
_stfps:
	mov	r5,-(sp)
	mov	sp,r5
	stfps	r0
	mov	(sp)+,r5
	rts	pc
_ldfps:
	mov	r5,-(sp)
	mov	sp,r5
	ldfps	4(r5)
	mov	(sp)+,r5
	rts	pc
_stst:
	mov	r5,-(sp)
	mov	sp,r5
	stst	*4(r5)
	mov	(sp)+,r5
	rts	pc
End of mch.s
\n#==========================
#= Here is file misc.c =
#==========================\n
echo misc.c 1>&2
cat >misc.c <<'End of misc.c'
/*
 * mainline and other stuff
 * that doesn't go anywhere
 * else in particular
 * (in alphabetical order!)
 * Copyright Nicholas B. Tufillaro, 1982-1994. All rights reserved.
 */

#include "ode.h"

double atof();
FILE	*infile = NULL;

main(argc,argv)
char **argv;
{
	prog = *argv;
	args(argc,argv);
	yyparse();
	exit(0);
}

/*
 * handle command line args
 * NOARG is destructive of ac and av.  It is used where
 * a numeric arg is required and failure is usually fatal.
 * LOOKARG is not destructive and is used when a numeric arg
 * is suspected of being there.  Both of these macros assume
 * that a numeric arg starts with something other than '-'.
 * This means it can't be negative, which is almost always
 * true anyway.
 */

#define NOARG (--ac==0 || **++av=='-')
#define LOOKARG (ac>1 && *av[1]!='-')

args(ac,av)
char **av;
{
	while (--ac) {
		if (**++av != '-')
			fatal("unexpected non-flag arg '%s'\n",*av);
		if ((*av)[1]=='\0' || (*av)[2]!='\0')
			fatal("bad arg '%s'\n",*av);
		switch ((*av)[1]) {

		case 'A':
			Aflag = 1;
			goto CONSTANT;
		case 'E':
			Eflag = 1;
			conflag = 1;
			tstep = .1;
			goto CONSTANT;
		case 'R':
			Rflag = 1;
CONSTANT:
			if (LOOKARG) {
				conflag = 1;
				--ac;
				tstep = atof(*++av);
			}
			break;

		case 'f':
			if (NOARG)
				fatal("no file name for -f\n");
			if ((infile=fopen(*av,"r")) == NULL) {
				perror(*av);
				exit(1);
			}
			break;

		case 'p':
			if (NOARG)
				fatal("no value for -p\n");
			prec = atoi(*av) - 1;
			if (prec <= 0 || prec > 18)
				fatal("bad value for -p\n");
			pflag = 1;
			fwd = prec + 7;
			if (fwd < 9)
				fwd = 9;
			break;

		case 't':
			tflag = 1;
			if (!pflag) {
				prec = 6;
				fwd = 13;
			}
			break;

		case 'r':
			rflag = 1;
			if (NOARG)
				fatal("no value for -r\n");
			ssmax = atof(*av);
			if (LOOKARG) {
				--ac;
				ssmin = atof(*++av);
			} else 
				ssmin = ssmax * SCALE;
			if ( ssmax < HMIN)
				fatal("-r: max value too small\n");
			break;

		case 'e':
			eflag = 1;
			if (NOARG)
				fatal("no value for -e\n");
			abmax = atof(*av);
			if (LOOKARG) {
				--ac;
				abmin = atof(*++av);
			} else
				abmin = abmax * SCALE;
			if ( abmax < HMIN)
				fatal("-e: max value too small\n");
			break;

		case 'h':
			if (NOARG)
				fatal("no minimum with -h\n");
			hmin = atof(*av);
			if (LOOKARG) {
				--ac;
				hmax = atof(*++av);
				hflag = 1;
			}
			if ( hmin < HMIN)
				fatal("-h: value too small\n");
			break;

		case 's':
			sflag = 1;
			break;

		default:
			fatal("unknown flag '%s'\n",*av);
		}
	}
	if (Rflag+Eflag+Aflag > 1)
		fatal("only one of -R, -E, and -A allowed\n");
	if (Eflag & (eflag|rflag))
		fatal("-E illegal with -e or -r\n");
}

/*
 * check checks internal consistency prior to calling the
 * numerical routine.  Tasks it performs include:
 *	ensures one and only one independent variable
 *	the symbol table is coherent
 *	the print queue is coherent
 *	the independent variable has a derivative == 1.
 *	each dependent variable has an ODE and an initial value
 */
check()
{
	register struct sym *sp, *ivp, *prevp;
	struct prt *pp;

	/*
	 * discard any previous entry for "(indep)"
	 */
	prevp = NULL;
	for (sp=symtab; sp!=NULL; sp=sp->sy_link) {
		if (strncmp(sp->sy_name,"(indep)",NAMMAX) == 0) {
			if (prevp == NULL)
				symtab = sp->sy_link;
			else
				prevp->sy_link = sp->sy_link;
			sfree(sp);
			break;
		}
		prevp = sp;
	}

	/*
	 * check for only one independent variable
	 */
	ivp = prevp = NULL;
	for (sp=symtab; sp!=NULL; sp=sp->sy_link) {
		if (!(sp->sy_flags & SF_DEPV)) {
			if (ivp != NULL) {
				error("both '%.*s' and '%.*s' are independent\n",
					NAMMAX,sp->sy_name,
					NAMMAX,ivp->sy_name);
				return !OK;
			}
			ivp = sp;
		}
		if (ivp == NULL)
			prevp = sp;
	}

	/*
	 * invent one if it's missing
	 */
	if (ivp == NULL) {
		ivp = salloc();
		strncpy(ivp->sy_name,"(indep)",NAMMAX);
	} else if (prevp != NULL) {

		/*
		 * link the independent var at the
		 * head of the symtab queue
		 */
		prevp->sy_link = ivp->sy_link;
		ivp->sy_link = symtab;
		symtab = ivp;
	}

	/*
	 * now ivp points to the ind. var. entry
	 * make sure the independent var gets
	 * printed when there's no print statement
	 */
	if (!sawprint) {
		for (pp=pqueue; pp!=NULL; pp=pp->pr_link)
			if (pp->pr_sym == ivp)
				goto found;
		pp = palloc();
		pp->pr_link = pqueue;
		pqueue = pp;
		pp->pr_sym = ivp;
	}
found:
	/*
	 * indep var has a derivative of 1.0
	 */
	ivp->sy_expr = &exprone;

	/*
	 * ensure an expr and value for each dep var
	 */
	for (sp=symtab; sp!=NULL; sp=sp->sy_link) {
		switch (sp->sy_flags&SF_DEPV) {
		case SF_INIT:
			sp->sy_expr = &exprzero;
			sp->sy_flags |= SF_ISEQN;
			break;
		case SF_ISEQN:
			sp->sy_value = 0;
			sp->sy_flags |= SF_INIT;
			break;
		}
	}

	/*
	 * dependent variables start here
	 */
	dqueue = symtab->sy_link;
	return OK;
}

/*
 * set default values
 * determine step direction (forgive confused users)
 * initialize values for printq() and numerical routines
 */
defalt()
{
	if (!sawfrom)
		tfrom = tstart;
	if (!sawevery)
		tevery = 1;
	if (tstart>tstop && tstep>0)
		tstep = -tstep;
	else if (tstart<tstop && tstep<0)
		tstep = -tstep;
	printnum = NO;
}

/*
 * non-fatal error message
 */
error(fmt,a,b,c,d,e,f,g,h,i,j)
char *fmt;
{
	fprintf(stderr,"%s: ",prog);
	fprintf(stderr,fmt,a,b,c,d,e,f,g,h,i,j);
}

/*
 * fatal error message
 */
fatal(fmt,a,b,c,d,e,f,g,h,i,j)
char *fmt;
{
	fprintf(stderr,"%s: ",prog);
	fprintf(stderr,fmt,a,b,c,d,e,f,g,h,i,j);
	exit(1);
}

/*
 * evaluate all the derivatives
 */
field()
{
	for (fsp=symtab->sy_link; fsp!=NULL; fsp=fsp->sy_link)
		fsp->sy_prime = eval(fsp->sy_expr);
}

/*
 * internal error (fatal)
 */
panic(fmt,a,b,c,d,e,f,g,h,i,j)
char *fmt;
{
	fprintf(stderr,"%s panic: ",prog);
	fprintf(stderr,fmt,a,b,c,d,e,f,g,h,i,j);
	exit(1);
}

# define	LASTVAL (tstep>0 ? t>=tstop-0.0625*tstep : t<=tstop-0.0625*tstep)
# define	TFROM	( tfrom - 0.0625*tstep)
# define	PRFROM	(tstep>0 ? t >= TFROM : t<= TFROM)
printq()
{
	double f;
	double t;
	register struct prt *pp;

	t = symtab->sy_value;
	if ( !printnum && PRFROM)
		printnum = YES;
	if ( (!( it % tevery) && printnum) || LASTVAL) {
		pp = pqueue;
		if (pp != NULL) for (;;) {
			switch (pp->pr_which) {
			case P_VALUE:
				f = pp->pr_sym->sy_value;
				break;
			case P_PRIME:
				f = pp->pr_sym->sy_prime;
				break;
			case P_ACERR:
				f = pp->pr_sym->sy_acerr;
				break;
			case P_ABERR:
				f = pp->pr_sym->sy_aberr;
				break;
			case P_SSERR:
				f = pp->pr_sym->sy_sserr;
				break;
			default:
				panic("bad pr_which (%d) in printq\n",
					pp->pr_which);
			}
			prval(f);
			pp = pp->pr_link;
			if (pp == NULL)
				break;
			putchar(' ');
		}
		putchar('\n');
		fflush(stdout);
	}
	if ( it == LONGMAX)
		it = 0;
}

/*
 * print a value to current precision
 * kludge for Pascal compatibility
 */
prval(x)
double x;
{
	if (prec < 0) {
		char outbuf[20];
		if (x < 0) {
			putchar('-');
			x = -x;
		}
		sprintf(outbuf,"%.7g",x);
		if (*outbuf == '.')
			putchar('0');
		printf("%s",outbuf);
	} else
		printf("%*.*e",fwd,prec,x);
}

/*
 * print a diagnostic for run-time errors.
 * uses fsp to decide which dep. var. was being worked
 */
rterror(s)
char *s;
{
	if (fsp == NULL)
		error("%s\n",s);
	else
		error("%s while calculating %.*s'\n",s,NAMMAX,fsp->sy_name);
	longjmp(mark,1);
}

/*
 * Run the numerical stuff.
 * This gets called from the grammar
 * because we want to solve on each
 * 'step' statement.
 */
solve()
{
	register struct sym *sp;
	int adapt;

	if (check() != OK)
		return;
	defalt();
	if (tflag)
		title();
	setflt();
	if (!setjmp(mark)) {
		adapt = eflag|rflag|!conflag;
		if ( tstart == tstop)
			trivial();
		else if (Eflag)
			eu();
		else if (Aflag) {
			if (adapt || prerr)
				ama();
			else
				am();
		} else {
			if (adapt || prerr)
				rka();
			else
				rk();
		}
	}
	resetflt();
	for (sp=symtab; sp!=NULL; sp=sp->sy_link) {
		sp->sy_prime = sp->sy_pri[0];
		sp->sy_value = sp->sy_val[0];
	}
}

/* 
 * choose step size at tstart
 */
startstep()
{
	if ( !hflag)
		hmax = fabs((tstop-tstart)/2);
	tstep = fabs((tstop-tstart)/MESH);
	if ( tstep > hmax)
		tstep = hmax;
	if ( tstep < hmin)
		tstep = hmin;
	while( tstep >= HMAX)
		tstep *= 0.5;
	while( tstart + tstep == tstart)
		tstep *= 2;
}

/*
 * print a header
 * Try to center the headings over the columns
 */
title()
{
	register struct prt *pp;
	register char tag;

	pp = pqueue;
	if (pp != NULL) for (;;) {
		switch (pp->pr_which) {
		case P_PRIME: 
			tag = '\''; 
			break;
		case P_VALUE: 
			tag = ' '; 
			break;
		case P_SSERR: 
			tag = '?'; 
			break;
		case P_ABERR:
			tag = '!';
			break;
		case P_ACERR: 
			tag = '~'; 
			break;
		default: 
			panic("bad pr_which (%d) in title\n",pp->pr_which);
		}
		printf(" %*.*s%c",fwd-2,NAMMAX,pp->pr_sym->sy_name,tag);
		if ((pp=pp->pr_link) == NULL)
			break;
		putchar(' ');
	}
	putchar('\n');
	fflush(stdout);
}
End of misc.c
\n#==========================
#= Here is file num.h =
#==========================\n
echo num.h 1>&2
cat >num.h <<'End of num.h'
/*
 * Constants used in adaptive-step-size numerical
 * routines.
 * Copyright Nicholas B. Tufillaro, 1982-1994. All rights reserved.
 */

#define ECONST  .07037037037037037037   /* 19/270       */
#define A0      .11574074074074074074   /* 25/216       */
#define A2      .54892787524366471734   /* 1408/2565    */
#define A3      .53533138401559454191   /* 2197/4104    */
#define A4      -.20000000000000000000  /* -1/5         */
#define B0      .11851851851851851851   /* 16/135       */
#define B2      .51898635477582846003   /* 6656/12825   */
#define B3      .50613149034201665780   /* 28561/56430  */
#define B4      -.18000000000000000000  /* -9/50        */
#define B5      .03636363636363636363   /* 2/55         */
#define C2t     .25                     /* 1/4          */
#define C20     .25                     /* 1/4          */
#define C3t     .37500000000000000000   /* 3/8          */
#define C30     .09375000000000000000   /* 3/32         */
#define C31     .28125000000000000000   /* 9/32         */
#define C4t     .92307692307692307692   /* 12/13        */
#define C40     .87938097405553026854   /* 1932/2197    */
#define C41     -3.27719617660446062812 /* -7200/2197   */
#define C42     3.32089212562585343650  /* 7296/2197    */
#define C50     2.03240740740740740740  /* 439/216      */
#define C51     -8.0                    /* -8           */
#define C52     7.17348927875243664717  /* 3680/513     */
#define C53     -.20589668615984405458  /* -845/4104    */
#define C6t     0.5                     /* 1/2          */
#define C60     -.29629629629629629629  /* -8/27        */
#define C61     2.0                     /* 2            */
#define C62     -1.38167641325536062378 /* -3544/2565   */
#define C63     .45297270955165692007   /* 1859/4104    */
#define C64     -.27500000000000000000  /* -11/40       */
End of num.h
\n#==========================
#= Here is file ode.h =
#==========================\n
echo ode.h 1>&2
cat >ode.h <<'End of ode.h'
/*
 * definitions and type declarations for ode
 * Copyright Nicholas B. Tufillaro, 1982-1994. All rights reserved.
 */

/*
 * VAX_4BSD sets up floating point error recovery for that
 * system.  It and V7_1170 should not both be defined at once.
 */
/*
#define VAX_4BSD 1
*/

/*
#define V7_1170 1
*/

#ifdef VAX_4BSD
/* our DEC version of 4BSD (Ultrix) doesn't have libjobs.a or sigsys() */
#define sigsys  signal
#endif

#include <stdio.h>
#include <math.h>
#include <setjmp.h>

#define NAMMAX  8       /* size of identifiers */
#define KMAX    7       /* size of k vector */
#define PASTMAX 7       /* size of history vector */

/*
 * symbol table entry
 */
struct sym {
        char    sy_name[NAMMAX];
        double  sy_value;
        double  sy_val[PASTMAX];
        double  sy_prime;
        double  sy_pri[PASTMAX];
        double  sy_predi;       /* predictor value */
        double  sy_sserr;       /* relative single step error */
        double  sy_aberr;       /* absolute single step error */
        double  sy_acerr;       /* accumulated error */
        double  sy_k[KMAX];
        int     sy_flags;
        struct  expr    *sy_expr;
        struct  sym     *sy_link;
};

/*
 * bits in the flags word
 * of a symbol table entry
 */
#define SF_INIT 1       /* set when the initial value is given */
#define SF_ISEQN 2      /* set when an ODE is given */
#define SF_DEPV (SF_INIT|SF_ISEQN)

/*
 * the print queue is made
 * of these
 */
struct prt {
        short   pr_which;
        struct  prt     *pr_link;
        struct  sym     *pr_sym;
};

/*
 * values for pr_which
 * (which cell to print?)
 */
#define P_VALUE 1
#define P_PRIME 2
#define P_SSERR 3
#define P_ACERR 4
#define P_ABERR 5

/*
 * yylex returns a pointer to this.
 * The receiving yacc action must
 * free the space when it's done with it.
 */
struct  lex {
        union {
                double  lxu_value;
                char    lxu_name[NAMMAX];
        } lx_u;
        int     lx_lino;
};

/*
 * an operation in an expression list
 */
struct expr {
        int     ex_oper;
        double  ex_value;
        struct  sym     *ex_sym;
        struct  expr    *ex_next;
};

/*
 * values for ex_oper
 */
#define E_NOOP  0
#define E_PLUS  1
#define E_MINUS 2
#define E_MULT  3
#define E_DIV   4
#define E_POWER 5
#define E_SQRT  6
#define E_EXP   7
#define E_LOG   8
#define E_SIN   9
#define E_COS   10
#define E_ATAN  11
#define E_IDENT 12
#define E_CONST 13
#define E_NEG   14
#define E_ABS   15
#define E_SINH  16
#define E_COSH  17
#define E_TANH  18
#define E_SQAR  19
#define E_CUBE  20
#define E_INV   21

#define NSTACK 30
double  stack[NSTACK];

struct  sym     *symtab, *fsp;
struct  sym     *dqueue;
struct  prt     *pqueue;
struct  expr    exprzero, exprone;
jmp_buf mark;
int     prerr;
double  hmin, hmax, ssmin, ssmax, abmin, abmax, acmax;
int     sawstep, sawprint, sawevery, sawfrom;
int     printnum;
long    it;
int     Aflag, Rflag, Eflag, tflag;
int     pflag, sflag, eflag, rflag, hflag;
int     conflag;
int     prec, fwd;
double  tstart, tstop, tstep, tfrom;
int     tevery;
FILE    *infile;
char    *prog;

struct  sym     *salloc();
struct  sym     *lookup();
struct  expr    *ealloc();
struct  lex     *lalloc();
struct  prt     *palloc();
char    *lineno();
double  eval();
long    time();

/*
 * other misc. consts
 */
# define        OK      ( 0)
# define        YES     ( 1)
# define        NO      ( 0)
# define        TWO     ( 2.0)
# define        HALF    ( 0.5)
# define        SCALE   ( 1e-4)
# define        HMIN    ( 1e-36) /* Minimum step size */
# define        HMAX    ( 0.5)  /* Maximum step size */
# define        LONGMAX ( 2147483647) /* 2^31-1 */
# define        MESH    ( 2)
# define        STOPR   (tstep>0 ? \
                        t-0.0625*tstep>=tstop : t-0.0625*tstep<=tstop )
# define        STOPA   (tstep>0 ? \
                        t+0.9375*tstep>=tstop : t+0.9375*tstep<=tstop )
End of ode.h
\n#==========================
#= Here is file prt.c =
#==========================\n
echo prt.c 1>&2
cat >prt.c <<'End of prt.c'
/*
 * print queue memory management
 * Copyright Nicholas B. Tufillaro, 1982-1994. All rights reserved.
 */

#include "ode.h"

struct prt *
palloc()
{
        register struct prt *pp;

        pp = (struct prt *)malloc(sizeof(struct prt));
        if (pp == NULL)
                fatal("out of core\n");
        pp->pr_sym = NULL;
        pp->pr_link = NULL;
        pp->pr_which = P_VALUE;
        return pp;
}

pfree(pp)
struct prt *pp;
{
        if (pp != NULL) {
                pfree(pp->pr_link);
                free(pp);
        }
}
End of prt.c
\n#==========================
#= Here is file rk.c =
#==========================\n
echo rk.c 1>&2
cat >rk.c <<'End of rk.c'
#include "ode.h"

/*
 * Fourth-Order Runge-Kutta
 * Copyright Nicholas B. Tufillaro, 1982-1994. All rights reserved.
 */
rk()
{
        double t;
        double halfstep = HALF * tstep;
        double onesixth = 1.0 / 6.0;

        for ( it = 0, t = tstart; !STOPR; t = tstart + ( ++it) * tstep) {
                symtab->sy_val[0] = symtab->sy_value = t;
                field();
                for( fsp = dqueue; fsp != NULL; fsp = fsp->sy_link) {
                        fsp->sy_val[0] = fsp->sy_value;
                        fsp->sy_pri[0] = fsp->sy_prime;
                }
                /* output */
                printq();
                for ( fsp = dqueue; fsp != NULL; fsp = fsp->sy_link)   {
                        fsp->sy_k[0] = tstep * fsp->sy_prime;
                        fsp->sy_value = fsp->sy_val[0] + HALF * fsp->sy_k[0];
                }
                symtab->sy_value = t + halfstep;
                field();
                for ( fsp = dqueue; fsp != NULL; fsp = fsp->sy_link)   {
                        fsp->sy_k[1] = tstep * fsp->sy_prime;
                        fsp->sy_value = fsp->sy_val[0] + HALF * fsp->sy_k[1];
                }
                symtab->sy_value = t + halfstep;
                field();
                for ( fsp = dqueue; fsp != NULL; fsp = fsp->sy_link)   {
                        fsp->sy_k[2] = tstep * fsp->sy_prime;
                        fsp->sy_value = fsp->sy_val[0] + fsp->sy_k[2];
                }
                symtab->sy_value = t + tstep;
                field();
                for ( fsp = dqueue; fsp != NULL; fsp = fsp->sy_link)
                        fsp->sy_k[3] = tstep * fsp->sy_prime;
                for ( fsp = dqueue; fsp != NULL; fsp = fsp->sy_link)
                        fsp->sy_value = fsp->sy_val[0] + onesixth *
                                        ( fsp->sy_k[0]
                                        + TWO * fsp->sy_k[1]
                                        + TWO * fsp->sy_k[2]
                                        + fsp->sy_k[3]);
        }
}

End of rk.c
\n#==========================
#= Here is file rka.c =
#==========================\n
echo rka.c 1>&2
cat >rka.c <<'End of rka.c'
#include "ode.h"
#include "num.h"

#define T_LT_TSTOP (tstep>0 ? t<tstop : t>tstop)

/*
 * Fifth-Order Runge-Kutta-Fehlberg with adaptive step size
 * Copyright Nicholas B. Tufillaro, 1982-1994. All rights reserved.
 */
rka()
{
        register gdval, overtime;
        double t;
        double prevstep; /* previous step */

        gdval = YES; 
        prevstep = 0.0;
        overtime = 1;
        for ( it = 0, t = tstart; T_LT_TSTOP || overtime--;) {
                symtab->sy_value = symtab->sy_val[0] = t;
                field();
                for ( fsp = dqueue; fsp != NULL; fsp = fsp->sy_link) {
                        fsp->sy_val[0] = fsp->sy_value;
                        fsp->sy_pri[0] = fsp->sy_prime;
                }
                if ( gdval)
                        printq();       /* output */
                if (tstep * (t+tstep-tstop) > 0)
                        tstep = tstop - t;
                for ( fsp = dqueue; fsp != NULL; fsp = fsp->sy_link) {
                        fsp->sy_k[0] = tstep * fsp->sy_prime;
                        fsp->sy_value = fsp->sy_val[0] + C20 * fsp->sy_k[0];
                }
                symtab->sy_value = t + C2t * tstep;
                field();
                for ( fsp = dqueue; fsp != NULL; fsp = fsp->sy_link) {
                        fsp->sy_k[1] = tstep * fsp->sy_prime;
                        fsp->sy_value = fsp->sy_val[0] + C30 * fsp->sy_k[0]
                            + C31 * fsp->sy_k[1];
                }
                symtab->sy_value = t + C3t * tstep;
                field();
                for ( fsp = dqueue; fsp != NULL; fsp = fsp->sy_link) {
                        fsp->sy_k[2] = tstep * fsp->sy_prime;
                        fsp->sy_value = fsp->sy_val[0] + C40 * fsp->sy_k[0]
                            + C41 * fsp->sy_k[1]
                            + C42 * fsp->sy_k[2];
                }
                symtab->sy_value = t + C4t * tstep;
                field();
                for ( fsp = dqueue; fsp != NULL; fsp = fsp->sy_link) {
                        fsp->sy_k[3] = tstep * fsp->sy_prime;
                        fsp->sy_value = fsp->sy_val[0] + C50 * fsp->sy_k[0]
                            + C51 * fsp->sy_k[1]
                            + C52 * fsp->sy_k[2]
                            + C53 * fsp->sy_k[3];
                }
                symtab->sy_value = t + tstep;
                field();
                for ( fsp = dqueue; fsp != NULL; fsp = fsp->sy_link) {
                        fsp->sy_k[4] = tstep * fsp->sy_prime;
                        fsp->sy_value = fsp->sy_val[0] + C60 * fsp->sy_k[0]
                            + C61 * fsp->sy_k[1]
                            + C62 * fsp->sy_k[2]
                            + C63 * fsp->sy_k[3]
                            + C64 * fsp->sy_k[4];
                }
                symtab->sy_value = t + C6t * tstep;
                field();
                for ( fsp = dqueue; fsp != NULL; fsp = fsp->sy_link)
                        fsp->sy_k[5] = tstep * fsp->sy_prime;
                for ( fsp = dqueue; fsp != NULL; fsp = fsp->sy_link) {
                        fsp->sy_predi = fsp->sy_val[0] + A0 * fsp->sy_k[0]
                            + A2 * fsp->sy_k[2]
                            + A3 * fsp->sy_k[3]
                            + A4 * fsp->sy_k[4];
                        fsp->sy_value = fsp->sy_val[0] + B0 * fsp->sy_k[0]
                            + B2 * fsp->sy_k[2]
                            + B3 * fsp->sy_k[3]
                            + B4 * fsp->sy_k[4]
                            + B5 * fsp->sy_k[5];
                        if ( fsp->sy_value != 0.0)
                                fsp->sy_sserr = fabs( 1.0 - fsp->sy_predi /
                                fsp->sy_value);
                        fsp->sy_aberr = fabs( fsp->sy_value - fsp->sy_predi);
                }
                if ( !conflag && T_LT_TSTOP) {
                        maxerr();
                        if ( hierror()) {
                                tstep *= 0.5;
                                for (fsp=dqueue; fsp!=NULL; fsp=fsp->sy_link)
                                        fsp->sy_value = fsp->sy_val[0];
                                gdval = NO;
                                continue;
                        }
                        else if ( lowerror() && prevstep != tstep) {
                                prevstep = tstep; /* prevent infinite loops */
                                tstep *= 2.0;
                                for (fsp=dqueue; fsp!=NULL; fsp=fsp->sy_link)
                                        fsp->sy_value = fsp->sy_val[0];
                                gdval = NO;
                                continue;
                        }
                }
                gdval = YES;
                prevstep = 0.0;
                ++it;
                t += tstep; /* the round off error is gross */
        }
}
End of rka.c
\n#==========================
#= Here is file stperr.c =
#==========================\n
echo stperr.c 1>&2
cat >stperr.c <<'End of stperr.c'
#include "ode.h"
/* find maximum errors
 * Copyright Nicholas B. Tufillaro, 1982-1994. All rights reserved.
*/

static double   ssemax, abemax, acemax;
static char     *ssenam, *abenam, *acenam;

maxerr()
{
        register struct sym *sp, *dq;

        dq = symtab->sy_link;
        ssemax = abemax = acemax = 0.0;
        for ( sp = dq; sp != NULL; sp = sp->sy_link) {
                if ( ssemax < sp->sy_sserr) {
                        ssemax = sp->sy_sserr;
                        ssenam = sp->sy_name;
                }
                if ( abemax < sp->sy_aberr) {
                        abemax = sp->sy_aberr;
                        abenam = sp->sy_name;
                }
                if ( acmax < sp->sy_acerr) {
                        acemax = sp->sy_acerr;
                        acenam = sp->sy_name;
                }
        }
}

hierror() /* not enough accuracy */
{
        double t = symtab->sy_val[0];

        if ( t + tstep == t) {
                error("step size below lower limit\n");
                longjmp(mark,1);
        }
        if ( ssemax <= ssmax && abemax <= abmax && acemax <= acmax)
                return NO;
        if (fabs(tstep) >= fabs(hmin))
                return YES;
        if (sflag)
                return NO;
        if ( ssemax > ssmax)
                error("rel. error limit exceeded while calculating %.*s'\n",
                        NAMMAX,ssenam);
        if ( abemax > abmax)
                error("abs. error limit exceeded while calculating %.*s'\n",
                        NAMMAX,abenam);
        if ( acemax > acmax)
                error("accum. error limit exceeded while calculating %.*s'\n",
                        NAMMAX,acenam);
        longjmp(mark,1);
}

lowerror() /* more than enough accuracy */
{
        if ( ssemax < ssmin || abemax < abmin)
                if (fabs(tstep) <= fabs(hmax))
                        return YES;
        return( NO);
}

/*
 * interpolate to tstop in Runge-Kutta routines
 */
# define PASTSTOP(stepvar) (t+0.9375*stepvar > tstop && \
                                t+0.0625*stepvar < tstop)
# define BEFORESTOP(stepvar) (t+0.9375*stepvar < tstop && \
                                t+0.0625*stepvar > tstop)
intpr( t)
double t;
{
        if ( tstep > 0)
                if ( !PASTSTOP(tstep))
                        return(NO);
        if ( tstep < 0)
                if ( !BEFORESTOP(tstep))
                        return(NO);
        if ( tstep > 0)
                while ( PASTSTOP(tstep))
                        tstep = 0.5*tstep;
        if ( tstep < 0)
                while ( BEFORESTOP(tstep))
                        tstep = 0.5*tstep;
        return(YES);
}
End of stperr.c
\n#==========================
#= Here is file sym.c =
#==========================\n
echo sym.c 1>&2
cat >sym.c <<'End of sym.c'
/*
 * symbol table space management
 * routines
 * Copyright Nicholas B. Tufillaro, 1982-1994. All rights reserved.
 */

#include "ode.h"

struct sym *
lookup(nam)
char *nam;
{
        register struct sym *sp;

        for (sp=symtab; sp!=NULL; sp=sp->sy_link)
                if (strncmp(sp->sy_name,nam,NAMMAX) == 0)
                        return sp;
        sp = salloc();
        strncpy(sp->sy_name,nam,NAMMAX);
        return sp;
}

struct sym *
salloc()
{
        register struct sym *sp;

        sp = (struct sym *)malloc(sizeof(struct sym));
        if (sp == NULL)
                fatal("out of core\n");
        sp->sy_link = symtab;
        symtab = sp;
        sp->sy_expr = NULL;
        sp->sy_value = sp->sy_prime = 0.0;
        sp->sy_sserr = sp->sy_aberr = sp->sy_acerr = 0.0;
        sp->sy_flags = 0;
        return sp;
}

sfree(sp)
struct sym *sp;
{
        if (sp != NULL)
                free((char *)sp);
}
End of sym.c
\n#==========================
#= Here is file x.bead.example =
#==========================\n
echo x.bead.example 1>&2
cat >x.bead.example <<'End of x.bead.example'
# bead sliding on a smooth circular wire of radius a :
# equation of motion is

g = 10 # acceleration due to gravity
a = 1 # radius of hoop
w = 10 # angular velocity of hoop

the' = vthe
vthe' = (w^2)*sin(the)*cos(the) - (g/a)*sin(the)
the = 0.1
vthe = 0

print t, the
End of x.bead.example
\n#==========================
#= Here is file x.chem.example =
#==========================\n
echo x.chem.example 1>&2
cat >x.chem.example <<'End of x.chem.example'
# chemical kinetics simulation
#
# A + 2B <==> C --> D
#
# rate constants are:
#	kf : A + 2B --> C
#	kb : C --> A + 2B
#	kd : C --> D
#

a' = kb*c - kf*a*b^2
b' = kb*c - kf*a*b^2
c' = kf*a*b^2 - kb*c - kd*c
d' = kd*c

c = 0
d = 0
a = 0.1
b = 1

kf = 1
kb = 1
kd = 1
print t,c every 10
step 0,10
End of x.chem.example
\n#==========================
#= Here is file x.coupled.example =
#==========================\n
echo x.coupled.example 1>&2
cat >x.coupled.example <<'End of x.coupled.example'
# Two similar pendulums coupled with a horizontal spring
# if the motion is undamped then the equations of motion 
# are aproximated by
# m*x1'' = - m*g*x1/l + k(x2 - x1)
# m*x2'' = - m*g*x2/l - k(x2 - x1)

xone' = vxone
xtwo' = vxtwo

vxone' = -g*xone/lone + k/mone * ( xtwo - xone)
vxtwo' = -g*xtwo/ltwo - k/ltwo * ( xtwo - xone)

k = 1
g = 9.8
lone = 5
ltwo = 5
mone = 1
mtwo = 1

xone = 0.0
xtwo = 0.3
vxone = 0
vxtwo = 0

print t, xone
End of x.coupled.example
\n#==========================
#= Here is file x.ddho.example =
#==========================\n
echo x.ddho.example 1>&2
cat >x.ddho.example <<'End of x.ddho.example'
# the dam driven harmonic oscillator with damping
# y'' = -k/m * y - R/m * y' + cos(w*t)
# if R^2 > 4km, overdamped
# if R^2 = 4km, critically damped
# if R^2 < 4km, damped
# this equation reduces to two first order equations, to wit;

y'	= vy
vy'	= -k/m * y - R/m * vy + cos(w*t)
y	= 1
vy	= 0
k	= 1
m	= 1
R	= 0.5
w	= 2*PI

print t, y
End of x.ddho.example
\n#==========================
#= Here is file x.dynamo.example =
#==========================\n
echo x.dynamo.example 1>&2
cat >x.dynamo.example <<'End of x.dynamo.example'
# equations arising in Mike Steele's thesis
# a disk dynamo simulation for the earths magnetics field
# an attempt to simulate the dramtic switch in the polarity
# every eon or so.
# w' = Q -zy - Vw
# y' = S( Az - y)
# z' = wy - z
# S = 5
# V = 1
# Q = 14.625
# A = 1

A = 1
V = 1
Q = 14.625
S = 5

w' = Q - z * y - V * w
y' = S * ( A * z - y)
z' = w * y - z

w = 1
y = 1
z = 1

print t, z
step 0, 10, .01

End of x.dynamo.example
\n#==========================
#= Here is file x.henon.example =
#==========================\n
echo x.henon.example 1>&2
cat >x.henon.example <<'End of x.henon.example'
# A simple center with complicated orbits
# the system investigated by Henon and Heiles
# Poincare plot is a graph of p1(t) vs. q1(t).
# initial condations:
# stable :  p1 = 1/3, q1 = 1/4, p2 = 0.1293144, q2 = 1/5
# unstable :  p1 = 0.1, q1 = 0.1, p2 = 0.467618, q2 = 0.1

qone' = pone
qtwo' = ptwo
pone' = -qone - 2*qone*qtwo
ptwo' = -qtwo - qone*qone + qtwo*qtwo

pone = 1/3
qone = 1/4
ptwo = 0.1293144
qtwo = 1/5

print pone, qone
End of x.henon.example
\n#==========================
#= Here is file x.limcyc.example =
#==========================\n
echo x.limcyc.example 1>&2
cat >x.limcyc.example <<'End of x.limcyc.example'
# nonlinear system with a limit cycle

yone' = yone + ytwo - yone*(yone*yone+ytwo*ytwo)
ytwo' = -yone + ytwo - ytwo*(yone*yone+ytwo*ytwo)

yone = 1
ytwo = 2
print yone, ytwo
End of x.limcyc.example
\n#==========================
#= Here is file x.lorentz.example =
#==========================\n
echo x.lorentz.example 1>&2
cat >x.lorentz.example <<'End of x.lorentz.example'
# Lorentz model , a third order system
# interesting cases are r = 26 , 2.5< t< 30, x = z = 0, y = 1
# and r = 17, 1< t < 50, x = z = 0, y = 1
r = 26
x' = -3*(x-y)
y' = -x*z+r*x-y
z' = x*y-z
x = 0
y = 1
z = 0
print x, y every 9
End of x.lorentz.example
\n#==========================
#= Here is file x.lunar.example =
#==========================\n
echo x.lunar.example 1>&2
cat >x.lunar.example <<'End of x.lunar.example'
: lunar lander shell script using ode to solve equations of motion
echo "*** lunar lander ***"
echo -n "grav (m/sec/sec): "; read grav
echo "velocity starts at zero
empty weight 1000 kg"
vel=0
echo -n "available impulse is 10^4 n-sec/kg
	total fuel (kg): "
if read fuel; then :; else exit 0; fi
echo -n "	starting height (m): "
if read height; then :; else exit 0; fi
(echo "y'=v;v'=thrust*1000/m-$grav;m'=-thrust/10"
dur=0
while : ; do
	sleep 2
	sleep $dur
	echo -n "	thrust (kn): " 1>&2
	if read thrust; then :; else break; fi
	echo -n "	duration (sec): " 1>&2
	if read dur; then :; else break; fi
	echo "y=$height;v=$vel;thrust=$thrust;m=$fuel+1000
		print y,v,m from $dur;step 0,$dur"
done) | ode -r 1e-5 | while read height vel mass; do
	echo $mass
	echo "velocity=${vel}m/sec; mass=${mass}kg"
	echo $mass
	fuel=`echo "5k$mass 1000-pq"|dc`
	case $fuel in
	-*|0)	echo out of fuel; exit 1 ;;
	*)	echo fuel=${fuel}kg ;;
	esac
	case $height in
	-*|0)	case $vel in
			-[0-9][0-9]*) echo too fast; exit 1 ;;
			*) echo good landing; exit 0 ;;
		esac ;;
	*)	echo height=${height}m ;;
	esac
done
End of x.lunar.example
\n#==========================
#= Here is file x.orbit.example =
#==========================\n
echo x.orbit.example 1>&2
cat >x.orbit.example <<'End of x.orbit.example'
# planetary orbit simulation
# two suns situated at (0,0) and (-5,0)
# one planet starting out at (1,0)
# x and y are positions
# vx and vy are velocities

vx' = -x/((x^2+y^2)^(3/2)) -(x+5)/(((x+5)^2+y^2)^(3/2))
vy' = -y/((x^2+y^2)^(3/2)) -y/(((x+5)^2+y^2)^(3/2))
y' = vy
x' = vx

x = 1
y = 0

print x,y every 10

# these values seem to give a nice orbit:
# vx = 0
# vy = 1.142

# a more exciting result can be obtained from:
vx = 0
vy = 1.165

End of x.orbit.example
\n#==========================
#= Here is file x.pen.example =
#==========================\n
echo x.pen.example 1>&2
cat >x.pen.example <<'End of x.pen.example'
# coupled oscillator : two pendula simulation
m = 1
M = 1.0625

a = 0.5
adot = 0
l = 10
ldot = 0
ldot' = ( m * l * adot * adot - M * 9.8 + m * 9.8 * cos( a) ) / ( m + M)
l'    = ldot
adot' = ( -1/ l) * ( 9.8 * sin( a) +  2 * adot * ldot)
a'    = adot

print l, ldot every 10000
step 0, 1000000000
End of x.pen.example
\n#==========================
#= Here is file x.pop.example =
#==========================\n
echo x.pop.example 1>&2
cat >x.pop.example <<'End of x.pop.example'
# An example from population biology
# Predator-prey equations of Volterra and Lotka
# x' = ( A - By) x
# y' = ( Cx - D) y
# A,B,C,D > 0

x' = ( A - B*y) * x
y' = ( C*x - D) * y

A = 1
B = 1
C = 1
D = 1

x = 3
y = 1

print t, x
End of x.pop.example
\n#==========================
#= Here is file x.qcd.example =
#==========================\n
echo x.qcd.example 1>&2
cat >x.qcd.example <<'End of x.qcd.example'
# Quantun Chromo-Dynamics
# cf. Griffiths et al., J. Math. Phy.

f' = fp
fp' = -f*g*g
g' = gp
gp' = g*f*f

f = 0
fp = -1
g = 1
gp = -1

print t, f
step 0, 5
End of x.qcd.example
\n#==========================
#= Here is file x.rumor.example =
#==========================\n
echo x.rumor.example 1>&2
cat >x.rumor.example <<'End of x.rumor.example'
# A rumor spreads through a closed population of constant size
# N+1. At time t the total population can be classified into three categories:
#	x persons who are ignorant of the rumor;
#	y persons who are actively spreading the rumor;
#	z persons who have heard the rumor but have stopped spreading it;
#	if two persons who are spreading the rumor meet then they stop
#	spreading it.
# The contact rate between any two categories is constant, u.
#	Show that the equations
#		x' = -u * x * y,
#		y' = u * ( x*y - y*(y - 1) - y*z)
# give a deterministic model of the problem.  Find the equation of the
# phase paths and sketch the phase diagram.
#	Show that, when initially y = 1 and x = N, the number of people
# who ultimately never hear the rumor is s, where
#	2N + 1 -2s + N log( s/N) = 0

x' = -u * x * y
y' = u * ( x*y - y*(y-1) - y*( 100 + 1 - y - x))

y = 1
x = 100

u = 1

print t, x
End of x.rumor.example
\n#==========================
#= Here is file x.solit1.example =
#==========================\n
echo x.solit1.example 1>&2
cat >x.solit1.example <<'End of x.solit1.example'
# equation arising in john selker's soliton thesis
# x'' = 2 / t * x' + B * x
# B = -1
# this reduces to two first order equations, to wit:
# x' = vx
# vx' = 2 / t * vx + B * x

B = -1
x' = vx
vx' = 2 / t * vx + B * x

x = 0
vx = 1

print t, x
step 5, 15
End of x.solit1.example
\n#==========================
#= Here is file x.solit2.example =
#==========================\n
echo x.solit2.example 1>&2
cat >x.solit2.example <<'End of x.solit2.example'
# solitons
y' = z
z' = -2*y*(1+ln(y^2+.0000001))
print t,y
z=0
End of x.solit2.example
\n#==========================
#= Here is file x.viscous.example =
#==========================\n
echo x.viscous.example 1>&2
cat >x.viscous.example <<'End of x.viscous.example'
# bead sliding on a smooth circular wire of radius a :
# there is viscious damping so the bead should home in 
# on the equilbrium point.
# equation of motion is

g = 10 # acceleration due to gravity
a = 1  # radius of hoop
w = 10 # angular velocity of hoop
b = 1  # damping coefficient

the' = vthe
vthe' = (w^2)*sin(the)*cos(the) - (g/a)*sin(the) - b * vthe
the = 0.1
vthe = 0

print t, the 

step 0, 20
End of x.viscous.example
